paymentPage.extend( function() {

    var paymentMethodChosen = false;

    var additionalCards = [
        {
            type: 'maestro',
            patterns: [50, 5018, 502, 503, 506, 56, 58, 639, 6220, 67],
            format: /(\d{1,4})/g,
            length: [12, 13, 14, 15, 16, 17, 18, 19],
            cvcLength: [3],
            luhn: true
        },
        {
            type: 'boncard',
            patterns: [604993,629875,629912,629920,637594,637595,637685,637736,637804],
            format: /(\d{1,4})(\d{1,4})(\d{1,4})(\d{1,4})(\d{1,3})?/,
            length: [16,19],
            cvcLength: [3,4],
            luhn: false
        },
        {
            type: 'lunchcheck',
            patterns: [604993,629875,629912,629920,637594,637595,637685,637736,637804],
            format: /(\d{1,4})(\d{1,4})(\d{1,4})(\d{1,4})(\d{1,3})?/,
            length: [16,19],
            cvcLength: [3,4],
            luhn: false
        },
        {
            type:"amex",
            patterns:[34,37],
            format:/(\d{1,4})(\d{1,6})?(\d{1,5})?/,
            length:[15],
            cvcLength:[4],
            luhn:true
        },
        {
            type: 'cosy-mastercard',
            patterns: [51365910],
            format: /(\d{1,4})/g,
            length: [16],
            cvcLength: [3],
            luhn: true
        },
        {
            type: 'uatp',
            patterns: [1],
            format: /(\d{1,4})(\d{1,5})?(\d{1,6})?/,
            length: [15],
            cvcLength: [0],
            luhn: true
        },
        {
            type: 'myone',
            patterns: [60045202],
            format: /(\d{1,9})(\d{1,7})?(\d{1,3})?/,
            length: [19],
            cvcLength: [3],
            luhn: true
        },
        {
            type: 'paycard',
            patterns: [975616300],
            format: /(\d{1,19})/,
            length: [19],
            cvcLength: [0],
            luhn: false
        },
        {
            type: 'unionpay',
            patterns: [''],
            format: /(\d{1,19})/,
            length: [14, 15, 16, 17, 18, 19],
            cvcLength: [3],
            luhn: true
        },
        {
            type: 'mfg-giftcard',
            patterns: [''],
            format: /(\d{1,19})/,
            length: [19],
            cvcLength: [0],
            luhn: false
        },
        {
            type: 'cartebleue',
            patterns: [4,5,6],
            format: /(\d{1,4})/g,
            length: [16,17,18,19],
            cvcLength: [3],
            luhn: true
        },
        {
            type: 'mediamarkt',
            patterns: [60045205,60045210],
            format: /(\d{1,4})/g,
            length: [19],
            cvcLength: [3],
            luhn: true
        },
        {
            type: 'butterfly',
            patterns: [60045203],
            format: /(\d{1,4})/g,
            length: [19],
            cvcLength: [3],
            luhn: true
        },
        {
            type: 'elocard',
            patterns: [''],
            format: /(\d{1,4})/g,
            length: [16],
            cvcLength: [3],
            luhn: true
        },
        {
            type: 'supercard',
            patterns: [4,5],
            format: /(\d{1,4})/g,
            length: [16],
            cvcLength: [3],
            luhn: true
        },
        {
            type: 'vpay',
            patterns: [4],
            format: /(\d{1,4})/g,
            length: [16],
            cvcLength: [3],
            luhn: true
        }
    ];

    // update overloaded types
    for ( var i = $.payment.cards.length; i--;) {
        var stripeCard = $.payment.cards[i];

        for (var j = 0; j < additionalCards.length; j++ ) {
            if (stripeCard.type === additionalCards[j].type) {
                $.payment.cards[i] = additionalCards[j];
                additionalCards.splice(j,1);
                break;
            }
        }
    }

    $.merge($.payment.cards, additionalCards);

    var paymentCards = $.payment.cards.slice(); // back up cards

    function formatCardnumberWithBullets( cardno )
    {
        var fcn = $.payment.formatCardNumber( cardno.toString().replace(/x/g, "0") );
        var result = "";
        var cPos = 0;
        for ( var i = 0; i < fcn.length; i++ )
        {
            if ( fcn[i] == ' ') {
                result += ' ';
                continue;
            }

            if ( (cardno[cPos] == 'x' || cardno[cPos] == 'X') && fcn[i] == '0' ) {
                result += String.fromCharCode(8226); //bullet
            }
            else
                result += fcn[i];

            cPos++
        }

        return result;
    }

    var cardnoAutocompletedValue;
    var cardSwitchAllowed;
    var cardValidationEnabled;
    var cardType;

    function prepareView( view )
    {
        $( "#cardExpirationYear" ).on('change',function() {
            $( "#expiry" ).val( $("#cardExpirationMonth").val() + " / " + $( "#cardExpirationYear").val()).trigger("paste");
        });

        var $cardno = $("#cardNumber");
        cardnoAutocompletedValue = $cardno.data("autocompletedValue");
        cardSwitchAllowed = $cardno.data("cardSwitchAllowed");
        cardValidationEnabled = $cardno.data("cardValidationEnabled");
        cardType = $cardno.data("cardType");
        paymentPage.mandatoryCVV = $("#cvv").data("mandatoryCvv");

        if ( view == "list" )
            delete paymentPage.pmethod;
        else
            paymentPage.pmethod = $("#pmethod").val();

        if ( paymentPage.iframed )
        {
            // back button is visible if the current view is 'form' but the flow was not started from a form
            var showBackButton = view === "form" && paymentPage.view !== "form";

            E.emit('app-toggle-actions', {'back': showBackButton } );
        }

        if ( $cardno.length == 1 && cardnoAutocompletedValue != null ) {

            setTimeout( function() {

                $cardno
                    .removeAttr("cardno")
                    .val(formatCardnumberWithBullets(cardnoAutocompletedValue))
                    .prop("readonly", true)
                    .prop("disabled", true)
                    .trigger("app-input-prefill");

                $("<input/>")
                    .attr({
                        "name": "cardno",
                        "type": "hidden",
                        "value": cardnoAutocompletedValue
                    })
                    .insertAfter($cardno);
            }, 100);
        }

        if ( view == "form" ) {

            if ( paymentPage.pmethod != null && $("#cvv").length ) {
                $.each(paymentPage.stripeMap, function (stripeType, value) {
                    if (value == paymentPage.pmethod) {
                        $.each($.payment.cards, function (index, card) {
                            if (card.type == stripeType) {
                                if ( card.cvcLength != null && card.cvcLength.length > 0 ) {
                                    var placeholder = "12345".substring(0, card.cvcLength );
                                    $("#cvv")[0].placeholder = placeholder;
                                }
                                return false;
                            }
                        });
                        return false;
                    }
                });
            }

            if ( $("#cvv")[0] && $("#cvv")[0].placeholder === "" ) {
                $("#cvv")[0].placeholder = "123";
            }

            $.payment.cards.splice(0, $.payment.cards.length); // empty it
            $.merge($.payment.cards, paymentCards); // add all again

            $.each(paymentPage.stripeMap, function (key, value) {

                if (value === paymentPage.pmethod) {
                    $.each($.payment.cards, function (index, value2) {

                        if (value2.type === key) {
                            // moving the selected paymentmethod to position 0 of the cards array
                            paymentPage.log('hit! moving ' + value2.type + ' to postion 0 of the cards array');
                            $.payment.cards.splice(index, 1);
                            $.payment.cards.unshift(value2);
                        }

                    });
                }
            });

            $("[data-after-init]").each( function(a,b) {
                var afterInit = $(this).data("afterInit");
                paymentPage.log("calling " + afterInit);
                paymentPage[afterInit]();

            });

            if(paymentPage.optionalPaymentCharge){
                $("#cardNumber").focusout(function (){
                    paymentPage.updateOptionalPaymentCharge();
                })
            }

            if(paymentPage.visaCtpConfig){
                if($("#cardsList").length === 1){
                    $('input:checkbox').click(function() {
                        $('input:checkbox').not(this).prop('checked', false);
                    });

                    $('input:checkbox:first').prop('checked',true);
                }
            }

            paymentPage.setupValidation();
        } else if (view == "dccmite") {
            handleDCCScreen();
        } else if (view == "accarda") {
            handlePaymentOptionsScreen( null, "accarda" );
        } else if (view == "byjuno") {
            handlePaymentOptionsScreen( null, "byjuno" );
        }
        //to be compliant with WAI-ARIA so that screenreaders dont read all the options --> DT-9549
        $("#verifyingLabel").hide();
        $("#verifiedLabel").hide();
    }

    var E = window.datatrans.emitter;
    var walletCheckEvents = 0;

    E.on('applepay-check-done',function(){
        walletCheckEvents++;
        if(walletCheckEvents == paymentPage.getNumberOfOnScreenWallets()){
            E.emit("wallets-ready");
        }
    });
    E.on('googlepay-check-done',function(){
        walletCheckEvents++;
        if(walletCheckEvents == paymentPage.getNumberOfOnScreenWallets()){
            E.emit("wallets-ready");
        }
    });

    E.on('dt-ready', function() {
        paymentPage.log("dt-ready Event received!");

        if(parent != null) {
            parent.postMessage("frameReady", "*" );
        }

        if(paymentPage.hasOnScreenWallet()){
            E.on("wallets-ready",function () {
                if (paymentPage.applePayAvailable) {
                    if (paymentPage.applePayConfig.walletMode) {
                        paymentPage.renderApplePayButton();
                        $(".onscreen-wallets").show();
                    } else {
                        $(".precheckable-APL").show();
                    }
                } else if (paymentPage.googlePayAvailable && !paymentPage.applePayAvailable) {
                    if (paymentPage.paymentMethodPresent && paymentPage.googlePayConfig.walletMode) {
                        paymentPage.renderGooglePayButton();
                        $(".onscreen-wallets").show();
                    } else {
                        $(".precheckable-PAY").show();
                    }
                }

                if(paymentPage.walletDirectCall && !paymentPage.applePayAvailable && !paymentPage.googlePayAvailable){
                    paymentPage.returnError("-999", "payment method not available", "No wallet is available");
                }

                E.emit('app-adapt-height');
                prepareView( paymentPage.view );
                E.emit('app-reveal');
            })
        }else{
            if(paymentPage.paymentMethod){
                var paymentMethod = paymentPage.paymentMethod;
                if(("PAY" == paymentMethod && !paymentPage.googlePayAvailable) || ("APL" == paymentMethod && !paymentPage.applePayAvailable) ){
                    paymentPage.returnError("-999", "payment method not available", "No wallet is available");
                }

                var timeOut = false;

                if ("VIS" === paymentMethod && paymentPage.visaCtpConfig) {
                    let timer = setTimeout(function () {
                        paymentPage.log("Failed to initialize Visa CTP within timeout period of " + paymentPage.visaCtpConfig.visaCtpInitTimeOut);
                        prepareAndRenderPaymentPage();
                        timeOut = true;
                        return;

                    },  paymentPage.visaCtpConfig.visaCtpInitTimeOut);

                    paymentPage.initVisaCtP();
                    E.on("visa-ctp-initialized", function (data) {
                        if(!timeOut){
                            clearTimeout(timer)
                            if (data.recognized === true) {
                                paymentPage.setSrcProfile()
                                E.on("visa-ctp-profile-loaded",function (){
                                    paymentPage.renderVisaCtPCardList('directCall');
                                    E.on("visa-ctp-card-list-ready", function () {
                                        prepareAndRenderPaymentPage();
                                        return;
                                    });
                                })
                            } else {
                                prepareAndRenderPaymentPage();
                                return;
                            }
                        }
                    })
                    return;
                }
            }

            prepareView(paymentPage.view);
            E.emit('app-reveal');
        }

        if (paymentPage.view === "waiting") {
            E.emit('app-toggle-actions', {'back': false, 'close': false});

            $.ajax(paymentPage.contextPath + "/jsp/upPayTrx_submit.jsp",
                {
                    cache: false,
                    data: {"datatransTrxId": paymentPage.transactionId}
                }
            ).done(function (data) {
                    var returnUrl = $.trim(data);

                    if ( returnUrl == "STOP_ALERT" ) {
                        self.location.href = paymentPage.contextPath + "/jsp/stop.html";
                        return;
                    }

                    if (returnUrl != "3D_submit_1u.jsp" && returnUrl != "3D2_start.jsp" && returnUrl != "upReturn.jsp")
                        returnUrl = "upStoreTrx_result.jsp";

                    self.location.href = paymentPage.contextPath + "/jsp/" + returnUrl + "?" + paymentPage.getDatatransTrxId();
                })
                .fail(function (data) {
                    self.location.href = paymentPage.contextPath + "/jsp/" + "upPayTrx_submit.jsp?noAjaxSupp=1&" + paymentPage.getDatatransTrxId();
                });
        }
        else
        if ( paymentPage.view === "waiting3d" )
        {
            E.emit('app-toggle-actions', {'back': false, 'close': false});

            setTimeout( function() {
                $.ajax(paymentPage.contextPath + "/jsp/upPayment.jsp",
                    {
                        cache: false,
                        data: {"datatransTrxId": paymentPage.transactionId}
                    }
                ).done(function (data) {
                        var returnUrl = $.trim(data);

                        if ( returnUrl == "STOP_ALERT" ) {
                        	self.location.href = paymentPage.contextPath + "/jsp/stop.html";
                        	return;
						}
                        if ( returnUrl != "upReturn.jsp" )
                            returnUrl = "upStoreTrx_result.jsp";

                        self.location.href = paymentPage.contextPath + "/jsp/" + returnUrl + "?" + paymentPage.getDatatransTrxId();
                    })
                    .fail(function (data) {
                        // ???
                    });
            }, 1000 );
        }
        else
        if (paymentPage.view === "waiting3dVer2") {
            E.emit('app-toggle-actions', {'back': false, 'close': false});

            $.ajax(paymentPage.contextPath + "/jsp/3D2_DS_request.jsp",
                {
                    cache: false,
                    data: {"datatransTrxId": paymentPage.transactionId}
                }
            ).done(function (data) {
                    var returnUrl = $.trim(data);

                    if ( returnUrl == "STOP_ALERT" ) {
                        self.location.href = paymentPage.contextPath + "/jsp/stop.html";
                        return;
                    }

                    if (returnUrl != "3D2_submit.jsp" && returnUrl != "upReturn.jsp" &&
                    		returnUrl != "3D2_start.jsp" && returnUrl != "3D_submit_1u.jsp")
                        returnUrl = "upStoreTrx_result.jsp";

                    self.location.href = paymentPage.contextPath + "/jsp/" + returnUrl + "?" + paymentPage.getDatatransTrxId();
                })
		.fail(function (data) {
		// ???
		});
        }
        else
        {
            setTimeout(function () {
                    E.emit('app-animate');
                    waitForPrecheckTasks( function() {

                        setTimeout(function () {

                            E.emit('app-state', {
                                group: 'mantle',
                                uid: 'main'
                            });
                        }, 100);

                        showInfoTexts();
                        paymentPage.startAutoCancelTimeout();
                        paymentPage.focusFirstField(500);
                    });
            },
                window.datatrans.config.mode == "overlay" ? 800 : 400);
        }

        E.emit("dt-initialized");
    });


    function waitForPrecheckTasks( callback )
    {
        if ( paymentPage.precheck == null || paymentPage.precheck["itemCount"] === 0 )
        {
            paymentPage.log( "No prechecks needed. Continuing... ");
            return callback();
        }

        $.ajax( {
            method: "GET",
            dataType: "json",
            cache: false,
            url: paymentPage.contextPath + "/payment/precheck",
            data: { "datatransTrxId": paymentPage.transactionId },
            timeout: 1000 * paymentPage.precheck["ajaxTimeout"]
        }).done( function( result ) {
            var possibleHeightChange = false;
            for (var methodCode in result ) {
                if (result.hasOwnProperty(methodCode)) {
                    if ( result[methodCode] == "SUCCESS" ) {
                        $(".precheckable-" + methodCode).show();
                        possibleHeightChange = true;
                        paymentPage.log( "Prechecked show " + methodCode );
                    }
                }
            }
            if ( possibleHeightChange )
                E.emit('app-adapt-height');
        }).fail( function() {
            // do nothing, expect always callback
        }).always( function() {
            callback(); // always signal the completion
        });
    }

    this.focusFirstField = function (timeout) {
        if(paymentPage.autoFocusFirstInputField) {
            setTimeout(function(){
                $('#paymentForm')
                    .find('input:blank:not(:checkbox,:button)')
                    .filter(":visible:enabled")
                    .first()
                    .trigger("focus");

            },timeout);
        }
    }

    function bind( emitterEvent, callback )
    {
        E.on( emitterEvent, function( data ) {
            callback.apply( paymentPage, [data] );
        } )
    }

    this.methodChoosenDispatcher = function (data) {
        if(data.method == "PAY"){
            paymentPage.startGooglePayRequest();
        }else if(data.method === "APL"){
            paymentPage.startApplePayRequest();
        }else if(data.method === "PAP" && paymentPage.usesPayPalRestApi){
            return;
        }
        else{
            this.methodChosen(data);

        }
    };

    // On payment method chosen
    this.methodChosen = function( data ) {

        if(paymentMethodChosen)
            return;

        paymentMethodChosen = true;

        E.emit('app-state', {
            group: 'mantle',
            uid: 'loader'
        });

        var url = this.contextPath + "/payment?pmethod=" + data.method;

        var parameters = {
            "datatransTrxId": this.transactionId,
            "pmethod": data.method,
            "paymentMethodChoosen": true,
        };

        if(paymentPage.visaCtpConfig){
            let ctPStatus = this.getVisaCtPStatus();
            if(ctPStatus.visaCtPInitialized){
                parameters.visaCtPInitialized = true;
                parameters.visaCtPisRecognized = ctPStatus.visaCtPisRecognized;
            }
            if(ctPStatus.visaCtPisRecognized === true){
                parameters.visaCtPProfile = JSON.stringify(ctPStatus.visaCtPprofileData);
                parameters.visaCtPCardsList = true;
            }
        }



        if(data.paymentData){  // google pay data
            parameters.paymentData = JSON.stringify(data.paymentData);
        }

        if(paymentPage.visaCtpConfig){
            let ctPStatus = this.getVisaCtPStatus();

            if(ctPStatus.visaCtPFullyLoaded === true){
                renderMethodSelection(url,parameters);
            }

            if(!ctPStatus.visaCtPInitialized){
                let timedOut = false;
                var timer = setTimeout(function (){
                    timedOut = true;
                    paymentPage.log("Visa CtP -Failed to initialize Visa CtP within timeout period")
                    paymentPage.visaCtPInitialized = false;
                    parameters.visaCtPInitialized = false;
                    paymentPage.visaCtPCardsList = false;
                    parameters.visaCtPCardsList = false;
                    renderMethodSelection(url,parameters)
                },paymentPage.visaCtpConfig.visaCtpInitTimeOut);

                E.on("visa-ctp-initialized", function () {
                    if(!timedOut){
                        clearTimeout(timer)
                        paymentPage.visaCtPInitialized = true;
                        let ctPStatus = paymentPage.getVisaCtPStatus();
                        if(ctPStatus.visaCtPInitialized){
                            parameters.visaCtPInitialized = true;
                            parameters.visaCtPisRecognized = ctPStatus.visaCtPisRecognized;
                        }
                        if(ctPStatus.visaCtPisRecognized === true){
                            parameters.visaCtPProfile = JSON.stringify(ctPStatus.visaCtPprofileData);
                            parameters.visaCtPCardsList = true;
                        }
                        renderMethodSelection(url,parameters)
                    }
                });
            }else if(!ctPStatus.visaCtPisRecognized){
                let timedOut = false;
                var recognizedEventTimer = setTimeout(function (){
                    timedOut = true;
                    paymentPage.log("Visa CtP -Failed to load recgonize data within timeout period")
                    paymentPage.visaCtPInitialized = false;
                    parameters.visaCtPInitialized = false;
                    paymentPage.visaCtPCardsList = false;
                    parameters.visaCtPCardsList = false;
                    renderMethodSelection(url,parameters)
                },paymentPage.visaCtpConfig.visaCtpInitTimeOut);

                E.on("visa-ctp-recognized", function () {
                    if(!timedOut){
                        clearTimeout(recognizedEventTimer)
                        var timer = setTimeout(function (){
                            timedOut = true;
                            paymentPage.log("Visa CtP -Failed to load profile data within timeout period")
                            paymentPage.visaCtPInitialized = false;
                            parameters.visaCtPInitialized = false;
                            paymentPage.visaCtPCardsList = false;
                            parameters.visaCtPCardsList = false;
                            renderMethodSelection(url,parameters)
                        },paymentPage.visaCtpConfig.visaCtpInitTimeOut);

                        E.on("visa-ctp-profile-loaded", function () {
                            if(!timedOut){
                                clearTimeout(timer)
                                paymentPage.visaCtPInitialized = true;
                                let ctPStatus = paymentPage.getVisaCtPStatus();
                                if(ctPStatus.visaCtPInitialized){
                                    parameters.visaCtPInitialized = true;
                                    parameters.visaCtPisRecognized = ctPStatus.visaCtPisRecognized;
                                }
                                if(ctPStatus.visaCtPisRecognized === true){
                                    parameters.visaCtPProfile = JSON.stringify(ctPStatus.visaCtPprofileData);
                                    parameters.visaCtPCardsList = true;
                                }
                                renderMethodSelection(url,parameters)
                            }
                        });
                    }
                });
            }
            else{
                if(!ctPStatus.visaCtPprofileData && ctPStatus.visaCtPisRecognized){
                    let timedOut = false;
                    var timer = setTimeout(function (){
                        timedOut = true;
                        paymentPage.log("Visa CtP -Failed to load profile data within timeout period")
                        paymentPage.visaCtPInitialized = false;
                        parameters.visaCtPInitialized = false;
                        paymentPage.visaCtPCardsList = false;
                        parameters.visaCtPCardsList = false;
                        renderMethodSelection(url,parameters)
                    },paymentPage.visaCtpConfig.visaCtpInitTimeOut);

                    E.on("visa-ctp-profile-loaded", function () {
                        if(!timedOut){
                            clearTimeout(timer)
                            paymentPage.visaCtPInitialized = true;
                            let ctPStatus = paymentPage.getVisaCtPStatus();
                            if(ctPStatus.visaCtPInitialized){
                                parameters.visaCtPInitialized = true;
                                parameters.visaCtPisRecognized = ctPStatus.visaCtPisRecognized;
                            }
                            if(ctPStatus.visaCtPisRecognized === true){
                                parameters.visaCtPProfile = JSON.stringify(ctPStatus.visaCtPprofileData);
                                parameters.visaCtPCardsList = true;
                            }
                            renderMethodSelection(url,parameters)
                        }
                    });
                }else{
                    renderMethodSelection(url,parameters)
                }
            }
        }else{
            renderMethodSelection(url,parameters)
        }
    };

    var firstSubmitTried = false;

    E.on('dt-input-exit', function(data) {
        if ( firstSubmitTried )
            $("#paymentForm").valid();
    } );

    function showInfoTexts() {
        $(".form--input").each(function () {
            var info = $(this).data("info");
            if (info != null && info.length > 0)
                E.emit('app-info', {message: info, uid: $(this).attr("id")});
        });
    }

    this.doCancel = function()
    {
        E.emit('app-toggle-actions', {'back': false, 'close': false});

        if ( this.iframed )
        {
            setTimeout( function() {
                parent.postMessage("cancel", "*" )
            }, 1000 );
        }
        else
            location.href = this.contextPath + "/jsp/upCancel.jsp?" + this.getDatatransTrxId();
    };

    this.doBack = function( data )
    {
        paymentMethodChosen = false;

        if ( this.storeTrx != null )
        {
            this.storeTrx.abort();
            delete this.storeTrx;
        }

        if(this.view == 'visaCtP'){
            this.handleVisaCtpBack();
            this.startAutoCancelTimeout();
            return;
        }

        if( this.view == 'list' && data.order === 1 || this.view == 'form' && data.order === 2 || this.dcc)
            this.doCancel();
        else
        {
            prepareView("list");
            this.startAutoCancelTimeout();
        }
    };

    this.cardTypeChanged = function( data )
    {
        $("input[name=pmethod]").val( paymentPage.stripeMap[ data.cardtype ] );
    };

    this.defaultFlow = { storeURL : "/jsp/upStoreTrx.jsp", responseHandler: storeTrxResponseHandler };
    this.accardaFlow = { storeURL: "/jsp/upAccardaStart.jsp", responseHandler: handlePaymentOptionsConfirm };
    this.byjunoFlow = { storeURL: "/jsp/upIntrumStart", responseHandler: handlePaymentOptionsConfirm };

    this.doSubmit = function( data )
    {

        if(!data.visaCtPCheckoutResponse && $("form[name='visaCtpForm']").is(":visible")){
            if($("#cardsList").is(":visible")){
                this.handleVisaCtPCardSelection(data);
                return;
            }else{
                this.handleVisaCtPSignIn(data);
                return;
            }

        }

        if($("#visaCtpEnroll").is(":checked") && !data.failedCtpEncrytion && !data.visaCtPCancelledRegistration && !data.visaCtPSuccessfulRegistration){
            this.handleVisaCtpRegistration(data);
            return;
        }

        var paymentForm = $("#paymentForm");

        if ( !paymentForm.valid() ) {
            $(".message--content").attr("role","alert");
            $(".message__error").first().siblings(":input").focus();
            E.emit('app-submit-response', {success: false, uid: data.uid});
            firstSubmitTried = true;
            return;
        }

        if ( this.previewOnly ) {
            alert( paymentPage.getTranslation("paym_preview_end") );
            return;
        }

        E.emit('app-toggle-actions', {'back': false, 'close': false});

        var $cardno = $("#cardNumber");
        if(cardSwitchAllowed && cardnoAutocompletedValue == null && $cardno.length == 1 ) {
            var cardno = $cardno.val();
            paymentPage.cardTypeChanged(  { "cardtype": $.payment.cardType(cardno) });
        }

        var flow = this.defaultFlow;
        if ( paymentPage.accarda ) {
            flow = this.accardaFlow;
        } else if ( paymentPage.byjuno ) {
            flow = this.byjunoFlow;
        }

        if(data.visaCtPCheckoutResponse){
            var input = document.createElement("input");
            input.type = "hidden";
            input.name = "visaCtPCheckoutResponse";
            input.value = JSON.stringify(data.visaCtPCheckoutResponse);
            paymentForm.append(input);
        }

        this.storeTrx = $.ajax({
            method: "POST",
            headers: {"x-no-redirect": "true"},
            url: this.contextPath + flow.storeURL,
            data: paymentForm.serialize()
        });



        var minimumTimeout = $.Deferred();

        setTimeout(function () {
            minimumTimeout.resolve();
        }, 1000);

        //to be compliant with WAI-ARIA so that screenreaders dont read all the options --> DT-9549
        $("#verifyingLabel").hide();
        $("#verifiedLabel").hide();
        $("#payLabel").hide();
        $("#verifyingLabel").show();

        paymentForm.submit(false); // disable form submit
        $.when(this.storeTrx.promise(), minimumTimeout).then( flow.responseHandler );
    };

    function storeTrxResponseHandler( response ) {
        var ajaxData = response.length ? response[0] : response;
        var targetFrame = paymentPage.startOnTop ? top : self;

        if ( typeof ajaxData === 'object' )
        {
            if ( ajaxData.redirect != null )
            {
                $(".form--state__verifying .form--state-inner .label").text( paymentPage.getTranslation("paym_redirecting") );
                targetFrame.location.href = ajaxData.redirect;
            }
            else
            if ( ajaxData.result === "waiting" )
            {
                var submitTrx = $.ajax( paymentPage.contextPath + "/jsp/upPayTrx_submit.jsp", {
                    cache: false,
                    data: { "datatransTrxId": paymentPage.transactionId }
                }).promise();

                submitTrx.done( function( data ) {
                    var returnUrl = $.trim(data);

					if ( returnUrl == "STOP_ALERT" ) {
						self.location.href = paymentPage.contextPath + "/jsp/stop.html";
						return;
					}

                    if ( returnUrl != "3D_submit_1u.jsp" && returnUrl != "3D2_start.jsp" && returnUrl != "upReturn.jsp" )
                        returnUrl = "upStoreTrx_result.jsp";

                    targetFrame.location.href = paymentPage.contextPath + "/jsp/" + returnUrl + "?" + paymentPage.getDatatransTrxId();
                });
            }
        }
        else
        {
            var formHolder = $("<div/>").css("display","none").append( ajaxData );

            /*
                Check if JS on the returned page has to be executed.
                If yes, the corresponding payment method has to include the following:
                <div id="runscript" data-function="functionToRun();"></div>

                Current paymentmethods using this:
                - MPW (upMpwStart.jsp)
             */
            var runScript = formHolder.find('#runscript').data('function');

            if (typeof runScript !== "undefined") {

                if( runScript.length ) {
                    //local first, remote afterwards
                    formHolder.find("script").each(function() {
                        var scriptSource = $(this).attr('src');
                        if (typeof scriptSource === "undefined" && $(this).attr('id')!=='dt--runscript') {
                            paymentPage.log("evaluating local js=>" + $(this).text());
                            window.eval($(this).text());
                        }
                    });

                    var requests = [];
                    formHolder.find("script").each(function() {
                        var scriptSource = $(this).attr('src');
                        if (typeof scriptSource !== "undefined") {
                            paymentPage.log("loading script source " + scriptSource);
                            requests.push($.getScript( scriptSource));
                        }

                    });

                    $.when.apply(undefined, requests).done(function(){
                        paymentPage.log("loading remote js scripts finished. running function specified in 'data-function'!");
                        eval(runScript);
                    });

                    return;
                }
            }

            var form = formHolder.find("form").eq(0);

            // GET params - in the uppWebResponseMethod case its not a <form> but an <a> tag
            if(form.length == 0) {
                var link = formHolder.find("a");
                if(link !== undefined) {
                    targetFrame.location.href = link.attr("href");
                }
            }

            // POST params - submit the <form>
            if( form.data("dcc") === true) {
                handleDCCScreen(form);
            } else if ( form.data("accarda") === true ) {
                handlePaymentOptionsScreen( form, "accarda" );
            } else if ( form.data("byjuno") === true ) {
                handlePaymentOptionsScreen( form, "byjuno" );
            } else {
                $(".form--state__verifying .form--state-inner .label").text( paymentPage.getTranslation("paym_redirecting") );
                $("body").append( formHolder );

                if ( paymentPage.startOnTop )
                    form.attr("target", "_top");

                form.submit();
            }

        }
    }

    function handleDCCScreen(form) {
        E.emit('app-toggle-actions', {'back': true, 'close': false});
        paymentPage.dcc=true;

        if(form != null) {
            $('.js-form--wrap').replaceWith(form);
        }

        E.emit('app-form-init');

        E.emit('app-state', {
            group: 'step',
            uid: 'payment-form'
        });

        function payHandler( dccDecision ) {
           $.ajax({
               method: "POST",
               headers: {"x-no-redirect": "true"},
               url: paymentPage.contextPath + "/jsp/dccDialogResult.jsp",
                data: { "datatransTrxId": paymentPage.transactionId, "dccDecision": dccDecision }
            }).done( storeTrxResponseHandler );
        }

        $("#dccPay").click( function() { payHandler(1)  });
        $("#originalPay").click( function() { payHandler(0)  });
    }

    function handlePaymentOptionsConfirm( response ) {
        var targetFrame = paymentPage.startOnTop ? top : self;
        response = response.length ? response[0] : response;
        if ( response.success ) {
            if ( response.redirectURL )
                targetFrame.location.href = response.redirectURL;
            else
                targetFrame.location.href = paymentPage.contextPath + "/jsp/upReturn.jsp?" +
                    paymentPage.getDatatransTrxId() + "&uppActionResult=success";
        } else {
            targetFrame.location.href = paymentPage.contextPath + "/jsp/upReturn.jsp?" +
                paymentPage.getDatatransTrxId() + "&uppActionResult=error" +
                "&errorCode=" + response.errorCode +
                "&errorMessage=" + response.errorMessage +
                "&errorDetail=" + response.errorDetail;
        }
    }

    function handlePaymentOptionsScreen( form, paymentMethod ) {

        E.emit('app-toggle-actions', {'back': false, 'close': false});
        paymentPage[paymentMethod] = true;

        if(form != null) {
           $('.js-form--wrap').replaceWith(form);
        }

        E.emit('app-form-init');
        E.emit('app-state', {
            group: 'step',
            uid: 'payment-form'
        });

        prepareView( "form" );

        $('input:radio[name=paymentMethod]:first').prop('checked', true);
    }

    bind("dt-back", function (data) {
        this.doBack.apply(this, [data]);
    });
    bind( "dt-close", this.doCancel );
    bind( "dt-submit", this.doSubmit );
    bind( "dt-method-chosen", this.methodChoosenDispatcher );
    E.on( 'dt-cardtype', this.cardTypeChanged );

    this.setupValidation = function() {
        var validationParams = this.getValidationParams();

        /*
         Validate the form on submit. Set to false to use only other events for validation.
         Set to a Function to decide for yourself when to run validation.
         */
        //validationParams.debug = true;
        validationParams.onsubmit = false;

        validationParams.showErrors = function (errorMap, errorList) {
            // clean up existing errors
            $(".form--field.hasmessage__error").find("input").each(function () {
                var errorFieldName = $(this).attr("name");
                //DT-10728 - invalid has the single truth, errorMap and errorList only contain the CURRENT validated field
                if (errorFieldName && !paymentPage.validator.invalid[errorFieldName]) {
                    E.emit('app-error', {
                        'uid': errorFieldName,
                        'message': null /* null is for remove */,
                        'show': false
                    });
                }
            });

            $.each(errorMap, function (fieldName, errorMessage) {
                var input = $(":input[name=" + fieldName + "]");
                var inputType = input.attr("type") || "select";
                paymentPage.log("Validation error: " + fieldName + " message: " + errorMessage + " type: " + inputType);
                if (inputType === "checkbox") {
                    $("label.checkfield--box[for=" + fieldName + "]").addClass("form--input__error").click(function () {
                        $(this).removeClass("form--input__error");
                    });
                }
                else
                if (inputType === "select") {
                    $(input).addClass("form--input__error").change(function () {
                        $(this).removeClass("form--input__error");
                    });
                }
                else {
                    E.emit('app-error', {'uid': fieldName, 'message': errorMessage, 'show': false});
                }
            });
        };

        $.validator.addMethod("MTANCode", function(value, element) {
            return this.optional(element) || $.payment.validateCardCVC(value);
        } );
        validationParams.messages.MTANCode = this.getTranslation("paym_invalid_MTANCode");
        validationParams.rules.MTANCode = { required: true, MTANCode: true };

        $.validator.addMethod("cardno", function (value, element) {

            if ( !cardValidationEnabled )
                return true;

            if (cardnoAutocompletedValue != null )
                return true;

            var stripeMapValue = paymentPage.stripeMap[$.payment.cardType(value)];

            if (stripeMapValue == null)
                return false;

            if (!cardSwitchAllowed && paymentPage.pmethod != stripeMapValue)
                return false;

            if (cardSwitchAllowed && !$("#cvv").length && paymentPage.pmethod != stripeMapValue)
                return false;

            return this.optional(element) || $.payment.validateCardNumber(value);
        });

        validationParams.rules.cardno = {required: true, cardno: true};

        $.validator.addMethod("cvv", function (value, element) {
            var currentCardType;
            if ( cardValidationEnabled ) {
                currentCardType = $.payment.cardType( $("input[name=cardno]").val() );
            }
            if ( currentCardType == null && cardType != null ) {
                currentCardType = cardType;
            }
            return this.optional(element) || $.payment.validateCardCVC( value, currentCardType );
        });
        validationParams.messages.cvv = this.getTranslation("paym_invalid_cvv");
        validationParams.rules.cvv = { required: paymentPage.mandatoryCVV, cvv: true};

        $.validator.addMethod("expiry", function (value, element) {
            var expiryVal = $.payment.cardExpiryVal(value);
            var valid = $.payment.validateCardExpiry(expiryVal);
            if (valid) {
                $("input[name=expy]").val(expiryVal.year % 100);
                $("input[name=expm]").val(expiryVal.month);
            }
            return valid;
        });
        validationParams.messages.expiry = this.getTranslation("paym_invalid_expiry");
        validationParams.rules.expiry = {expiry: true};


        this.validator =  $("#paymentForm").validate(validationParams);
    }

    this.cleanupErrors = function (){
        $(".form--field.hasmessage__error").find("input").each(function () {
            var errorFieldName = $(this).attr("name");
            //DT-10728 - invalid has the single truth, errorMap and errorList only contain the CURRENT validated field
            if (errorFieldName && !paymentPage.validator.invalid[errorFieldName]) {
                E.emit('app-error', {
                    'uid': errorFieldName,
                    'message': null /* null is for remove */,
                    'show': false
                });
            }
        });
    }

   function prepareAndRenderPaymentPage(){
        prepareView(paymentPage.view);
        E.emit('app-reveal');
        setTimeout(function () {
                E.emit('app-animate');
                waitForPrecheckTasks(function () {

                    setTimeout(function () {

                        E.emit('app-state', {
                            group: 'mantle',
                            uid: 'main'
                        });
                    }, 100);

                    showInfoTexts();
                    paymentPage.startAutoCancelTimeout();
                    paymentPage.focusFirstField(500);
                });
            },
            window.datatrans.config.mode == "overlay" ? 800 : 400);

        E.emit("dt-initialized");
    }

    function renderMethodSelection(url,parameters){
        var promise = $.post(url, parameters).promise();
        setTimeout(function () {

            promise.done(function (data) {

                E.emit('app-state', {
                    group: 'mantle',
                    uid: 'main'
                });

                setTimeout(function () {
                    var htmlData = $(data);
                    $('.js-form--wrap').replaceWith(htmlData.find(".js-form--wrap"));

                    E.emit('app-form-init');

                    E.emit('app-state', {
                        group: 'step',
                        uid: 'payment-form'
                    });

                    firstSubmitTried = false;

                    showInfoTexts();

                    prepareView("form");

                    if ($("#paymentNoNeedLabel").length == 1) {
                        setTimeout(function () {
                            $("#paymentForm").submit();
                        }, 300);
                        return;
                    }
                    paymentPage.startAutoCancelTimeout();
                    paymentPage.focusFirstField(500);
                }, 200);

            });

        }, 500);
    }
});

