paymentPage.extend(function () {
    var config = this.googlePayConfig;
    var allowedPaymentMethods = ['PAN_ONLY', 'CRYPTOGRAM_3DS'];
    var allowedCardNetworks = config.creditCardRequirements.allowedNetworks;
    var googlePayClient = new google.payments.api.PaymentsClient({environment: config.environment});

    var tokenizationParameters = {
        type: 'PAYMENT_GATEWAY',
        parameters: {
            'gateway': config.gatewayParams.processorName,
            'gatewayMerchantId': config.gatewayParams.merchantId
        }
    };

    var baseRequest = {
        apiVersion: 2,
        apiVersionMinor: 0
    };

    var baseCardPaymentMethod = {
        type: 'CARD',
        parameters: {
            allowedAuthMethods: allowedPaymentMethods,
            allowedCardNetworks: allowedCardNetworks,
            assuranceDetailsRequired: true
        }
    };

    var cardPaymentMethod = $.extend(
        {},
        baseCardPaymentMethod,
        {
            tokenizationSpecification: tokenizationParameters
        }
    );

    var isReadToPayRequestObject = $.extend({},baseRequest,{allowedPaymentMethods: [baseCardPaymentMethod]});

    var isSafariBasedBrowser = navigator.vendor && navigator.vendor.indexOf('Apple') > -1;

    if(config.walletMode && !(window.navigator.userAgent.indexOf("Edg") > -1) && !isSafariBasedBrowser){
        isReadToPayRequestObject.existingPaymentMethodRequired = true;
    }

    var E = window.datatrans.emitter;
    E.on("dt-initialized",function () {
        googlePayClient.isReadyToPay(isReadToPayRequestObject)
            .then(function (response) {
                if (response.result) {
                    paymentPage.googlePayAvailable = true;
                    if(response.paymentMethodPresent){
                        paymentPage.paymentMethodPresent = true;
                    }
                }
                E.emit("googlepay-check-done");
            })
            .catch(function (err) {
                // show error in developer console for debugging
                paymentPage.log("Failed to determine google pay readyness: " + err);
                E.emit("googlepay-check-done");
            });
    });

    this.renderGooglePayButton = function(){
        var container = document.getElementById("walletsContainer");
        var button = googlePayClient.createButton({ onClick:  paymentPage.startGooglePayRequest, buttonType: ""+getButtonType()+"", buttonColor: "black" })
        button.firstChild.setAttribute('style','width:100%');
        container.appendChild(button)
    };

    function getButtonType() {
        var buttonType = "long";
        if(config.transactionInfo.amount == 0){
            buttonType = "short";
        }
        return buttonType;
    }

    function getGoogleTransactionInfo() {
        var transactionInfo =  {
            currencyCode: config.transactionInfo.currency,
            totalPriceStatus: 'FINAL',
            totalPrice: config.transactionInfo.amount
        };
        if(config.transactionInfo.countryCode){
            transactionInfo.countryCode = config.transactionInfo.countryCode;
        }
        return transactionInfo;
    }


    function getGooglePaymentDataConfiguration() {
        var paymentDataRequest = $.extend({}, baseRequest);
        paymentDataRequest.allowedPaymentMethods = [cardPaymentMethod];
        paymentDataRequest.merchantInfo = {
            merchantId: '06110866861057621228',
            merchantName: config.merchantName
        };

        paymentDataRequest.transactionInfo = getGoogleTransactionInfo();


        if (config.shippingRequirements) {
            if (config.shippingRequirements.shippingRequired) {
                paymentDataRequest.shippingAddressRequired = true;
                paymentDataRequest.shippingAddressParameters = {};
            }
        }

        if (config.emailRequired) {
            paymentDataRequest.emailRequired = true;
        }

        var billingAddressParameters ={};

        if (config.phoneRequired) {
            billingAddressParameters.phoneNumberRequired = true;
        }

        if (config.creditCardRequirements.billingRequirements.billingAddressRequired) {
            paymentDataRequest.allowedPaymentMethods[0].parameters.billingAddressRequired = true;
            if (config.creditCardRequirements.billingRequirements.billingAddressFormat) {
                billingAddressParameters.format = config.creditCardRequirements.billingRequirements.billingAddressFormat;
            }
        }

        paymentDataRequest.allowedPaymentMethods[0].parameters.billingAddressParameters = billingAddressParameters;

        return paymentDataRequest;
    }

    this.startGooglePayRequest = function () {
        if (config.interceptGooglePay) {
            var url = this.contextPath + "/payment?pmethod=PAY"

            var parameters = {
                "datatransTrxId": this.transactionId,
            };

            $.post(url, parameters);
        } else {
            var paymentDataRequest = getGooglePaymentDataConfiguration();
            paymentDataRequest.transactionInfo = getGoogleTransactionInfo();
            googlePayClient.loadPaymentData(paymentDataRequest)
                .then(function (paymentData) {
                    var data = {method: "PAY", paymentData: paymentData}
                    paymentPage.methodChosen(data)
                })
                .catch(function (err) {
                    // show error in developer console for debugging
                    paymentPage.log(err);
                });
        }
    }


});