webpackSbbInternetFrontendJsonp([16,3],[
/* 0 */,
/* 1 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * Helper class to register modules as jQuery plugins and do some other magic
 *
 * @license APLv2
 */



var $ = __webpack_require__(0),
	log = estatico.helpers.log(name),
	uniqueId = __webpack_require__(33),
	isElement = __webpack_require__(34),
	promisesToResolve = {},
	resolvePromises,
	getByElement,
	destroyModules,
	initializeModules = __webpack_require__(35);

/**
 * Helper class
 * @param {object} config
 * @param {string} config.name - Module name (used as namespace for all kinds of things)
 * @param {object} config.defaults - Default options
 * @param {object} config.element - DOM element to init the module on
 * @param {object} config.options - Custom options
 * @param {object} config.data - Custom data
 */
function SuperClass(config) {
	// Save things
	this.name = config.name;

	this.$element = $(config.element);

	// Handle options
	this._defaults = config.defaults;
	this._options = config.options;
	this._metaOptions = this.$element.data(this.name + 'Options');
	this._globalOptions = estatico.options[this.name];

	this.options = $.extend(true, {}, this._defaults, this._options, this._globalOptions, this._metaOptions);

	// Handle data
	this._data = config.data;
	this._globalData = estatico.data[this.name];
	this._metaData = this.$element.data(this.name + 'Data');

	// Module events
	this._moduleEvents = config.moduleEvents;

	this.data = $.extend(true, {}, this._data, this._globalData, this._metaData);

	// Identify instance by UUID
	this.uuid = uniqueId(this.name);

	// Save instance
	estatico.modules[this.name].instances[this.uuid] = this;

	// Have fun
	this.init();

	resolvePromises(this);
}

/**
 * Init method
 *
 * Should likely be overwritten in module (otherwise nothing will happen on init)
 */
SuperClass.prototype.init = function() {
	log('Module initialized', this.$element);
};

/**
 * Destroy method
 *
 * Should be overwritten in module if there are additional DOM elements, DOM data, event listeners to remove
 *
 * Use cases:
 * - Unbind (namespaced) event listeners
 * - Remove data from DOM elements
 * - Remove elements from DOM
 */
SuperClass.prototype.destroy = function() {
	// Remove event listeners connected to this instance
	this.$element.off('.' + this.uuid);

	$(document).off('.' + this.uuid);

	// make sure that on destroy, registered event handlers on the window
	// object are removed as well
	$(window).off('.' + this.uuid);

	// Delete references to instance
	this.$element.removeData(this.name + 'Instance');

	delete estatico.modules[this.name].instances[this.uuid];
	this.destroyed = true;
};

SuperClass.prototype.on = function(eventName, cb) {

	if (this._$callbacks && this._$callbacks[eventName]) {
		this._$callbacks[eventName].add(cb);
		return;
	}

	if (!this._moduleEvents || this._moduleEvents.indexOf(eventName) === -1) {
		throw new Error(
			this.name + ' module does not support the following event: ' + eventName
		);
	}

	this._$callbacks = this._$callbacks || {};
	this._$callbacks[eventName] = $.Callbacks();
	this.on(eventName, cb);

	return this;
};

SuperClass.prototype.trigger = function(eventName, data) {

	if (!this._$callbacks || !this._$callbacks[eventName]) {
		return;
	}

	this._$callbacks[eventName].fireWith(this, [data]);
	return this;
};

/**
 * Block default click event on <a> tags, when we listen to tap, and want to prevent second click.
 * @param event {jQuery}
 */
SuperClass.prototype._blockDefaultClick = function(event) {
	event.preventDefault();
};

/**
 * Allows quickly generating selector string based only on partial attribute name.
 *
 * Instead of creating this string by hand ie. [data-megadropdown="some_element"]
 * pass only "some_element" to get the rest for jquery selector.
 *
 * @param attributeName string to build attribute name for dom selector
 * @returns {string} selector name based on attribute
 */
SuperClass.prototype.getDomSelector = function(attributeName) {
	return '[data-' + this.name + '="' + attributeName + '"]';
};

/**
 * Register the module as jQuery plugin, auto-init at specified events
 *
 * Allows to call all public methods (not starting with underscore).
 * E.g. $(element).moduleName('methodName', arg1, arg2)
 *
 * @param {function} Class - Constructor
 * @param {string} name - Module name (used as namespace for all kind of things)
 * @param {object} config
 * @param {array} config.events - List of custom events
 */
SuperClass.register = function(Class, name, config) {
	// Register jQuery plugin
	$.fn[name] = function(options) {
		var args = arguments,
			value;

		if (options === undefined || typeof options === 'object') {
			// Init module
			return this.each(function() {
				var $this = $(this),
					instance = $this.data(name + 'Instance');

				// Check if the module was already instantiated
				if (instance === undefined) {
					instance = new Class(this, options);

					// Save instance to DOM element data
					$this.data(name + 'Instance', instance);
				}

			});
		} else if (typeof options === 'string') {
			// Private method, throw error
			if (options.substr(0, 1) === '_') {
				throw '"' + options + '" is a private method';
			}

			// Loop through elements
			this.each(function() {
				var instance = $(this).data(name + 'Instance');

				// No module instance found, throw error
				if (!(instance instanceof Class)) {
					throw 'Instance of "' + name + '" module not found';
				}

				// Method not found, throw error
				if (typeof instance[options] !== 'function') {
					throw '"' + name + '" has no method "' + options + '"';
				}

				value = instance[options].apply(instance, Array.prototype.slice.call(args, 1));

				// Stop if method returned a value
				if (value) {
					return false;
				}
			});

			// Return either value or jQuery collection
			if (value) {
				return value;
			} else {
				return this;
			}
		}
	};

	// Save to global namespace
	estatico.modules[name] = {
		instances: {},
		events: config.events,
		Class: Class
	};

	initializeModules(name, config.initEvents);
};

/**
 * This method is intended to be used with 3rd party code that does not have
 * access to our instance of jQuery. It returns a promise that is resolved with
 * the initialized module, once it is initialized. This function gets exposed
 * globally through `estatico.modules.getByElement('name', element)`.
 *
 * Usage example:
 *
 * ```
 * estatico.modules.getByElement('accordion', $('.some.element')[0])
 * .then(function(module) {
 *   // Do what you need to do.
 * })
 *
 * ```
 * @param {string} name - Module name (used as namespace for all kind of things)
 * @param {DOMElement} element
 */
getByElement = function(name, element) {
	var deferred, module;

	if (!isElement(element)) {
		throw new Error('The second parameter of `estatico.modules.getByElement()` should be a DOMElement');
	}

	deferred = $.Deferred();

	module = $(element).data(name + 'Instance');

	if (module) {
		deferred.resolve(module);
	} else {
		promisesToResolve[name] = promisesToResolve[name] || [];
		promisesToResolve[name].push({
			element: element,
			deferred: deferred
		});
	}

	return deferred.promise();
};

destroyModules = function(element) {
	var $element = $(element),
		currentElementNeedsDestruction = $element.is('[data-init]'),
		$elementsThatNeedDestruction = $element.find('[data-init]');

	if (currentElementNeedsDestruction) {
		$elementsThatNeedDestruction = $elementsThatNeedDestruction.addBack();
	}

	$elementsThatNeedDestruction.each(function(index, elementToDestroy) {
		var $elementToDestroy = $(elementToDestroy),
			modules = $elementToDestroy.data('init').split(' ');

		modules.forEach(function(module) {
			getByElement(module, elementToDestroy).then(function(module) {
				module.destroy();
			});
		});
	});
};

resolvePromises = function(module) {
	var $element = module.$element,
		element = $element[0],
		name = module.name;

	if (!promisesToResolve[name]) {
		return;
	}

	promisesToResolve[name] = promisesToResolve[name].filter(function(object) {
		var currentElement = object.element,
			deferred = object.deferred;

		if (element !== currentElement) {
			return true;
		}

		deferred.resolve(module);
	});
};

// Save to global namespace
$.extend(true, estatico, {
	modules: {
		getByElement: getByElement,
		destroyModules: destroyModules
	},
	helpers: {
		SuperClass: SuperClass
	}
});

module.exports = SuperClass;


/***/ }),
/* 2 */,
/* 3 */
/***/ (function(module, exports, __webpack_require__) {

/* WEBPACK VAR INJECTION */(function(global) {/*!
* Parsley.js
* Version 2.9.2 - built Tue, Dec 10th 2019, 6:18 pm
* http://parsleyjs.org
* Guillaume Potier - <guillaume@wisembly.com>
* Marc-Andre Lafortune - <petroselinum@marc-andre.ca>
* MIT Licensed
*/

// The source code below is generated by babel as
// Parsley is written in ECMAScript 6
//

(function (global, factory) {
   true ? module.exports = factory(__webpack_require__(0)) :
  typeof define === 'function' && define.amd ? define(['jquery'], factory) :
  (global = global || self, global.parsley = factory(global.jQuery));
}(this, (function ($) { 'use strict';

  function _typeof(obj) {
    if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
      _typeof = function (obj) {
        return typeof obj;
      };
    } else {
      _typeof = function (obj) {
        return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
      };
    }

    return _typeof(obj);
  }

  function _extends() {
    _extends = Object.assign || function (target) {
      for (var i = 1; i < arguments.length; i++) {
        var source = arguments[i];

        for (var key in source) {
          if (Object.prototype.hasOwnProperty.call(source, key)) {
            target[key] = source[key];
          }
        }
      }

      return target;
    };

    return _extends.apply(this, arguments);
  }

  function _slicedToArray(arr, i) {
    return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
  }

  function _toConsumableArray(arr) {
    return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
  }

  function _arrayWithoutHoles(arr) {
    if (Array.isArray(arr)) {
      for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

      return arr2;
    }
  }

  function _arrayWithHoles(arr) {
    if (Array.isArray(arr)) return arr;
  }

  function _iterableToArray(iter) {
    if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
  }

  function _iterableToArrayLimit(arr, i) {
    if (!(Symbol.iterator in Object(arr) || Object.prototype.toString.call(arr) === "[object Arguments]")) {
      return;
    }

    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"] != null) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  function _nonIterableSpread() {
    throw new TypeError("Invalid attempt to spread non-iterable instance");
  }

  function _nonIterableRest() {
    throw new TypeError("Invalid attempt to destructure non-iterable instance");
  }

  var globalID = 1;
  var pastWarnings = {};
  var Utils = {
    // Parsley DOM-API
    // returns object from dom attributes and values
    attr: function attr(element, namespace, obj) {
      var i;
      var attribute;
      var attributes;
      var regex = new RegExp('^' + namespace, 'i');
      if ('undefined' === typeof obj) obj = {};else {
        // Clear all own properties. This won't affect prototype's values
        for (i in obj) {
          if (obj.hasOwnProperty(i)) delete obj[i];
        }
      }
      if (!element) return obj;
      attributes = element.attributes;

      for (i = attributes.length; i--;) {
        attribute = attributes[i];

        if (attribute && attribute.specified && regex.test(attribute.name)) {
          obj[this.camelize(attribute.name.slice(namespace.length))] = this.deserializeValue(attribute.value);
        }
      }

      return obj;
    },
    checkAttr: function checkAttr(element, namespace, _checkAttr) {
      return element.hasAttribute(namespace + _checkAttr);
    },
    setAttr: function setAttr(element, namespace, attr, value) {
      element.setAttribute(this.dasherize(namespace + attr), String(value));
    },
    getType: function getType(element) {
      return element.getAttribute('type') || 'text';
    },
    generateID: function generateID() {
      return '' + globalID++;
    },

    /** Third party functions **/
    deserializeValue: function deserializeValue(value) {
      var num;

      try {
        return value ? value == "true" || (value == "false" ? false : value == "null" ? null : !isNaN(num = Number(value)) ? num : /^[\[\{]/.test(value) ? JSON.parse(value) : value) : value;
      } catch (e) {
        return value;
      }
    },
    // Zepto camelize function
    camelize: function camelize(str) {
      return str.replace(/-+(.)?/g, function (match, chr) {
        return chr ? chr.toUpperCase() : '';
      });
    },
    // Zepto dasherize function
    dasherize: function dasherize(str) {
      return str.replace(/::/g, '/').replace(/([A-Z]+)([A-Z][a-z])/g, '$1_$2').replace(/([a-z\d])([A-Z])/g, '$1_$2').replace(/_/g, '-').toLowerCase();
    },
    warn: function warn() {
      var _window$console;

      if (window.console && 'function' === typeof window.console.warn) (_window$console = window.console).warn.apply(_window$console, arguments);
    },
    warnOnce: function warnOnce(msg) {
      if (!pastWarnings[msg]) {
        pastWarnings[msg] = true;
        this.warn.apply(this, arguments);
      }
    },
    _resetWarnings: function _resetWarnings() {
      pastWarnings = {};
    },
    trimString: function trimString(string) {
      return string.replace(/^\s+|\s+$/g, '');
    },
    parse: {
      date: function date(string) {
        var parsed = string.match(/^(\d{4,})-(\d\d)-(\d\d)$/);
        if (!parsed) return null;

        var _parsed$map = parsed.map(function (x) {
          return parseInt(x, 10);
        }),
            _parsed$map2 = _slicedToArray(_parsed$map, 4),
            _ = _parsed$map2[0],
            year = _parsed$map2[1],
            month = _parsed$map2[2],
            day = _parsed$map2[3];

        var date = new Date(year, month - 1, day);
        if (date.getFullYear() !== year || date.getMonth() + 1 !== month || date.getDate() !== day) return null;
        return date;
      },
      string: function string(_string) {
        return _string;
      },
      integer: function integer(string) {
        if (isNaN(string)) return null;
        return parseInt(string, 10);
      },
      number: function number(string) {
        if (isNaN(string)) throw null;
        return parseFloat(string);
      },
      'boolean': function _boolean(string) {
        return !/^\s*false\s*$/i.test(string);
      },
      object: function object(string) {
        return Utils.deserializeValue(string);
      },
      regexp: function regexp(_regexp) {
        var flags = ''; // Test if RegExp is literal, if not, nothing to be done, otherwise, we need to isolate flags and pattern

        if (/^\/.*\/(?:[gimy]*)$/.test(_regexp)) {
          // Replace the regexp literal string with the first match group: ([gimy]*)
          // If no flag is present, this will be a blank string
          flags = _regexp.replace(/.*\/([gimy]*)$/, '$1'); // Again, replace the regexp literal string with the first match group:
          // everything excluding the opening and closing slashes and the flags

          _regexp = _regexp.replace(new RegExp('^/(.*?)/' + flags + '$'), '$1');
        } else {
          // Anchor regexp:
          _regexp = '^' + _regexp + '$';
        }

        return new RegExp(_regexp, flags);
      }
    },
    parseRequirement: function parseRequirement(requirementType, string) {
      var converter = this.parse[requirementType || 'string'];
      if (!converter) throw 'Unknown requirement specification: "' + requirementType + '"';
      var converted = converter(string);
      if (converted === null) throw "Requirement is not a ".concat(requirementType, ": \"").concat(string, "\"");
      return converted;
    },
    namespaceEvents: function namespaceEvents(events, namespace) {
      events = this.trimString(events || '').split(/\s+/);
      if (!events[0]) return '';
      return $.map(events, function (evt) {
        return "".concat(evt, ".").concat(namespace);
      }).join(' ');
    },
    difference: function difference(array, remove) {
      // This is O(N^2), should be optimized
      var result = [];
      $.each(array, function (_, elem) {
        if (remove.indexOf(elem) == -1) result.push(elem);
      });
      return result;
    },
    // Alter-ego to native Promise.all, but for jQuery
    all: function all(promises) {
      // jQuery treats $.when() and $.when(singlePromise) differently; let's avoid that and add spurious elements
      return $.when.apply($, _toConsumableArray(promises).concat([42, 42]));
    },
    // Object.create polyfill, see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/create#Polyfill
    objectCreate: Object.create || function () {
      var Object = function Object() {};

      return function (prototype) {
        if (arguments.length > 1) {
          throw Error('Second argument not supported');
        }

        if (_typeof(prototype) != 'object') {
          throw TypeError('Argument must be an object');
        }

        Object.prototype = prototype;
        var result = new Object();
        Object.prototype = null;
        return result;
      };
    }(),
    _SubmitSelector: 'input[type="submit"], button:submit'
  };

  // All these options could be overriden and specified directly in DOM using
  // `data-parsley-` default DOM-API
  // eg: `inputs` can be set in DOM using `data-parsley-inputs="input, textarea"`
  // eg: `data-parsley-stop-on-first-failing-constraint="false"`
  var Defaults = {
    // ### General
    // Default data-namespace for DOM API
    namespace: 'data-parsley-',
    // Supported inputs by default
    inputs: 'input, textarea, select',
    // Excluded inputs by default
    excluded: 'input[type=button], input[type=submit], input[type=reset], input[type=hidden]',
    // Stop validating field on highest priority failing constraint
    priorityEnabled: true,
    // ### Field only
    // identifier used to group together inputs (e.g. radio buttons...)
    multiple: null,
    // identifier (or array of identifiers) used to validate only a select group of inputs
    group: null,
    // ### UI
    // Enable\Disable error messages
    uiEnabled: true,
    // Key events threshold before validation
    validationThreshold: 3,
    // Focused field on form validation error. 'first'|'last'|'none'
    focus: 'first',
    // event(s) that will trigger validation before first failure. eg: `input`...
    trigger: false,
    // event(s) that will trigger validation after first failure.
    triggerAfterFailure: 'input',
    // Class that would be added on every failing validation Parsley field
    errorClass: 'parsley-error',
    // Same for success validation
    successClass: 'parsley-success',
    // Return the `$element` that will receive these above success or error classes
    // Could also be (and given directly from DOM) a valid selector like `'#div'`
    classHandler: function classHandler(Field) {},
    // Return the `$element` where errors will be appended
    // Could also be (and given directly from DOM) a valid selector like `'#div'`
    errorsContainer: function errorsContainer(Field) {},
    // ul elem that would receive errors' list
    errorsWrapper: '<ul class="parsley-errors-list"></ul>',
    // li elem that would receive error message
    errorTemplate: '<li></li>'
  };

  var Base = function Base() {
    this.__id__ = Utils.generateID();
  };

  Base.prototype = {
    asyncSupport: true,
    // Deprecated
    _pipeAccordingToValidationResult: function _pipeAccordingToValidationResult() {
      var _this = this;

      var pipe = function pipe() {
        var r = $.Deferred();
        if (true !== _this.validationResult) r.reject();
        return r.resolve().promise();
      };

      return [pipe, pipe];
    },
    actualizeOptions: function actualizeOptions() {
      Utils.attr(this.element, this.options.namespace, this.domOptions);
      if (this.parent && this.parent.actualizeOptions) this.parent.actualizeOptions();
      return this;
    },
    _resetOptions: function _resetOptions(initOptions) {
      this.domOptions = Utils.objectCreate(this.parent.options);
      this.options = Utils.objectCreate(this.domOptions); // Shallow copy of ownProperties of initOptions:

      for (var i in initOptions) {
        if (initOptions.hasOwnProperty(i)) this.options[i] = initOptions[i];
      }

      this.actualizeOptions();
    },
    _listeners: null,
    // Register a callback for the given event name
    // Callback is called with context as the first argument and the `this`
    // The context is the current parsley instance, or window.Parsley if global
    // A return value of `false` will interrupt the calls
    on: function on(name, fn) {
      this._listeners = this._listeners || {};
      var queue = this._listeners[name] = this._listeners[name] || [];
      queue.push(fn);
      return this;
    },
    // Deprecated. Use `on` instead
    subscribe: function subscribe(name, fn) {
      $.listenTo(this, name.toLowerCase(), fn);
    },
    // Unregister a callback (or all if none is given) for the given event name
    off: function off(name, fn) {
      var queue = this._listeners && this._listeners[name];

      if (queue) {
        if (!fn) {
          delete this._listeners[name];
        } else {
          for (var i = queue.length; i--;) {
            if (queue[i] === fn) queue.splice(i, 1);
          }
        }
      }

      return this;
    },
    // Deprecated. Use `off`
    unsubscribe: function unsubscribe(name, fn) {
      $.unsubscribeTo(this, name.toLowerCase());
    },
    // Trigger an event of the given name
    // A return value of `false` interrupts the callback chain
    // Returns false if execution was interrupted
    trigger: function trigger(name, target, extraArg) {
      target = target || this;
      var queue = this._listeners && this._listeners[name];
      var result;

      if (queue) {
        for (var i = queue.length; i--;) {
          result = queue[i].call(target, target, extraArg);
          if (result === false) return result;
        }
      }

      if (this.parent) {
        return this.parent.trigger(name, target, extraArg);
      }

      return true;
    },
    asyncIsValid: function asyncIsValid(group, force) {
      Utils.warnOnce("asyncIsValid is deprecated; please use whenValid instead");
      return this.whenValid({
        group: group,
        force: force
      });
    },
    _findRelated: function _findRelated() {
      return this.options.multiple ? $(this.parent.element.querySelectorAll("[".concat(this.options.namespace, "multiple=\"").concat(this.options.multiple, "\"]"))) : this.$element;
    }
  };

  var convertArrayRequirement = function convertArrayRequirement(string, length) {
    var m = string.match(/^\s*\[(.*)\]\s*$/);
    if (!m) throw 'Requirement is not an array: "' + string + '"';
    var values = m[1].split(',').map(Utils.trimString);
    if (values.length !== length) throw 'Requirement has ' + values.length + ' values when ' + length + ' are needed';
    return values;
  };

  var convertExtraOptionRequirement = function convertExtraOptionRequirement(requirementSpec, string, extraOptionReader) {
    var main = null;
    var extra = {};

    for (var key in requirementSpec) {
      if (key) {
        var value = extraOptionReader(key);
        if ('string' === typeof value) value = Utils.parseRequirement(requirementSpec[key], value);
        extra[key] = value;
      } else {
        main = Utils.parseRequirement(requirementSpec[key], string);
      }
    }

    return [main, extra];
  }; // A Validator needs to implement the methods `validate` and `parseRequirements`


  var Validator = function Validator(spec) {
    $.extend(true, this, spec);
  };

  Validator.prototype = {
    // Returns `true` iff the given `value` is valid according the given requirements.
    validate: function validate(value, requirementFirstArg) {
      if (this.fn) {
        // Legacy style validator
        if (arguments.length > 3) // If more args then value, requirement, instance...
          requirementFirstArg = [].slice.call(arguments, 1, -1); // Skip first arg (value) and last (instance), combining the rest

        return this.fn(value, requirementFirstArg);
      }

      if (Array.isArray(value)) {
        if (!this.validateMultiple) throw 'Validator `' + this.name + '` does not handle multiple values';
        return this.validateMultiple.apply(this, arguments);
      } else {
        var instance = arguments[arguments.length - 1];

        if (this.validateDate && instance._isDateInput()) {
          arguments[0] = Utils.parse.date(arguments[0]);
          if (arguments[0] === null) return false;
          return this.validateDate.apply(this, arguments);
        }

        if (this.validateNumber) {
          if (!value) // Builtin validators all accept empty strings, except `required` of course
            return true;
          if (isNaN(value)) return false;
          arguments[0] = parseFloat(arguments[0]);
          return this.validateNumber.apply(this, arguments);
        }

        if (this.validateString) {
          return this.validateString.apply(this, arguments);
        }

        throw 'Validator `' + this.name + '` only handles multiple values';
      }
    },
    // Parses `requirements` into an array of arguments,
    // according to `this.requirementType`
    parseRequirements: function parseRequirements(requirements, extraOptionReader) {
      if ('string' !== typeof requirements) {
        // Assume requirement already parsed
        // but make sure we return an array
        return Array.isArray(requirements) ? requirements : [requirements];
      }

      var type = this.requirementType;

      if (Array.isArray(type)) {
        var values = convertArrayRequirement(requirements, type.length);

        for (var i = 0; i < values.length; i++) {
          values[i] = Utils.parseRequirement(type[i], values[i]);
        }

        return values;
      } else if ($.isPlainObject(type)) {
        return convertExtraOptionRequirement(type, requirements, extraOptionReader);
      } else {
        return [Utils.parseRequirement(type, requirements)];
      }
    },
    // Defaults:
    requirementType: 'string',
    priority: 2
  };

  var ValidatorRegistry = function ValidatorRegistry(validators, catalog) {
    this.__class__ = 'ValidatorRegistry'; // Default Parsley locale is en

    this.locale = 'en';
    this.init(validators || {}, catalog || {});
  };

  var typeTesters = {
    email: /^((([a-zA-Z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+(\.([a-zA-Z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+)*)|((\x22)((((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(([\x01-\x08\x0b\x0c\x0e-\x1f\x7f]|\x21|[\x23-\x5b]|[\x5d-\x7e]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(\\([\x01-\x09\x0b\x0c\x0d-\x7f]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))))*(((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(\x22)))@((([a-zA-Z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-zA-Z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-zA-Z]|\d|-|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-zA-Z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))\.)+(([a-zA-Z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-zA-Z]|\d|-|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-zA-Z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))$/,
    // Follow https://www.w3.org/TR/html5/infrastructure.html#floating-point-numbers
    number: /^-?(\d*\.)?\d+(e[-+]?\d+)?$/i,
    integer: /^-?\d+$/,
    digits: /^\d+$/,
    alphanum: /^\w+$/i,
    date: {
      test: function test(value) {
        return Utils.parse.date(value) !== null;
      }
    },
    url: new RegExp("^" + // protocol identifier
    "(?:(?:https?|ftp)://)?" + // ** mod: make scheme optional
    // user:pass authentication
    "(?:\\S+(?::\\S*)?@)?" + "(?:" + // IP address exclusion
    // private & local networks
    // "(?!(?:10|127)(?:\\.\\d{1,3}){3})" +   // ** mod: allow local networks
    // "(?!(?:169\\.254|192\\.168)(?:\\.\\d{1,3}){2})" +  // ** mod: allow local networks
    // "(?!172\\.(?:1[6-9]|2\\d|3[0-1])(?:\\.\\d{1,3}){2})" +  // ** mod: allow local networks
    // IP address dotted notation octets
    // excludes loopback network 0.0.0.0
    // excludes reserved space >= 224.0.0.0
    // excludes network & broacast addresses
    // (first & last IP address of each class)
    "(?:[1-9]\\d?|1\\d\\d|2[01]\\d|22[0-3])" + "(?:\\.(?:1?\\d{1,2}|2[0-4]\\d|25[0-5])){2}" + "(?:\\.(?:[1-9]\\d?|1\\d\\d|2[0-4]\\d|25[0-4]))" + "|" + // host name
    "(?:(?:[a-zA-Z\\u00a1-\\uffff0-9]-*)*[a-zA-Z\\u00a1-\\uffff0-9]+)" + // domain name
    "(?:\\.(?:[a-zA-Z\\u00a1-\\uffff0-9]-*)*[a-zA-Z\\u00a1-\\uffff0-9]+)*" + // TLD identifier
    "(?:\\.(?:[a-zA-Z\\u00a1-\\uffff]{2,}))" + ")" + // port number
    "(?::\\d{2,5})?" + // resource path
    "(?:/\\S*)?" + "$")
  };
  typeTesters.range = typeTesters.number; // See http://stackoverflow.com/a/10454560/8279

  var decimalPlaces = function decimalPlaces(num) {
    var match = ('' + num).match(/(?:\.(\d+))?(?:[eE]([+-]?\d+))?$/);

    if (!match) {
      return 0;
    }

    return Math.max(0, // Number of digits right of decimal point.
    (match[1] ? match[1].length : 0) - ( // Adjust for scientific notation.
    match[2] ? +match[2] : 0));
  }; // parseArguments('number', ['1', '2']) => [1, 2]


  var parseArguments = function parseArguments(type, args) {
    return args.map(Utils.parse[type]);
  }; // operatorToValidator returns a validating function for an operator function, applied to the given type


  var operatorToValidator = function operatorToValidator(type, operator) {
    return function (value) {
      for (var _len = arguments.length, requirementsAndInput = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        requirementsAndInput[_key - 1] = arguments[_key];
      }

      requirementsAndInput.pop(); // Get rid of `input` argument

      return operator.apply(void 0, [value].concat(_toConsumableArray(parseArguments(type, requirementsAndInput))));
    };
  };

  var comparisonOperator = function comparisonOperator(operator) {
    return {
      validateDate: operatorToValidator('date', operator),
      validateNumber: operatorToValidator('number', operator),
      requirementType: operator.length <= 2 ? 'string' : ['string', 'string'],
      // Support operators with a 1 or 2 requirement(s)
      priority: 30
    };
  };

  ValidatorRegistry.prototype = {
    init: function init(validators, catalog) {
      this.catalog = catalog; // Copy prototype's validators:

      this.validators = _extends({}, this.validators);

      for (var name in validators) {
        this.addValidator(name, validators[name].fn, validators[name].priority);
      }

      window.Parsley.trigger('parsley:validator:init');
    },
    // Set new messages locale if we have dictionary loaded in ParsleyConfig.i18n
    setLocale: function setLocale(locale) {
      if ('undefined' === typeof this.catalog[locale]) throw new Error(locale + ' is not available in the catalog');
      this.locale = locale;
      return this;
    },
    // Add a new messages catalog for a given locale. Set locale for this catalog if set === `true`
    addCatalog: function addCatalog(locale, messages, set) {
      if ('object' === _typeof(messages)) this.catalog[locale] = messages;
      if (true === set) return this.setLocale(locale);
      return this;
    },
    // Add a specific message for a given constraint in a given locale
    addMessage: function addMessage(locale, name, message) {
      if ('undefined' === typeof this.catalog[locale]) this.catalog[locale] = {};
      this.catalog[locale][name] = message;
      return this;
    },
    // Add messages for a given locale
    addMessages: function addMessages(locale, nameMessageObject) {
      for (var name in nameMessageObject) {
        this.addMessage(locale, name, nameMessageObject[name]);
      }

      return this;
    },
    // Add a new validator
    //
    //    addValidator('custom', {
    //        requirementType: ['integer', 'integer'],
    //        validateString: function(value, from, to) {},
    //        priority: 22,
    //        messages: {
    //          en: "Hey, that's no good",
    //          fr: "Aye aye, pas bon du tout",
    //        }
    //    })
    //
    // Old API was addValidator(name, function, priority)
    //
    addValidator: function addValidator(name, arg1, arg2) {
      if (this.validators[name]) Utils.warn('Validator "' + name + '" is already defined.');else if (Defaults.hasOwnProperty(name)) {
        Utils.warn('"' + name + '" is a restricted keyword and is not a valid validator name.');
        return;
      }
      return this._setValidator.apply(this, arguments);
    },
    hasValidator: function hasValidator(name) {
      return !!this.validators[name];
    },
    updateValidator: function updateValidator(name, arg1, arg2) {
      if (!this.validators[name]) {
        Utils.warn('Validator "' + name + '" is not already defined.');
        return this.addValidator.apply(this, arguments);
      }

      return this._setValidator.apply(this, arguments);
    },
    removeValidator: function removeValidator(name) {
      if (!this.validators[name]) Utils.warn('Validator "' + name + '" is not defined.');
      delete this.validators[name];
      return this;
    },
    _setValidator: function _setValidator(name, validator, priority) {
      if ('object' !== _typeof(validator)) {
        // Old style validator, with `fn` and `priority`
        validator = {
          fn: validator,
          priority: priority
        };
      }

      if (!validator.validate) {
        validator = new Validator(validator);
      }

      this.validators[name] = validator;

      for (var locale in validator.messages || {}) {
        this.addMessage(locale, name, validator.messages[locale]);
      }

      return this;
    },
    getErrorMessage: function getErrorMessage(constraint) {
      var message; // Type constraints are a bit different, we have to match their requirements too to find right error message

      if ('type' === constraint.name) {
        var typeMessages = this.catalog[this.locale][constraint.name] || {};
        message = typeMessages[constraint.requirements];
      } else message = this.formatMessage(this.catalog[this.locale][constraint.name], constraint.requirements);

      return message || this.catalog[this.locale].defaultMessage || this.catalog.en.defaultMessage;
    },
    // Kind of light `sprintf()` implementation
    formatMessage: function formatMessage(string, parameters) {
      if ('object' === _typeof(parameters)) {
        for (var i in parameters) {
          string = this.formatMessage(string, parameters[i]);
        }

        return string;
      }

      return 'string' === typeof string ? string.replace(/%s/i, parameters) : '';
    },
    // Here is the Parsley default validators list.
    // A validator is an object with the following key values:
    //  - priority: an integer
    //  - requirement: 'string' (default), 'integer', 'number', 'regexp' or an Array of these
    //  - validateString, validateMultiple, validateNumber: functions returning `true`, `false` or a promise
    // Alternatively, a validator can be a function that returns such an object
    //
    validators: {
      notblank: {
        validateString: function validateString(value) {
          return /\S/.test(value);
        },
        priority: 2
      },
      required: {
        validateMultiple: function validateMultiple(values) {
          return values.length > 0;
        },
        validateString: function validateString(value) {
          return /\S/.test(value);
        },
        priority: 512
      },
      type: {
        validateString: function validateString(value, type) {
          var _ref = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {},
              _ref$step = _ref.step,
              step = _ref$step === void 0 ? 'any' : _ref$step,
              _ref$base = _ref.base,
              base = _ref$base === void 0 ? 0 : _ref$base;

          var tester = typeTesters[type];

          if (!tester) {
            throw new Error('validator type `' + type + '` is not supported');
          }

          if (!value) return true; // Builtin validators all accept empty strings, except `required` of course

          if (!tester.test(value)) return false;

          if ('number' === type) {
            if (!/^any$/i.test(step || '')) {
              var nb = Number(value);
              var decimals = Math.max(decimalPlaces(step), decimalPlaces(base));
              if (decimalPlaces(nb) > decimals) // Value can't have too many decimals
                return false; // Be careful of rounding errors by using integers.

              var toInt = function toInt(f) {
                return Math.round(f * Math.pow(10, decimals));
              };

              if ((toInt(nb) - toInt(base)) % toInt(step) != 0) return false;
            }
          }

          return true;
        },
        requirementType: {
          '': 'string',
          step: 'string',
          base: 'number'
        },
        priority: 256
      },
      pattern: {
        validateString: function validateString(value, regexp) {
          if (!value) return true; // Builtin validators all accept empty strings, except `required` of course

          return regexp.test(value);
        },
        requirementType: 'regexp',
        priority: 64
      },
      minlength: {
        validateString: function validateString(value, requirement) {
          if (!value) return true; // Builtin validators all accept empty strings, except `required` of course

          return value.length >= requirement;
        },
        requirementType: 'integer',
        priority: 30
      },
      maxlength: {
        validateString: function validateString(value, requirement) {
          return value.length <= requirement;
        },
        requirementType: 'integer',
        priority: 30
      },
      length: {
        validateString: function validateString(value, min, max) {
          if (!value) return true; // Builtin validators all accept empty strings, except `required` of course

          return value.length >= min && value.length <= max;
        },
        requirementType: ['integer', 'integer'],
        priority: 30
      },
      mincheck: {
        validateMultiple: function validateMultiple(values, requirement) {
          return values.length >= requirement;
        },
        requirementType: 'integer',
        priority: 30
      },
      maxcheck: {
        validateMultiple: function validateMultiple(values, requirement) {
          return values.length <= requirement;
        },
        requirementType: 'integer',
        priority: 30
      },
      check: {
        validateMultiple: function validateMultiple(values, min, max) {
          return values.length >= min && values.length <= max;
        },
        requirementType: ['integer', 'integer'],
        priority: 30
      },
      min: comparisonOperator(function (value, requirement) {
        return value >= requirement;
      }),
      max: comparisonOperator(function (value, requirement) {
        return value <= requirement;
      }),
      range: comparisonOperator(function (value, min, max) {
        return value >= min && value <= max;
      }),
      equalto: {
        validateString: function validateString(value, refOrValue) {
          if (!value) return true; // Builtin validators all accept empty strings, except `required` of course

          var $reference = $(refOrValue);
          if ($reference.length) return value === $reference.val();else return value === refOrValue;
        },
        priority: 256
      },
      euvatin: {
        validateString: function validateString(value, refOrValue) {
          if (!value) {
            return true; // Builtin validators all accept empty strings, except `required` of course
          }

          var re = /^[A-Z][A-Z][A-Za-z0-9 -]{2,}$/;
          return re.test(value);
        },
        priority: 30
      }
    }
  };

  var UI = {};

  var diffResults = function diffResults(newResult, oldResult, deep) {
    var added = [];
    var kept = [];

    for (var i = 0; i < newResult.length; i++) {
      var found = false;

      for (var j = 0; j < oldResult.length; j++) {
        if (newResult[i].assert.name === oldResult[j].assert.name) {
          found = true;
          break;
        }
      }

      if (found) kept.push(newResult[i]);else added.push(newResult[i]);
    }

    return {
      kept: kept,
      added: added,
      removed: !deep ? diffResults(oldResult, newResult, true).added : []
    };
  };

  UI.Form = {
    _actualizeTriggers: function _actualizeTriggers() {
      var _this = this;

      this.$element.on('submit.Parsley', function (evt) {
        _this.onSubmitValidate(evt);
      });
      this.$element.on('click.Parsley', Utils._SubmitSelector, function (evt) {
        _this.onSubmitButton(evt);
      }); // UI could be disabled

      if (false === this.options.uiEnabled) return;
      this.element.setAttribute('novalidate', '');
    },
    focus: function focus() {
      this._focusedField = null;
      if (true === this.validationResult || 'none' === this.options.focus) return null;

      for (var i = 0; i < this.fields.length; i++) {
        var field = this.fields[i];

        if (true !== field.validationResult && field.validationResult.length > 0 && 'undefined' === typeof field.options.noFocus) {
          this._focusedField = field.$element;
          if ('first' === this.options.focus) break;
        }
      }

      if (null === this._focusedField) return null;
      return this._focusedField.focus();
    },
    _destroyUI: function _destroyUI() {
      // Reset all event listeners
      this.$element.off('.Parsley');
    }
  };
  UI.Field = {
    _reflowUI: function _reflowUI() {
      this._buildUI(); // If this field doesn't have an active UI don't bother doing something


      if (!this._ui) return; // Diff between two validation results

      var diff = diffResults(this.validationResult, this._ui.lastValidationResult); // Then store current validation result for next reflow

      this._ui.lastValidationResult = this.validationResult; // Handle valid / invalid / none field class

      this._manageStatusClass(); // Add, remove, updated errors messages


      this._manageErrorsMessages(diff); // Triggers impl


      this._actualizeTriggers(); // If field is not valid for the first time, bind keyup trigger to ease UX and quickly inform user


      if ((diff.kept.length || diff.added.length) && !this._failedOnce) {
        this._failedOnce = true;

        this._actualizeTriggers();
      }
    },
    // Returns an array of field's error message(s)
    getErrorsMessages: function getErrorsMessages() {
      // No error message, field is valid
      if (true === this.validationResult) return [];
      var messages = [];

      for (var i = 0; i < this.validationResult.length; i++) {
        messages.push(this.validationResult[i].errorMessage || this._getErrorMessage(this.validationResult[i].assert));
      }

      return messages;
    },
    // It's a goal of Parsley that this method is no longer required [#1073]
    addError: function addError(name) {
      var _ref = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {},
          message = _ref.message,
          assert = _ref.assert,
          _ref$updateClass = _ref.updateClass,
          updateClass = _ref$updateClass === void 0 ? true : _ref$updateClass;

      this._buildUI();

      this._addError(name, {
        message: message,
        assert: assert
      });

      if (updateClass) this._errorClass();
    },
    // It's a goal of Parsley that this method is no longer required [#1073]
    updateError: function updateError(name) {
      var _ref2 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {},
          message = _ref2.message,
          assert = _ref2.assert,
          _ref2$updateClass = _ref2.updateClass,
          updateClass = _ref2$updateClass === void 0 ? true : _ref2$updateClass;

      this._buildUI();

      this._updateError(name, {
        message: message,
        assert: assert
      });

      if (updateClass) this._errorClass();
    },
    // It's a goal of Parsley that this method is no longer required [#1073]
    removeError: function removeError(name) {
      var _ref3 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {},
          _ref3$updateClass = _ref3.updateClass,
          updateClass = _ref3$updateClass === void 0 ? true : _ref3$updateClass;

      this._buildUI();

      this._removeError(name); // edge case possible here: remove a standard Parsley error that is still failing in this.validationResult
      // but highly improbable cuz' manually removing a well Parsley handled error makes no sense.


      if (updateClass) this._manageStatusClass();
    },
    _manageStatusClass: function _manageStatusClass() {
      if (this.hasConstraints() && this.needsValidation() && true === this.validationResult) this._successClass();else if (this.validationResult.length > 0) this._errorClass();else this._resetClass();
    },
    _manageErrorsMessages: function _manageErrorsMessages(diff) {
      if ('undefined' !== typeof this.options.errorsMessagesDisabled) return; // Case where we have errorMessage option that configure an unique field error message, regardless failing validators

      if ('undefined' !== typeof this.options.errorMessage) {
        if (diff.added.length || diff.kept.length) {
          this._insertErrorWrapper();

          if (0 === this._ui.$errorsWrapper.find('.parsley-custom-error-message').length) this._ui.$errorsWrapper.append($(this.options.errorTemplate).addClass('parsley-custom-error-message'));

          this._ui.$errorClassHandler.attr('aria-describedby', this._ui.errorsWrapperId);

          return this._ui.$errorsWrapper.addClass('filled').attr('aria-hidden', 'false').find('.parsley-custom-error-message').html(this.options.errorMessage);
        }

        this._ui.$errorClassHandler.removeAttr('aria-describedby');

        return this._ui.$errorsWrapper.removeClass('filled').attr('aria-hidden', 'true').find('.parsley-custom-error-message').remove();
      } // Show, hide, update failing constraints messages


      for (var i = 0; i < diff.removed.length; i++) {
        this._removeError(diff.removed[i].assert.name);
      }

      for (i = 0; i < diff.added.length; i++) {
        this._addError(diff.added[i].assert.name, {
          message: diff.added[i].errorMessage,
          assert: diff.added[i].assert
        });
      }

      for (i = 0; i < diff.kept.length; i++) {
        this._updateError(diff.kept[i].assert.name, {
          message: diff.kept[i].errorMessage,
          assert: diff.kept[i].assert
        });
      }
    },
    _addError: function _addError(name, _ref4) {
      var message = _ref4.message,
          assert = _ref4.assert;

      this._insertErrorWrapper();

      this._ui.$errorClassHandler.attr('aria-describedby', this._ui.errorsWrapperId);

      this._ui.$errorsWrapper.addClass('filled').attr('aria-hidden', 'false').append($(this.options.errorTemplate).addClass('parsley-' + name).html(message || this._getErrorMessage(assert)));
    },
    _updateError: function _updateError(name, _ref5) {
      var message = _ref5.message,
          assert = _ref5.assert;

      this._ui.$errorsWrapper.addClass('filled').find('.parsley-' + name).html(message || this._getErrorMessage(assert));
    },
    _removeError: function _removeError(name) {
      this._ui.$errorClassHandler.removeAttr('aria-describedby');

      this._ui.$errorsWrapper.removeClass('filled').attr('aria-hidden', 'true').find('.parsley-' + name).remove();
    },
    _getErrorMessage: function _getErrorMessage(constraint) {
      var customConstraintErrorMessage = constraint.name + 'Message';
      if ('undefined' !== typeof this.options[customConstraintErrorMessage]) return window.Parsley.formatMessage(this.options[customConstraintErrorMessage], constraint.requirements);
      return window.Parsley.getErrorMessage(constraint);
    },
    _buildUI: function _buildUI() {
      // UI could be already built or disabled
      if (this._ui || false === this.options.uiEnabled) return;
      var _ui = {}; // Give field its Parsley id in DOM

      this.element.setAttribute(this.options.namespace + 'id', this.__id__);
      /** Generate important UI elements and store them in this **/
      // $errorClassHandler is the $element that woul have parsley-error and parsley-success classes

      _ui.$errorClassHandler = this._manageClassHandler(); // $errorsWrapper is a div that would contain the various field errors, it will be appended into $errorsContainer

      _ui.errorsWrapperId = 'parsley-id-' + (this.options.multiple ? 'multiple-' + this.options.multiple : this.__id__);
      _ui.$errorsWrapper = $(this.options.errorsWrapper).attr('id', _ui.errorsWrapperId); // ValidationResult UI storage to detect what have changed bwt two validations, and update DOM accordingly

      _ui.lastValidationResult = [];
      _ui.validationInformationVisible = false; // Store it in this for later

      this._ui = _ui;
    },
    // Determine which element will have `parsley-error` and `parsley-success` classes
    _manageClassHandler: function _manageClassHandler() {
      // Class handled could also be determined by function given in Parsley options
      if ('string' === typeof this.options.classHandler && $(this.options.classHandler).length) return $(this.options.classHandler); // Class handled could also be determined by function given in Parsley options

      var $handlerFunction = this.options.classHandler; // It might also be the function name of a global function

      if ('string' === typeof this.options.classHandler && 'function' === typeof window[this.options.classHandler]) $handlerFunction = window[this.options.classHandler];

      if ('function' === typeof $handlerFunction) {
        var $handler = $handlerFunction.call(this, this); // If this function returned a valid existing DOM element, go for it

        if ('undefined' !== typeof $handler && $handler.length) return $handler;
      } else if ('object' === _typeof($handlerFunction) && $handlerFunction instanceof jQuery && $handlerFunction.length) {
        return $handlerFunction;
      } else if ($handlerFunction) {
        Utils.warn('The class handler `' + $handlerFunction + '` does not exist in DOM nor as a global JS function');
      }

      return this._inputHolder();
    },
    _inputHolder: function _inputHolder() {
      // if simple element (input, texatrea, select...) it will perfectly host the classes and precede the error container
      if (!this.options.multiple || this.element.nodeName === 'SELECT') return this.$element; // But if multiple element (radio, checkbox), that would be their parent

      return this.$element.parent();
    },
    _insertErrorWrapper: function _insertErrorWrapper() {
      var $errorsContainer = this.options.errorsContainer; // Nothing to do if already inserted

      if (0 !== this._ui.$errorsWrapper.parent().length) return this._ui.$errorsWrapper.parent();

      if ('string' === typeof $errorsContainer) {
        if ($($errorsContainer).length) return $($errorsContainer).append(this._ui.$errorsWrapper);else if ('function' === typeof window[$errorsContainer]) $errorsContainer = window[$errorsContainer];else Utils.warn('The errors container `' + $errorsContainer + '` does not exist in DOM nor as a global JS function');
      }

      if ('function' === typeof $errorsContainer) $errorsContainer = $errorsContainer.call(this, this);
      if ('object' === _typeof($errorsContainer) && $errorsContainer.length) return $errorsContainer.append(this._ui.$errorsWrapper);
      return this._inputHolder().after(this._ui.$errorsWrapper);
    },
    _actualizeTriggers: function _actualizeTriggers() {
      var _this2 = this;

      var $toBind = this._findRelated();

      var trigger; // Remove Parsley events already bound on this field

      $toBind.off('.Parsley');
      if (this._failedOnce) $toBind.on(Utils.namespaceEvents(this.options.triggerAfterFailure, 'Parsley'), function () {
        _this2._validateIfNeeded();
      });else if (trigger = Utils.namespaceEvents(this.options.trigger, 'Parsley')) {
        $toBind.on(trigger, function (event) {
          _this2._validateIfNeeded(event);
        });
      }
    },
    _validateIfNeeded: function _validateIfNeeded(event) {
      var _this3 = this;

      // For keyup, keypress, keydown, input... events that could be a little bit obstrusive
      // do not validate if val length < min threshold on first validation. Once field have been validated once and info
      // about success or failure have been displayed, always validate with this trigger to reflect every yalidation change.
      if (event && /key|input/.test(event.type)) if (!(this._ui && this._ui.validationInformationVisible) && this.getValue().length <= this.options.validationThreshold) return;

      if (this.options.debounce) {
        window.clearTimeout(this._debounced);
        this._debounced = window.setTimeout(function () {
          return _this3.validate();
        }, this.options.debounce);
      } else this.validate();
    },
    _resetUI: function _resetUI() {
      // Reset all event listeners
      this._failedOnce = false;

      this._actualizeTriggers(); // Nothing to do if UI never initialized for this field


      if ('undefined' === typeof this._ui) return; // Reset all errors' li

      this._ui.$errorsWrapper.removeClass('filled').children().remove(); // Reset validation class


      this._resetClass(); // Reset validation flags and last validation result


      this._ui.lastValidationResult = [];
      this._ui.validationInformationVisible = false;
    },
    _destroyUI: function _destroyUI() {
      this._resetUI();

      if ('undefined' !== typeof this._ui) this._ui.$errorsWrapper.remove();
      delete this._ui;
    },
    _successClass: function _successClass() {
      this._ui.validationInformationVisible = true;

      this._ui.$errorClassHandler.removeClass(this.options.errorClass).addClass(this.options.successClass);
    },
    _errorClass: function _errorClass() {
      this._ui.validationInformationVisible = true;

      this._ui.$errorClassHandler.removeClass(this.options.successClass).addClass(this.options.errorClass);
    },
    _resetClass: function _resetClass() {
      this._ui.$errorClassHandler.removeClass(this.options.successClass).removeClass(this.options.errorClass);
    }
  };

  var Form = function Form(element, domOptions, options) {
    this.__class__ = 'Form';
    this.element = element;
    this.$element = $(element);
    this.domOptions = domOptions;
    this.options = options;
    this.parent = window.Parsley;
    this.fields = [];
    this.validationResult = null;
  };

  var statusMapping = {
    pending: null,
    resolved: true,
    rejected: false
  };
  Form.prototype = {
    onSubmitValidate: function onSubmitValidate(event) {
      var _this = this;

      // This is a Parsley generated submit event, do not validate, do not prevent, simply exit and keep normal behavior
      if (true === event.parsley) return; // If we didn't come here through a submit button, use the first one in the form

      var submitSource = this._submitSource || this.$element.find(Utils._SubmitSelector)[0];
      this._submitSource = null;
      this.$element.find('.parsley-synthetic-submit-button').prop('disabled', true);
      if (submitSource && null !== submitSource.getAttribute('formnovalidate')) return;
      window.Parsley._remoteCache = {};
      var promise = this.whenValidate({
        event: event
      });

      if ('resolved' === promise.state() && false !== this._trigger('submit')) ; else {
        // Rejected or pending: cancel this submit
        event.stopImmediatePropagation();
        event.preventDefault();
        if ('pending' === promise.state()) promise.done(function () {
          _this._submit(submitSource);
        });
      }
    },
    onSubmitButton: function onSubmitButton(event) {
      this._submitSource = event.currentTarget;
    },
    // internal
    // _submit submits the form, this time without going through the validations.
    // Care must be taken to "fake" the actual submit button being clicked.
    _submit: function _submit(submitSource) {
      if (false === this._trigger('submit')) return; // Add submit button's data

      if (submitSource) {
        var $synthetic = this.$element.find('.parsley-synthetic-submit-button').prop('disabled', false);
        if (0 === $synthetic.length) $synthetic = $('<input class="parsley-synthetic-submit-button" type="hidden">').appendTo(this.$element);
        $synthetic.attr({
          name: submitSource.getAttribute('name'),
          value: submitSource.getAttribute('value')
        });
      }

      this.$element.trigger(_extends($.Event('submit'), {
        parsley: true
      }));
    },
    // Performs validation on fields while triggering events.
    // @returns `true` if all validations succeeds, `false`
    // if a failure is immediately detected, or `null`
    // if dependant on a promise.
    // Consider using `whenValidate` instead.
    validate: function validate(options) {
      if (arguments.length >= 1 && !$.isPlainObject(options)) {
        Utils.warnOnce('Calling validate on a parsley form without passing arguments as an object is deprecated.');

        var _arguments = Array.prototype.slice.call(arguments),
            group = _arguments[0],
            force = _arguments[1],
            event = _arguments[2];

        options = {
          group: group,
          force: force,
          event: event
        };
      }

      return statusMapping[this.whenValidate(options).state()];
    },
    whenValidate: function whenValidate() {
      var _this2 = this,
          _Utils$all$done$fail$;

      var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          group = _ref.group,
          force = _ref.force,
          event = _ref.event;

      this.submitEvent = event;

      if (event) {
        this.submitEvent = _extends({}, event, {
          preventDefault: function preventDefault() {
            Utils.warnOnce("Using `this.submitEvent.preventDefault()` is deprecated; instead, call `this.validationResult = false`");
            _this2.validationResult = false;
          }
        });
      }

      this.validationResult = true; // fire validate event to eventually modify things before every validation

      this._trigger('validate'); // Refresh form DOM options and form's fields that could have changed


      this._refreshFields();

      var promises = this._withoutReactualizingFormOptions(function () {
        return $.map(_this2.fields, function (field) {
          return field.whenValidate({
            force: force,
            group: group
          });
        });
      });

      return (_Utils$all$done$fail$ = Utils.all(promises).done(function () {
        _this2._trigger('success');
      }).fail(function () {
        _this2.validationResult = false;

        _this2.focus();

        _this2._trigger('error');
      }).always(function () {
        _this2._trigger('validated');
      })).pipe.apply(_Utils$all$done$fail$, _toConsumableArray(this._pipeAccordingToValidationResult()));
    },
    // Iterate over refreshed fields, and stop on first failure.
    // Returns `true` if all fields are valid, `false` if a failure is detected
    // or `null` if the result depends on an unresolved promise.
    // Prefer using `whenValid` instead.
    isValid: function isValid(options) {
      if (arguments.length >= 1 && !$.isPlainObject(options)) {
        Utils.warnOnce('Calling isValid on a parsley form without passing arguments as an object is deprecated.');

        var _arguments2 = Array.prototype.slice.call(arguments),
            group = _arguments2[0],
            force = _arguments2[1];

        options = {
          group: group,
          force: force
        };
      }

      return statusMapping[this.whenValid(options).state()];
    },
    // Iterate over refreshed fields and validate them.
    // Returns a promise.
    // A validation that immediately fails will interrupt the validations.
    whenValid: function whenValid() {
      var _this3 = this;

      var _ref2 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          group = _ref2.group,
          force = _ref2.force;

      this._refreshFields();

      var promises = this._withoutReactualizingFormOptions(function () {
        return $.map(_this3.fields, function (field) {
          return field.whenValid({
            group: group,
            force: force
          });
        });
      });

      return Utils.all(promises);
    },
    refresh: function refresh() {
      this._refreshFields();

      return this;
    },
    // Reset UI
    reset: function reset() {
      // Form case: emit a reset event for each field
      for (var i = 0; i < this.fields.length; i++) {
        this.fields[i].reset();
      }

      this._trigger('reset');
    },
    // Destroy Parsley instance (+ UI)
    destroy: function destroy() {
      // Field case: emit destroy event to clean UI and then destroy stored instance
      this._destroyUI(); // Form case: destroy all its fields and then destroy stored instance


      for (var i = 0; i < this.fields.length; i++) {
        this.fields[i].destroy();
      }

      this.$element.removeData('Parsley');

      this._trigger('destroy');
    },
    _refreshFields: function _refreshFields() {
      return this.actualizeOptions()._bindFields();
    },
    _bindFields: function _bindFields() {
      var _this4 = this;

      var oldFields = this.fields;
      this.fields = [];
      this.fieldsMappedById = {};

      this._withoutReactualizingFormOptions(function () {
        _this4.$element.find(_this4.options.inputs).not(_this4.options.excluded).not("[".concat(_this4.options.namespace, "excluded=true]")).each(function (_, element) {
          var fieldInstance = new window.Parsley.Factory(element, {}, _this4); // Only add valid and not excluded `Field` and `FieldMultiple` children

          if ('Field' === fieldInstance.__class__ || 'FieldMultiple' === fieldInstance.__class__) {
            var uniqueId = fieldInstance.__class__ + '-' + fieldInstance.__id__;

            if ('undefined' === typeof _this4.fieldsMappedById[uniqueId]) {
              _this4.fieldsMappedById[uniqueId] = fieldInstance;

              _this4.fields.push(fieldInstance);
            }
          }
        });

        $.each(Utils.difference(oldFields, _this4.fields), function (_, field) {
          field.reset();
        });
      });

      return this;
    },
    // Internal only.
    // Looping on a form's fields to do validation or similar
    // will trigger reactualizing options on all of them, which
    // in turn will reactualize the form's options.
    // To avoid calling actualizeOptions so many times on the form
    // for nothing, _withoutReactualizingFormOptions temporarily disables
    // the method actualizeOptions on this form while `fn` is called.
    _withoutReactualizingFormOptions: function _withoutReactualizingFormOptions(fn) {
      var oldActualizeOptions = this.actualizeOptions;

      this.actualizeOptions = function () {
        return this;
      };

      var result = fn();
      this.actualizeOptions = oldActualizeOptions;
      return result;
    },
    // Internal only.
    // Shortcut to trigger an event
    // Returns true iff event is not interrupted and default not prevented.
    _trigger: function _trigger(eventName) {
      return this.trigger('form:' + eventName);
    }
  };

  var Constraint = function Constraint(parsleyField, name, requirements, priority, isDomConstraint) {
    var validatorSpec = window.Parsley._validatorRegistry.validators[name];
    var validator = new Validator(validatorSpec);
    priority = priority || parsleyField.options[name + 'Priority'] || validator.priority;
    isDomConstraint = true === isDomConstraint;

    _extends(this, {
      validator: validator,
      name: name,
      requirements: requirements,
      priority: priority,
      isDomConstraint: isDomConstraint
    });

    this._parseRequirements(parsleyField.options);
  };

  var capitalize = function capitalize(str) {
    var cap = str[0].toUpperCase();
    return cap + str.slice(1);
  };

  Constraint.prototype = {
    validate: function validate(value, instance) {
      var _this$validator;

      return (_this$validator = this.validator).validate.apply(_this$validator, [value].concat(_toConsumableArray(this.requirementList), [instance]));
    },
    _parseRequirements: function _parseRequirements(options) {
      var _this = this;

      this.requirementList = this.validator.parseRequirements(this.requirements, function (key) {
        return options[_this.name + capitalize(key)];
      });
    }
  };

  var Field = function Field(field, domOptions, options, parsleyFormInstance) {
    this.__class__ = 'Field';
    this.element = field;
    this.$element = $(field); // Set parent if we have one

    if ('undefined' !== typeof parsleyFormInstance) {
      this.parent = parsleyFormInstance;
    }

    this.options = options;
    this.domOptions = domOptions; // Initialize some properties

    this.constraints = [];
    this.constraintsByName = {};
    this.validationResult = true; // Bind constraints

    this._bindConstraints();
  };

  var statusMapping$1 = {
    pending: null,
    resolved: true,
    rejected: false
  };
  Field.prototype = {
    // # Public API
    // Validate field and trigger some events for mainly `UI`
    // @returns `true`, an array of the validators that failed, or
    // `null` if validation is not finished. Prefer using whenValidate
    validate: function validate(options) {
      if (arguments.length >= 1 && !$.isPlainObject(options)) {
        Utils.warnOnce('Calling validate on a parsley field without passing arguments as an object is deprecated.');
        options = {
          options: options
        };
      }

      var promise = this.whenValidate(options);
      if (!promise) // If excluded with `group` option
        return true;

      switch (promise.state()) {
        case 'pending':
          return null;

        case 'resolved':
          return true;

        case 'rejected':
          return this.validationResult;
      }
    },
    // Validate field and trigger some events for mainly `UI`
    // @returns a promise that succeeds only when all validations do
    // or `undefined` if field is not in the given `group`.
    whenValidate: function whenValidate() {
      var _this$whenValid$alway,
          _this = this;

      var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          force = _ref.force,
          group = _ref.group;

      // do not validate a field if not the same as given validation group
      this.refresh();
      if (group && !this._isInGroup(group)) return;
      this.value = this.getValue(); // Field Validate event. `this.value` could be altered for custom needs

      this._trigger('validate');

      return (_this$whenValid$alway = this.whenValid({
        force: force,
        value: this.value,
        _refreshed: true
      }).always(function () {
        _this._reflowUI();
      }).done(function () {
        _this._trigger('success');
      }).fail(function () {
        _this._trigger('error');
      }).always(function () {
        _this._trigger('validated');
      })).pipe.apply(_this$whenValid$alway, _toConsumableArray(this._pipeAccordingToValidationResult()));
    },
    hasConstraints: function hasConstraints() {
      return 0 !== this.constraints.length;
    },
    // An empty optional field does not need validation
    needsValidation: function needsValidation(value) {
      if ('undefined' === typeof value) value = this.getValue(); // If a field is empty and not required, it is valid
      // Except if `data-parsley-validate-if-empty` explicitely added, useful for some custom validators

      if (!value.length && !this._isRequired() && 'undefined' === typeof this.options.validateIfEmpty) return false;
      return true;
    },
    _isInGroup: function _isInGroup(group) {
      if (Array.isArray(this.options.group)) return -1 !== $.inArray(group, this.options.group);
      return this.options.group === group;
    },
    // Just validate field. Do not trigger any event.
    // Returns `true` iff all constraints pass, `false` if there are failures,
    // or `null` if the result can not be determined yet (depends on a promise)
    // See also `whenValid`.
    isValid: function isValid(options) {
      if (arguments.length >= 1 && !$.isPlainObject(options)) {
        Utils.warnOnce('Calling isValid on a parsley field without passing arguments as an object is deprecated.');

        var _arguments = Array.prototype.slice.call(arguments),
            force = _arguments[0],
            value = _arguments[1];

        options = {
          force: force,
          value: value
        };
      }

      var promise = this.whenValid(options);
      if (!promise) // Excluded via `group`
        return true;
      return statusMapping$1[promise.state()];
    },
    // Just validate field. Do not trigger any event.
    // @returns a promise that succeeds only when all validations do
    // or `undefined` if the field is not in the given `group`.
    // The argument `force` will force validation of empty fields.
    // If a `value` is given, it will be validated instead of the value of the input.
    whenValid: function whenValid() {
      var _this2 = this;

      var _ref2 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          _ref2$force = _ref2.force,
          force = _ref2$force === void 0 ? false : _ref2$force,
          value = _ref2.value,
          group = _ref2.group,
          _refreshed = _ref2._refreshed;

      // Recompute options and rebind constraints to have latest changes
      if (!_refreshed) this.refresh(); // do not validate a field if not the same as given validation group

      if (group && !this._isInGroup(group)) return;
      this.validationResult = true; // A field without constraint is valid

      if (!this.hasConstraints()) return $.when(); // Value could be passed as argument, needed to add more power to 'field:validate'

      if ('undefined' === typeof value || null === value) value = this.getValue();
      if (!this.needsValidation(value) && true !== force) return $.when();

      var groupedConstraints = this._getGroupedConstraints();

      var promises = [];
      $.each(groupedConstraints, function (_, constraints) {
        // Process one group of constraints at a time, we validate the constraints
        // and combine the promises together.
        var promise = Utils.all($.map(constraints, function (constraint) {
          return _this2._validateConstraint(value, constraint);
        }));
        promises.push(promise);
        if (promise.state() === 'rejected') return false; // Interrupt processing if a group has already failed
      });
      return Utils.all(promises);
    },
    // @returns a promise
    _validateConstraint: function _validateConstraint(value, constraint) {
      var _this3 = this;

      var result = constraint.validate(value, this); // Map false to a failed promise

      if (false === result) result = $.Deferred().reject(); // Make sure we return a promise and that we record failures

      return Utils.all([result]).fail(function (errorMessage) {
        if (!(_this3.validationResult instanceof Array)) _this3.validationResult = [];

        _this3.validationResult.push({
          assert: constraint,
          errorMessage: 'string' === typeof errorMessage && errorMessage
        });
      });
    },
    // @returns Parsley field computed value that could be overrided or configured in DOM
    getValue: function getValue() {
      var value; // Value could be overriden in DOM or with explicit options

      if ('function' === typeof this.options.value) value = this.options.value(this);else if ('undefined' !== typeof this.options.value) value = this.options.value;else value = this.$element.val(); // Handle wrong DOM or configurations

      if ('undefined' === typeof value || null === value) return '';
      return this._handleWhitespace(value);
    },
    // Reset UI
    reset: function reset() {
      this._resetUI();

      return this._trigger('reset');
    },
    // Destroy Parsley instance (+ UI)
    destroy: function destroy() {
      // Field case: emit destroy event to clean UI and then destroy stored instance
      this._destroyUI();

      this.$element.removeData('Parsley');
      this.$element.removeData('FieldMultiple');

      this._trigger('destroy');
    },
    // Actualize options and rebind constraints
    refresh: function refresh() {
      this._refreshConstraints();

      return this;
    },
    _refreshConstraints: function _refreshConstraints() {
      return this.actualizeOptions()._bindConstraints();
    },
    refreshConstraints: function refreshConstraints() {
      Utils.warnOnce("Parsley's refreshConstraints is deprecated. Please use refresh");
      return this.refresh();
    },

    /**
    * Add a new constraint to a field
    *
    * @param {String}   name
    * @param {Mixed}    requirements      optional
    * @param {Number}   priority          optional
    * @param {Boolean}  isDomConstraint   optional
    */
    addConstraint: function addConstraint(name, requirements, priority, isDomConstraint) {
      if (window.Parsley._validatorRegistry.validators[name]) {
        var constraint = new Constraint(this, name, requirements, priority, isDomConstraint); // if constraint already exist, delete it and push new version

        if ('undefined' !== this.constraintsByName[constraint.name]) this.removeConstraint(constraint.name);
        this.constraints.push(constraint);
        this.constraintsByName[constraint.name] = constraint;
      }

      return this;
    },
    // Remove a constraint
    removeConstraint: function removeConstraint(name) {
      for (var i = 0; i < this.constraints.length; i++) {
        if (name === this.constraints[i].name) {
          this.constraints.splice(i, 1);
          break;
        }
      }

      delete this.constraintsByName[name];
      return this;
    },
    // Update a constraint (Remove + re-add)
    updateConstraint: function updateConstraint(name, parameters, priority) {
      return this.removeConstraint(name).addConstraint(name, parameters, priority);
    },
    // # Internals
    // Internal only.
    // Bind constraints from config + options + DOM
    _bindConstraints: function _bindConstraints() {
      var constraints = [];
      var constraintsByName = {}; // clean all existing DOM constraints to only keep javascript user constraints

      for (var i = 0; i < this.constraints.length; i++) {
        if (false === this.constraints[i].isDomConstraint) {
          constraints.push(this.constraints[i]);
          constraintsByName[this.constraints[i].name] = this.constraints[i];
        }
      }

      this.constraints = constraints;
      this.constraintsByName = constraintsByName; // then re-add Parsley DOM-API constraints

      for (var name in this.options) {
        this.addConstraint(name, this.options[name], undefined, true);
      } // finally, bind special HTML5 constraints


      return this._bindHtml5Constraints();
    },
    // Internal only.
    // Bind specific HTML5 constraints to be HTML5 compliant
    _bindHtml5Constraints: function _bindHtml5Constraints() {
      // html5 required
      if (null !== this.element.getAttribute('required')) this.addConstraint('required', true, undefined, true); // html5 pattern

      if (null !== this.element.getAttribute('pattern')) this.addConstraint('pattern', this.element.getAttribute('pattern'), undefined, true); // range

      var min = this.element.getAttribute('min');
      var max = this.element.getAttribute('max');
      if (null !== min && null !== max) this.addConstraint('range', [min, max], undefined, true); // HTML5 min
      else if (null !== min) this.addConstraint('min', min, undefined, true); // HTML5 max
        else if (null !== max) this.addConstraint('max', max, undefined, true); // length

      if (null !== this.element.getAttribute('minlength') && null !== this.element.getAttribute('maxlength')) this.addConstraint('length', [this.element.getAttribute('minlength'), this.element.getAttribute('maxlength')], undefined, true); // HTML5 minlength
      else if (null !== this.element.getAttribute('minlength')) this.addConstraint('minlength', this.element.getAttribute('minlength'), undefined, true); // HTML5 maxlength
        else if (null !== this.element.getAttribute('maxlength')) this.addConstraint('maxlength', this.element.getAttribute('maxlength'), undefined, true); // html5 types

      var type = Utils.getType(this.element); // Small special case here for HTML5 number: integer validator if step attribute is undefined or an integer value, number otherwise

      if ('number' === type) {
        return this.addConstraint('type', ['number', {
          step: this.element.getAttribute('step') || '1',
          base: min || this.element.getAttribute('value')
        }], undefined, true); // Regular other HTML5 supported types
      } else if (/^(email|url|range|date)$/i.test(type)) {
        return this.addConstraint('type', type, undefined, true);
      }

      return this;
    },
    // Internal only.
    // Field is required if have required constraint without `false` value
    _isRequired: function _isRequired() {
      if ('undefined' === typeof this.constraintsByName.required) return false;
      return false !== this.constraintsByName.required.requirements;
    },
    // Internal only.
    // Shortcut to trigger an event
    _trigger: function _trigger(eventName) {
      return this.trigger('field:' + eventName);
    },
    // Internal only
    // Handles whitespace in a value
    // Use `data-parsley-whitespace="squish"` to auto squish input value
    // Use `data-parsley-whitespace="trim"` to auto trim input value
    _handleWhitespace: function _handleWhitespace(value) {
      if (true === this.options.trimValue) Utils.warnOnce('data-parsley-trim-value="true" is deprecated, please use data-parsley-whitespace="trim"');
      if ('squish' === this.options.whitespace) value = value.replace(/\s{2,}/g, ' ');
      if ('trim' === this.options.whitespace || 'squish' === this.options.whitespace || true === this.options.trimValue) value = Utils.trimString(value);
      return value;
    },
    _isDateInput: function _isDateInput() {
      var c = this.constraintsByName.type;
      return c && c.requirements === 'date';
    },
    // Internal only.
    // Returns the constraints, grouped by descending priority.
    // The result is thus an array of arrays of constraints.
    _getGroupedConstraints: function _getGroupedConstraints() {
      if (false === this.options.priorityEnabled) return [this.constraints];
      var groupedConstraints = [];
      var index = {}; // Create array unique of priorities

      for (var i = 0; i < this.constraints.length; i++) {
        var p = this.constraints[i].priority;
        if (!index[p]) groupedConstraints.push(index[p] = []);
        index[p].push(this.constraints[i]);
      } // Sort them by priority DESC


      groupedConstraints.sort(function (a, b) {
        return b[0].priority - a[0].priority;
      });
      return groupedConstraints;
    }
  };

  var Multiple = function Multiple() {
    this.__class__ = 'FieldMultiple';
  };

  Multiple.prototype = {
    // Add new `$element` sibling for multiple field
    addElement: function addElement($element) {
      this.$elements.push($element);
      return this;
    },
    // See `Field._refreshConstraints()`
    _refreshConstraints: function _refreshConstraints() {
      var fieldConstraints;
      this.constraints = []; // Select multiple special treatment

      if (this.element.nodeName === 'SELECT') {
        this.actualizeOptions()._bindConstraints();

        return this;
      } // Gather all constraints for each input in the multiple group


      for (var i = 0; i < this.$elements.length; i++) {
        // Check if element have not been dynamically removed since last binding
        if (!$('html').has(this.$elements[i]).length) {
          this.$elements.splice(i, 1);
          continue;
        }

        fieldConstraints = this.$elements[i].data('FieldMultiple')._refreshConstraints().constraints;

        for (var j = 0; j < fieldConstraints.length; j++) {
          this.addConstraint(fieldConstraints[j].name, fieldConstraints[j].requirements, fieldConstraints[j].priority, fieldConstraints[j].isDomConstraint);
        }
      }

      return this;
    },
    // See `Field.getValue()`
    getValue: function getValue() {
      // Value could be overriden in DOM
      if ('function' === typeof this.options.value) return this.options.value(this);else if ('undefined' !== typeof this.options.value) return this.options.value; // Radio input case

      if (this.element.nodeName === 'INPUT') {
        var type = Utils.getType(this.element);
        if (type === 'radio') return this._findRelated().filter(':checked').val() || ''; // checkbox input case

        if (type === 'checkbox') {
          var values = [];

          this._findRelated().filter(':checked').each(function () {
            values.push($(this).val());
          });

          return values;
        }
      } // Select multiple case


      if (this.element.nodeName === 'SELECT' && null === this.$element.val()) return []; // Default case that should never happen

      return this.$element.val();
    },
    _init: function _init() {
      this.$elements = [this.$element];
      return this;
    }
  };

  var Factory = function Factory(element, options, parsleyFormInstance) {
    this.element = element;
    this.$element = $(element); // If the element has already been bound, returns its saved Parsley instance

    var savedparsleyFormInstance = this.$element.data('Parsley');

    if (savedparsleyFormInstance) {
      // If the saved instance has been bound without a Form parent and there is one given in this call, add it
      if ('undefined' !== typeof parsleyFormInstance && savedparsleyFormInstance.parent === window.Parsley) {
        savedparsleyFormInstance.parent = parsleyFormInstance;

        savedparsleyFormInstance._resetOptions(savedparsleyFormInstance.options);
      }

      if ('object' === _typeof(options)) {
        _extends(savedparsleyFormInstance.options, options);
      }

      return savedparsleyFormInstance;
    } // Parsley must be instantiated with a DOM element or jQuery $element


    if (!this.$element.length) throw new Error('You must bind Parsley on an existing element.');
    if ('undefined' !== typeof parsleyFormInstance && 'Form' !== parsleyFormInstance.__class__) throw new Error('Parent instance must be a Form instance');
    this.parent = parsleyFormInstance || window.Parsley;
    return this.init(options);
  };

  Factory.prototype = {
    init: function init(options) {
      this.__class__ = 'Parsley';
      this.__version__ = '2.9.2';
      this.__id__ = Utils.generateID(); // Pre-compute options

      this._resetOptions(options); // A Form instance is obviously a `<form>` element but also every node that is not an input and has the `data-parsley-validate` attribute


      if (this.element.nodeName === 'FORM' || Utils.checkAttr(this.element, this.options.namespace, 'validate') && !this.$element.is(this.options.inputs)) return this.bind('parsleyForm'); // Every other element is bound as a `Field` or `FieldMultiple`

      return this.isMultiple() ? this.handleMultiple() : this.bind('parsleyField');
    },
    isMultiple: function isMultiple() {
      var type = Utils.getType(this.element);
      return type === 'radio' || type === 'checkbox' || this.element.nodeName === 'SELECT' && null !== this.element.getAttribute('multiple');
    },
    // Multiples fields are a real nightmare :(
    // Maybe some refactoring would be appreciated here...
    handleMultiple: function handleMultiple() {
      var _this = this;

      var name;
      var parsleyMultipleInstance; // Handle multiple name

      this.options.multiple = this.options.multiple || (name = this.element.getAttribute('name')) || this.element.getAttribute('id'); // Special select multiple input

      if (this.element.nodeName === 'SELECT' && null !== this.element.getAttribute('multiple')) {
        this.options.multiple = this.options.multiple || this.__id__;
        return this.bind('parsleyFieldMultiple'); // Else for radio / checkboxes, we need a `name` or `data-parsley-multiple` to properly bind it
      } else if (!this.options.multiple) {
        Utils.warn('To be bound by Parsley, a radio, a checkbox and a multiple select input must have either a name or a multiple option.', this.$element);
        return this;
      } // Remove special chars


      this.options.multiple = this.options.multiple.replace(/(:|\.|\[|\]|\{|\}|\$)/g, ''); // Add proper `data-parsley-multiple` to siblings if we have a valid multiple name

      if (name) {
        $('input[name="' + name + '"]').each(function (i, input) {
          var type = Utils.getType(input);
          if (type === 'radio' || type === 'checkbox') input.setAttribute(_this.options.namespace + 'multiple', _this.options.multiple);
        });
      } // Check here if we don't already have a related multiple instance saved


      var $previouslyRelated = this._findRelated();

      for (var i = 0; i < $previouslyRelated.length; i++) {
        parsleyMultipleInstance = $($previouslyRelated.get(i)).data('Parsley');

        if ('undefined' !== typeof parsleyMultipleInstance) {
          if (!this.$element.data('FieldMultiple')) {
            parsleyMultipleInstance.addElement(this.$element);
          }

          break;
        }
      } // Create a secret Field instance for every multiple field. It will be stored in `data('FieldMultiple')`
      // And will be useful later to access classic `Field` stuff while being in a `FieldMultiple` instance


      this.bind('parsleyField', true);
      return parsleyMultipleInstance || this.bind('parsleyFieldMultiple');
    },
    // Return proper `Form`, `Field` or `FieldMultiple`
    bind: function bind(type, doNotStore) {
      var parsleyInstance;

      switch (type) {
        case 'parsleyForm':
          parsleyInstance = $.extend(new Form(this.element, this.domOptions, this.options), new Base(), window.ParsleyExtend)._bindFields();
          break;

        case 'parsleyField':
          parsleyInstance = $.extend(new Field(this.element, this.domOptions, this.options, this.parent), new Base(), window.ParsleyExtend);
          break;

        case 'parsleyFieldMultiple':
          parsleyInstance = $.extend(new Field(this.element, this.domOptions, this.options, this.parent), new Multiple(), new Base(), window.ParsleyExtend)._init();
          break;

        default:
          throw new Error(type + 'is not a supported Parsley type');
      }

      if (this.options.multiple) Utils.setAttr(this.element, this.options.namespace, 'multiple', this.options.multiple);

      if ('undefined' !== typeof doNotStore) {
        this.$element.data('FieldMultiple', parsleyInstance);
        return parsleyInstance;
      } // Store the freshly bound instance in a DOM element for later access using jQuery `data()`


      this.$element.data('Parsley', parsleyInstance); // Tell the world we have a new Form or Field instance!

      parsleyInstance._actualizeTriggers();

      parsleyInstance._trigger('init');

      return parsleyInstance;
    }
  };

  var vernums = $.fn.jquery.split('.');

  if (parseInt(vernums[0]) <= 1 && parseInt(vernums[1]) < 8) {
    throw "The loaded version of jQuery is too old. Please upgrade to 1.8.x or better.";
  }

  if (!vernums.forEach) {
    Utils.warn('Parsley requires ES5 to run properly. Please include https://github.com/es-shims/es5-shim');
  } // Inherit `on`, `off` & `trigger` to Parsley:


  var Parsley = _extends(new Base(), {
    element: document,
    $element: $(document),
    actualizeOptions: null,
    _resetOptions: null,
    Factory: Factory,
    version: '2.9.2'
  }); // Supplement Field and Form with Base
  // This way, the constructors will have access to those methods


  _extends(Field.prototype, UI.Field, Base.prototype);

  _extends(Form.prototype, UI.Form, Base.prototype); // Inherit actualizeOptions and _resetOptions:


  _extends(Factory.prototype, Base.prototype); // ### jQuery API
  // `$('.elem').parsley(options)` or `$('.elem').psly(options)`


  $.fn.parsley = $.fn.psly = function (options) {
    if (this.length > 1) {
      var instances = [];
      this.each(function () {
        instances.push($(this).parsley(options));
      });
      return instances;
    } // Return undefined if applied to non existing DOM element


    if (this.length == 0) {
      return;
    }

    return new Factory(this[0], options);
  }; // ### Field and Form extension
  // Ensure the extension is now defined if it wasn't previously


  if ('undefined' === typeof window.ParsleyExtend) window.ParsleyExtend = {}; // ### Parsley config
  // Inherit from ParsleyDefault, and copy over any existing values

  Parsley.options = _extends(Utils.objectCreate(Defaults), window.ParsleyConfig);
  window.ParsleyConfig = Parsley.options; // Old way of accessing global options
  // ### Globals

  window.Parsley = window.psly = Parsley;
  Parsley.Utils = Utils;
  window.ParsleyUtils = {};
  $.each(Utils, function (key, value) {
    if ('function' === typeof value) {
      window.ParsleyUtils[key] = function () {
        Utils.warnOnce('Accessing `window.ParsleyUtils` is deprecated. Use `window.Parsley.Utils` instead.');
        return Utils[key].apply(Utils, arguments);
      };
    }
  }); // ### Define methods that forward to the registry, and deprecate all access except through window.Parsley

  var registry = window.Parsley._validatorRegistry = new ValidatorRegistry(window.ParsleyConfig.validators, window.ParsleyConfig.i18n);
  window.ParsleyValidator = {};
  $.each('setLocale addCatalog addMessage addMessages getErrorMessage formatMessage addValidator updateValidator removeValidator hasValidator'.split(' '), function (i, method) {
    window.Parsley[method] = function () {
      return registry[method].apply(registry, arguments);
    };

    window.ParsleyValidator[method] = function () {
      var _window$Parsley;

      Utils.warnOnce("Accessing the method '".concat(method, "' through Validator is deprecated. Simply call 'window.Parsley.").concat(method, "(...)'"));
      return (_window$Parsley = window.Parsley)[method].apply(_window$Parsley, arguments);
    };
  }); // ### UI
  // Deprecated global object

  window.Parsley.UI = UI;
  window.ParsleyUI = {
    removeError: function removeError(instance, name, doNotUpdateClass) {
      var updateClass = true !== doNotUpdateClass;
      Utils.warnOnce("Accessing UI is deprecated. Call 'removeError' on the instance directly. Please comment in issue 1073 as to your need to call this method.");
      return instance.removeError(name, {
        updateClass: updateClass
      });
    },
    getErrorsMessages: function getErrorsMessages(instance) {
      Utils.warnOnce("Accessing UI is deprecated. Call 'getErrorsMessages' on the instance directly.");
      return instance.getErrorsMessages();
    }
  };
  $.each('addError updateError'.split(' '), function (i, method) {
    window.ParsleyUI[method] = function (instance, name, message, assert, doNotUpdateClass) {
      var updateClass = true !== doNotUpdateClass;
      Utils.warnOnce("Accessing UI is deprecated. Call '".concat(method, "' on the instance directly. Please comment in issue 1073 as to your need to call this method."));
      return instance[method](name, {
        message: message,
        assert: assert,
        updateClass: updateClass
      });
    };
  }); // ### PARSLEY auto-binding
  // Prevent it by setting `ParsleyConfig.autoBind` to `false`

  if (false !== window.ParsleyConfig.autoBind) {
    $(function () {
      // Works only on `data-parsley-validate`.
      if ($('[data-parsley-validate]').length) $('[data-parsley-validate]').parsley();
    });
  }

  var o = $({});

  var deprecated = function deprecated() {
    Utils.warnOnce("Parsley's pubsub module is deprecated; use the 'on' and 'off' methods on parsley instances or window.Parsley");
  }; // Returns an event handler that calls `fn` with the arguments it expects


  function adapt(fn, context) {
    // Store to allow unbinding
    if (!fn.parsleyAdaptedCallback) {
      fn.parsleyAdaptedCallback = function () {
        var args = Array.prototype.slice.call(arguments, 0);
        args.unshift(this);
        fn.apply(context || o, args);
      };
    }

    return fn.parsleyAdaptedCallback;
  }

  var eventPrefix = 'parsley:'; // Converts 'parsley:form:validate' into 'form:validate'

  function eventName(name) {
    if (name.lastIndexOf(eventPrefix, 0) === 0) return name.substr(eventPrefix.length);
    return name;
  } // $.listen is deprecated. Use Parsley.on instead.


  $.listen = function (name, callback) {
    var context;
    deprecated();

    if ('object' === _typeof(arguments[1]) && 'function' === typeof arguments[2]) {
      context = arguments[1];
      callback = arguments[2];
    }

    if ('function' !== typeof callback) throw new Error('Wrong parameters');
    window.Parsley.on(eventName(name), adapt(callback, context));
  };

  $.listenTo = function (instance, name, fn) {
    deprecated();
    if (!(instance instanceof Field) && !(instance instanceof Form)) throw new Error('Must give Parsley instance');
    if ('string' !== typeof name || 'function' !== typeof fn) throw new Error('Wrong parameters');
    instance.on(eventName(name), adapt(fn));
  };

  $.unsubscribe = function (name, fn) {
    deprecated();
    if ('string' !== typeof name || 'function' !== typeof fn) throw new Error('Wrong arguments');
    window.Parsley.off(eventName(name), fn.parsleyAdaptedCallback);
  };

  $.unsubscribeTo = function (instance, name) {
    deprecated();
    if (!(instance instanceof Field) && !(instance instanceof Form)) throw new Error('Must give Parsley instance');
    instance.off(eventName(name));
  };

  $.unsubscribeAll = function (name) {
    deprecated();
    window.Parsley.off(eventName(name));
    $('form,input,textarea,select').each(function () {
      var instance = $(this).data('Parsley');

      if (instance) {
        instance.off(eventName(name));
      }
    });
  }; // $.emit is deprecated. Use jQuery events instead.


  $.emit = function (name, instance) {
    var _instance;

    deprecated();
    var instanceGiven = instance instanceof Field || instance instanceof Form;
    var args = Array.prototype.slice.call(arguments, instanceGiven ? 2 : 1);
    args.unshift(eventName(name));

    if (!instanceGiven) {
      instance = window.Parsley;
    }

    (_instance = instance).trigger.apply(_instance, _toConsumableArray(args));
  };

  $.extend(true, Parsley, {
    asyncValidators: {
      'default': {
        fn: function fn(xhr) {
          // By default, only status 2xx are deemed successful.
          // Note: we use status instead of state() because responses with status 200
          // but invalid messages (e.g. an empty body for content type set to JSON) will
          // result in state() === 'rejected'.
          return xhr.status >= 200 && xhr.status < 300;
        },
        url: false
      },
      reverse: {
        fn: function fn(xhr) {
          // If reverse option is set, a failing ajax request is considered successful
          return xhr.status < 200 || xhr.status >= 300;
        },
        url: false
      }
    },
    addAsyncValidator: function addAsyncValidator(name, fn, url, options) {
      Parsley.asyncValidators[name] = {
        fn: fn,
        url: url || false,
        options: options || {}
      };
      return this;
    }
  });
  Parsley.addValidator('remote', {
    requirementType: {
      '': 'string',
      'validator': 'string',
      'reverse': 'boolean',
      'options': 'object'
    },
    validateString: function validateString(value, url, options, instance) {
      var data = {};
      var ajaxOptions;
      var csr;
      var validator = options.validator || (true === options.reverse ? 'reverse' : 'default');
      if ('undefined' === typeof Parsley.asyncValidators[validator]) throw new Error('Calling an undefined async validator: `' + validator + '`');
      url = Parsley.asyncValidators[validator].url || url; // Fill current value

      if (url.indexOf('{value}') > -1) {
        url = url.replace('{value}', encodeURIComponent(value));
      } else {
        data[instance.element.getAttribute('name') || instance.element.getAttribute('id')] = value;
      } // Merge options passed in from the function with the ones in the attribute


      var remoteOptions = $.extend(true, options.options || {}, Parsley.asyncValidators[validator].options); // All `$.ajax(options)` could be overridden or extended directly from DOM in `data-parsley-remote-options`

      ajaxOptions = $.extend(true, {}, {
        url: url,
        data: data,
        type: 'GET'
      }, remoteOptions); // Generate store key based on ajax options

      instance.trigger('field:ajaxoptions', instance, ajaxOptions);
      csr = $.param(ajaxOptions); // Initialise querry cache

      if ('undefined' === typeof Parsley._remoteCache) Parsley._remoteCache = {}; // Try to retrieve stored xhr

      var xhr = Parsley._remoteCache[csr] = Parsley._remoteCache[csr] || $.ajax(ajaxOptions);

      var handleXhr = function handleXhr() {
        var result = Parsley.asyncValidators[validator].fn.call(instance, xhr, url, options);
        if (!result) // Map falsy results to rejected promise
          result = $.Deferred().reject();
        return $.when(result);
      };

      return xhr.then(handleXhr, handleXhr);
    },
    priority: -1
  });
  Parsley.on('form:submit', function () {
    Parsley._remoteCache = {};
  });

  Base.prototype.addAsyncValidator = function () {
    Utils.warnOnce('Accessing the method `addAsyncValidator` through an instance is deprecated. Simply call `Parsley.addAsyncValidator(...)`');
    return Parsley.addAsyncValidator.apply(Parsley, arguments);
  };

  // This is included with the Parsley library itself,
  Parsley.addMessages('en', {
    defaultMessage: "This value seems to be invalid.",
    type: {
      email: "This value should be a valid email.",
      url: "This value should be a valid url.",
      number: "This value should be a valid number.",
      integer: "This value should be a valid integer.",
      digits: "This value should be digits.",
      alphanum: "This value should be alphanumeric."
    },
    notblank: "This value should not be blank.",
    required: "This value is required.",
    pattern: "This value seems to be invalid.",
    min: "This value should be greater than or equal to %s.",
    max: "This value should be lower than or equal to %s.",
    range: "This value should be between %s and %s.",
    minlength: "This value is too short. It should have %s characters or more.",
    maxlength: "This value is too long. It should have %s characters or fewer.",
    length: "This value length is invalid. It should be between %s and %s characters long.",
    mincheck: "You must select at least %s choices.",
    maxcheck: "You must select %s choices or fewer.",
    check: "You must select between %s and %s choices.",
    equalto: "This value should be the same.",
    euvatin: "It's not a valid VAT Identification Number."
  });
  Parsley.setLocale('en');

  function InputEvent() {
    var _this = this;

    var globals = window || global; // Slightly odd way construct our object. This way methods are force bound.
    // Used to test for duplicate library.

    _extends(this, {
      // For browsers that do not support isTrusted, assumes event is native.
      isNativeEvent: function isNativeEvent(evt) {
        return evt.originalEvent && evt.originalEvent.isTrusted !== false;
      },
      fakeInputEvent: function fakeInputEvent(evt) {
        if (_this.isNativeEvent(evt)) {
          $(evt.target).trigger('input');
        }
      },
      misbehaves: function misbehaves(evt) {
        if (_this.isNativeEvent(evt)) {
          _this.behavesOk(evt);

          $(document).on('change.inputevent', evt.data.selector, _this.fakeInputEvent);

          _this.fakeInputEvent(evt);
        }
      },
      behavesOk: function behavesOk(evt) {
        if (_this.isNativeEvent(evt)) {
          $(document) // Simply unbinds the testing handler
          .off('input.inputevent', evt.data.selector, _this.behavesOk).off('change.inputevent', evt.data.selector, _this.misbehaves);
        }
      },
      // Bind the testing handlers
      install: function install() {
        if (globals.inputEventPatched) {
          return;
        }

        globals.inputEventPatched = '0.0.3';

        for (var _i = 0, _arr = ['select', 'input[type="checkbox"]', 'input[type="radio"]', 'input[type="file"]']; _i < _arr.length; _i++) {
          var selector = _arr[_i];
          $(document).on('input.inputevent', selector, {
            selector: selector
          }, _this.behavesOk).on('change.inputevent', selector, {
            selector: selector
          }, _this.misbehaves);
        }
      },
      uninstall: function uninstall() {
        delete globals.inputEventPatched;
        $(document).off('.inputevent');
      }
    });
  }
  var inputevent = new InputEvent();

  inputevent.install();

  return Parsley;

})));
//# sourceMappingURL=parsley.js.map

/* WEBPACK VAR INJECTION */}.call(exports, __webpack_require__(2)))

/***/ }),
/* 4 */,
/* 5 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * Add debounced global resize and scroll events
 *
 * @license APLv2
 *
 * @example
 * var globalEvents = require('./events.js');
 *
 * // Listen to debounced scroll event:
 * $(document}.on(globalEvents.scroll.key, function(event, originalEvent) {
 * 	console.log(originalEvent); // original scroll event
 * });
 */



var $ = __webpack_require__(0),
debounce = __webpack_require__(29),
$document = $(document),
mediaqueries = __webpack_require__(6),
interval = {
	resize: 50,
	scroll: 50
},
events = {
	mq: mediaqueries.event,
	resize: {
		key: 'debouncedresize.estatico',
		attach: function() {
			$(window).on('resize.estatico', debounce(function(event) {
				$document.triggerHandler(this.key, event);
			}.bind(this), interval.resize));
		}
	},
	scroll: {
		key: 'debouncedscroll.estatico',
		attach: function() {
			$(window).on('scroll.estatico', debounce(function(event) {
				$document.triggerHandler(this.key, event);
			}.bind(this), interval.scroll));
		}
	},
	formchange: {
		key: 'formchange.estatico',
		attach: function() {}
	},
	contentchange: {
		key: 'contentchange.estatico',
		attach: function() {}
	}
};

Object.keys(events).forEach(function(eventName) {
	events[eventName].attach();
});

module.exports = events;


/***/ }),
/* 6 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * Trigger custom events when changing breakpoint, get breakpoints from CSS properties
 *
 * @license APLv2
 *
 * @example
 * var mediaqueries = require('./mediaqueries.js');
 *
 * // Listen to custom (debounced) event to react to viewport changes:
 * $(document).on(mediaqueries.event.key, function(event, prevBreakpoint, currentBreakpoint) {
 * 	console.log(prevBreakpoint); // { name: "small", value: "768px" }
 * 	console.log(parseInt(prevBreakpoint.value)); // "768"
 * });
 *
 * // Check the current viewport against a specific breakpoint:
 * if (mediaqueries.query({ from: 'small' })) {
 * 	this.destroySmall();
 * 	this.initLarge();
 * }
 * // or
 * if (mediaqueries.query({ from: 'small', to: 'medium' })) {
 * 	this.destroySmall();
 * 	this.initMedium();
 * }
 */



var $ = __webpack_require__(0);

function parseCssProperty(str) {
	return JSON.parse($.trim(str.replace(/^('|")|(\\)|('|")$/g, '')));
}

var $document = $(document),
	key = 'estatico.mq',
	breakpointsString = window.getComputedStyle(document.body, ':after').getPropertyValue('content'),
	currentBreakpointString = window.getComputedStyle(document.body, ':before').getPropertyValue('content'),
	breakpoints = parseCssProperty(breakpointsString),
	currentBreakpoint = parseCssProperty(currentBreakpointString);

function getBreakpointValue(breakpoint) {
	var value = 0;

	if (typeof breakpoints[breakpoint] !== 'undefined') {
		value = parseInt(breakpoints[breakpoint], 10);
	} else {
		throw 'Breakpoint not found: "' + breakpoint + '"';
	}

	return value;
}

function query(options) {
	var breakpointFrom, breakpointTo,
		breakpointCurrent = parseInt(currentBreakpoint.value, 10);

	if (typeof options !== 'object') {
		// No or wrong arguments passed
		throw 'Illegal argument of type "' + typeof options + '", expected "object"';
	} else if (typeof options.to !== 'undefined' && typeof options.from !== 'undefined') {
		breakpointFrom = getBreakpointValue(options.from);
		breakpointTo = getBreakpointValue(options.to);

		// "from" cannot be larger than "to"
		if (breakpointFrom > breakpointTo) {
			throw 'Breakpoint ' + breakpointFrom + ' is larger than ' + breakpointTo + '';
		}

		// The breakpoint needs to smaller than the "to" (exclusive)
		// but larger or the same as "from" (inclusive)
		return breakpointFrom <= breakpointCurrent && breakpointCurrent < breakpointTo;
	} else if (typeof options.to !== 'undefined') {
		breakpointTo = getBreakpointValue(options.to);

		// Breakpoint needs to smaller than the "to" (exclusive)
		return breakpointCurrent < breakpointTo;
	} else if (typeof options.from !== 'undefined') {
		breakpointFrom = getBreakpointValue(options.from);

		// Breakpoint needs larger or the same as "from" (inclusive)
		return breakpointCurrent >= breakpointFrom;
	} else {
		throw 'No values for "to" or "from" received';
	}
}

/**
 * Based on breakpointCurrent pixel value return name of that breakpoint as string
 */
function getCurrentBreakpointName() {
	var ret,
		breakpointsNameArray = Object.keys(breakpoints);

	breakpointsNameArray.forEach(function(breakpointName) {
		var queryResult = query({ from: breakpointName});

		if (queryResult) {
			ret = breakpointName;
		}
	});

	return ret;
}

/**
 * triggers event
 */
function triggerBreakpointChange() {
	var breakpoint = parseCssProperty(window.getComputedStyle(document.body, ':before').getPropertyValue('content')),
		prevBreakpoint = currentBreakpoint;

	currentBreakpoint = breakpoint;

	// on iOS at least, we encountered that this method is triggered twice
	// per viewport change. If we rotate for example the iPhone from landscape
	// to portrait, it first triggers correctly the change from tabletPortrait
	// to mobilePortrait. But then, another trigger is made from mobilePortrait
	// to mobilePortrait, which is useless.
	// Since we might do extensive rendering  stuff on breakpoint changes we
	// must prevent the 2nd useless call here.
	// console.log(prevBreakpoint.name + ' -> ' prevBreakpoint.value);
	// console.log(currentBreakpoint.name + ' -> ' currentBreakpoint.value);
	if (prevBreakpoint.name === currentBreakpoint.name && prevBreakpoint.value === currentBreakpoint.value) {
		return;
	}

	$document.triggerHandler(key, [prevBreakpoint, breakpoint]);
}

/**
 * mql parameter is passed to query against different viewports
 */
/*function handleOrientationChange(mql) {
 triggerBreakpointChange(mql);
 }*/

/**
 * mql parameter is passed to query against different viewports
 */
function handleViewportSizeChange(mql) {
	triggerBreakpointChange(mql);
}

/**
 * Build queries list based on breakpoints array passed from css
 */
function buildMatchMediaList() {
	var minWidth = 0,
		maxWidth,
		width,
		mql,
		query;

	// Add listener for orientation change
	// mql = window.matchMedia('(orientation: landscape)');
	// mql.addListener(handleOrientationChange);

	$.each(breakpoints, function(element) {
		width = breakpoints[element];

		// query = '(min-width:' + minWidth + ')' + ' and ' + '(max-width:' + maxWidth + ')';
		query = '(min-width:' + width + ')';
		minWidth = maxWidth;
		mql = window.matchMedia(query);
		mql.addListener(handleViewportSizeChange);
	});
}

module.exports = {
	query: query,
	getCurrentBreakpointName: getCurrentBreakpointName,
	breakpoints: breakpoints,
	event: {
		key: key,
		attach: function() {
			buildMatchMediaList();
		}
	}
};


/***/ }),
/* 7 */
/***/ (function(module, exports) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */

/** Used for native method references */
var objectProto = Object.prototype;

/** Used to resolve the internal [[Class]] of values */
var toString = objectProto.toString;

/** Used to detect if a method is native */
var reNative = RegExp('^' +
  String(toString)
    .replace(/[.*+?^${}()|[\]\\]/g, '\\$&')
    .replace(/toString| for [^\]]+/g, '.*?') + '$'
);

/**
 * Checks if `value` is a native function.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if the `value` is a native function, else `false`.
 */
function isNative(value) {
  return typeof value == 'function' && reNative.test(value);
}

module.exports = isNative;


/***/ }),
/* 8 */
/***/ (function(module, exports, __webpack_require__) {

// Create a simple path alias to allow browserify to resolve
// the runtime on a supported path.
module.exports = __webpack_require__(71)['default'];


/***/ }),
/* 9 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var objectTypes = __webpack_require__(12);

/**
 * Checks if `value` is the language type of Object.
 * (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
 *
 * @static
 * @memberOf _
 * @category Objects
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if the `value` is an object, else `false`.
 * @example
 *
 * _.isObject({});
 * // => true
 *
 * _.isObject([1, 2, 3]);
 * // => true
 *
 * _.isObject(1);
 * // => false
 */
function isObject(value) {
  // check if the value is the ECMAScript language type of Object
  // http://es5.github.io/#x8
  // and avoid a V8 bug
  // http://code.google.com/p/v8/issues/detail?id=2291
  return !!(value && objectTypes[typeof value]);
}

module.exports = isObject;


/***/ }),
/* 10 */
/***/ (function(module, exports) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */

/**
 * Slices the `collection` from the `start` index up to, but not including,
 * the `end` index.
 *
 * Note: This function is used instead of `Array#slice` to support node lists
 * in IE < 9 and to ensure dense arrays are returned.
 *
 * @private
 * @param {Array|Object|string} collection The collection to slice.
 * @param {number} start The start index.
 * @param {number} end The end index.
 * @returns {Array} Returns the new array.
 */
function slice(array, start, end) {
  start || (start = 0);
  if (typeof end == 'undefined') {
    end = array ? array.length : 0;
  }
  var index = -1,
      length = end - start || 0,
      result = Array(length < 0 ? 0 : length);

  while (++index < length) {
    result[index] = array[start + index];
  }
  return result;
}

module.exports = slice;


/***/ }),
/* 11 */,
/* 12 */
/***/ (function(module, exports) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */

/** Used to determine if values are of the language type Object */
var objectTypes = {
  'boolean': false,
  'function': true,
  'object': true,
  'number': false,
  'string': false,
  'undefined': false
};

module.exports = objectTypes;


/***/ }),
/* 13 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var isNative = __webpack_require__(7),
    noop = __webpack_require__(18);

/** Used as the property descriptor for `__bindData__` */
var descriptor = {
  'configurable': false,
  'enumerable': false,
  'value': null,
  'writable': false
};

/** Used to set meta data on functions */
var defineProperty = (function() {
  // IE 8 only accepts DOM elements
  try {
    var o = {},
        func = isNative(func = Object.defineProperty) && func,
        result = func(o, o, o) && func;
  } catch(e) { }
  return result;
}());

/**
 * Sets `this` binding data on a given function.
 *
 * @private
 * @param {Function} func The function to set data on.
 * @param {Array} value The data array to set.
 */
var setBindData = !defineProperty ? noop : function(func, value) {
  descriptor.value = value;
  defineProperty(func, '__bindData__', descriptor);
};

module.exports = setBindData;


/***/ }),
/* 14 */
/***/ (function(module, exports, __webpack_require__) {

//     keymaster.js
//     (c) 2011-2013 Thomas Fuchs
//     keymaster.js may be freely distributed under the MIT license.

;(function(global){
  var k,
    _handlers = {},
    _mods = { 16: false, 18: false, 17: false, 91: false },
    _scope = 'all',
    // modifier keys
    _MODIFIERS = {
      '⇧': 16, shift: 16,
      '⌥': 18, alt: 18, option: 18,
      '⌃': 17, ctrl: 17, control: 17,
      '⌘': 91, command: 91
    },
    // special keys
    _MAP = {
      backspace: 8, tab: 9, clear: 12,
      enter: 13, 'return': 13,
      esc: 27, escape: 27, space: 32,
      left: 37, up: 38,
      right: 39, down: 40,
      del: 46, 'delete': 46,
      home: 36, end: 35,
      pageup: 33, pagedown: 34,
      ',': 188, '.': 190, '/': 191,
      '`': 192, '-': 189, '=': 187,
      ';': 186, '\'': 222,
      '[': 219, ']': 221, '\\': 220
    },
    code = function(x){
      return _MAP[x] || x.toUpperCase().charCodeAt(0);
    },
    _downKeys = [];

  for(k=1;k<20;k++) _MAP['f'+k] = 111+k;

  // IE doesn't support Array#indexOf, so have a simple replacement
  function index(array, item){
    var i = array.length;
    while(i--) if(array[i]===item) return i;
    return -1;
  }

  // for comparing mods before unassignment
  function compareArray(a1, a2) {
    if (a1.length != a2.length) return false;
    for (var i = 0; i < a1.length; i++) {
        if (a1[i] !== a2[i]) return false;
    }
    return true;
  }

  var modifierMap = {
      16:'shiftKey',
      18:'altKey',
      17:'ctrlKey',
      91:'metaKey'
  };
  function updateModifierKey(event) {
      for(k in _mods) _mods[k] = event[modifierMap[k]];
  };

  // handle keydown event
  function dispatch(event) {
    var key, handler, k, i, modifiersMatch, scope;
    key = event.keyCode;

    if (index(_downKeys, key) == -1) {
        _downKeys.push(key);
    }

    // if a modifier key, set the key.<modifierkeyname> property to true and return
    if(key == 93 || key == 224) key = 91; // right command on webkit, command on Gecko
    if(key in _mods) {
      _mods[key] = true;
      // 'assignKey' from inside this closure is exported to window.key
      for(k in _MODIFIERS) if(_MODIFIERS[k] == key) assignKey[k] = true;
      return;
    }
    updateModifierKey(event);

    // see if we need to ignore the keypress (filter() can can be overridden)
    // by default ignore key presses if a select, textarea, or input is focused
    if(!assignKey.filter.call(this, event)) return;

    // abort if no potentially matching shortcuts found
    if (!(key in _handlers)) return;

    scope = getScope();

    // for each potential shortcut
    for (i = 0; i < _handlers[key].length; i++) {
      handler = _handlers[key][i];

      // see if it's in the current scope
      if(handler.scope == scope || handler.scope == 'all'){
        // check if modifiers match if any
        modifiersMatch = handler.mods.length > 0;
        for(k in _mods)
          if((!_mods[k] && index(handler.mods, +k) > -1) ||
            (_mods[k] && index(handler.mods, +k) == -1)) modifiersMatch = false;
        // call the handler and stop the event if neccessary
        if((handler.mods.length == 0 && !_mods[16] && !_mods[18] && !_mods[17] && !_mods[91]) || modifiersMatch){
          if(handler.method(event, handler)===false){
            if(event.preventDefault) event.preventDefault();
              else event.returnValue = false;
            if(event.stopPropagation) event.stopPropagation();
            if(event.cancelBubble) event.cancelBubble = true;
          }
        }
      }
    }
  };

  // unset modifier keys on keyup
  function clearModifier(event){
    var key = event.keyCode, k,
        i = index(_downKeys, key);

    // remove key from _downKeys
    if (i >= 0) {
        _downKeys.splice(i, 1);
    }

    if(key == 93 || key == 224) key = 91;
    if(key in _mods) {
      _mods[key] = false;
      for(k in _MODIFIERS) if(_MODIFIERS[k] == key) assignKey[k] = false;
    }
  };

  function resetModifiers() {
    for(k in _mods) _mods[k] = false;
    for(k in _MODIFIERS) assignKey[k] = false;
  };

  // parse and assign shortcut
  function assignKey(key, scope, method){
    var keys, mods;
    keys = getKeys(key);
    if (method === undefined) {
      method = scope;
      scope = 'all';
    }

    // for each shortcut
    for (var i = 0; i < keys.length; i++) {
      // set modifier keys if any
      mods = [];
      key = keys[i].split('+');
      if (key.length > 1){
        mods = getMods(key);
        key = [key[key.length-1]];
      }
      // convert to keycode and...
      key = key[0]
      key = code(key);
      // ...store handler
      if (!(key in _handlers)) _handlers[key] = [];
      _handlers[key].push({ shortcut: keys[i], scope: scope, method: method, key: keys[i], mods: mods });
    }
  };

  // unbind all handlers for given key in current scope
  function unbindKey(key, scope) {
    var multipleKeys, keys,
      mods = [],
      i, j, obj;

    multipleKeys = getKeys(key);

    for (j = 0; j < multipleKeys.length; j++) {
      keys = multipleKeys[j].split('+');

      if (keys.length > 1) {
        mods = getMods(keys);
      }

      key = keys[keys.length - 1];
      key = code(key);

      if (scope === undefined) {
        scope = getScope();
      }
      if (!_handlers[key]) {
        return;
      }
      for (i = 0; i < _handlers[key].length; i++) {
        obj = _handlers[key][i];
        // only clear handlers if correct scope and mods match
        if (obj.scope === scope && compareArray(obj.mods, mods)) {
          _handlers[key][i] = {};
        }
      }
    }
  };

  // Returns true if the key with code 'keyCode' is currently down
  // Converts strings into key codes.
  function isPressed(keyCode) {
      if (typeof(keyCode)=='string') {
        keyCode = code(keyCode);
      }
      return index(_downKeys, keyCode) != -1;
  }

  function getPressedKeyCodes() {
      return _downKeys.slice(0);
  }

  function filter(event){
    var tagName = (event.target || event.srcElement).tagName;
    // ignore keypressed in any elements that support keyboard data input
    return !(tagName == 'INPUT' || tagName == 'SELECT' || tagName == 'TEXTAREA');
  }

  // initialize key.<modifier> to false
  for(k in _MODIFIERS) assignKey[k] = false;

  // set current scope (default 'all')
  function setScope(scope){ _scope = scope || 'all' };
  function getScope(){ return _scope || 'all' };

  // delete all handlers for a given scope
  function deleteScope(scope){
    var key, handlers, i;

    for (key in _handlers) {
      handlers = _handlers[key];
      for (i = 0; i < handlers.length; ) {
        if (handlers[i].scope === scope) handlers.splice(i, 1);
        else i++;
      }
    }
  };

  // abstract key logic for assign and unassign
  function getKeys(key) {
    var keys;
    key = key.replace(/\s/g, '');
    keys = key.split(',');
    if ((keys[keys.length - 1]) == '') {
      keys[keys.length - 2] += ',';
    }
    return keys;
  }

  // abstract mods logic for assign and unassign
  function getMods(key) {
    var mods = key.slice(0, key.length - 1);
    for (var mi = 0; mi < mods.length; mi++)
    mods[mi] = _MODIFIERS[mods[mi]];
    return mods;
  }

  // cross-browser events
  function addEvent(object, event, method) {
    if (object.addEventListener)
      object.addEventListener(event, method, false);
    else if(object.attachEvent)
      object.attachEvent('on'+event, function(){ method(window.event) });
  };

  // set the handlers globally on document
  addEvent(document, 'keydown', function(event) { dispatch(event) }); // Passing _scope to a callback to ensure it remains the same by execution. Fixes #48
  addEvent(document, 'keyup', clearModifier);

  // reset modifiers to false whenever the window is (re)focused.
  addEvent(window, 'focus', resetModifiers);

  // store previously defined key
  var previousKey = global.key;

  // restore previously defined key and return reference to our key object
  function noConflict() {
    var k = global.key;
    global.key = previousKey;
    return k;
  }

  // set window.key and window.key.set/get/deleteScope, and the default filter
  global.key = assignKey;
  global.key.setScope = setScope;
  global.key.getScope = getScope;
  global.key.deleteScope = deleteScope;
  global.key.filter = filter;
  global.key.isPressed = isPressed;
  global.key.getPressedKeyCodes = getPressedKeyCodes;
  global.key.noConflict = noConflict;
  global.key.unbind = unbindKey;

  if(true) module.exports = assignKey;

})(this);


/***/ }),
/* 15 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.3 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var baseIndexOf = __webpack_require__(36),
    forOwn = __webpack_require__(37),
    isArray = __webpack_require__(48),
    isString = __webpack_require__(49);

/* Native method shortcuts for methods with the same name as other `lodash` methods */
var nativeMax = Math.max;

/**
 * Checks if a given value is present in a collection using strict equality
 * for comparisons, i.e. `===`. If `fromIndex` is negative, it is used as the
 * offset from the end of the collection.
 *
 * @static
 * @memberOf _
 * @alias include
 * @category Collections
 * @param {Array|Object|string} collection The collection to iterate over.
 * @param {*} target The value to check for.
 * @param {number} [fromIndex=0] The index to search from.
 * @returns {boolean} Returns `true` if the `target` element is found, else `false`.
 * @example
 *
 * _.contains([1, 2, 3], 1);
 * // => true
 *
 * _.contains([1, 2, 3], 1, 2);
 * // => false
 *
 * _.contains({ 'name': 'fred', 'age': 40 }, 'fred');
 * // => true
 *
 * _.contains('pebbles', 'eb');
 * // => true
 */
function contains(collection, target, fromIndex) {
  var index = -1,
      indexOf = baseIndexOf,
      length = collection ? collection.length : 0,
      result = false;

  fromIndex = (fromIndex < 0 ? nativeMax(0, length + fromIndex) : fromIndex) || 0;
  if (isArray(collection)) {
    result = indexOf(collection, target, fromIndex) > -1;
  } else if (typeof length == 'number') {
    result = (isString(collection) ? collection.indexOf(target, fromIndex) : indexOf(collection, target, fromIndex)) > -1;
  } else {
    forOwn(collection, function(value) {
      if (++index >= fromIndex) {
        return !(result = value === target);
      }
    });
  }
  return result;
}

module.exports = contains;


/***/ }),
/* 16 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


/**
 * This helper accepts a url as parameter and returns a promise. After
 * successfully resolving the promise, it will return the contents of the
 * provided url.
 */

var $ = __webpack_require__(0);

module.exports = function(requestUrl, cache, data, method, contentType) {

	var httpMethod = method || 'GET',
		options = {
			url: requestUrl,
			cache: cache,
			data: data,
			method: httpMethod
		};

	if (contentType) {
		options.contentType = contentType;
	}

	return $.ajax(options);

};


/***/ }),
/* 17 */
/***/ (function(module, exports, __webpack_require__) {

/* WEBPACK VAR INJECTION */(function(global) {/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var isNative = __webpack_require__(7),
    isObject = __webpack_require__(9),
    noop = __webpack_require__(18);

/* Native method shortcuts for methods with the same name as other `lodash` methods */
var nativeCreate = isNative(nativeCreate = Object.create) && nativeCreate;

/**
 * The base implementation of `_.create` without support for assigning
 * properties to the created object.
 *
 * @private
 * @param {Object} prototype The object to inherit from.
 * @returns {Object} Returns the new object.
 */
function baseCreate(prototype, properties) {
  return isObject(prototype) ? nativeCreate(prototype) : {};
}
// fallback for browsers without `Object.create`
if (!nativeCreate) {
  baseCreate = (function() {
    function Object() {}
    return function(prototype) {
      if (isObject(prototype)) {
        Object.prototype = prototype;
        var result = new Object;
        Object.prototype = null;
      }
      return result || global.Object();
    };
  }());
}

module.exports = baseCreate;

/* WEBPACK VAR INJECTION */}.call(exports, __webpack_require__(2)))

/***/ }),
/* 18 */
/***/ (function(module, exports) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */

/**
 * A no-operation function.
 *
 * @static
 * @memberOf _
 * @category Utilities
 * @example
 *
 * var object = { 'name': 'fred' };
 * _.noop(object) === undefined;
 * // => true
 */
function noop() {
  // no operation performed
}

module.exports = noop;


/***/ }),
/* 19 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


var $ = __webpack_require__(0);

function removeTabIndex(instance) {
	$(instance).find('.mCustomScrollBox').removeAttr('tabindex');
}

module.exports = {
	removeTabIndex: removeTabIndex
};


/***/ }),
/* 20 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


exports.__esModule = true;
exports.extend = extend;

// Older IE versions do not directly support indexOf so we must implement our own, sadly.
exports.indexOf = indexOf;
exports.escapeExpression = escapeExpression;
exports.isEmpty = isEmpty;
exports.blockParams = blockParams;
exports.appendContextPath = appendContextPath;
var escape = {
  '&': '&amp;',
  '<': '&lt;',
  '>': '&gt;',
  '"': '&quot;',
  '\'': '&#x27;',
  '`': '&#x60;'
};

var badChars = /[&<>"'`]/g,
    possible = /[&<>"'`]/;

function escapeChar(chr) {
  return escape[chr];
}

function extend(obj /* , ...source */) {
  for (var i = 1; i < arguments.length; i++) {
    for (var key in arguments[i]) {
      if (Object.prototype.hasOwnProperty.call(arguments[i], key)) {
        obj[key] = arguments[i][key];
      }
    }
  }

  return obj;
}

var toString = Object.prototype.toString;

exports.toString = toString;
// Sourced from lodash
// https://github.com/bestiejs/lodash/blob/master/LICENSE.txt
/*eslint-disable func-style, no-var */
var isFunction = function isFunction(value) {
  return typeof value === 'function';
};
// fallback for older versions of Chrome and Safari
/* istanbul ignore next */
if (isFunction(/x/)) {
  exports.isFunction = isFunction = function (value) {
    return typeof value === 'function' && toString.call(value) === '[object Function]';
  };
}
var isFunction;
exports.isFunction = isFunction;
/*eslint-enable func-style, no-var */

/* istanbul ignore next */
var isArray = Array.isArray || function (value) {
  return value && typeof value === 'object' ? toString.call(value) === '[object Array]' : false;
};exports.isArray = isArray;

function indexOf(array, value) {
  for (var i = 0, len = array.length; i < len; i++) {
    if (array[i] === value) {
      return i;
    }
  }
  return -1;
}

function escapeExpression(string) {
  if (typeof string !== 'string') {
    // don't escape SafeStrings, since they're already safe
    if (string && string.toHTML) {
      return string.toHTML();
    } else if (string == null) {
      return '';
    } else if (!string) {
      return string + '';
    }

    // Force a string conversion as this will be done by the append regardless and
    // the regex test will do this transparently behind the scenes, causing issues if
    // an object's to string has escaped characters in it.
    string = '' + string;
  }

  if (!possible.test(string)) {
    return string;
  }
  return string.replace(badChars, escapeChar);
}

function isEmpty(value) {
  if (!value && value !== 0) {
    return true;
  } else if (isArray(value) && value.length === 0) {
    return true;
  } else {
    return false;
  }
}

function blockParams(params, ids) {
  params.path = ids;
  return params;
}

function appendContextPath(contextPath, id) {
  return (contextPath ? contextPath + '.' : '') + id;
}

/***/ }),
/* 21 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


exports.__esModule = true;

var errorProps = ['description', 'fileName', 'lineNumber', 'message', 'name', 'number', 'stack'];

function Exception(message, node) {
  var loc = node && node.loc,
      line = undefined,
      column = undefined;
  if (loc) {
    line = loc.start.line;
    column = loc.start.column;

    message += ' - ' + line + ':' + column;
  }

  var tmp = Error.prototype.constructor.call(this, message);

  // Unfortunately errors are not enumerable in Chrome (at least), so `for prop in tmp` doesn't work.
  for (var idx = 0; idx < errorProps.length; idx++) {
    this[errorProps[idx]] = tmp[errorProps[idx]];
  }

  if (Error.captureStackTrace) {
    Error.captureStackTrace(this, Exception);
  }

  if (loc) {
    this.lineNumber = line;
    this.column = column;
  }
}

Exception.prototype = new Error();

exports['default'] = Exception;
module.exports = exports['default'];

/***/ }),
/* 22 */,
/* 23 */,
/* 24 */,
/* 25 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * Lock Tab Focus helper
 *
 */



var $ = __webpack_require__(0),
	$focusTarget,
	focusableChildElements = {},
	includeInvisibleChildElements = false,
	focusableElementsLength = 0,
	keyCodes = {
		tab: 9
	},
	didInitialize = false,
	observer,
	mutationObserverConfig;

function getFocusableElements() {

	if (includeInvisibleChildElements) {
		focusableChildElements = $focusTarget.find(
			'a:not([tabindex="-1"]), button:not([tabindex="-1"]):not([disabled]), input:not([tabindex="-1"]):not([type="hidden"]):not([disabled]), div.mod_media_gallery_image_wrapper'
		);
	} else {
		focusableChildElements = $focusTarget.find(
			'a:not([tabindex="-1"]):visible, button:not([tabindex="-1"]):not([disabled]):visible, input:not([tabindex="-1"]):not([type="hidden"]):not([disabled]):visible, div.mod_media_gallery_image_wrapper'
		);
	}

	focusableElementsLength = focusableChildElements.length - 1;

}

function setKeyHandlingForFirstElement(e) {

	if (e.which === keyCodes.tab && e.shiftKey) {
		e.preventDefault();
		focusableChildElements[focusableElementsLength].focus();
	}

}

function setKeyHandlingForLastElement(e) {

	if (e.which === keyCodes.tab && e.shiftKey === false) {
		e.preventDefault();
		focusableChildElements[0].focus();
	}

}

function updateKeyHandling() {

	focusableChildElements[0].addEventListener(
		'keydown',
		setKeyHandlingForFirstElement
	);

	focusableChildElements[focusableElementsLength].addEventListener(
		'keydown',
		setKeyHandlingForLastElement
	);

}

function removeKeyHandlers() {

	focusableChildElements[0].removeEventListener(
		'keydown',
		setKeyHandlingForFirstElement
	);

	focusableChildElements[focusableElementsLength].removeEventListener(
		'keydown',
		setKeyHandlingForLastElement
	);

}

function initMutationObserver() {

	observer = new MutationObserver(function() {

		removeKeyHandlers();

		requestAnimationFrame(function() {
			getFocusableElements();
			updateKeyHandling();
		}.bind(this));

	});

	observer.observe($focusTarget[0], mutationObserverConfig);

}

function destroyMutationObserver() {

	observer.disconnect();

}

function selectFirstElement() {

	if (focusableChildElements.length > 0) {
		focusableChildElements[0].focus();
	}

}

function getDirectBodyChildren() {
	var children,
		$children;

	// Since the clock overlay is a child element of data-init="footer" we have to create an exception here
	// otherwise the clock overlay could not get reached due to the aria-hidden="true" and role="presentation"
	// attributes.
	if ($focusTarget[0].getAttribute('data-init') === 'clock') {

		$children = $('body').children('div:not([data-init="marketresearchoverlay"]):not([data-init="accordion footer"]):not(.footerconfig), main');

		children = $.merge($children, [
			$('[data-footer="heading_footer"]')[0],
			$('[data-footer="heading_clock"]')[0],
			$('[data-footer="wrapper"]')[0],
			$('[data-footer="langnav"]')[0],
			$('[data-footer="bottomcontainer_disclaimer"]')[0],
			$('[data-footer="bottomcontainer_logo"]')[0]
		]);

	// the multiselect checkboxes can't be reached if we would set the aria-hidden="true"
	// and role="presentation" attributes on the main container
	} else if ($focusTarget[0].getAttribute('data-multiselect') === 'container') {
		return;

	} else {
		children = $('body').children('div:not([data-init="marketresearchoverlay"]), main');
	}

	// make sure we have no undefined children
	children = children.filter(function(index, child) {
		return child !== undefined;
	});

	children.splice($.inArray($focusTarget, children), 1);

	return children;
}

function hideBodyChildrenExceptThisForScreenreaders() {
	var children = getDirectBodyChildren();

	requestAnimationFrame(function() {
		if (children !== undefined) {
			children.attr({
				'aria-hidden': true,
				'role': 'presentation'
			});
		}
	});

}

function showBodyChildrenExceptThisForScreenreaders() {
	var children = getDirectBodyChildren();

	requestAnimationFrame(function() {

		if (children !== undefined) {
			children.attr({
				'aria-hidden': false
			})
			.removeAttr('role');
		}
	});

}

function init() {

	if (didInitialize) {
		return;
	}

	didInitialize = true;

	// Make sure that we receive a DOM Element to work with
	if (!$.contains(document, $focusTarget[0])) {
		throw new Error('lockTabFocus expects an object, received ' + typeof $focusTarget);
	}

	// Get all focusable elements within the focus target
	getFocusableElements();

	// Set Keyboard Handling
	updateKeyHandling();

	if ('MutationObserver' in window) {
		initMutationObserver();
	}

}

module.exports = {

	setFocusOnElement: function($element, elementToFocus, observerConfig, _includeInvisibleChildElements) {

		var possibleElementToFocus;

		includeInvisibleChildElements = _includeInvisibleChildElements;

		// cache the element
		$focusTarget = $element;

		// define which properties should be watched by MutationObserver
		mutationObserverConfig = observerConfig || {
			childList: true,
			subtree: true,
			attributes: true
		};

		init();

		hideBodyChildrenExceptThisForScreenreaders();

		// if the parameter is not defined, we can focus the first element
		if (!elementToFocus) {
			selectFirstElement();

			return;
		}

		// so at this point, we know that delegate wants to focus specific element
		// check if we get passed a number ("select nth child") or specific selector
		if ($.isNumeric(elementToFocus)) { // we got a number
			if (elementToFocus < focusableChildElements.length) {
				focusableChildElements[elementToFocus].focus();
			}
		} else { // we might have got a selector

			// just for the case that this method gets called in a promised-based
			// event chain, we throw the following commands on the callstack
			// by using setTimeout
			window.setTimeout(function() {
				possibleElementToFocus = $element.find(elementToFocus);

				if (possibleElementToFocus.length > 0) {
					possibleElementToFocus[0].focus();
				} else {
					selectFirstElement();
				}
			}, 0);

		}

	},

	destroy: function() {
		showBodyChildrenExceptThisForScreenreaders();
		removeKeyHandlers();
		destroyMutationObserver();
		didInitialize = false;

	}

};


/***/ }),
/* 26 */,
/* 27 */
/***/ (function(module, exports, __webpack_require__) {

/* WEBPACK VAR INJECTION */(function(global) {/**
 * Lodash (Custom Build) <https://lodash.com/>
 * Build: `lodash modularize exports="npm" -o ./`
 * Copyright OpenJS Foundation and other contributors <https://openjsf.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */
var reInterpolate = __webpack_require__(32);

/** Used as references for various `Number` constants. */
var INFINITY = 1 / 0;

/** `Object#toString` result references. */
var nullTag = '[object Null]',
    symbolTag = '[object Symbol]',
    undefinedTag = '[object Undefined]';

/** Used to match HTML entities and HTML characters. */
var reUnescapedHtml = /[&<>"']/g,
    reHasUnescapedHtml = RegExp(reUnescapedHtml.source);

/** Used to match template delimiters. */
var reEscape = /<%-([\s\S]+?)%>/g,
    reEvaluate = /<%([\s\S]+?)%>/g;

/** Used to map characters to HTML entities. */
var htmlEscapes = {
  '&': '&amp;',
  '<': '&lt;',
  '>': '&gt;',
  '"': '&quot;',
  "'": '&#39;'
};

/** Detect free variable `global` from Node.js. */
var freeGlobal = typeof global == 'object' && global && global.Object === Object && global;

/** Detect free variable `self`. */
var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

/** Used as a reference to the global object. */
var root = freeGlobal || freeSelf || Function('return this')();

/**
 * A specialized version of `_.map` for arrays without support for iteratee
 * shorthands.
 *
 * @private
 * @param {Array} [array] The array to iterate over.
 * @param {Function} iteratee The function invoked per iteration.
 * @returns {Array} Returns the new mapped array.
 */
function arrayMap(array, iteratee) {
  var index = -1,
      length = array == null ? 0 : array.length,
      result = Array(length);

  while (++index < length) {
    result[index] = iteratee(array[index], index, array);
  }
  return result;
}

/**
 * The base implementation of `_.propertyOf` without support for deep paths.
 *
 * @private
 * @param {Object} object The object to query.
 * @returns {Function} Returns the new accessor function.
 */
function basePropertyOf(object) {
  return function(key) {
    return object == null ? undefined : object[key];
  };
}

/**
 * Used by `_.escape` to convert characters to HTML entities.
 *
 * @private
 * @param {string} chr The matched character to escape.
 * @returns {string} Returns the escaped character.
 */
var escapeHtmlChar = basePropertyOf(htmlEscapes);

/** Used for built-in method references. */
var objectProto = Object.prototype;

/** Used to check objects for own properties. */
var hasOwnProperty = objectProto.hasOwnProperty;

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
 * of values.
 */
var nativeObjectToString = objectProto.toString;

/** Built-in value references. */
var Symbol = root.Symbol,
    symToStringTag = Symbol ? Symbol.toStringTag : undefined;

/** Used to convert symbols to primitives and strings. */
var symbolProto = Symbol ? Symbol.prototype : undefined,
    symbolToString = symbolProto ? symbolProto.toString : undefined;

/**
 * By default, the template delimiters used by lodash are like those in
 * embedded Ruby (ERB) as well as ES2015 template strings. Change the
 * following template settings to use alternative delimiters.
 *
 * @static
 * @memberOf _
 * @type {Object}
 */
var templateSettings = {

  /**
   * Used to detect `data` property values to be HTML-escaped.
   *
   * @memberOf _.templateSettings
   * @type {RegExp}
   */
  'escape': reEscape,

  /**
   * Used to detect code to be evaluated.
   *
   * @memberOf _.templateSettings
   * @type {RegExp}
   */
  'evaluate': reEvaluate,

  /**
   * Used to detect `data` property values to inject.
   *
   * @memberOf _.templateSettings
   * @type {RegExp}
   */
  'interpolate': reInterpolate,

  /**
   * Used to reference the data object in the template text.
   *
   * @memberOf _.templateSettings
   * @type {string}
   */
  'variable': '',

  /**
   * Used to import variables into the compiled template.
   *
   * @memberOf _.templateSettings
   * @type {Object}
   */
  'imports': {

    /**
     * A reference to the `lodash` function.
     *
     * @memberOf _.templateSettings.imports
     * @type {Function}
     */
    '_': { 'escape': escape }
  }
};

/**
 * The base implementation of `getTag` without fallbacks for buggy environments.
 *
 * @private
 * @param {*} value The value to query.
 * @returns {string} Returns the `toStringTag`.
 */
function baseGetTag(value) {
  if (value == null) {
    return value === undefined ? undefinedTag : nullTag;
  }
  return (symToStringTag && symToStringTag in Object(value))
    ? getRawTag(value)
    : objectToString(value);
}

/**
 * The base implementation of `_.toString` which doesn't convert nullish
 * values to empty strings.
 *
 * @private
 * @param {*} value The value to process.
 * @returns {string} Returns the string.
 */
function baseToString(value) {
  // Exit early for strings to avoid a performance hit in some environments.
  if (typeof value == 'string') {
    return value;
  }
  if (isArray(value)) {
    // Recursively convert values (susceptible to call stack limits).
    return arrayMap(value, baseToString) + '';
  }
  if (isSymbol(value)) {
    return symbolToString ? symbolToString.call(value) : '';
  }
  var result = (value + '');
  return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
}

/**
 * A specialized version of `baseGetTag` which ignores `Symbol.toStringTag` values.
 *
 * @private
 * @param {*} value The value to query.
 * @returns {string} Returns the raw `toStringTag`.
 */
function getRawTag(value) {
  var isOwn = hasOwnProperty.call(value, symToStringTag),
      tag = value[symToStringTag];

  try {
    value[symToStringTag] = undefined;
    var unmasked = true;
  } catch (e) {}

  var result = nativeObjectToString.call(value);
  if (unmasked) {
    if (isOwn) {
      value[symToStringTag] = tag;
    } else {
      delete value[symToStringTag];
    }
  }
  return result;
}

/**
 * Converts `value` to a string using `Object.prototype.toString`.
 *
 * @private
 * @param {*} value The value to convert.
 * @returns {string} Returns the converted string.
 */
function objectToString(value) {
  return nativeObjectToString.call(value);
}

/**
 * Checks if `value` is classified as an `Array` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an array, else `false`.
 * @example
 *
 * _.isArray([1, 2, 3]);
 * // => true
 *
 * _.isArray(document.body.children);
 * // => false
 *
 * _.isArray('abc');
 * // => false
 *
 * _.isArray(_.noop);
 * // => false
 */
var isArray = Array.isArray;

/**
 * Checks if `value` is object-like. A value is object-like if it's not `null`
 * and has a `typeof` result of "object".
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
 * @example
 *
 * _.isObjectLike({});
 * // => true
 *
 * _.isObjectLike([1, 2, 3]);
 * // => true
 *
 * _.isObjectLike(_.noop);
 * // => false
 *
 * _.isObjectLike(null);
 * // => false
 */
function isObjectLike(value) {
  return value != null && typeof value == 'object';
}

/**
 * Checks if `value` is classified as a `Symbol` primitive or object.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
 * @example
 *
 * _.isSymbol(Symbol.iterator);
 * // => true
 *
 * _.isSymbol('abc');
 * // => false
 */
function isSymbol(value) {
  return typeof value == 'symbol' ||
    (isObjectLike(value) && baseGetTag(value) == symbolTag);
}

/**
 * Converts `value` to a string. An empty string is returned for `null`
 * and `undefined` values. The sign of `-0` is preserved.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to convert.
 * @returns {string} Returns the converted string.
 * @example
 *
 * _.toString(null);
 * // => ''
 *
 * _.toString(-0);
 * // => '-0'
 *
 * _.toString([1, 2, 3]);
 * // => '1,2,3'
 */
function toString(value) {
  return value == null ? '' : baseToString(value);
}

/**
 * Converts the characters "&", "<", ">", '"', and "'" in `string` to their
 * corresponding HTML entities.
 *
 * **Note:** No other characters are escaped. To escape additional
 * characters use a third-party library like [_he_](https://mths.be/he).
 *
 * Though the ">" character is escaped for symmetry, characters like
 * ">" and "/" don't need escaping in HTML and have no special meaning
 * unless they're part of a tag or unquoted attribute value. See
 * [Mathias Bynens's article](https://mathiasbynens.be/notes/ambiguous-ampersands)
 * (under "semi-related fun fact") for more details.
 *
 * When working with HTML you should always
 * [quote attribute values](http://wonko.com/post/html-escaping) to reduce
 * XSS vectors.
 *
 * @static
 * @since 0.1.0
 * @memberOf _
 * @category String
 * @param {string} [string=''] The string to escape.
 * @returns {string} Returns the escaped string.
 * @example
 *
 * _.escape('fred, barney, & pebbles');
 * // => 'fred, barney, &amp; pebbles'
 */
function escape(string) {
  string = toString(string);
  return (string && reHasUnescapedHtml.test(string))
    ? string.replace(reUnescapedHtml, escapeHtmlChar)
    : string;
}

module.exports = templateSettings;

/* WEBPACK VAR INJECTION */}.call(exports, __webpack_require__(2)))

/***/ }),
/* 28 */
/***/ (function(module, exports) {

/**
 * lodash 4.0.0 (Custom Build) <https://lodash.com/>
 * Build: `lodash modularize exports="npm" -o ./`
 * Copyright 2012-2016 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2016 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <https://lodash.com/license>
 */

/**
 * Checks if `value` is classified as an `Array` object.
 *
 * @static
 * @memberOf _
 * @type Function
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is correctly classified, else `false`.
 * @example
 *
 * _.isArray([1, 2, 3]);
 * // => true
 *
 * _.isArray(document.body.children);
 * // => false
 *
 * _.isArray('abc');
 * // => false
 *
 * _.isArray(_.noop);
 * // => false
 */
var isArray = Array.isArray;

module.exports = isArray;


/***/ }),
/* 29 */
/***/ (function(module, exports, __webpack_require__) {

/* WEBPACK VAR INJECTION */(function(global) {/**
 * lodash (Custom Build) <https://lodash.com/>
 * Build: `lodash modularize exports="npm" -o ./`
 * Copyright jQuery Foundation and other contributors <https://jquery.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */

/** Used as the `TypeError` message for "Functions" methods. */
var FUNC_ERROR_TEXT = 'Expected a function';

/** Used as references for various `Number` constants. */
var NAN = 0 / 0;

/** `Object#toString` result references. */
var symbolTag = '[object Symbol]';

/** Used to match leading and trailing whitespace. */
var reTrim = /^\s+|\s+$/g;

/** Used to detect bad signed hexadecimal string values. */
var reIsBadHex = /^[-+]0x[0-9a-f]+$/i;

/** Used to detect binary string values. */
var reIsBinary = /^0b[01]+$/i;

/** Used to detect octal string values. */
var reIsOctal = /^0o[0-7]+$/i;

/** Built-in method references without a dependency on `root`. */
var freeParseInt = parseInt;

/** Detect free variable `global` from Node.js. */
var freeGlobal = typeof global == 'object' && global && global.Object === Object && global;

/** Detect free variable `self`. */
var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

/** Used as a reference to the global object. */
var root = freeGlobal || freeSelf || Function('return this')();

/** Used for built-in method references. */
var objectProto = Object.prototype;

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
 * of values.
 */
var objectToString = objectProto.toString;

/* Built-in method references for those with the same name as other `lodash` methods. */
var nativeMax = Math.max,
    nativeMin = Math.min;

/**
 * Gets the timestamp of the number of milliseconds that have elapsed since
 * the Unix epoch (1 January 1970 00:00:00 UTC).
 *
 * @static
 * @memberOf _
 * @since 2.4.0
 * @category Date
 * @returns {number} Returns the timestamp.
 * @example
 *
 * _.defer(function(stamp) {
 *   console.log(_.now() - stamp);
 * }, _.now());
 * // => Logs the number of milliseconds it took for the deferred invocation.
 */
var now = function() {
  return root.Date.now();
};

/**
 * Creates a debounced function that delays invoking `func` until after `wait`
 * milliseconds have elapsed since the last time the debounced function was
 * invoked. The debounced function comes with a `cancel` method to cancel
 * delayed `func` invocations and a `flush` method to immediately invoke them.
 * Provide `options` to indicate whether `func` should be invoked on the
 * leading and/or trailing edge of the `wait` timeout. The `func` is invoked
 * with the last arguments provided to the debounced function. Subsequent
 * calls to the debounced function return the result of the last `func`
 * invocation.
 *
 * **Note:** If `leading` and `trailing` options are `true`, `func` is
 * invoked on the trailing edge of the timeout only if the debounced function
 * is invoked more than once during the `wait` timeout.
 *
 * If `wait` is `0` and `leading` is `false`, `func` invocation is deferred
 * until to the next tick, similar to `setTimeout` with a timeout of `0`.
 *
 * See [David Corbacho's article](https://css-tricks.com/debouncing-throttling-explained-examples/)
 * for details over the differences between `_.debounce` and `_.throttle`.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Function
 * @param {Function} func The function to debounce.
 * @param {number} [wait=0] The number of milliseconds to delay.
 * @param {Object} [options={}] The options object.
 * @param {boolean} [options.leading=false]
 *  Specify invoking on the leading edge of the timeout.
 * @param {number} [options.maxWait]
 *  The maximum time `func` is allowed to be delayed before it's invoked.
 * @param {boolean} [options.trailing=true]
 *  Specify invoking on the trailing edge of the timeout.
 * @returns {Function} Returns the new debounced function.
 * @example
 *
 * // Avoid costly calculations while the window size is in flux.
 * jQuery(window).on('resize', _.debounce(calculateLayout, 150));
 *
 * // Invoke `sendMail` when clicked, debouncing subsequent calls.
 * jQuery(element).on('click', _.debounce(sendMail, 300, {
 *   'leading': true,
 *   'trailing': false
 * }));
 *
 * // Ensure `batchLog` is invoked once after 1 second of debounced calls.
 * var debounced = _.debounce(batchLog, 250, { 'maxWait': 1000 });
 * var source = new EventSource('/stream');
 * jQuery(source).on('message', debounced);
 *
 * // Cancel the trailing debounced invocation.
 * jQuery(window).on('popstate', debounced.cancel);
 */
function debounce(func, wait, options) {
  var lastArgs,
      lastThis,
      maxWait,
      result,
      timerId,
      lastCallTime,
      lastInvokeTime = 0,
      leading = false,
      maxing = false,
      trailing = true;

  if (typeof func != 'function') {
    throw new TypeError(FUNC_ERROR_TEXT);
  }
  wait = toNumber(wait) || 0;
  if (isObject(options)) {
    leading = !!options.leading;
    maxing = 'maxWait' in options;
    maxWait = maxing ? nativeMax(toNumber(options.maxWait) || 0, wait) : maxWait;
    trailing = 'trailing' in options ? !!options.trailing : trailing;
  }

  function invokeFunc(time) {
    var args = lastArgs,
        thisArg = lastThis;

    lastArgs = lastThis = undefined;
    lastInvokeTime = time;
    result = func.apply(thisArg, args);
    return result;
  }

  function leadingEdge(time) {
    // Reset any `maxWait` timer.
    lastInvokeTime = time;
    // Start the timer for the trailing edge.
    timerId = setTimeout(timerExpired, wait);
    // Invoke the leading edge.
    return leading ? invokeFunc(time) : result;
  }

  function remainingWait(time) {
    var timeSinceLastCall = time - lastCallTime,
        timeSinceLastInvoke = time - lastInvokeTime,
        result = wait - timeSinceLastCall;

    return maxing ? nativeMin(result, maxWait - timeSinceLastInvoke) : result;
  }

  function shouldInvoke(time) {
    var timeSinceLastCall = time - lastCallTime,
        timeSinceLastInvoke = time - lastInvokeTime;

    // Either this is the first call, activity has stopped and we're at the
    // trailing edge, the system time has gone backwards and we're treating
    // it as the trailing edge, or we've hit the `maxWait` limit.
    return (lastCallTime === undefined || (timeSinceLastCall >= wait) ||
      (timeSinceLastCall < 0) || (maxing && timeSinceLastInvoke >= maxWait));
  }

  function timerExpired() {
    var time = now();
    if (shouldInvoke(time)) {
      return trailingEdge(time);
    }
    // Restart the timer.
    timerId = setTimeout(timerExpired, remainingWait(time));
  }

  function trailingEdge(time) {
    timerId = undefined;

    // Only invoke if we have `lastArgs` which means `func` has been
    // debounced at least once.
    if (trailing && lastArgs) {
      return invokeFunc(time);
    }
    lastArgs = lastThis = undefined;
    return result;
  }

  function cancel() {
    if (timerId !== undefined) {
      clearTimeout(timerId);
    }
    lastInvokeTime = 0;
    lastArgs = lastCallTime = lastThis = timerId = undefined;
  }

  function flush() {
    return timerId === undefined ? result : trailingEdge(now());
  }

  function debounced() {
    var time = now(),
        isInvoking = shouldInvoke(time);

    lastArgs = arguments;
    lastThis = this;
    lastCallTime = time;

    if (isInvoking) {
      if (timerId === undefined) {
        return leadingEdge(lastCallTime);
      }
      if (maxing) {
        // Handle invocations in a tight loop.
        timerId = setTimeout(timerExpired, wait);
        return invokeFunc(lastCallTime);
      }
    }
    if (timerId === undefined) {
      timerId = setTimeout(timerExpired, wait);
    }
    return result;
  }
  debounced.cancel = cancel;
  debounced.flush = flush;
  return debounced;
}

/**
 * Checks if `value` is the
 * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
 * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an object, else `false`.
 * @example
 *
 * _.isObject({});
 * // => true
 *
 * _.isObject([1, 2, 3]);
 * // => true
 *
 * _.isObject(_.noop);
 * // => true
 *
 * _.isObject(null);
 * // => false
 */
function isObject(value) {
  var type = typeof value;
  return !!value && (type == 'object' || type == 'function');
}

/**
 * Checks if `value` is object-like. A value is object-like if it's not `null`
 * and has a `typeof` result of "object".
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
 * @example
 *
 * _.isObjectLike({});
 * // => true
 *
 * _.isObjectLike([1, 2, 3]);
 * // => true
 *
 * _.isObjectLike(_.noop);
 * // => false
 *
 * _.isObjectLike(null);
 * // => false
 */
function isObjectLike(value) {
  return !!value && typeof value == 'object';
}

/**
 * Checks if `value` is classified as a `Symbol` primitive or object.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
 * @example
 *
 * _.isSymbol(Symbol.iterator);
 * // => true
 *
 * _.isSymbol('abc');
 * // => false
 */
function isSymbol(value) {
  return typeof value == 'symbol' ||
    (isObjectLike(value) && objectToString.call(value) == symbolTag);
}

/**
 * Converts `value` to a number.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to process.
 * @returns {number} Returns the number.
 * @example
 *
 * _.toNumber(3.2);
 * // => 3.2
 *
 * _.toNumber(Number.MIN_VALUE);
 * // => 5e-324
 *
 * _.toNumber(Infinity);
 * // => Infinity
 *
 * _.toNumber('3.2');
 * // => 3.2
 */
function toNumber(value) {
  if (typeof value == 'number') {
    return value;
  }
  if (isSymbol(value)) {
    return NAN;
  }
  if (isObject(value)) {
    var other = typeof value.valueOf == 'function' ? value.valueOf() : value;
    value = isObject(other) ? (other + '') : other;
  }
  if (typeof value != 'string') {
    return value === 0 ? value : +value;
  }
  value = value.replace(reTrim, '');
  var isBinary = reIsBinary.test(value);
  return (isBinary || reIsOctal.test(value))
    ? freeParseInt(value.slice(2), isBinary ? 2 : 8)
    : (reIsBadHex.test(value) ? NAN : +value);
}

module.exports = debounce;

/* WEBPACK VAR INJECTION */}.call(exports, __webpack_require__(2)))

/***/ }),
/* 30 */,
/* 31 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


var _interopRequireWildcard = function (obj) { return obj && obj.__esModule ? obj : { 'default': obj }; };

exports.__esModule = true;
exports.HandlebarsEnvironment = HandlebarsEnvironment;
exports.createFrame = createFrame;

var _import = __webpack_require__(20);

var Utils = _interopRequireWildcard(_import);

var _Exception = __webpack_require__(21);

var _Exception2 = _interopRequireWildcard(_Exception);

var VERSION = '3.0.1';
exports.VERSION = VERSION;
var COMPILER_REVISION = 6;

exports.COMPILER_REVISION = COMPILER_REVISION;
var REVISION_CHANGES = {
  1: '<= 1.0.rc.2', // 1.0.rc.2 is actually rev2 but doesn't report it
  2: '== 1.0.0-rc.3',
  3: '== 1.0.0-rc.4',
  4: '== 1.x.x',
  5: '== 2.0.0-alpha.x',
  6: '>= 2.0.0-beta.1'
};

exports.REVISION_CHANGES = REVISION_CHANGES;
var isArray = Utils.isArray,
    isFunction = Utils.isFunction,
    toString = Utils.toString,
    objectType = '[object Object]';

function HandlebarsEnvironment(helpers, partials) {
  this.helpers = helpers || {};
  this.partials = partials || {};

  registerDefaultHelpers(this);
}

HandlebarsEnvironment.prototype = {
  constructor: HandlebarsEnvironment,

  logger: logger,
  log: log,

  registerHelper: function registerHelper(name, fn) {
    if (toString.call(name) === objectType) {
      if (fn) {
        throw new _Exception2['default']('Arg not supported with multiple helpers');
      }
      Utils.extend(this.helpers, name);
    } else {
      this.helpers[name] = fn;
    }
  },
  unregisterHelper: function unregisterHelper(name) {
    delete this.helpers[name];
  },

  registerPartial: function registerPartial(name, partial) {
    if (toString.call(name) === objectType) {
      Utils.extend(this.partials, name);
    } else {
      if (typeof partial === 'undefined') {
        throw new _Exception2['default']('Attempting to register a partial as undefined');
      }
      this.partials[name] = partial;
    }
  },
  unregisterPartial: function unregisterPartial(name) {
    delete this.partials[name];
  }
};

function registerDefaultHelpers(instance) {
  instance.registerHelper('helperMissing', function () {
    if (arguments.length === 1) {
      // A missing field in a {{foo}} constuct.
      return undefined;
    } else {
      // Someone is actually trying to call something, blow up.
      throw new _Exception2['default']('Missing helper: "' + arguments[arguments.length - 1].name + '"');
    }
  });

  instance.registerHelper('blockHelperMissing', function (context, options) {
    var inverse = options.inverse,
        fn = options.fn;

    if (context === true) {
      return fn(this);
    } else if (context === false || context == null) {
      return inverse(this);
    } else if (isArray(context)) {
      if (context.length > 0) {
        if (options.ids) {
          options.ids = [options.name];
        }

        return instance.helpers.each(context, options);
      } else {
        return inverse(this);
      }
    } else {
      if (options.data && options.ids) {
        var data = createFrame(options.data);
        data.contextPath = Utils.appendContextPath(options.data.contextPath, options.name);
        options = { data: data };
      }

      return fn(context, options);
    }
  });

  instance.registerHelper('each', function (context, options) {
    if (!options) {
      throw new _Exception2['default']('Must pass iterator to #each');
    }

    var fn = options.fn,
        inverse = options.inverse,
        i = 0,
        ret = '',
        data = undefined,
        contextPath = undefined;

    if (options.data && options.ids) {
      contextPath = Utils.appendContextPath(options.data.contextPath, options.ids[0]) + '.';
    }

    if (isFunction(context)) {
      context = context.call(this);
    }

    if (options.data) {
      data = createFrame(options.data);
    }

    function execIteration(field, index, last) {
      if (data) {
        data.key = field;
        data.index = index;
        data.first = index === 0;
        data.last = !!last;

        if (contextPath) {
          data.contextPath = contextPath + field;
        }
      }

      ret = ret + fn(context[field], {
        data: data,
        blockParams: Utils.blockParams([context[field], field], [contextPath + field, null])
      });
    }

    if (context && typeof context === 'object') {
      if (isArray(context)) {
        for (var j = context.length; i < j; i++) {
          execIteration(i, i, i === context.length - 1);
        }
      } else {
        var priorKey = undefined;

        for (var key in context) {
          if (context.hasOwnProperty(key)) {
            // We're running the iterations one step out of sync so we can detect
            // the last iteration without have to scan the object twice and create
            // an itermediate keys array.
            if (priorKey) {
              execIteration(priorKey, i - 1);
            }
            priorKey = key;
            i++;
          }
        }
        if (priorKey) {
          execIteration(priorKey, i - 1, true);
        }
      }
    }

    if (i === 0) {
      ret = inverse(this);
    }

    return ret;
  });

  instance.registerHelper('if', function (conditional, options) {
    if (isFunction(conditional)) {
      conditional = conditional.call(this);
    }

    // Default behavior is to render the positive path if the value is truthy and not empty.
    // The `includeZero` option may be set to treat the condtional as purely not empty based on the
    // behavior of isEmpty. Effectively this determines if 0 is handled by the positive path or negative.
    if (!options.hash.includeZero && !conditional || Utils.isEmpty(conditional)) {
      return options.inverse(this);
    } else {
      return options.fn(this);
    }
  });

  instance.registerHelper('unless', function (conditional, options) {
    return instance.helpers['if'].call(this, conditional, { fn: options.inverse, inverse: options.fn, hash: options.hash });
  });

  instance.registerHelper('with', function (context, options) {
    if (isFunction(context)) {
      context = context.call(this);
    }

    var fn = options.fn;

    if (!Utils.isEmpty(context)) {
      if (options.data && options.ids) {
        var data = createFrame(options.data);
        data.contextPath = Utils.appendContextPath(options.data.contextPath, options.ids[0]);
        options = { data: data };
      }

      return fn(context, options);
    } else {
      return options.inverse(this);
    }
  });

  instance.registerHelper('log', function (message, options) {
    var level = options.data && options.data.level != null ? parseInt(options.data.level, 10) : 1;
    instance.log(level, message);
  });

  instance.registerHelper('lookup', function (obj, field) {
    return obj && obj[field];
  });
}

var logger = {
  methodMap: { 0: 'debug', 1: 'info', 2: 'warn', 3: 'error' },

  // State enum
  DEBUG: 0,
  INFO: 1,
  WARN: 2,
  ERROR: 3,
  level: 1,

  // Can be overridden in the host environment
  log: function log(level, message) {
    if (typeof console !== 'undefined' && logger.level <= level) {
      var method = logger.methodMap[level];
      (console[method] || console.log).call(console, message); // eslint-disable-line no-console
    }
  }
};

exports.logger = logger;
var log = logger.log;

exports.log = log;

function createFrame(object) {
  var frame = Utils.extend({}, object);
  frame._parent = object;
  return frame;
}

/* [args, ]options */

/***/ }),
/* 32 */
/***/ (function(module, exports) {

/**
 * lodash 3.0.0 (Custom Build) <https://lodash.com/>
 * Build: `lodash modern modularize exports="npm" -o ./`
 * Copyright 2012-2015 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.7.0 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2015 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <https://lodash.com/license>
 */

/** Used to match template delimiters. */
var reInterpolate = /<%=([\s\S]+?)%>/g;

module.exports = reInterpolate;


/***/ }),
/* 33 */
/***/ (function(module, exports, __webpack_require__) {

/* WEBPACK VAR INJECTION */(function(global) {/**
 * lodash (Custom Build) <https://lodash.com/>
 * Build: `lodash modularize exports="npm" -o ./`
 * Copyright jQuery Foundation and other contributors <https://jquery.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */

/** Used as references for various `Number` constants. */
var INFINITY = 1 / 0;

/** `Object#toString` result references. */
var symbolTag = '[object Symbol]';

/** Detect free variable `global` from Node.js. */
var freeGlobal = typeof global == 'object' && global && global.Object === Object && global;

/** Detect free variable `self`. */
var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

/** Used as a reference to the global object. */
var root = freeGlobal || freeSelf || Function('return this')();

/** Used for built-in method references. */
var objectProto = Object.prototype;

/** Used to generate unique IDs. */
var idCounter = 0;

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/6.0/#sec-object.prototype.tostring)
 * of values.
 */
var objectToString = objectProto.toString;

/** Built-in value references. */
var Symbol = root.Symbol;

/** Used to convert symbols to primitives and strings. */
var symbolProto = Symbol ? Symbol.prototype : undefined,
    symbolToString = symbolProto ? symbolProto.toString : undefined;

/**
 * The base implementation of `_.toString` which doesn't convert nullish
 * values to empty strings.
 *
 * @private
 * @param {*} value The value to process.
 * @returns {string} Returns the string.
 */
function baseToString(value) {
  // Exit early for strings to avoid a performance hit in some environments.
  if (typeof value == 'string') {
    return value;
  }
  if (isSymbol(value)) {
    return symbolToString ? symbolToString.call(value) : '';
  }
  var result = (value + '');
  return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
}

/**
 * Checks if `value` is object-like. A value is object-like if it's not `null`
 * and has a `typeof` result of "object".
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
 * @example
 *
 * _.isObjectLike({});
 * // => true
 *
 * _.isObjectLike([1, 2, 3]);
 * // => true
 *
 * _.isObjectLike(_.noop);
 * // => false
 *
 * _.isObjectLike(null);
 * // => false
 */
function isObjectLike(value) {
  return !!value && typeof value == 'object';
}

/**
 * Checks if `value` is classified as a `Symbol` primitive or object.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
 * @example
 *
 * _.isSymbol(Symbol.iterator);
 * // => true
 *
 * _.isSymbol('abc');
 * // => false
 */
function isSymbol(value) {
  return typeof value == 'symbol' ||
    (isObjectLike(value) && objectToString.call(value) == symbolTag);
}

/**
 * Converts `value` to a string. An empty string is returned for `null`
 * and `undefined` values. The sign of `-0` is preserved.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to process.
 * @returns {string} Returns the string.
 * @example
 *
 * _.toString(null);
 * // => ''
 *
 * _.toString(-0);
 * // => '-0'
 *
 * _.toString([1, 2, 3]);
 * // => '1,2,3'
 */
function toString(value) {
  return value == null ? '' : baseToString(value);
}

/**
 * Generates a unique ID. If `prefix` is given, the ID is appended to it.
 *
 * @static
 * @since 0.1.0
 * @memberOf _
 * @category Util
 * @param {string} [prefix=''] The value to prefix the ID with.
 * @returns {string} Returns the unique ID.
 * @example
 *
 * _.uniqueId('contact_');
 * // => 'contact_104'
 *
 * _.uniqueId();
 * // => '105'
 */
function uniqueId(prefix) {
  var id = ++idCounter;
  return toString(prefix) + id;
}

module.exports = uniqueId;

/* WEBPACK VAR INJECTION */}.call(exports, __webpack_require__(2)))

/***/ }),
/* 34 */
/***/ (function(module, exports) {

/**
 * lodash (Custom Build) <https://lodash.com/>
 * Build: `lodash modularize exports="npm" -o ./`
 * Copyright jQuery Foundation and other contributors <https://jquery.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */

/** `Object#toString` result references. */
var objectTag = '[object Object]';

/**
 * Checks if `value` is a host object in IE < 9.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a host object, else `false`.
 */
function isHostObject(value) {
  // Many host objects are `Object` objects that can coerce to strings
  // despite having improperly defined `toString` methods.
  var result = false;
  if (value != null && typeof value.toString != 'function') {
    try {
      result = !!(value + '');
    } catch (e) {}
  }
  return result;
}

/**
 * Creates a unary function that invokes `func` with its argument transformed.
 *
 * @private
 * @param {Function} func The function to wrap.
 * @param {Function} transform The argument transform.
 * @returns {Function} Returns the new function.
 */
function overArg(func, transform) {
  return function(arg) {
    return func(transform(arg));
  };
}

/** Used for built-in method references. */
var funcProto = Function.prototype,
    objectProto = Object.prototype;

/** Used to resolve the decompiled source of functions. */
var funcToString = funcProto.toString;

/** Used to check objects for own properties. */
var hasOwnProperty = objectProto.hasOwnProperty;

/** Used to infer the `Object` constructor. */
var objectCtorString = funcToString.call(Object);

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
 * of values.
 */
var objectToString = objectProto.toString;

/** Built-in value references. */
var getPrototype = overArg(Object.getPrototypeOf, Object);

/**
 * Checks if `value` is likely a DOM element.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a DOM element, else `false`.
 * @example
 *
 * _.isElement(document.body);
 * // => true
 *
 * _.isElement('<body>');
 * // => false
 */
function isElement(value) {
  return !!value && value.nodeType === 1 && isObjectLike(value) && !isPlainObject(value);
}

/**
 * Checks if `value` is object-like. A value is object-like if it's not `null`
 * and has a `typeof` result of "object".
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
 * @example
 *
 * _.isObjectLike({});
 * // => true
 *
 * _.isObjectLike([1, 2, 3]);
 * // => true
 *
 * _.isObjectLike(_.noop);
 * // => false
 *
 * _.isObjectLike(null);
 * // => false
 */
function isObjectLike(value) {
  return !!value && typeof value == 'object';
}

/**
 * Checks if `value` is a plain object, that is, an object created by the
 * `Object` constructor or one with a `[[Prototype]]` of `null`.
 *
 * @static
 * @memberOf _
 * @since 0.8.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a plain object, else `false`.
 * @example
 *
 * function Foo() {
 *   this.a = 1;
 * }
 *
 * _.isPlainObject(new Foo);
 * // => false
 *
 * _.isPlainObject([1, 2, 3]);
 * // => false
 *
 * _.isPlainObject({ 'x': 0, 'y': 0 });
 * // => true
 *
 * _.isPlainObject(Object.create(null));
 * // => true
 */
function isPlainObject(value) {
  if (!isObjectLike(value) ||
      objectToString.call(value) != objectTag || isHostObject(value)) {
    return false;
  }
  var proto = getPrototype(value);
  if (proto === null) {
    return true;
  }
  var Ctor = hasOwnProperty.call(proto, 'constructor') && proto.constructor;
  return (typeof Ctor == 'function' &&
    Ctor instanceof Ctor && funcToString.call(Ctor) == objectCtorString);
}

module.exports = isElement;


/***/ }),
/* 35 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


var $ = __webpack_require__(0),
	$document = $(document),
	isArray = __webpack_require__(28),
	registeredInitEvents = {},

	initializeModulesInDom = function(event) {
		var eventType = event.type,
			initModules = registeredInitEvents[eventType];

		$('[data-init]').each(function() {
			var $this = $(this),
				modules = $this.data('init').split(' ');

			modules.forEach(function(module) {
				if (initModules.indexOf(module) !== -1 && $.fn[module]) {
					$.fn[module].apply($this);
				}
			});
		});

	};

// Expose modules events (ready, ajaxload, ...) to the global scope
$.extend(true, estatico, {
	moduleEvents: {
		trigger: function(eventName) {
			$document.trigger(eventName);
		}
	}
});

module.exports = function(moduleName, initEvents) {

	if (!isArray(initEvents)) {
		// Fallback to sane defaults for legacy modules.
		initEvents = ['ready', 'ajaxload'];
	}

	initEvents.forEach(function(registeredInitEvent) {
		if (!registeredInitEvents[registeredInitEvent]) {
			registeredInitEvents[registeredInitEvent] = [];
			if (registeredInitEvent === 'ready') {
				// As of jQuery 3 the synthetic event named "ready" is deprecated
				$(function() {
					initializeModulesInDom({type: 'ready'});
				});
			} else {
				$document.on(registeredInitEvent, initializeModulesInDom);
			}
		}

		registeredInitEvents[registeredInitEvent].push(moduleName);
	});
};

$document.on('target-dom-loaded', function() {
	$document.triggerHandler('ajaxload');
});


/***/ }),
/* 36 */
/***/ (function(module, exports) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */

/**
 * The base implementation of `_.indexOf` without support for binary searches
 * or `fromIndex` constraints.
 *
 * @private
 * @param {Array} array The array to search.
 * @param {*} value The value to search for.
 * @param {number} [fromIndex=0] The index to search from.
 * @returns {number} Returns the index of the matched value or `-1`.
 */
function baseIndexOf(array, value, fromIndex) {
  var index = (fromIndex || 0) - 1,
      length = array ? array.length : 0;

  while (++index < length) {
    if (array[index] === value) {
      return index;
    }
  }
  return -1;
}

module.exports = baseIndexOf;


/***/ }),
/* 37 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var baseCreateCallback = __webpack_require__(38),
    keys = __webpack_require__(46),
    objectTypes = __webpack_require__(12);

/**
 * Iterates over own enumerable properties of an object, executing the callback
 * for each property. The callback is bound to `thisArg` and invoked with three
 * arguments; (value, key, object). Callbacks may exit iteration early by
 * explicitly returning `false`.
 *
 * @static
 * @memberOf _
 * @type Function
 * @category Objects
 * @param {Object} object The object to iterate over.
 * @param {Function} [callback=identity] The function called per iteration.
 * @param {*} [thisArg] The `this` binding of `callback`.
 * @returns {Object} Returns `object`.
 * @example
 *
 * _.forOwn({ '0': 'zero', '1': 'one', 'length': 2 }, function(num, key) {
 *   console.log(key);
 * });
 * // => logs '0', '1', and 'length' (property order is not guaranteed across environments)
 */
var forOwn = function(collection, callback, thisArg) {
  var index, iterable = collection, result = iterable;
  if (!iterable) return result;
  if (!objectTypes[typeof iterable]) return result;
  callback = callback && typeof thisArg == 'undefined' ? callback : baseCreateCallback(callback, thisArg, 3);
    var ownIndex = -1,
        ownProps = objectTypes[typeof iterable] && keys(iterable),
        length = ownProps ? ownProps.length : 0;

    while (++ownIndex < length) {
      index = ownProps[ownIndex];
      if (callback(iterable[index], index, collection) === false) return result;
    }
  return result
};

module.exports = forOwn;


/***/ }),
/* 38 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var bind = __webpack_require__(39),
    identity = __webpack_require__(44),
    setBindData = __webpack_require__(13),
    support = __webpack_require__(45);

/** Used to detected named functions */
var reFuncName = /^\s*function[ \n\r\t]+\w/;

/** Used to detect functions containing a `this` reference */
var reThis = /\bthis\b/;

/** Native method shortcuts */
var fnToString = Function.prototype.toString;

/**
 * The base implementation of `_.createCallback` without support for creating
 * "_.pluck" or "_.where" style callbacks.
 *
 * @private
 * @param {*} [func=identity] The value to convert to a callback.
 * @param {*} [thisArg] The `this` binding of the created callback.
 * @param {number} [argCount] The number of arguments the callback accepts.
 * @returns {Function} Returns a callback function.
 */
function baseCreateCallback(func, thisArg, argCount) {
  if (typeof func != 'function') {
    return identity;
  }
  // exit early for no `thisArg` or already bound by `Function#bind`
  if (typeof thisArg == 'undefined' || !('prototype' in func)) {
    return func;
  }
  var bindData = func.__bindData__;
  if (typeof bindData == 'undefined') {
    if (support.funcNames) {
      bindData = !func.name;
    }
    bindData = bindData || !support.funcDecomp;
    if (!bindData) {
      var source = fnToString.call(func);
      if (!support.funcNames) {
        bindData = !reFuncName.test(source);
      }
      if (!bindData) {
        // checks if `func` references the `this` keyword and stores the result
        bindData = reThis.test(source);
        setBindData(func, bindData);
      }
    }
  }
  // exit early if there are no `this` references or `func` is bound
  if (bindData === false || (bindData !== true && bindData[1] & 1)) {
    return func;
  }
  switch (argCount) {
    case 1: return function(value) {
      return func.call(thisArg, value);
    };
    case 2: return function(a, b) {
      return func.call(thisArg, a, b);
    };
    case 3: return function(value, index, collection) {
      return func.call(thisArg, value, index, collection);
    };
    case 4: return function(accumulator, value, index, collection) {
      return func.call(thisArg, accumulator, value, index, collection);
    };
  }
  return bind(func, thisArg);
}

module.exports = baseCreateCallback;


/***/ }),
/* 39 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var createWrapper = __webpack_require__(40),
    slice = __webpack_require__(10);

/**
 * Creates a function that, when called, invokes `func` with the `this`
 * binding of `thisArg` and prepends any additional `bind` arguments to those
 * provided to the bound function.
 *
 * @static
 * @memberOf _
 * @category Functions
 * @param {Function} func The function to bind.
 * @param {*} [thisArg] The `this` binding of `func`.
 * @param {...*} [arg] Arguments to be partially applied.
 * @returns {Function} Returns the new bound function.
 * @example
 *
 * var func = function(greeting) {
 *   return greeting + ' ' + this.name;
 * };
 *
 * func = _.bind(func, { 'name': 'fred' }, 'hi');
 * func();
 * // => 'hi fred'
 */
function bind(func, thisArg) {
  return arguments.length > 2
    ? createWrapper(func, 17, slice(arguments, 2), null, thisArg)
    : createWrapper(func, 1, null, null, thisArg);
}

module.exports = bind;


/***/ }),
/* 40 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var baseBind = __webpack_require__(41),
    baseCreateWrapper = __webpack_require__(42),
    isFunction = __webpack_require__(43),
    slice = __webpack_require__(10);

/**
 * Used for `Array` method references.
 *
 * Normally `Array.prototype` would suffice, however, using an array literal
 * avoids issues in Narwhal.
 */
var arrayRef = [];

/** Native method shortcuts */
var push = arrayRef.push,
    unshift = arrayRef.unshift;

/**
 * Creates a function that, when called, either curries or invokes `func`
 * with an optional `this` binding and partially applied arguments.
 *
 * @private
 * @param {Function|string} func The function or method name to reference.
 * @param {number} bitmask The bitmask of method flags to compose.
 *  The bitmask may be composed of the following flags:
 *  1 - `_.bind`
 *  2 - `_.bindKey`
 *  4 - `_.curry`
 *  8 - `_.curry` (bound)
 *  16 - `_.partial`
 *  32 - `_.partialRight`
 * @param {Array} [partialArgs] An array of arguments to prepend to those
 *  provided to the new function.
 * @param {Array} [partialRightArgs] An array of arguments to append to those
 *  provided to the new function.
 * @param {*} [thisArg] The `this` binding of `func`.
 * @param {number} [arity] The arity of `func`.
 * @returns {Function} Returns the new function.
 */
function createWrapper(func, bitmask, partialArgs, partialRightArgs, thisArg, arity) {
  var isBind = bitmask & 1,
      isBindKey = bitmask & 2,
      isCurry = bitmask & 4,
      isCurryBound = bitmask & 8,
      isPartial = bitmask & 16,
      isPartialRight = bitmask & 32;

  if (!isBindKey && !isFunction(func)) {
    throw new TypeError;
  }
  if (isPartial && !partialArgs.length) {
    bitmask &= ~16;
    isPartial = partialArgs = false;
  }
  if (isPartialRight && !partialRightArgs.length) {
    bitmask &= ~32;
    isPartialRight = partialRightArgs = false;
  }
  var bindData = func && func.__bindData__;
  if (bindData && bindData !== true) {
    // clone `bindData`
    bindData = slice(bindData);
    if (bindData[2]) {
      bindData[2] = slice(bindData[2]);
    }
    if (bindData[3]) {
      bindData[3] = slice(bindData[3]);
    }
    // set `thisBinding` is not previously bound
    if (isBind && !(bindData[1] & 1)) {
      bindData[4] = thisArg;
    }
    // set if previously bound but not currently (subsequent curried functions)
    if (!isBind && bindData[1] & 1) {
      bitmask |= 8;
    }
    // set curried arity if not yet set
    if (isCurry && !(bindData[1] & 4)) {
      bindData[5] = arity;
    }
    // append partial left arguments
    if (isPartial) {
      push.apply(bindData[2] || (bindData[2] = []), partialArgs);
    }
    // append partial right arguments
    if (isPartialRight) {
      unshift.apply(bindData[3] || (bindData[3] = []), partialRightArgs);
    }
    // merge flags
    bindData[1] |= bitmask;
    return createWrapper.apply(null, bindData);
  }
  // fast path for `_.bind`
  var creater = (bitmask == 1 || bitmask === 17) ? baseBind : baseCreateWrapper;
  return creater([func, bitmask, partialArgs, partialRightArgs, thisArg, arity]);
}

module.exports = createWrapper;


/***/ }),
/* 41 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var baseCreate = __webpack_require__(17),
    isObject = __webpack_require__(9),
    setBindData = __webpack_require__(13),
    slice = __webpack_require__(10);

/**
 * Used for `Array` method references.
 *
 * Normally `Array.prototype` would suffice, however, using an array literal
 * avoids issues in Narwhal.
 */
var arrayRef = [];

/** Native method shortcuts */
var push = arrayRef.push;

/**
 * The base implementation of `_.bind` that creates the bound function and
 * sets its meta data.
 *
 * @private
 * @param {Array} bindData The bind data array.
 * @returns {Function} Returns the new bound function.
 */
function baseBind(bindData) {
  var func = bindData[0],
      partialArgs = bindData[2],
      thisArg = bindData[4];

  function bound() {
    // `Function#bind` spec
    // http://es5.github.io/#x15.3.4.5
    if (partialArgs) {
      // avoid `arguments` object deoptimizations by using `slice` instead
      // of `Array.prototype.slice.call` and not assigning `arguments` to a
      // variable as a ternary expression
      var args = slice(partialArgs);
      push.apply(args, arguments);
    }
    // mimic the constructor's `return` behavior
    // http://es5.github.io/#x13.2.2
    if (this instanceof bound) {
      // ensure `new bound` is an instance of `func`
      var thisBinding = baseCreate(func.prototype),
          result = func.apply(thisBinding, args || arguments);
      return isObject(result) ? result : thisBinding;
    }
    return func.apply(thisArg, args || arguments);
  }
  setBindData(bound, bindData);
  return bound;
}

module.exports = baseBind;


/***/ }),
/* 42 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var baseCreate = __webpack_require__(17),
    isObject = __webpack_require__(9),
    setBindData = __webpack_require__(13),
    slice = __webpack_require__(10);

/**
 * Used for `Array` method references.
 *
 * Normally `Array.prototype` would suffice, however, using an array literal
 * avoids issues in Narwhal.
 */
var arrayRef = [];

/** Native method shortcuts */
var push = arrayRef.push;

/**
 * The base implementation of `createWrapper` that creates the wrapper and
 * sets its meta data.
 *
 * @private
 * @param {Array} bindData The bind data array.
 * @returns {Function} Returns the new function.
 */
function baseCreateWrapper(bindData) {
  var func = bindData[0],
      bitmask = bindData[1],
      partialArgs = bindData[2],
      partialRightArgs = bindData[3],
      thisArg = bindData[4],
      arity = bindData[5];

  var isBind = bitmask & 1,
      isBindKey = bitmask & 2,
      isCurry = bitmask & 4,
      isCurryBound = bitmask & 8,
      key = func;

  function bound() {
    var thisBinding = isBind ? thisArg : this;
    if (partialArgs) {
      var args = slice(partialArgs);
      push.apply(args, arguments);
    }
    if (partialRightArgs || isCurry) {
      args || (args = slice(arguments));
      if (partialRightArgs) {
        push.apply(args, partialRightArgs);
      }
      if (isCurry && args.length < arity) {
        bitmask |= 16 & ~32;
        return baseCreateWrapper([func, (isCurryBound ? bitmask : bitmask & ~3), args, null, thisArg, arity]);
      }
    }
    args || (args = arguments);
    if (isBindKey) {
      func = thisBinding[key];
    }
    if (this instanceof bound) {
      thisBinding = baseCreate(func.prototype);
      var result = func.apply(thisBinding, args);
      return isObject(result) ? result : thisBinding;
    }
    return func.apply(thisBinding, args);
  }
  setBindData(bound, bindData);
  return bound;
}

module.exports = baseCreateWrapper;


/***/ }),
/* 43 */
/***/ (function(module, exports) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */

/**
 * Checks if `value` is a function.
 *
 * @static
 * @memberOf _
 * @category Objects
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if the `value` is a function, else `false`.
 * @example
 *
 * _.isFunction(_);
 * // => true
 */
function isFunction(value) {
  return typeof value == 'function';
}

module.exports = isFunction;


/***/ }),
/* 44 */
/***/ (function(module, exports) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */

/**
 * This method returns the first argument provided to it.
 *
 * @static
 * @memberOf _
 * @category Utilities
 * @param {*} value Any value.
 * @returns {*} Returns `value`.
 * @example
 *
 * var object = { 'name': 'fred' };
 * _.identity(object) === object;
 * // => true
 */
function identity(value) {
  return value;
}

module.exports = identity;


/***/ }),
/* 45 */
/***/ (function(module, exports, __webpack_require__) {

/* WEBPACK VAR INJECTION */(function(global) {/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var isNative = __webpack_require__(7);

/** Used to detect functions containing a `this` reference */
var reThis = /\bthis\b/;

/**
 * An object used to flag environments features.
 *
 * @static
 * @memberOf _
 * @type Object
 */
var support = {};

/**
 * Detect if functions can be decompiled by `Function#toString`
 * (all but PS3 and older Opera mobile browsers & avoided in Windows 8 apps).
 *
 * @memberOf _.support
 * @type boolean
 */
support.funcDecomp = !isNative(global.WinRTError) && reThis.test(function() { return this; });

/**
 * Detect if `Function#name` is supported (all but IE).
 *
 * @memberOf _.support
 * @type boolean
 */
support.funcNames = typeof Function.name == 'string';

module.exports = support;

/* WEBPACK VAR INJECTION */}.call(exports, __webpack_require__(2)))

/***/ }),
/* 46 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var isNative = __webpack_require__(7),
    isObject = __webpack_require__(9),
    shimKeys = __webpack_require__(47);

/* Native method shortcuts for methods with the same name as other `lodash` methods */
var nativeKeys = isNative(nativeKeys = Object.keys) && nativeKeys;

/**
 * Creates an array composed of the own enumerable property names of an object.
 *
 * @static
 * @memberOf _
 * @category Objects
 * @param {Object} object The object to inspect.
 * @returns {Array} Returns an array of property names.
 * @example
 *
 * _.keys({ 'one': 1, 'two': 2, 'three': 3 });
 * // => ['one', 'two', 'three'] (property order is not guaranteed across environments)
 */
var keys = !nativeKeys ? shimKeys : function(object) {
  if (!isObject(object)) {
    return [];
  }
  return nativeKeys(object);
};

module.exports = keys;


/***/ }),
/* 47 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var objectTypes = __webpack_require__(12);

/** Used for native method references */
var objectProto = Object.prototype;

/** Native method shortcuts */
var hasOwnProperty = objectProto.hasOwnProperty;

/**
 * A fallback implementation of `Object.keys` which produces an array of the
 * given object's own enumerable property names.
 *
 * @private
 * @type Function
 * @param {Object} object The object to inspect.
 * @returns {Array} Returns an array of property names.
 */
var shimKeys = function(object) {
  var index, iterable = object, result = [];
  if (!iterable) return result;
  if (!(objectTypes[typeof object])) return result;
    for (index in iterable) {
      if (hasOwnProperty.call(iterable, index)) {
        result.push(index);
      }
    }
  return result
};

module.exports = shimKeys;


/***/ }),
/* 48 */
/***/ (function(module, exports, __webpack_require__) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */
var isNative = __webpack_require__(7);

/** `Object#toString` result shortcuts */
var arrayClass = '[object Array]';

/** Used for native method references */
var objectProto = Object.prototype;

/** Used to resolve the internal [[Class]] of values */
var toString = objectProto.toString;

/* Native method shortcuts for methods with the same name as other `lodash` methods */
var nativeIsArray = isNative(nativeIsArray = Array.isArray) && nativeIsArray;

/**
 * Checks if `value` is an array.
 *
 * @static
 * @memberOf _
 * @type Function
 * @category Objects
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if the `value` is an array, else `false`.
 * @example
 *
 * (function() { return _.isArray(arguments); })();
 * // => false
 *
 * _.isArray([1, 2, 3]);
 * // => true
 */
var isArray = nativeIsArray || function(value) {
  return value && typeof value == 'object' && typeof value.length == 'number' &&
    toString.call(value) == arrayClass || false;
};

module.exports = isArray;


/***/ }),
/* 49 */
/***/ (function(module, exports) {

/**
 * Lo-Dash 2.4.1 (Custom Build) <http://lodash.com/>
 * Build: `lodash modularize modern exports="npm" -o ./npm/`
 * Copyright 2012-2013 The Dojo Foundation <http://dojofoundation.org/>
 * Based on Underscore.js 1.5.2 <http://underscorejs.org/LICENSE>
 * Copyright 2009-2013 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 * Available under MIT license <http://lodash.com/license>
 */

/** `Object#toString` result shortcuts */
var stringClass = '[object String]';

/** Used for native method references */
var objectProto = Object.prototype;

/** Used to resolve the internal [[Class]] of values */
var toString = objectProto.toString;

/**
 * Checks if `value` is a string.
 *
 * @static
 * @memberOf _
 * @category Objects
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if the `value` is a string, else `false`.
 * @example
 *
 * _.isString('fred');
 * // => true
 */
function isString(value) {
  return typeof value == 'string' ||
    value && typeof value == 'object' && toString.call(value) == stringClass || false;
}

module.exports = isString;


/***/ }),
/* 50 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


var speculative = __webpack_require__(30),
	preconnectedHrefs = {};

// Create a <link> tag per href that needs to preconnect
module.exports = function(href) {
	if (typeof href !== 'string') {
		throw new Error('Preconnect expects a string, recived ' + href);
	}

	var hrefEmpty = !href,
		hrefAlreadyPreconnected = preconnectedHrefs[href];

	if (hrefEmpty || hrefAlreadyPreconnected) {
		return;
	}

	preconnectedHrefs[href] = true;

	speculative({
		href: href,
		rel: 'preconnect'
	});

};


/***/ }),
/* 51 */
/***/ (function(module, exports, __webpack_require__) {

/* WEBPACK VAR INJECTION */(function(global, module) {/**
 * Lodash (Custom Build) <https://lodash.com/>
 * Build: `lodash modularize exports="npm" -o ./`
 * Copyright OpenJS Foundation and other contributors <https://openjsf.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */
var reInterpolate = __webpack_require__(32),
    templateSettings = __webpack_require__(27);

/** Used to detect hot functions by number of calls within a span of milliseconds. */
var HOT_COUNT = 800,
    HOT_SPAN = 16;

/** Used as references for various `Number` constants. */
var INFINITY = 1 / 0,
    MAX_SAFE_INTEGER = 9007199254740991;

/** `Object#toString` result references. */
var argsTag = '[object Arguments]',
    arrayTag = '[object Array]',
    asyncTag = '[object AsyncFunction]',
    boolTag = '[object Boolean]',
    dateTag = '[object Date]',
    domExcTag = '[object DOMException]',
    errorTag = '[object Error]',
    funcTag = '[object Function]',
    genTag = '[object GeneratorFunction]',
    mapTag = '[object Map]',
    numberTag = '[object Number]',
    nullTag = '[object Null]',
    objectTag = '[object Object]',
    proxyTag = '[object Proxy]',
    regexpTag = '[object RegExp]',
    setTag = '[object Set]',
    stringTag = '[object String]',
    symbolTag = '[object Symbol]',
    undefinedTag = '[object Undefined]',
    weakMapTag = '[object WeakMap]';

var arrayBufferTag = '[object ArrayBuffer]',
    dataViewTag = '[object DataView]',
    float32Tag = '[object Float32Array]',
    float64Tag = '[object Float64Array]',
    int8Tag = '[object Int8Array]',
    int16Tag = '[object Int16Array]',
    int32Tag = '[object Int32Array]',
    uint8Tag = '[object Uint8Array]',
    uint8ClampedTag = '[object Uint8ClampedArray]',
    uint16Tag = '[object Uint16Array]',
    uint32Tag = '[object Uint32Array]';

/** Used to match empty string literals in compiled template source. */
var reEmptyStringLeading = /\b__p \+= '';/g,
    reEmptyStringMiddle = /\b(__p \+=) '' \+/g,
    reEmptyStringTrailing = /(__e\(.*?\)|\b__t\)) \+\n'';/g;

/**
 * Used to match `RegExp`
 * [syntax characters](http://ecma-international.org/ecma-262/7.0/#sec-patterns).
 */
var reRegExpChar = /[\\^$.*+?()[\]{}|]/g;

/**
 * Used to match
 * [ES template delimiters](http://ecma-international.org/ecma-262/7.0/#sec-template-literal-lexical-components).
 */
var reEsTemplate = /\$\{([^\\}]*(?:\\.[^\\}]*)*)\}/g;

/** Used to detect host constructors (Safari). */
var reIsHostCtor = /^\[object .+?Constructor\]$/;

/** Used to detect unsigned integer values. */
var reIsUint = /^(?:0|[1-9]\d*)$/;

/** Used to ensure capturing order of template delimiters. */
var reNoMatch = /($^)/;

/** Used to match unescaped characters in compiled string literals. */
var reUnescapedString = /['\n\r\u2028\u2029\\]/g;

/** Used to identify `toStringTag` values of typed arrays. */
var typedArrayTags = {};
typedArrayTags[float32Tag] = typedArrayTags[float64Tag] =
typedArrayTags[int8Tag] = typedArrayTags[int16Tag] =
typedArrayTags[int32Tag] = typedArrayTags[uint8Tag] =
typedArrayTags[uint8ClampedTag] = typedArrayTags[uint16Tag] =
typedArrayTags[uint32Tag] = true;
typedArrayTags[argsTag] = typedArrayTags[arrayTag] =
typedArrayTags[arrayBufferTag] = typedArrayTags[boolTag] =
typedArrayTags[dataViewTag] = typedArrayTags[dateTag] =
typedArrayTags[errorTag] = typedArrayTags[funcTag] =
typedArrayTags[mapTag] = typedArrayTags[numberTag] =
typedArrayTags[objectTag] = typedArrayTags[regexpTag] =
typedArrayTags[setTag] = typedArrayTags[stringTag] =
typedArrayTags[weakMapTag] = false;

/** Used to escape characters for inclusion in compiled string literals. */
var stringEscapes = {
  '\\': '\\',
  "'": "'",
  '\n': 'n',
  '\r': 'r',
  '\u2028': 'u2028',
  '\u2029': 'u2029'
};

/** Detect free variable `global` from Node.js. */
var freeGlobal = typeof global == 'object' && global && global.Object === Object && global;

/** Detect free variable `self`. */
var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

/** Used as a reference to the global object. */
var root = freeGlobal || freeSelf || Function('return this')();

/** Detect free variable `exports`. */
var freeExports = typeof exports == 'object' && exports && !exports.nodeType && exports;

/** Detect free variable `module`. */
var freeModule = freeExports && typeof module == 'object' && module && !module.nodeType && module;

/** Detect the popular CommonJS extension `module.exports`. */
var moduleExports = freeModule && freeModule.exports === freeExports;

/** Detect free variable `process` from Node.js. */
var freeProcess = moduleExports && freeGlobal.process;

/** Used to access faster Node.js helpers. */
var nodeUtil = (function() {
  try {
    // Use `util.types` for Node.js 10+.
    var types = freeModule && freeModule.require && freeModule.require('util').types;

    if (types) {
      return types;
    }

    // Legacy `process.binding('util')` for Node.js < 10.
    return freeProcess && freeProcess.binding && freeProcess.binding('util');
  } catch (e) {}
}());

/* Node.js helper references. */
var nodeIsTypedArray = nodeUtil && nodeUtil.isTypedArray;

/**
 * A faster alternative to `Function#apply`, this function invokes `func`
 * with the `this` binding of `thisArg` and the arguments of `args`.
 *
 * @private
 * @param {Function} func The function to invoke.
 * @param {*} thisArg The `this` binding of `func`.
 * @param {Array} args The arguments to invoke `func` with.
 * @returns {*} Returns the result of `func`.
 */
function apply(func, thisArg, args) {
  switch (args.length) {
    case 0: return func.call(thisArg);
    case 1: return func.call(thisArg, args[0]);
    case 2: return func.call(thisArg, args[0], args[1]);
    case 3: return func.call(thisArg, args[0], args[1], args[2]);
  }
  return func.apply(thisArg, args);
}

/**
 * A specialized version of `_.map` for arrays without support for iteratee
 * shorthands.
 *
 * @private
 * @param {Array} [array] The array to iterate over.
 * @param {Function} iteratee The function invoked per iteration.
 * @returns {Array} Returns the new mapped array.
 */
function arrayMap(array, iteratee) {
  var index = -1,
      length = array == null ? 0 : array.length,
      result = Array(length);

  while (++index < length) {
    result[index] = iteratee(array[index], index, array);
  }
  return result;
}

/**
 * The base implementation of `_.times` without support for iteratee shorthands
 * or max array length checks.
 *
 * @private
 * @param {number} n The number of times to invoke `iteratee`.
 * @param {Function} iteratee The function invoked per iteration.
 * @returns {Array} Returns the array of results.
 */
function baseTimes(n, iteratee) {
  var index = -1,
      result = Array(n);

  while (++index < n) {
    result[index] = iteratee(index);
  }
  return result;
}

/**
 * The base implementation of `_.unary` without support for storing metadata.
 *
 * @private
 * @param {Function} func The function to cap arguments for.
 * @returns {Function} Returns the new capped function.
 */
function baseUnary(func) {
  return function(value) {
    return func(value);
  };
}

/**
 * The base implementation of `_.values` and `_.valuesIn` which creates an
 * array of `object` property values corresponding to the property names
 * of `props`.
 *
 * @private
 * @param {Object} object The object to query.
 * @param {Array} props The property names to get values for.
 * @returns {Object} Returns the array of property values.
 */
function baseValues(object, props) {
  return arrayMap(props, function(key) {
    return object[key];
  });
}

/**
 * Used by `_.template` to escape characters for inclusion in compiled string literals.
 *
 * @private
 * @param {string} chr The matched character to escape.
 * @returns {string} Returns the escaped character.
 */
function escapeStringChar(chr) {
  return '\\' + stringEscapes[chr];
}

/**
 * Gets the value at `key` of `object`.
 *
 * @private
 * @param {Object} [object] The object to query.
 * @param {string} key The key of the property to get.
 * @returns {*} Returns the property value.
 */
function getValue(object, key) {
  return object == null ? undefined : object[key];
}

/**
 * Creates a unary function that invokes `func` with its argument transformed.
 *
 * @private
 * @param {Function} func The function to wrap.
 * @param {Function} transform The argument transform.
 * @returns {Function} Returns the new function.
 */
function overArg(func, transform) {
  return function(arg) {
    return func(transform(arg));
  };
}

/** Used for built-in method references. */
var funcProto = Function.prototype,
    objectProto = Object.prototype;

/** Used to detect overreaching core-js shims. */
var coreJsData = root['__core-js_shared__'];

/** Used to resolve the decompiled source of functions. */
var funcToString = funcProto.toString;

/** Used to check objects for own properties. */
var hasOwnProperty = objectProto.hasOwnProperty;

/** Used to detect methods masquerading as native. */
var maskSrcKey = (function() {
  var uid = /[^.]+$/.exec(coreJsData && coreJsData.keys && coreJsData.keys.IE_PROTO || '');
  return uid ? ('Symbol(src)_1.' + uid) : '';
}());

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
 * of values.
 */
var nativeObjectToString = objectProto.toString;

/** Used to infer the `Object` constructor. */
var objectCtorString = funcToString.call(Object);

/** Used to detect if a method is native. */
var reIsNative = RegExp('^' +
  funcToString.call(hasOwnProperty).replace(reRegExpChar, '\\$&')
  .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
);

/** Built-in value references. */
var Buffer = moduleExports ? root.Buffer : undefined,
    Symbol = root.Symbol,
    getPrototype = overArg(Object.getPrototypeOf, Object),
    propertyIsEnumerable = objectProto.propertyIsEnumerable,
    symToStringTag = Symbol ? Symbol.toStringTag : undefined;

var defineProperty = (function() {
  try {
    var func = getNative(Object, 'defineProperty');
    func({}, '', {});
    return func;
  } catch (e) {}
}());

/* Built-in method references for those with the same name as other `lodash` methods. */
var nativeIsBuffer = Buffer ? Buffer.isBuffer : undefined,
    nativeKeys = overArg(Object.keys, Object),
    nativeMax = Math.max,
    nativeNow = Date.now;

/** Used to convert symbols to primitives and strings. */
var symbolProto = Symbol ? Symbol.prototype : undefined,
    symbolToString = symbolProto ? symbolProto.toString : undefined;

/**
 * Creates an array of the enumerable property names of the array-like `value`.
 *
 * @private
 * @param {*} value The value to query.
 * @param {boolean} inherited Specify returning inherited property names.
 * @returns {Array} Returns the array of property names.
 */
function arrayLikeKeys(value, inherited) {
  var isArr = isArray(value),
      isArg = !isArr && isArguments(value),
      isBuff = !isArr && !isArg && isBuffer(value),
      isType = !isArr && !isArg && !isBuff && isTypedArray(value),
      skipIndexes = isArr || isArg || isBuff || isType,
      result = skipIndexes ? baseTimes(value.length, String) : [],
      length = result.length;

  for (var key in value) {
    if ((inherited || hasOwnProperty.call(value, key)) &&
        !(skipIndexes && (
           // Safari 9 has enumerable `arguments.length` in strict mode.
           key == 'length' ||
           // Node.js 0.10 has enumerable non-index properties on buffers.
           (isBuff && (key == 'offset' || key == 'parent')) ||
           // PhantomJS 2 has enumerable non-index properties on typed arrays.
           (isType && (key == 'buffer' || key == 'byteLength' || key == 'byteOffset')) ||
           // Skip index properties.
           isIndex(key, length)
        ))) {
      result.push(key);
    }
  }
  return result;
}

/**
 * Assigns `value` to `key` of `object` if the existing value is not equivalent
 * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
 * for equality comparisons.
 *
 * @private
 * @param {Object} object The object to modify.
 * @param {string} key The key of the property to assign.
 * @param {*} value The value to assign.
 */
function assignValue(object, key, value) {
  var objValue = object[key];
  if (!(hasOwnProperty.call(object, key) && eq(objValue, value)) ||
      (value === undefined && !(key in object))) {
    baseAssignValue(object, key, value);
  }
}

/**
 * The base implementation of `assignValue` and `assignMergeValue` without
 * value checks.
 *
 * @private
 * @param {Object} object The object to modify.
 * @param {string} key The key of the property to assign.
 * @param {*} value The value to assign.
 */
function baseAssignValue(object, key, value) {
  if (key == '__proto__' && defineProperty) {
    defineProperty(object, key, {
      'configurable': true,
      'enumerable': true,
      'value': value,
      'writable': true
    });
  } else {
    object[key] = value;
  }
}

/**
 * The base implementation of `getTag` without fallbacks for buggy environments.
 *
 * @private
 * @param {*} value The value to query.
 * @returns {string} Returns the `toStringTag`.
 */
function baseGetTag(value) {
  if (value == null) {
    return value === undefined ? undefinedTag : nullTag;
  }
  return (symToStringTag && symToStringTag in Object(value))
    ? getRawTag(value)
    : objectToString(value);
}

/**
 * The base implementation of `_.isArguments`.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an `arguments` object,
 */
function baseIsArguments(value) {
  return isObjectLike(value) && baseGetTag(value) == argsTag;
}

/**
 * The base implementation of `_.isNative` without bad shim checks.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a native function,
 *  else `false`.
 */
function baseIsNative(value) {
  if (!isObject(value) || isMasked(value)) {
    return false;
  }
  var pattern = isFunction(value) ? reIsNative : reIsHostCtor;
  return pattern.test(toSource(value));
}

/**
 * The base implementation of `_.isTypedArray` without Node.js optimizations.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
 */
function baseIsTypedArray(value) {
  return isObjectLike(value) &&
    isLength(value.length) && !!typedArrayTags[baseGetTag(value)];
}

/**
 * The base implementation of `_.keys` which doesn't treat sparse arrays as dense.
 *
 * @private
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property names.
 */
function baseKeys(object) {
  if (!isPrototype(object)) {
    return nativeKeys(object);
  }
  var result = [];
  for (var key in Object(object)) {
    if (hasOwnProperty.call(object, key) && key != 'constructor') {
      result.push(key);
    }
  }
  return result;
}

/**
 * The base implementation of `_.keysIn` which doesn't treat sparse arrays as dense.
 *
 * @private
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property names.
 */
function baseKeysIn(object) {
  if (!isObject(object)) {
    return nativeKeysIn(object);
  }
  var isProto = isPrototype(object),
      result = [];

  for (var key in object) {
    if (!(key == 'constructor' && (isProto || !hasOwnProperty.call(object, key)))) {
      result.push(key);
    }
  }
  return result;
}

/**
 * The base implementation of `_.rest` which doesn't validate or coerce arguments.
 *
 * @private
 * @param {Function} func The function to apply a rest parameter to.
 * @param {number} [start=func.length-1] The start position of the rest parameter.
 * @returns {Function} Returns the new function.
 */
function baseRest(func, start) {
  return setToString(overRest(func, start, identity), func + '');
}

/**
 * The base implementation of `setToString` without support for hot loop shorting.
 *
 * @private
 * @param {Function} func The function to modify.
 * @param {Function} string The `toString` result.
 * @returns {Function} Returns `func`.
 */
var baseSetToString = !defineProperty ? identity : function(func, string) {
  return defineProperty(func, 'toString', {
    'configurable': true,
    'enumerable': false,
    'value': constant(string),
    'writable': true
  });
};

/**
 * The base implementation of `_.toString` which doesn't convert nullish
 * values to empty strings.
 *
 * @private
 * @param {*} value The value to process.
 * @returns {string} Returns the string.
 */
function baseToString(value) {
  // Exit early for strings to avoid a performance hit in some environments.
  if (typeof value == 'string') {
    return value;
  }
  if (isArray(value)) {
    // Recursively convert values (susceptible to call stack limits).
    return arrayMap(value, baseToString) + '';
  }
  if (isSymbol(value)) {
    return symbolToString ? symbolToString.call(value) : '';
  }
  var result = (value + '');
  return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
}

/**
 * Copies properties of `source` to `object`.
 *
 * @private
 * @param {Object} source The object to copy properties from.
 * @param {Array} props The property identifiers to copy.
 * @param {Object} [object={}] The object to copy properties to.
 * @param {Function} [customizer] The function to customize copied values.
 * @returns {Object} Returns `object`.
 */
function copyObject(source, props, object, customizer) {
  var isNew = !object;
  object || (object = {});

  var index = -1,
      length = props.length;

  while (++index < length) {
    var key = props[index];

    var newValue = customizer
      ? customizer(object[key], source[key], key, object, source)
      : undefined;

    if (newValue === undefined) {
      newValue = source[key];
    }
    if (isNew) {
      baseAssignValue(object, key, newValue);
    } else {
      assignValue(object, key, newValue);
    }
  }
  return object;
}

/**
 * Creates a function like `_.assign`.
 *
 * @private
 * @param {Function} assigner The function to assign values.
 * @returns {Function} Returns the new assigner function.
 */
function createAssigner(assigner) {
  return baseRest(function(object, sources) {
    var index = -1,
        length = sources.length,
        customizer = length > 1 ? sources[length - 1] : undefined,
        guard = length > 2 ? sources[2] : undefined;

    customizer = (assigner.length > 3 && typeof customizer == 'function')
      ? (length--, customizer)
      : undefined;

    if (guard && isIterateeCall(sources[0], sources[1], guard)) {
      customizer = length < 3 ? undefined : customizer;
      length = 1;
    }
    object = Object(object);
    while (++index < length) {
      var source = sources[index];
      if (source) {
        assigner(object, source, index, customizer);
      }
    }
    return object;
  });
}

/**
 * Used by `_.defaults` to customize its `_.assignIn` use to assign properties
 * of source objects to the destination object for all destination properties
 * that resolve to `undefined`.
 *
 * @private
 * @param {*} objValue The destination value.
 * @param {*} srcValue The source value.
 * @param {string} key The key of the property to assign.
 * @param {Object} object The parent object of `objValue`.
 * @returns {*} Returns the value to assign.
 */
function customDefaultsAssignIn(objValue, srcValue, key, object) {
  if (objValue === undefined ||
      (eq(objValue, objectProto[key]) && !hasOwnProperty.call(object, key))) {
    return srcValue;
  }
  return objValue;
}

/**
 * Gets the native function at `key` of `object`.
 *
 * @private
 * @param {Object} object The object to query.
 * @param {string} key The key of the method to get.
 * @returns {*} Returns the function if it's native, else `undefined`.
 */
function getNative(object, key) {
  var value = getValue(object, key);
  return baseIsNative(value) ? value : undefined;
}

/**
 * A specialized version of `baseGetTag` which ignores `Symbol.toStringTag` values.
 *
 * @private
 * @param {*} value The value to query.
 * @returns {string} Returns the raw `toStringTag`.
 */
function getRawTag(value) {
  var isOwn = hasOwnProperty.call(value, symToStringTag),
      tag = value[symToStringTag];

  try {
    value[symToStringTag] = undefined;
    var unmasked = true;
  } catch (e) {}

  var result = nativeObjectToString.call(value);
  if (unmasked) {
    if (isOwn) {
      value[symToStringTag] = tag;
    } else {
      delete value[symToStringTag];
    }
  }
  return result;
}

/**
 * Checks if `value` is a valid array-like index.
 *
 * @private
 * @param {*} value The value to check.
 * @param {number} [length=MAX_SAFE_INTEGER] The upper bounds of a valid index.
 * @returns {boolean} Returns `true` if `value` is a valid index, else `false`.
 */
function isIndex(value, length) {
  var type = typeof value;
  length = length == null ? MAX_SAFE_INTEGER : length;

  return !!length &&
    (type == 'number' ||
      (type != 'symbol' && reIsUint.test(value))) &&
        (value > -1 && value % 1 == 0 && value < length);
}

/**
 * Checks if the given arguments are from an iteratee call.
 *
 * @private
 * @param {*} value The potential iteratee value argument.
 * @param {*} index The potential iteratee index or key argument.
 * @param {*} object The potential iteratee object argument.
 * @returns {boolean} Returns `true` if the arguments are from an iteratee call,
 *  else `false`.
 */
function isIterateeCall(value, index, object) {
  if (!isObject(object)) {
    return false;
  }
  var type = typeof index;
  if (type == 'number'
        ? (isArrayLike(object) && isIndex(index, object.length))
        : (type == 'string' && index in object)
      ) {
    return eq(object[index], value);
  }
  return false;
}

/**
 * Checks if `func` has its source masked.
 *
 * @private
 * @param {Function} func The function to check.
 * @returns {boolean} Returns `true` if `func` is masked, else `false`.
 */
function isMasked(func) {
  return !!maskSrcKey && (maskSrcKey in func);
}

/**
 * Checks if `value` is likely a prototype object.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a prototype, else `false`.
 */
function isPrototype(value) {
  var Ctor = value && value.constructor,
      proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto;

  return value === proto;
}

/**
 * This function is like
 * [`Object.keys`](http://ecma-international.org/ecma-262/7.0/#sec-object.keys)
 * except that it includes inherited enumerable properties.
 *
 * @private
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property names.
 */
function nativeKeysIn(object) {
  var result = [];
  if (object != null) {
    for (var key in Object(object)) {
      result.push(key);
    }
  }
  return result;
}

/**
 * Converts `value` to a string using `Object.prototype.toString`.
 *
 * @private
 * @param {*} value The value to convert.
 * @returns {string} Returns the converted string.
 */
function objectToString(value) {
  return nativeObjectToString.call(value);
}

/**
 * A specialized version of `baseRest` which transforms the rest array.
 *
 * @private
 * @param {Function} func The function to apply a rest parameter to.
 * @param {number} [start=func.length-1] The start position of the rest parameter.
 * @param {Function} transform The rest array transform.
 * @returns {Function} Returns the new function.
 */
function overRest(func, start, transform) {
  start = nativeMax(start === undefined ? (func.length - 1) : start, 0);
  return function() {
    var args = arguments,
        index = -1,
        length = nativeMax(args.length - start, 0),
        array = Array(length);

    while (++index < length) {
      array[index] = args[start + index];
    }
    index = -1;
    var otherArgs = Array(start + 1);
    while (++index < start) {
      otherArgs[index] = args[index];
    }
    otherArgs[start] = transform(array);
    return apply(func, this, otherArgs);
  };
}

/**
 * Sets the `toString` method of `func` to return `string`.
 *
 * @private
 * @param {Function} func The function to modify.
 * @param {Function} string The `toString` result.
 * @returns {Function} Returns `func`.
 */
var setToString = shortOut(baseSetToString);

/**
 * Creates a function that'll short out and invoke `identity` instead
 * of `func` when it's called `HOT_COUNT` or more times in `HOT_SPAN`
 * milliseconds.
 *
 * @private
 * @param {Function} func The function to restrict.
 * @returns {Function} Returns the new shortable function.
 */
function shortOut(func) {
  var count = 0,
      lastCalled = 0;

  return function() {
    var stamp = nativeNow(),
        remaining = HOT_SPAN - (stamp - lastCalled);

    lastCalled = stamp;
    if (remaining > 0) {
      if (++count >= HOT_COUNT) {
        return arguments[0];
      }
    } else {
      count = 0;
    }
    return func.apply(undefined, arguments);
  };
}

/**
 * Converts `func` to its source code.
 *
 * @private
 * @param {Function} func The function to convert.
 * @returns {string} Returns the source code.
 */
function toSource(func) {
  if (func != null) {
    try {
      return funcToString.call(func);
    } catch (e) {}
    try {
      return (func + '');
    } catch (e) {}
  }
  return '';
}

/**
 * Performs a
 * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
 * comparison between two values to determine if they are equivalent.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to compare.
 * @param {*} other The other value to compare.
 * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
 * @example
 *
 * var object = { 'a': 1 };
 * var other = { 'a': 1 };
 *
 * _.eq(object, object);
 * // => true
 *
 * _.eq(object, other);
 * // => false
 *
 * _.eq('a', 'a');
 * // => true
 *
 * _.eq('a', Object('a'));
 * // => false
 *
 * _.eq(NaN, NaN);
 * // => true
 */
function eq(value, other) {
  return value === other || (value !== value && other !== other);
}

/**
 * Checks if `value` is likely an `arguments` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an `arguments` object,
 *  else `false`.
 * @example
 *
 * _.isArguments(function() { return arguments; }());
 * // => true
 *
 * _.isArguments([1, 2, 3]);
 * // => false
 */
var isArguments = baseIsArguments(function() { return arguments; }()) ? baseIsArguments : function(value) {
  return isObjectLike(value) && hasOwnProperty.call(value, 'callee') &&
    !propertyIsEnumerable.call(value, 'callee');
};

/**
 * Checks if `value` is classified as an `Array` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an array, else `false`.
 * @example
 *
 * _.isArray([1, 2, 3]);
 * // => true
 *
 * _.isArray(document.body.children);
 * // => false
 *
 * _.isArray('abc');
 * // => false
 *
 * _.isArray(_.noop);
 * // => false
 */
var isArray = Array.isArray;

/**
 * Checks if `value` is array-like. A value is considered array-like if it's
 * not a function and has a `value.length` that's an integer greater than or
 * equal to `0` and less than or equal to `Number.MAX_SAFE_INTEGER`.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is array-like, else `false`.
 * @example
 *
 * _.isArrayLike([1, 2, 3]);
 * // => true
 *
 * _.isArrayLike(document.body.children);
 * // => true
 *
 * _.isArrayLike('abc');
 * // => true
 *
 * _.isArrayLike(_.noop);
 * // => false
 */
function isArrayLike(value) {
  return value != null && isLength(value.length) && !isFunction(value);
}

/**
 * Checks if `value` is a buffer.
 *
 * @static
 * @memberOf _
 * @since 4.3.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a buffer, else `false`.
 * @example
 *
 * _.isBuffer(new Buffer(2));
 * // => true
 *
 * _.isBuffer(new Uint8Array(2));
 * // => false
 */
var isBuffer = nativeIsBuffer || stubFalse;

/**
 * Checks if `value` is an `Error`, `EvalError`, `RangeError`, `ReferenceError`,
 * `SyntaxError`, `TypeError`, or `URIError` object.
 *
 * @static
 * @memberOf _
 * @since 3.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an error object, else `false`.
 * @example
 *
 * _.isError(new Error);
 * // => true
 *
 * _.isError(Error);
 * // => false
 */
function isError(value) {
  if (!isObjectLike(value)) {
    return false;
  }
  var tag = baseGetTag(value);
  return tag == errorTag || tag == domExcTag ||
    (typeof value.message == 'string' && typeof value.name == 'string' && !isPlainObject(value));
}

/**
 * Checks if `value` is classified as a `Function` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a function, else `false`.
 * @example
 *
 * _.isFunction(_);
 * // => true
 *
 * _.isFunction(/abc/);
 * // => false
 */
function isFunction(value) {
  if (!isObject(value)) {
    return false;
  }
  // The use of `Object#toString` avoids issues with the `typeof` operator
  // in Safari 9 which returns 'object' for typed arrays and other constructors.
  var tag = baseGetTag(value);
  return tag == funcTag || tag == genTag || tag == asyncTag || tag == proxyTag;
}

/**
 * Checks if `value` is a valid array-like length.
 *
 * **Note:** This method is loosely based on
 * [`ToLength`](http://ecma-international.org/ecma-262/7.0/#sec-tolength).
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a valid length, else `false`.
 * @example
 *
 * _.isLength(3);
 * // => true
 *
 * _.isLength(Number.MIN_VALUE);
 * // => false
 *
 * _.isLength(Infinity);
 * // => false
 *
 * _.isLength('3');
 * // => false
 */
function isLength(value) {
  return typeof value == 'number' &&
    value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER;
}

/**
 * Checks if `value` is the
 * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
 * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an object, else `false`.
 * @example
 *
 * _.isObject({});
 * // => true
 *
 * _.isObject([1, 2, 3]);
 * // => true
 *
 * _.isObject(_.noop);
 * // => true
 *
 * _.isObject(null);
 * // => false
 */
function isObject(value) {
  var type = typeof value;
  return value != null && (type == 'object' || type == 'function');
}

/**
 * Checks if `value` is object-like. A value is object-like if it's not `null`
 * and has a `typeof` result of "object".
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
 * @example
 *
 * _.isObjectLike({});
 * // => true
 *
 * _.isObjectLike([1, 2, 3]);
 * // => true
 *
 * _.isObjectLike(_.noop);
 * // => false
 *
 * _.isObjectLike(null);
 * // => false
 */
function isObjectLike(value) {
  return value != null && typeof value == 'object';
}

/**
 * Checks if `value` is a plain object, that is, an object created by the
 * `Object` constructor or one with a `[[Prototype]]` of `null`.
 *
 * @static
 * @memberOf _
 * @since 0.8.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a plain object, else `false`.
 * @example
 *
 * function Foo() {
 *   this.a = 1;
 * }
 *
 * _.isPlainObject(new Foo);
 * // => false
 *
 * _.isPlainObject([1, 2, 3]);
 * // => false
 *
 * _.isPlainObject({ 'x': 0, 'y': 0 });
 * // => true
 *
 * _.isPlainObject(Object.create(null));
 * // => true
 */
function isPlainObject(value) {
  if (!isObjectLike(value) || baseGetTag(value) != objectTag) {
    return false;
  }
  var proto = getPrototype(value);
  if (proto === null) {
    return true;
  }
  var Ctor = hasOwnProperty.call(proto, 'constructor') && proto.constructor;
  return typeof Ctor == 'function' && Ctor instanceof Ctor &&
    funcToString.call(Ctor) == objectCtorString;
}

/**
 * Checks if `value` is classified as a `Symbol` primitive or object.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
 * @example
 *
 * _.isSymbol(Symbol.iterator);
 * // => true
 *
 * _.isSymbol('abc');
 * // => false
 */
function isSymbol(value) {
  return typeof value == 'symbol' ||
    (isObjectLike(value) && baseGetTag(value) == symbolTag);
}

/**
 * Checks if `value` is classified as a typed array.
 *
 * @static
 * @memberOf _
 * @since 3.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
 * @example
 *
 * _.isTypedArray(new Uint8Array);
 * // => true
 *
 * _.isTypedArray([]);
 * // => false
 */
var isTypedArray = nodeIsTypedArray ? baseUnary(nodeIsTypedArray) : baseIsTypedArray;

/**
 * Converts `value` to a string. An empty string is returned for `null`
 * and `undefined` values. The sign of `-0` is preserved.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to convert.
 * @returns {string} Returns the converted string.
 * @example
 *
 * _.toString(null);
 * // => ''
 *
 * _.toString(-0);
 * // => '-0'
 *
 * _.toString([1, 2, 3]);
 * // => '1,2,3'
 */
function toString(value) {
  return value == null ? '' : baseToString(value);
}

/**
 * This method is like `_.assignIn` except that it accepts `customizer`
 * which is invoked to produce the assigned values. If `customizer` returns
 * `undefined`, assignment is handled by the method instead. The `customizer`
 * is invoked with five arguments: (objValue, srcValue, key, object, source).
 *
 * **Note:** This method mutates `object`.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @alias extendWith
 * @category Object
 * @param {Object} object The destination object.
 * @param {...Object} sources The source objects.
 * @param {Function} [customizer] The function to customize assigned values.
 * @returns {Object} Returns `object`.
 * @see _.assignWith
 * @example
 *
 * function customizer(objValue, srcValue) {
 *   return _.isUndefined(objValue) ? srcValue : objValue;
 * }
 *
 * var defaults = _.partialRight(_.assignInWith, customizer);
 *
 * defaults({ 'a': 1 }, { 'b': 2 }, { 'a': 3 });
 * // => { 'a': 1, 'b': 2 }
 */
var assignInWith = createAssigner(function(object, source, srcIndex, customizer) {
  copyObject(source, keysIn(source), object, customizer);
});

/**
 * Creates an array of the own enumerable property names of `object`.
 *
 * **Note:** Non-object values are coerced to objects. See the
 * [ES spec](http://ecma-international.org/ecma-262/7.0/#sec-object.keys)
 * for more details.
 *
 * @static
 * @since 0.1.0
 * @memberOf _
 * @category Object
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property names.
 * @example
 *
 * function Foo() {
 *   this.a = 1;
 *   this.b = 2;
 * }
 *
 * Foo.prototype.c = 3;
 *
 * _.keys(new Foo);
 * // => ['a', 'b'] (iteration order is not guaranteed)
 *
 * _.keys('hi');
 * // => ['0', '1']
 */
function keys(object) {
  return isArrayLike(object) ? arrayLikeKeys(object) : baseKeys(object);
}

/**
 * Creates an array of the own and inherited enumerable property names of `object`.
 *
 * **Note:** Non-object values are coerced to objects.
 *
 * @static
 * @memberOf _
 * @since 3.0.0
 * @category Object
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property names.
 * @example
 *
 * function Foo() {
 *   this.a = 1;
 *   this.b = 2;
 * }
 *
 * Foo.prototype.c = 3;
 *
 * _.keysIn(new Foo);
 * // => ['a', 'b', 'c'] (iteration order is not guaranteed)
 */
function keysIn(object) {
  return isArrayLike(object) ? arrayLikeKeys(object, true) : baseKeysIn(object);
}

/**
 * Creates a compiled template function that can interpolate data properties
 * in "interpolate" delimiters, HTML-escape interpolated data properties in
 * "escape" delimiters, and execute JavaScript in "evaluate" delimiters. Data
 * properties may be accessed as free variables in the template. If a setting
 * object is given, it takes precedence over `_.templateSettings` values.
 *
 * **Note:** In the development build `_.template` utilizes
 * [sourceURLs](http://www.html5rocks.com/en/tutorials/developertools/sourcemaps/#toc-sourceurl)
 * for easier debugging.
 *
 * For more information on precompiling templates see
 * [lodash's custom builds documentation](https://lodash.com/custom-builds).
 *
 * For more information on Chrome extension sandboxes see
 * [Chrome's extensions documentation](https://developer.chrome.com/extensions/sandboxingEval).
 *
 * @static
 * @since 0.1.0
 * @memberOf _
 * @category String
 * @param {string} [string=''] The template string.
 * @param {Object} [options={}] The options object.
 * @param {RegExp} [options.escape=_.templateSettings.escape]
 *  The HTML "escape" delimiter.
 * @param {RegExp} [options.evaluate=_.templateSettings.evaluate]
 *  The "evaluate" delimiter.
 * @param {Object} [options.imports=_.templateSettings.imports]
 *  An object to import into the template as free variables.
 * @param {RegExp} [options.interpolate=_.templateSettings.interpolate]
 *  The "interpolate" delimiter.
 * @param {string} [options.sourceURL='templateSources[n]']
 *  The sourceURL of the compiled template.
 * @param {string} [options.variable='obj']
 *  The data object variable name.
 * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
 * @returns {Function} Returns the compiled template function.
 * @example
 *
 * // Use the "interpolate" delimiter to create a compiled template.
 * var compiled = _.template('hello <%= user %>!');
 * compiled({ 'user': 'fred' });
 * // => 'hello fred!'
 *
 * // Use the HTML "escape" delimiter to escape data property values.
 * var compiled = _.template('<b><%- value %></b>');
 * compiled({ 'value': '<script>' });
 * // => '<b>&lt;script&gt;</b>'
 *
 * // Use the "evaluate" delimiter to execute JavaScript and generate HTML.
 * var compiled = _.template('<% _.forEach(users, function(user) { %><li><%- user %></li><% }); %>');
 * compiled({ 'users': ['fred', 'barney'] });
 * // => '<li>fred</li><li>barney</li>'
 *
 * // Use the internal `print` function in "evaluate" delimiters.
 * var compiled = _.template('<% print("hello " + user); %>!');
 * compiled({ 'user': 'barney' });
 * // => 'hello barney!'
 *
 * // Use the ES template literal delimiter as an "interpolate" delimiter.
 * // Disable support by replacing the "interpolate" delimiter.
 * var compiled = _.template('hello ${ user }!');
 * compiled({ 'user': 'pebbles' });
 * // => 'hello pebbles!'
 *
 * // Use backslashes to treat delimiters as plain text.
 * var compiled = _.template('<%= "\\<%- value %\\>" %>');
 * compiled({ 'value': 'ignored' });
 * // => '<%- value %>'
 *
 * // Use the `imports` option to import `jQuery` as `jq`.
 * var text = '<% jq.each(users, function(user) { %><li><%- user %></li><% }); %>';
 * var compiled = _.template(text, { 'imports': { 'jq': jQuery } });
 * compiled({ 'users': ['fred', 'barney'] });
 * // => '<li>fred</li><li>barney</li>'
 *
 * // Use the `sourceURL` option to specify a custom sourceURL for the template.
 * var compiled = _.template('hello <%= user %>!', { 'sourceURL': '/basic/greeting.jst' });
 * compiled(data);
 * // => Find the source of "greeting.jst" under the Sources tab or Resources panel of the web inspector.
 *
 * // Use the `variable` option to ensure a with-statement isn't used in the compiled template.
 * var compiled = _.template('hi <%= data.user %>!', { 'variable': 'data' });
 * compiled.source;
 * // => function(data) {
 * //   var __t, __p = '';
 * //   __p += 'hi ' + ((__t = ( data.user )) == null ? '' : __t) + '!';
 * //   return __p;
 * // }
 *
 * // Use custom template delimiters.
 * _.templateSettings.interpolate = /{{([\s\S]+?)}}/g;
 * var compiled = _.template('hello {{ user }}!');
 * compiled({ 'user': 'mustache' });
 * // => 'hello mustache!'
 *
 * // Use the `source` property to inline compiled templates for meaningful
 * // line numbers in error messages and stack traces.
 * fs.writeFileSync(path.join(process.cwd(), 'jst.js'), '\
 *   var JST = {\
 *     "main": ' + _.template(mainText).source + '\
 *   };\
 * ');
 */
function template(string, options, guard) {
  // Based on John Resig's `tmpl` implementation
  // (http://ejohn.org/blog/javascript-micro-templating/)
  // and Laura Doktorova's doT.js (https://github.com/olado/doT).
  var settings = templateSettings.imports._.templateSettings || templateSettings;

  if (guard && isIterateeCall(string, options, guard)) {
    options = undefined;
  }
  string = toString(string);
  options = assignInWith({}, options, settings, customDefaultsAssignIn);

  var imports = assignInWith({}, options.imports, settings.imports, customDefaultsAssignIn),
      importsKeys = keys(imports),
      importsValues = baseValues(imports, importsKeys);

  var isEscaping,
      isEvaluating,
      index = 0,
      interpolate = options.interpolate || reNoMatch,
      source = "__p += '";

  // Compile the regexp to match each delimiter.
  var reDelimiters = RegExp(
    (options.escape || reNoMatch).source + '|' +
    interpolate.source + '|' +
    (interpolate === reInterpolate ? reEsTemplate : reNoMatch).source + '|' +
    (options.evaluate || reNoMatch).source + '|$'
  , 'g');

  // Use a sourceURL for easier debugging.
  // The sourceURL gets injected into the source that's eval-ed, so be careful
  // with lookup (in case of e.g. prototype pollution), and strip newlines if any.
  // A newline wouldn't be a valid sourceURL anyway, and it'd enable code injection.
  var sourceURL = hasOwnProperty.call(options, 'sourceURL')
    ? ('//# sourceURL=' +
       (options.sourceURL + '').replace(/[\r\n]/g, ' ') +
       '\n')
    : '';

  string.replace(reDelimiters, function(match, escapeValue, interpolateValue, esTemplateValue, evaluateValue, offset) {
    interpolateValue || (interpolateValue = esTemplateValue);

    // Escape characters that can't be included in string literals.
    source += string.slice(index, offset).replace(reUnescapedString, escapeStringChar);

    // Replace delimiters with snippets.
    if (escapeValue) {
      isEscaping = true;
      source += "' +\n__e(" + escapeValue + ") +\n'";
    }
    if (evaluateValue) {
      isEvaluating = true;
      source += "';\n" + evaluateValue + ";\n__p += '";
    }
    if (interpolateValue) {
      source += "' +\n((__t = (" + interpolateValue + ")) == null ? '' : __t) +\n'";
    }
    index = offset + match.length;

    // The JS engine embedded in Adobe products needs `match` returned in
    // order to produce the correct `offset` value.
    return match;
  });

  source += "';\n";

  // If `variable` is not specified wrap a with-statement around the generated
  // code to add the data object to the top of the scope chain.
  // Like with sourceURL, we take care to not check the option's prototype,
  // as this configuration is a code injection vector.
  var variable = hasOwnProperty.call(options, 'variable') && options.variable;
  if (!variable) {
    source = 'with (obj) {\n' + source + '\n}\n';
  }
  // Cleanup code by stripping empty strings.
  source = (isEvaluating ? source.replace(reEmptyStringLeading, '') : source)
    .replace(reEmptyStringMiddle, '$1')
    .replace(reEmptyStringTrailing, '$1;');

  // Frame code as the function body.
  source = 'function(' + (variable || 'obj') + ') {\n' +
    (variable
      ? ''
      : 'obj || (obj = {});\n'
    ) +
    "var __t, __p = ''" +
    (isEscaping
       ? ', __e = _.escape'
       : ''
    ) +
    (isEvaluating
      ? ', __j = Array.prototype.join;\n' +
        "function print() { __p += __j.call(arguments, '') }\n"
      : ';\n'
    ) +
    source +
    'return __p\n}';

  var result = attempt(function() {
    return Function(importsKeys, sourceURL + 'return ' + source)
      .apply(undefined, importsValues);
  });

  // Provide the compiled function's source by its `toString` method or
  // the `source` property as a convenience for inlining compiled templates.
  result.source = source;
  if (isError(result)) {
    throw result;
  }
  return result;
}

/**
 * Attempts to invoke `func`, returning either the result or the caught error
 * object. Any additional arguments are provided to `func` when it's invoked.
 *
 * @static
 * @memberOf _
 * @since 3.0.0
 * @category Util
 * @param {Function} func The function to attempt.
 * @param {...*} [args] The arguments to invoke `func` with.
 * @returns {*} Returns the `func` result or error object.
 * @example
 *
 * // Avoid throwing errors for invalid selectors.
 * var elements = _.attempt(function(selector) {
 *   return document.querySelectorAll(selector);
 * }, '>_>');
 *
 * if (_.isError(elements)) {
 *   elements = [];
 * }
 */
var attempt = baseRest(function(func, args) {
  try {
    return apply(func, undefined, args);
  } catch (e) {
    return isError(e) ? e : new Error(e);
  }
});

/**
 * Creates a function that returns `value`.
 *
 * @static
 * @memberOf _
 * @since 2.4.0
 * @category Util
 * @param {*} value The value to return from the new function.
 * @returns {Function} Returns the new constant function.
 * @example
 *
 * var objects = _.times(2, _.constant({ 'a': 1 }));
 *
 * console.log(objects);
 * // => [{ 'a': 1 }, { 'a': 1 }]
 *
 * console.log(objects[0] === objects[1]);
 * // => true
 */
function constant(value) {
  return function() {
    return value;
  };
}

/**
 * This method returns the first argument it receives.
 *
 * @static
 * @since 0.1.0
 * @memberOf _
 * @category Util
 * @param {*} value Any value.
 * @returns {*} Returns `value`.
 * @example
 *
 * var object = { 'a': 1 };
 *
 * console.log(_.identity(object) === object);
 * // => true
 */
function identity(value) {
  return value;
}

/**
 * This method returns `false`.
 *
 * @static
 * @memberOf _
 * @since 4.13.0
 * @category Util
 * @returns {boolean} Returns `false`.
 * @example
 *
 * _.times(2, _.stubFalse);
 * // => [false, false]
 */
function stubFalse() {
  return false;
}

module.exports = template;

/* WEBPACK VAR INJECTION */}.call(exports, __webpack_require__(2), __webpack_require__(80)(module)))

/***/ }),
/* 52 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


var $ = __webpack_require__(0),
	onIntersect, observer, animateContent;

onIntersect = function(intersectionObserverEntries) {
	intersectionObserverEntries.forEach(function(entry) {
		if (!entry.isIntersecting) {
			return;
		}

		var $trigger = $(entry.target);

		observer.unobserve($trigger[0]);

		var selectors = $trigger.data('scrollAnimationSelectors'),
		$parent = $trigger.data('scrollAnimationParent');

		animateContent($parent, selectors);

	});
};

observer = new window.IntersectionObserver(onIntersect);

animateContent = function($element, selectors) {
	window.requestAnimationFrame(function() {
		var startTime = 0;

		selectors.forEach(function(selector) {
			var $elementsToAnimate = $element.find(selector);

			$elementsToAnimate.each(function(index, element) {
				setTimeout(function() {
					window.requestAnimationFrame(function() {
						$(element).addClass('is_animating_scroll');
					});
				}, startTime);

				startTime = startTime + 125;
			});
		});
	});
};

module.exports = function(options) {
	var $trigger = options.$trigger,
	$parent = options.$parent,
	selectors = options.selectors,
	delta = options.delta;

	if (typeof selectors === 'string') {
		selectors = [selectors];
	}

	if (!$trigger) {
		// If no trigger is passed, we use the first element in the dom that
		// will be animatated.
		$trigger = selectors.join(', ');
	}

	if (typeof $trigger === 'string') {
		$trigger = $parent.find($trigger).first();

		if (!$trigger.length) {
			$trigger = $parent;
		}
	}

	$trigger.data('scrollAnimationSelectors', selectors);
	$trigger.data('scrollAnimationParent', $parent);
	$trigger.data('scrollAnimationDelta', delta);
	observer.observe($trigger[0]);
};


/***/ }),
/* 53 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(54);


/***/ }),
/* 54 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 55 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * Helper class for slider
 */



var $ = __webpack_require__(0),
	mediaqueries = __webpack_require__(6),
	keymaster = __webpack_require__(14),
	cssTransitionEnd = __webpack_require__(4),
	globalEvents = __webpack_require__(5);

var $document = $(document),
	events = {
		showCurrentSlide: 'showCurrentSlide.estatico.' + name
	},
	options = {
		domSelectors: {
			slider: {}// this is filled dynamically (see below)
		},
		stateClasses: {
			isHidden: 'is_hidden',
			isActive: 'is_active',
			isVisible: 'is_visible',
			isHover: 'is_hover',
			isSet: 'is_set'
		},
		sliderSettings: {
			itemsQuantity: {
				tiny: 1,
				mobilePortrait: 1,
				mobileLandscape: 1,
				tabletPortrait: 2,
				tabletLandscape: 2,
				desktop: 4,
				desktopLarge: 4,
				desktopLargePlus: 4,
				desktopMaxWidth: 4,
				desktop4k: 4,
				desktop5k: 4
			},
			gutter: {
				tiny: 8,
				mobilePortrait: 8,
				mobileLandscape: 8,
				tabletPortrait: 8,
				tabletLandscape: 8,
				desktop: 8,
				desktopLarge: 8,
				desktopLargePlus: 8,
				desktopMaxWidth: 8,
				desktop4k: 12,
				desktop5k: 16
			},
			transitionTime: 0.3,
			transitionEase: 'Power3.easeOut'
		}
	},
	isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints;

function SliderModule() {}

SliderModule.prototype.init = function() {

	// General initialization code
	this.options = $.extend(true, {}, options, this.options);

	this.options.domSelectors.slider = {
		container: '[data-' + this.name + '="container"]',
		scrollableContainer: '[data-' + this.name + '="scrollable_container"]',
		list: '[data-' + this.name + '="list"]',
		item: '[data-' + this.name + '="item"]',
		itemLink: '[data-teaserelement="link"]',
		navigation: '[data-' + this.name + '="slider_navigation"]',
		navigationItem: '[data-' + this.name + '="slider_navigation_item"]',
		navigationItemButton: '[data-' + this.name + '="slider_navigation_item_button"]',
		navigationItemButtonText: '[data-' + this.name + '="slider_navigation_item_button_text"]'
	};

	this.$sliderContainer = this.$element.find(this.options.domSelectors.slider.container);
	this.$sliderList = this.$element.find(this.options.domSelectors.slider.list);
	this.$sliderItem = this.$element.find(this.options.domSelectors.slider.item);
	this.$sliderItemLink = this.$element.find(this.options.domSelectors.slider.itemLink);
	this.$sliderNavigation = this.$element.find(this.options.domSelectors.slider.navigation);
	this.$sliderNavigationItem = this.$element.find(this.options.domSelectors.slider.navigationItem);
	this.$sliderNavigationItemButtons = this.$element.find(this.options.domSelectors.slider.navigationItemButton);
	this.currentBreakpointName = (mediaqueries.getCurrentBreakpointName()) ? mediaqueries.getCurrentBreakpointName() : Object.keys(this.options.sliderSettings.itemsQuantity)[0];

	this.isInitialized = false;

	this._getSliderData();
	this._setSlider();
	this._moveToCurrentSlide();

	// Try to initialize touch events
	this._initKeyboardEvents();
	this._initDesktopEvents();

	if (isTouchDevice) {
		this._initTouchEvents();
	}

	// Bind global events for breakpoint change and window resize
	$document.on(mediaqueries.event.key, this._handleBreakpointChange.bind(this));
	$document.on(globalEvents.resize.key + '.' + this.uuid, this._handleResizeChange.bind(this));

};

SliderModule.prototype._initTouchEvents = function() {
	__webpack_require__.e/* require.ensure */(1/* duplicate */).then((function(require) {
			__webpack_require__(22);
			this.$sliderList.hammer()
				.on('swipeleft.' + this.uuid, this._changeSlideToNext.bind(this))
				.on('swiperight.' + this.uuid, this._changeSlideToPrev.bind(this));
		}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);
};

SliderModule.prototype._initDesktopEvents = function() {
	this.$element.on('click.' + this.uuid, this.options.domSelectors.slider.navigationItem, this._changeSlide.bind(this));
	this.$element.on(events.showCurrentSlide + '.' + this.uuid, this._moveToCurrentSlide.bind(this));
	this.$element.on('mouseenter.' + this.uuid, this.options.domSelectors.slider.item, this._slideMouseEnter.bind(this));
	this.$element.on('mouseleave.' + this.uuid, this.options.domSelectors.slider.item, this._slideMouseLeave.bind(this));
};

SliderModule.prototype._initKeyboardEvents = function() {

	// set the scope of the module to
	var moduleScope = this.name + '-' + this.uuid;

	this.$sliderList.on('focusin.' + this.uuid, function() {
		keymaster.setScope(moduleScope);
	}.bind(this)).on('focusout.' + this.uuid, function() {
		keymaster.setScope('all');
	});

	// handle events when appropriate
	keymaster('left, pageup', moduleScope, this._changeSlideToPrev.bind(this));
	keymaster('right, pagedown', moduleScope, this._changeSlideToNext.bind(this));
	keymaster('home', moduleScope, this._changeSlideToFirst.bind(this));
	keymaster('end', moduleScope, this._changeSlideToLast.bind(this));

	keymaster('shift+tab, esc', moduleScope, function(e) {
		if (this.$element[0].contains(document.activeElement)) {
			e.preventDefault();
			this._setFocusBackToActivePaging();
		}
	}.bind(this));

};

/**
 * Function handling breakpoint changes.
 */
SliderModule.prototype._handleBreakpointChange = function(event, prevBreakpoint, currentBreakpoint) {
	this.currentBreakpointName = (mediaqueries.getCurrentBreakpointName()) ? mediaqueries.getCurrentBreakpointName() : Object.keys(this.options.sliderSettings.itemsQuantity)[0];

	// Check if we need rebuild slider
	if (this.options.sliderSettings.itemsQuantity[currentBreakpoint.name] !== this.options.sliderSettings.itemsQuantity[prevBreakpoint.name]) {
		this.sliderData.currentSlide = 0;
		this._updateSliderData();
		this._setSlider();
		window.requestAnimationFrame(function() {
			this.$element.trigger(events.showCurrentSlide);
		}.bind(this));
	}
};

/**
 * Function handling resize changes.
 */
SliderModule.prototype._handleResizeChange = function() {
	this._updateSliderData();
	this._setSliderItemWidth();

	window.requestAnimationFrame(function() {
		this.$element.trigger(events.showCurrentSlide);
	}.bind(this));

};

/**
 * Get slider data and assign it to this.sliderData.
 */
SliderModule.prototype._getSliderData = function() {
	var itemsQuantity = this.$sliderItem.length,
		itemWidth = (this.$sliderContainer.outerWidth() + 2 * this.options.sliderSettings.gutter[this.currentBreakpointName]) / this.options.sliderSettings.itemsQuantity[this.currentBreakpointName],
		slidesQuantity = Math.ceil(this.$sliderItem.length / this.options.sliderSettings.itemsQuantity[this.currentBreakpointName]),
		itemsPerSlide = Math.ceil(itemsQuantity / slidesQuantity);

	this.sliderData = {
		'itemsQuantity': itemsQuantity,
		'currentSlide': 0,
		'slidesQuantity': slidesQuantity,
		'sliderWidth': itemWidth * itemsQuantity,
		'itemWidth': itemWidth,
		'slideWidth': itemWidth * this.options.sliderSettings.itemsQuantity[this.currentBreakpointName],
		'itemsPerSlide': itemsPerSlide
	};

};

/**
 * Update slider data.
 * Function triggered on resize event.
 * itemWidth is adjusted to current viewport size based on slider settings.
 */
SliderModule.prototype._updateSliderData = function() {
	var itemWidth = (this.$sliderContainer.outerWidth() + 2 * this.options.sliderSettings.gutter[this.currentBreakpointName]) / this.options.sliderSettings.itemsQuantity[this.currentBreakpointName];

	this.sliderData.slidesQuantity = Math.ceil(this.$sliderItem.length / this.options.sliderSettings.itemsQuantity[this.currentBreakpointName]);
	this.sliderData.sliderWidth = itemWidth * this.sliderData.itemsQuantity;
	this.sliderData.itemWidth = itemWidth;
	this.sliderData.slideWidth = itemWidth * this.options.sliderSettings.itemsQuantity[this.currentBreakpointName];
	this.sliderData.itemsPerSlide = Math.ceil(this.sliderData.itemsQuantity / this.sliderData.slidesQuantity);
};

/**
 * Set all slider elements.
 * Function triggered on init and on resize.
 */
SliderModule.prototype._setSlider = function() {

	// set some setting on init, on breakpoint change
	window.requestAnimationFrame(function() {
		this._setSliderItemWidth();
		this._setSliderNavigation();
	}.bind(this));

	this.$sliderList.one(cssTransitionEnd + '.' + this.uuid, function() {
		requestAnimationFrame(function() {
			this.$sliderList.addClass(this.options.stateClasses.isSet);
		}.bind(this));
	}.bind(this));

};

/**
 * Function triggered on mouse enter on slider item element
 */
SliderModule.prototype._slideMouseEnter = function(event) {
	var $el = $(event.currentTarget);

	$el.addClass(this.options.stateClasses.isHover);
};

/**
 * Function triggered on mouse leave on slider item element
 */
SliderModule.prototype._slideMouseLeave = function(event) {
	var $el = $(event.currentTarget);

	$el.removeClass(this.options.stateClasses.isHover);
};

/**
 * Set slider item width.
 */
SliderModule.prototype._setSliderItemWidth = function() {
	this.$sliderList.css('width', this.sliderData.sliderWidth);
	this.$sliderItem.css('width', this.sliderData.itemWidth);
};

/**
 * Add slider dots navigation if there is more then one slide.
 */
SliderModule.prototype._setSliderNavigation = function() {

	// if we have only one slide, hide navigation bullets and return
	if (this.sliderData.slidesQuantity <= 1) {
		this.$sliderNavigation.addClass(this.options.stateClasses.isHidden);
		return;
	} else {
		this.$sliderNavigation.removeClass(this.options.stateClasses.isHidden);
	}

	this.$sliderNavigationItem.hide();

	// hide/show dots based on slidesQuantity
	for (var i = 0; i < this.sliderData.slidesQuantity; i++) {
		$(this.$sliderNavigationItem[i]).show();
	}

	this._modifySliderNavigationItemsText();

};

SliderModule.prototype._modifySliderNavigationItemsText = function() {

	var $listItem,
		$sliderNavigationItemButton,
		showFrom,
		showTo,
		newText;

	if (this.sliderData.itemsPerSlide === 2 && this.$sliderItem.length > 2) {

		for (var j = 1; j <= this.sliderData.slidesQuantity; j++) {

			$sliderNavigationItemButton = $(this.$sliderNavigationItemButtons[j - 1]);
			showTo = j * this.sliderData.itemsPerSlide;
			showFrom = showTo - 1;
			newText = this.options.i18n.textToSlide + ' ' + showFrom + '-' + showTo + ' ' + this.options.i18n.textOf + ' ' + this.$sliderItem.length;

			$sliderNavigationItemButton.attr({
				'aria-label': newText.trim()
			})
			.find(this.options.domSelectors.slider.navigationItemButtonText)
			.text(newText);

		}

	} else {

		var items = this.sliderData.slidesQuantity;

		while (items--) {

			$listItem = $(this.$sliderNavigationItem[items]);
			$sliderNavigationItemButton = $(this.$sliderNavigationItemButtons[items]);
			showFrom = items + 1;

			$listItem.removeClass(this.options.stateClasses.isActive);

			$sliderNavigationItemButton
				.removeAttr('aria-label data-original-text data-original-index')
				.find(this.options.domSelectors.slider.navigationItemButtonText)
				.text((this.options.i18n.textToSlide + ' ' + showFrom + ' ' + this.options.i18n.textOf + ' ' + this.$sliderItem.length).trim());

		}

	}

};

/**
 * Set current navigation item proper classes and aria attributes.
 */
SliderModule.prototype._setCurrentNavigationItem = function() {

	var $navigationItems = this.$sliderNavigation.find(this.options.domSelectors.slider.navigationItem),
		$currentItem = $navigationItems.eq(this.sliderData.currentSlide),
		$currentItemButton = $currentItem.find(this.options.domSelectors.slider.navigationItemButton),
		$prevItemButton = $navigationItems.eq(this.sliderData.prevSlide).find(this.options.domSelectors.slider.navigationItemButton),
		currentItemText = $currentItemButton.text().trim() + ' ' + this.options.i18n.textCurrentlyActive;

	$navigationItems.removeClass(this.options.stateClasses.isActive);

	if ($prevItemButton.length > 0 && this.sliderData.itemsPerSlide === 2 && this.$sliderItem.length > 2) {
		$prevItemButton.attr({
			'aria-label': $prevItemButton.text().trim()
		});
	} else {
		$prevItemButton.removeAttr('aria-label');
	}

	$currentItem.addClass(this.options.stateClasses.isActive);

	$currentItemButton.attr({
		'aria-label': currentItemText
	});

};

SliderModule.prototype._setFocusBackToActivePaging = function() {
	this.$sliderNavigationItemButtons[this.sliderData.currentSlide].focus();
};

/**
 * Show all slider items.
 * Function used when user scrolls further than to the next item.
 */
SliderModule.prototype._showAllSliderItems = function() {
	window.requestAnimationFrame(function() {
		this.$sliderItem
			.addClass(this.options.stateClasses.isVisible);
	}.bind(this));
};

/**
 * Set proper classes and aria attributes for current slider item.
 */
SliderModule.prototype._setCurrentSliderItemAttribute = function() {
	this._hideAllSlidesFromScreenReader();
	this._showCurrentSlideForScreenReader();
};

SliderModule.prototype._hideAllSlidesFromScreenReader = function() {

	this.$sliderItem
		.removeClass(this.options.stateClasses.isVisible)
		.attr({
			'aria-hidden': true,
			'role': 'presentation'
		})
		.removeAttr('tabindex')
		.find('a, button, iframe')
		.attr({
			'tabindex': '-1'
		});

};

SliderModule.prototype._showCurrentSlideForScreenReader = function() {

	var $slideToShow = this.$sliderItem.slice(this.options.sliderSettings.itemsQuantity[this.currentBreakpointName] * this.sliderData.currentSlide, this.options.sliderSettings.itemsQuantity[this.currentBreakpointName] * (this.sliderData.currentSlide + 1));

	window.requestAnimationFrame(function() {

		$slideToShow
			.addClass(this.options.stateClasses.isVisible)
			.removeAttr('role aria-hidden')
			.find('a, button, iframe')
			.removeAttr('tabindex');

		if (this.sliderData.slidesQuantity === 2) {
			$slideToShow = $($slideToShow[0]);
		}

		if (this.$sliderItemLink.length === 0) {
			$slideToShow.attr({
				'tabindex': '0'
			});
		} else {
			$slideToShow = $slideToShow.find(this.options.domSelectors.slider.itemLink);
		}

		if (this.$element[0].contains(document.activeElement)) {
			$slideToShow.focus();
		}

		this.isInitialized = true;

	}.bind(this));

};

/**
 * Update current slide index data.
 */
SliderModule.prototype._changeSlide = function(event) {
	event.preventDefault();

	var $el = $(event.currentTarget),
		$items = this.$element.find(this.options.domSelectors.slider.navigationItem);

	this.sliderData.prevSlide = this.sliderData.currentSlide;
	this.sliderData.currentSlide = $items.index($el);
	this._triggerSlideChange();
};

/**
 * Update current slide index data to next one.
 */
SliderModule.prototype._changeSlideToNext = function(event) {
	event.preventDefault();
	this.sliderData.prevSlide = this.sliderData.currentSlide;
	this.sliderData.currentSlide = (this.sliderData.currentSlide + 1 > this.sliderData.slidesQuantity - 1) ? 0 : this.sliderData.currentSlide + 1;
	this._triggerSlideChange();
};

/**
 * Adjusts UI and triggers event to display proper slide
 */
SliderModule.prototype._triggerSlideChange = function() {
	this.$element.trigger(events.showCurrentSlide);
};

/**
 * Update current slide index data to previous one.
 */
SliderModule.prototype._changeSlideToPrev = function(event) {
	event.preventDefault();
	this.sliderData.prevSlide = this.sliderData.currentSlide;
	this.sliderData.currentSlide = (this.sliderData.currentSlide - 1 >= 0) ? this.sliderData.currentSlide - 1 : this.sliderData.slidesQuantity - 1;
	this._triggerSlideChange();
};

SliderModule.prototype._changeSlideToFirst = function(event) {
	event.preventDefault();
	this.sliderData.prevSlide = this.sliderData.currentSlide;
	this.sliderData.currentSlide = 0;
	this._triggerSlideChange();
};

SliderModule.prototype._changeSlideToLast = function(event) {
	event.preventDefault();
	this.sliderData.prevSlide = this.sliderData.currentSlide;
	this.sliderData.currentSlide = this.sliderData.slidesQuantity - 1;
	this._triggerSlideChange();
};

/**
 * Do animation to current slide.
 * Before animation show all slide, onComplete show only current one, hide rest of the slides.
 */
SliderModule.prototype._moveToCurrentSlide = function() {
	var leftPosition = -1 * this.sliderData.currentSlide * this.sliderData.slideWidth;

	window.requestAnimationFrame(function() {
		this._setCurrentNavigationItem();
		this._showAllSliderItems();

		$(this.$sliderList).css({
			transform: 'translate3d(' + leftPosition + 'px,0,0)'
		});

		this.$sliderList.one(cssTransitionEnd + '.' + this.uuid, function() {
			this._setCurrentSliderItemAttribute();
		}.bind(this));

	}.bind(this));
};

module.exports = SliderModule;


/***/ }),
/* 56 */
/***/ (function(module, exports, __webpack_require__) {

var __WEBPACK_AMD_DEFINE_FACTORY__, __WEBPACK_AMD_DEFINE_ARRAY__, __WEBPACK_AMD_DEFINE_RESULT__;/**
 * lscache library
 * Copyright (c) 2011, Pamela Fox
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* jshint undef:true, browser:true, node:true */
/* global define */

(function (root, factory) {
    if (true) {
        // AMD. Register as an anonymous module.
        !(__WEBPACK_AMD_DEFINE_ARRAY__ = [], __WEBPACK_AMD_DEFINE_FACTORY__ = (factory),
				__WEBPACK_AMD_DEFINE_RESULT__ = (typeof __WEBPACK_AMD_DEFINE_FACTORY__ === 'function' ?
				(__WEBPACK_AMD_DEFINE_FACTORY__.apply(exports, __WEBPACK_AMD_DEFINE_ARRAY__)) : __WEBPACK_AMD_DEFINE_FACTORY__),
				__WEBPACK_AMD_DEFINE_RESULT__ !== undefined && (module.exports = __WEBPACK_AMD_DEFINE_RESULT__));
    } else if (typeof module !== "undefined" && module.exports) {
        // CommonJS/Node module
        module.exports = factory();
    } else {
        // Browser globals
        root.lscache = factory();
    }
}(this, function () {

  // Prefix for all lscache keys
  var CACHE_PREFIX = 'lscache-';

  // Suffix for the key name on the expiration items in localStorage
  var CACHE_SUFFIX = '-cacheexpiration';

  // expiration date radix (set to Base-36 for most space savings)
  var EXPIRY_RADIX = 10;

  // time resolution in milliseconds
  var expiryMilliseconds = 60 * 1000;
  // ECMAScript max Date (epoch + 1e8 days)
  var maxDate = calculateMaxDate(expiryMilliseconds);

  var cachedStorage;
  var cachedJSON;
  var cacheBucket = '';
  var warnings = false;

  // Determines if localStorage is supported in the browser;
  // result is cached for better performance instead of being run each time.
  // Feature detection is based on how Modernizr does it;
  // it's not straightforward due to FF4 issues.
  // It's not run at parse-time as it takes 200ms in Android.
  function supportsStorage() {
    var key = '__lscachetest__';
    var value = key;

    if (cachedStorage !== undefined) {
      return cachedStorage;
    }

    // some browsers will throw an error if you try to access local storage (e.g. brave browser)
    // hence check is inside a try/catch
    try {
      if (!localStorage) {
        return false;
      }
    } catch (ex) {
      return false;
    }

    try {
      setItem(key, value);
      removeItem(key);
      cachedStorage = true;
    } catch (e) {
        // If we hit the limit, and we don't have an empty localStorage then it means we have support
        if (isOutOfSpace(e) && localStorage.length) {
            cachedStorage = true; // just maxed it out and even the set test failed.
        } else {
            cachedStorage = false;
        }
    }
    return cachedStorage;
  }

  // Check to set if the error is us dealing with being out of space
  function isOutOfSpace(e) {
    return e && (
      e.name === 'QUOTA_EXCEEDED_ERR' ||
      e.name === 'NS_ERROR_DOM_QUOTA_REACHED' ||
      e.name === 'QuotaExceededError'
    );
  }

  // Determines if native JSON (de-)serialization is supported in the browser.
  function supportsJSON() {
    /*jshint eqnull:true */
    if (cachedJSON === undefined) {
      cachedJSON = (window.JSON != null);
    }
    return cachedJSON;
  }

  /**
   * Returns a string where all RegExp special characters are escaped with a \.
   * @param {String} text
   * @return {string}
   */
  function escapeRegExpSpecialCharacters(text) {
    return text.replace(/[[\]{}()*+?.\\^$|]/g, '\\$&');
  }

  /**
   * Returns the full string for the localStorage expiration item.
   * @param {String} key
   * @return {string}
   */
  function expirationKey(key) {
    return key + CACHE_SUFFIX;
  }

  /**
   * Returns the number of minutes since the epoch.
   * @return {number}
   */
  function currentTime() {
    return Math.floor((new Date().getTime())/expiryMilliseconds);
  }

  /**
   * Wrapper functions for localStorage methods
   */

  function getItem(key) {
    return localStorage.getItem(CACHE_PREFIX + cacheBucket + key);
  }

  function setItem(key, value) {
    // Fix for iPad issue - sometimes throws QUOTA_EXCEEDED_ERR on setItem.
    localStorage.removeItem(CACHE_PREFIX + cacheBucket + key);
    localStorage.setItem(CACHE_PREFIX + cacheBucket + key, value);
  }

  function removeItem(key) {
    localStorage.removeItem(CACHE_PREFIX + cacheBucket + key);
  }

  function eachKey(fn) {
    var prefixRegExp = new RegExp('^' + CACHE_PREFIX + escapeRegExpSpecialCharacters(cacheBucket) + '(.*)');
    // Loop in reverse as removing items will change indices of tail
    for (var i = localStorage.length-1; i >= 0 ; --i) {
      var key = localStorage.key(i);
      key = key && key.match(prefixRegExp);
      key = key && key[1];
      if (key && key.indexOf(CACHE_SUFFIX) < 0) {
        fn(key, expirationKey(key));
      }
    }
  }

  function flushItem(key) {
    var exprKey = expirationKey(key);

    removeItem(key);
    removeItem(exprKey);
  }

  function flushExpiredItem(key) {
    var exprKey = expirationKey(key);
    var expr = getItem(exprKey);

    if (expr) {
      var expirationTime = parseInt(expr, EXPIRY_RADIX);

      // Check if we should actually kick item out of storage
      if (currentTime() >= expirationTime) {
        removeItem(key);
        removeItem(exprKey);
        return true;
      }
    }
  }

  function warn(message, err) {
    if (!warnings) return;
    if (!('console' in window) || typeof window.console.warn !== 'function') return;
    window.console.warn("lscache - " + message);
    if (err) window.console.warn("lscache - The error was: " + err.message);
  }

  function calculateMaxDate(expiryMilliseconds) {
    return Math.floor(8.64e15/expiryMilliseconds);
  }

  var lscache = {
    /**
     * Stores the value in localStorage. Expires after specified number of minutes.
     * @param {string} key
     * @param {Object|string} value
     * @param {number} time
     * @return true if the value was inserted successfully
     */
    set: function(key, value, time) {
      if (!supportsStorage()) return false;

      // If we don't get a string value, try to stringify
      // In future, localStorage may properly support storing non-strings
      // and this can be removed.

      if (!supportsJSON()) return false;
      try {
        value = JSON.stringify(value);
      } catch (e) {
        // Sometimes we can't stringify due to circular refs
        // in complex objects, so we won't bother storing then.
        return false;
      }

      try {
        setItem(key, value);
      } catch (e) {
        if (isOutOfSpace(e)) {
          // If we exceeded the quota, then we will sort
          // by the expire time, and then remove the N oldest
          var storedKeys = [];
          var storedKey;
          eachKey(function(key, exprKey) {
            var expiration = getItem(exprKey);
            if (expiration) {
              expiration = parseInt(expiration, EXPIRY_RADIX);
            } else {
              // TODO: Store date added for non-expiring items for smarter removal
              expiration = maxDate;
            }
            storedKeys.push({
              key: key,
              size: (getItem(key) || '').length,
              expiration: expiration
            });
          });
          // Sorts the keys with oldest expiration time last
          storedKeys.sort(function(a, b) { return (b.expiration-a.expiration); });

          var targetSize = (value||'').length;
          while (storedKeys.length && targetSize > 0) {
            storedKey = storedKeys.pop();
            warn("Cache is full, removing item with key '" + key + "'");
            flushItem(storedKey.key);
            targetSize -= storedKey.size;
          }
          try {
            setItem(key, value);
          } catch (e) {
            // value may be larger than total quota
            warn("Could not add item with key '" + key + "', perhaps it's too big?", e);
            return false;
          }
        } else {
          // If it was some other error, just give up.
          warn("Could not add item with key '" + key + "'", e);
          return false;
        }
      }

      // If a time is specified, store expiration info in localStorage
      if (time) {
        setItem(expirationKey(key), (currentTime() + time).toString(EXPIRY_RADIX));
      } else {
        // In case they previously set a time, remove that info from localStorage.
        removeItem(expirationKey(key));
      }
      return true;
    },

    /**
     * Retrieves specified value from localStorage, if not expired.
     * @param {string} key
     * @return {string|Object}
     */
    get: function(key) {
      if (!supportsStorage()) return null;

      // Return the de-serialized item if not expired
      if (flushExpiredItem(key)) { return null; }

      // Tries to de-serialize stored value if its an object, and returns the normal value otherwise.
      var value = getItem(key);
      if (!value || !supportsJSON()) {
        return value;
      }

      try {
        // We can't tell if its JSON or a string, so we try to parse
        return JSON.parse(value);
      } catch (e) {
        // If we can't parse, it's probably because it isn't an object
        return value;
      }
    },

    /**
     * Removes a value from localStorage.
     * Equivalent to 'delete' in memcache, but that's a keyword in JS.
     * @param {string} key
     */
    remove: function(key) {
      if (!supportsStorage()) return;

      flushItem(key);
    },

    /**
     * Returns whether local storage is supported.
     * Currently exposed for testing purposes.
     * @return {boolean}
     */
    supported: function() {
      return supportsStorage();
    },

    /**
     * Flushes all lscache items and expiry markers without affecting rest of localStorage
     */
    flush: function() {
      if (!supportsStorage()) return;

      eachKey(function(key) {
        flushItem(key);
      });
    },

    /**
     * Flushes expired lscache items and expiry markers without affecting rest of localStorage
     */
    flushExpired: function() {
      if (!supportsStorage()) return;

      eachKey(function(key) {
        flushExpiredItem(key);
      });
    },

    /**
     * Appends CACHE_PREFIX so lscache will partition data in to different buckets.
     * @param {string} bucket
     */
    setBucket: function(bucket) {
      cacheBucket = bucket;
    },

    /**
     * Resets the string being appended to CACHE_PREFIX so lscache will use the default storage behavior.
     */
    resetBucket: function() {
      cacheBucket = '';
    },

    /**
     * @returns {number} The currently set number of milliseconds each time unit represents in
     *   the set() function's "time" argument.
     */
    getExpiryMilliseconds: function() {
      return expiryMilliseconds;
    },

    /**
     * Sets the number of milliseconds each time unit represents in the set() function's
     *   "time" argument.
     * Sample values:
     *  1: each time unit = 1 millisecond
     *  1000: each time unit = 1 second
     *  60000: each time unit = 1 minute (Default value)
     *  360000: each time unit = 1 hour
     * @param {number} milliseconds
     */
    setExpiryMilliseconds: function(milliseconds) {
        expiryMilliseconds = milliseconds;
        maxDate = calculateMaxDate(expiryMilliseconds);
    },

    /**
     * Sets whether to display warnings when an item is removed from the cache or not.
     */
    enableWarnings: function(enabled) {
      warnings = enabled;
    }
  };

  // Return the module
  return lscache;
}));


/***/ }),
/* 57 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(58);


/***/ }),
/* 58 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 59 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(60);


/***/ }),
/* 60 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/* global YT */

/**
 * M17 - Media
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(61);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	ConsentHelper = __webpack_require__(26),
	cssTransitionEnd = __webpack_require__(4),
	preconnect = __webpack_require__(50),
	$document = $(document);

var name = 'media',
	log = estatico.helpers.log(name),
	events = {
		closeVideo: 'close.estatico.' + name,
		ytApiLoaded: 'ytapiloaded.estatico.' + name
	},
	defaults = {
		domSelectors: {
			videoPlayer: '[data-' + name + '="video_player"]',
			videoWrapper: '[data-' + name + '="video_wrapper"]',
			videoNotification: '[data-' + name + '="video_notification"]',
			stillImageWrapper: '> [data-' + name + '="image_wrapper"]',
			stillImage: '[data-' + name + '="image"]',
			buttonPlay: '[data-' + name + '="button_play"]'
		},
		stateClasses: {
			videoRatioModifier: 'var_16x9',
			isShowingVideo: 'is_showing_video',
			isHidden: 'is_hidden',
			isVisible: 'is_visible'
		},
		interfaceColor: 'EB0000',
		delay: 200
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 /**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	var oneTrustIsIntegrated = ConsentHelper.consentIsIntegrated();

	this.$videoWrapper = this.$element.find(this.options.domSelectors.videoWrapper);
	this.$videoPlayer = this.$element.find(this.options.domSelectors.videoPlayer);
	this.$videoNotification = this.$element.find(this.options.domSelectors.videoNotification);
	this.$stillImageWrapper = this.$element.find(this.options.domSelectors.stillImageWrapper);
	this.$buttonPlay = this.$element.find(this.options.domSelectors.buttonPlay);

	if (oneTrustIsIntegrated) {

		ConsentHelper.activeGroups().then(function(groups) {

			// Allowance for marketing cookies seem to fall into the targeting category
			if (groups.allowedMarketing) {
				this.initWithConsent();
			} else {
				this.initWithOutConsent();
			}

		}.bind(this)).catch(function() {
			this.initWithOutConsent();
		}.bind(this));

	} else {
		// If OneTrust is not integrated, we assume that we can load the player without any restrictions
		this.initWithConsent();
	}

};

Module.prototype.initWithOutConsent = function() {

	this.$buttonPlay.attr({
		disabled: 'disabled',
		tabindex: '-1'
	});

	this.$videoNotification.removeClass(this.options.stateClasses.isHidden);

	requestAnimationFrame(function() {
		this.$videoNotification.addClass(this.options.stateClasses.isVisible);
	}.bind(this));

};

Module.prototype.initWithConsent = function() {
	this.$element.addClass('is_initialized');
	this._setupEventHandlers();
	this._createApiPreConnect();
};

Module.prototype._setupEventHandlers = function() {
	this.$element.on('click.' + this.uuid, this.options.domSelectors.buttonPlay, this._getVideo.bind(this));
	this.$element.on(events.closeVideo + '.' + this.uuid, this._closeVideo.bind(this));
};

/**
 * Create Pre-Connect to API
 */
Module.prototype._createApiPreConnect = function() {
	var tag,
		firstScriptTag,
		url = this.options.api.host + this.options.api.path;

	preconnect(this.options.api.host);

	if (this.options.api.name === 'youtube') {

		if ($('script[src="' + url + '"]').length === 0) {
			// Create promise
			this.YTApiPromise = this._getYTPromise();
			this.YTApiPromise.done(function() {
				this._loadYouTubeVideo();
				$document.trigger(events.ytApiLoaded);
			}.bind(this));

			tag = document.createElement('script');
			tag.src = url;
			firstScriptTag = document.getElementsByTagName('script')[0];
			firstScriptTag.parentNode.insertBefore(tag, firstScriptTag);
		} else {
			// We already have the youtube api loaded, proceed with loading video
			$document.one(events.ytApiLoaded + '.' + this.uuid, this._loadYouTubeVideo.bind(this));
		}
	}

};

/**
 * Get YouTube Promise
 * @returns {*}
 * @private
 */
Module.prototype._getYTPromise = function() {
	var deferred = $.Deferred(),
		promise = deferred.promise();

	window.onYouTubeIframeAPIReady = function() {
		deferred.resolve();
	};

	return promise;
};

/**
 * Get Video
 */
Module.prototype._getVideo = function() {
	if (this.options.api.name === 'vimeo') {
		$.ajax({
			url: encodeURI('https://vimeo.com/api/oembed.json?url=https://vimeo.com/' + this.options.videoID + '&autoplay=true&api=true&color=' + this.options.interfaceColor),
			success: this._showVimeoVideo.bind(this),
			error: function() {
				log('error occurred while fetching video');
			}
		});
	} else if (this.options.api.name === 'youtube') {
		if (this.YTPlayerReady) {
			this._showYouTubeVideo();
		}
	}
};

/**
 * Close video
 * @private
 */
Module.prototype._closeVideo = function() {
	if (this.options.api.name === 'vimeo') {
		this.$element.addClass(this.options.stateClasses.isShowingVideo);
		this._hideVimeoVideo();
	} else if (this.options.api.name === 'youtube') {
		this.$element.addClass(this.options.stateClasses.isShowingVideo);
		this.player.stopVideo();
		this._hideYouTubeVideo();
	}
};

/**
 * Load YouTube Video
 * @private
 */
Module.prototype._loadYouTubeVideo = function() {
	this.player = new YT.Player(this.$videoPlayer[0], {
		videoId: this.options.videoID,
		host: 'https://www.youtube-nocookie.com',
		height: '100%',
		width: '100%',
		playerVars: {
			color: '#EB0000',
			rel: 0,
			showinfo: 0,
			modestbranding: 1,
			enablejsapi: 1,
			wmode: 'opaque',
			autoplay: 0,
			hl: localStorage.getItem('sbblang') || 'de'
		},
		events: {
			onReady: function() {
				this.YTPlayerReady = true;
			}.bind(this),
			onStateChange: function(playerState) {
				if (playerState.data === 0) {
					this._hideYouTubeVideo();
				}
			}.bind(this)
		}
	});
};

/**
 * Show YouTube Video Response
 * @private
 */
Module.prototype._showYouTubeVideo = function() {
	log('_showYouTubeVideo');
	this.player.playVideo();
	this._toggleModules();
};

Module.prototype._hideYouTubeVideo = function() {
	log('_hideYouTubeVideo');
	this._toggleModules();
};

/**
 * Populate Video Container
 * @param videoResponse
 * @private
 */
Module.prototype._showVimeoVideo = function(videoResponse) {
	this.$videoWrapper.append(videoResponse.html);
	this.iframeCallback = this._onVimeoPlayerMessage.bind(this);
	window.addEventListener('message', this.iframeCallback, false);
	this._toggleModules();
};

/**
 * Handler for Vimeo player iframe messages
 * @param event Vimeo API player event
 * @private
 */
Module.prototype._onVimeoPlayerMessage = function(event) {

	// Check if event origin matches vimeo player, if not, exit function
	if (!(/^https?:\/\/player.vimeo.com/).test(event.origin)) {
		return false;
	}

	// Checking for typeof has it's purpose, as sometimes Vimeo returns json objects and sometimes it's stringified.
	var data = (typeof event.data === 'string') ? JSON.parse(event.data) : event.data;

	switch (data.event) {
		case 'ready':

			// Post message to vimeo player to notify us when video is done
			var message = {
				method: 'addEventListener',
				value: 'finish'
			};

			this.$videoWrapper.find('iframe')[0].contentWindow.postMessage(JSON.stringify(message), '*');
			break;
		case 'finish':
			this._hideVimeoVideo();
			break;
	}
};

/**
 * Hide Vimeo Video
 * @private
 */
Module.prototype._hideVimeoVideo = function() {
	this._toggleModules();
	window.removeEventListener('message', this.iframeCallback, false);
	this.$videoWrapper.find('iframe').remove();
};

/**
 * Toggle Image and Video Module
 */
Module.prototype._toggleModules = function() {
	// If video is hidden
	if (!this.$element.hasClass(this.options.stateClasses.isShowingVideo)) {

		window.requestAnimationFrame(function() {
			this.$element.find(this.options.domSelectors.videoPlayer).removeClass('is_hidden');
			this.$element.addClass(this.options.stateClasses.isShowingVideo + ' ' + this.options.stateClasses.videoRatioModifier);

			this.$stillImageWrapper.one(cssTransitionEnd + '.' + this.uuid, function() {
				this.$stillImageWrapper.hide().attr('aria-hidden', 'true');
				this.$buttonPlay.hide().attr('aria-hidden', 'true');
			}.bind(this));
		}.bind(this));

	} else { // If video is visible

		window.requestAnimationFrame(function() {
			this.$element.find(this.options.domSelectors.videoPlayer).addClass('is_hidden');
			this.$stillImageWrapper.show().attr('aria-hidden', 'false');
			this.$buttonPlay.show().attr('aria-hidden', 'false');

			// After changing display from none to block on an element, there's a required one keyframe pause at least
			// before css animations can be triggered - hence the setTimeout call.
			setTimeout(function() {
				this.$element.removeClass(this.options.stateClasses.isShowingVideo + ' ' + this.options.stateClasses.videoRatioModifier);
			}.bind(this), 30);
		}.bind(this));

	}
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 61 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 62 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";

/**
 * Cookie Helper
 *
 */



module.exports = {

	setCookie: function(elementID, cookieExpirationTimeInDays) {

		var hostname = window.location.hostname,
			domain = '',
			expirationDate = new Date(),
			expirationTimeInDays = cookieExpirationTimeInDays || 30,
			expires;

		if (!elementID) {
			throw new Error('setCookie expects an ID as first parameter');
		}

		// if we are on sbb.ch, cff.ch or ffs.ch, we want to save the cookie
		// across subdomains to share things like ghettobox clicks. cff.ch maps to sbb.ch/fr/,
		// ffs.ch maps to sbb.ch/it/
		if (hostname.indexOf('sbb.ch') !== -1) {
			domain = ';domain=' + hostname.replace('www.', '');
		}

		// define expiration date
		expirationDate.setTime(expirationDate.getTime() + (expirationTimeInDays * 24 * 60 * 60 * 1000));

		// define expiration date as string
		expires = 'expires=' + expirationDate.toUTCString();

		// set cookie
		document.cookie = elementID + '= true' + '; ' + expires + domain;

	},

	hasCookie: function(elementID) {

		var name,
			cookieName,
			cookieArray,
			iterator,
			cookie;

		if (!elementID) {
			return false;
		}

		name = elementID + '=';

		cookieArray = document.cookie.split(';');

		for (iterator = 0; iterator < cookieArray.length; iterator++) {
			cookie = cookieArray[iterator];

			while (cookie.charAt(0) === ' ') {
				cookie = cookie.substring(1);
			}

			if (cookie.indexOf(name) === 0) {
				cookieName = cookie.substring(name.length, cookie.length);
				if (cookieName) {
					return true;
				}
			}
		}

		return false;
	}

};


/***/ }),
/* 63 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(64);


/***/ }),
/* 64 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 65 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


/**
 * This helper checks to see if a hash-value is present in the url and
 * returns it if there is one.
 */

function _getHash(url) {
	var hash = window.location.hash,
		urlObject;

	if (url) {
		urlObject = new URL(url);
		hash = urlObject.hash;
	}

	if (!hash) {
		return false;
	}

	// per default, location.hash returns the value including #. We remove #.
	if (hash.indexOf('#') === 0) {
		hash = hash.substring(1);
	} else {
		return false;
	}

	// browsers might add a slash behind #. We remove it.
	if (hash.indexOf('/') === 0) {
		hash = hash.substring(1);
	}

	if (hash.length > 0) {
		return hash;
	}

	return false;
}

module.exports = _getHash;


/***/ }),
/* 66 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


/**
 * This helper scrolls the html document to a specified element.
 * @param {string} elementAttribute - attribute of the element to search for
 * @param {string} hash - the hash to scroll to. if empty string is passed, the helper will search for the hash in the url
 * @param {integer} animationSpeed - time in ms over which the animation will occur
 * @param {boolean} focus - if set to true, the defined element will be focussed
 * @param {string} accessibilityHint - visually hidden text that will be added to the focussed element. if focus is set to false, this parameter is not handled
 *
 *
 * About elementAttribute:
 * Let's assume elementAttribute is 'data-tab-id' and the hash in
 * the url is 'panel-1'. In this case, this helper will scroll
 * to the element '[data-tab-id="panel-1"]'
 */

var $ = __webpack_require__(0),

	// mobile header is just slightly less tall then desktop header,
	// so we can live with these values for all viewports
	defaults = {
		rolledUpHeaderHeight: 68,
		scrollToPaddingTop: 20
	},
	_scrollToHash;

// helper method to set focus
function _setFocusOnElement($element) {

	// unfortunately, some of our modules are full of timeouts and rafs. therefore,
	// we need to wait for the tabs to be ready before we can set the focus on
	// the active tab.
	setTimeout(function() {
		$element.focus();
	}, 500);
}

// helper method to add visually hidden text
function _addVisuallyHiddenTextToElement($element, text) {

	var html;

	if (!text) {
		return;
	}

	html = '<span class="visuallyhidden">' + text + '</span>';

	$element.append(html);
}

// helper to get the hash value from the current window location
function _getHashValueFromWindowLocation() {
	var hash = window.location.hash,
		hashName;

	// make sure we have a hash
	if (!hash) {
		return false;
	}

	// get hash value
	hashName = hash.substring(1);
	if (hashName.length < 1) {
		return false;
	}

	return hashName;

}

// main method
_scrollToHash = function(elementAttribute, hash, animationSpeed, focus, accessibilityHint) {

	var hashName = (hash.length === 0 || !hash) ? _getHashValueFromWindowLocation() : hash,
		elements,
		$element,
		scrollToPosition = 0;

	// make sure we have a hash
	if (!hashName) {
		return;
	}

	// make sure we get an attribute passed as parameter
	if (!elementAttribute) {
		return;
	}

	// define animation speed
	if (!animationSpeed) {
		animationSpeed = 0;
	}

	// let's try to find the element
	elements = $(document).find('[' + elementAttribute + '="' + hashName + '"]');

	// we can return if we did not find any element
	if (elements.length < 1) {
		return;
	}

	// define element to which we want to scroll
	$element = $(elements[0]);

	if (focus) {
		if (accessibilityHint) {
			_addVisuallyHiddenTextToElement($element, accessibilityHint);
		}

		_setFocusOnElement($element);
	}

	// set position to which we want to scroll
	scrollToPosition = parseInt($element.offset().top) - defaults.rolledUpHeaderHeight - defaults.scrollToPaddingTop;

	// scroll to that element
	$('html, body').animate({
		scrollTop: scrollToPosition
	}, animationSpeed);

};

module.exports = _scrollToHash;


/***/ }),
/* 67 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(68);


/***/ }),
/* 68 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 69 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


/**
 * This helper accepts a url as parameter. It uses a regex to parse the
 * url and returning an object from the url parameters and values
 */

module.exports = function(url) {

	var params = {};

	if (!url) {
		return params;
	}

	url.replace(/[?&]+([^=&]+)=([^&]*)/gi, function(str, key, value) {
		params[key] = value;
	});

	return params;

};


/***/ }),
/* 70 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(76);


/***/ }),
/* 71 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


var _interopRequireWildcard = function (obj) { return obj && obj.__esModule ? obj : { 'default': obj }; };

exports.__esModule = true;

var _import = __webpack_require__(31);

var base = _interopRequireWildcard(_import);

// Each of these augment the Handlebars object. No need to setup here.
// (This is done to easily share code between commonjs and browse envs)

var _SafeString = __webpack_require__(72);

var _SafeString2 = _interopRequireWildcard(_SafeString);

var _Exception = __webpack_require__(21);

var _Exception2 = _interopRequireWildcard(_Exception);

var _import2 = __webpack_require__(20);

var Utils = _interopRequireWildcard(_import2);

var _import3 = __webpack_require__(73);

var runtime = _interopRequireWildcard(_import3);

var _noConflict = __webpack_require__(74);

var _noConflict2 = _interopRequireWildcard(_noConflict);

// For compatibility and usage outside of module systems, make the Handlebars object a namespace
function create() {
  var hb = new base.HandlebarsEnvironment();

  Utils.extend(hb, base);
  hb.SafeString = _SafeString2['default'];
  hb.Exception = _Exception2['default'];
  hb.Utils = Utils;
  hb.escapeExpression = Utils.escapeExpression;

  hb.VM = runtime;
  hb.template = function (spec) {
    return runtime.template(spec, hb);
  };

  return hb;
}

var inst = create();
inst.create = create;

_noConflict2['default'](inst);

inst['default'] = inst;

exports['default'] = inst;
module.exports = exports['default'];

/***/ }),
/* 72 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


exports.__esModule = true;
// Build out our basic SafeString type
function SafeString(string) {
  this.string = string;
}

SafeString.prototype.toString = SafeString.prototype.toHTML = function () {
  return '' + this.string;
};

exports['default'] = SafeString;
module.exports = exports['default'];

/***/ }),
/* 73 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


var _interopRequireWildcard = function (obj) { return obj && obj.__esModule ? obj : { 'default': obj }; };

exports.__esModule = true;
exports.checkRevision = checkRevision;

// TODO: Remove this line and break up compilePartial

exports.template = template;
exports.wrapProgram = wrapProgram;
exports.resolvePartial = resolvePartial;
exports.invokePartial = invokePartial;
exports.noop = noop;

var _import = __webpack_require__(20);

var Utils = _interopRequireWildcard(_import);

var _Exception = __webpack_require__(21);

var _Exception2 = _interopRequireWildcard(_Exception);

var _COMPILER_REVISION$REVISION_CHANGES$createFrame = __webpack_require__(31);

function checkRevision(compilerInfo) {
  var compilerRevision = compilerInfo && compilerInfo[0] || 1,
      currentRevision = _COMPILER_REVISION$REVISION_CHANGES$createFrame.COMPILER_REVISION;

  if (compilerRevision !== currentRevision) {
    if (compilerRevision < currentRevision) {
      var runtimeVersions = _COMPILER_REVISION$REVISION_CHANGES$createFrame.REVISION_CHANGES[currentRevision],
          compilerVersions = _COMPILER_REVISION$REVISION_CHANGES$createFrame.REVISION_CHANGES[compilerRevision];
      throw new _Exception2['default']('Template was precompiled with an older version of Handlebars than the current runtime. ' + 'Please update your precompiler to a newer version (' + runtimeVersions + ') or downgrade your runtime to an older version (' + compilerVersions + ').');
    } else {
      // Use the embedded version info since the runtime doesn't know about this revision yet
      throw new _Exception2['default']('Template was precompiled with a newer version of Handlebars than the current runtime. ' + 'Please update your runtime to a newer version (' + compilerInfo[1] + ').');
    }
  }
}

function template(templateSpec, env) {
  /* istanbul ignore next */
  if (!env) {
    throw new _Exception2['default']('No environment passed to template');
  }
  if (!templateSpec || !templateSpec.main) {
    throw new _Exception2['default']('Unknown template object: ' + typeof templateSpec);
  }

  // Note: Using env.VM references rather than local var references throughout this section to allow
  // for external users to override these as psuedo-supported APIs.
  env.VM.checkRevision(templateSpec.compiler);

  function invokePartialWrapper(partial, context, options) {
    if (options.hash) {
      context = Utils.extend({}, context, options.hash);
    }

    partial = env.VM.resolvePartial.call(this, partial, context, options);
    var result = env.VM.invokePartial.call(this, partial, context, options);

    if (result == null && env.compile) {
      options.partials[options.name] = env.compile(partial, templateSpec.compilerOptions, env);
      result = options.partials[options.name](context, options);
    }
    if (result != null) {
      if (options.indent) {
        var lines = result.split('\n');
        for (var i = 0, l = lines.length; i < l; i++) {
          if (!lines[i] && i + 1 === l) {
            break;
          }

          lines[i] = options.indent + lines[i];
        }
        result = lines.join('\n');
      }
      return result;
    } else {
      throw new _Exception2['default']('The partial ' + options.name + ' could not be compiled when running in runtime-only mode');
    }
  }

  // Just add water
  var container = {
    strict: function strict(obj, name) {
      if (!(name in obj)) {
        throw new _Exception2['default']('"' + name + '" not defined in ' + obj);
      }
      return obj[name];
    },
    lookup: function lookup(depths, name) {
      var len = depths.length;
      for (var i = 0; i < len; i++) {
        if (depths[i] && depths[i][name] != null) {
          return depths[i][name];
        }
      }
    },
    lambda: function lambda(current, context) {
      return typeof current === 'function' ? current.call(context) : current;
    },

    escapeExpression: Utils.escapeExpression,
    invokePartial: invokePartialWrapper,

    fn: function fn(i) {
      return templateSpec[i];
    },

    programs: [],
    program: function program(i, data, declaredBlockParams, blockParams, depths) {
      var programWrapper = this.programs[i],
          fn = this.fn(i);
      if (data || depths || blockParams || declaredBlockParams) {
        programWrapper = wrapProgram(this, i, fn, data, declaredBlockParams, blockParams, depths);
      } else if (!programWrapper) {
        programWrapper = this.programs[i] = wrapProgram(this, i, fn);
      }
      return programWrapper;
    },

    data: function data(value, depth) {
      while (value && depth--) {
        value = value._parent;
      }
      return value;
    },
    merge: function merge(param, common) {
      var obj = param || common;

      if (param && common && param !== common) {
        obj = Utils.extend({}, common, param);
      }

      return obj;
    },

    noop: env.VM.noop,
    compilerInfo: templateSpec.compiler
  };

  function ret(context) {
    var options = arguments[1] === undefined ? {} : arguments[1];

    var data = options.data;

    ret._setup(options);
    if (!options.partial && templateSpec.useData) {
      data = initData(context, data);
    }
    var depths = undefined,
        blockParams = templateSpec.useBlockParams ? [] : undefined;
    if (templateSpec.useDepths) {
      depths = options.depths ? [context].concat(options.depths) : [context];
    }

    return templateSpec.main.call(container, context, container.helpers, container.partials, data, blockParams, depths);
  }
  ret.isTop = true;

  ret._setup = function (options) {
    if (!options.partial) {
      container.helpers = container.merge(options.helpers, env.helpers);

      if (templateSpec.usePartial) {
        container.partials = container.merge(options.partials, env.partials);
      }
    } else {
      container.helpers = options.helpers;
      container.partials = options.partials;
    }
  };

  ret._child = function (i, data, blockParams, depths) {
    if (templateSpec.useBlockParams && !blockParams) {
      throw new _Exception2['default']('must pass block params');
    }
    if (templateSpec.useDepths && !depths) {
      throw new _Exception2['default']('must pass parent depths');
    }

    return wrapProgram(container, i, templateSpec[i], data, 0, blockParams, depths);
  };
  return ret;
}

function wrapProgram(container, i, fn, data, declaredBlockParams, blockParams, depths) {
  function prog(context) {
    var options = arguments[1] === undefined ? {} : arguments[1];

    return fn.call(container, context, container.helpers, container.partials, options.data || data, blockParams && [options.blockParams].concat(blockParams), depths && [context].concat(depths));
  }
  prog.program = i;
  prog.depth = depths ? depths.length : 0;
  prog.blockParams = declaredBlockParams || 0;
  return prog;
}

function resolvePartial(partial, context, options) {
  if (!partial) {
    partial = options.partials[options.name];
  } else if (!partial.call && !options.name) {
    // This is a dynamic partial that returned a string
    options.name = partial;
    partial = options.partials[partial];
  }
  return partial;
}

function invokePartial(partial, context, options) {
  options.partial = true;

  if (partial === undefined) {
    throw new _Exception2['default']('The partial ' + options.name + ' could not be found');
  } else if (partial instanceof Function) {
    return partial(context, options);
  }
}

function noop() {
  return '';
}

function initData(context, data) {
  if (!data || !('root' in data)) {
    data = data ? _COMPILER_REVISION$REVISION_CHANGES$createFrame.createFrame(data) : {};
    data.root = context;
  }
  return data;
}

/***/ }),
/* 74 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/* WEBPACK VAR INJECTION */(function(global) {

exports.__esModule = true;
/*global window */

exports['default'] = function (Handlebars) {
  /* istanbul ignore next */
  var root = typeof global !== 'undefined' ? global : window,
      $Handlebars = root.Handlebars;
  /* istanbul ignore next */
  Handlebars.noConflict = function () {
    if (root.Handlebars === Handlebars) {
      root.Handlebars = $Handlebars;
    }
  };
};

module.exports = exports['default'];
/* WEBPACK VAR INJECTION */}.call(exports, __webpack_require__(2)))

/***/ }),
/* 75 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M15 - Accordion
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(84);

var $ = __webpack_require__(0),
	contains = __webpack_require__(15),
	SuperClass = __webpack_require__(1),
	getHash = __webpack_require__(65),
	scrollToHash = __webpack_require__(66),
	gsap = __webpack_require__(11).gsap,
	$document = $(document);

var name = 'accordion',
	log = estatico.helpers.log(name),
	events = {
		openItem: 'open.estatico.' + name,
		closeItem: 'close.estatico.' + name,
		forceOpenItem: 'forceOpenItem.estatico.' + name, // triggered from outside
		globalCloseItem: 'm15-accordion.close-item'
	},
	moduleEvents = ['open', 'close'],
	defaults = {
		animationDuration: 0.3,
		domSelectors: {
			item: '[data-' + name + '="item"]',
			link: '[data-' + name + '="link"]',
			content: '[data-' + name + '="content"]',
			heading: '[data-' + name + '="heading"]',
			moduleTitle: '[data-' + name + '="module_title"]',
			timetable: 'data-timetable',
			timetableAccordionSet: '.mod_accordion_set'
		},
		stateClasses: {
			isOpen: 'is_open',
			close: 'is_closed',
			isVisible: 'is_visible'
		},
		singleTabOpen: false, // if false multiple tabs can be opened
		closeKeys: [27], // ESC
		observeDOM: false
	},
	data = {
		button: {}
	};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		moduleEvents: moduleEvents,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	var $hashTarget;

	this.hashValue = getHash();
	this.$accordionItems = false;
	this.firstExpansionTriggeredByHashValue = false;

	// Set flag single or multiple tabs opened
	this.options.singleTabOpen = this.$element.data('singleTab');

	this.$element.on('keyup.' + this.uuid, this.options.domSelectors.item, this._handleKeyboard.bind(this));

	this.$moduleTitle = this.$element.prev(this.options.domSelectors.moduleTitle);

	// we shall not use a delegated event handler on this.$element here.
	// the reason being that JAWS will read the content of the accordion
	// as clickable since the parent container of the content would have
	// a click handler attached.
	this._attachClickHandlers();

	if (this.hashValue) {

		this.firstExpansionTriggeredByHashValue = true;

		$hashTarget = this.$element.find('[data-accordion-id="' + this.hashValue + '"]');

		// set accordion if element exist
		if ($hashTarget.length) {
			this._handleHash($hashTarget);
		}
	} else {
		this.trigger('open', $(this.options.domSelectors.content)[0]);

		// listen to openItem event from outside
		this.$element.on(events.forceOpenItem, this._handleForceOpen.bind(this));
	}

	if (this.options.observeDOM) {
		this._registerMutationObserver();
	}

};

Module.prototype._handleHash = function($hashTarget) {
	var $item = $hashTarget.closest(this.options.domSelectors.item);

	if (!$item.hasClass(this.options.stateClasses.isOpen)) {
		this._toggleSection($item);
	}
};

Module.prototype._handleForceOpen = function(evt, hash) {
	var $hashTarget = this.$element.find('[data-accordion-id="' + hash + '"]'),
		$itemToOpen;

	if ($hashTarget.length < 1) {
		return;
	}

	$itemToOpen = $hashTarget.closest(this.options.domSelectors.item);

	if (!$itemToOpen.hasClass(this.options.stateClasses.isOpen)) {
		this._toggleSection($itemToOpen);
	}
};

Module.prototype._attachClickHandlers = function() {
	this.$accordionItems = $(this.$element.find(this.options.domSelectors.link));
	this._detachClickHandlers();
	this.$accordionItems.on('click.' + this.uuid, this._handleCollapsibleLink.bind(this));

};

Module.prototype._detachClickHandlers = function() {
	if (this.$accordionItems) {
		this.$accordionItems.off('click');
	}
};

Module.prototype._registerMutationObserver = function() {
	var isTimetable = this.$element[0].hasAttribute(this.options.domSelectors.timetable),
		elementToObserve = this.$element[0],
		config = {
			childList: true
		};

	if (!('MutationObserver' in window)) {
		return;
	}

	if (isTimetable) {
		elementToObserve = this.$element.find(this.options.domSelectors.timetableAccordionSet)[0];
	}

	this.observer = new MutationObserver(function(mutations) {
		mutations.forEach(function(mutation) {
			if (mutation.addedNodes.length > 0) {
				this._attachClickHandlers();
			}
		}.bind(this));
	}.bind(this));

	this.observer.observe(elementToObserve, config);

};

/**
 * Sets a close validator.
 * @param $stateChangeValidator {function} if called, the stateChangeValidator should return a promise.
 * @private
 */
Module.prototype.setStateChangeValidator = function(stateChangeValidator) {
	if (typeof stateChangeValidator !== 'function') {
		throw new Error('accordion.setStateChangeValidator() requires a function');
	}

	this.stateChangeValidator = stateChangeValidator;
};

/**
 * Sets aria-attributes for the given collapsible element.
 * @param $collapsible {jQuery} Object containing the link and the content of the collapsible.
 * @private
 */
Module.prototype._updateAria = function($collapsible, isOpen, setFocus) {
	var $content = $collapsible.find(this.options.domSelectors.content);

	$collapsible.find(this.options.domSelectors.link).attr('aria-expanded', isOpen);
	$content.attr('aria-hidden', !isOpen);

	if (this.hashValue && this.firstExpansionTriggeredByHashValue) {
		scrollToHash('data-accordion-id', this.hashValue, 0, true, false);
	} else {
		if (setFocus) {
			$collapsible.find(this.options.domSelectors.link).focus();
		}
	}
};

/**
 * Handles the click-event on the link and handles the toggle of the section.
 * @param event {jQuery.Event} The original click-event object.
 * @private
 */
Module.prototype._handleCollapsibleLink = function(event) {

	// Preventing the default event on the <button> as the shop put
	// accordions within a <form>
	event.preventDefault();

	var $link = $(event.target),
		$item = $link.closest(this.options.domSelectors.item);

	this._toggleSection($item);
};

/**
 * Handles some keyboard-navigation-options. (closing per ESC)
 * @private
 */
Module.prototype._handleKeyboard = function(event) {
	if (contains(this.options.closeKeys, event.keyCode)) {
		var $item = $(event.currentTarget);

		if ($item.hasClass(this.options.stateClasses.isOpen)) {
			this._toggleSection($item);
		}
	}
};

Module.prototype._openSection = function($item) {
	var $content = $item.find(this.options.domSelectors.content),
		titleElement = $item.find(this.options.domSelectors.link),
		title = titleElement.length > 0 ? titleElement[0].innerText : '',
		trackingData = {
			action: 'open',
			component: $item[0],
			label: title,
			headline: ''
		};

	if (this.$moduleTitle.length > 0) {
		trackingData.headline = this.$moduleTitle[0].innerText;
	}

	window.requestAnimationFrame(function() {
		$item.removeClass(this.options.stateClasses.close); // Show before animation
		$item.addClass(this.options.stateClasses.isOpen); // Change icons

		gsap.fromTo(
			$content,
			{
				height: 0

			},
			{
				duration: this.options.animationDuration,
				height: 'auto',
				ease: 'power1.easeInOut',
				onComplete: function() {
					$document.trigger('flying-focus_trigger');
					$item
						.addClass(this.options.stateClasses.isVisible)
						.trigger(events.openItem);

					if (window.tracking) {
						window.tracking.dataLayerApi.trackAccordionAction(trackingData);
					}
				}.bind(this)
			}
		);

		// Check if only single tab can be open
		if (this.options.singleTabOpen) {
			this._closeActiveSections($item);
		}

		this._updateAria($item, true, true);
		this.firstExpansionTriggeredByHashValue = false;
		this.trigger('open', $item[0]);
	}.bind(this));
};

Module.prototype._closeSection = function($item) {
	var $content = $item.find(this.options.domSelectors.content),
		titleElement = $item.find(this.options.domSelectors.link),
		title = titleElement.length > 0 ? titleElement[0].innerText : '',
		trackingData = {
			action: 'close',
			component: $item[0],
			label: title,
			headline: ''
		};

	if (this.$moduleTitle.length > 0) {
		trackingData.headline = this.$moduleTitle[0].innerText;
	}

	window.requestAnimationFrame(function() {
		$item.removeClass(this.options.stateClasses.isVisible); // Set overflow hidden before animation

		gsap.fromTo(
			$content,
			{
				height: 'auto'
			},
			{
				duration: this.options.animationDuration,
				height: 0,
				ease: 'power1.easeInOut',
				onComplete: function() {
					$item.addClass(this.options.stateClasses.close) // Hide after animation
						.trigger('flying-focus_trigger')
						.trigger(events.closeItem);

					if (window.tracking) {
						window.tracking.dataLayerApi.trackAccordionAction(trackingData);
					}

					// sbbch-12819: when accordions are closing, trigger event to allow components (m19) to recalc position
					setTimeout(function() {
						$document.trigger(events.globalCloseItem);
					});
				}.bind(this)
			}
		);

		this._updateAria($item, false, true);
		$item.removeClass(this.options.stateClasses.isOpen); // Change icons
		this.trigger('close', $item[0]);
	}.bind(this));
};

Module.prototype._closeActiveSections = function($currentlyOpenedSection) {
	var $itemToClose = $currentlyOpenedSection.siblings('.' + this.options.stateClasses.isOpen);

	if ($itemToClose.length > 0) {
		this._closeSection($itemToClose);
	}
};

/**
 * Toggles a Section in the Footer.
 * @param $item {jQuery} The collapsible-container containing link and content.
 * @param hasBeenValidated
 * @private
 */
Module.prototype._toggleSection = function($item, hasBeenValidated) {

	if (!hasBeenValidated && typeof this.stateChangeValidator === 'function') {
		var promise = this.stateChangeValidator($item);

		if (typeof promise.then !== 'function') {
			// If the stateChangeValidator dose not return a promise, we close the accordion
			// and throw an error.
			this._toggleSection($item, true);
			throw new Error('accordion.setStateChangeValidator() expects a function that returns a promise');
		}

		promise.then(function(isValid) {
			if (!isValid) {
				// If the promise gets resolve with false, we do not close the accordion.
				return;
			}

			this._toggleSection($item, true);
		}.bind(this), function() {
			// If the stateChangeValidator fails we catch it here, close the accordion and
			// throw an error.
			this._toggleSection($item, true);
			throw new Error('accordion.setStateChangeValidator() promise faild to be resolved.');
		}.bind(this));

		return;
	} else {
		log('no stateChangeValidator present');
	}

	if (!$item.hasClass(this.options.stateClasses.isOpen)) {
		this._openSection($item);
	} else {
		this._closeSection($item);
	}
};

// Destroy module
Module.prototype.destroy = function() {
	SuperClass.prototype.destroy.apply(this);

	if (this.observer) {
		this.observer.disconnect();
	}

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	initEvents: ['ready', 'ajaxload'],
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 76 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.15 - Notification
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2019.
 */



__webpack_require__(77);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'notification',
	events = {
		loggedInUser: 'loggedInUser'
	},
	defaults = {
		domSelectors: {},
		stateClasses: {
			isHidden: 'is_hidden',
			isVisible: 'is_visible'
		},
		attributes: {
			context: 'data-' + name + '-context'
		}
	},
	data = {},
	identifiers = {
		businessCostumer: 'b2b',
		privateCostumer: 'b2c',
		loginResponseBusinessIdentifier: 'businessPartner'
	};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	// Listen to Login event
	$(document).on(events.loggedInUser + '.' + this.uuid, this.handleLogin.bind(this));
};

Module.prototype.handleLogin = function(_, data) {
	var isBusinessUser,
		context;

	if (!data) {
		return;
	}

	if (!data.userInfo) {
		return;
	}

	isBusinessUser = this.userInfoRepresentsBusinessUser(data.userInfo);
	context = this.$element.attr(this.options.attributes.context);

	if (isBusinessUser && context === identifiers.businessCostumer) {
		return;
	}

	if (!isBusinessUser && context === identifiers.privateCostumer) {
		return;
	}

	this.showNotification();
};

Module.prototype.userInfoRepresentsBusinessUser = function(userInfo) {
	var userInfoKeys = Object.keys(userInfo);

	// usually, we should get a boolean for the key 'businessPartner'. But to
	// be sure, we first check if it is there.
	if (userInfoKeys.indexOf(identifiers.loginResponseBusinessIdentifier) === -1) {
		return false;
	}

	return userInfo[identifiers.loginResponseBusinessIdentifier];
};

Module.prototype.showNotification = function() {
	this.$element.removeClass(this.options.stateClasses.isHidden);

	requestAnimationFrame(function() {
		this.$element.addClass(this.options.stateClasses.isVisible);
	}.bind(this));

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 77 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 78 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 79 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


/**
 * This helper checks to see if a hash-value is present in the url and
 * returns it if there is one.
 */

function createUrlParamsFromObject(object) {
	var params = '',
		filterValue,
		isArray;

	function addNextUrlParam(_params) {
		if (_params.length > 0) {
			_params += '&';
		} else {
			_params += '?';
		}

		return _params;
	}

	Object.keys(object).forEach(function(key) {
		filterValue = object[key];

		if (!filterValue) {
			return;
		}

		isArray = Array.isArray(filterValue);

		if (!isArray) {
			params = addNextUrlParam(params);
			params += key + '=' + encodeURIComponent(filterValue);
		} else {
			if (filterValue.length < 1) {
				return;
			}

			params = addNextUrlParam(params);
			params += key + '=' + encodeURIComponent(filterValue.join(','));
		}

	});

	return params;
}

module.exports = createUrlParamsFromObject;


/***/ }),
/* 80 */
/***/ (function(module, exports) {

module.exports = function(module) {
	if(!module.webpackPolyfill) {
		module.deprecate = function() {};
		module.paths = [];
		// module.parent = undefined by default
		if(!module.children) module.children = [];
		Object.defineProperty(module, "loaded", {
			enumerable: true,
			get: function() {
				return module.l;
			}
		});
		Object.defineProperty(module, "id", {
			enumerable: true,
			get: function() {
				return module.i;
			}
		});
		module.webpackPolyfill = 1;
	}
	return module;
};


/***/ }),
/* 81 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(82);


/***/ }),
/* 82 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 83 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(75);


/***/ }),
/* 84 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 85 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(86);


/***/ }),
/* 86 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 87 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(88);


/***/ }),
/* 88 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.4 - Radio/Checkbox
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(78);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	$document = $(document);

var name = 'radiocheckbox',
	events = {
		checkboxChanged: 'checkboxChanged.estatico.' + name
	},
	defaults = {
		domSelectors: {
			form: 'form',
			wrapper: '[data-' + name + '="wrapper"]',
			checkbox: '[data-' + name + '*="checkbox_item"]',
			checkboxReset: '[data-' + name + '*="checkbox_reset"]'
		},
		stateClasses: {
			checkAllBefore: 'is_checking_all_boxes_before',
			uncheckOnReset: 'is_unchecking_all_boxes'
		},
		attributes: {
			conditionalControlsTarget: 'data-formconditional-controls-target',
			conditionalTargetHiddenAttr: 'data-formconditional-target-hidden'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	this.$wrapper = this.$element.parent('div').find(this.options.domSelectors.wrapper);
	this.$inputs = this.$element.find(this.options.domSelectors.checkbox);
	this.possibiliytOfSeveralCollapsible = this.options.possibiliytOfSeveralCollapsible || false;

	if (!this.possibiliytOfSeveralCollapsible) {
		this._updateTargetElement();
	}

	if (this.targetAttr && this.targetAttr !== 'false' || this.possibiliytOfSeveralCollapsible) {
		// ok, these radios are controlling a target.
		// add a change event on the radios.
		// don't use delegated event here, so that we're able to dispatch
		// custom event from outside and handle it here (e.g. for testing).
		this.$element.on('change.' + this.uuid, this._radioCheckboxSelectionChanged.bind(this));

		// initially check if we need to hide a target
		this._radioCheckboxSelectionChanged();
	}

	// if wrapper has state class checkAllBefore, then attach value changed handler to input elements
	if (this.$wrapper.hasClass(this.options.stateClasses.checkAllBefore) === true) {
		this.$element.on('change.' + this.uuid, this.options.domSelectors.checkbox, this._checkboxChanged.bind(this));
	}

	if (this.$wrapper.hasClass(this.options.stateClasses.uncheckOnReset) === true) {
		this.$element.on('change.' + this.uuid, this.options.domSelectors.checkbox, this._checkboxReset.bind(this));
	}

	// load custom validation for checkbox/radio group
	if ($('[data-parsley-custom-group-validation]').length) {
		__webpack_require__(89);
	}
};

/**
 * Handle checkbox change event: if a checkbox is selected, select all before
 * @method
 * @private
 */
Module.prototype._checkboxChanged = function(event) {

	var $currentElement = $(event.currentTarget);

	// return if this element is not a checkbox
	if ($currentElement.prop('type') !== 'checkbox') {
		return;
	}

	if ($currentElement.prop('checked')) {
		this._checkboxChecked(event);
	} else {
		this._checkboxUnchecked(event);
	}
};

/**
 * Handle checkbox change event: if a checkbox is selected, select all before
 * @method
 * @private
 */
Module.prototype._checkboxReset = function(event) {

	var $currentElement = $(event.currentTarget),
		checkedCount = 0;

	// return if this element is not a checkbox
	if ($currentElement.prop('type') !== 'checkbox') {
		return;
	}

	if ($currentElement.prop('checked')) {
		if ($currentElement.attr('data-radiocheckbox').indexOf('checkbox_reset') >= 0) {
			// reset checkbox is checked, uncheck all other checkboxes.
			this.$element.find(this.options.domSelectors.checkbox).prop('checked', false);
			$currentElement.prop('checked', true);
		} else {
			// a checkbox with a value is checked, uncheck reset checkbox
			this.$element.find(this.options.domSelectors.checkboxReset).prop('checked', false);
		}
	} else {
		// clicking the 'Alle' checkbox must always select this item, it does not
		// matter if it is already selected or not
		if ($currentElement.attr('data-radiocheckbox').indexOf('checkbox_reset') >= 0) {
			$currentElement.prop('checked', true);
		}
	}

	// At this point, if no checkbox is selected, we need to select the 'Alle' checkbox
	checkedCount = this.$element.find(this.options.domSelectors.checkbox + ':checked').length;
	if (checkedCount === 0) {
		this.$element.find(this.options.domSelectors.checkboxReset).prop('checked', true);
	}
};

/**
 * Handle checked state of checkbox, so the checkbox will be unchecked: there are 3 checkboxes that are selected. If then the first or second is unchecked, the following checkboxes will be unchecked as well.
 * @method
 * @private
 */
Module.prototype._checkboxChecked = function(event) {

	var $currentElement = $(event.currentTarget),
		iterator = 0,
		$elem;

	// iterate and select all checkboxes before
	for (iterator; iterator < this.$inputs.length; iterator++) {
		$elem = $(this.$inputs[iterator]);
		if ($elem.get(0).id === $currentElement.get(0).id) {
			return;
		} else {
			$elem.prop('checked', true);
		}
	}
};

/**
 * Handle unchecked state of checkbox, so the checkbox will be checked: there are 3 checkboxes that are not selected. If then the 3rd checkbox is checked, the 1st and 2nd are checked as well.
 * @method
 * @private
 */
Module.prototype._checkboxUnchecked = function(event) {
	var $currentElement = $(event.currentTarget),
		iterator = 0,
		$elem,
		doDeselection = false;

	// iterate and deselect all checkboxes after
	for (iterator; iterator < this.$inputs.length; iterator++) {
		$elem = $(this.$inputs[iterator]);
		if ($elem.get(0).id === $currentElement.get(0).id) {
			doDeselection = true;
		} else {
			if (doDeselection) {
				$elem.prop('checked', false);
			}
		}
	}
};

/**
 * Update target element
 */
Module.prototype._updateTargetElement = function() {
	this.targetAttr = this.$element.attr(this.options.attributes.conditionalControlsTarget);

	if (this.targetAttr && this.targetAttr !== 'false') {
		// cache the target element
		this.$targetElement = $('#' + this.targetAttr);
	}
};

/**
 * Handles visibility of multiple collapsibles
 * @method
 * @private
 */
 Module.prototype._visibilityOfMultipleCollapsibles = function() {
	Array.prototype.forEach.call(this.$inputs, function(input) {
		if (input.hasAttribute(this.options.attributes.conditionalControlsTarget) && input.getAttribute(this.options.attributes.conditionalControlsTarget) !== '') {
			var formId = this.$element.closest(this.options.domSelectors.form).attr('id');

			this.$collapsibleContentItem = $('#' + input.getAttribute(this.options.attributes.conditionalControlsTarget));
			this.$collapsibleContentItem.hide();

			if (input.checked) {
				this.$collapsibleContentItem.show();
				$document.triggerHandler('formchange', [{id: formId}]);
			}
		}
	}.bind(this));
};

/**
 * Handle radio changed
 */
Module.prototype._radioCheckboxSelectionChanged = function() {

	var itemsAreCheckboxes = this.itemsAreOfType(this.$inputs, 'checkbox'),
		$checkedItem = $(this.$element.find(':checked')),
		targetHiddenAttrValue = $checkedItem.attr(this.options.attributes.conditionalTargetHiddenAttr),
		showOrHide = targetHiddenAttrValue;

	if (itemsAreCheckboxes) {
		showOrHide = !($checkedItem.length === 1 && targetHiddenAttrValue);
	}

	// run helper method to show/hide the conditional block
	if (!showOrHide && !this.possibiliytOfSeveralCollapsible) {
		this.$targetElement.show();
	} else if (!this.possibiliytOfSeveralCollapsible) {
		this.$targetElement.hide();
	} else {
		// has multiple collapsible content
		this._visibilityOfMultipleCollapsibles();
	}

	$document.trigger(events.checkboxChanged, [{showOrHide: showOrHide}]);
};

/**
 * Helper method to check if all of the passed input items are of a certain type
 */
Module.prototype.itemsAreOfType = function($items, type) {
	var allItemsAreOfThatType = true;

	$items.each(function(index, item) {
		if ($(item).attr('type') !== type) {
			allItemsAreOfThatType = false;
			return false; // break the loop
		}
	});

	return allItemsAreOfThatType;
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 89 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


__webpack_require__.e/* require.ensure */(3).then((function(require) {
	var $ = __webpack_require__(0);

	__webpack_require__(3);

	var checkGroupSelection = function(value, parsleyField) {
		var groupToValidate = parsleyField.$element,
			selectedElements = $(groupToValidate).find('[data-radiocheckbox="checkbox_item"]:checked');

		// at least one element in the group must be selected
		return selectedElements.length > 0;
	};

	window.Parsley.addValidator('customGroupValidation', {
		requirementType: 'string',
		validate: function(value, state, parsley) {
			return checkGroupSelection(value, parsley);
		}
	});

	// parsley needs to know which fields need to be validated in addition to the default fields (input, textarea, select)
	window.Parsley.options.inputs += ', [data-parsley-custom-group-validation]';
}).bind(null, __webpack_require__)).catch(__webpack_require__.oe);


/***/ }),
/* 90 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(91);


/***/ }),
/* 91 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M21 - Ghettobox
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(92);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	CookieHelper = __webpack_require__(62),
	gsap = __webpack_require__(11).gsap;

var name = 'ghettobox',
	$document = $(document),
	events = {
		closeGhettoBox: 'closeGhettoBox.estatico.' + name
	},
	defaults = {
		domSelectors: {
			closeButton: '[data-' + name + '="close"]',
			messageLink: '[data-' + name + '="message_link"]'
		},
		stateClasses: {},
		cookieExpirationTimeInDays: 30,
		fadeInAnimationDelayFactor: 0.6,
		fadeInAnimationEasing: 'power1.easeInOut',
		fadeOutAnimationEasing: 'power1.easeInOut',
		crisisBoxVarClass: 'var_crisisbox',
		isNotClosable: false
	},
	data = {};

/**
 * Create an instance of the module
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 */
Module.prototype.init = function() {

	this.ghettoboxId = this.$element.attr('id');
	this.ghettoBoxIndex = this.getGhettoBoxIndex();
	this.isCrisisBox = this.checkIfCrisisBox();

	// bind event handlers
	this.bindEventHandlers();

	if (this.options.isNotClosable) {
		return;
	}

	// potentially open ghettobox
	this.potentiallyShowGhettoBox();
};

/**
 * Bind event handlers
 */
Module.prototype.bindEventHandlers = function() {

	// bind close button handler
	this.$element.on('click.' + this.uuid, this.options.domSelectors.closeButton, this.closeGhettoBox.bind(this));

	// bind message link handler to the whole wrapper
	if (this.$element.hasClass('is_clickable')) {
		this.$element.on('click.' + this.uuid, this.$element, this.openMessage.bind(this));
	}

	// bind close event to close method handler
	$document.on(events.closeGhettoBox + '.' + this.uuid, this.closeGhettoBox.bind(this));
};

Module.prototype.getGhettoBoxHeight = function() {
	return this.$element.outerHeight();
};

Module.prototype.getGhettoBoxIndex = function() {
	return this.$element.index();
};

Module.prototype.checkIfCrisisBox = function() {
	return this.$element.hasClass(this.options.crisisBoxVarClass);
};

/**
 * Either show or remove ghettobox
 */
Module.prototype.potentiallyShowGhettoBox = function() {
	if (this.shouldShowGhettoBox()) {
		this.animationDuration = this.calculateAnimationDuration(this.$element, true);
		this.showGhettoBox();
	} else {
		this.removeGhettoBox();
	}
};

/**
 * Show the ghettobox
 */
Module.prototype.showGhettoBox = function() {

	var $ghettoBoxes = this.$element.parent().find('[data-init="ghettobox"]'),
		ghettoBoxHeight = this.getGhettoBoxHeight(),
		fadeInAnimationDelay = 0;

	// In case that there is more than one ghettobox, we need to make sure
	// that they don't animate in at the same time.
	// Among all present ghettoboxes, we find the ghettobox which is represented
	// by this instance. Then we can determine an appropriate animation delay for
	// this instance.
	$.each($ghettoBoxes, function(index, value) {
		if ($(value).attr('id') === this.ghettoboxId) {
			fadeInAnimationDelay = ($ghettoBoxes.length - 1 - index) * this.animationDuration;
		}
	}.bind(this));

	// it does not look that good if we delay this instance by the whole animation
	// duration of the previous one. So we decimate it by a certain factor.
	fadeInAnimationDelay = fadeInAnimationDelay * this.options.fadeInAnimationDelayFactor;

	// animate from initial properties
	requestAnimationFrame(function() {
		gsap.fromTo(this.$element, {
				'display': 'block',
				'margin-top': -ghettoBoxHeight
			},
			{
				'margin-top': 0,
				duration: this.animationDuration,
				ease: this.options.fadeInAnimationEasing,
				delay: fadeInAnimationDelay,
				immediateRender: true,
				onComplete: function() {

					// set focus on Crisis Box
					if (this.isCrisisBox) {
						requestAnimationFrame(function() {
							this.$element.focus();
						}.bind(this));
					}

				}.bind(this)
			});
	}.bind(this));

};

/**
 * Remove the ghettobox
 */
Module.prototype.removeGhettoBox = function() {
	this.$element.remove();
};

/**
 * Close Ghettobox
 */
Module.prototype.closeGhettoBox = function() {

	// determine the height of the ghettobox
	var ghettoBoxHeight = this.getGhettoBoxHeight();

	// animate to target properties and remove ghettobox when animation is finished
	gsap.to(this.$element, {
		duration: this.animationDuration,
		'display': 'none',
		'margin-top': -ghettoBoxHeight,
		ease: this.options.fadeOutAnimationEasing,
		onComplete: this.removeGhettoBox.bind(this)
	});

	// set cookie so that we don't show this ghettobox again
	CookieHelper.setCookie(this.ghettoboxId, this.options.cookieExpirationTimeInDays);

};

/**
 * Helper method to decide whether the ghettobox should be shown
 */
Module.prototype.shouldShowGhettoBox = function() {

	var currentTime = this.getCurrentUnixTimeStamp(),
		options;

	try {
		options = JSON.parse(this.$element.attr('data-options'));
	} catch (e) {
		options = {};
	}

	// if we already have a cookie, we don't want to show the ghettobox.
	if (CookieHelper.hasCookie(this.ghettoboxId)) {
		return false;
	}

	// Check if current client time is within the range in which the ghettobox should get displayed
	if (currentTime > options.showFrom && currentTime < options.showTo) {
		return true;
	} else {
		return false;
	}
};

/**
 * Get animation duration for fadeIn/fadeOut dependant from element-height
 * and viewport width
 */
Module.prototype.calculateAnimationDuration = function($element, isForFadeIn) {
	var fadeDuration = 0.0,
		animationConstant = 0.000015,
		elementHeight = this.getGhettoBoxHeight(),
		exponent = 0.2,
		fadeInCorrectionFactor = 1.4;

	// make the duration dependant from element height
	fadeDuration = Math.pow(elementHeight * animationConstant, exponent);

	// correct the duration for fadeIn
	if (isForFadeIn) {
		fadeDuration *= fadeInCorrectionFactor;
	}

	return fadeDuration;
};

/**
 * Open message.
 */
Module.prototype.openMessage = function(event) {

	var redirectLink;

	event.preventDefault();

	// only proceed if the user did not click on the close button
	var eventTargetDataAttribute = $(event.originalEvent.target).attr('data-' + name),
		closeButtonDataAttribute = this.$element.find(this.options.domSelectors.closeButton).attr('data-' + name),
		$messageLink = this.$element.find(this.options.domSelectors.messageLink);

	// a close button exists and the user did click on it. so ignore this action and return
	if (eventTargetDataAttribute === closeButtonDataAttribute && !this.options.isNotClosable) {
		return;
	}

	// get the link from the a tag
	redirectLink = $messageLink.attr('href');

	// set new location to window object
	if (window.tracking) {
		this.sendTrackingData($messageLink, redirectLink);
	} else {
		if (this.options.isNotClosable) {
			window.open(redirectLink, '_blank');
			return;
		}

		window.location.href = redirectLink;
	}
};

/**
 * Send tracking data
 */
Module.prototype.sendTrackingData = function($messageLink, redirectLink) {
	var regExp = /^\.*\//,
		destinationUrl = redirectLink,
		openInNewTab = this.options.isNotClosable;

	// check if the redirect link is an absolute or a relative path
	if (redirectLink.match(regExp) !== null) {
		destinationUrl = window.location.protocol + '//' + window.location.hostname + redirectLink;
	}

	window.tracking.dataLayerApi.trackGenericNavigation({
		label: $messageLink[0].innerText,
		component: this.$element[0],
		destination: destinationUrl,
		callback: function() {
			if (openInNewTab) {
				window.open(redirectLink, '_blank');
				return;
			}

			window.location.href = redirectLink;
		}
	});
};

/**
 * Get the current time
 */
Module.prototype.getCurrentUnixTimeStamp = function() {
	return Math.floor((performance.timing.navigationStart + performance.now()) / 1000);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 92 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 93 */,
/* 94 */,
/* 95 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(96);


/***/ }),
/* 96 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 97 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(98);


/***/ }),
/* 98 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M08 - Breadcrumb
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(99);

var $ = __webpack_require__(0),
	events = __webpack_require__(5),
	SuperClass = __webpack_require__(1),
	keymaster = __webpack_require__(14);

var name = 'breadcrumb',
	defaults = {
		domSelectors: {
			button: '[data-' + name + '="button"]',
			buttonNoSibling: '[data-' + name + '="button-no-siblings"]',
			dropdown: '[data-' + name + '="dropdown"]',
			wrapper: '[data-' + name + '="inner-wrapper"]',
			itemsLev1: '[data-' + name + '="list-item"]',
			siblings: '[data-' + name + '="siblings-url"]',
			expandLinks: '[data-' + name + '="expand-links"]',
			homeLink: '[data-' + name + '="home-link"]',
			putOnTop: '[data-' + name + '="put-on-top"]'
		},
		stateClasses: {
			isExpanded: 'is_expanded'
		},
		dropdownOuterSpacing: 15
	},
	isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints,
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @private
 */
Module.prototype.init = function() {
	var domSelectors = this.options.domSelectors,
		$element = this.$element;

	this.$document = $(document);
	this.$buttons = this.$element.find(domSelectors.button);

	this.allItemsVisible = false;

	$element.on('click.'+ this.uuid, domSelectors.button, function(event) {
		event.preventDefault();
		this.toggleNavigation($(event.currentTarget));
	}.bind(this));

	this.setupScroll();
	this.setupKeyboardNavigation();
	this.setupAriaAttributes();

	// Used to fade in the breadcrumb on moblie once it is initialized.
	this.$element.addClass('is_initialized');
};

Module.prototype.setupAriaAttributes = function() {

	this.$buttons.each(function(index, element) {
		var $button = $(element),
			$dropdown = this.getRelatedDropdown($button),
			id = $dropdown.uniqueId().attr('id');

		$button.attr({
			'aria-haspopup': 'true',
			'aria-controls': id,
			role: 'button'
		});

		this.setDropdownAriaTags($dropdown, false);
		this.setButtonAriaTags($button, false);

	}.bind(this));
};

Module.prototype.setDropdownAriaTags = function($dropdown, isOpen) {
	$dropdown.attr({
		'aria-expanded': !!isOpen
	});
};

Module.prototype.setButtonAriaTags = function($button, isOpen) {
	$button.attr('aria-expanded', !!isOpen);
};

Module.prototype.setupKeyboardNavigation = function() {
	var moduleScope = this.name + '-' + this.uuid,
		$siblings = this.$element.find(this.options.domSelectors.siblings),
		$buttonsNoSiblings = this.$element.find(this.options.domSelectors.buttonNoSibling),
		$homeLink = this.$element.find(this.options.domSelectors.homeLink);

	this.$focusElements = this.$buttons.add($homeLink).add($buttonsNoSiblings);

	$siblings.attr('tabindex', '-1');

	this.$element.on('focusin.' + this.uuid, function() {
		keymaster.setScope(moduleScope);
	}).on('focusout.' + this.uuid, function() {
		keymaster.setScope('all');
	});

	// handle events when appropriate
	keymaster('esc', moduleScope, function() {
		this.closeDropdown();
	}.bind(this));

	keymaster('right, pagedown', moduleScope, this.focusNext.bind(this));
	keymaster('left, pageup', moduleScope, this.focusPrevious.bind(this));
	keymaster('home', moduleScope, this.focusFirst.bind(this));
	keymaster('end', moduleScope, this.focusLast.bind(this));
	keymaster('down', moduleScope, this.focusNextSibling.bind(this));
	keymaster('up', moduleScope, this.focusPreviousSibling.bind(this));
	keymaster('space', moduleScope, this.handleSpace.bind(this));
};

Module.prototype.handleSpace = function(event) {
	var $focusElement = this.getCurrentBreadcrumbElement();

	if ($focusElement.length < 1) {
		return;
	}

	event.preventDefault();
	this.toggleNavigation($focusElement);
};

Module.prototype.focusNext = function(event) {
	this.focus(event, {delta: +1});
};

Module.prototype.focusPrevious = function(event) {
	this.focus(event, {delta: -1});
};

Module.prototype.focusFirst = function(event) {
	this.focus(event, {nextItemIndex: 0});
};

Module.prototype.focusLast = function(event) {
	this.focus(event, {nextItemIndex: -1});
};

Module.prototype.focusNextSibling = function(event) {
	this.focusSibling(event, {delta: +1});
};

Module.prototype.focusPreviousSibling = function(event) {
	this.focusSibling(event, {delta: -1});
};

Module.prototype.focusSibling = function(event, options) {
	var $button = this.getCurrentBreadcrumbElement(),
		focusSibling = function() {
			var $siblings = $button.siblings(this.options.domSelectors.dropdown).find(this.options.domSelectors.siblings),
				$currentlyFocusedSibling = $siblings.filter(':focus'),
				size = $siblings.length,
				currentlyFocusedSiblingIndex, siblingIndexToFocus;

			if ($currentlyFocusedSibling.length) {
				currentlyFocusedSiblingIndex = $siblings.index($currentlyFocusedSibling);
				siblingIndexToFocus = (size + currentlyFocusedSiblingIndex + options.delta) % size;
			} else {
				if (options.delta === -1) {
					siblingIndexToFocus = size - 1;
				} else {
					siblingIndexToFocus = 0;
				}
			}

			$($siblings[siblingIndexToFocus]).focus();
		}.bind(this);

	event.preventDefault();

	if (this.isExpanded($button)) {
		focusSibling();
	} else {
		this.onOpenDropdown = focusSibling;
		this.toggleNavigation($button);
	}
};

Module.prototype.getCurrentBreadcrumbElement = function() {
	return this.$focusElements.filter('.is_in_focus');
};

Module.prototype.closeDropdown = function() {
	// This will be overwritten when needed, but should always be callable.
};

Module.prototype.onOpenDropdown = function() {
	// This will be overwritten when needed, but should always be callable.
};

Module.prototype.focus = function(event, options) {
	var delta = options.delta,
		nextItemIndex = options.nextItemIndex,
		$currentItem = this.getCurrentBreadcrumbElement(),
		currentItemIndex = this.$focusElements.index($currentItem),
		size = this.$focusElements.length,
		$nextItem;

	event.preventDefault();
	this.closeDropdown();

	if (typeof nextItemIndex === 'number') {
		// pass -1 for the last element.
		nextItemIndex = (nextItemIndex + size) % size;
	} else if (delta) {
		nextItemIndex = (size + currentItemIndex + delta) % size;
	}

	$nextItem = $(this.$focusElements[nextItemIndex]);

	this.focusFromHere($nextItem);
	$nextItem.focus();
};

Module.prototype.focusFromHere = function($newFocusableItem, $currentItem) {
	$currentItem = $currentItem || this.getCurrentBreadcrumbElement();
	$currentItem.removeClass('is_in_focus');
	$newFocusableItem.addClass('is_in_focus').focus();
};

Module.prototype.setupScroll = function() {
	var $wrapper = this.$element.find(this.options.domSelectors.wrapper);

	this.isScrolling = $wrapper.css('overflow') !== 'visible';

	if (!this.isScrolling) {
		return;
	}

	this.$wrapper = $wrapper;

	this.setupExpandLinks();
};

Module.prototype.setupExpandLinks = function() {
	var buttonSelector = this.options.domSelectors.expandLinks,
		listItemsSelector = this.options.domSelectors.itemsLev1,
		needsExpandButton = this.$element.find(listItemsSelector).length > 1;

	if (needsExpandButton) {
		this.$element.find(buttonSelector)
			.attr('aria-haspopup', 'true')
			.attr('aria-expanded', 'false')
			.click(function() {
				this.$element.addClass('var_has_expanded_links');
				this.allItemsVisible = true;
			}.bind(this));
	} else {
		this.$element.addClass('var_has_expanded_links');
	}
};

Module.prototype.isExpanded = function($element) {
	return this.getRelatedItem($element).hasClass(this.options.stateClasses.isExpanded);
};

Module.prototype.getRelatedDropdown = function($element) {
	return this.getRelatedItem($element).find(this.options.domSelectors.dropdown);
};

Module.prototype.getRelatedItem = function($element) {
	return $element.closest(this.options.domSelectors.itemsLev1);
};

Module.prototype.toggleNavigation = function($button, amountOfRecursion) {

	var isExpandedClass = this.options.stateClasses.isExpanded,
		$item = this.getRelatedItem($button),
		$dropdown = this.getRelatedDropdown($item),
		extraPadding = 24,
		elementWidth,
		titleWidth,
		dropDownWidth,
		parentOffset,
		moduleWidth;

	amountOfRecursion = amountOfRecursion || 0;
	amountOfRecursion++;

	if (this.isExpanded($button)) {
		return;
	}

	parentOffset = this.$element.offset();
	moduleWidth = this.$element.width();
	titleWidth = $item.find(this.options.domSelectors.putOnTop).width();

	if (titleWidth > moduleWidth) {
		$item.addClass('var_full_width');
	}

	// Narrows the item to a max width of 100%
	$item.addClass('is_opening_dropdown');

	// if dropdown is out of the viewport, opening to the left side
	if ((moduleWidth + parentOffset.left) < ($dropdown.outerWidth() + $item.offset().left)) {
		$dropdown.addClass('is_opening_dropdown_left');
	}

	// ensure that the title is not bigger than the dropdown.
	elementWidth = $item.width();
	dropDownWidth = parseInt($dropdown.outerWidth());

	// this is for the portrait size (e.g. ipad)
	if (isTouchDevice) {
		var realElementWidth = parseInt(elementWidth + extraPadding);

		if (realElementWidth > dropDownWidth) {
			$dropdown.css({
				width: realElementWidth
			});
		}
	}

	if (this.isScrolling) {
		// Ensure position of the dropdown on mobile. I want to specially thank
		// Apple and Moblie Safari for being s.th. I will not name here.
		var scrollPosition = this.$wrapper.scrollLeft(),
			itemWidth = $dropdown.outerWidth(),
			itemHeight = $item.outerHeight(),
			itemOffset = $item.offset(),
			textPosition = $item.find('.mod_breadcrumb_button').offset().left,
			isUnderflowing = parentOffset.left > textPosition,
			isOverflowing = (moduleWidth + parentOffset.left) < (itemWidth + itemOffset.left),
			recurse, amount;

		if (amountOfRecursion < 3) {
			recurse = function() {
				// Wait for the scrolling to be completely done.
				requestAnimationFrame(function() {
					// Recursively call the function after the animation.
					this.toggleNavigation($button, amountOfRecursion);
				}.bind(this));
			}.bind(this);

			if (isUnderflowing) {
				amount = parentOffset.left - textPosition;

				this.$wrapper.animate({
					scrollLeft: Math.round(scrollPosition - amount)
				}, 300, recurse);
				return;
			}

			if (isOverflowing) {
				amount = (itemWidth + itemOffset.left) - (moduleWidth + parentOffset.left);

				this.$wrapper.animate({
					scrollLeft: Math.round(scrollPosition + amount)
				}, 300, recurse);
				return;
			}
		}

		if ($item.is(':last-child') && this.allItemsVisible) {
			requestAnimationFrame(function() {
				$dropdown.css({
					left: 'auto',
					right: '0.5px',
					top: (itemOffset.top - parentOffset.top + itemHeight) + 'px'
				});
				$item.addClass('var_is_last');
			});
		} else {
			requestAnimationFrame(function() {
				$dropdown.css({
					left: (itemOffset.left - parentOffset.left) + 'px',
					top: (itemOffset.top - parentOffset.top + itemHeight) + 'px'
				});
			});
		}
	}

	this.ensureVisibleTitle($button);

	// Setup dropdown close events.
	setTimeout(function() {
		// This is in an setTimeout, as we don't want the click that triggered
		// the toggleNavigation() to bubble up and close the dropdown again.
		var callback = this.closeDropdown = function() {
			this.$element.removeClass(isExpandedClass);
			$item.removeClass(isExpandedClass);
			$item.removeClass('is_opening_dropdown');
			$dropdown.removeClass('is_opening_dropdown_left');
			$item.removeClass('var_is_last');

			this.$document.off('click', callback);

			if (this.isScrolling) {
				this.$wrapper.off('scroll', callback);
				this.$wrapper.removeClass(isExpandedClass);
			}

			$button.focus();
			var isOpen = false;

			this.setDropdownAriaTags($dropdown, isOpen);
			this.setButtonAriaTags($button, isOpen);

			this.closeDropdown = function() {};

		}.bind(this);

		this.$document.one('click', callback);

		if (this.isScrolling) {
			this.$wrapper.addClass(isExpandedClass);
			this.$wrapper.one('scroll', callback);
		}

		$item.addClass(isExpandedClass);
		this.focusFromHere($button);

		var isOpen = true;

		this.setDropdownAriaTags($dropdown, isOpen);
		this.setButtonAriaTags($button, isOpen);

		this.onOpenDropdown();
		this.onOpenDropdown = function() {};

	}.bind(this), 0);

};

Module.prototype.ensureVisibleTitle = function($button) {
	var selector = this.options.domSelectors.putOnTop,
		classToAdd = 'var_is_on_top';

	// Remove class from previous interaction
	this.$element.find(selector + '.' + classToAdd)
		.removeClass(classToAdd);
	$button.find(selector).addClass(classToAdd);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 99 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 100 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(101);


/***/ }),
/* 101 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M22 - Lightbox
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(102);

var $ = __webpack_require__(0),
	template = __webpack_require__(51),
	templateSettings = __webpack_require__(27),
	cssTransitionEnd = __webpack_require__(4),
	LockTabFocus = __webpack_require__(25),
	SuperClass = __webpack_require__(1),
	RemoveTabIndex = __webpack_require__(19),
	globalEvents = __webpack_require__(5),
	contains = __webpack_require__(15);

var name = 'lightbox',
	log = estatico.helpers.log(name),
	$document = $(document),
	$body = $('body'),
	isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints,
	events = {
		scrollBarUpdate: 'scrollBarUpdated.estatico',
		scrollBarScrolled: 'scrollBarScrolled.estatico',
		scrollBarInitialized: 'scrollBarInitialized.estatico'
	},
	defaults = {
		domSelectors: {
			lightboxCloseButton: '[data-' + name + '="close"]',
			lightboxContainer: '[data-' + name + '="container"]',
			lightboxContentContainer: '[data-' + name + '="content"]',
			lightboxTriggerButton: '[data-' + name + '-trigger]',
			lightboxHeader: '[data-' + name + '="header"]',
			lightboxBackground: '[data-' + name + '="background"]',
			lightboxAlertTemplate: '[data-' + name + '="alert_template"]',
			lightboxAlertWindow: '[data-' + name + '="alert"]',
			lightboxAlertCancel: '[data-' + name + '="alert_cancel_button"]',
			lightboxAlertConfirm: '[data-' + name + '="alert_confirm_button"]',
			lightboxAttacheableHeader: '[data-' + name + '="attacheable_header"]',
			lightboxAttacheableFooter: '[data-' + name + '="attacheable_footer"]'
		},
		stateClasses: {
			isVisible: 'is_visible',
			isBottom: 'is_bottom',
			isOpen: 'is_open',
			isContentLoaded: 'is_content_loaded',
			isContentCenteredHorizontally: 'is_content_centered_horizontally',
			isContentCenteredVertically: 'is_content_centered_vertically',
			isPreloaderVisible: 'is_preloader_visible',
			isBodyLightboxOpen: 'is_body_lightbox_open',
			isAnimationEnded: 'is_animation_ended',
			hasAlert: 'has_alert',
			isDisabled: 'is_disabled'
		},
		scrollBarPluginConfig: {
			axis: 'y',
			advanced: {
				updateOnContentResize: true,
				autoUpdateTimeout: 300
			}
		},
		$lightbox: undefined,
		$lightboxNodes: {},
		$triggerButtons: {},
		$lightboxContentContainer: undefined,
		$lightboxClosingAlert: undefined,
		resizeCallback: undefined,
		askBeforeClosing: false,
		closeKeys: [27], // ESC
		closeAlert: {
			confirmCloseClick: {
				action: 'closeOverlay'
			}
		},
		i18n: {
			closeAlert: {
				message: 'Nachricht mit Anweisungen',
				cancelButton: 'Zurück zum Lightbox Inhhalt',
				confirmButton: 'Lightbox schliessen'
			}
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element.innerHTML, // because original element is enclosed in <script> tag, use innerHtml to get template and pass it down to constructor
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	this.isInitialLoad = true;

	// The trigger for the role selection overlay happens aysnc and may occur
	// before the _init function is executed. So we start to listen here but
	// can not execute _onJSLightboxTrigger immediately since the lazy loaded
	// and necessary dependencies of this module might not be ready by that time.
	//
	// So we store the dataObject and access it from within the _init function once
	// we are sure that everything is ready
	$document.on('lightboxJSTrigger', function(event, dataObject) {
		this.missedTriggerPayLoad = dataObject;
	}.bind(this));

	// Check if there's already estatico instance of this module and do not create more than one instance
	if (Object.keys(estatico.modules[name].instances).length > 1) {

		// Lightbox module cannot have more than one instance
		return false;
	}

	// Interpolate for hbs templates syntax when parsing with lodash.template function
	templateSettings.interpolate = /{{([\s\S]+?)}}/g;

	this.elementToFocus = 0;

	if (!isTouchDevice) {
		// The following code will tell webpack to create a file named wm02_-_processflow_plugins.js (name defined as last argument)
		// and save it to assets/js/async/ (specified in gulp/js/default.js). This file will then get loaded on demand only for devices
		// where neither touchevents nor pointerevents are available
		__webpack_require__.e/* require.ensure */(2/* duplicate */).then((function(require) {
				__webpack_require__(24);
				__webpack_require__(23)($);
				setTimeout(function() {
					this._init();
				}.bind(this), 1);
			}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);
	} else {
		this._init();
	}
};

/**
 * Private init function for when the custom scrollbar is created
 * @private
 */
Module.prototype._init = function() {

	// set current state
	history.pushState({
		name: 'lightBoxOpen'
	}, 'lightbox');

	// We either consume what is there yet or set the new listener
	if (this.missedTriggerPayLoad) {
		this._onJSLightboxTrigger({}, this.missedTriggerPayLoad);
	}

	$document.on('lightboxJSTrigger', this._onJSLightboxTrigger.bind(this));

	$.extend(true, events, {
		click: 'click.estatico.' + this.uuid,
		keydown: 'keydown.estatico.' + this.uuid
	});

	// Save instance to DOM element data
	this.$element.data(name + 'Instance', this);

	// Cache all trigger buttons
	this.options.$triggerButtons = $document.find(this.options.domSelectors.lightboxTriggerButton);

	// Add listener to lightbox close button
	this.$closeButton = this.$element.find(this.options.domSelectors.lightboxCloseButton);
	this.$closeButton.on(events.click, this._closeLightbox.bind(this));

	// Cache lightbox content container
	this.options.$lightboxContentContainer = this.$element.find(this.options.domSelectors.lightboxContentContainer);

	// Cache lightbox header and content container
	this.options.$lightboxContainer = this.$element.find(this.options.domSelectors.lightboxContainer);

	// Add aria attributes to all trigger buttons
	this._addAriaAttributesToTriggerButtons();

	// Add event listeners to trigger buttons
	this._addEventListenersToTriggerButtons();

	this._listenForBrowserBackButton();

	// Append lightbox to body element
	if (this.$element.data('appendToBody')) {
		$body.append(this.$element);
	}

	// Cache _resizeCallback function with binded context, so event listener can be removed when not type of iframe
	this.options.resizeCallback = this._resizeCallback.bind(this);

	// Initialize custom scrollbar if necessary
	// trigger only on no-touch devices
	if (!isTouchDevice) {

		this.options.scrollBarPluginConfig.callbacks = {
			onInit: function() {
				window.requestAnimationFrame(function() {
					$document.trigger(events.scrollBarInitialized);
				});
			},

			whileScrolling: function() {
				window.requestAnimationFrame(function() {
					$document.trigger(events.scrollBarScrolled);
				});
			},

			onUpdate: function() {
				window.requestAnimationFrame(function() {
					$document.trigger(events.scrollBarUpdate);
				});

				RemoveTabIndex.removeTabIndex(this);
			}

		};
	}
};

/**
 * For the time when XHR request is being processed, show preloader div
 * @private
 */
Module.prototype._showPreloader = function() {

};

/**
 * Check if a header object we can attach to lightbox header,
 * exists within newly rendered/loaded template.
 * @param $content
 * @private
 */
Module.prototype._attachAdditionalHeaderIfPresent = function($content) {
	var $header = $content.find(this.options.domSelectors.lightboxAttacheableHeader);

	if ($header.length) {
		this.$element.find(this.options.domSelectors.lightboxHeader).append($header);
	}
};

/**
 * Check if a footer object we can attach to lightbox,
 * exists within newly rendered/loaded template.
 * @param $content
 * @private
 */
Module.prototype._attachAdditionalFooterIfPresent = function($content) {
	var $footer = $content.find(this.options.domSelectors.lightboxAttacheableFooter);

	if ($footer.length) {
		this.options.$lightboxContainer.append($footer);
	}
};

/**
 * Check if header has been attached, and if yes detach it.
 * @private
 */
Module.prototype._detachAdditionalHeaderIfPresent = function() {
	var $header = this.$element.find(this.options.domSelectors.lightboxAttacheableHeader);

	if ($header) {
		$header.remove();
	}
};

/**
 * Check if footer has been attached, and if yes detach it.
 * @private
 */
Module.prototype._detachAdditionalFooterIfPresent = function() {
	var $footer = this.$element.find(this.options.domSelectors.lightboxAttacheableFooter);

	if ($footer) {
		$footer.remove();
	}
};

/**
 * hide preloader div when loading of content is complete
 * @private
 */
Module.prototype._hidePreloader = function() {

};

/**
 * Adjust iframe height to content height
 * @private
 */
Module.prototype._adjustIframeHeight = function() {
	var $el = this.$element.find('iframe'),
		elementHeight = this.$element.height(),
		headerHeight = this.$element.find(this.options.domSelectors.lightboxHeader).outerHeight();

	$el.css({height: elementHeight - headerHeight});
};

/**
 * Function triggered on window resize when iframe contnet is displayed
 * @private
 */
Module.prototype._resizeCallback = function() {
	this._adjustIframeHeight();
};

/**
 * Add event listeners to all buttons with lightbox-trigger attribute defined and to
 * document to listen to trigger event
 * @private
 */
Module.prototype._addEventListenersToTriggerButtons = function() {
	this.options.$triggerButtons.on(events.click, this._onLightboxButtonClick.bind(this));
};

Module.prototype._listenForBrowserBackButton = function() {
	window.addEventListener('popstate', function(event) {
		event.preventDefault();

		if ('scrollRestoration' in history) {
			history.scrollRestoration = 'manual';
		}

		window.removeEventListener('popstate', function() {});

		this._closeLightbox();

	}.bind(this));
};

/**
 * Function triggered when event lightboxJSTrigger is sent from document scope
 * Example trigger call: $document.triggerHandler('lightboxJSTrigger', { type: 'node', source: '#lightbox1'});
 * Object structure comes from data.js file, and assumes that there's a node with id #lightbox1 generated as a template - in this particular case.
 * @param dataObject - object constructed the same way as in the data.js files
 * @method
 * @private
 */
Module.prototype._onJSLightboxTrigger = function(event, dataObject) {

	if (!dataObject) {
		// dataObject is empty if the lightboxJSTrigger event was triggered natively with a CustomEvent
		dataObject = (event && event.originalEvent ? event.originalEvent.detail : undefined) || {};
	}

	// Overwrite existing option values
	if (typeof dataObject.options === 'object' && dataObject.options !== null) {

		Object.keys(dataObject.options).forEach(function(key) {
			this.options[key] = dataObject.options[key];
		}.bind(this));

		// save element to focus for later
		if (dataObject.options.elementToFocus) {
			this.elementToFocus = dataObject.options.elementToFocus;
		}

		if (dataObject.options.elementToFocusAfterClose) {
			this.$elementToFocusAfterClose = dataObject.options.elementToFocusAfterClose;
		} else {
			this.$elementToFocusAfterClose = $body;
		}

	}

	this._openLightbox();

	requestAnimationFrame(function() {
		this._attachContentToLightbox(dataObject);
	}.bind(this));

};

/**
 * Lightbox button click event handler
 * @param event
 * @private
 */
Module.prototype._onLightboxButtonClick = function(event) {

	var trigger = $(event.currentTarget).data('lightboxTrigger');

	event.preventDefault();

	this.$elementToFocusAfterClose = $(event.currentTarget);

	if (trigger.additionalClasses) {
		this.$element.addClass(trigger.additionalClasses);
	}

	this._openLightbox();

	setTimeout(function() {
		this._attachContentToLightbox(trigger);
	}.bind(this), 10);

};

/**
 * Listen to keyboard input for lightbox
 * @private
 */
Module.prototype._addKeyboardListeners = function() {
	$document.on(events.keydown, this._closeLightboxKeyboard.bind(this));
};

/**
 * Remove event listener when lightbox is not present
 * @private
 */
Module.prototype._removeKeyboardListeners = function() {
	$document.off(events.keydown, this._closeLightboxKeyboard.bind(this));
};

/**
 * Close lightbox on pressing "esc" button
 * @param event
 * @private
 */
Module.prototype._closeLightboxKeyboard = function(event) {
	if (contains(this.options.closeKeys, event.keyCode)) {

		// in some cases we have elements within the lightbox which
		// should close themselves when the users press esc. in this
		// case we do not want to close the lightbox
		if (document.activeElement.className.indexOf('mod_tooltip') !== -1) {
			return;
		}

		this._closeLightbox();
	}
};

/**
 * Adds aria-has-popup attribute to all trigger buttons in the document
 * @private
 */
Module.prototype._addAriaAttributesToTriggerButtons = function() {
	this.options.$triggerButtons.attr('aria-haspopup', true);
};

/**
 * Prevents scrolling of page in the background on mobile devices,
 * when lightbox is open
 * @private
 */
Module.prototype._disableScrollingOfPageOnMobile = function() {
	if (isTouchDevice) {
		$('#main , ' + this.options.domSelectors.lightboxHeader + ' , ' + this.options.domSelectors.lightboxAttacheableFooter).on('touchmove.lightbox', function(event) {
			event.preventDefault();
		});
	}
};

/**
 * Enables scrolling of the page on mobile devices when lightbox is
 * closed.
 * @private
 */
Module.prototype._enableScrollingOfPageOnMobile = function() {
	if (isTouchDevice) {
		$('#main , ' + this.options.domSelectors.lightboxHeader + ' , ' + this.options.domSelectors.lightboxAttacheableFooter).off('touchmove.lightbox');
	}
};

/**
 * Open lightbox - show it on screen with content already loaded
 * @private
 */
Module.prototype._openLightbox = function() {

	// Set object display to block
	this.$element.addClass(this.options.stateClasses.isVisible);

	// Add body class
	$body.addClass(this.options.stateClasses.isBodyLightboxOpen);

	// Wait a frame before adding new class - that will trigger animation.
	requestAnimationFrame(function() {

		// Add class that trigger opening animation
		this.$element.addClass(this.options.stateClasses.isOpen);

		if (cssTransitionEnd) {
			this.$element.one(cssTransitionEnd + '.' + this.uuid, function() {
				this.$element.addClass(this.options.stateClasses.isAnimationEnded);
			}.bind(this));
		} else {
			setTimeout(function() { // wait until transition will end
				this.$element.addClass(this.options.stateClasses.isAnimationEnded);
			}.bind(this), 500);
		}

	}.bind(this));

	setTimeout(function() {

		// Update aria attributes
		this.$element.attr('aria-hidden', false).attr('aria-live', 'polite').attr('aria-relevant', 'additions');

		// Add keyboard listeners
		this._addKeyboardListeners();

	}.bind(this), 100);
};

/**
 * Hide lightbox on screen - play animation and eventually remove display
 * @private
 */
Module.prototype._closeLightbox = function() {

	// Check if there's a modal asking for permission to close the lightbox
	if (this.options.askBeforeClosing) {
		this._showClosingAlert();
	} else {

		// Enable scrolling on mobile devices
		this._enableScrollingOfPageOnMobile();

		// Use RequestAnimationFrame to streamline actions within browser rendering cycle
		window.requestAnimationFrame(function() {
			this.$element.attr('aria-hidden', true).removeAttr('aria-live').removeAttr('aria-relevant');
			this._removeKeyboardListeners();

			// Remove body class
			$body.removeClass(this.options.stateClasses.isBodyLightboxOpen);

			if (cssTransitionEnd) {
				this.$element.one('transitionend', function() {
					this.$element.removeClass(this.options.stateClasses.isVisible);
					this._detachCurrentContent();

					if (this.$elementToFocusAfterClose) {
						this.$elementToFocusAfterClose.focus();
					}

				}.bind(this));
			} else {
				this._detachCurrentContent();

				if (this.$elementToFocusAfterClose) {
					this.$elementToFocusAfterClose.focus();
				}
			}

			LockTabFocus.destroy();

			this.$element.removeClass(this.options.stateClasses.isAnimationEnded);
			this.$element.removeClass(this.options.stateClasses.isOpen);

		}.bind(this));
	}

};

/**
 * Function triggered on successful loading of content with AJAX from _attachExternalCall function
 * @param response from AJAX call to fill content node
 * @private
 */
Module.prototype._onAttachContentSuccess = function(response) {
	this.options.$lightboxContentContainer.html(response);
	this._hidePreloader();
	$document.triggerHandler('ajaxload');

	// Try finding additional trigger buttons for closing
	// This needs to be called before attaching additional header and footer, because then elements might not be present in $lightboxContentContainer
	this._findAdditionalCloseButtons(this.options.$lightboxContentContainer);

	// Try attaching header
	this._attachAdditionalHeaderIfPresent(this.options.$lightboxContentContainer);

	// Try attaching footer
	this._attachAdditionalFooterIfPresent(this.options.$lightboxContentContainer);

	// Add class indicating that content has been loaded
	this.$element.addClass(this.options.stateClasses.isContentLoaded);

	// Disable scrolling of page on mobile devices
	this._disableScrollingOfPageOnMobile();

	// Initialize custom scroll bar
	this._initCustomScrollbar();

	LockTabFocus.setFocusOnElement(this.$element, this.elementToFocus);
};

/**
 * Initialize custom scroll bar plugin.
 * @private
 */
Module.prototype._initCustomScrollbar = function() {
	if (isTouchDevice) {
		return;
	}

	this.$element.find('.mod_lightbox_content_container').mCustomScrollbar(this.options.scrollBarPluginConfig);
};

/**
 * Searches for additional close lightbox triggers within given content
 * @param $content
 * @private
 */
Module.prototype._findAdditionalCloseButtons = function($content) {
	$content.find(this.options.domSelectors.lightboxCloseButton).on(events.click, this._closeLightbox.bind(this));
};

/**
 * Remove event listener off of the additional close buttons in given content
 * @param $content
 * @private
 */
Module.prototype._removeAdditionalCloseButtons = function($content) {
	$content.find(this.options.domSelectors.lightboxCloseButton).off(events.click);
};

/**
 * Function triggered when AJAX call fails
 * @param response from AJAX call
 * @private
 */
Module.prototype._onAttachContentError = function(response) {
	log(response);
};

/**
 * Attach resource from resource url - loaded with ajax call
 * @param resourceURL
 * @private
 */
Module.prototype._attachExternalContent = function(resourceURL) {
	$.ajax({
		url: resourceURL,
		success: this._onAttachContentSuccess.bind(this),
		error: this._onAttachContentError.bind(this)}
	);
	this._showPreloader();
};

/**
 * Attach iframe content based on passed url parameter
 * @param iframeContentURL string passed with qulified uri to iframe content
 * @private
 */
Module.prototype._attachIframe = function(iframeContentURL) {
	var node = document.createElement('iframe');

	node.onload = this._onIframeLoaded.bind(this);
	node.src= iframeContentURL;
	node.style.width = '100%';
	node.style.height = '100%';

	this.options.$lightboxContentContainer.html(node);
	this._adjustIframeHeight();
	$document.on(globalEvents.resize.key + '.' + this.uuid, this.options.resizeCallback);

	// Initialize custom scroll bar
	this._initCustomScrollbar();
};

/**
 * Function triggered when iframe content finished loading
 * @private
 */
Module.prototype._onIframeLoaded = function() {
	this.$element.addClass(this.options.stateClasses.isContentLoaded);
	LockTabFocus.setFocusOnElement(this.$element, this.elementToFocus);
};

/**
 * Removes currently embedded content from div container
 * @private
 */
Module.prototype._detachCurrentContent = function() {
	// find all data-init options and get module instances, and then destroy them.
	var createdModules = this.options.$lightboxContentContainer.find('[data-init]'),
		$el,
		modules,
		instance;

	createdModules.each(function(index, element) {
		$el = $(element);
		modules = $el.data('init').split(' ');
		modules.forEach(function(module) {
			instance = $el.data(module + 'Instance');
			if (instance) {
				instance.destroy();
			}
		});
	}.bind(this));

	// remove centering classes
	this.options.$lightboxContainer.removeClass(this.options.stateClasses.isContentCenteredHorizontally + ' ' + this.options.stateClasses.isContentCenteredVertically);

	// remove tab index from content container
	this.options.$lightboxContentContainer.attr('tabindex', -1);

	// Make sure LightboxContentContainer is initially visible and the closing alert is hidden.
	if (this.options.askBeforeClosing) {
		this._showLightboxContentContainer();
		this._hideClosingAlert();
	}

	// detach additional header
	this._detachAdditionalHeaderIfPresent();

	// detach additional footer
	this._detachAdditionalFooterIfPresent();

	// detach additional close buttons from content
	this._findAdditionalCloseButtons(this.options.$lightboxContentContainer);

	// remove children from lightbox
	this.options.$lightboxContentContainer.empty();

	// remove resize event listener
	$document.off(globalEvents.resize.key, this.options.resizeCallback);

	if (!isTouchDevice) {
		this.$element.find('.mod_lightbox_content_container').mCustomScrollbar('destroy');
	}

};

/**
 * Attach DOM node based on attribute
 * @param $domObject
 * @private
 */
Module.prototype._attachDocumentNode = function($domObject) {
	this.options.$lightboxContentContainer.html($domObject.html());
	$document.triggerHandler('ajaxload');

	// Try attaching header
	this._attachAdditionalHeaderIfPresent(this.options.$lightboxContentContainer);

	// Try attaching footer
	this._attachAdditionalFooterIfPresent(this.options.$lightboxContentContainer);

	// Add class indicating that content has been loaded
	this.$element.addClass(this.options.stateClasses.isContentLoaded);

	// Disable scrolling of other content than lightbox
	this._disableScrollingOfPageOnMobile();

	// Initialize custom scroll bar
	this._initCustomScrollbar();

	LockTabFocus.setFocusOnElement(this.$element, this.elementToFocus);
};

/**
 * Checks if closing alert should be shown
 * @param options
 * @private
 */
Module.prototype._checkForClosingAlert = function() {
	// check if there was an closing alert modal and remove
	if (this.options.$lightboxClosingAlert) {
		this._destroyClosingAlertTemplate();
	}

	if (this.options.askBeforeClosing) {
		this._renderClosingAlertTemplate();
	}
};

Module.prototype._renderClosingAlertTemplate = function() {

	var templateFunction = template($document.find(this.options.domSelectors.lightboxAlertTemplate).html());

	this.options.askBeforeClosing = true;

	this.options.$lightboxClosingAlert = $(templateFunction({
		message: this.options.i18n.closeAlert.message,
		cancelButton: this.options.i18n.closeAlert.cancelButton,
		confirmButton: this.options.i18n.closeAlert.confirmButton
	}));

	this.options.$lightboxContainer.append(this.options.$lightboxClosingAlert);

	this.options.$lightboxClosingAlertButtons = this.options.$lightboxClosingAlert.find('button');

	this._addEventListenersToClosingAlertButtons();
	this._removeTabIndexFromClosingAlertButtons();

};

Module.prototype._addEventListenersToClosingAlertButtons = function() {
	$(this.options.$lightboxClosingAlertButtons[0]).on(events.click, this._onCancelClosingClick.bind(this));
	$(this.options.$lightboxClosingAlertButtons[1]).on(events.click, this._onConfirmClosingClick.bind(this));
};

Module.prototype._removeTabIndexFromClosingAlertButtons = function() {
	this.options.$lightboxClosingAlertButtons.attr({
		'tabindex': '-1'
	});
};

Module.prototype._addTabIndexToClosingAlertButtons = function() {
	this.options.$lightboxClosingAlertButtons.removeAttr('tabindex');
};

/**
 * Confirm closing alert modal handler
 * @private
 * @method
 */
Module.prototype._onConfirmClosingClick = function() {

	switch (this.options.closeAlert.confirmCloseClick.action) {
		case 'redirect':
			window.location = this.options.closeAlert.confirmCloseClick.redirectUrl;
			break;
		default:
			this.options.askBeforeClosing = false;
			this._closeLightbox();
			this.$closeButton.removeClass(this.options.stateClasses.isDisabled);
			break;
	}

};

/**
 * Cancel alert modal handler
 * @private
 * @method
 */
Module.prototype._onCancelClosingClick = function() {
	this._hideClosingAlert();
};

/**
 * Destroys closing alert rendered template
 * @private
 * @method
 */
Module.prototype._destroyClosingAlertTemplate = function() {
	// Remove listeners
	this.options.$lightboxClosingAlertButtons.off(events.click);

	// Remove object from DOM
	this.options.$lightboxClosingAlert.remove();
	this.options.$lightboxClosingAlert = undefined;
	this.$element.removeClass(this.options.stateClasses.hasAlert);
};

/**
 * Display closing alert modal
 * @private
 * @method
 */
Module.prototype._showClosingAlert = function() {

	this._addTabIndexToClosingAlertButtons();
	this._hideLightboxContentContainer();

	LockTabFocus.destroy();
	LockTabFocus.setFocusOnElement(this.options.$lightboxClosingAlert);

	this.$element.addClass(this.options.stateClasses.hasAlert);

	// disable the main close button, since we now have a modal overlay
	// and expect the user to interact with it instead of close button
	// on lightbox directly.
	this.$closeButton.addClass(this.options.stateClasses.isDisabled);
};

/**
 * Hide closing alert modal
 * @private
 */
Module.prototype._hideClosingAlert = function() {

	this._removeTabIndexFromClosingAlertButtons();
	this._showLightboxContentContainer();

	LockTabFocus.destroy();
	LockTabFocus.setFocusOnElement(this.$element);

	this.$element.removeClass(this.options.stateClasses.hasAlert);
	this.$closeButton.removeClass(this.options.stateClasses.isDisabled);
};

Module.prototype._hideLightboxContentContainer = function() {
	this.options.$lightboxContentContainer.attr({
		'hidden': 'hidden'
	});
};

Module.prototype._showLightboxContentContainer = function() {
	this.options.$lightboxContentContainer.removeAttr('hidden');
};

/**
 * Function triggered by event that should open lightbox
 * @param options object passed to function. Based on parameters it can call _attachDocumentNode of _attachExternalContent
 * @private
 */
Module.prototype._attachContentToLightbox = function(config) {

	if (typeof config === 'object') {

		Object.keys(config).forEach(function(key) {
			this.options[key] = config[key];
		}.bind(this));
	}

	if (config.elementToFocus) {
		this.elementToFocus = config.elementToFocus;
	}

	this._checkForClosingAlert();

	// Only do a cleanup if we have to
	if (!this.isInitialLoad) {
		this._detachCurrentContent();
	}

	this._addStateClasses(config);

	switch (config.type) {
		case 'node':
			this._attachDocumentNode($(config.source));
			break;
		case 'iframe':
			this._attachIframe(config.source);
			break;
		case 'url':
			this._attachExternalContent(config.source);
			break;
		default:
	}

	this.isInitialLoad = false;
};

/**
 * Adds state classes to lightbox or its elements
 * @param lightboxOptions
 * @private
 */
Module.prototype._addStateClasses = function(lightboxOptions) {

	var options;

	if (lightboxOptions.options) {
		options = lightboxOptions.options;
	} else {
		options = lightboxOptions;
	}

	if (options.centerContentHorizontally) {
		this.options.$lightboxContainer.addClass(this.options.stateClasses.isContentCenteredHorizontally);
	}

	if (options.centerContentVertically) {
		this.options.$lightboxContainer.addClass(this.options.stateClasses.isContentCenteredVertically);
	}

};

/**
 * Unbind events, remove data, custom teardown
 * @method
 * @public
 */
Module.prototype.destroy = function() {
	// Unbind events, remove data
	SuperClass.prototype.destroy.apply(this);

	// Custom teardown (removing added DOM elements etc.)
	// If there is no need for a custom teardown, this method can be removed
	this._detachCurrentContent();
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events,
	event: {
		key: 'lightbox.open.estatico'
	}
};


/***/ }),
/* 102 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 103 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(104);


/***/ }),
/* 104 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M32 - Tooltip
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(105);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	ToolTip;

var name = 'tooltip',
	events = {
		closeTooltip: 'closeTooltip.estatico.' + name
	},
	defaults = {
		domSelectors: {
			layoutCenter: '.layout_center:first-child',
			tableModule: '[data-init="tables"]',
			parentTable: '[data-tables="table"]',
			button: '[data-' + name + '="button"]',
			closeButton: '[data-' + name + '="close-button"]',
			text: '[data-' + name + '="text"]',
			arrow: '[data-' + name + '="arrow"]',
			window: '[data-' + name + '="window"]'
		},
		stateClasses: {
			isOpen: 'is_open',
			isHidden: 'is_hidden',
			varBelow: 'var_below'
		},
		keys: {
			close: [27]
		}
	},
	data = {
		offsetSmall: 24,
		tablePadding: 0
	};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	__webpack_require__.e/* require.ensure */(4).then((function(require) {
			ToolTip = __webpack_require__(94);

			Module.prototype = $.extend(true, Module.prototype, ToolTip.prototype);

			ToolTip.prototype.init.apply(this);

		}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	initEvents: ['ready', 'ajaxload'],
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 105 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 106 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.10 - Form Validation
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	ajaxRequest = __webpack_require__(16);

var name = 'formValidation',
	log = estatico.helpers.log(name),
	$document = $(document),
	moduleEvents = [],
	confirmationStatus = {
		NEEDS_CONFIRMATION: 'NEEDS_CONFIRMATION'
	},
	events = {
		formValidationSubmit: 'formValidationSubmit.estatico'
	},
	defaults = {
		domSelectors: {
			header: '[data-init="header"]',
			confirmationInput: '[data-formValidation="confirmation"]',
			conditionalBlockToggleElement: 'select[data-formconditional-controls-target], [data-formconditional-controls-target] input, input[data-formconditional-controls-target]'
		},
		stateClasses: {
			hasError: 'has_error'
		},
		parsleyOptions: {},
		triggerChanges: false,
		paymentOptions: false
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		moduleEvents: moduleEvents,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Init by lazy loading the dependencies first and then do the actual init
 */
Module.prototype.init = function() {

	// load Parsley async and give webpack ability to handle it correctly
	// jquery is in here for the webshop
	__webpack_require__.e/* require.ensure */(3/* duplicate */).then((function(require) {

		__webpack_require__(3);

		// call real init
		this._init();

	}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);

};

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype._init = function() {

	var errorsListId = this.$element.attr('id') + '_errors_list';

	// Create event name for clicking error link
	this.localEvents = {
		errorLinkClick: 'click.estatico.' + this.uuid
	};

	// Create array for fields with errors
	this._resetFieldsWithErrors();

	// Error container
	this.globalErrorContainer = this.$element.find('#' + this.$element.attr('id') + '_error_message');

	// Cache confirmation input
	this.$confirmationInputs = this.$element.find(this.options.domSelectors.confirmationInput);

	// Cache Submit Button
	this.$submitButton = this.$element.find('button[type="submit"]');

	// Is there a server side rendered error message list present?
	this.errorsList = this.$element.find('#' + errorsListId)[0];

	// Append error list container if we do not have one rendered by the backend yet
	if (this.errorsList === undefined) {
		this.errorsList = document.createElement('ul');
		this.errorsList.id = errorsListId;
		this.globalErrorContainer.find('#' + this.$element.attr('id') + '_error_message_text').append(this.errorsList);
	}

	// Create Parsley object with proper options
	this.$parsley = this.$element.parsley(this.options.parsleyOptions);

	// Initialize Parsley listeners
	this._initializeParsleyListeners();

	// Initialize Trigger Changes if option is set -> defaults to false -> Needed within the Webshop
	// We inform the shop about the validation state of each parsley validated field
	if (this.options.triggerChanges) {
		this._initializeChangeTriggers();
	}

	// Listen to changes if fields are added subtracted
	$document.on('formchange' + '.' + this.uuid, this._updateForm.bind(this));

};

Module.prototype._resetFieldsWithErrors = function() {
	this.fieldsWithErrors = {};
};

/**
 *
 * @param event {jQuery} event object
 * @param data contains property id to update only one form
 * @private
 */
Module.prototype._updateForm = function(event, data) {

	if (data.id === this.$element.attr('id')) {

		// Set Error List back to empty state
		this._resetFieldsWithErrors();

		// Clear Parsley listeners if they exists
		if (this.$parsley) {
			this._clearParsleyListeners();
		}

		// Update current form if sent id proper
		this.$parsley = this.$element.parsley(this.options.parsleyOptions);

		// Add listeners to reinitialized form
		this._initializeParsleyListeners();

		// Trigger listeners - only way I managed to do that
		this.$parsley.isValid();
	}
};

/**
 * Initialize Parsleyjs listeners on various events
 * @private
 */
Module.prototype._initializeParsleyListeners = function() {
	// Register listener on validation error to update field aria attributes
	this.$parsley.on('field:error', this._fieldValidationError.bind(this));

	// Register listener on validation success to update field aria attributes
	this.$parsley.on('field:success', this._fieldValidationSuccess.bind(this));

	// Register listener on form validation error
	this.$parsley.on('form:error', this._formValidationError.bind(this));

	// Register listener on form validation success
	this.$parsley.on('form:success', this._formValidationSuccess.bind(this));

	// Register listener on form submit to prevent additional unintended submit
	this.$parsley.on('form:submit', this._handleFormSubmit.bind(this));

	// Register listener for conditional form block toggle buttons
	this.$element.on('change.' + this.uuid, this.options.domSelectors.conditionalBlockToggleElement, function() {
		this._resetFieldsWithErrors();
	}.bind(this));

	// Set validation messages, set in data.js file config if present
	if (this.options.errorMessages) {
		window.Parsley.addMessages(this.options.errorMessagesLocale, this.options.errorMessages);
	}

	// Set Parsley locale to override automatic settings
	if (this.options.errorMessagesLocale) {
		window.Parsley.setLocale(this.options.errorMessagesLocale);
	}
};

/**
 * Removes old reference to Parsley object, to clear all form listeners before creating new one
 */
Module.prototype._clearParsleyListeners = function() {

	// Register listener on validation error to update field aria attributes
	this.$parsley.off('field:error');

	// Register listener on validation success to update field aria attributes
	this.$parsley.off('field:success');

	// Register listener on form validation error
	this.$parsley.off('form:error');

	// Register listener on form validation success
	this.$parsley.off('form:success');

	// Destroy $parsley reference
	delete this.$parsley;
};

/**
 * Initialize Change Triggers to inform other plugins or frameworks
 * about changes for form fields without parsley validation
 * @private
 */
Module.prototype._initializeChangeTriggers = function() {

	var $domElements;

	moduleEvents.push('changed', 'changedAndValidated', 'changedAndInvalidated');
	this.options.domSelectors.plainSelectElements = 'select:not([data-parsley-trigger]):not([data-multiselect])';
	this.options.domSelectors.plainInputElements = 'input:not([data-parsley-trigger]):not([data-datepicker])';
	this.options.domSelectors.plainTextareaElements = 'textarea:not([data-parsley-trigger])';

	$domElements = this.options.domSelectors.plainSelectElements + ', ' + this.options.domSelectors.plainInputElements + ', ' + this.options.domSelectors.plainTextareaElements;

	// this.$element.on('select.' + this.uuid + ' textInput.' + this.uuid + ' input.' + this.uuid, $domElements, function(event) {
	this.$element.on('keyup.' + this.uuid, $domElements, function(event) {
		var $element = event.target;

		this.trigger('changed', $element);

	}.bind(this));

};

/**
 * Function triggered when validation for field fails.
 * @param parsleyField
 * @private
 */
Module.prototype._formValidationError = function() {
	this.globalErrorContainer.addClass(this.options.stateClasses.hasError);
	this._constructFieldsWithErrorsList();

	// Scroll to beginning of the list to see the errors
	this._scrollToTopOfForm();
};

/**
 * Function triggered when validation for field succeed.
 * @param parsleyField
 * @private
 */
Module.prototype._formValidationSuccess = function() {
	this.globalErrorContainer.removeClass(this.options.stateClasses.hasError);
};

Module.prototype._handleFormSubmit = function() {
	var shouldPreventSubmitEvent = false;

	this.$confirmationInputs.each(function(i, el) {
		var $confirmationInput = $(el);

		if ($confirmationInput.val() === confirmationStatus.NEEDS_CONFIRMATION) {
			this.$element.trigger(events.formValidationSubmit);
			shouldPreventSubmitEvent = true;
		}
	}.bind(this));

	// Handle the payment, if the payment options are defined
	if (this.options.paymentOptions) {
		shouldPreventSubmitEvent = true;
		this._handlePayment();
	}

	if (shouldPreventSubmitEvent) {
		return false;
	}

	this.$submitButton.attr({
		'disabled': 'disabled'
	});
};

/**
 * Autoscroll to top of the form including sticky header height
 * @private
 */
Module.prototype._scrollToTopOfForm = function() {
	$document.scrollTop(this.$element.offset().top - this._getHeightOfTheHeader() - 40);
};

/**
 * Function triggered when validation for field fails.
 * @param parsleyField
 * @private
 */
Module.prototype._fieldValidationError = function(parsleyField) {
	var currentValue = parsleyField.$element.val();

	// Pass on the invalidated form field reference to potential listeners
	if (this.options.triggerChanges) {
		log('Changed and invalidated form field "' + $(parsleyField.$element).attr('name') + '" value to: "' + currentValue + '"');
		this.trigger('changedAndInvalidated', parsleyField.$element);
	}

	parsleyField.$element
		.attr('aria-invalid', true)
		.attr('aria-describedby', parsleyField.domOptions.errorsContainer.replace('#', ''))
		.data('previousValue', currentValue);

	this._removeIfInputHasBackendInjectedErrors(parsleyField);
	this.fieldsWithErrors[parsleyField.$element.data('parsleyId')] = parsleyField;
};

/**
 * Returns height of the header depending
 * @returns {*|jQuery}
 * @private
 */
Module.prototype._getHeightOfTheHeader = function() {
	var ret = $(this.options.domSelectors.header).outerHeight();

	return ret;
};

/**
 * When validating some input it will check for backend injected error in the
 * error container for that particular field and remove it.
 * @private
 */
Module.prototype._removeIfInputHasBackendInjectedErrors = function(parsleyField) {
	var $errorContainer = $($(parsleyField.$element).attr('data-parsley-errors-container')),
		$errorsList,
		id,
		$errorLists = $errorContainer.find('ul');

	$errorLists.each(function(index, errorsList) {
		$errorsList = $(errorsList);
		id = $errorsList.attr('id');
		if (typeof id === typeof undefined || id === false) {
			$errorsList.remove();
		}
	});
};

/**
 * Function triggered when validation for field succeeds.
 * @param parsleyField
 * @private
 */
Module.prototype._fieldValidationSuccess = function(parsleyField) {

	var labelName = '#' + parsleyField.$element.attr('id') + '_label',
		previousValue = parsleyField.$element.data('previousValue'),
		currentValue = parsleyField.$element.val();

	// Pass on the invalidated form field reference to potential listeners
	if (this.options.triggerChanges) {

		// Trigger event only if previously validated value is different or does not exist
		// which happens when validating successfully for the first time
		if (!previousValue || previousValue !== currentValue) {

			// setTimeout,0 is used to make sure that all changes of syncFormField
			// fields are updated before the event is triggered.
			setTimeout(function() {
				this.trigger('changedAndValidated', parsleyField.$element);
			}.bind(this), 0);
		}
	}

	this._removeIfInputHasBackendInjectedErrors(parsleyField);
	parsleyField.$element.removeAttr('aria-invalid').attr('aria-describedby', labelName).data('previousValue', currentValue);
	delete this.fieldsWithErrors[parsleyField.$element.data('parsleyId')];
};

/**
 * Creates clickable list or errors for the form object
 * @private
 */
Module.prototype._constructFieldsWithErrorsList = function() {
	var list = this.errorsList,
		listItem,
		labelName,
		inputType,
		errorLink;

	// remove existing errors
	$(list).empty();

	$.each(this.fieldsWithErrors, function(index, parsleyField) {

		var $el = parsleyField.$element,
			errorElementInnerText,
			errorText = '';

		inputType = $el.attr('type');

		if (inputType === 'radio' || inputType === 'checkbox') {
			labelName = '#' + $el.attr('data-radiocheckbox-legend');
		} else {
			labelName = '#' + $el.attr('id') + '_label';
		}

		if ($el.attr('data-parsley-custom-group-validation')) {
			labelName = '#' + $el.attr('id') + '_legend';
		}

		errorElementInnerText = this.$element.find(labelName);

		if (errorElementInnerText.length > 0) {
			for (var i = 0, l = errorElementInnerText[0].childNodes.length; i < l; i++) {
				var $childNode = $(errorElementInnerText[0].childNodes[i]);

				if (!$childNode.hasClass('visuallyhidden')) {
					errorText += $childNode.text().trim();
				}
			}
		}

		listItem = document.createElement('li');
		errorLink = document.createElement('a');
		errorLink.innerText = errorText;
		errorLink.href = '#' + $el.attr('id');
		listItem.appendChild(errorLink);
		$(listItem).on(this.localEvents.errorLinkClick, this._onErrorLinkClick.bind(this));
		list.appendChild(listItem);

	}.bind(this));
};

/**
 * Event handler for error link click, that redirects user to input field position on the page
 * @param event
 * @private
 */
Module.prototype._onErrorLinkClick = function(event) {
	var $el = $(event.currentTarget),
		scrollTop = this.$element.find($el.find('a').attr('href')).offset().top - this._getHeightOfTheHeader() - 40;

	event.preventDefault();
	$document.scrollTop(scrollTop);
};

/**
 * Function triggered when payment options are given
 * @private
 */
Module.prototype._handlePayment = function() {
	var formData = this.$element.serializeArray(),
		params = {};

	// pass form data tru the payment process
	formData.forEach(function(data) {
		params[data.name] = data.value;
	}.bind(this));

	params.lang = localStorage.getItem('sbblang') || 'de';

	// for testing purposes we need the fallback for the old datatrans integration
	if (this.options.paymentOptions.provider === 'datatrans') {
		if (this.options.paymentOptions.version === '2') {
			params.refno = this.$element.attr('data-refno');
			params.amount = this.$element.attr('data-amount');

			ajaxRequest(this.options.paymentOptions.apiUrl, false, params)
				.then(function(response) {
					window.Datatrans.startPayment({
						transactionId: response.transactionId
					});
				}.bind(this))
				.catch(function() {
					return;
				});

		} else {
			window.Datatrans.startPayment({
				form: '#' + this.$element.attr('id'),
				params: params
			});
		}
	}
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events,
	confirmationStatus: confirmationStatus
};


/***/ }),
/* 107 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(108);


/***/ }),
/* 108 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.6 - Select
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(109);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'select',
	events = {},
	defaults = {
		attributes: {
			conditionalControlsTarget: 'data-formconditional-controls-target',
			conditionalTargetHiddenAttr: 'data-formconditional-target-hidden'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	var targetAttr = this.$element.attr(this.options.attributes.conditionalControlsTarget);

	// immediately return if there is no target attribute
	if (!targetAttr) {
		return;
	}

	// cache the target element
	this.$targetElement = $('#' + targetAttr);

	// add change event listener to the select
	this.$element.on('change.' + this.uuid, this.$element, this.selectionChanged.bind(this));

	// initially check if we need to hide a target
	this.selectionChanged();
};

/**
 * Handle selection change
 */
Module.prototype.selectionChanged = function() {

	// get the hidden attribute of the currently selected option
	var targetHiddenAttrValue = $('option:selected', this.$element).attr(this.options.attributes.conditionalTargetHiddenAttr);

	// run the transition helper method
	if (!targetHiddenAttrValue) {
		this.$targetElement.show();
	} else {
		this.$targetElement.hide();
	}
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 109 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 110 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(111);


/***/ }),
/* 111 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 112 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(106);


/***/ }),
/* 113 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(114);


/***/ }),
/* 114 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 115 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(116);


/***/ }),
/* 116 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M18.1 - Media Slider
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(117);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	SliderModule = __webpack_require__(55);

var name = 'mediaslider',
	events = {
		showCurrentSlide: 'showCurrentSlide.estatico.' + name
	},
	defaults = {
		domSelectors: {
		},
		sliderSettings: {
			itemsQuantity: {
				tiny: 1,
				mobilePortrait: 1,
				mobileLandscape: 1,
				tabletPortrait: 1,
				tabletLandscape: 1,
				desktop: 1,
				desktopLarge: 1,
				desktopLargePlus: 1,
				desktopMaxWidth: 1,
				desktop4k: 1,
				desktop5k: 1
			}
		}
	},
	data = {
	};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype, SliderModule.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	SliderModule.prototype.init.apply(this);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 117 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 118 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(135);


/***/ }),
/* 119 */,
/* 120 */,
/* 121 */,
/* 122 */,
/* 123 */,
/* 124 */,
/* 125 */,
/* 126 */,
/* 127 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(128);


/***/ }),
/* 128 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 129 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(130);


/***/ }),
/* 130 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 131 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(132);


/***/ }),
/* 132 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 133 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(134);


/***/ }),
/* 134 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 135 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 136 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(137);


/***/ }),
/* 137 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.3 - Datepicker
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(138);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	$document = $(document);

var name = 'datepicker',
	events = {
		updateAllowedDates: 'datepicker.updateAllowedDates',
		removeAllowedTimes: 'datepicker.removeAllowedTimes'
	},
	defaults = {
		domSelectors: {
			picker: '[data-' + name + '="picker"]',
			fieldWrapper: '[data-' + name + '="input_wrapper"]',
			field: '[data-' + name + '="input"]',
			monthPrev: '[data-' + name + '="month_prev"]',
			monthNext: '[data-' + name + '="month_next"]',
			monthCurrent: '[data-' + name + '="month_current"]',
			yearPrev: '[data-' + name + '="year_prev"]',
			yearNext: '[data-' + name + '="year_next"]',
			yearCurrent: '[data-' + name + '="year_current"]',
			cal: '[data-' + name + '="cal"]',
			arrow: '[data-' + name + '="arrow"]',
			keyboardFeatures: '[data-' + name + '="keyboardfeatures"]',
			notification: '[data-' + name + '="notification"]',
			gridNotification: '[data-' + name + '="gridNotification"]',
			heading: '[data-' + name + '="heading"]',
			allowedTimesWrapper: '[data-' + name + '="allowed_times_wrapper"]',
			allowedTimeItem: '[data-' + name + '="allowed_time_item"]'
		},
		stateClasses: {
			inputIsActive: 'input_is_active',
			isVisible: 'is_visible'
		},
		showInputCalendarIcon: false,
		showInputArrowIcons: false,
		i18n: {
			months: ['Januar', 'Februar', 'März', 'April', 'Mai', 'Juni', 'Juli', 'August', 'September', 'Oktober', 'November', 'Dezember'],
			days: ['Montag', 'Dienstag', 'Mittwoch', 'Donnerstag', 'Freitag', 'Samstag','Sonntag'],
			daysShort: ['Mo', 'Di', 'Mi', 'Do', 'Fr', 'Sa','So'],
			prevDay: 'Zum vorherigen Tag wechseln',
			nextDay: 'Zum nächsten Tag wechseln',
			prevMonth: 'Zum vorherigen Monat wechseln',
			nextMonth: 'Zum nächsten Monat wechseln',
			prevYear: 'Zum vorherigen Jahr wechseln',
			nextYear: 'Zum nächsten Jahr wechseln',
			showCalendar: 'Kalender anzeigen',
			errorDateNotAllowed: 'Dieses Datum ist nicht erlaubt',
			errorDateNotValid: 'Das ist kein gültiges Datum',
			keyboardFeatureDescription: 'Sie können folgende Tasten verwenden um den Kalender zu bedienen:',
			keyboardFeatures: [
				'Drücken Sie die Leertaste um einen Tag im Kalender auszuwählen',
				'Drücken Sie die rechte Pfeiltaste um zum nächsten Tag zu springen',
				'Drücken Sie die linke Pfeiltaste um zum vorherigen Tag zu springen',
				'Drücken Sie die Pfeiltaste nach unten um eine Woche vor zu springen',
				'Drücken Sie die Pfeiltaste nach oben um eine Woche zurück zu springen',
				'Drücken Sie die Taste nächste Seite, um zum nächsten Monat zu wechseln',
				'Drücken Sie die Taste vorherige Seite, um zum vorherigen Monat zu wechseln',
				'Benutzen Sie die Tabulatortaste um zwischen dem Kalender und den Schaltern für die nächsten und vorherigen Monate und Jahre zu wechseln',
				'Drücken Sie die Taste Escape um den Kalender zu schliessen'
			],
			currentlySelected: 'derzeit ausgewählt',
			displayChangedTo: 'Kalender zeigt',
			dateChangedTo: 'Datum geändert zu',
			activeDayInGrid: 'gewählt',
			allowedTimes: 'Freie Termine für den'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	__webpack_require__.e/* require.ensure */(5).then((function(require) {

		__webpack_require__(3);

		this.Datepicker = __webpack_require__(124);

		this.templates = {
			picker: __webpack_require__(120),
			button: __webpack_require__(121),
			buttonPrevDay: __webpack_require__(122),
			buttonNextDay: __webpack_require__(123)
		};

		setTimeout(function() {
			this._init();
		}.bind(this), 1);
	}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);
};

Module.prototype._init = function() {
	var templateData = {
			id: this.uuid,
			i18n: this.options.i18n
		},
		templates = {
			picker: this.templates.picker(templateData),
			button: this.templates.button(templateData),
			buttonPrevDay: this.templates.buttonPrevDay(templateData),
			buttonNextDay: this.templates.buttonNextDay(templateData)
		};

	this.$fieldWrapper = this.$element.find(this.options.domSelectors.fieldWrapper);
	this.$notification = this.$element.find(this.options.domSelectors.notification);
	this.$field = this.$element.find(this.options.domSelectors.field);
	this.$renderedPicker = this.$element.find(this.options.domSelectors.picker);

	// render picker and buttons only once
	if (this.$renderedPicker.length === 1) {
		return;
	}

	this.$datepicker = $(templates.picker).appendTo(this.$element);
	this.$button = $(templates.button);
	this.$buttonPrevDay = $(templates.buttonPrevDay);
	this.$buttonNextDay = $(templates.buttonNextDay);

	if (this.options.showInputCalendarIcon) {
		this.$button.prependTo(this.$fieldWrapper);
	}

	if (this.options.showInputArrowIcons) {
		this.$buttonPrevDay.prependTo(this.$fieldWrapper);
		this.$buttonNextDay.appendTo(this.$fieldWrapper);
	}

	// decide wether this is a single/independent datepicker or if it is a datepicker depending on another
	if (this.options.hasSlavePicker || this.options.hasMasterPicker) { // this is a master-date-picker or a slave date-picker

		// set the slave or master picker to the options object.
		if (this.options.hasSlavePicker) { // this is the master picker. open the slave as well.
			this.$slavePicker = $('#' + this.options.hasSlavePicker);

			setTimeout(function() {
				this.options.syncedPicker = this.$slavePicker.data('datepickerInstance').datepicker;
			}.bind(this), 1);
		} else if (this.options.hasMasterPicker) { // this is the slave picker. open the master as well.
			this.$masterPicker = $('#' + this.options.hasMasterPicker);

			setTimeout(function() {
				this.options.syncedPicker = this.$masterPicker.data('datepickerInstance').datepicker;
			}.bind(this), 1);
		}
	}

	// create object that contains all elements of input field to pass to the datepicker-init method
	var inputElements = {
		'inputPreviousDay': this.$buttonPrevDay.get(0),
		'inputNextDay': this.$buttonNextDay.get(0),
		'inputWrapper': this.$fieldWrapper.get(0),
		'input': this.$field.get(0),
		'notification': this.$notification.get(0),
		'calendarButton': this.$button.get(0)
	};

	// create instance of the Datepicker
	this.datepicker = new this.Datepicker(this.$element, this.$datepicker.get(0), this.$field.get(0), this.uuid, this.options, inputElements);

	// handle click on calendar button
	this.$button.on('click.' + name, function(event) {
		event.preventDefault();

		// open datepicker
		this.datepicker.showDlg();
	}.bind(this));

	if (this.options.allowedTimes) {
		$document.on(
			events.updateAllowedDates + '.' + this.uuid,
			function(_, allowedDates) {
				// convert the passed string of dates to an array of date objects
				var allowedDatesArray = this.datepicker.convertStringOfDatesToArrayOfDates(allowedDates);

				// sort array of dates
				allowedDatesArray = this.datepicker.sortArrayOfDatesAscending(allowedDatesArray);

				this.datepicker.allowDates(allowedDatesArray);
			}
		.bind(this));

		$document.on(
			events.removeAllowedTimes + '.' + this.uuid,
			function() {
				this.datepicker.removeAllowedTimes();
				this.$field.val('');
				this.datepicker.resetLocallyStoredDate();
			}
		.bind(this));
	}
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	initEvents: ['ready', 'ajaxload'],
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 138 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 139 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * WM22 - Webshop Checkbox Radio Panel
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(140);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	cssTransitionEnd = __webpack_require__(4);

var name = 'checkboxradiopanel',
	events = {
		checkboxRadioPanelClicked: 'checkboxRadioPanel.clicked'
	},
	defaults = {
		domSelectors: {
			input: '[data-' + name + '="input"]',
			item: '[data-' + name + '="item"]',
			nestedCheckbox: '[data-radiocheckbox="checkbox_item"]',
			deselectButton: '[data-' + name + '="deselect_all"]',
			selectAllButton: '[data-' + name + '="select_all"]',
			selectSpecific: '[data-' + name + '="select_specific"]'
		},
		stateClasses: {
			isChecked: 'is_checked',
			isHidden: 'is_hidden'
		},
		toggleNestedCheckboxes: false,
		notifyChange: false,
		multistepForm: false
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	var events = {
		change: 'change.' + this.uuid,
		click: 'click.' + this.uuid
	};

	// cache radio/checkbox items
	this.$items = this.$element.find(this.options.domSelectors.item);

	// cache radio/checkbox inputs
	this.$inputs = this.$element.find(this.options.domSelectors.input);

	// cache deselect all button
	this.$deselectButton = this.$element.find(this.options.domSelectors.deselectButton);

	// cache select all button
	this.$selectButton = this.$element.find(this.options.domSelectors.selectAllButton);

	// cache specific selection button
	this.$selectSpecificButton = this.$element.find(this.options.domSelectors.selectSpecific);

	// bind change event to inputs
	this.$inputs.on(events.change, this._handleInputStateChange.bind(this));

	// bind click events to buttons
	this.$deselectButton.on(events.click, this._handleDeselectAll.bind(this));
	this.$selectButton.on(events.click, this._handleSelectAll.bind(this));
	this.$selectSpecificButton.on(events.click, this._handleSelectSpecific.bind(this));

	// trigger helper to manage selection button visability, but only if on of the complementary buttons are found. these should only be used, if type is checkbox. They shall not be used for radios.
	if (this.$deselectButton.length) {
		this._handleSelectionButtonVisability();
	}

	if (this.$selectSpecificButton.length) {
		this._handleSelectSpecificButtonVisability();
	}
};

/**
 * Event listener handler for input state change
 * @param event
 * @private
 */
Module.prototype._handleInputStateChange = function(event) {
	var $el = $(event.currentTarget),
		$item = $el.closest(this.options.domSelectors.item),

		// Select all panels with the same name
		$siblings = $item.siblings().find('input[name="' + $el.attr('name') + '"]').closest(this.options.domSelectors.item);

	if ($el.prop('checked')) {
		this._markAsChecked($item);
	} else {
		this._markAsUnchecked($item);
	}

	// remove classes from previously selected input, as jquery event is not firing
	// only with radios
	if ($el.attr('type') === 'radio') {
		this._markAsUnchecked($siblings);
	}

	// handle visability of buttons select/deselect all, but only if the deselect element is present in the dom. This means, that both select and deselect buttons are present.
	if (this.$deselectButton.length) {
		this._handleSelectionButtonVisability();
	}

	// handle visability of button to select all checkboxes of specific kind, but only if the selectSpecific element is present in the dom.
	if (this.$selectSpecificButton.length) {
		this._handleSelectSpecificButtonVisability();
	}

	// set the value of the checked checkbox to a data-attribute of another element
	if (this.options.notifyChange) {
		if (this.options.notifyChange.type === 'datatrans') {
			this.$element.parents(this.options.notifyChange.target).attr('data-paymentmethod', $el.val());
		}
	}

	// to show the correct next step in the multistep form, get step target id from the attribute and trigger an event with it
	if (this.options.multistepForm) {
		this.$element.trigger(events.checkboxRadioPanelClicked, $el);
	}
};

/**
 * Method to deselect all checkboxes
 * @method
 * @private
 */
Module.prototype._handleDeselectAll = function() {

	this._markAsUnchecked(this.$items);

	this._handleSelectionButtonVisability();

	this._handleSelectSpecificButtonVisability();
};

/**
 * Method to select all checkboxes
 * @method
 * @private
 */
Module.prototype._handleSelectAll = function() {

	this._markAsChecked(this.$items);

	this._handleSelectionButtonVisability();

	this._handleSelectSpecificButtonVisability();
};

/**
 * Method to select all checkboxes with a specific key
 * @method
 * @private
 */
Module.prototype._handleSelectSpecific = function(event) {
	var keyToSelect = $(event.currentTarget).attr('data-checkboxradiopanel-select-key'),
		thisKey,
		$item;

	// deselect all
	this._handleDeselectAll();

	// check specific checkboxes
	$.each(this.$inputs, function(key, value) {
		thisKey = $(value).attr('data-checkboxradiopanel-type-key');
		if (thisKey === keyToSelect) {
			$item = $(value).closest(this.options.domSelectors.item);
			$(value).prop('checked', true);
			this._markAsChecked($item);
		}
	}.bind(this));

	this._handleSelectionButtonVisability();

	this._handleSelectSpecificButtonVisability();
};

/**
 * Mark selected panel state as checked adding appropriate class
 * @param $item
 * @private
 */
Module.prototype._markAsChecked = function($item) {

	$item.off(cssTransitionEnd);
	$item.removeClass(this.options.stateClasses.isHidden);
	requestAnimationFrame(function() {
		$item.addClass(this.options.stateClasses.isChecked);
	}.bind(this));

	$item.find(this.options.domSelectors.input).prop('checked', true);
};

/**
 * Mark passed panel reference as unselected removing appropriate class
 * @param $item
 * @private
 */
Module.prototype._markAsUnchecked = function($item) {

	if (cssTransitionEnd) {
		$item.one(cssTransitionEnd, function() {
			$item.addClass(this.options.stateClasses.isHidden);
		}.bind(this));
	} else {
		$item.addClass(this.options.stateClasses.isHidden);
	}

	$item.removeClass(this.options.stateClasses.isChecked);
	$item.find(this.options.domSelectors.input).prop('checked', false);

	if (this.options.toggleNestedCheckboxes) {
		$item.find(this.options.domSelectors.nestedCheckbox).prop('checked', false);
	}
};

/**
 * Show/hide select all / deselect all button. If at least one checkbox is selected, show deselect button. If no checkbox is selected, show select all.
 * @method
 * @private
 */
Module.prototype._handleSelectionButtonVisability = function() {
	var iterator = 0,
		itemsCount = this.$inputs.length,
		$item,
		checkedCount = 0;

	// count how many checkboxes are checked
	for (iterator; iterator < itemsCount; iterator++) {
		$item = $(this.$inputs[iterator]);
		if ($item.prop('checked')) {
			checkedCount++;
		}
	}

	// based on checked class, add button hidden class to select/deselect buttons
	if (checkedCount === 0) { // nothing is selected
		this.$selectButton
			.removeClass(this.options.stateClasses.isHidden)
			.attr('aria-hidden', 'false');
		this.$deselectButton
			.addClass(this.options.stateClasses.isHidden)
			.attr('aria-hidden', 'true');
	} else {
		this.$selectButton
			.addClass(this.options.stateClasses.isHidden)
			.attr('aria-hidden', 'true');
		this.$deselectButton
			.removeClass(this.options.stateClasses.isHidden)
			.attr('aria-hidden', 'false');
	}
};

/**
 * Show/hide button to select specific items. If all checkboxes of the defined kind are selected, hide the button, otherwise show it.
 * @method
 * @private
 */
Module.prototype._handleSelectSpecificButtonVisability = function() {
	var keyToSelect = this.$selectSpecificButton.attr('data-checkboxradiopanel-select-key'),
		thisKey,
		itemsCount = 0,
		selectedCount = 0,
		forceShowButton = false;

	// count total items with specific key, and how many of them are checked
	$.each(this.$inputs, function(key, value) {
		thisKey = $(value).attr('data-checkboxradiopanel-type-key');
		if (thisKey === keyToSelect) {
			itemsCount++;
			if ($(value).prop('checked')) {
				selectedCount++;
			}
		} else { // if at least one other checkbox is checked which has not that specific key, then show the select specific button again
			if ($(value).prop('checked')) {
				forceShowButton = true;
			}
		}
	});

	if (itemsCount !== selectedCount || forceShowButton) {
		this.$selectSpecificButton
			.removeClass(this.options.stateClasses.isHidden)
			.attr('aria-hidden', 'false');
	} else {
		this.$selectSpecificButton
			.addClass(this.options.stateClasses.isHidden)
			.attr('aria-hidden', 'true');
	}
};

/**
 * Unbind events, remove data, custom teardown
 * @method
 * @public
 */
Module.prototype.destroy = function() {
	// Unbind events, remove data
	SuperClass.prototype.destroy.apply(this);

	// Custom teardown (removing added DOM elements etc.)
	// If there is no need for a custom teardown, this method can be removed
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 140 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 141 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(142);


/***/ }),
/* 142 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 143 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M25 - Tabs
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(165);

var $ = __webpack_require__(0),
	gsap = __webpack_require__(11).gsap,
	globalEvents = __webpack_require__(5),
	SuperClass = __webpack_require__(1),
	RemoveTabIndex = __webpack_require__(19);

var name = 'tabs',
	$document = $(document),
	events = {
		click: 'click.estatico',
		tap: 'tap.estatico',
		keydown: 'keydown.estatico',
		switch: 'switch.estatico.' + name,
		tabSwitched: 'tabSwitched.estatico.' + name,
		tabsInitialized: 'initialized.estatico.' + name
	},
	defaults = {
		domSelectors: {
			tabItem: '[data-' + name + '="tab_list_item"]',
			tabPanel: '[data-' + name + '="tab_panel_item"]',
			tabsList: '[data-' + name + '="tabs_list"]',
			tabsWrapper: '[data-' + name + '="tabs_wrapper"]'
		},
		stateClasses: {
			isActive: 'is_active',
			isFocus: 'is_focused',
			isHidden: 'is_hidden',
			isScrolling: 'is_scrolling',
			isScrollingStart: 'is_scrolling_start',
			isScrollingEnd: 'is_scrolling_end',
			hasNoAnimation: 'has_no_animation',
			isNotClickable: 'is_not_clickable',
			isInitialized: 'is_initialized'
		},
		scrollBarPluginConfig: {
			axis: 'x',
			scrollInertia: 100,
			advanced: {
				autoExpandHorizontalScroll: true
			}
		},
		activeTabIndex: 0,
		$tabs: undefined,
		$panels: undefined,
		triggerEventAfterInit: false
	},
	isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints,
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	if (!isTouchDevice) {
		__webpack_require__.e/* require.ensure */(2/* duplicate */).then((function(require) {
				__webpack_require__(24);
				__webpack_require__(23)($);
				setTimeout(function() {
					__webpack_require__(19);
					this._init();
				}.bind(this), 1);
			}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);
	} else {
		setTimeout(function() {
			__webpack_require__(19);
			this._init();
		}.bind(this), 1);
	}

};

Module.prototype._init = function() {
	var tapClick = events.click + '.' + this.uuid + ' ' + events.tap + '.' + this.uuid,
		thisProxy = this,
		$hashTarget;

	this.$tabs = this.$element.find(this.options.domSelectors.tabItem);
	this.options.activeTabIndex = this.$tabs.filter('.' + this.options.stateClasses.isActive).index() || 0;
	this.$panels = this.$element.find(this.options.domSelectors.tabPanel);
	this.$tabsWrapper = this.$element.find(this.options.domSelectors.tabsWrapper);

	if (this.$tabs.length > 1) {
		this.$tabs.on(tapClick, this._tabItemClick.bind(this));
	}

	// we initially set aria-hidden to true. at the and of this method,
	// we set the correct aria-attributes on the currently active tab item / tab
	// panel. We do this to make sure that we have always correct aria-attributes
	// independent from the initial attributes in the markup.
	this.$panels.attr('aria-hidden', true);

	$(document).on(events.switch + '.' + this.uuid, function(event, tabIndex) {
		this._setTabIndex(tabIndex);

		// remove the focus of the previous user-action, because clicking the
		// back-button (mouse) does not remove the focus inside the document.
		this.$tabs.blur();
	}.bind(this));

	// trigger only on no-touch devices
	if (!isTouchDevice) {
		this.options.scrollBarPluginConfig.callbacks = {
			whileScrolling: function() {
				this.$tabsWrapper.removeClass(this.options.stateClasses.isScrollingStart + ' ' + this.options.stateClasses.isScrollingEnd);
			}.bind(this),

			onTotalScrollOffset: 0,
			onTotalScroll: function() {
				this.$tabsWrapper.addClass(this.options.stateClasses.isScrollingEnd);
			}.bind(this),

			onTotalScrollBackOffset: 0,
			onTotalScrollBack: function() {
				this.$tabsWrapper.addClass(this.options.stateClasses.isScrollingStart);
			}.bind(this),

			onOverflowX: function() {
				this.$tabsWrapper.addClass(this.options.stateClasses.isScrolling);
			}.bind(this),

			onOverflowXNone: function() {
				this.$tabsWrapper.removeClass(this.options.stateClasses.isScrolling);
			}.bind(this),

			onInit: function() {
				var $dragBar = thisProxy.$element.find('.mCSB_dragger_bar');

				if ($dragBar.length > 0) {
					if ($dragBar.position().left === 0) {
						thisProxy.$tabsWrapper.addClass(thisProxy.options.stateClasses.isScrollingStart);
					}
				}
			},

			onUpdate: function() {
				RemoveTabIndex.removeTabIndex(this);
			}
		};

		this.$element.find(this.options.domSelectors.tabsList).mCustomScrollbar(this.options.scrollBarPluginConfig);

	} else {
		this._addScrollGradientsOnTouchDevice();
	}

	// Listen to onresize
	$document.on(globalEvents.resize.key + '.' + this.uuid, this._onResizeHandler.bind(this));

	this._bindKeyboardHandlers();

	if (window.location.hash) {
		$hashTarget = this.$element.find('[data-tab-id="' + window.location.hash.substring(1) + '"]');

		// set tab if element exist
		if ($hashTarget.length) {
			this._setHash($hashTarget);
		}
	}

	if (!$hashTarget && this.options.activeTabIndex >= 0) {
		this._setTabIndex(this.options.activeTabIndex);
	}

	if (this.$tabs.length === 1) {
		this.$element.find(this.options.domSelectors.tabsList).removeAttr('role');

		this.$tabs
			.removeAttr('role aria-controls aria-selected aria-disabled')
			.addClass(this.options.stateClasses.isNotClickable);

		this.$panels.removeAttr('role');
	}

	if (this.options.triggerEventAfterInit) {
		$document.trigger(events.tabsInitialized);
	}

	this.$element.addClass(this.options.stateClasses.isInitialized);
};

Module.prototype._setHash = function($hashTarget) {
	var hashIndex = $hashTarget.index();

	this._setTabIndexWithoutAnimation(hashIndex);
};

Module.prototype._addScrollGradientsOnTouchDevice = function() {
	var hasOverflow = this._hasOverflow(),
		$tabList = this.$element.find(this.options.domSelectors.tabsList);

	if (hasOverflow) {
		this.$tabsWrapper.addClass(this.options.stateClasses.isScrolling);

		$tabList.on('scroll.' + this.uuid, function(event) {
			var scrollPosition = $(event.currentTarget).scrollLeft() || 0;

			if (scrollPosition <= 0) {
				this.$tabsWrapper.addClass(this.options.stateClasses.isScrollingStart);
			} else {
				this.$tabsWrapper.removeClass(this.options.stateClasses.isScrollingStart);
			}

			if (scrollPosition + this.$tabsWrapper.width() >= this.$element.find(this.options.domSelectors.tabsList)[0].scrollWidth - 1) {
				this.$tabsWrapper.addClass(this.options.stateClasses.isScrollingEnd);
			} else {
				this.$tabsWrapper.removeClass(this.options.stateClasses.isScrollingEnd);
			}

		}.bind(this));
	} else {
		$tabList.off('scroll');
	}

	this._scrollActiveTabIntoView(this.$tabs.eq(this.options.activeTabIndex));
};

/**
 * Adjust slider position on tabs panel
 * @private
 */
Module.prototype._onResizeHandler = function() {
	this._scrollActiveTabIntoView(this.$tabs.eq(this.options.activeTabIndex));
	if (!isTouchDevice) {
		this._addScrollGradientsOnTouchDevice();
	}
};

/**
 * Bind keyboard handlers for tab interactions
 * @private
 */
Module.prototype._bindKeyboardHandlers = function() {
	this.$tabs.on(events.keydown + '.' + this.uuid, this._handleKeypress.bind(this));
};

/**
 * Handler for keypress event
 * @param event {jQuery} object
 * @private
 */
Module.prototype._handleKeypress = function(event) {
	var keycode = event.keyCode;

	switch (keycode) {
		case 13:

			this._setTabIndex($(event.currentTarget).index());
			break;

		case 37:
		case 38:

			// left
			if ($(event.currentTarget).index() > 0) {
				event.preventDefault();
				$(event.currentTarget).prev().focus();
			}

			break;
		case 39:
		case 40:

			// right
			if ($(event.currentTarget).index() < this.$tabs.length - 1) {
				event.preventDefault();
				$(event.currentTarget).next().focus();
			}

			break;
	}
};

/**
 * Check if Tablist has a horizontal overflow
 * @returns {boolean}
 * @private
 */
Module.prototype._hasOverflow = function() {
	var invisibleItems = [],
		$tabList = this.$element.find(this.options.domSelectors.tabsList),
		tabListOffsetLeft = $tabList[0].offsetLeft,
		tabListOffsetWidth = $tabList[0].offsetWidth,
		childElementsLength = $tabList[0].childElementCount;

	while (childElementsLength--) {
		if ($tabList[0].children[childElementsLength].offsetLeft + $tabList[0].children[childElementsLength].offsetWidth > tabListOffsetLeft + tabListOffsetWidth) {
			invisibleItems.push($tabList[0].children[childElementsLength]);
		}
	}

	return (invisibleItems.length > 0);

};

/**
 * Trigger active tab by API call
 * @param activeTabIndex integer starting from index 0 as first tab
 * @public
 */
Module.prototype.setActiveTabByIndex = function(activeTabIndex) {
	this._setTabIndex(activeTabIndex);
};

/**
 * Click handler for click event
 * @param event {jQuery}
 * @private
 */
Module.prototype._tabItemClick = function(event) {
	var $el = $(event.currentTarget),
		index = $el.index();

	this._setTabIndex(index);
};

/**
 * Set active tab but without animation on tabs and panels
 * @param tabIndex - integer, defaults to 0
 * @private
 */
Module.prototype._setTabIndexWithoutAnimation = function(tabIndex) {
	this.$element.addClass(this.options.stateClasses.hasNoAnimation);
	this._setTabIndex(tabIndex);
	this.$element.removeClass(this.options.stateClasses.hasNoAnimation);
};

/**
 * Set active tab
 * @param tabIndex - integer, defaults to 0
 * @private
 */
Module.prototype._setTabIndex = function(tabIndex) {
	if (tabIndex < 0) {
		tabIndex = 0;
	}

	var $curTab = this.$tabs.eq(this.options.activeTabIndex),
		$curPanel = this.$panels.eq(this.options.activeTabIndex),
		$newTab = this.$tabs.eq(tabIndex),
		$newPanel = this.$panels.eq(tabIndex);

	// set tabindex to -1 on all tabs and panels
	this.$tabs.attr('tabindex', '-1');
	this.$panels.attr('tabindex', '-1');

	$curTab.removeClass(this.options.stateClasses.isActive).attr('aria-selected', false);
	$curPanel.removeClass(this.options.stateClasses.isActive).addClass(this.options.stateClasses.isHidden).attr('aria-hidden', true);
	$newTab.addClass(this.options.stateClasses.isActive).attr({'aria-selected': true, 'tabindex': '0'});
	$newPanel.removeClass(this.options.stateClasses.isHidden).attr({'aria-hidden': false, 'tabindex': '0'});

	if (this.$tabs.length === 1) {
		$newTab.attr({'tabindex': '-1'});
		$newPanel.attr({'tabindex': '-1'});
	}

	// Add slight delay to opening animation so the browser gets time to remove display:none from object before animating opacity
	setTimeout(function() {
		$newPanel.addClass(this.options.stateClasses.isActive);
		$document.trigger(events.tabSwitched, $newTab.attr('id'));
	}.bind(this), 30);

	// Assign activeTabIndex value passed to function
	this.options.activeTabIndex = tabIndex;

	this._scrollActiveTabIntoView(this.$tabs.eq(tabIndex));
};

/**
 * Scrolls active tab into center of viewport on startup or on prev/next key
 * click, whenever tab item is out of visible bounds
 * @param $activeTab
 * @private
 */
Module.prototype._scrollActiveTabIntoView = function($activeTab) {
	var activeTabOffset, componentWidth, scrollPosition;

	if (!$activeTab.length) {
		// return if no active tab found
		return;
	}

	activeTabOffset = $activeTab.position().left;
	componentWidth = this.$element.width();

	// scroll position cannot be negative, so we'll always have to check for < 0 and scroll to 0
	// special case for custom scroll, as native scroll could handle that
	scrollPosition = Math.max(parseInt(activeTabOffset - (componentWidth / 2) + ($activeTab.width() / 2)), 0);

	if (isTouchDevice) {

		if (this.options.activeTabIndex === 0) {
			this.$tabsWrapper.addClass(this.options.stateClasses.isScrollingStart);
		} else {
			this.$tabsWrapper.removeClass(this.options.stateClasses.isScrollingStart);
		}

		if (this.options.activeTabIndex >= this.$tabs.length - 1) {
			this.$tabsWrapper.addClass(this.options.stateClasses.isScrollingEnd);
		} else {
			this.$tabsWrapper.removeClass(this.options.stateClasses.isScrollingEnd);
		}

		gsap.to($activeTab.parent(), {duration: 0.5, scrollLeft: scrollPosition});
	} else {
		this.$element.find(this.options.domSelectors.tabsList).mCustomScrollbar('scrollTo', scrollPosition);
	}
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	initEvents: ['ready', 'ajaxload'],
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 144 */,
/* 145 */,
/* 146 */,
/* 147 */,
/* 148 */,
/* 149 */,
/* 150 */,
/* 151 */,
/* 152 */,
/* 153 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(154);


/***/ }),
/* 154 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 155 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(156);


/***/ }),
/* 156 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M00 - Map
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(157);

// require css from leaflet
__webpack_require__(158);
__webpack_require__(159);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	cssTransitionEnd = __webpack_require__(4),
	keymaster = __webpack_require__(14);

var name = 'mapboxmap',
	$document = $(document),
	events = {
		markerClick: 'markerClick.estatico.' + name,

		// consuming events. other modules may trigger these events.
		invalidateSize: 'invalidateSize.estatico.' + name,
		setAllMarkersActive: 'setMarkersActive.' + name,
		removeMarkers: 'removeMarkers.' + name,
		addMarkers: 'addMarkers.' + name,
		addInitialMarkers: 'addInitialMarkers.' + name,
		setZoomAroundMarkers: 'setZoomAroundMarkers.' + name
	},
	defaults = {
		domSelectors: {
			map: '[data-' + name + '="map"]',
			overlay: '[data-' + name + '="overlay"]',
			mobileOverlay: '[data-' + name + '="overlay_mobile"]',
			copyrightButton: '[data-' + name + '="copyright_button"]',
			copyrightContainer: '[data-' + name + '="copyright_container"]',
			copyrightButtonClose: '[data-' + name + '="copyright_button_close"]',
			accordionContent: '[data-accordion="content"]',
			tabContent: '[data-tabs="tab_panel_item"]'
		},
		stateClasses: {
			isExpanded: 'is_expanded',
			isVisible: 'is_visible',
			isOpen: 'is_open'
		},
		markerClass: 'mod_map_marker',
		zoomButtonsClass: 'mod_map_button',
		zoomInButtonClass: 'var_zoom_in',
		zoomOutButtonClass: 'var_zoom_out',
		zoomButtonInnerClass: 'mod_map_button_inner',
		clusterClass: 'mod_map_cluster',
		specialDisplayInitClass: 'display_for_init',
		markersPath: 'https://cdn.app.sbb.ch/png/map/v1/',
		token: 'pk.eyJ1IjoiZ2VvcHMiLCJhIjoiVnlkeXdzayJ9.aGtp4w0C-0vZZae1HneLGg',
		style: 'geops',
		maxZoom: 18,
		lang: 'de',
		apiUrl: 'https://api.maptiler.com/maps/477a7ed2-6617-4940-a27e-fe2ed5d6842d/{z}/{x}/{y}.png?key=eBiJeMUyegs8KTOnktMB',
		expandingMap: false,
		markers: [],
		defaultBounds: [
			{
				lat: 47.84634,
				lng: 10.53589
			},
			{
				lat: 45.80966,
				lng: 5.83923
			}
		]
	},
	data = {},
	isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints;

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ***********************************************************************
// Initializing
// ***********************************************************************

// Module Init.
// Please call only once!
Module.prototype.init = function() {

	// load mapbox script async and give webpack ability to handle it correctly
	__webpack_require__.e/* require.ensure */(6).then((function(require) {

		// return if leaflet is already initialized
		if (this.leaflet) {
			return;
		}

		// retrieve the required modules
		this.leaflet = __webpack_require__(144);
		__webpack_require__(145);
		__webpack_require__(146);

		// Due to a bug in leaflet source, we need to hack the pointer propoerty
		// on leaflets Browser object
		// (see https://github.com/Leaflet/Leaflet/issues/5310, answer from 'druncore').
		// If we don't do this, chrome on android touch devices will think
		// that it has no touch capabilities.
		if (this.leaflet.Browser.chrome && this.leaflet.Browser.touch && isTouchDevice) {
			this.leaflet.Browser.pointer = false;
		}

		// call real init
		this.callbackInit();

	}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);
};

// Real initializer
Module.prototype.callbackInit = function() {

	// handle case when we have expanding map
	this.handleExpandingMap();

	// various stuff to do when map is loaded
	this.handleMapLoaded();

	// bind event listeners
	this.bindEventListeners();

	// handle drag/pan for touch devices
	this.handleTouchDrag();
};

// Handling click on map overlay when we have an expaning map
Module.prototype.handleExpandingMap = function() {

	var $overlay;

	// if we have an expanding map, add a corresponding listener to overlay
	if (this.options.expandingMap === true) {

		// get overlay
		$overlay = this.$element.find(this.options.domSelectors.overlay);

		// handle click on overlay
		$overlay.on('click.' + this.uuid, function() {

			// remove overlay
			$overlay.remove();

			// add class indicating that map is expanded
			this.$element.addClass(this.options.stateClasses.isExpanded);

			// show zoom buttons
			this.$element.find('.' + this.options.zoomButtonsClass).addClass(this.options.stateClasses.isVisible);

			// call invalidateSize to let leaflet know that we changed
			// the size of the container
			if (cssTransitionEnd) {
				$(this.$element).one(cssTransitionEnd, function() {
					this.handleInvalidateSize();
				}.bind(this));
			} else {
				// if browser does not support transition end, call it after 1s
				setTimeout(function() {
					this.handleInvalidateSize();
				}.bind(this), 1000);
			}

		}.bind(this));
	}
};

// Things to do when map has finished loading
Module.prototype.handleMapLoaded = function() {

	var $accordionContent,
		$tabContent;

	// we don't want to manipulate the center during setup
	this.dontManipulateCenterWhileZooming = true;

	// wait for the map to be loaded
	this.customMap().whenReady(function() {

		// handle zoom event and make sure that every zoom event results in an
		// int-zoom-level. float-zoom-levels will result in a blurry map.
		this.customMap().on('zoom', function() {
			this.ensureIntZoomLevel();
			this.eventuallyCenterMapOnMarkerWhileZooming();
		}.bind(this));

		// set the initial zoom level
		this.setInitialZoomLevel();

		// keep track of leafletMarkers
		this.markers = [];

		// add default markers
		this.addMarkers(this.options.markers);

		window.requestAnimationFrame(function() {

			// this seems to be a bit hacky. unfortunately, there is no other way
			// to set either data attributes or tabindex on the buttons.
			$('.' + this.options.zoomButtonsClass).attr('tabindex', '-1');

		}.bind(this));

		// if we don't have an expanding map, initially show zoom buttons
		if (this.options.expandingMap !== true) {
			this.$element.find('.' + this.options.zoomButtonsClass).addClass(this.options.stateClasses.isVisible);
		}

		// if the map is inside an accordion or tabs, we need to remove the special
		// init class on the content of the accordion
		// (s. m15_-_accordion.hbs for explanation)
		$accordionContent = this.$element.parents(this.options.domSelectors.accordionContent);
		if ($accordionContent.length > 0) {
			$accordionContent.removeClass(this.options.specialDisplayInitClass);
		}

		$tabContent = this.$element.parents(this.options.domSelectors.tabContent);
		if ($tabContent.length > 0) {
			$tabContent.removeClass(this.options.specialDisplayInitClass);
		}

	}.bind(this));

	this.dontManipulateCenterWhileZooming = false;
};

// Bind event listeners
Module.prototype.bindEventListeners = function() {

	var scope = this.name + '-' + this.uuid;

	this.$copyright = this.$element.find(this.options.domSelectors.copyrightContainer);

	// prevent force touch on map for ios
	this.$element.on('touchforcechange', function(evt) {
		if (evt.originalEvent.changedTouches[0].force > 0.1) {
			evt.preventDefault();
		}
	});

	// listen to clicks on copyright open / close
	this.$element.on('click.' + this.uuid, this.options.domSelectors.copyrightButton, this.toggleCopyright.bind(this));

	this.$element.on('click.' + this.uuid, this.options.domSelectors.copyrightButtonClose, this.toggleCopyright.bind(this));

	// listen to escape key
	this.$copyright.on('focusin.' + this.uuid, function() {
		keymaster.setScope(scope);
	}).on('focusout.' + this.uuid, function() {
		keymaster.setScope('all');
	});

	// handle escape key
	keymaster('esc', scope, function() {
		this.closeCopyright();
	}.bind(this));

	// the following events can be triggered from outside.

	// listen to event 'invalidateSize' to accomodate a new container size.
	$document.on(events.invalidateSize + '.' + this.uuid, this.handleInvalidateSize.bind(this));

	// listen to event 'setMarkersActive' which will then set all markers
	// to active again
	$document.on(events.setAllMarkersActive + '.' + this.uuid, this.setAllMarkersActive.bind(this));

	// listen to `removeMarkers`
	$document.on(events.removeMarkers + '.' + this.uuid, this.removeMarkers.bind(this));

	// listen to `addMarkers`
	$document.on(events.addMarkers + '.' + this.uuid, function(undefined, markers) {
		this.addMarkers(markers);
	}.bind(this));

	// listen to `addInitialMarkers`
	$document.on(events.addInitialMarkers + '.' + this.uuid, function() {
		this.addMarkers(this.options.markers);
	}.bind(this));

	// listen to `setZoomAroundMarkers`
	$document.on(events.setZoomAroundMarkers + '.' + this.uuid, this.setZoomAroundMarkers.bind(this));

};

// ***********************************************************************
// Marker specific stuff
// ***********************************************************************

// Add markers by passing an array of markers
Module.prototype.addMarkers = function(markers) {

	window.requestAnimationFrame(function() {
		var iterator = markers.length;

		// iterate passed markers and add them as marker to map
		while (iterator--) {
			this.addMarker(markers[iterator]);
		}

	}.bind(this));

};

// Add a marker by passing a $markerIcon and coordinates
Module.prototype.addMarker = function(marker) {

	var icon;

	// get icon
	if (marker.click || marker.url) {
		icon = this.markerIconActive(marker.smallIcon);
	} else {
		icon = this.markerIconInactive(marker.smallIcon);
	}

	// create marker
	var newMarker = this.leaflet.marker(
		[marker.lat, marker.lng],
		{
			icon: icon,
			keyboard: false
		}
	);

	// listen to clicks on marker if interactive option is true
	if (marker.click || marker.url) {
		newMarker.on('click', function(evt) {
			this.handleMarkerClick(evt, marker);
		}.bind(this));
	}

	// add marker to cluster group and to local array
	this.clusterGroup().addLayer(newMarker);
	this.markers.push(newMarker);

};

// Handle click on marker
Module.prototype.handleMarkerClick = function(evt, markerObject) {

	// if `click` is not an object, we expect a link which we will open
	if (markerObject.url) {
		window.location = markerObject.url;
		return;
	}

	// if `click` is empty, we can return
	if (!markerObject.click) {
		return;
	}

	// set active icon on all markers again
	this.setAllMarkersActive();

	// set inactive icon on clicked marker
	evt.target.setIcon(this.markerIconInactive());

	// trigger event mapMarkerClicked. Modules that use this module may listen
	// and handle to this event.
	$document.trigger(events.markerClick, markerObject);

	// if we have bounds inside the click object, we use those the set the
	// view of the map. Otherwise we expect center coordinates and a zoom level.
	if (markerObject.click.bounds) {
		this.setZoomByBounds(markerObject.click.bounds);
	} else {
		this.customMap().setView([markerObject.click.lat, markerObject.click.lng], markerObject.click.zoom);
	}

};

// Setting all markers to active
Module.prototype.setAllMarkersActive = function() {

	var iterator = this.markers.length;

	while (iterator--) {
		this.markers[iterator].setIcon(this.markerIconActive());
	}
};

// Remove all markers
Module.prototype.removeMarkers = function() {
	this.clusterGroup().clearLayers();
	this.markers = [];
};

// ***********************************************************************
// Convenience getter methods
// ***********************************************************************

// Get the map object
Module.prototype.customMap = function() {

	var buffer = 0,
		ZoomInButton,
		ZoomOutButton,
		container;

	// If we have already created the map, return it
	if (this._customMap) {
		return this._customMap;
	}

	// Since we don't have a map at this point, initialize and configure it.

	// if we have an expanding map, we should set buffer to 1 so that
	// the map can load 1 additional tile row at top and bottom
	if (this.options.expandingMap === true) {
		buffer = 1;
	}

	// get id of map container
	container = this.$element.find(this.options.domSelectors.map).attr('id');

	// Initialize the map object
	this._customMap = this.leaflet.map(container, {
		attributionControl: false,
		zoomControl: false,
		center: [0, 0],
		zoom: 1, // we need to set a value here, but it does not matter, since
		// we will set zoom based on bounds later on.
		zoomSnap: 0,
		scrollWheelZoom: false,
		keyboard: false
	});

	// add the layer with the tiles of specific style from mapbox
	this.leaflet.tileLayer(this.options.apiUrl, {
		maxZoom: this.options.maxZoom,
		id: this.options.style,
		accessToken: this.options.token,
		customBufferTiles: buffer // see leaflet_buffer.js in js/helpers
	}).addTo(this._customMap);

	// get zoom button Constructors from native control extension
	ZoomInButton = this.zoomInOutButton(true);
	ZoomOutButton = this.zoomInOutButton(false);

	// add custom zoom in button
	this._customMap.addControl(new ZoomOutButton());
	this._customMap.addControl(new ZoomInButton());

	// return the map
	return this._customMap;

};

// get the cluster group
Module.prototype.clusterGroup = function() {

	// if we already have a cluster group, return it
	if (this._clusterGroup) {
		return this._clusterGroup;
	}

	// init new cluster group and add it as layer to map
	this._clusterGroup = this.leaflet.markerClusterGroup({
		showCoverageOnHover: false,
		removeOutsideVisibleBounds: false,
		iconCreateFunction: function(cluster) {
			return this.leaflet.divIcon(
				{
					html: cluster.getChildCount(),
					className: this.options.clusterClass,
					iconSize: this.leaflet.point(38, 38)
				}
			);
		}.bind(this)
	});

	this.customMap().addLayer(this._clusterGroup);

	// listen to spiderfied event for marker cluster andmake sure marker
	// clusters can not be focussed
	this._clusterGroup.on('animationend', function() {
		$('.' + this.options.clusterClass).attr('tabindex', '-1');
	}.bind(this));

	return this._clusterGroup;
};

// get active icon for marker
Module.prototype.markerIconActive = function(isForSmallIcon) {

	// if we already have an icon, return it
	if (this._markerIconActive && !isForSmallIcon) {
		return this._markerIconActive;
	} else if (this._markerIconActiveSmall && isForSmallIcon) {
		return this._markerIconActiveSmall;
	}

	// init new icon
	if (isForSmallIcon) {
		this._markerIconActiveSmall = this.markerIconActiveInactive(true, true);
		return this._markerIconActiveSmall;
	} else {
		this._markerIconActive = this.markerIconActiveInactive(true, false);
		return this._markerIconActive;
	}
};

// get inactive icon for marker
Module.prototype.markerIconInactive = function(isForSmallIcon) {

	// if we already have an icon, return it
	if (this._markerIconInactive && !isForSmallIcon) {
		return this._markerIconInactive;
	} else if (this._markerIconInactiveSmall && isForSmallIcon) {
		return this._markerIconInactiveSmall;
	}

	// init new icon
	if (isForSmallIcon) {
		this._markerIconInactiveSmall = this.markerIconActiveInactive(false, true);
		return this._markerIconInactiveSmall;
	} else {
		this._markerIconInactive = this.markerIconActiveInactive(false, false);
		return this._markerIconInactive;
	}
};

// helper to get active or inactive marker
Module.prototype.markerIconActiveInactive = function(isForActive, isForSmallIcon) {

	var icon,
		imageName,
		iconSize = isForSmallIcon ? [16, 23.5] : [32, 47],
		iconAnchor = isForSmallIcon ? [8, 23.5] : [16, 47],
		popupAnchor = isForSmallIcon ? [8, 0] : [16, 0];

	if (isForActive) {
		imageName = 'marker-red.png';
	} else {
		imageName = 'marker-black.png';
	}

	icon = this.leaflet.icon({
		iconUrl: this.options.markersPath + imageName,
		iconSize: iconSize,
		iconAnchor: iconAnchor,
		popupAnchor: popupAnchor,
		className: this.options.markerClass
	});

	return icon;
};

// ***********************************************************************
// Map control methods
// ***********************************************************************

// ensure int zoom level: if we have decimal number as zoom, the map will most
// likely be blurry
Module.prototype.ensureIntZoomLevel = function() {
	var currentZoom = this.customMap()._zoom,
		intZoom = parseInt(currentZoom);

	if (currentZoom === intZoom) {
		return;
	}

	this.customMap().setZoom(intZoom);
};

// If we have only one marker on the map or if we have only one marker
// in the visible area of the map, we want to make sure that the map
// is centering on that map while zooming
Module.prototype.eventuallyCenterMapOnMarkerWhileZooming = function() {
	var visibleMarkers;

	if (this.dontManipulateCenterWhileZooming) {
		return;
	}

	if (this.options.markers.length === 1) {
		this.setCenterOnMarker(this.options.markers[0]);
		return;
	}

	visibleMarkers = this.getVisibleMarkers();

	if (visibleMarkers.length !== 1) {
		return;
	}

	this.setCenterOnMarker(visibleMarkers[0]);

};

// Helper to set the map to center on the passed marker
Module.prototype.setCenterOnMarker = function(marker) {
	var markerCoords = this.leaflet.latLng(marker.lat, marker.lng),
		zoom = this.customMap().getZoom();

	this.customMap().setView(markerCoords, zoom, { animate: true });
};

// set initial zoom level. either by using provided bounds or zoom
Module.prototype.setInitialZoomLevel = function() {

	// 3 Methods to set initial zoom:
	// 1.: Set zoom and center based on passed bounds
	// 2.: Set zoom and center so that all passed markers are visible
	// 3.: Set zoom and center based on passed lat/lng and zoom values
	// 4.: Fallback, set center on passed marker (there can only be
	// one single marker if we arrive at step 4) and set zoom to some default
	// value

	if (this.options.bounds && this.options.bounds.length === 2) { // 1
		this.setInitialZoomLevelByBounds();
	} else if (this.options.markers.length > 1) { // 2
		this.setInitialZoomAroundMarkers();
	} else if (this.options.center) { // 3
		this.setInitialZoomLevelByZoomAndCenter();
	} else {
		this.setInitialZoomAroundMarkerFallback();
	}

};

// set initial zoom by provided bounds
Module.prototype.setInitialZoomLevelByBounds = function() {
	this.setZoomByBounds(this.options.bounds);
};

Module.prototype.setZoomByBounds = function(boundsPoints) {

	var bounds;

	if (boundsPoints.length !== 2) {
		return;
	}

	bounds = this.leaflet.latLngBounds(boundsPoints[0], boundsPoints[1]);
	this.customMap().fitBounds(bounds);
};

// set initial zoom by setting view so that all markers are visible
Module.prototype.setInitialZoomAroundMarkers = function(_markers) {
	var bounds,
		markers = _markers || this.options.markers;

	bounds = this.getBoundsForMarkers(markers);

	this.customMap().fitBounds(bounds, {
		padding: [50, 50]
	});
};

// set initial zoom by provided zoom and center coordinates
Module.prototype.setInitialZoomLevelByZoomAndCenter = function() {

	if (!this.options.center) {
		return;
	}

	var lat = this.options.center.lat,
		lng = this.options.center.lng,
		zoom = this.options.center.zoom;

	if (lat && lng && zoom) {
		this.customMap().setView([lat, lng], zoom);
	}
};

// Fallback to set the center to the passed marker location, and the
// zoom to some default value
Module.prototype.setInitialZoomAroundMarkerFallback = function() {

	// we should arrive here only, if we have just one marker.
	// we grap the default bounds, which are the bounds for
	// switzerland. we then determine the zoom level so that those
	// bounds would be visible. in that way we get an acceptable zoom
	// level. then, we set the map to that zoom level and the center
	// to the coordinate of the marker.

	var bounds = this.getBoundsForMarkers(this.options.defaultBounds),
		marker,
		markerCoords,
		zoom = this.customMap().getBoundsZoom(bounds);

	if (this.options.markers.length < 1) {
		return;
	}

	marker = this.options.markers[0];
	markerCoords = this.leaflet.latLng(marker.lat, marker.lng);
	this.customMap().setView(markerCoords, zoom);

};

// set the zoom level so that the passed markers are visible
Module.prototype.setZoomAroundMarkers = function(evt) {

	var bounds,
		containing;

	if (!evt) {
		return;
	}

	if (!evt.markers) {
		return;
	}

	if (evt.markers.length <= 0) {
		return;
	}

	// let's see if the passed markers are visible. To do so, we first need
	// the bounds defined by the makers
	bounds = this.getBoundsForMarkers(evt.markers);

	// now let's see if the bounds are inside currently visible area of the map...
	containing = this.customMap().getBounds().contains(bounds);

	// ... if not, we set the zoom around the markers
	if (!containing) {
		this.setInitialZoomAroundMarkers(evt.markers);
	}

};

// zoom in the map by one level
Module.prototype.zoomIn = function() {
	this.customMap().zoomIn();
};

// zoom out the map by one level
Module.prototype.zoomOut = function() {
	this.customMap().zoomOut();
};

// Handle event 'invalidateSize'
Module.prototype.handleInvalidateSize = function() {
	this.customMap().invalidateSize(false);
};

// Handle drag/pan for touch devices
Module.prototype.handleTouchDrag = function() {

	if (!isTouchDevice || !this.leaflet.Browser.mobile) {
		return;
	}

	// on touch devices, disable drag and tap events and implement 2-finger pan

	this.customMap().dragging.disable();
	if (this.customMap().tap) {
		this.customMap().tap.disable();
	}

	// on movestart we set this flag to true. that means, during the timeout
	// phase, the map moved, which indicates that the user currently has 2
	// fingers on, so we don't want to show the overlay.
	this.customMap().on('move', function() {
		this.movestartCancelsMobileOverlay = true;

		// in this case - should the overlay be visible - we definitely want
		// to hide it
		this.handleTouchEnd();

	}.bind(this));

	// listen to touchstart event
	this.$element.on('touchstart.' + this.uuid, this.handleTouchStart.bind(this));

	// listen to touch end event
	this.$element.on('touchend.' + this.uuid, this.handleTouchEnd.bind(this));

};

// Handle touch start.
// - Show an informative overlay when user does a 1-finger touch
// - Pan and zoom the map like normal if user does a 2-finger touch
Module.prototype.handleTouchStart = function(evt) {

	this.movestartCancelsMobileOverlay = false;

	// we only want to proceed if we have a 1-finger touch
	if (evt.originalEvent.targetTouches.length !== 1) {
		return;
	}

	// show overlay

	// The user pretty sure won't lay down the 2 fingers exactly at the same
	// time. For this reason, we add a timeout here to have the possibility
	// to cancel the showing of the overlay if the user lays down 2 fingers
	// during the timeout phase.
	// Timeout value of 150 is experimental. Experiences:
	// - 150 is too short: if a user taps on a marker, it may be that the overlay
	// appears unintentionally
	// - 250 is too long: if the user scrolls the page with the finger on the map,
	// the overlay appears too late
	setTimeout(function() {
		if (this.movestartCancelsMobileOverlay) {
			return;
		}

		this.$element.find(this.options.domSelectors.mobileOverlay).addClass(this.options.stateClasses.isVisible);
	}.bind(this), 200);

};

// Handle touch end
// - Hide the informative overlay when user ends a 1-finger touch
Module.prototype.handleTouchEnd = function() {

	this.movestartCancelsMobileOverlay = true;

	// hide overlay. We want to hide the overlay in any case, no matter if we
	// have 1 or more touches.
	this.$element.find(this.options.domSelectors.mobileOverlay).removeClass(this.options.stateClasses.isVisible);

};

// ***********************************************************************
// Leaflet extensions
// ***********************************************************************

// extend the native control class for zoom in/out button
Module.prototype.zoomInOutButton = function(isForZoomIn) {

	var buttonElement,
		buttonControl,
		buttonClasses = this.options.zoomButtonsClass + ' ',
		clickHandler;

	if (isForZoomIn) {
		buttonClasses += this.options.zoomInButtonClass;
		clickHandler = this.zoomIn.bind(this);
	} else {
		buttonClasses += this.options.zoomOutButtonClass;
		clickHandler = this.zoomOut.bind(this);
	}

	buttonControl = this.leaflet.Control.extend({
		options: {
			position: 'bottomright'
		},
		onAdd: function() {

			// create buttonElement with leaflefts DomUtil constructor
			buttonElement = this.leaflet.DomUtil.create('button', buttonClasses);
			this.leaflet.DomUtil.create('span', this.options.zoomButtonInnerClass, buttonElement);

			// make sure that clicks or doubleclicks do not pan or zoom
			// the map
			this.leaflet.DomEvent.disableClickPropagation(buttonElement);

			// add click event and handler
			$(buttonElement).on('click', clickHandler);

			return buttonElement;

		}.bind(this)
	});

	return buttonControl;
};

// ***********************************************************************
// Copyright
// ***********************************************************************

Module.prototype.closeCopyright = function() {
	if (this.$copyright.hasClass(this.options.stateClasses.isOpen)) {
		this.toggleCopyright();
	}
};

Module.prototype.toggleCopyright = function() {
	window.requestAnimationFrame(function() {
		if (this.$copyright.hasClass(this.options.stateClasses.isOpen)) {
			this.$copyright.removeClass(this.options.stateClasses.isOpen);

			if (cssTransitionEnd) {
				this.$copyright.one(cssTransitionEnd, function() {
					this.$copyright.css({display: 'none'});
				}.bind(this));
			} else {
				setTimeout(function() {
					this.$copyright.css({display: 'none'});
				}.bind(this), 300);
			}
		} else {
			this.$copyright.css({display: 'block'});

			setTimeout(function() {
				this.$copyright.addClass(this.options.stateClasses.isOpen);
			}.bind(this), 1);

		}

	}.bind(this));

};

// ***********************************************************************
// Helpers
// ***********************************************************************

Module.prototype.getBoundsForMarkers = function(markers) {

	var markerPoints,
		markerPoint;

	markerPoints = markers.reduce(function(points, val) {
		markerPoint = {lat: val.lat, lng: val.lng};
		points.push(markerPoint);
		return points;
	}, []);

	return this.leaflet.latLngBounds(markerPoints);
};

Module.prototype.getVisibleMarkers = function() {
	var iterator,
		marker,
		markerCoords,
		bounds = this.customMap().getBounds(),
		visibleMarkers = [];

	iterator = this.options.markers.length;

	// iterate passed markers and add them as marker to map
	while (iterator--) {
		marker = this.options.markers[iterator];
		markerCoords = this.leaflet.latLng(marker.lat, marker.lng);

		if (bounds.contains(markerCoords)) {
			visibleMarkers.push(marker);
		}
	}

	return visibleMarkers;
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 157 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 158 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 159 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 160 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(161);


/***/ }),
/* 161 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M00 - Journey Map
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2021.
 */



__webpack_require__(162);
__webpack_require__(163);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'journeymap',
	$document = $(document),
	events = {
		markerClick: 'markerClick.estatico.' + name,

		// consuming events. other modules may trigger these events.
		unselectAllMarkers: 'unselectMarkers.' + name,
		removeMarkers: 'removeMarkers.' + name,
		addMarkers: 'addMarkers.' + name
	},
	defaults = {
		domSelectors: {
			map: '[data-' + name + '="map"]',
			overlay: '[data-' + name + '="overlay"]'
		},
		stateClasses: {
			isExpanded: 'is_expanded'
		},
		expandingMap: false
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ***********************************************************************
// Initializing
// ***********************************************************************

// Module Init.
// Please call only once!
Module.prototype.init = function() {
	/* jshint ignore:start */
	var intersectionObserver = new IntersectionObserver(function(changes) {

		changes.forEach(function(change) {
			// Journey map is in the viewport
			if (change.isIntersecting) {
				__webpack_require__.e/* require.ensure */(7).then((function(require) {

					// retrieve the required modules
					__webpack_require__(147);
					__webpack_require__(148);

					// call real init
					this.callbackInit();

					intersectionObserver.unobserve(this.$element[0]);

				}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);
			}
		}.bind(this));

	}.bind(this), { rootMargin: '2500px 0px 0px 0px' });

	// start observing the element
	intersectionObserver.observe(this.$element[0]);

	/* jshint ignore:end */
};

// Real initializer
Module.prototype.callbackInit = function() {
	this.mapComponent = document.createElement('journey-maps-client');

	this.mapComponent.language = this.options.lang;
	this.mapComponent.apiKey = this.options.apiKey;
	this.mapComponent.viewportOptions = this.options.viewportOptions;
	this.mapComponent.controlOptions = this.options.controlOptions;
	this.mapComponent.markerOptions = this.options.markerOptions;

	this.$element.find(this.options.domSelectors.map).append(this.mapComponent);

	// handle case when we have expanding map
	this.handleExpandingMap();

	// bind event listeners
	this.bindEventListeners();

};

// Handling click on map overlay when we have an expaning map
Module.prototype.handleExpandingMap = function() {

	var $overlay;

	// if we have an expanding map, add a corresponding listener to overlay
	if (this.options.expandingMap === true) {

		// get overlay
		$overlay = this.$element.find(this.options.domSelectors.overlay);

		// handle click on overlay
		$overlay.on('click.' + this.uuid, function() {

			// hide overlay
			$overlay.hide();

			// add class indicating that map is expanded
			this.$element.addClass(this.options.stateClasses.isExpanded);

		}.bind(this));
	}
};

// Bind event listeners
Module.prototype.bindEventListeners = function() {

	// prevent force touch on map for ios
	this.$element.on('touchforcechange', function(evt) {
		if (evt.originalEvent.changedTouches[0].force > 0.1) {
			evt.preventDefault();
		}
	});

	// listen to a marker selection
	this.mapComponent.addEventListener('selectedMarkerIdChange', function(event) {
		this.handleMarkerClick(event.detail);
	}.bind(this));

	// the following events can be triggered from outside.

	// listen to event 'unselectAllMarkers' which will then unselect all markers
	$document.on(events.unselectAllMarkers + '.' + this.uuid, this.unselectAllMarkers.bind(this));

	// listen to `removeMarkers`
	$document.on(events.removeMarkers + '.' + this.uuid, this.removeMarkers.bind(this));

	// listen to `addMarkers`
	$document.on(events.addMarkers + '.' + this.uuid, function(undefined, markers) {
		this.addMarkers(markers);
	}.bind(this));

};

// ***********************************************************************
// Marker specific stuff
// ***********************************************************************

// Add markers by passing an array of markers
Module.prototype.addMarkers = function(markers) {

	window.requestAnimationFrame(function() {
		var iterator = markers.length;

		// iterate passed markers and add them as marker to map
		while (iterator--) {
			this.addMarker(markers[iterator]);
		}

		this.mapComponent.markerOptions = {
			markers: this.markers
		};

	}.bind(this));

};

// Add a marker by passing a $markerIcon and coordinates
Module.prototype.addMarker = function(marker) {

	this.markers.push({
		id: marker.id,
		title: marker.label,
		position: [marker.lng, marker.lat],
		category: 'NEUTRAL',
		markerUrl: marker.url
	});
};

// Handle click on marker
Module.prototype.handleMarkerClick = function(clickedMarkerId) {
	var markerOptions = this.options.markerOptions.markers,
		markerObject;

	markerObject = markerOptions.filter(function(marker) {
		return marker.id === clickedMarkerId;
	});

	// trigger event mapMarkerClicked. Modules that use this module may listen
	// and handle to this event.
	$document.trigger(events.markerClick, markerObject[0]);
};

// Unselect all markers
Module.prototype.unselectAllMarkers = function() {
	this.mapComponent.selectedMarkerId = '';
};

// Remove all markers
Module.prototype.removeMarkers = function() {
	this.mapComponent.markerOptions = {};

	this.markers = [];
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 162 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 163 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 164 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(143);


/***/ }),
/* 165 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 166 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(167);


/***/ }),
/* 167 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 168 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(169);


/***/ }),
/* 169 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M05 - Crosslink Teaser
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(170);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	SliderModule = __webpack_require__(55);

var name = 'crosslinkteaser',
	events = {
		showCurrentSlide: 'showCurrentSlide.estatico.' + name
	},
	defaults = {
		domSelectors: {
		},
		sliderSettings: {
			itemsQuantity: {
				tiny: 1,
				mobilePortrait: 1,
				mobileLandscape: 1,
				tabletPortrait: 2,
				tabletLandscape: 2,
				desktop: 4,
				desktopLarge: 4,
				desktop4k: 4,
				desktop5k: 4
			}
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype, SliderModule.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	SliderModule.prototype.init.apply(this);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	initEvents: ['ready', 'ajaxload'],
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 170 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 171 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(172);


/***/ }),
/* 172 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 173 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


/**
 * This helper accepts a file input. It is returning an object
 * with the width and the height of the image
 */

module.exports = function(fileUpload) {
	var promise,
		image = new Image(),
		size;

	image.src = window.URL.createObjectURL(fileUpload.files[0]);

	promise = new Promise(function(resolve) {
		image.onload = function() {
			size = {
				width: image.naturalWidth,
				height: image.naturalHeight
			};

			window.URL.revokeObjectURL(image.src);

			resolve(size);
		};
	}.bind(this));

	return promise;

};


/***/ }),
/* 174 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * Helper class SuperStick - sticking any element on page
 */



var $ = __webpack_require__(0),
	cssTransitionEnd = __webpack_require__(4),
	mediaqueries = __webpack_require__(6),
	contains = __webpack_require__(15),
	gsap = __webpack_require__(11).gsap,
	globalEvents = __webpack_require__(5);

var name = 'nbc',
	$document = $(document),
	$window = $(window),
	$body = $('body'),
	events = {
		update: 'update.estatico.' + name,
		recountTotalPrice: 'resize.estatico.totalprice'
	},
	options = {
		domSelectors: {
			moduleTitle: '[data-' + name + '="moduletitle"]',
			item: '[data-' + name + '="item"]',
			link: '[data-' + name + '="link"]',
			title: '[data-' + name + '="title"]',
			itemContent: '[data-' + name + '="content"]',
			toggleButton: '[data-' + name + '="toggle_button"]',
			header: '[data-init="header"]',
			buttonContainer: '[data-' + name + '="button_container"]'
		},
		dataBodyPosition: name + 'BodyPosition',
		stateClasses: {
			isFixed: 'is_fixed',
			isSet: 'is_set',
			isScrolled: 'is_scrolled',
			isExpanded: 'is_expanded',
			isOverflowDisabled: 'is_overflow_disabled',
			hasAnimation: 'has_animation',
			varMobile: 'var_has_mobile_version',
			bodyOverflow: 'is_layout_overflow_hidden'
		},
		stickerSettings: {
			defaultTopGutter: 80,
			defaultBottomGutter: 0,
			gutterTop: 80,
			gutterBottom: 0,
			defaultItemSpace: 0,
			itemSpace: 0,
			itemsBorderWidth: 1,
			toggleFunctionalityOnMobile: false
		},
		waitForTarget: false,
		keys: {
			close: [27]
		}
	},
	data = {
		itemsData: [],
		containerData: {},
		windowHeight: 0,
		scrollTop: 0,
		scrapValue: 0,
		scrapObjectTopMargin: 0,
		heightChangeListen: true
	};

function Nbc() {

	this.init();

}

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Nbc.prototype.init = function() {
	this.options = $.extend(true, {}, options, this.options);
	this.data = $.extend(true, {}, data, this.data);
	this.inDesktopViewport = this._checkIfInDesktopViewport();

	this.cacheElements();
	this._setInitData();

	// get init data
	if (this.inDesktopViewport) {

		this._setDesktopListeners();
		this._handleDesktopScrollAction();

	} else if (!this.options.waitForTarget || this.$toggleButton.hasClass('at-element-click-tracking')) {
		this._mobilePreInit();
	} else {
		document.addEventListener(adobe.target.event.CONTENT_RENDERING_SUCCEEDED, function() { // jshint ignore:line
			this._mobilePreInit();
			document.removeEventListener(adobe.target.event.CONTENT_RENDERING_SUCCEEDED, function() {}); // jshint ignore:line
		}.bind(this));
	}

	// add listeners
	$document
	.on(mediaqueries.event.key + '.' + this.uuid, this._handleBreakpointChange.bind(this))
	.on(globalEvents.contentchange.key + '.' + this.uuid, this._handleMinHeight.bind(this))
	.on(events.update + '.' + this.uuid, function() {
		this._rebuildModule();
	}.bind(this));

	// sbbch-12819: when accordions of content (m15) are closing, content becomes shorter, recalc position
	$document.on('m15-accordion.close-item', function() {
		setTimeout(function() {
			this._setCurrentUI();
		}.bind(this), 500);
	}.bind(this));

	this._checkBodyHeight();

	// calc position on load of the page
	setTimeout(function() {
		this._setCurrentUI();
	}.bind(this), 1200);
};

Nbc.prototype.cacheElements = function() {

	this.$moduleTitle = this.$element.find(this.options.domSelectors.moduleTitle);
	this.moduleTitle = this.$moduleTitle.length > 0 ? this.$moduleTitle[0].innerText : '';
	this.$item = this.$element.find(this.options.domSelectors.item);
	this.$elementParent = this.$element.parent();
	this.$itemTitles = this.$element.find(this.options.domSelectors.link);
	this.$itemContent = this.$element.find(this.options.domSelectors.itemContent);
	this.$scrapObject = this.$elementParent;
	this.bodyHeight = $body.outerHeight();

	this.$toggleButton = this.$element.next(this.options.domSelectors.buttonContainer).find(this.options.domSelectors.toggleButton);
};

/**
 * Check whether we are in desktop viewport
 * @returns {*}
 * @private
 */
Nbc.prototype._checkIfInDesktopViewport = function() {
	return mediaqueries.query({
		from: 'desktop'
	});
};

/**
 * Rebuild the module
 * @private
 */
Nbc.prototype._rebuildModule = function() {

	if (this.inDesktopViewport) {
		this._resetSettings(); // reset css settings only in desktop
	}

	requestAnimationFrame(function() {
		this.data.itemsData = [];
		this.$item.removeClass(this.options.stateClasses.isSet);
		this._setInitData();
	}.bind(this));
};

Nbc.prototype._mobilePreInit = function() {
	this._setMobileListeners();
	this._setToggleButtonTrackingAttributeValue();

	if (!this.options.stickerSettings.toggleFunctionalityOnMobile) {
		this._setMobileInit();
	} else {
		this._handleMobileScrollAction();
	}
};

Nbc.prototype._setToggleButtonTrackingAttributeValue = function() {
	var viewport = mediaqueries.query({ from: 'tabletPortrait' }) ? 'tablet' : 'mobile';

	this.$toggleButton.attr({
		'data-sit-component': 'nbc ' + viewport
	});
};

/**
 * Set init state for mobile devices.
 * @private
 */
Nbc.prototype._setMobileInit = function() {
	requestAnimationFrame(function() {
		this.$itemContent.css({
			height: 0
		});
	}.bind(this));
};

/**
 * Function listening for body height change.
 * On height change recalculate module position and data.
 * @private
 */
Nbc.prototype._checkBodyHeight = function() {
	if (this.destroyed) {
		return;
	}

	if (this.inDesktopViewport) {

		this.$element.parent().siblings().on(cssTransitionEnd + '.' + this.uuid, function() {
			if (this.data.heightChangeListen) {
				this._handleMinHeight();
			}
		}.bind(this));
	} else {
		this._resetSettings();
	}

};

Nbc.prototype._handleMinHeight = function() {

	var height = 0;

	this.$element.parent().siblings().children().each(function() {
		height += $(this).height();
	});

	if (this.$element.height() < height) {
		this.$elementParent.css({
			'min-height': 0
		});
	}

	if (this.bodyHeight !== $body.outerHeight()) {
		this.bodyHeight = $body.outerHeight();
		this._handleResizeChange();
	}
};

/**
 * Function handling window size changes.
 * @private
 */
Nbc.prototype._handleResizeChange = function() {
	this.data.windowHeight = window.innerHeight ? window.innerHeight : $window.height();

	if (this.inDesktopViewport) {
		this._setProperGutterValues();
		this._recalculatePositions();
		this._updateNbcItemsDisplay();
		this._checkScrapValue();
	} else {
		this._resetSettings();

		if (!this.options.stickerSettings.toggleFunctionalityOnMobile) {
			this._setMobileInit();
		}
	}

};

/**
 * Function handling breakpoint changes
 * @private
 */
Nbc.prototype._handleBreakpointChange = function() {

	this.inDesktopViewport = this._checkIfInDesktopViewport();

	this._disableListeners();
	this._resetMobileSettings();

	if (this.inDesktopViewport) {

		if (this.data.heightChangeListen) {
			this._handleMinHeight();
		}

		this._setProperGutterValues();
		this._setDesktopListeners();

	} else {

		this._resetSettings();
		this._setMobileListeners();
		this._setToggleButtonTrackingAttributeValue();

		if (!this.options.stickerSettings.toggleFunctionalityOnMobile) {
			this._setMobileInit();
		}

	}

	this._recalculatePositions();
	this._setCurrentUI();

};

/**
 * Hanndle toggle button click action.
 * @param event {jQuery.Event} The original click-event object.
 * @private
 */
Nbc.prototype._handleToggleButtonClick = function(event) {

	if (event !== null) {
		event.preventDefault();
	}

	this.$element.toggleClass(this.options.stateClasses.isExpanded);

	if (!$body.hasClass(this.options.stateClasses.bodyOverflow) && this.$element.hasClass(this.options.stateClasses.isExpanded)) {

		this._blockScrolling(true);
		this.$element.on('click.' + this.uuid, this.options.domSelectors.link, this._handleMobileStickCollapsibleLink.bind(this));

		$document.on('keyup.' + this.uuid, this._handleKeyboard.bind(this));

		if (mediaqueries.query({ to: 'tabletPortrait' })) {

			this.$itemContent.css({
				height: 0
			});
		}

		// add gradient click handler
		this.$element.on('click.' + this.uuid, this._setGradientClick.bind(this));

	} else {

		this.$item.removeClass(this.options.stateClasses.isExpanded);
		this.$toggleButton.removeClass(this.options.stateClasses.isExpanded);

		this._blockScrolling(false);

		setTimeout(function() {
			this._resetMobileSettings(true);
		}.bind(this), 250);

		this.$element.off('click.' + this.uuid, this.options.domSelectors.link);

		$document.off('keyup.' + this.uuid);

		if (mediaqueries.query({ to: 'tabletPortrait' })) {

			this.$itemContent.css({
				height: 0
			});

		}

		// remove gradient click handler
		this.$element.off('click.' + this.uuid);
	}

	this._setCurrentUI();

};

/**
 * Check for forsake click on tablet/mobile
 * @private
 */
Nbc.prototype._setGradientClick = function(e) {

	// set exit classes on different views
	if (mediaqueries.query({from: 'tabletPortrait'})) {

		// Tablet
		this.forsakeClasses = [
			'mod_nbc'
		];
	} else {

		// Mobile
		this.forsakeClasses = [
			'mod_nbc_scroll',
			'box'
		];
	}

	// close module if click target with matching class (click on gradient background)
	for (var i = 0, l = this.forsakeClasses.length; i < l; i++) {
		if ($(e.target).hasClass(this.forsakeClasses[i])) {
			// call module close
			this._handleToggleButtonClick(null);
			break;
		}
	}
};

/**
 * Adding events handlers for desktop only.
 * @private
 */
Nbc.prototype._setDesktopListeners = function() {
	this.data.heightChangeListen = true;

	this.$element.on('click.' + this.uuid, this.options.domSelectors.link, this._handleCollapsibleLink.bind(this));
	$window.on('scroll.' + this.uuid, this._handleDesktopScrollAction.bind(this));
	$document.on(globalEvents.resize.key + '.' + this.uuid, this._handleResizeChange.bind(this));

};

/**
 * Adding evnets handlers for mobile only.
 * @private
 */
Nbc.prototype._setMobileListeners = function() {
	this.data.heightChangeListen = true;

	if (this.options.stickerSettings.toggleFunctionalityOnMobile) {
		$window.on('scroll.' + this.uuid, this._handleMobileScrollAction.bind(this));
		$document.on('click.' + this.uuid, this.options.domSelectors.toggleButton, this._handleToggleButtonClick.bind(this));
	}

	if (!this.options.stickerSettings.toggleFunctionalityOnMobile) {
		this.data.heightChangeListen = false;
		this.$element.on('click.' + this.uuid, this.options.domSelectors.link, this._handleMobileCollapsibleLink.bind(this));
	}
};

/**
 * Removing events handlers.
 * @private
 */
Nbc.prototype._disableListeners = function() {
	this.$element.off('click.' + this.uuid, this.options.domSelectors.link);
	$window.off('scroll.' + this.uuid);
	$document.off(globalEvents.resize.key + this.uuid);
	$document.off('click.' + this.uuid, this.options.domSelectors.toggleButton);
};

/**
 * Function is setting gutterTop and gutterBottom values depending from viewport size.
 * @private
 */
Nbc.prototype._setProperGutterValues = function() {

	if (mediaqueries.query({from: 'desktop5k'})) {

		this.options.stickerSettings.gutterTop = this.options.stickerSettings.defaultTopGutter * 2;
		this.options.stickerSettings.gutterBottom = this.options.stickerSettings.defaultBottomGutter * 2;
		this.options.stickerSettings.itemSpace = this.options.stickerSettings.defaultItemSpace * 2;

	} else {

		if (mediaqueries.query({from: 'desktop4k'})) {

			this.options.stickerSettings.gutterTop = this.options.stickerSettings.defaultTopGutter * 1.5;
			this.options.stickerSettings.gutterBottom = this.options.stickerSettings.defaultBottomGutter * 1.5;
			this.options.stickerSettings.itemSpace = this.options.stickerSettings.defaultItemSpace * 1.5;

		} else {

			this.options.stickerSettings.gutterTop = this.options.stickerSettings.defaultTopGutter;
			this.options.stickerSettings.gutterBottom = this.options.stickerSettings.defaultBottomGutter;
			this.options.stickerSettings.itemSpace = this.options.stickerSettings.defaultItemSpace;

		}

	}

};

/**
 * Recalculate values for container and item.
 * Function triggered on window resize.
 * @private
 */
Nbc.prototype._recalculatePositions = function() {
	var leftPadding = this.$elementParent.css('padding-left');

	this.data.windowHeight = window.innerHeight ? window.innerHeight : $window.height();

	this.$elementClone.css({
		'position': 'absolute',
		'top': 0,
		'left': leftPadding,
		'visibility': 'hidden',
		'width': 'calc(100% - ' + leftPadding + ')'
	});

	this.$elementParent.append(this.$elementClone).css({
		position: 'relative'
	});

	this._getContainerSize(this.$elementClone);

	this.$cloneElementItem.each(this._updateItemData.bind(this));

	this.$elementClone.remove();

	this.$elementParent.css({
		position: 'inherit'
	});

};

/**
 * Function setting proper position of container and height of item elements.
 * @private
 */
Nbc.prototype._handleDesktopScrollAction = function() {

	this.data.scrollTop = $window.scrollTop();
	this._setCurrentUI();

};

/**
 * Handle scroll action on mobile devices
 * @private
 */
Nbc.prototype._handleMobileScrollAction = function() {

	// determine the top offset value from nbc's parent on every scroll iteration
	var parentOffsetTop = this.$elementParent.offset().top,
		scrollLimit = parentOffsetTop - this.data.windowHeight;

	this.data.scrollTop = $window.scrollTop();

	if (this.data.scrollTop >= scrollLimit) {

		if (this.$element.hasClass(this.options.stateClasses.varMobile)) {
			this.$itemContent.css({
				height: 'auto'
			});
			this.$element
			.removeClass(this.options.stateClasses.varMobile)
			.removeClass(this.options.stateClasses.hasAnimation);
		}
	} else {
		if (!this.$element.hasClass(this.options.stateClasses.varMobile)) {

			this.$element.addClass(this.options.stateClasses.varMobile);
			setTimeout(function() {
				this.$element.addClass(this.options.stateClasses.hasAnimation);

				if (mediaqueries.query({ to: 'tabletPortrait' })) {
					this.$itemContent.css({
						height: 0
					});
				}
			}.bind(this), 250);
		}
	}
};

/**
 * Setting init data. Getting container default data, getting init data for items.
 * After getting init data trigger setCurrentUI function.
 * @private
 */
Nbc.prototype._setInitData = function() {
	this.data.windowHeight = window.innerHeight ? window.innerHeight : $window.height();

	this.$elementClone = this.$element.clone();
	this.$cloneElementItem = this.$elementClone.find(this.options.domSelectors.item);

	this._getContainerSize(this.$element);
	this.$item.each(this._getItemData.bind(this));

	this.forsakeClasses = [];

	this._setProperGutterValues();
	this._setCurrentUI();

};

/**
 * Toggles overflow on body to prevent scrolling and hide scroll if one exists
 * @param disableScroll {boolean} true if scroll should be disabled.
 * @private
 */
Nbc.prototype._blockScrolling = function(disableScroll) {
	var bodyPosition = $window.scrollTop();

	clearTimeout(this.options.blockScrollTimeout);
	this.options.blockScrollTimeout = setTimeout(function() {
		if (disableScroll) {
			window.requestAnimationFrame(function() {
				$body.addClass(this.options.stateClasses.bodyOverflow);
				if (bodyPosition !== 0) {
					$body.data(this.options.dataBodyPosition, bodyPosition);
					gsap.set($body, {top: -bodyPosition});
				}
			}.bind(this));
		} else {
			window.requestAnimationFrame(function() {
				$body.removeClass(this.options.stateClasses.bodyOverflow);
				if ($body.data(this.options.dataBodyPosition)) {
					gsap.set($body, {top: 0});
					$window.scrollTop($body.data(this.options.dataBodyPosition));
					$body.removeData(this.options.dataBodyPosition);
				}
			}.bind(this));
		}
	}.bind(this), 300);
};

/**
 * Set current container position
 * Update item height / position
 * @private
 */
Nbc.prototype._setCurrentUI = function() {

	if (!this.data.containerData.state.isFixed) {
		this._updateParentContainerHeight();
	} else {
		this._updateNbcItemsDisplay();
		this._checkScrapValue();
	}

	this._checkContainerState();
};

/**
 * Loop over all NBC items to update their appearance
 * @private
 */
Nbc.prototype._updateNbcItemsDisplay = function() {
	this.$item.each(this._updateItemDisplay.bind(this));
};

/**
 * Update item height and top margin. Function triggered on scroll action.
 * @param item
 * @param index
 * @private
 */
Nbc.prototype._updateItemDisplay = function(index) {

	this._updateItemHeight(index);
	this._updateItemTop(index);

};

/**
 * Update item vertical position. If it's necessary change margin-top.
 * @param index - index of item.
 * @private
 */
Nbc.prototype._updateItemTop = function(index) {
	var $el = this.$item.eq(index);

	if (this.data.scrollTop > this.data.itemsData[index].stickPosition && index > 0) {
		this.data.itemsData[index].top = Math.min(Math.round(this.data.scrollTop - this.data.itemsData[index].stickPosition), this.data.itemsData[index].maxTopMarginMove);
	} else {
		this.data.itemsData[index].top = 0;
	}

	requestAnimationFrame(function() {
		$el.css({
			'margin-top': -this.data.itemsData[index].top
		});
	}.bind(this));
};

/**
 * Updating item current height. Height is updated only when there is no user interaction.
 * When user trigger open action then height isn't changing until item will back to default state or uses trigger close action.
 * @param index - index of item
 * @private
 */
Nbc.prototype._updateItemHeight = function(index) {
	var $item = this.$item.eq(index),
		$content = $item.find(this.options.domSelectors.itemContent),
		visibleContent = this.data.scrollTop > this.data.itemsData[index].stickPosition + this.data.itemsData[index].maxTopMarginMove;

	if (visibleContent) {
		// set height proportional to scrolled page

		this.data.itemsData[index].content.currentHeight = Math.max(0, Math.round(this.data.itemsData[index].content.height + this.data.itemsData[index].stickPosition + this.data.itemsData[index].maxTopMarginMove - this.data.scrollTop));
		this.data.itemsData[index].currentHeight = Math.round(this.data.itemsData[index].height - this.data.itemsData[index].content.height + this.data.itemsData[index].content.currentHeight);

		requestAnimationFrame(function() {
			this.$item.eq(index).addClass(this.options.stateClasses.isScrolled);
		}.bind(this));

		if (!this.data.itemsData[index].state.isExpanded) {
			$document.trigger(events.recountTotalPrice);
		}

	} else {
		// set default content height

		this.data.itemsData[index].content.currentHeight = Math.round(this.data.itemsData[index].content.height);
		this.data.itemsData[index].currentHeight = Math.round(this.data.itemsData[index].height - this.data.itemsData[index].content.height + this.data.itemsData[index].content.currentHeight);

		requestAnimationFrame(function() {
			this.$item.eq(index).removeClass(this.options.stateClasses.isScrolled);
			this.$item.eq(index).removeClass(this.options.stateClasses.isExpanded);
		}.bind(this));

		if (this.data.itemsData[index].state.isExpanded) {
			this._updateScrapObjectPosition();
		}

		// reset user open data
		this.data.itemsData[index].state.isExpanded = false;

	}

	if (!this.data.itemsData[index].state.isExpanded) {

		if (this.inDesktopViewport) {

			requestAnimationFrame(function() {
				$content.css({
					height: this.data.itemsData[index].content.currentHeight
				});
			}.bind(this));

			if (!this.$item.hasClass(this.options.stateClasses.isSet)) {
				requestAnimationFrame(function() {
					this.$item.addClass(this.options.stateClasses.isSet);
				}.bind(this));
			}

		} else {
			requestAnimationFrame(function() {
				$content.css({
					height: 'auto'
				});
			}.bind(this));
		}
	}
};

/**
 * Update scrap object position. Reducing margin top when it's not needed.
 * @private
 */
Nbc.prototype._updateScrapObjectPosition = function() {

	if (this.data.scrapValue > 0 && this.data.scrapObjectTopMargin > 0) {
		this.data.scrapObjectTopMargin = 0;
	}

};

/**
 * Get init data for item
 * @param item
 * @param index
 * @private
 */
Nbc.prototype._getItemData = function(index, item) {
	var itemData;

	itemData = this._getItemDataObject(item, index);

	itemData.state = {
		isFixed: false,
		isUserOpen: false,
		isExpanded: false
	};

	this.data.itemsData.push(itemData);

};

/**
 * Update item data. Function triggered on window resize event.
 * @param item
 * @param index
 * @private
 */
Nbc.prototype._updateItemData = function(index, item) {
	var itemData;

	itemData = this._getItemDataObject(item, index);
	$.extend(this.data.itemsData[index], itemData);

};

/**
 * Get init data for item
 * @param item
 * @param index
 * @returns itemData {Object} - itemData object
 * @private
 */
Nbc.prototype._getItemDataObject = function(item, index) {

	var $el = $(item),
		elHeight = $el.outerHeight(),
		$content = $el.find(this.options.domSelectors.itemContent),
		contentHeight = $content.css({ height: 'auto' }).outerHeight() || 0, // we use 0 as a fallback in cases where a nbc box does not have a content element
		stickPosition,
		maxTopMarginMove,
		itemData;

	if (index === 0) {

		stickPosition = $el.offset().top - this.options.stickerSettings.gutterTop;
		maxTopMarginMove = 0;

	} else {

		stickPosition = this.data.itemsData[index - 1].stickPosition + this.data.itemsData[index - 1].content.height + this.data.itemsData[index - 1].maxTopMarginMove + this.options.stickerSettings.itemsBorderWidth + this.options.stickerSettings.itemSpace;
		maxTopMarginMove = Math.round(this.$item.eq(index - 1).css('margin-bottom').replace(',', '.').replace('px', '') * 1 + this.options.stickerSettings.itemsBorderWidth - this.options.stickerSettings.itemSpace);

	}

	itemData = {
		height: elHeight,
		currentHeight: elHeight,
		maxTopMarginMove: maxTopMarginMove,
		stickPosition: stickPosition,
		content: {
			height: contentHeight,
			currentHeight: contentHeight
		}
	};

	return itemData;
};

/**
 * Get init container sizes and position.
 * @private
 */
Nbc.prototype._getContainerSize = function($container) {
	var itemTop = $container.offset().top,
		containerHeight = $container.outerHeight();

	this.data.containerData = {
		top: itemTop,
		left: Math.round($container.offset().left),
		width: $container.outerWidth(),
		parentHeight: Math.round(this.$elementParent.outerHeight() - this.options.stickerSettings.gutterTop),
		scrolledToStick: itemTop - this.options.stickerSettings.gutterTop,
		scrapValue: this.$scrapObject.offset().top + this.$scrapObject.outerHeight() - this.options.stickerSettings.gutterTop - containerHeight,
		state: {
			isFixed: false
		}
	};

};

/**
 * Update the height of the nbc parent container
 */
Nbc.prototype._updateParentContainerHeight = function() {

	if (this.inDesktopViewport || this.options.stickerSettings.toggleFunctionalityOnMobile) {

		window.requestAnimationFrame(function() {
			this.$elementParent.css({
				'min-height': this.data.containerData.parentHeight + this.options.stickerSettings.gutterTop
			});

		}.bind(this));

	}
};

/**
 * Update container stick position. Set correct size, translate if fixed container should be pushed form stage.
 * @private
 */
Nbc.prototype._checkContainerState = function() {

	if (this.inDesktopViewport && this.data.scrollTop >= Math.ceil(this.data.containerData.scrolledToStick)) {
		this._setFixedContainerState();
		this._checkIfNegativeTopOffsetIsNecessary();
	} else {
		this._revokeFixedContainerState();
	}

};

/**
 * Set Fixed Container State
 * @private
 */
Nbc.prototype._setFixedContainerState = function() {

	this.data.containerData.state.isFixed = true;

	requestAnimationFrame(function() {
		this.$element
		.addClass(this.options.stateClasses.isFixed)
		.css({
			top: this.options.stickerSettings.gutterTop,
			left: this.data.containerData.left,
			width: this.data.containerData.width
		});
	}.bind(this));

};

/**
 * Revoke Fixed Container State
 * @private
 */
Nbc.prototype._revokeFixedContainerState = function() {

	this.data.containerData.state.isFixed = false;

	if (this.$element.hasClass(this.options.stateClasses.isFixed)) {
		requestAnimationFrame(function() {
			this.$element
			.removeClass(this.options.stateClasses.isFixed)
			.css({
				top: 'auto',
				left: 'auto',
				width: 'auto'
			});
		}.bind(this));
	}

};

/**
 * Check if we need to adjust the top position of the
 * fixed container to be able to display all items
 * @private
 */
Nbc.prototype._checkIfNegativeTopOffsetIsNecessary = function() {

	if (this.data.scrapValue < 0) {

		requestAnimationFrame(function() {
			gsap.set(this.$element, {
				top: this.data.scrapValue + this.options.stickerSettings.gutterTop
			});
		}.bind(this));

	} else {
		requestAnimationFrame(function() {
			gsap.set(this.$element, {
				top: this.options.stickerSettings.gutterTop
			});
		}.bind(this));
	}

};

Nbc.prototype._trackAccordionAction = function(action, item, title) {

	if (window.tracking) {
		window.tracking.dataLayerApi.trackAccordionAction({
			action: action,
			component: item[0],
			headline: this.moduleTitle,
			label: title
		});
	}
};

/**
 * Toggle open/close item. Function triggered on click.
 * @param itemIndex
 * @private
 */
Nbc.prototype._toggleSection = function(itemIndex) {

	if (this.data.itemsData[itemIndex].state.isExpanded) {
		this._contractItem(itemIndex);
	} else {
		this._expandItem(itemIndex);
	}

};

/**
 * Contracting item content.
 * @param itemIndex - item index
 * @private
 */
Nbc.prototype._contractItem = function(itemIndex) {
	var $item = this.$item.eq(itemIndex),
		title = this.$itemTitles.eq(itemIndex).attr('aria-label'),
		$content = this.$itemContent.eq(itemIndex);

	if ($item.hasClass(this.options.stateClasses.isScrolled)) {
		this._trackAccordionAction('close', $item, title);
	}

	if (this.data.itemsData[itemIndex].state.isExpanded) {
		this._updateFooterPosition(itemIndex, false);
		this.data.heightChangeListen = false;
	}

	gsap.to($content, {
		duration: 0.3,
		'height': this.data.itemsData[itemIndex].content.currentHeight,
		onComplete: function() {
			$item.removeClass(this.options.stateClasses.isExpanded);
			$document.trigger(events.recountTotalPrice);
			this.data.heightChangeListen = true;
		}.bind(this)
	});

	this.data.itemsData[itemIndex].state.isExpanded = false;

};

/**
 * Expanding item content.
 * @param itemIndex - item index
 * @private
 */
Nbc.prototype._expandItem = function(itemIndex) {
	var $item = this.$item.eq(itemIndex),
		title = this.$itemTitles.eq(itemIndex).attr('aria-label'),
		$content = this.$itemContent.eq(itemIndex);

	this._contractAllOpenedItems();
	this._updateFooterPosition(itemIndex, true);

	if ($item.hasClass(this.options.stateClasses.isScrolled)) {
		this._trackAccordionAction('open', $item, title);
	}

	gsap.to($content, {
		duration: 0.3,
		'height': this.data.itemsData[itemIndex].content.height,
		delay: 0.1,
		onComplete: function() {
			$item.addClass(this.options.stateClasses.isExpanded);
			$document.trigger(events.recountTotalPrice);
		}.bind(this)
	});

	this.data.itemsData[itemIndex].state.isExpanded = true;

};

/**
 * Contracting all expanded items content.
 * @private
 */
Nbc.prototype._contractAllOpenedItems = function() {

	this.$item.each(function(index) {
		if (this.data.itemsData[index].state.isExpanded) {
			this._contractItem(index);
		}
	}.bind(this));

};

/**
 * Updating footer position. Function is used when page is scrolled to bottom and used extending item content.
 * In this case we need do some space for sticked module.
 * @param itemIndex
 * @param open
 * @private
 */
Nbc.prototype._updateFooterPosition = function(itemIndex, open) {
	var vector = (open) ? 1 : -1,
		heightDifference = this.data.itemsData[itemIndex].content.height - this.data.itemsData[itemIndex].content.currentHeight,
		move;

	if (this.data.scrapValue <= heightDifference) {
		this.data.scrapObjectTopMargin = Math.max(this.data.scrapObjectTopMargin + vector * heightDifference, 0);
		this.data.heightChangeListen = false;

		if (vector === 1) {
			move = this.data.containerData.parentHeight + this.options.stickerSettings.gutterTop + heightDifference - this.data.scrapValue;

			if (this.data.scrapValue < 0) {
				move = move + this.data.scrapValue;
			}

		} else {
			move = this.data.containerData.parentHeight + this.options.stickerSettings.gutterTop;
		}

		gsap.to(this.$scrapObject, {
			duration: 0.3,
			'min-height': move,
			onComplete: function() {

				this._setCurrentUI();

			}.bind(this)
		});
	}

};

/**
 * Handles the click-event on the link and handles the toggle of the section. For desktop only.
 * @param event {jQuery.Event} The original click-event object.
 * @private
 */
Nbc.prototype._handleCollapsibleLink = function(event) {
	event.preventDefault();

	var $link = $(event.target),
		$item = $link.closest(this.options.domSelectors.item),
		itemIndex = this.$item.index($item);

	this._toggleSection(itemIndex);
};

/**
 * Handles the click-event on the link. For mobile with sticky functionality activated only.
 * @param event {jQuery.Event} The original click-event object.
 * @private
 */
Nbc.prototype._handleMobileStickCollapsibleLink = function(event) {

	if (!mediaqueries.query({ to: 'tabletPortrait' })) {
		return;
	}

	event.preventDefault();

	var $link = $(event.target),
		$item = $link.closest(this.options.domSelectors.item),
		title = $link.attr('aria-label'),
		$content = $item.find(this.options.domSelectors.itemContent),
		$itemSiblingsContent = this.$itemContent.not($content),
		itemTop,
		headerHeight = $item.find(this.options.domSelectors.title).outerHeight() + 1,
		maxHeight = this.data.windowHeight + 1 - headerHeight - $(this.options.domSelectors.buttonContainer).outerHeight() - $(this.options.domSelectors.header).outerHeight(),
		currentIndex = this.$item.index($item);

	if ($item.hasClass(this.options.stateClasses.isExpanded)) {

		gsap.to($content, { duration: 0.4, 'height': '0', ease: 'Power1.easeInOut' });
		gsap.to(this.$item, {
			duration: 0.4,
			y: 0,
			ease: 'Power1.easeInOut',
			onComplete: function() {
				this.$element.removeClass(this.options.stateClasses.isOverflowDisabled);
				$item.removeClass(this.options.stateClasses.isExpanded);
			}.bind(this)
		});

		this._trackAccordionAction('close', $item, title);

	} else {
		itemTop = $item.position().top;

		this.$element.addClass(this.options.stateClasses.isOverflowDisabled);

		$item.addClass(this.options.stateClasses.isExpanded);

		$itemSiblingsContent.removeClass(this.options.stateClasses.isExpanded);

		gsap.set($content, { 'height': 'auto' });
		gsap.from($content, { duration: 0.4, 'height': '0', ease: 'Power1.easeInOut', delay: 0 });

		this.$item.each(function(index, item) {
			if (index < currentIndex) {

				gsap.to(item, { duration: 0.4, y: -1 * itemTop, ease: 'Power1.easeInOut' });

			} else {
				if (index > currentIndex) {

					gsap.to(item, { duration: 0.4, y: this.data.windowHeight - headerHeight + itemTop, ease: 'Power1.easeInOut'});

				} else {

					gsap.to(item, { duration: 0.4, y: this._getItemRangeHeaderHeight(currentIndex + 1, this.$item.length), ease: 'Power1.easeInOut'});
					gsap.set($content, {'max-height': maxHeight});

				}
			}
		}.bind(this));

		this._trackAccordionAction('open', $item, title);

	}

};

/**
 * Handles the click-event on the link. For mobile without sticky functionality activated only.
 * @param event {jQuery.Event} The original click-event object.
 * @private
 */
Nbc.prototype._handleMobileCollapsibleLink = function(event) {
	event.preventDefault();

	var $link = $(event.target),
		$item = $link.closest(this.options.domSelectors.item),
		$content = $item.find(this.options.domSelectors.itemContent);

	if ($item.hasClass(this.options.stateClasses.isExpanded)) {

		gsap.to($content, {
			duration: 0.4,
			'height': '0',
			ease: 'Power1.easeInOut',
			onComplete: function() {
				$item.removeClass(this.options.stateClasses.isExpanded);
				$document.trigger(events.recountTotalPrice);
			}.bind(this)
		});

	} else {

		$item.addClass(this.options.stateClasses.isExpanded);

		gsap.set($content, { 'height': 'auto' });
		gsap.from($content, {
			duration: 0.4,
			'height': '0',
			ease: 'Power1.easeInOut',
			delay: 0.1,
			onComplete: function() {
				$document.trigger(events.recountTotalPrice);
			}.bind(this)
		});

	}

};

/**
 * Checking is container should be pushed up by scrapObject. If it should be, then calculate how much.
 * @private
 */
Nbc.prototype._checkScrapValue = function() {

	this.data.containerData.scrapValue = this.$scrapObject.offset().top + this.$scrapObject.outerHeight() - this.options.stickerSettings.gutterTop - this._getAllItemsCurrentHeight() - this.options.stickerSettings.gutterBottom;
	this.data.scrapValue = Math.round(this.data.containerData.scrapValue - this.data.scrollTop);

};

/**
 * Get current container current height.
 * @returns {number}
 * @private
 */
Nbc.prototype._getAllItemsCurrentHeight = function() {
	var returnHeight = 0;

	this.$item.each(function(index) {
		if (this.data.itemsData[index].state.isExpanded) {
			returnHeight += this.data.itemsData[index].height;
		} else {
			returnHeight += this.data.itemsData[index].currentHeight;
		}

	}.bind(this));

	return returnHeight;
};

/**
 *
 * @param start {Integer} - An integer indicating the 0-based position at which the elements begin to be selected. If negative, it indicates an offset from the end of the set.
 * @param count {Integer} - An integer indicating the 0-based position at which the elements stop being selected. If negative, it indicates an offset from the end of the set.
 * @returns {number}
 * @private
 */
Nbc.prototype._getItemRangeHeaderHeight = function(start, count) {
	var returnHeight = 0;

	this.$item.slice(start, count).each(function(index, item) {

		returnHeight += $(item).find(this.options.domSelectors.title).outerHeight() + 1;

	}.bind(this));

	return returnHeight;
};

/**
 * Reset all desktop css settings. Reset class states for each elements.
 * @private
 */
Nbc.prototype._resetSettings = function() {

	requestAnimationFrame(function() {

		this.$element.removeClass(this.options.stateClasses.isFixed);

		gsap.set(this.$element, { clearProps: 'top,left,width' });
		gsap.set(this.$elementParent, { clearProps: 'min-height' });

		gsap.set(this.$element.find(this.options.domSelectors.itemContent), { clearProps: 'height' });

	}.bind(this));

	this.$item.each(function(index, el) {
		var $el = $(el);

		requestAnimationFrame(function() {
			gsap.set($el, { clearProps: 'margin-top,height' });
			$el
			.removeClass(this.options.stateClasses.isSet)
			.removeClass(this.options.stateClasses.isScrolled);
		}.bind(this));

	}.bind(this));

};

/**
 * Reset all mobile settings.
 * @private
 */
Nbc.prototype._resetMobileSettings = function(resetFromToggleClosing) {

	window.requestAnimationFrame(function() {

		if (!resetFromToggleClosing) {
			this.$element
			.removeClass(this.options.stateClasses.varMobile)
			.removeClass(this.options.stateClasses.isExpanded)
			.removeClass(this.options.stateClasses.hasAnimation)
			.removeClass(this.options.stateClasses.isOverflowDisabled);
		}

		$document.off('keyup.' + this.uuid);

		this._blockScrolling(false);

		this.$elementParent.removeAttr('style');

		gsap.set(this.$item, {
			clearProps: 'transform'
		});

	}.bind(this));

};

Nbc.prototype._handleKeyboard = function(event) {
	event.preventDefault();
	if (contains(this.options.keys.close, event.keyCode)) {
		this._handleToggleButtonClick(null);
	}
};

module.exports = Nbc;


/***/ }),
/* 175 */,
/* 176 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(177);


/***/ }),
/* 177 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 178 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(179);


/***/ }),
/* 179 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M19 - NBC
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(180);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	Nbc = __webpack_require__(174);

var name = 'nbc',
	events = {},
	defaults = {
		domSelectors: {
			item: '[data-' + name + '="item"]',
			buttonContainer: '[data-' + name + '="button_container"]'
		},
		stateClasses: {},
		ids: {
			isInAemEditor: 'aemauthor'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype, Nbc.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	var hasContent = this._hasContent();

	if (!hasContent) {
		this._cleanUp();
		return;
	}

	Nbc.prototype.init.apply(this);
};

// Check for content. If there is no content, we remove the divs
// ['data-init="nbc"'] and ['data-nbc="button_container"'].
// Return true if we have content, otherwise return false,
// except for the case that we're in aem editor
Module.prototype._hasContent = function() {

	// if we're in aem, we can immediately return true
	if (this._isInAemEditor()) {
		return true;
	}

	// get the nbc item
	this.$item = this.$element.find(this.options.domSelectors.item);

	// if we don't have items at all, we can return anyway
	if (!this.$item) {
		return false;
	}

	return this.$item.length > 0 ? true : false;
};

// If an empty page in aem is created, the nbc will also be empty. In
// this case we must not remove the nbc. Otherwise editors could not
// add content to the nbc
Module.prototype._isInAemEditor = function() {

	// if the page is inside aem editor, we will find an empty span with the
	// id aemauthor inside the body.
	var aemSpan = document.getElementById(this.options.ids.isInAemEditor);

	if (aemSpan) {
		return true;
	}

	return false;
};

// Clean up routine for the case when we don't have content in the nbc
Module.prototype._cleanUp = function() {
	this.$element.remove();
	$(document).find(this.options.domSelectors.buttonContainer).remove();
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 180 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 181 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(182);


/***/ }),
/* 182 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 183 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(184);


/***/ }),
/* 184 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.14 - Timepicker
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(185);

var $ = __webpack_require__(0),
	$document = $(document),
	SuperClass = __webpack_require__(1);

var name = 'timepicker',
	events = {
		setCurrentTime: 'setCurrentTime.estatico.' + name
	},
	defaults = {
		domSelectors: {
			// item: '[data-' + name + '="item"]'
		},
		stateClasses: {
			// isActive: 'is_active'
		}
	},
	data = { /* items: ["Item 1", "Item 2"] */ };

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	this.setCurrentTime();

	// attach event handler for setting the current time
	$document.on(events.setCurrentTime, this.setCurrentTime.bind(this));
};

Module.prototype.setCurrentTime = function() {
	var currentDate = new Date(),
		currentHour,
		currentMinute;

	this.$input = this.$element.find('input');

	// select text when focusing input
	this.$input
		.on('focus.' + this.uuid, function(event) {
			$(event.currentTarget).select();
		}.bind(this));

	if (this.options.currentTime) {

		currentHour = currentDate.getHours();
		currentMinute = currentDate.getMinutes();

		if (currentHour < 10) {
			currentHour = '0' + currentHour;
		}

		if (currentMinute < 10) {
			currentMinute = '0' + currentMinute;
		}

		this.$input.val(currentHour + ':' + currentMinute);
	}
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 185 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 186 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(187);


/***/ }),
/* 187 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.15 - Toggle Button
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(188);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'togglebutton',
	events = {
		setValueActive: 'setValueActive.toggleButton'
	},
	defaults = {
		domSelectors: {
			inputRadios: 'input',
			checkedRadio: 'input[checked]',
			unCheckedRadio: 'input:not(:checked)'
		},
		keys: {
			space: 32
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	this.$radios = this.$element.find(this.options.domSelectors.inputRadios);

	if (this.options.conditionalContainer) {
		this.$radios.on('change.' + this.uuid, this._updateConditionalContent.bind(this));
	}

	this.$element.on('keyup.' + this.uuid, this._toggleSelection.bind(this));

	this.$element.on(events.setValueActive + '.' + this.uuid, this._setValueActive.bind(this));

};

/**
 * Update Conditional content
 */
Module.prototype._updateConditionalContent = function() {
	$('#' + this.options.conditionalContainer).toggleClass('var_hidden');
};

Module.prototype._setValueActive = function(_, value) {
	var activeValue = this.$element.find(this.options.domSelectors.checkedRadio).val();

	if (activeValue !== value) {
		this._toggleSelection();
	}
};

Module.prototype._toggleSelection = function(e) {

	if (e && e.keyCode !== this.options.keys.space) {
		return false;
	}

	var $unCheckedRadio = this.$element.find(this.options.domSelectors.unCheckedRadio),
		$checkedRadio = this.$element.find(this.options.domSelectors.checkedRadio);

	requestAnimationFrame(function() {
		$checkedRadio.removeAttr('checked');
		$unCheckedRadio.prop('checked', true);
	});

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 188 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 189 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(190);


/***/ }),
/* 190 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 191 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(192);


/***/ }),
/* 192 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.6.1 - Tabs
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(193);

var $ = __webpack_require__(0),
	contains = __webpack_require__(15),
	isArray = __webpack_require__(28),
	LockTabFocus = __webpack_require__(25),
	SuperClass = __webpack_require__(1);

var name = 'multiselect',
	$document = $(document),
	events = {
		closeMultiselect: 'closeMultiselect',
		triggerSelectionOnItem: 'multiselect.triggerSelectionOnItem',
		clearSelection: 'multiselect.clearSelection',
		multiSelectIsReady: 'multiselect.multiSelectIsReady',
		multiselectOptionIsSelected: 'multiselect.multiselectOptionIsSelected'
	},
	defaults = {
		domSelectors: {
			lightBoxContainer: '[data-lightbox="container"]',
			option: '[data-' + name + '="option"]',
			options: '[data-' + name + '="options"]',
			container: '[data-' + name + '="container"]',
			reset: '[data-' + name + '="reset"]',
			select: '[data-' + name + '="select"]',
			selection: '[data-' + name + '="selection"]',
			button: '[data-' + name + '="button"]'
		},
		stateClasses: {
			isExpanded: 'is_expanded',
			isDisabled: 'is_disabled'
		},
		keys: {
			close: [27],
			tab: [9]
		},
		dataAttributeForSelection: 'data-multiselect-values'
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Load required dependencies
 * @method
 * @public
 */
Module.prototype.init = function() {
	__webpack_require__.e/* require.ensure */(8).then((function(require) {
			this.template = __webpack_require__(175);
			this._init();
		}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);
};

/**
 * Initialize Selectors and trigger followup methods
 * @method
 * @private
 */
Module.prototype._init = function() {

	this.$container = this.$element.find(this.options.domSelectors.container);
	this.$select = this.$element.find(this.options.domSelectors.select);
	this.$button = this.$element.find(this.options.domSelectors.button);
	this.$selection = this.$element.find(this.options.domSelectors.selection);
	this.placeholder = this.$selection.eq(0).text();
	this.isUpdated = false;
	this.inputType = this.options.inputType || 'checkbox';

	this.$lightBoxContainer = this._hasLightBoxParentElement();

	this._parseData();

	this.$options = this.$element.find(this.options.domSelectors.options);
	this.$option = this.$element.find(this.options.domSelectors.option);
	this.$reset = this.$element.find(this.options.domSelectors.reset);

	if (this.$select.prop('disabled')) {
		this.$element.addClass(this.options.stateClasses.isDisabled);

		// Check for selected elements, even though user interaction is disabled, to display preselected
		// labels as selected in the dropdown
		this._validateSelection();

		// exit
		return;
	}

	this._bindEvents();
	this._validateSelection();

	$document.trigger(events.multiSelectIsReady);

};

Module.prototype._hasLightBoxParentElement = function() {
	return this.$element.closest(this.options.domSelectors.lightBoxContainer);
};

/**
 * Bind Events used in this module and/or around it
 * @method
 * @private
 */
Module.prototype._bindEvents = function() {

	this.$element
		.on('click.' + this.uuid, function(event) {
			event.stopPropagation();
		})

		.on('click.' + this.uuid, this.options.domSelectors.button, function() {

			if (this.$element.hasClass(this.options.stateClasses.isExpanded)) {
				this._closeOptions();
				return;
			}

			this._openOptions();

		}.bind(this))

		.on('change.' + this.uuid, this.options.domSelectors.option, function() {
			this.isUpdated = true;
			this._optionSelection();
		}.bind(this))

		.on('change.' + this.uuid, this.options.domSelectors.reset, function() {
			this.isUpdated = true;
			this._resetSelection();
		}.bind(this));

	// emitted by another instance of multiselect.
	$document.on(events.closeMultiselect, function(evt, data) {

		// we only want to handle this event if it has been triggered by another
		// instance of this Module
		if (data.$element === this.$element) {
			return;
		}

		// close the multiselect if it is open. We only want to have
		// one multiselect open at a time on the same page.
		if (this.$element.hasClass(this.options.stateClasses.isExpanded)) {
			this._closeOptions();
		}

	}.bind(this));

	this.registerSyncFormFieldListeners();

	// trigger item selection from outside
	this.$element.on(events.triggerSelectionOnItem + '.' + this.uuid, this.handleTriggerSelectionOnItem.bind(this));

	// clear selection from outside
	this.$element.on(events.clearSelection + '.' + this.uuid, function() {
		this.isUpdated = true;
		this._resetSelection();
		this.$option.prop('checked', false);
	}.bind(this));
};

Module.prototype.registerSyncFormFieldListeners = function() {
	if (!this.options.syncFormFieldId) {
		return;
	}

	var eventName = 'sync-form-field-changed.' + this.options.syncFormFieldId + '.' + this.uuid;

	$document.on(eventName, function(event, item) {
		var newValues;

		if (this.options.showLabel && item[this.options.showLabel]) {
			newValues = item[this.options.showLabel];
		} else {
			newValues = item.label;
		}

		if (!isArray(newValues)) {
			if (typeof newValues !== 'string') {
				return;
			}

			newValues = [newValues];
		}

		var selection = {
			labels: [],
			values: []
		};

		this.$element.find(this.options.domSelectors.option + ', ' + this.options.domSelectors.reset)
		.prop('checked', false)
		.each(function(index, checkbox) {
			var $checkbox = $(checkbox),
				checkboxValue = $checkbox.val();

			if (!contains(newValues, checkboxValue)) {
				return;
			}

			$checkbox.prop('checked', true);

			var label = $checkbox.attr('data-multiselect-label');

			selection.labels.push(label);
			selection.values.push(checkboxValue);
		});

		this._saveSelection(selection.values);
		this._updateSelection(selection.labels, selection.values);

		// Validate the new value and inform listeners of the `changedAndValidated` event.
		this.$select.trigger('change');
	}.bind(this));
};

/**
 * Iterate through the select object provided by backend
 * @method
 * @private
 */
Module.prototype._parseData = function() {

	var data = {
			id: this.$element.attr('id') || name,
			options: []
		};

	this.$select.children().each(function(index, item) {

		var optgroup;

		if ($(item).prop('tagName') === 'OPTION') {
			data.options.push(this._parseOption($(item)));
		}

		if ($(item).prop('tagName') === 'OPTGROUP') {
			// push option

			optgroup = {
				group: $(item).attr('label'),
				options: []
			};

			$(item).children('option').each(function(index, item) {
				optgroup.options.push(this._parseOption($(item)));
			}.bind(this));

			data.options.push(optgroup);
		}

	}.bind(this));

	this.$data = this.template(data);
	this.$container.append(this.$data);

};

/**
 * Iterate through options in a optgroup of the select
 * @method
 * @private
 * @param {jquery-object} $item - jquery object of <option>
 */
Module.prototype._parseOption = function($item) {
	var option = {
		value: $item.attr('value'),
		label: $item.text(),
		selected: $item.prop('selected'),
		disabled: $item.prop('disabled'),
		id: this.$element.attr('id') || name,
		type: this.inputType
	};

	return option;
};

/**
 * Iterate through all checked inputs and push the values to an array
 * @method
 * @private
 */
Module.prototype._validateSelection = function() {
	var selection = {
			labels: [],
			values: []
		};

	this.$option.filter(':checked').each(function() {
		var $option = $(this);

		selection.values.push($option.attr('value'));
		selection.labels.push($option.attr('data-multiselect-label'));

	});

	this.$reset.filter(':checked').each(function() {
		var $option = $(this);

		selection.values.push($option.attr('value'));
		selection.labels.push($option.attr('data-multiselect-label'));

	});

	this._saveSelection(selection.values);
	this._updateSelection(selection.labels, selection.values);
};

/**
 * Iterate through all checked option inputs and push the values to an array
 * uncheck reset option of validation is fine
 * @method
 * @private
 */
Module.prototype._optionSelection = function() {
	var selection = {
			labels: [],
			values: []
		};

	this.$option.filter(':checked').each(function() {
		var $option = $(this);

		selection.values.push($option.attr('value'));
		selection.labels.push($option.attr('data-multiselect-label'));

	});

	if (selection.values.length >= 1) {
		this.$reset.prop('checked', false);
	}

	this._saveSelection(selection.values);
	this._updateSelection(selection.labels, selection.values);

	$document.trigger(events.multiselectOptionIsSelected);
};

/**
 * Saves the selection to the select component
 * @method
 * @private
 * @param {array} values - array with checked values
 */
Module.prototype._saveSelection = function(values) {
	if (!values.length) {
		this.$select.find('option').prop('selected', false);
	} else {
		this.$select.find('option').each(function(index, item) {
			var $option = $(item),
				state = false;

			if (values.indexOf($option.attr('value')) > -1) {
				state = true;
			}

			$option.prop('selected', state);
		}.bind(this));
	}
};

/**
 * Replaces the Label value with the values in the passed array
 * @method
 * @private
 * @param {array} values - array with checked values
 */
Module.prototype._updateSelection = function(labels, values) {
	if (this.options.reportSelection) {
		if (this.options.inputType === 'radio') {
			this.$element.attr(this.options.dataAttributeForSelection, JSON.stringify(values));
		} else {
			if (this.options.reportSelectionValues) {
				this.$element.attr(this.options.dataAttributeForSelection, JSON.stringify(values));
			} else {
				this.$element.attr(this.options.dataAttributeForSelection, JSON.stringify(labels));
			}
		}
	}

	if (!labels.length) {
		labels = [this.placeholder];
		this.$selection.removeClass('is_active');
	} else {
		this.$selection.addClass('is_active');
	}

	this.$selection.text(labels.join(', '));
};

/**
 * Resets all checkboxes when reset option is clicked/checked and updates the
 * label
 * @method
 * @private
 */
Module.prototype._resetSelection = function() {
	var selection = {
			labels: [],
			values: []
		};

	if (this.$reset.is(':checked')) {
		this.$option.prop('checked', false);
		selection.values.push(this.$reset.attr('value'));
		selection.labels.push(this.$reset.attr('data-multiselect-label'));
	}

	this._saveSelection(selection.values);
	this._updateSelection(selection.labels, selection.values);
};

/**
 * Opens the optionslist and updates all required attributes
 * @method
 * @private
 */
Module.prototype._openOptions = function() {

	// for the case that other multiselects are on thepage, we trigger this
	// event so that we're able to close them.
	$document.trigger(events.closeMultiselect, this);

	this.isUpdated = false;

	this.$element.addClass(this.options.stateClasses.isExpanded);
	this.$button.attr('aria-expanded', 'true');
	this.$options.attr('aria-hidden', 'false');

	this.$container.css('max-height', this.$container.css('max-height') || 360);

	// Lock Tab Focus within the options container
	if (this.options.inputType !== 'radio') {

		if (this.$lightBoxContainer.length > 0) {
			LockTabFocus.destroy();
		}

		requestAnimationFrame(function() {
			LockTabFocus.setFocusOnElement(this.$container);
		}.bind(this));

	}

	// temporary events for open state;
	this.$element.on('keydown.' + this.uuid, this._handleKeyboard.bind(this));
	this.$container.on('keydown.' + this.uuid, this._handleTabKeyOnRadios.bind(this));

	$document
		.one('click.' + this.uuid, this._closeOptions.bind(this))
		.css({
			cursor: 'pointer' // Ridiculous CSS cursor change hack needed due to mobile safari: https://stackoverflow.com/questions/14795944/jquery-click-events-not-working-in-ios
		});
};

/**
 * Closes the optionslist and updates all required attributes
 * @method
 * @private
 */
Module.prototype._closeOptions = function() {

	this.$element.removeClass(this.options.stateClasses.isExpanded);
	this.$button.attr('aria-expanded', 'false');
	this.$options.attr('aria-hidden', 'true');

	if (this.isUpdated === true) {
		this.$select.trigger('change');
	}

	// remove temporal events
	this.$element.off('keydown.' + this.uuid);
	this.$container.off('keydown.' + this.uuid);

	$document
		.off('click.' + this.uuid)
		.css({
			cursor: 'default'
		});

	// Make sure that the tab focus restriction is given back to the lightbox
	if (this.options.inputType !== 'radio') {
		LockTabFocus.destroy();
	}

	if (this.$lightBoxContainer.length > 0) {
		LockTabFocus.setFocusOnElement(this.$lightBoxContainer, this.$button);
	} else {
		// Set the focus to the button so that the user can continue to navigate
		// using the keyboard.
		this.$button.focus();
	}

};

/**
 * Handles Close Action Keys
 * @method
 * @private
 */
Module.prototype._handleKeyboard = function(event) {

	if (contains(this.options.keys.close, event.keyCode)) {
		event.stopPropagation();
		this._closeOptions();
	}

};

Module.prototype._handleTabKeyOnRadios = function(event) {

	if (this.options.inputType === 'radio') {

		var tabKey = contains(this.options.keys.tab, event.keyCode);

		if (tabKey || tabKey + event.shiftKey) {
			event.preventDefault();
		}

	}

};

Module.prototype.handleTriggerSelectionOnItem = function(_, item) {
	var $checkbox,
		currentCheckedState;

	if (this.options.reportSelectionValues) {
		$checkbox = this.$options.find('[value="' + item + '"]');

	} else {
		$checkbox = this.$options.find('[data-' + name + '-label="' + item + '"]');
	}

	if ($checkbox.length !== 1) {
		return;
	}

	// trigger selection on checkbox
	currentCheckedState = $checkbox.prop('checked');
	$checkbox.prop('checked', !currentCheckedState);

	// update instance state
	this._optionSelection();

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	initEvents: ['ready', 'ajaxload'],
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 193 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 194 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(195);


/***/ }),
/* 195 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M18.2 - Media Gallery
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(196);

var $ = __webpack_require__(0),
	mediaqueries = __webpack_require__(6),
	keymaster = __webpack_require__(14),
	cssTransitionEnd = __webpack_require__(4),
	SuperClass = __webpack_require__(1),
	globalEvents = __webpack_require__(5);

var name = 'mediagallery',
	$document = $(document),
	isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints,
	events = {},
	defaults = {
		domSelectors: {
			imagesWrapper: '[data-' + name + '="image_wrapper"]',
			thumbnailButton: '[data-' + name + '="thumbnail_button"]',
			thumbnail: '[data-' + name + '="thumbnail"]',
			lightBoxButton: '[data-' + name + '="lightbox_button"]',
			imageCounter: '[data-' + name + '="image_counter"]',
			imageCounterAria: '[data-' + name + '="image_counter_aria"]',
			imageWrapper: '[data-' + name + '="image"]',
			image: '[data-media="image"]',
			mobileLoadMoreButton: '[data-' + name + '="mobile_load_more_button"]',
			prevImageArrow: '[data-' + name + '="goto_prev_slide"]',
			nextImageArrow: '[data-' + name + '="goto_next_slide"]',
			videoWrapper: '[data-media="video_wrapper"]'
		},
		classSelectors: {
			mediumImage: 'mod_media_gallery_image_wrapper'
		},
		stateClasses: {
			isSelected: 'is_selected',
			isVisible: 'is_visible',
			isHidden: 'is_hidden',
			inCalculation: 'in_calculation',
			hasVideo: 'has_video'
		},
		i18n: {
			counterLabel: 'Zeige Inhalt',
			counterOf: 'von',
			counter: '{0} / {1}'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	this.initialFadeIn = true;

	if (!this.options.activeItem) {
		this._resetIndexValues();
	} else {
		this.currentIndex = parseInt(this.options.activeItem, 10);
		this.previousIndex = this.currentIndex;
	}

	this.$thumbnails = this.$element.find(this.options.domSelectors.thumbnail);
	this.$thumbnailsButtons = this.$element.find(this.options.domSelectors.thumbnailButton);
	this.$lightBoxButtons = this.$element.find(this.options.domSelectors.lightBoxButton);
	this.$imagesWrapper = this.$element.find(this.options.domSelectors.imagesWrapper);
	this.$mediumImageWrapper = this.$element.find(this.options.domSelectors.imageWrapper);
	this.$mediumImages = this.$element.find(this.options.domSelectors.image);
	this.$imageCounterAria = this.$element.find(this.options.domSelectors.imageCounterAria);
	this.$imageCounter = this.$element.find(this.options.domSelectors.imageCounter);
	this.$mobileLoadMoreButton = this.$element.find(this.options.domSelectors.mobileLoadMoreButton);
	this.$prevArrow = this.$element.find(this.options.domSelectors.prevImageArrow);
	this.$nextArrow = this.$element.find(this.options.domSelectors.nextImageArrow);
	this.numberOfImages = this.$mediumImageWrapper.length;

	this._getHeightOfMediumImages();
	this._initializeMouseEvents();

	// Check for mobile viewport mode and it query returns true, build it instantly
	if (mediaqueries.query({ to: 'tabletPortrait' })) {
		this._buildGalleryMobileMode();
	} else {
		this._buildGalleryDesktopMode();
	}

	$document.on(mediaqueries.event.key, this._handleBreakpointChange.bind(this));
	$document.on(globalEvents.resize.key + '.' + this.uuid, this._handleResizeChange.bind(this));

};

Module.prototype._resetIndexValues = function() {
	this.currentIndex = 0;
	this.previousIndex = 0;
	this.initialCounterValue = 1;
};

/**
 * Build desktop viewport mode
 * @private
 */
Module.prototype._buildGalleryDesktopMode = function() {

	// Set mobile view flag to false
	this.mobileView = false;

	// Hide all images
	this.$mediumImageWrapper
		.removeClass(this.options.stateClasses.isVisible)
		.addClass(this.options.stateClasses.isHidden);

	// Hide load more button
	this.$mobileLoadMoreButton.addClass(this.options.stateClasses.isHidden);

	this._revertFakeImagesToGalleryButtons();
	this._setActiveThumbnail();
	this._initializeKeyboardEvents();

	// Show first image
	this._fadeInImage(this.currentIndex);

	// Update image counter
	this._setImageCounterToValue(this.currentIndex + 1);

};

/**
 * Build mobile viewport mode
 * @private
 */
Module.prototype._buildGalleryMobileMode = function() {

	var showFromTo;

	// Set mobile viewport flag
	this.mobileView = true;

	// Set default visible images number
	this.visibleImages = (this.numberOfImages >= 3) ? 3 : this.numberOfImages;

	// Check if we can show load more button
	if (this.numberOfImages > 3) {
		this.$mobileLoadMoreButton.removeClass(this.options.stateClasses.isHidden);
	}

	// Hide all images
	this.$mediumImageWrapper
		.removeClass(this.options.stateClasses.isVisible)
		.addClass(this.options.stateClasses.isHidden);

	$(this.$thumbnails[this.currentIndex])
		.removeClass(this.options.stateClasses.isSelected);

	this._convertGalleryButtonsToFakeImages();

	// Show first 3 images or less if there's less than 3 in total
	for (var i = 0; i < this.visibleImages; i++) {
		this._fadeInImage(i, false);
	}

	showFromTo = '1-' + this.visibleImages;

	// Update image counter
	this._setImageCounterToValue(showFromTo);
};

/**
 * Breakpoint change handler
 * @param event,
 * @param prevBreakpoint,
 * @param currentBreakpoint
 * @private
 */
Module.prototype._handleBreakpointChange = function() {

	this._resetIndexValues();

	if (mediaqueries.query({ to: 'tabletPortrait' }) && !this.mobileView) {
		this._buildGalleryMobileMode();
	} else if (mediaqueries.query({ from: 'tabletPortrait' }) && this.mobileView) {
		this._buildGalleryDesktopMode();
	}

};

Module.prototype._handleResizeChange = function() {
	this._getHeightOfMediumImages();
	this._setWrapperHeight();
};

/**
 * In small viewport we do not want to have a lightbox on click but
 * we still need to make the images focusable somehow. We therefore
 * change the role of the button to an img and hide the images inside
 * to avoid that the screen readers will read the alt attribute as well.
 */
Module.prototype._convertGalleryButtonsToFakeImages = function() {

	var i = this.numberOfImages;

	while (i--) {

		var $button = $(this.$lightBoxButtons[i]),
			$img = $(this.$mediumImages[i]),
			ariaLabel = $button.attr('aria-label'),
			alt = $img.attr('alt');

		$button.attr({
			'role': 'img',
			'aria-haspopup': 'false',
			'aria-label': alt,
			'data-aria-label': ariaLabel
		});

		$img.attr({
			'role': 'presentation',
			'aria-hidden': 'true'
		});

	}

};

Module.prototype._revertFakeImagesToGalleryButtons = function() {

	var i = this.numberOfImages;

	while (i--) {

		var $button = $(this.$lightBoxButtons[i]),
			$img = $(this.$mediumImages[i]),
			ariaLabel = $button.attr('data-aria-label');

		$button.attr({
			'aria-haspopup': 'true',
			'aria-label': ariaLabel
		}).removeAttr('role data-aria-label');

		$img.removeAttr('role aria-hidden');

	}

};

/**
 * Initialize keyboard events
 * @private
 */
Module.prototype._initializeKeyboardEvents = function() {
	// set the scope of the module to
	var moduleScope = this.name + '-' + this.uuid;

	this.$lightBoxButtons.on('focusin.' + this.uuid, function() {
		keymaster.setScope(moduleScope);
	}).on('focusout.' + this.uuid, function() {
		keymaster.setScope('all');
	});

	// handle events when appropriate
	keymaster('left, pageup', moduleScope, this._goToPrevImageCallback.bind(this));
	keymaster('right, pagedown', moduleScope, this._goToNextImageCallback.bind(this));
	keymaster('home', moduleScope, this._goToFirstImageCallback.bind(this));
	keymaster('end', moduleScope, this._goToLastImageCallback.bind(this));

	keymaster('esc', moduleScope, function() {

		if (this.options.isInLightbox) {
			return;
		}

		this._setActiveThumbnail();
		this.$thumbnailsButtons[this.currentIndex].focus();

	}.bind(this));

	// handle enter/space on medium size image
	keymaster('enter, space', moduleScope, function(event) {

		if (this.options.isInLightbox) {
			return;
		}

		// prevent page scroll if user presses space
		event.preventDefault();

		// open image in lightbox
		this._lightBoxTriggerCallback(event);

	}.bind(this));
};

Module.prototype._goToNextImageCallback = function(event) {
	event.preventDefault();

	this._toggleActiveImage('next');

};

Module.prototype._goToPrevImageCallback = function(event) {
	event.preventDefault();

	this._toggleActiveImage('prev');

};

Module.prototype._goToFirstImageCallback = function(event) {
	event.preventDefault();

	this._toggleActiveImage('first');

};

Module.prototype._goToLastImageCallback = function(event) {
	event.preventDefault();

	this._toggleActiveImage('last');

};

Module.prototype._lightBoxTriggerCallback = function(event) {

	if (this.mobileView) {
		return;
	}

	event.preventDefault();

	var source = this.options.lightboxUrl + '?index=' + this.currentIndex;

	$document.triggerHandler('lightboxJSTrigger', {
		type: 'url',
		source: source,
		options: {
			elementToFocusAfterClose: this.$lightBoxButtons[this.currentIndex],
			centerContentHorizontally: true,
			centerContentVertically: true
		}
	});

};

Module.prototype._thumbnailClickCallback = function(event) {

	var index = $(event.currentTarget).parent().index();

	if (index === this.currentIndex) {
		return;
	}

	if (this.currentIndex >= 0) {
		this.previousIndex = this.currentIndex;
		this._fadeOutImage(this.previousIndex);
	}

	this._fadeInImage(index);

	this._setActiveThumbnail();
	this._setImageCounterToValue(index + 1);

};

Module.prototype._showMoreImagesCallback = function(event) {
	event.preventDefault();

	var i = this.visibleImages,
		start = i,
		$elementToFocus = this.$lightBoxButtons.eq(this.visibleImages),
		len = (this.visibleImages + 3 < this.numberOfImages) ? this.visibleImages + 3 : this.numberOfImages,
		counterValue = '1-' + this.visibleImages;

	// Hide "load more" mobile button
	if (this.visibleImages + 3 >= this.numberOfImages) {
		this.$mobileLoadMoreButton.addClass(this.options.stateClasses.isHidden);
	}

	// Show next 3 images
	for (i; i < len; i++) {
		this._fadeInImage(i, i === start);
		this.visibleImages++;
	}

	$elementToFocus.one(cssTransitionEnd + '.' + this.uuid, function() {
		if (this.$element[0].contains(document.activeElement)) {
			$elementToFocus.focus();
		}
	}.bind(this));

	this._setImageCounterToValue(counterValue);

};

/**
 * Initialize mouse events
 * @private
 */
Module.prototype._initializeMouseEvents = function() {
	this.$mobileLoadMoreButton.on('click.' + this.uuid, this._showMoreImagesCallback.bind(this));
	this.$thumbnailsButtons.on('click.' + this.uuid, this._thumbnailClickCallback.bind(this));

	// Add medium images listener if gallery is not already embedded in lightbox
	if (!this.options.isInLightbox) {
		this.$lightBoxButtons.on('click.' + this.uuid, this._lightBoxTriggerCallback.bind(this));
	} else {
		this.$nextArrow.on('click.' + this.uuid, this._goToNextImageCallback.bind(this));
		this.$prevArrow.on('click.' + this.uuid, this._goToPrevImageCallback.bind(this));
	}
};

/**
 * Initialize touch events
 * @private
 */
Module.prototype._initializeTouchEvents = function() {

	if (!isTouchDevice) {
		return;
	}

	__webpack_require__.e/* require.ensure */(1).then((function(require) {
		__webpack_require__(22);
		this.$imagesWrapper.hammer()
			.on('swipeleft.' + this.uuid, this._goToNextImageCallback.bind(this))
			.on('swiperight.' + this.uuid, this._goToPrevImageCallback.bind(this));
	}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);

};

Module.prototype._getHeightOfMediumImages = function() {

	var i = this.numberOfImages;

	this.mediumImagesHeights = [];

	this.$imagesWrapper.addClass(this.options.stateClasses.inCalculation);

	while (i--) {
		this.mediumImagesHeights.push(this.$mediumImageWrapper.eq(i)[0].getBoundingClientRect().height);
	}

	this.$imagesWrapper.removeClass(this.options.stateClasses.inCalculation);

	this.mediumImagesHeights.reverse();

};

/**
 * Animates height of the images wrapper and then changes value to auto after animation is done
 * @param index
 * @private
 */
Module.prototype._setWrapperHeight = function() {

	if (mediaqueries.query({ from: 'tabletPortrait' })) {

		if (this.options.imageFreeRatio) {

			if (!this.options.isInLightbox) {
				return;
			}

			this.$mediumImages[0].onload = function() {
				requestAnimationFrame(function() {
					$(this.$mediumImages[0]).css({
						maxWidth: this.$mediumImages[0].naturalWidth
					});
				}.bind(this));
			}.bind(this);

		} else {
			requestAnimationFrame(function() {
				this.$imagesWrapper.css({
					height: this.mediumImagesHeights[this.currentIndex]
				});
			}.bind(this));
		}

	} else {
		requestAnimationFrame(function() {
			this.$imagesWrapper.removeAttr('style');
		}.bind(this));
	}

};

Module.prototype._toggleActiveImage = function(callBackName) {

	if (this.mobileView) {
		return;
	}

	var counterValue;

	this.previousIndex = this.currentIndex;

	switch (callBackName) {

		case 'next':
			this.currentIndex = (this.currentIndex < this.numberOfImages - 1) ? this.currentIndex + 1 : 0;
			counterValue = this.currentIndex + 1;
			break;

		case 'prev':
			this.currentIndex = (this.currentIndex > 0) ? this.currentIndex - 1 : this.numberOfImages - 1;
			counterValue = this.currentIndex + 1;
			break;

		case 'first':
			this.currentIndex = 0;
			counterValue = this.currentIndex + 1;
			break;

		case 'last':
			this.currentIndex = this.numberOfImages - 1;
			counterValue = this.numberOfImages;
			break;

	}

	this._setActiveThumbnail();
	this._fadeOutImage();
	this._fadeInImage(this.currentIndex);
	this._setImageCounterToValue(counterValue);

};

/**
 * Fade out currently visible image, return promise which will be resolved once animation is complete
 * Promise will return reference to faded image
 * @private
 */
Module.prototype._fadeOutImage = function() {
	var $elementToFadeOut = this.$mediumImageWrapper.eq(this.previousIndex);

	$elementToFadeOut
		.removeClass(this.options.stateClasses.isVisible)
		.attr({
			'aria-hidden': true,
			'role': 'presentation'
		})
		.one(cssTransitionEnd + '.' + this.uuid, function() {
			$elementToFadeOut.addClass(this.options.stateClasses.isHidden);
		}.bind(this));
};

/**
 * Fade in image on index
 * @private
 */
Module.prototype._fadeInImage = function(index, focus) {
	var i = index,
		$elementToFadeIn = this.$mediumImageWrapper.eq(i);

	$elementToFadeIn.removeClass(this.options.stateClasses.isHidden);

	this._setWrapperHeight();

	setTimeout(function() {
		requestAnimationFrame(function() {
			$elementToFadeIn.addClass(this.options.stateClasses.isVisible);

			if (this.initialFadeIn) {
				this.initialFadeIn = false;
			} else {
				$elementToFadeIn.one(cssTransitionEnd + '.' + this.uuid, function() {
					$elementToFadeIn
						.attr('aria-hidden', false)
						.removeAttr('role');

					if (focus !== false && this.$element[0].contains(document.activeElement)) {
						this.$lightBoxButtons[index].focus();
					}

				}.bind(this));
			}

		}.bind(this));
	}.bind(this), 0);

	this.currentIndex = index;

};

/**
 * Highlights thumbnail on given index
 * @param index {int}
 * @private
 */
Module.prototype._setActiveThumbnail = function() {

	if (!this.mobileView) {
		requestAnimationFrame(function() {
			$(this.$thumbnails[this.previousIndex]).removeClass(this.options.stateClasses.isSelected);
			$(this.$thumbnails[this.currentIndex]).addClass(this.options.stateClasses.isSelected);
		}.bind(this));
	}

};

/**
 * Sets image counter to given value
 * @param value {int}
 * @private
 */
Module.prototype._setImageCounterToValue = function(value) {

	var label = this.options.i18n.counter.replace('\{0\}', value),
		ariaLabel = this.options.i18n.counterLabel + ' ' + value + ' ' + this.options.i18n.counterOf + ' ' + this.numberOfImages;

	label = label.replace('\{1\}', this.numberOfImages);

	requestAnimationFrame(function() {
		this.$imageCounterAria.text(ariaLabel);
		this.$imageCounter.text(label.trim());
	}.bind(this));

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 196 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 197 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(198);


/***/ }),
/* 198 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M00 - Contract Selection
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2019.
 */



__webpack_require__(199);

var $ = __webpack_require__(0),
	cssTransitionEnd = __webpack_require__(4),
	SuperClass = __webpack_require__(1);

var name = 'contractselection',
	events = {},
	defaults = {
		domSelectors: {
			radioButtons: 'input[name="contract"]',
			mainRoleCheckbox: '[data-' + name + '="mainrolecheckbox"]',
			fieldsetSubmit: '[data-' + name + '="fieldsetsubmit"]'
		},
		stateClasses: {
			isVisible: 'is_visible',
			isMainRole: 'is_main_role',
			isChangingMainRole: 'is_changing_main_role'
		}
	},
	data = {
		button: {}
	};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

Module.prototype.init = function() {

	this._cacheElements();

	this.contractSelected = this.$element.find('input[name="contract"]:checked').length > 0;

	this._addEventListeners();

	if (this.contractSelected) {
		this._showFieldsetSubmit();
	} else {
		this._hideFieldsetSubmit();
	}

};

Module.prototype._cacheElements = function() {
	this.$contractRadioButtons = this.$element.find(this.options.domSelectors.radioButtons);
	this.$mainRoleCheckbox = this.$element.find(this.options.domSelectors.mainRoleCheckbox);
	this.$fieldsetSubmit = this.$element.find(this.options.domSelectors.fieldsetSubmit);
};

Module.prototype._addEventListeners = function() {

	/* this.$mainRoleCheckbox.on('change.' + this.uuid, function() {
		this.$element.toggleClass(this.options.stateClasses.isChangingMainRole);
	}.bind(this)); */

	this.$contractRadioButtons.on('change.' + this.uuid, function() {

		this._resetMainRoleCheckBox();

		// If there is no contract set initially, we want to listen for changes
		// and only then let the submit fieldset appear
		if (!this.contractSelected) {
			this._showFieldsetSubmit();

			this.$fieldsetSubmit.one(cssTransitionEnd + '.' + this.uuid, function() {
				this.scrollMainRoleCheckBoxIntoView();
			}.bind(this));

		} else {
			this.scrollMainRoleCheckBoxIntoView();
		}

	}.bind(this));

};

Module.prototype._hideFieldsetSubmit = function() {
	this.$fieldsetSubmit.attr({
		'hidden': 'hidden'
	});
};

Module.prototype._showFieldsetSubmit = function() {
	this.$fieldsetSubmit.removeAttr('hidden');

	requestAnimationFrame(function() {
		this.$fieldsetSubmit.addClass(this.options.stateClasses.isVisible);
	}.bind(this));

	this.contractSelected = true;

};

Module.prototype._resetMainRoleCheckBox = function() {
	this.$mainRoleCheckbox.prop('checked', false);
};

Module.prototype.scrollMainRoleCheckBoxIntoView = function() {
	// TODO re-factor to something more elegant. this is causing too much confusion when the user operates by keyboard
	// this.$mainRoleCheckbox.focus();
};

Module.prototype._hasMainRoleClassSet = function() {
	return this.$element.find(this.options.stateClasses.isMainRole).length;
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	initEvents: ['ready', 'ajaxload'],
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 199 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 200 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(201);


/***/ }),
/* 201 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M00 - Module Loader
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(202);

var $ = __webpack_require__(0),
	$document = $(document),
	SuperClass = __webpack_require__(1);

var name = 'moduleloader',
	events = {},
	defaults = {
		domSelectors: {},
		stateClasses: {}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, load content
 * @method
 * @public
 */
Module.prototype.init = function() {

	var data = {};

	if (!this.options.url) {
		return;
	}

	if (this.options.random) {
		data = { random: this._getRandomInt(0, this.options.random) };
	}

	$.ajax({
		url: this.options.url,
		data: data,
		success: function(response) {
			var $content = $(response);

			this.$element.html($content);
			$document.trigger('ajaxload', $content);

			if (window.tracking && this.options.initTrackingAfterLoad) {
				window.tracking.scanForElementsToTrack();
			}

		}.bind(this)
	});
};

/**
 * return random integer based on min, max params
 * @method
 * @private
 */
Module.prototype._getRandomInt = function(min, max) {
	min = Math.ceil(min);
	max = Math.floor(max);

	return Math.floor(Math.random() * (max - min + 1)) + min;
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 202 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 203 */,
/* 204 */,
/* 205 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(206);


/***/ }),
/* 206 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 207 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(208);


/***/ }),
/* 208 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 209 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(210);


/***/ }),
/* 210 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M19.5 - Live-Chat
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



// Ensure that the SASS file is built.
__webpack_require__(211);

var $ = __webpack_require__(0),
	mediaqueries = __webpack_require__(6),
	SuperClass = __webpack_require__(1),
	speculative = __webpack_require__(30);

var name = 'livechat',
	$document = $(document),
	events = {
		chat: {
			online: 'rp.status.text.online',
			offline: 'rp.status.text.offline'
		}
	},
	defaults = {
		domSelectors: {
			infoText: '[data-' + name + '="infotext"]',
			loader: '[data-' + name + '="loader"]'
		},
		stateClasses: {
			isHidden: 'is_hidden'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	speculative({
		href: this.options.preFetchURL,
		rel: 'dns-prefetch'
	});

	speculative({
		href: this.options.preFetchURL,
		rel: 'preconnect'
	});

	this.$infoText = this.$element.find(this.options.domSelectors.infoText);
	this.$loader = this.$element.find(this.options.domSelectors.loader);

	$document[0].addEventListener(events.chat.online, this._togglePanel.bind(this));
	$document[0].addEventListener(events.chat.offline, this._togglePanel.bind(this));

	$document.on(mediaqueries.event.key + '.' + this.uuid, this._updateNBC);

};

/**
 * Toggle Info Panel
 * @private
 */
Module.prototype._togglePanel = function(e) {

	this.$loader.remove();

	if (e.type === events.chat.online) {
		this.$infoText
			.removeClass(this.options.stateClasses.isHidden)
			.removeAttr('aria-hidden')
			.removeAttr('role');
	} else {
		this.$infoText
			.addClass(this.options.stateClasses.isHidden)
			.attr('aria-hidden', 'true')
			.attr('role', 'presentation');
	}

	this._updateNBC();

};

/**
 * Update NBC
 * @private
 */
Module.prototype._updateNBC = function() {
	if (mediaqueries.query({from: 'desktop'})) {
		requestAnimationFrame(function() {
			$document.trigger('update.estatico.nbc');
		});
	}
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 211 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 212 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(213);


/***/ }),
/* 213 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.2 - Textarea
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(214);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'textarea',
	events = {},
	defaults = {
		domSelectors: {
			textarea: '[data-' + name + '="textarea"]',
			counter: '[data-' + name + '="counter"]'
		},
		stateClasses: {
		},
		minHeight: 192,
		prevHeight: 192
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	this.$textareaEl = this.$element.find(this.options.domSelectors.textarea);
	this.$counter = this.$element.find(this.options.domSelectors.counter);

	if (this.$counter.length === 0) {
		return;
	}

	this.textareaEl = this.$textareaEl[0];

	this.fieldOffset = this.textareaEl.offsetHeight - this.textareaEl.clientHeight;
	this.maxAllowLength = parseInt(this.$textareaEl.attr('maxlength'));

	this._updateCounter(this.maxAllowLength);
	this.$element.on('input.' + this.uuid + ' keypress.' + this.uuid, this.options.domSelectors.textarea, this._inputValueIsChanged.bind(this));
};

/**
 * Handles keyup, input events
 * @private
 */
Module.prototype._inputValueIsChanged = function(evt) {

	if (!this._preventNewLineOnLastChar(evt)) {
		return false;
	}

	this._resizeTextarea();
	this._updateCounter();
};

/**
 * Adjust textarea height to the height of the text entered.
 * @private
 */
Module.prototype._resizeTextarea = function() {
	var newHeight;

	this.$textareaEl.css({
		height: 'auto'
	});

	newHeight = this.textareaEl.scrollHeight + this.fieldOffset;

	this.$textareaEl.css({
		height: this.options.prevHeight
	});

	window.requestAnimationFrame(function() {
		if (newHeight > this.options.minHeight) {
			this.$textareaEl.css({
				height: newHeight
			});
			this.options.prevHeight = newHeight;
		} else {
			this.$textareaEl.css({
				height: this.options.minHeight
			});
			this.options.prevHeight = this.options.minHeight;
		}
	}.bind(this));

};

/**
 * Updates the value of the remaining amount of acceptable characters.
 * @private
 */
Module.prototype._updateCounter = function() {
	var enteredText = this.$textareaEl.val(),
		lengthLeft = this._getLengthLeft(),
		newLines = enteredText.match(/(\r\n|\n|\r)/g);

	this._setMaxLengthOnTextareaForNewLines(newLines);

	this.$counter.text(lengthLeft);
};

// Update max length on textarea
Module.prototype._setMaxLengthOnTextareaForNewLines = function(newLines) {
	if (newLines !== null) {
		this.$textareaEl.attr({
			'maxlength': this.maxAllowLength - newLines.length
		});
	} else {
		this.$textareaEl.attr({
			'maxlength': this.maxAllowLength
		});
	}
};

// If there is only one character left to type, we need special logic to prevent
// user from entering a new line
Module.prototype._preventNewLineOnLastChar = function(evt) {

	var length = this._getLengthLeft();

	if (length === 1) {

		if (evt.keyCode === 13) {
			return false;
		}
	}

	return true;

};

// Get the length left and take new lines into consideration
Module.prototype._getLengthLeft = function() {
	var enteredText = this.$textareaEl.val(),
		lengthLeft = this.maxAllowLength - enteredText.length,
		newLines = enteredText.match(/(\r\n|\n|\r)/g);

	if (newLines !== null) {
		lengthLeft -= newLines.length;
	}

	return lengthLeft;
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 214 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 215 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(216);


/***/ }),
/* 216 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M07 - L&H Teaser
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(217);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'lhteaser',
	isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints,
	events = {},
	defaults = {
		domSelectors: {
			link: '[data-' + name + '="link"]'
		},
		stateClasses: {
			isLongPressed: 'is_long_pressed'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	// immediately return if we're not on a touch device
	if (!isTouchDevice) {
		return;
	}

	// chache link element
	this.$link = this.$element.find(this.options.domSelectors.link);

	// handle clicks and touches
	this.handleClick();
	this.handleTouches();
	this.handleHoverTriggered();

};

/**
 * Handle default clicks
 */
Module.prototype.handleClick = function() {

	// prevent default on clicks
	this.$element.on('click.' + this.uuid, this.options.domSelectors.link, function(event) {
		event.preventDefault();
	});

	// make sure the default context menu does not appear on android and
	// windows. ios is fine ;-)
	window.oncontextmenu = function(event) {
		event.preventDefault();
		event.stopPropagation();
		return false;
	};

};

/**
 * Handle touches
 */
Module.prototype.handleTouches = function() {

	var linkHref;

	// include hammerjs script
	__webpack_require__.e/* require.ensure */(1/* duplicate */).then((function(require) {
			__webpack_require__(22);

			// init hammer on teaser link
			this.$element.find(this.options.domSelectors.link).hammer()

				// simulate hover on long press
				.on('press.' + this.uuid, function() {

					// add/remove hover class
					this.$link.toggleClass(this.options.stateClasses.isLongPressed);

					// trigger event
					this.$element.trigger('hoverTriggered', [{object: this.uuid}]);

				}.bind(this))

				// open url on tap
				.on('tap.' + this.uuid, function() {
					linkHref = this.$link.attr('href');
					window.open(linkHref, '_self');
				}.bind(this));

		}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);

};

// Listen to triggered event omitted by other instances of this module
Module.prototype.handleHoverTriggered = function() {
	$(document).on('hoverTriggered', function(event, data) {
		if (data.object !== this.uuid) {
			this.$link.removeClass(this.options.stateClasses.isLongPressed);
		}
	}.bind(this));
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 217 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 218 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(219);


/***/ }),
/* 219 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 220 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(221);


/***/ }),
/* 221 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M13 - Tables
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(222);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	globalEvents = __webpack_require__(5),
	cssTransitionEnd = __webpack_require__(4);

var name = 'tables',
	$window = $(window),
	$document = $(document),
	isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints,
	events = {
		openAccordion: 'open.estatico.accordion'
	},
	defaults = {
		domSelectors: {
			accordionItem: '[data-accordion="item"]',
			tableWrapper: '[data-' + name + '="table_wrapper"]',
			table: '[data-' + name + '="table"]',
			tableHeaderCell: '[data-' + name + '="table_header_cell"]',
			shadowRight: '[data-' + name + '="table_shadow_right"]'
		},
		stateClasses: {
			hasShadow: 'has_shadow',
			isVisible: 'is_visible',
			hasNoFixedColumn: 'var_not_fixed'
		},
		$tableWrapper: undefined,
		$table: undefined,
		$tableSticky: undefined
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	var columns,
		bodyHeaderCells,
		$parentAccordionItem;

	// cache elements
	this.$tableWrapper = this.$element.find(this.options.domSelectors.tableWrapper);
	this.$table = this.$element.find(this.options.domSelectors.table);
	this.$shadowRight = this.$element.find(this.options.domSelectors.shadowRight);
	$parentAccordionItem = this.$element.closest(this.options.domSelectors.accordionItem);
	bodyHeaderCells = this.$element.find(this.options.domSelectors.tableHeaderCell);

	// setup shadow on the right side
	this._initShadowRight();

	// check if we're dealing with a sticky table.

	// a) If we don't have th elements in the tbody, we're definitely not dealing
	// with a sticky table.
	if (bodyHeaderCells.length === 0) {
		return;
	}

	// b) if we detect the stateClass var_not_fixed we don't want a sticky
	// header column
	if (this.$element.hasClass(this.options.stateClasses.hasNoFixedColumn)) {
		return;
	}

	// cache sticky elements
	this.$tableSticky = $('<table class="mod_table_table_sticky" cellspacing="0" aria-hidden="true" role="presentation"></table>');
	columns = this.$table.find('thead, tbody').clone();

	// append sticky table to the dom
	this.$table.after(this.$tableSticky);

	// fill the sticky column table
	this.$tableSticky.append(columns).find('thead th:gt(0)').remove().end().find('tbody td').remove();

	// attach scroll event to table
	this.$tableWrapper.on('scroll', function() {
		this._stickyColumnVisability();
	}.bind(this));

	// If we are inside of an accordion, listen to the open event and then
	// calculate the dimensions and set visibility of the sticky column
	if ($parentAccordionItem.length > 0) {
		$document.on(events.openAccordion + '.' + this.uuid, function() {
			this._setStickyColumnDimensions();
			this._stickyColumnVisability();
		}.bind(this));
	}

	// attach load event listener to the window
	$window.on('load', this._setStickyColumnDimensions());

	// attach resize event listener to the window
	$document.on(globalEvents.resize.key + '.' + this.uuid, function() {
		this._setStickyColumnDimensions();
		this._stickyColumnVisability();
	}.bind(this));

	// Show sticky table column if the table is wider than its wrapper
	if (this._tableHasOverflow()) {
		this.$tableSticky.css({
			opacity: 0.9999,
			display: 'block'
		});
	}

};

/**
 * Check if table should scroll
 * @method
 * @private
 */
Module.prototype._tableHasOverflow = function() {
	return this.$tableWrapper.width() < this.$table.outerWidth();
};

/**
 * Setup shadow pseudo element on the right side of the table
 * @method
 * @private
 */
Module.prototype._initShadowRight = function() {

	this._setupShadowRight();

	// listen to resize, show/hide shadow conditionally
	$document.on(globalEvents.resize.key + '.' + this.uuid, function() {
		this._setupShadowRight();
	}.bind(this));

	// listen to scroll, show/hide shadow conditionally
	this.$tableWrapper.on('scroll', function() {
		this._setupShadowRight();
	}.bind(this));
};

Module.prototype._setupShadowRight = function() {
	this._clipShadowElement(this.$shadowRight, true);
	this._showHideShadowRight();
};

/**
 * Show or hide the shadow on the right
 * @method
 * @private
 */
Module.prototype._showHideShadowRight = function() {
	var maxScrollDistance = parseInt(this.$table.outerWidth() - this.$tableWrapper.width()),
		scrollLeft = parseInt(this.$tableWrapper.scrollLeft()) + 10, // some pixels for reserve
		isScrolledToTheFarLeft = maxScrollDistance - scrollLeft < 0;

	if (!this._tableHasOverflow() || isScrolledToTheFarLeft) {
		this._hideShadowRight();

		return;
	}

	this._showShadowRight();
};

/**
 * Sets the height of the row from the original table to the row from the sticky table column. Also, setting the width of the cells
 * @method
 * @private
 */
Module.prototype._setStickyColumnDimensions = function() {

	var thWidth = 0,
		trHeight = 0;

	// set height of sticky table cells
	this.$table.find('tr').each(function(i, item) {

		// get the height of this row. make it an integer
		trHeight = parseInt($(item).height());

		// set the height of the sticky column row
		this.$tableSticky.find('tr').eq(i).height(trHeight);

		// it may be that the row has a fractional height. thus, setting to row height to an integer value,
		// we make sure that the height of this row and the height of to fixed column row are exactly the same.
		$(item).height(trHeight);
	}.bind(this));

	// get width of first cell in the header row. make it an integer.
	// Since we add a border to the sticky table, we also need to
	// compensate for that.
	thWidth = parseInt(this.$table.find('thead th:first-child').innerWidth() - 3);

	// set the width of the body cells from the sticky table
	this.$tableSticky.css('width', thWidth + 'px');

	this._clipShadowElement(this.$tableSticky);
};

/**
 * Clip shadow on element like sticky column or shadow on pseudo
 * element on the right side of the table
 * @method
 * @private
 */
Module.prototype._clipShadowElement = function($shadowElement, compensateForTopValue) {
	var topValue = 0,
		topRawValue;

	if (compensateForTopValue) {
		topRawValue = this.$shadowRight.css('top');
		topValue = Math.abs(parseInt(topRawValue.split('px')[0]));
	}

	// clip out the box shadow on the shadow element
	var tableHeight = this.$table.outerHeight() + topValue + 'px';

	// clip is deprecated, but still working in all browsers.
	// For sustainability, set clip-path as well,
	// although not yet supported by all browsers.
	$shadowElement.css('clip', 'rect(' + topValue + 'px,9000px,'+ tableHeight + ',-30px)');
	$shadowElement.css('clip-path', 'inset(0px -20px 0px 0px)');
};

/**
 * Handles the visibility of the sticky table column
 * @method
 * @private
 */
Module.prototype._stickyColumnVisability = function() {
	if (this.$tableWrapper.scrollLeft() > 0) {
		// show sticky table and add shadow
		this._showStickyColumn();
	} else {

		// hide sticky table and fade out shadow
		this._hideStickyColumn();
	}
};

/**
 * Helper method to hide the sticky column
 * @method
 * @private
 */
Module.prototype._hideStickyColumn = function() {
	if (cssTransitionEnd) {
		this.$tableSticky.removeClass(this.options.stateClasses.hasShadow);
		if (!isTouchDevice) {
			this.$tableSticky.one(cssTransitionEnd, function() {
				this.$tableSticky.css({opacity: 0, display: 'none'});
			}.bind(this));
		}
	} else {
		this.$tableSticky.removeClass(this.options.stateClasses.hasShadow);
		if (!isTouchDevice) {
			this.$tableSticky.css({opacity: 0, display: 'none'});
		}
	}
};

/**
 * Helper method to show the sticky column
 * @method
 * @public
 */
Module.prototype._showStickyColumn = function() {
	this.$tableSticky
		.addClass(this.options.stateClasses.hasShadow)
		.css({
			opacity: 0.9999,
			display: 'block'
		});
};

/**
 * Helper method to show the shadow on the right side
 * @method
 * @public
 */
Module.prototype._showShadowRight = function() {
	this.$shadowRight.addClass(this.options.stateClasses.isVisible);
};

/**
 * Helper method to hide the shadow on the right side
 * @method
 * @public
 */
Module.prototype._hideShadowRight = function() {
	this.$shadowRight.removeClass(this.options.stateClasses.isVisible);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 222 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 223 */,
/* 224 */,
/* 225 */,
/* 226 */,
/* 227 */,
/* 228 */,
/* 229 */,
/* 230 */,
/* 231 */,
/* 232 */,
/* 233 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(234);


/***/ }),
/* 234 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M11 - Pagination
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(235);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	name = 'pagination',
	events = {};

// Create an instance of the module
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: {},
		options: options,
		events: events,
		data: {}
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// Initialize module, bind events.
Module.prototype.init = function() {
	if (!this.options.loadJS) {
		return;
	}

	__webpack_require__.e/* require.ensure */(9).then((function(require) {

			this.Pagination = __webpack_require__(226);

			setTimeout(function() {
				this.Pagination = new this.Pagination(this.$element, this.uuid, name);
			}.bind(this), 1);

		}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 235 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 236 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(237);


/***/ }),
/* 237 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 238 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(239);


/***/ }),
/* 239 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M07 - L&H Teaser
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(240);

var $ = __webpack_require__(0),
	$document = $(document),
	SuperClass = __webpack_require__(1),
	mediaqueries = __webpack_require__(6),
	ConsentHelper = __webpack_require__(26);

var name = 'googleads',
	events = {},
	defaults = {
		domSelectors: {
			content: '[data-' + name + '="content"]'
		},
		stateClasses: {
			isHidden: 'is_hidden'
		}
	},
	data = {},
	eventNameGoogleAdsReady = 'googleAdsReady';

// Create an instance of the module
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// Initialize module, bind events.
Module.prototype.init = function() {
	document.addEventListener(eventNameGoogleAdsReady, function() {
		this.showHideAd();
	}.bind(this));

	this.$content = this.$element.find(this.options.domSelectors.content);

	// set initial breakpoint
	this.breakpoint = this.getCurrentBreakpoint();

	// conditionally show ads
	this.showHideAd();

	// listen to breakpoint changes
	$document.on(mediaqueries.event.key, this.handleBreakpointChange.bind(this));
};

Module.prototype.handleBreakpointChange = function() {
	var newBreakpoint = this.getCurrentBreakpoint();

	if (newBreakpoint === this.breakpoint) {
		return;
	}

	this.breakpoint = newBreakpoint;

	this.showHideAd();
};

Module.prototype.showHideAd = function() {
	var adsAreAllowed = false;

	if (ConsentHelper.consentIsIntegrated()) {
		ConsentHelper.activeGroups().then(function(groups) {
			adsAreAllowed = groups.allowedAds && groups.allowedMarketing;

			if (adsAreAllowed) {
				if (this.options.allowedViewports.indexOf(this.breakpoint) !== -1) {
					this.showAd();
				} else {
					this.hideAd();
				}
			} else {
				this.hideAd();
			}
		}.bind(this));
	} else {
		this.showAd();
	}

};

Module.prototype.getCurrentBreakpoint = function() {
	var currentBreakpoint = mediaqueries.getCurrentBreakpointName(),
		breakpointUndefined = currentBreakpoint === undefined,
		mobilePortrait = currentBreakpoint === 'mobilePortrait',
		mobileLandscape = currentBreakpoint === 'mobileLandscape',
		tabletPortrait = currentBreakpoint === 'tabletPortrait',
		tabletLandscape = currentBreakpoint === 'tabletLandscape',
		desktop = currentBreakpoint === 'desktop',
		desktopLarge = currentBreakpoint === 'desktopLarge',
		desktopLargePlus = currentBreakpoint === 'desktopLargePlus',
		viewport,
		viewportWidth;

	if (breakpointUndefined || mobilePortrait || mobileLandscape) {
		viewport = 'mobile';
	} else if (tabletPortrait || tabletLandscape) {
		viewport = 'tablet';
	} else if (desktop || desktopLarge || desktopLargePlus) {
		viewport = 'desktop';
	} else {
		viewport = 'desktop4k';
	}

	// we have an exception: ads beside should not show at 4k already, only
	// at window width of 2200px.
	if (viewport === 'desktop' || viewport === 'desktop4k') {
		viewportWidth = window.innerWidth;

		if (viewportWidth > 2200) {
			viewport = 'desktop4k';
		} else {
			viewport = 'desktop';
		}
	}

	return viewport;

};

Module.prototype.showAd = function() {
	var slot,
		slotId;

	// iterate all slots that are defined on the service and find the slot
	// that is defined for this ad instance
	if (window.googletag && window.googletag.pubads) {
		window.googletag.pubads().getSlots().forEach(function(_slot) {
			slotId = _slot.getSlotElementId();

			if (slotId === this.options.adsId) {
				slot = _slot;
			}
		}.bind(this));
	}

	if (!slot) {
		return;
	}

	// add the ad-id to the element and call refresh to load the ad
	if (this.$content.attr('id') !== this.options.adsId) {
		this.$content.attr('id', this.options.adsId);

		if (window.googletag.pubads) {
			window.googletag.pubads().refresh([slot]);
		}
	}

	this.$element.removeClass(this.options.stateClasses.isHidden);
};

Module.prototype.hideAd = function() {
	this.$content.removeAttr('id');
	this.$element.addClass(this.options.stateClasses.isHidden);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 240 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 241 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(242);


/***/ }),
/* 242 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 243 */,
/* 244 */,
/* 245 */,
/* 246 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(247);


/***/ }),
/* 247 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M39 - Social Media Stream
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(248);

var $ = __webpack_require__(0),
	Handlebars = __webpack_require__(8),
	SuperClass = __webpack_require__(1),
	ajaxRequest = __webpack_require__(16),
	mediaqueries = __webpack_require__(6);

var name = 'socialstream',
	$document = $(document),
	events = {},
	viewports = {
		mobile: [
			undefined, // smallest one seems to be undefined
			'mobilePortrait',
			'mobileLandscape'
		],
		tablet: [
			'tabletPortrait',
			'tabletLandscape'
		]
	},
	defaults = {
		domSelectors: {
			grid: '[data-columns]', // defers from coding guidelines to match masonry plugin
			masonryColumn: '.mod_socialstream_cols', // defers from coding guidelines to match masonry plugin
			loadMoreButton: '[data-socialstream="show_more"]',
			loadingAnimation: '.mod_loading_animation'
		},
		stateClasses: {
			hasAdditionalEntries: 'has_additional_entries'
		},
		i18n: {
			agoMin: 'Vor wenigen Minuten',
			minutes: 'Minuten',
			hours: 'Stunden',
			days: 'Tagen'
		},
		offset: 0,
		showLoadMoreButton: false
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	this.$gridContainer = this.$element.find(this.options.domSelectors.grid);
	this.$loadingAnimation = this.$element.find(this.options.domSelectors.loadingAnimation);
	this.$loadMoreButton = this.$element.find(this.options.domSelectors.loadMoreButton).on('click.' + this.uuid, this._getSocialMediaEntries.bind(this));

	this._getSocialMediaEntries();
};

/**
 * register handlebar helpers
 */
Module.prototype._registerHandlebarsHelpers = function() {

	Handlebars.registerHelper('is', function(arg1, arg2, options) {
		if (arg1 === arg2) {
			return options.fn(this);
		}

		return options.inverse(this);
	});

	Handlebars.registerHelper('switch', function(value, options) {
		this.switchValue = value;
		this.switchBreak = false;
		var html = options.fn(this);

		delete this.switchValue;
		return html;
	});

	Handlebars.registerHelper('case', function(value, options) {
		if (value === this.switchValue) {
			this.switchBreak = true;
			return options.fn(this);
		}
	});

	Handlebars.registerHelper('default', function(value) {
		if (!this.switchBreak) {
			return value.fn(this);
		}
	});

	Handlebars.registerHelper('isVideo', function(block) {
		// jscs:disable
		if (this.media_type === 'video') {
			// jscs:enable
			return block.fn(this);
		}
	});

	Handlebars.registerHelper('getRatio', function(value) {

		var aspectRatio = Math.floor(value.height / value.width * 100);

		if (aspectRatio > 100) {
			aspectRatio = Math.floor(value.width / value.height * 100);
		}

		return aspectRatio + '%';
	});

	Handlebars.registerHelper('transformHttp', function(url) {
		return url.replace('http://', 'https://');
	});

	Handlebars.registerHelper('checkTimestamp', function(timestamp) {
		var timestampNow = Math.floor(Date.now() / 1000),
			timestampDate = new Date(timestamp* 1000),
			offsetMinutes = (timestampNow - timestamp) / 60,
			offsetHours = offsetMinutes / 60,
			offsetDays = offsetHours / 24,
			retStr = '';

		if (offsetMinutes <= 30) {
			// show 'some minutes ago' if less than 30 minutes
			retStr = this.options.i18n.agoMin;
		} else if (offsetMinutes <= 120) {
			// show minutes if less or equal 2 hours ago
			retStr = Math.floor(offsetMinutes) + ' ' + this.options.i18n.minutes;
		} else if (offsetHours <= 24) {
			// show hours if less or equal 24 hours ago
			retStr = Math.floor(offsetHours) + ' ' + this.options.i18n.hours;
		} else if (offsetDays <= 14) {
			// show days if less or equal 14 days ago
			retStr = Math.ceil(offsetDays) + ' ' + this.options.i18n.days;
		} else {
			// show date
			retStr = timestampDate.toLocaleDateString('de-DE', {day: '2-digit', month: '2-digit', year: 'numeric'});
		}

		return retStr;

	}.bind(this));

};

Module.prototype._constructUrl = function() {

	var req = {
		'action': 'broadcasts',
		'room_id': this.options.rid,
		'limit': this._itemsAmount() + 1,
		'offset': this.options.offset
	};

	return this.options.api + JSON.stringify(req);

};

/**
 * get contentfry data
 */
Module.prototype._getSocialMediaEntries = function() {

	var requestUrl = this._constructUrl();

	// get content for url
	ajaxRequest(requestUrl).then(function(jsonResponse) {
			this.socialMediaEntries = jsonResponse.data;

			// Check if we have received 13 results. In this case we know that
			// there are more results to show and we can display the load more
			// button. We also remove the 13th entry from the array.
			if (this.socialMediaEntries.length > this._itemsAmount()) {
				this.showLoadMoreButton = true;
				this.socialMediaEntries.pop();
			} else {
				this.showLoadMoreButton = false;
			}

			if (this.options.offset === 0) {
				this.$loadingAnimation.remove();
				this._registerHandlebarsHelpers();
				this._requireDependenciesAndRenderEntries();

				// set aria roles on salvattore rebuild
				$document.on(mediaqueries.event.key + '.' + this.uuid, this._handleBreakpointChange.bind(this));
			} else {
				this._renderSocialMediaEntries();
			}

			this._toggleLoadMoreButton();

		}.bind(this),

		// handle errors
		function(error) {
			console.log(error);
		});

};

/**
 * Load Async Dependencies and render Social Media Entries
 * @private
 */
Module.prototype._requireDependenciesAndRenderEntries = function() {

	__webpack_require__.e/* require.ensure */(10).then((function(require) {
			this.salvattore = __webpack_require__(227);
			this.template = __webpack_require__(228);
			this._renderSocialMediaEntries();
		}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);

};

/**
 * Render the Social Media Entries that we have received from the
 * ContentFry API
 */
Module.prototype._renderSocialMediaEntries = function() {

	var i = 0,
		ids = [],
		len = this.socialMediaEntries.length,
		templateData,
		entry;

	while (i < len) {
		templateData = this.template(this.socialMediaEntries[i]);
		entry = document.createElement('div');
		ids.push(this.socialMediaEntries[i].id);
		entry.innerHTML = templateData;
		this.salvattore.appendElements(this.$gridContainer[0], [entry.firstChild]);
		i++;
	}

	// Check if the render function was triggered by using the load more button
	// we know so, since the offset would be 0 on initial. If so point the screenreader
	// to the first of the freshly added links and trigger the focus there.
	if (this.options.offset > 0) {
		this.$gridContainer.find('[data-id=' + ids[0] + '] a').focus();
	}

	this.options.offset += this._itemsAmount() - 1;

	// We set role presentation for the masonry column wrapper divs to hide them from the
	// screenreader and make the module announce itself as a list.
	window.requestAnimationFrame(function() {
		this.$element.find(this.options.domSelectors.masonryColumn).attr('role', 'presentation');
	}.bind(this));

	// TODO remove loading animation and set focus on first of the newly added entries
	// Add visuallyhidden block announcing that items are getting loaded

};

/**
 * Toggle Visibility of Load More Button
 */
Module.prototype._toggleLoadMoreButton = function() {
	if (this.showLoadMoreButton) {
		this.$element.addClass(this.options.stateClasses.hasAdditionalEntries);
		this.$loadMoreButton.removeAttr('aria-hidden role');
	} else {
		this.$loadMoreButton.remove();
	}
};

/**
 * Breakpoint change handler
 */
Module.prototype._handleBreakpointChange = function() {
	// wait for salvattore rebuild
	setTimeout(function() {
		this.$gridContainer.find(this.options.domSelectors.masonryColumn).attr('role', 'presentation');
	}.bind(this), 300);
};

// Get items amount for current viewport
Module.prototype._itemsAmount = function() {
	var currentBreakpoint = mediaqueries.getCurrentBreakpointName(),
		items = this.options.itemsAmount.desktop;

	if (viewports.mobile.indexOf(currentBreakpoint) !== -1) {
		items = this.options.itemsAmount.mobile;
	} else if (viewports.tablet.indexOf(currentBreakpoint) !== -1) {
		items = this.options.itemsAmount.tablet;
	}

	// in case BE delivers amount values as strings...
	return parseInt(items, 10);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 248 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 249 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(250);


/***/ }),
/* 250 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 251 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(252);


/***/ }),
/* 252 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 253 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(254);


/***/ }),
/* 254 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.7 - Upload
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(255);
__webpack_require__(3);

var $ = __webpack_require__(0),
	template = __webpack_require__(51),
	templateSettings = __webpack_require__(27),
	SuperClass = __webpack_require__(1),
	$document = $(document);

var name = 'upload',
	events = {
		fileselect: 'change.estatico',
		dragenter: 'dragenter.estatico',
		dragover: 'dragover.estatico',
		drop: 'drop.estatico',
		removeFile: 'click.estatico',
		multistepFormSubmitSuccess: 'multistepForm.submitSuccess'
	},
	filetypes = [
		{ expression: /^saskruploxbdgwuerst/, icon: 'sbb_116_doc', label: 'FILE'},
		{ expression: /^image\/jpeg/, icon: 'sbb_115_image', label: 'JPEG'},
		{ expression: /^image\/png/, icon: 'sbb_115_image', label: 'PNG'},
		{ expression: /^image\/gif/, icon: 'sbb_115_image', label: 'GIF'},
		{ expression: /^application\/postscript/, icon: 'sbb_115_image', label: 'PS'},
		{ expression: /^application\/text/, icon: 'sbb_116_doc', label: 'TEXT'},
		{ expression: /^text\/plain/, icon: 'sbb_116_doc', label: 'TXT'},
		{ expression: /^application\/pdf/, icon: 'SBB_38_pdf', label: 'PDF'},
		{ expression: /^sound\//, icon: 'sbb_118_sound', label: 'AUDIO'},
		{ expression: /^video\//, icon: 'sbb_119_video', label: 'VIDEO'},
		{ expression: /^application\/zip/, icon: 'sbb_120_zip', label: 'ZIP'},
		{ expression: /^application\/vnd.openxmlformats-officedocument.word/, icon: 'sbb_116_doc', label: 'DOC'},
		{ expression: /^application\/vnd.openxmlformats-officedocument.spreadsheet/, icon: 'sbb_116_doc', label: 'XLS'},
		{ expression: /^application\/vnd.openxmlformats-officedocument.presentation/, icon: 'sbb_116_doc', label: 'PPT'}
	],
	defaults = {
		domSelectors: {
			filesList: '[data-' + name + '=' + 'mod_upload_files_list' + ']',
			listItemTemplate: '[data-' + name + '=' + 'template' + ']',
			deleteButton: '[data-' + name + '=' + 'delete' + ']',
			input: '[data-' + name + '=' + 'input' + ']',
			filesListItem: '[data-' + name + '=' + 'files_list_item' + ']',
			hiddenTextfield: '[data-' + name + '=' + 'hidden_field' + ']',
			notification: '[data-' + name + '="notification"]',
			notificationMessage: '[data-notification="message"]'
		},
		stateClasses: {
			isVisible: 'is_visible',
			isHidden: 'is_hidden',
			hasError: 'has-error'
		},
		maxFileSize: '50000000000000',
		filesInLocalStorage: false,
		provideImageDimensionWarning: false,
		i18n: {
			maxFileSizeError: 'Die Datei $0 ist zu gross, die maximale Dateigrösse beträgt 512 Kilobyte.',
			maxLocalStorageSizeError: 'Die Datei $0 konnte nicht hinzugefügt werden.'
		},
		localStorageOptions: {
			cacheExpirationTime: 1800, // in seconds
			uploadedFilesKey: 'ls-uploaded-files'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	// Change interpolate settings to accept Handlebars syntax
	templateSettings.interpolate = /{{([\s\S]+?)}}/g;

	this.filesArray = [];
	this.filesArrayToStore = [];
	this.hasError = false;

	this._setupEventHandlers();

	if (this.options.validateImageDimensions) {
		__webpack_require__(256);
	}

	if (this.options.filesInLocalStorage) {
		this._setInitialState();
	}
};

Module.prototype._setupEventHandlers = function() {
	// Add listener to input=file to handle selecting one or multiple files
	this.$input = this.$element.find(this.options.domSelectors.input);
	this.$input.on('click.' + this.uuid, this._clearSelectedFiles.bind(this));
	this.$input.on(events.fileselect, this._onFileSelectHandler.bind(this));
	this.$input.on(events.dragenter, this._onDragEnter.bind(this));
	this.$input.on(events.dragover, this._onDragOver.bind(this));
	this.$input.on(events.drop, this._onDrop.bind(this));

	// remove files from local storage after successfully form submit
	$document.on(events.multistepFormSubmitSuccess + '.' + this.uuid, this._cleanLocalStorage.bind(this));

	// make sure that no more than one error message is displayed
	this.$input.parsley().on('field:error', function() {
		if (this.hasError) {
			this.$input.parsley().removeError('required', {updateClass: true});
		}
	}.bind(this));
};

/**
 * Prevents rendering dragged object in the browser window
 * @param event {jQuery}
 * @private
 */
Module.prototype._onDragEnter = function(event) {
	event.preventDefault();
	event.stopPropagation();
};

/**
 * Prevents rendering dragged object in the browser window
 * @param event {jQuery}
 * @private
 */
Module.prototype._onDragOver = function(event) {
	event.preventDefault();
	event.stopPropagation();
};

/**
 * Handles file drop event from dragging file over
 * @param event {jQuery}
 * @private
 */
Module.prototype._onDrop = function(event) {
	event.preventDefault();
	event.stopPropagation();

	var dt = event.originalEvent.dataTransfer;

	this._onFilesSelected(dt.files);
};

/**
 * Pass array of files that have been selected. Checks if there are any previously existing files
 * and clears the list.
 * @param files array of files selected or dropped into input field
 * @private
 */
Module.prototype._onFilesSelected = function(files) {

	// Clear previously selected files only if new array is not empty - meaning that some files have been selected
	if (files.length > 0) {

		if (this.filesArray && this.filesArray.length > 0 && !this.options.filesInLocalStorage) {
			this._clearFilesList();
		}

		this._displayFilesList(files);
	}
};

/**
 * Event handler triggered when user accepts files from popup system dialog
 * @private
 */
Module.prototype._onFileSelectHandler = function() {
	this._onFilesSelected(this.$input[0].files);
};

/**
 * Clear selected files form input file element.
 * @private
 */
Module.prototype._clearSelectedFiles = function() {
	this.$input[0].value = null;
};

/**
 * Loop that on each iteration adds file to the list of selected files below the button
 * @private
 */
Module.prototype._displayFilesList = function(files) {
	var i = 0,
		len = files.length,
		file;

	this._removeErrorMessage();

	for (i; i< len; i++) {
		file = files[i];
		if (typeof file.size !== 'undefined' && file.size <= this.options.maxFileSize) {
			this._prepareFileForLocalStorage(file);
			this.hasError = false;

			if (this.options.provideImageDimensionWarning) {
				this._getImageDimension();
			}

		} else {
			this.hasError = true;
		}
	}

	if (this.hasError) {
		this._showErrorMessage(file.name, this.options.i18n.maxFileSizeError);
	}
};

/**
 * Clear list of already selected files
 * @private
 */
Module.prototype._clearFilesList = function() {
	var $list = this.$element.find(this.options.domSelectors.filesList);

	$list.find(this.options.domSelectors.filesListItem).remove();
	$list.removeClass(this.options.stateClasses.isVisible);
	this.filesArray = [];
};

/**
 * Show error message
 * @private
 */
Module.prototype._showErrorMessage = function(fileName, errorMessage) {
	var instance = this.$input.parsley();

	this.$input[0].classList.add(this.options.stateClasses.hasError);
	instance.addError('error', { message: errorMessage.replace('$0', fileName), updateClass: true });
};

/**
 * Remove error message
 * @private
 */
 Module.prototype._removeErrorMessage = function() {
	var instance = this.$input.parsley(),
		$notification = this.$element.find(this.options.domSelectors.notification);

	this.$input[0].classList.remove(this.options.stateClasses.hasError);
	instance.removeError('error');

	$notification.addClass(this.options.stateClasses.isHidden);
};

/**
 * Appends info about selected file type to the list
 * @param fileName string
 * @param fileType based on value passed from input field i.e. image/jpeg or application/zip
 * @param fileSize based on value passed from input field
 * @private
 */
Module.prototype._appendFileToList = function(fileName, fileType, fileSize) {
	var $list = this.$element.find(this.options.domSelectors.filesList),
		$element,
		compiledTemplate = template(this.$element.find(this.options.domSelectors.listItemTemplate).html()),
		fileTypeObject = this._getFileType(fileType),
		data = {
			fileName: fileName,
			icon: fileTypeObject[0].icon,
			fileExtension: fileTypeObject[0].label,
			fileSize: this._getFormattedFileSize(fileSize)
		};

	if (typeof data.fileSize === 'undefined') {
		data.fileSize = '';
	}

	$element = $(compiledTemplate(data));

	// Listen to delete click on list item
	$list.append($element);
	$list.addClass(this.options.stateClasses.isVisible);
	$element.find(this.options.domSelectors.deleteButton).on(events.removeFile, this._removeSingleFile.bind(this));
};

/**
 * Pass file size in bytes and get nicely formatted file size
 * @param fileSize in bytes
 * @returns {*} formatted file size i.e. 1.2 MB
 * @private
 */
Module.prototype._getFormattedFileSize = function(fileSize) {
	var aMultiples = ['KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'],
		nMultiple = 0,
		ret,
		nApprox = fileSize / 1024;

	// if the file is smaller than 1KB
	if (fileSize < 1024) {
		ret = ', ' + fileSize + ' Bytes';
	}

	for (nMultiple; nApprox >= 1; nApprox /= 1024, nMultiple++) {
		ret = ', ' + nApprox.toFixed(2) + ' ' + aMultiples[nMultiple];
	}

	return ret;
};

/**
 * Checks regex stored in filetypes array to match proper icon name
 * @param fileTypeString i.e. application/zip
 * @returns {{expression, icon}|*}
 * @private
 */
Module.prototype._getFileType = function(fileTypeString) {
	var fileType;

	fileType = filetypes.filter(function(element) {
		return element.expression.test(fileTypeString);
	});

	if (fileType.length) {
		return fileType;
	}

	return filetypes;
};

Module.prototype.getFiles = function(input) {
	var files = new Array(input.files.length);

	for (var i = 0; i < input.files.length; i++) {
		files[i] = input.files.item(i);
	}

	return files;
};

/* jshint ignore:start */
Module.prototype.setFiles = function(input, files) {
	var dataTransfer = new DataTransfer();

	for (var file of files) {
		dataTransfer.items.add(file);
	}

	input.files = dataTransfer.files;
};
/* jshint ignore:end */

/**
 * Removes single file from the files list
 * @param $event
 * @private
 */
Module.prototype._removeSingleFile = function($event) {
	var $el = $($event.currentTarget).parent(),
		index = $el.index();

	this.filesArray.splice(index- 1, 1);
	this.filesArrayToStore.splice(index- 1, 1);
	this._saveFilesToLocalStorage();

	if (this.filesArray.length === 0) {
		$el.parent().removeClass(this.options.stateClasses.isVisible);
	}

	$el.remove();

	var $lastElement = this.$element.find(this.options.domSelectors.filesListItem).filter(':last').find(this.options.domSelectors.deleteButton);

	if ($lastElement.length) {
		$lastElement.focus();
	} else {
		this.$input.focus();
		this._cleanLocalStorage();
		this.$input.prop('required', true);
	}

	// remove file from FileList
	// https://stackoverflow.com/questions/5632629/how-to-change-a-file-inputs-filelist-programmatically/68182158#68182158
	var files = this.getFiles(this.$input[0]);

	files.splice(index - 1, 1);

	this.setFiles(this.$input[0], files);
};

/**
 * Get image dimensions (width/height)
 * @private
 */
 Module.prototype._getImageDimension = function() {
	var getImageDimensions = __webpack_require__(173);

	getImageDimensions(this.$input[0])
		.then(function(size) {
			this.imageSizeWidth = size.width;
			this.imageSizeHeight = size.height;

			if (this.options.provideImageDimensionWarning) {
				this._getImageAspectRatio();
				this._showImageDimensionWarning();
			}
		}.bind(this))
		.catch(function() {
			return;
		}.bind(this));
};

/**
 * Get image aspect ratio and save it to the hidden field
 * @private
 */
 Module.prototype._getImageAspectRatio = function() {
	var $hiddenField = this.$element.find(this.options.domSelectors.hiddenTextfield);

	this.imageIsPortrait = false;

	if (this.imageSizeWidth < this.imageSizeHeight) {
		this.imageIsPortrait = true;
	}

	$hiddenField.val(this.imageIsPortrait);
};

/**
 * Show a warning if the image doesn't reach the min. requirements
 * @private
 */
 Module.prototype._showImageDimensionWarning = function() {
	if (this.imageIsPortrait && this.imageSizeWidth < (this.options.minImageWidth / 2)) {
		this._showNotification(this.options.i18n.minFileWidthWarningPortrait);
	} else if (!this.imageIsPortrait && this.imageSizeWidth < this.options.minImageWidth) {
		this._showNotification(this.options.i18n.minFileWidthWarning);
	}
};

/**
 * Show notification warning message
 * @private
 */
Module.prototype._showNotification = function(notificationText) {
	var $notification = this.$element.find(this.options.domSelectors.notification),
		$notificationMessage = $notification.find(this.options.domSelectors.notificationMessage);

	$notificationMessage.text(notificationText);
	$notification.removeClass(this.options.stateClasses.isHidden);
};

// -------------------------------------------------------------
// LOCAL STORAGE
// -------------------------------------------------------------
Module.prototype._setInitialState = function() {
	// append localStorageOptions key suffix based on upload id
	this.options.localStorageOptions.uploadedFilesKey += '-' + this.$input.attr('id');

	// check if we have values in the local storage
	this.localStorageEntry = this._readFromLocalStorage();

	this._checkItemExpiration();

	if (this.localStorageEntry) {
		this.setInputsBasedOnLocalStorageContent();
		return;
	}
};

Module.prototype._readFromLocalStorage = function() {
	var files = JSON.parse(window.localStorage.getItem(this.options.localStorageOptions.uploadedFilesKey));

	if (files) {
		return files;
	}

	return false;
};

Module.prototype._checkItemExpiration = function() {
	var localStorageEntryExpiryTime = this.localStorageEntry.expiryTime;

	// if the item is expired, delete the files
	if (Date.now() > localStorageEntryExpiryTime) {
		this._cleanLocalStorage();
	}
};

Module.prototype.setInputsBasedOnLocalStorageContent = function() {
	// restore files
	this.localStorageEntry.files.forEach(function(file) {
		this.filesArray.push(file);
		this.filesArrayToStore.push(file);
		this._appendFileToList(file.name, file.type, file.size);
	}.bind(this));

	// set required to false if files from the local storage are restored
	this.$input.prop('required', false);
};

Module.prototype._prepareFileForLocalStorage = function(file) {
	this._readFileForLocalStorage(file)
		.then(function() {
			this.filesArray.push(file);
			this._appendFileToList(file.name, file.type, file.size);
		}.bind(this))
		.catch(function() {
			return;
		}.bind(this));
};

Module.prototype._readFileForLocalStorage = function(file) {
	var promise,
		fileReader,
		fileDataUrl,
		base64String,
		fileToStore;

	promise = new Promise(function(resolve, reject) {
		if (!this.options.filesInLocalStorage) {
			resolve();
			return;
		}

		fileReader = new FileReader();

		// read file as data url
		fileReader.readAsDataURL(file);

		fileReader.addEventListener('load', function() {
			fileDataUrl = fileReader.result;
			base64String = fileDataUrl.split(',')[1];
			fileToStore = {
				nameOfInput: this.$input.attr('name'),
				value: base64String,
				size: file.size,
				name: file.name,
				type: file.type
			};

			// push the base64 encoded file to the array
			this.filesArrayToStore.push(fileToStore);

			this._saveFilesToLocalStorage()
				.then(function() {
					resolve();

				}.bind(this))
				.catch(function() {
					// remove not added file from the array
					this.filesArrayToStore = this.filesArrayToStore.filter(function(file) {
						return file !== fileToStore;
					}.bind(this));

					// show an error message if the max local storage size is reached
					this._showErrorMessage(file.name, this.options.i18n.maxLocalStorageSizeError);

					reject();

				}.bind(this));

		}.bind(this));
	}.bind(this));

	return promise;
};

Module.prototype._saveFilesToLocalStorage = function() {
	var promise,
		data = {};

	promise = new Promise(function(resolve, reject) {
		// check if the max local storage size is reached
		try {
			this._removeErrorMessage();

			// add files and expiry time to the data object
			data.files = this.filesArrayToStore;
			data.expiryTime = Date.now() + (this.options.localStorageOptions.cacheExpirationTime * 1000);

			window.localStorage.setItem(this.options.localStorageOptions.uploadedFilesKey, JSON.stringify(data));

			this.hasError = false;

			resolve();

		} catch (e) {
			this.hasError = true;

			reject();
		}
	}.bind(this));

	return promise;
};

Module.prototype._cleanLocalStorage = function() {
	window.localStorage.removeItem(this.options.localStorageOptions.uploadedFilesKey);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 255 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 256 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


new Promise(function(resolve) { resolve(); }).then((function(require) {

	var $ = __webpack_require__(0),
		uploadedImageDimension = __webpack_require__(173);

	__webpack_require__(3);

	var validateImageDimensions = function($element) {

		var $deferred = $.Deferred(),
			options = JSON.parse($element.closest('[data-init="upload"]').attr('data-upload-options')),
			requiredDimensionsWidth = options.requiredImageDimensions.width,
			requiredDimensionsHeight = options.requiredImageDimensions.height;

		uploadedImageDimension($element[0])
			.then(function(size) {

				if (size.width < requiredDimensionsWidth || size.height < requiredDimensionsHeight) {
					$deferred.reject();
				} else {
					$deferred.resolve(true);
				}
			})
			.catch(function() {
				return;
			});

		return $deferred.promise();

	};

	window.Parsley.addValidator('customRequiredImageDimensionValidation', {
		requirementType: 'string',
		validateString: function(value, state, parsley) {
			return validateImageDimensions(parsley.$element);
		}
	});

}).bind(null, __webpack_require__)).catch(__webpack_require__.oe);


/***/ }),
/* 257 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(258);


/***/ }),
/* 258 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.8 - Captcha
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(259);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'captcha',
	events = {},
	defaults = {
		domSelectors: {
			container: '[data-' + name + '="container"]',
			loading: '[data-' + name + '="loading"]'
		},
		stateClasses: {}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	this._initLoading();
	this._createCaptchaContainer();
};

Module.prototype._initLoading = function() {
	var source = $('[data-template="preloader"]').html(),
		$loadingAnimationWrapper = this.$element.find(this.options.domSelectors.loading);

	$loadingAnimationWrapper.append($(source));
	$loadingAnimationWrapper.find('div:first-child').addClass('var_small');
};

Module.prototype._createCaptchaContainer = function() {
	var captcha = document.createElement('div'),
		script = document.createElement('script');

	// Add attribute and class
	captcha.setAttribute('data-sitekey', this.options.siteKey);
	captcha.setAttribute('data-size', window.innerWidth < 356 ? 'compact' : 'normal');
	captcha.setAttribute('async', '');
	captcha.setAttribute('defer', '');
	captcha.className = 'g-recaptcha';

	// Create script and load api
	script.src = this.options.apiUrl;

	// Append elements
	this.$element.find(this.options.domSelectors.container).append(captcha);
	this.$element.append(script);
};

/**
 * Unbind events, remove data, custom teardown
 * @method
 * @public
 */
Module.prototype.destroy = function() {
	// Unbind events, remove data
	SuperClass.prototype.destroy.apply(this);

	// Custom teardown (removing added DOM elements etc.)
	// If there is no need for a custom teardown, this method can be removed
	/* jshint ignore:start */
	grecaptcha.reset();
	/* jshint ignore:end */
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 259 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 260 */,
/* 261 */,
/* 262 */,
/* 263 */,
/* 264 */,
/* 265 */,
/* 266 */,
/* 267 */,
/* 268 */,
/* 269 */,
/* 270 */,
/* 271 */,
/* 272 */,
/* 273 */,
/* 274 */,
/* 275 */,
/* 276 */,
/* 277 */,
/* 278 */,
/* 279 */,
/* 280 */,
/* 281 */,
/* 282 */,
/* 283 */,
/* 284 */,
/* 285 */,
/* 286 */,
/* 287 */,
/* 288 */,
/* 289 */,
/* 290 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(291);


/***/ }),
/* 291 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 292 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(293);


/***/ }),
/* 293 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 294 */,
/* 295 */,
/* 296 */,
/* 297 */,
/* 298 */,
/* 299 */,
/* 300 */,
/* 301 */,
/* 302 */,
/* 303 */,
/* 304 */,
/* 305 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(306);


/***/ }),
/* 306 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M64 - General Supply Teaser
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(307);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	scrollAnimation = __webpack_require__(52);

var name = 'generalsupplyteasers',
	events = {},
	defaults = {
		domSelectors: {
			title: '[data-' + name + '="title"]',
			lead: '[data-' + name + '="lead"]',
			listItem: '[data-' + name + '="list_item"]'
		},
		stateClasses: {}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	this.initializeAnimations();
};

Module.prototype.initializeAnimations = function() {

	var hasAnimations = this.$element.hasClass('var_scroll_animation');

	if (!hasAnimations) {
		return;
	}

	scrollAnimation({
		$trigger: this.options.domSelectors.title,
		selectors: [
			this.options.domSelectors.title,
			this.options.domSelectors.lead,
			this.options.domSelectors.listItem
		],
		$parent: this.$element
	});

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 307 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 308 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(309);


/***/ }),
/* 309 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 310 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(311);


/***/ }),
/* 311 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 312 */,
/* 313 */,
/* 314 */,
/* 315 */,
/* 316 */,
/* 317 */,
/* 318 */,
/* 319 */,
/* 320 */,
/* 321 */,
/* 322 */,
/* 323 */,
/* 324 */,
/* 325 */,
/* 326 */,
/* 327 */,
/* 328 */,
/* 329 */,
/* 330 */,
/* 331 */,
/* 332 */,
/* 333 */,
/* 334 */,
/* 335 */,
/* 336 */,
/* 337 */,
/* 338 */,
/* 339 */,
/* 340 */,
/* 341 */,
/* 342 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(343);


/***/ }),
/* 343 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M01 - Timetable Input Form
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(344);

var $ = __webpack_require__(0),
	debounce = __webpack_require__(29),
	SuperClass = __webpack_require__(1),
	cssTransitionEnd = __webpack_require__(4),
	gsap = __webpack_require__(11).gsap,
	speculative = __webpack_require__(30),
	lscache = __webpack_require__(56),
	ConsentHelper = __webpack_require__(26);

var name = 'timetableInputForm',
	$document = $(document),
	events = {
		timetableInputFormFocus: 'timetable-input-form-focus',
		timetableInputFormBlur: 'timetable-input-form-blur',
		timetableInputFormPrefill: 'timetable-input-form-prefill',
		autocompleteSetOption: 'setOptionValue.autocomplete',
		datepickerDidInit: 'datepickerDidFinishInit',
		toggleButtonSetValue: 'setValueActive.toggleButton',
		focusHeaderLogo: 'focus.headerLogo'
	},
	defaults = {
		domSelectors: {
			viaLink: '[data-' + name + '="add_via"]',
			fieldsetSecondLine: '[data-' + name + '="fieldset_second_line"]',
			buttonWrapper: '[data-' + name + '="button_wrapper"]',
			form: '[data-init="formValidation"]:first',
			fromField: '[data-' + name + '="from_field"]',
			toField: '[data-' + name + '="to_field"]',
			viaField: '[data-' + name + '="via_field"]',
			swapButton: '[data-' + name + '="swap_button"]',
			timeField: '[data-init="textfield timepicker"]',
			deleteButton: '[data-' + name + '="delete_button"]',
			viaWrapper: '[data-' + name + '="via_wrapper"]',
			viasHidden: '[data-' + name + '="vias_hidden"]',
			extendedSearchButton: '[data-' + name + '="extendedsearch_button"]',
			extendedSearch: '[data-' + name + '="extendedsearch"]',
			datepicker: '[data-init="datepicker"]',
			datepickerInput: '[data-datepicker="input"]',
			toggleButton: '[data-init="togglebutton"]'
		},
		stateClasses: {
			isViaOpened: 'is_via_opened',
			isViaWrapperVisible: 'is_via_wrapper_visible',
			isExpanded: 'is_expanded',
			isFirstLineExpanded: 'is_first_line_expanded',
			isSwapped: 'is_swapped',
			isForExtendedSearch: 'is_for_extended_search',
			isViaTime: 'mod_timepicker_via',
			isViaButtonVisible: 'via_button_visible',
			fadeViaWrapperOut: 'fade_via_wrapper_out'
		},
		isForExtendedSearch: false,
		setFocus: false,
		cacheFormValues: false,
		cacheFormValuesExpirationTime: 3, // in minutes
		keysToIgnoreForExpanding: [
			8, 9, 13, 16, 17, 18, 19, 20, 27, 33, 34, 35, 36, 37, 38,
			39, 40, 46, 91, 92, 93, 106, 107, 109, 112, 113, 114, 115,
			116, 117, 118, 119, 120, 121, 123, 144, 145, 186, 187, 188
		]
	},
	isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints,
	data = {},
	localStorageOptions = {
		bucketName: 'timetable',
		dataKey: 'data'
	};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Init by lazy loading the dependencies first and then do the actual init
 */
Module.prototype.init = function() {

	// load ParsleyJS script async and give webpack ability to handle it correctly
	__webpack_require__.e/* require.ensure */(3/* duplicate */).then((function(require) {

		__webpack_require__(3);

		// call real init
		this._init();

	}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);

};

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype._init = function() {

	// Check for presence of market research overlay. In this case we are not going to set the focus
	// inside of the the timetable input form
	var marketResearchOverlay = document.querySelector('.mod_market_research_overlay'),
		metaPublisher = document.querySelector('meta[name="identifier"]'),
		metaPublisherIdentifier = '';

	if (metaPublisher) {
		metaPublisherIdentifier = '_' + metaPublisher.getAttribute('content');
	}

	// Create performance mark which will be interpreted in Speedcurve
	// Output publisher instance if meta element is present and has content
	if (window.performance.mark !== undefined) {
		performance.mark('timetable_input_form_ready');
		performance.mark('timetable_input_form_ready' + metaPublisherIdentifier);
	}

	// Create the same marks in the real user metrics monitoring if the library is present
	if (window.LUX !== undefined) {
		window.LUX.mark('timetable_input_form_ready');
		window.LUX.mark('timetable_input_form_ready' + metaPublisherIdentifier);
	}

	// store to options, if the module is used inside extended search (Webshop) or not
	if (this.$element.hasClass(this.options.stateClasses.isForExtendedSearch)) {
		this.options.isForExtendedSearch = true;
	}

	this.$form = this.$element.find(this.options.domSelectors.form);
	this.$from = this.$element.find(this.options.domSelectors.fromField);
	this.$to = this.$element.find(this.options.domSelectors.toField);
	this.$fromAndToField = this.$from.add(this.$to);
	this.$via = this.$element.find(this.options.domSelectors.viaField);
	this.$swap = this.$element.find(this.options.domSelectors.swapButton);
	this.$viaWrappers = this.$element.find(this.options.domSelectors.viaWrapper);
	this.$viaWrapper = $(this.$viaWrappers[0]);
	this.$viaLink = this.$element.find(this.options.domSelectors.viaLink);
	this.$viasHidden = this.$element.find(this.options.domSelectors.viasHidden);
	this.$viaFields = this.$viaWrappers.find('.mod_timetable_input_form_viafield input');
	this.$datepicker = this.$element.find(this.options.domSelectors.datepicker);
	this.$datepickerInput = this.$datepicker.find('[data-datepicker="input"]');
	this.$primaryTimepicker = this.$datepicker.siblings(this.options.domSelectors.timeField);
	this.$timepickerInput = this.$primaryTimepicker.find('input');
	this.$toggleButton = this.$element.find(this.options.domSelectors.toggleButton);

	this.$element.on('click.' + this.uuid, this.options.domSelectors.swapButton, this._swapNames.bind(this));

	if (!isTouchDevice && this.options.setFocus && marketResearchOverlay === null) {

		this.$element.on('keyup.' + this.uuid, this._expandForm.bind(this));

		// we need a timeout here. This makes sure, autocomplete module is
		// initialized before focus is set.
		setTimeout(function() {
			this.$from.focus();
		}.bind(this), 0);

		ConsentHelper.consentChangeCallback(function() {
			this.$from.focus();
		}.bind(this));

	} else if (!isTouchDevice) {

		this.$element.one('focusin.' + this.uuid, function(event, data) {

			if (data && data === 'onlyFocus') {

				// we need a timeout here. This makes sure, autocomplete module is
				// initialized before focus is set.
				setTimeout(function() {
					this.$from.focus();
				}.bind(this), 0);

				ConsentHelper.consentChangeCallback(function() {
					this.$from.focus();
				}.bind(this));

				this.$element.on('keyup.' + this.uuid, this._expandForm.bind(this));

			} else {
				this._expandForm();
			}

		}.bind(this));

	} else {

		// Here we assume that we are on a touch device
		// Unfortunately voice over is triggering focusin
		// when focus reaches a button. we therefore have
		// to listen exclusively on the two input fields
		this.$fromAndToField.one('focusin.' + this.uuid, function(event, data) {
			if (!data) {
				this._expandForm();
				this.$fromAndToField.off('focusin.' + this.uuid);
			}
		}.bind(this));

		ConsentHelper.consentChangeCallback(function() {
			$document.trigger(events.focusHeaderLogo);
		}.bind(this));

	}

	this.$element.one(cssTransitionEnd + '.' + this.uuid, function() {
		$document.trigger('flying-focus_trigger');
	}.bind(this));

	this.$element.on('focusin.' + this.uuid, function() {
		$document.trigger(events.timetableInputFormFocus);
	});

	this.$element.on('focusout.' + this.uuid, function() {
		$document.trigger(events.timetableInputFormBlur);
	});

	if (this.options.isForExtendedSearch) {
		this.$element
			.on('click.' + this.uuid, this.options.domSelectors.viaLink, this._addViaField.bind(this))
			.on('click.' + this.uuid, this.options.domSelectors.deleteButton, this._deleteViaField.bind(this))
			.on('click.' + this.uuid, this.options.domSelectors.extendedSearchButton, this._showExtendedSearch.bind(this));
		this._collectDefaultViaFieldsIds();
	} else {

		this.$element
			.on('click.' + this.uuid, this.options.domSelectors.viaLink, this._toggleViaField.bind(this))
			.on('click.' + this.uuid, this.options.domSelectors.extendedSearchButton, this._setExtendedSearchParameter.bind(this));

		this.$element.on('submit.' + this.uuid, this.$element[0].getElementsByTagName('form')[0], function(evt) {

			this._updateHiddenViasField();

			// Save how long it took to perform the journey request
			// to Speedcurve LUX metric
			if (window.LUX !== undefined) {
				window.LUX.mark('time_to_journey_request');
			}

			this.$datepickerInput.trigger('blur');

			if (!this.options.cacheFormValues) {
				return;
			}

			if (!this.didSaveInputValues) {
				evt.preventDefault();

				// save values to local storage
				this._saveInputValues();
				this.didSaveInputValues = true;

				// trigger form submit again
				$(this.$element[0].getElementsByTagName('form')[0]).trigger('submit');
			}

			return;

		}.bind(this));
	}

	if (this.$swap.length) {

		// Listen to input changes in from and to field and set swap button state
		this.$from.on('keyup.' + this.uuid, this._setSwapButtonState.bind(this));
		this.$to.on('keyup.' + this.uuid, this._setSwapButtonState.bind(this));

		// Set Initial SwapButton State
		this._setSwapButtonState();

	}

	this._setInitialAriaState();

	if (!this.$element.hasClass(this.options.stateClasses.isExpanded)) {
		this._hideElementsInitially();
	}

	this._listenToFormValidation();

	// listen if someone wants to prefill the timetable input form
	$document.on(events.timetableInputFormPrefill + '.' + this.uuid, function(evt, data, callback) {

		var $fieldToFocus;

		// show via field eventually
		if (data.showVia) {
			this._showViaField();
		}

		// fill in the passed values
		this.$from.val(data.start);
		this.$to.val(data.end);
		this.$via.val(data.via);

		// determine which field to focus
		if (!data.start) {
			$fieldToFocus = this.$from;
		} else if (!data.end) {
			$fieldToFocus = this.$to;
		} else {
			if (data.showVia) {
				$fieldToFocus = this.$via;
			}
		}

		// expand the form
		this._expandForm(null, callback, $fieldToFocus);

	}.bind(this));

	// init work to restore and save input values to local storage
	this._initCache();

};

/**
 * Set Initial Aria State
 * @private
 */
Module.prototype._setInitialAriaState = function() {

	var i = this.$viaWrappers.length,
		$wrapper;

	// Hide time field and delete buttons from screenreaders
	// in case where we only have one via field
	if (!this.options.isForExtendedSearch) {
		this._setAriaHiddenState(this.$viaWrapper.find(this.options.domSelectors.timeField));
		this._setAriaHiddenState(this.$viaWrapper.find(this.options.domSelectors.deleteButton));
	}

	// Hide all visually hidden via wrappers from screenreaders as well
	while (i--) {

		$wrapper = $(this.$viaWrappers[i]);

		if (!$wrapper.hasClass(this.options.stateClasses.isViaWrapperVisible)) {
			this._setAriaHiddenState($wrapper);
		}

	}

};

/**
 * Set display to none on visually hidden elements so they are not focusable
 */
Module.prototype._hideElementsInitially = function() {

	// hide the datepicker initially. Since the elements inside
	// datepicker are lazyloaded (e.g. arrows), we're not able to set
	// tabindex to -1 at this point. Furthermore, timepicker and toggle button
	// have their inputs nested inside the main modules, so we don't have
	// an easy way to select them to set the tabindex. Thus we're just
	// hiding them.
	this.datepickerDisplayInitValue = this.$datepicker.css('display');
	this.timepickerDisplayInitValue = this.$primaryTimepicker.css('display');
	this.toggleButtonDisplayInitValue = this.$toggleButton.css('display');

	window.requestAnimationFrame(function() {
		this.$datepicker.css({ 'display': 'none' });
		this.$primaryTimepicker.css({ 'display': 'none' });
		this.$toggleButton.css({ 'display': 'none' });
	}.bind(this));

};

/**
 * Set display to original value for elements that were initially hidden
 */
Module.prototype._showInitiallyHiddenElements = function() {

	// show the elements again that we did initially hide
	window.requestAnimationFrame(function() {
		this.$datepicker.css({ 'display': this.datepickerDisplayInitValue });
		this.$primaryTimepicker.css({ 'display': this.timepickerDisplayInitValue });
		this.$toggleButton.css({ 'display': this.toggleButtonDisplayInitValue });
	}.bind(this));

};

/**
 * Set Aria Hidden State
 * @param element
 */
Module.prototype._setAriaHiddenState = function($element) {
	$element.attr(
		{
			'aria-hidden': 'true',
			'role': 'presentation'
		}
	);
};

/**
 * Remove Aria Hidden State
 * @param element
 * @private
 */
Module.prototype._removeAriaHiddenState = function($element) {
	$element.removeAttr('aria-hidden role');
};

/**
 * Set pre-fetch Link to pre-fetch timetable in
 * supported browsers
 * @private
 */
Module.prototype._setPreFetchLink = function() {

	if (this.options.preloadURL !== undefined && !this.$element.data('preFetchUrlSet')) {

		speculative({
			href: this.options.preloadURL,
			rel: 'prefetch',
			as: 'html'
		});

		this.$element.data('preFetchUrlSet', 'true');

	}

};

/**
 * Set Swap Button Status (enabled/disabled)
 * @private
 */
Module.prototype._setSwapButtonState = function() {

	debounce(function() {

		if (this.$from.val().trim() === '' || this.$to.val().trim() === '') {
			this.$swap.prop('disabled', true);
		} else {
			this.$swap.prop('disabled', false);
		}

	}.bind(this), 150)();

};

/**
 * Creates array that will store all input ids for later assignment, so items
 * can keep id's that resemble via position in the list.
 * @private
 */
Module.prototype._collectDefaultViaFieldsIds = function() {
	this.ids = [];

	this.$viaFields.each(function(index, element) {
		this.ids.push($(element).attr('id'));
	}.bind(this));
};

/**
 * Update Hidden Vias Field
 * @private
 */
Module.prototype._updateHiddenViasField = function() {
	var inputValue = this.$via.val();

	this.$viasHidden.val('["' + inputValue + '"]');
};

/**
 * Function is showing "via" input form.
 * @param event {jQuery.Event} The original click-event-object.
 * @private
 */
Module.prototype._toggleViaField = function(event) {

	if (event) {
		event.preventDefault();
	}

	if (this.$element.hasClass(this.options.stateClasses.isViaOpened)) {

		this.$element.removeClass(this.options.stateClasses.isViaOpened);
		this._setAriaHiddenState(this.$viaWrapper);

	} else {

		this.$element.addClass(this.options.stateClasses.isViaOpened);
		this.$via.focus();
		this._removeAriaHiddenState(this.$viaWrapper);

	}
};

Module.prototype._showViaField = function() {

	if (!this.$element.hasClass(this.options.stateClasses.isViaOpened)) {

		this.$element.addClass(this.options.stateClasses.isViaOpened);
		this._removeAriaHiddenState(this.$viaWrapper);

	}
};

/**
 * Goes through all visible via fields and replaces attributes that contain
 * via field id. So element with an id "viaField1" which would get new id
 * "viaField2" will have to have it's children traversed and every occurence
 * of the string "viaField1" would be changed to "viaField2" and so on, to keep
 * the indexes in the proper order for the backend.
 * @private
 */
Module.prototype._changeIdAndNameForViaFields = function() {
	var $visibleViaFields = this.$viaWrappers.filter('.' + this.options.stateClasses.isViaWrapperVisible);

	$visibleViaFields.each(function(index, viaField) {
		var $el = $(viaField).find('.mod_timetable_input_form_viafield'),
			$input = $el.find('input'),
			$label = $el.find('label'),
			$button = $el.find('.mod_autocomplete_clear_button'),
			$error = $el.find('.mod_error_message'),
			$errorMessage = $el.find('.mod_error_message_text'),
			oldId = $input.attr('id'),
			newId = this.ids[index];

		// replace only if id's are different
		if (oldId !== newId) {

			$input.attr('name', newId)
				.attr('id', newId)
				.attr('data-parsley-errors-container', $input.attr('data-parsley-errors-container').replace(oldId, newId))
				.attr('aria-describedby', $input.attr('aria-describedby').replace(oldId, newId));
			$label.attr('id', $label.attr('id').replace(oldId, newId));
			$button.attr('aria-controls', $button.attr('aria-controls').replace(oldId, newId));
			$error.attr('id', $error.attr('id').replace(oldId, newId));
			$errorMessage.attr('id', $errorMessage.attr('id').replace(oldId, newId));
		}
	}.bind(this));
};

/**
 * Function is adding up to 5 via fields.
 * @param event {jQuery.Event} The original click-event-object.
 * @private
 */
Module.prototype._addViaField = function(event) {

	var nextVisibleFieldIndex = this.$viaWrappers.filter('.' + this.options.stateClasses.isViaWrapperVisible).length,
		$viaWrapperToShow = this.$viaWrappers.eq(nextVisibleFieldIndex);

	event.preventDefault();

	// make sure that the parent wrapper has this class to indicate that via's are open
	if (!this.$element.hasClass(this.options.stateClasses.isViaOpened)) {
		this.$element.addClass(this.options.stateClasses.isViaOpened);
	}

	// Proceed only if there's a button to add, otherwise hide "add via" button
	if (nextVisibleFieldIndex <= this.$viaWrappers.length - 1) {

		$viaWrapperToShow
			.addClass(this.options.stateClasses.isViaWrapperVisible)
			.find(this.options.domSelectors.viaField)
			.focus();

		this._removeAriaHiddenState($viaWrapperToShow);

	}

	if (nextVisibleFieldIndex === this.$viaWrappers.length - 1) {
		this.$viaLink.removeClass(this.options.stateClasses.isViaButtonVisible);
	}

	// Adjust via fields id's
	this._changeIdAndNameForViaFields();
};

/**
 * Function to remove a via field
 * @param event {jQuery.Event} The original click-event-object.
 * @private
 */
Module.prototype._deleteViaField = function(event) {
	event.preventDefault();

	var $viaWrapper,
		$removedElement;

	// get parent via wrapper
	$viaWrapper = $(event.currentTarget).parent('div.' + this.options.stateClasses.isViaWrapperVisible);

	// remove visible class from wrapper
	$viaWrapper.addClass(this.options.stateClasses.fadeViaWrapperOut);

	if (cssTransitionEnd) {
		$viaWrapper.one(cssTransitionEnd + '.' + this.uuid, function() {

			// remove class that wrapper did finish fading out and that it is visible
			$viaWrapper
				.removeClass(this.options.stateClasses.fadeViaWrapperOut)
				.removeClass(this.options.stateClasses.isViaWrapperVisible);

			this._setAriaHiddenState($viaWrapper);

			// reset value of via field
			$viaWrapper.find(this.options.domSelectors.viaField).val('');

			// remove the element from dom and store it to var
			$removedElement = $viaWrapper.remove();

			// refresh the cached wrappers
			this.$viaWrappers = this.$element.find(this.options.domSelectors.viaWrapper);

			// append it to the last via again
			this.$viaWrappers.last().after($removedElement);

			// refresh the cashed wrappers again
			this.$viaWrappers = this.$element.find(this.options.domSelectors.viaWrapper);

			// make sure the via button is shown
			this.$viaLink.addClass(this.options.stateClasses.isViaButtonVisible);

			// make sure the via button gets focus
			this.$viaLink.focus();

			// Adjust via fields id's
			this._changeIdAndNameForViaFields();

		}.bind(this));
	} else {
		// remove the element from dom and store it to var
		$removedElement = $viaWrapper.remove();
	}
};

/**
 * Function is responsible for opening element animation.
 * @private
 */
Module.prototype._expandForm = function(evt, callback, $fieldToFocus) {

	if (evt) {

		if ($.inArray(evt.keyCode, this.options.keysToIgnoreForExpanding) !== -1) {
			return;
		}

		this.$element.off('keyup.' + this.uuid);

	}

	if (!this.$element.hasClass(this.options.stateClasses.isExpanded)) {

		var startHeight = this.$element.outerHeight(),
			$fieldsetSecondLine = this.$element.find(this.options.domSelectors.fieldsetSecondLine),
			$buttonWrapper = this.$element.find(this.options.domSelectors.buttonWrapper);

		this._showInitiallyHiddenElements();

		requestAnimationFrame(function() {
			$fieldsetSecondLine.css('height', 'auto');
			this.$element
				.addClass(this.options.stateClasses.isExpanded)
				.css({
					'will-change': 'height',
					'transform': 'translateZ(0)'
				});
		}.bind(this));

		gsap.fromTo(this.$element, {
				height: startHeight
			},
			{
				duration: 0.2,
				height: 'auto',
				onComplete: function() {

					this._updatePlaceholder();
					this._setPreFetchLink();

					this.$element
						.addClass(this.options.stateClasses.isFirstLineExpanded)
						.css({
							'height': 'auto',
							'transform': 'none',
							'will-change': 'unset'
						});

					if ($fieldToFocus) {
						requestAnimationFrame(function() {
							$fieldToFocus.focus();
						});
					}

				}.bind(this)
			}
		);

		gsap.to($buttonWrapper, {
			duration: 0.2,
			x: '+=50',
			'opacity': '0',
			onComplete: function() {

				gsap.set($buttonWrapper, {'clearProps': 'transform'});

			}.bind(this)

		});

	} else {
		if (callback) {
			setTimeout(function() {
				callback();

				if ($fieldToFocus) {
					setTimeout(function() {
						$fieldToFocus.focus();
					}, 0);
				}
			}, 0);
		}
	}
};

/**
 * Function is swapping station names between from/to fields.
 * @private
 */
Module.prototype._swapNames = function() {
	var fromValue = this.$from.val(),
		toValue = this.$to.val();

	this.$from.val(toValue);
	this.$to.val(fromValue);

	this.$swap.toggleClass(this.options.stateClasses.isSwapped);
	this._reverseViasOrder();
};

/**
 * After swapping from/to fields, we need to reverse the order of the vias
 */
Module.prototype._reverseViasOrder = function() {
	var $visibleViaFields = this.$viaWrappers.filter('.' + this.options.stateClasses.isViaWrapperVisible),
		viaField,
		$visibleViaFieldsCopy = $visibleViaFields.clone(),
		viaStationName,
		viaTime;

	// we only need to proceed, if we have more than 1 via
	if ($visibleViaFields.length < 2) {
		return;
	}

	// iterate via fields and apply new values
	$visibleViaFieldsCopy.each(function(index, viaFieldCopy) {

		// get station name and time from copied via fields
		viaStationName = $(viaFieldCopy).find(this.options.domSelectors.viaField).val();
		viaTime = $(viaFieldCopy).find(this.options.domSelectors.timeField).find('input').val();

		// get new via field to set new values to
		viaField = $visibleViaFields[$visibleViaFields.length - index - 1];

		// set new values
		$(viaField).find(this.options.domSelectors.viaField).val(viaStationName);
		$(viaField).find(this.options.domSelectors.timeField).find('input').val(viaTime);

	}.bind(this));

};

/**
 * Function updating placeholder after animation.
 * @private
 */
Module.prototype._updatePlaceholder = function() {
	var placeholder = this.$via.attr('placeholder');

	if (!this.$from.val()) {
		this.$from.attr('placeholder', placeholder);
	}

	if (!this.$to.val()) {
		this.$to.attr('placeholder', placeholder);
	}
};

Module.prototype._showExtendedSearch = function(event) {
	if (event) {
		event.preventDefault();
	}

	this.$element.find(this.options.domSelectors.extendedSearchButton).addClass('var_hidden');

	$(this.options.domSelectors.extendedSearch).removeClass('var_hidden');
};

Module.prototype._setExtendedSearchParameter = function(event) {

	event.preventDefault();

	var $hiddenSearchField = this.$form.find('[name="suche"]');

	if ($hiddenSearchField.length > 0) {
		$hiddenSearchField.attr({
			'name': 'erweiterteSuche'
		});
	}

	this.$form.submit();

};

// Listen to form validation errors
Module.prototype._listenToFormValidation = function() {

	if (this.$form.length > 0) { // this handles the exception we have on ws02 where the form is placed outside
		this.$form.parsley().on('form:error', function(form) {
			this._handleFormValidationError(form);
		}.bind(this));
	}

};

// Set focus to specific field after form validation error
Module.prototype._handleFormValidationError = function(form) {

	var itemsCount = form.fields.length,
		iterator = 0,
		formField;

	// iterate all form fields and find the first element with validation error.
	// we explicitly don't use while(itemsCount--) here, because in most cases
	// we will get an error on one of the first fields. So iterating from first
	// to last item makes more sense
	for (iterator; iterator < itemsCount; iterator++) {

		formField = form.fields[iterator];

		if (formField.validationResult !== true) {
			break;
		}

		formField = undefined;

	}

	// if formField isn't defined, we return since we don't have an invalid field
	if (formField === undefined) {
		return;
	}

	// at this point we have an invalid field. We set the focus into it
	$(formField.element).focus();

};

// init work to restore and save input values to local storage
Module.prototype._initCache = function() {
	if (!this.options.cacheFormValues) {
		return;
	}

	// instance properties
	this.didSaveInputValues = false;
	this.didChangeTime = false;

	// lsCache-specific
	lscache.setBucket(localStorageOptions.bucketName);
	lscache.flushExpired();

	// input handling
	this._listenToInputChangesForCache();
	this._restoreInputValues();
};

Module.prototype._listenToInputChangesForCache = function() {

	// get informed when time changes
	this.$element.one('input', this.options.domSelectors.timeField, function() {
		this.didChangeTime = true;
	}.bind(this));

};

// Save input values to local storage
Module.prototype._saveInputValues = function() {
	var values = {},
		savedValues;

	if (!lscache.supported()) {
		return;
	}

	savedValues = lscache.get(localStorageOptions.dataKey);

	// set from and to values
	values[this.options.domSelectors.fromField] = this.$from.val();
	values[this.options.domSelectors.toField] = this.$to.val();

	// set via
	if (this.$via.val().trim().length > 0) {
		values[this.options.domSelectors.viaField] = this.$via.val();
	} else {
		values[this.options.domSelectors.viaField] = false;
	}

	// set date
	values[this.options.domSelectors.datepicker] = this.$datepickerInput.val();

	// set time
	if (this.didChangeTime) {
		values[this.options.domSelectors.timeField] = this.$timepickerInput.val();
	} else if (savedValues && savedValues[this.options.domSelectors.timeField]) {
		values[this.options.domSelectors.timeField] = savedValues[this.options.domSelectors.timeField];
	}

	// set toggle button
	values[this.options.domSelectors.toggleButton] = this.$toggleButton.find('input:checked').val();

	// save values
	lscache.set(localStorageOptions.dataKey, values, this.options.cacheFormValuesExpirationTime);

};

// Restore input values from local storage
Module.prototype._restoreInputValues = function() {
	var values;

	if (!lscache.supported()) {
		return;
	}

	// get values
	values = lscache.get(localStorageOptions.dataKey);

	if (!values) {
		return;
	}

	// expand form
	this._expandForm(null, null, this.$from);

	// restore from value. on load, we want to have the text in the input selected, but the
	// autocomplete should be hidden.
	this.$from.trigger(events.autocompleteSetOption, { 'hideAutocompleteOnFocus': true});
	this.$from.val(values[this.options.domSelectors.fromField]);
	this.$from.one('blur', function() {
		this.$from.trigger(events.autocompleteSetOption, { 'hideAutocompleteOnFocus': false});
	}.bind(this));

	// restore to value
	this.$to.val(values[this.options.domSelectors.toField]);

	// highlight text inside from field
	this.$from[0].select();

	// toggle the via field and set value
	if (values[this.options.domSelectors.viaField]) {
		this.$via.val(values[this.options.domSelectors.viaField]);
		this._toggleViaField(null);
	}

	// restore date
	this.$element.one(events.datepickerDidInit, this.options.domSelectors.datepicker, function() {
		this.$datepickerInput.val(values[this.options.domSelectors.datepicker]).trigger('input');
	}.bind(this));

	// restore time
	if (values[this.options.domSelectors.timeField]) {
		this.$timepickerInput.val(values[this.options.domSelectors.timeField]);
	}

	// restore ab/an toggle
	if (values[this.options.domSelectors.toggleButton]) {
		this.$toggleButton.trigger(events.toggleButtonSetValue, values[this.options.domSelectors.toggleButton]);
	}
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 344 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 345 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(346);


/***/ }),
/* 346 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 347 */,
/* 348 */,
/* 349 */,
/* 350 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(351);


/***/ }),
/* 351 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M00 - Geolocation
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2020.
 */



__webpack_require__(352);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	$document = $(document),
	Geolocation = __webpack_require__(353);

__webpack_require__(3);

var name = 'geolocation',
	events = {
		geolocationInitState: 'geolocationInitState',
		geolocationHasSucceeded: 'geolocationHasSucceeded',
		geolocationHasFailed: 'geolocationHasFailed',
		setCoordinatesInHiddenFields: 'setCoordinatesInHiddenFields'
	},
	defaults = {
		domSelectors: {
			geoButton: '[data-' + name + '="button"]',
			hiddenLongitude: '[data-' + name + '="longitude"]',
			hiddenLatitude: '[data-' + name + '="latitude"]',
			geoTarget: '[data-' + name + '="target"]'
		},
		dataAttributes: {},
		stateClasses: {
			hasError: 'has_error',
			isAnimating: 'is_animating'
		},
		geoLocationRequestOptions: {
			timeout: 10 * 1000,
			enableHighAccuracy: false,
			maximumAge: 30000 // 30 seconds
		},
		i18n: {
			geolocationSuccessText: 'We could successfully detect your location.',
			geolocationErrorText: 'Sorry, we could not detect your current location. Please verify the permissions in your browser settings.',
			geolocationSuccessVisualLabel: 'My location'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ---------------------------------------------------------------------------
// INIT
// ---------------------------------------------------------------------------
Module.prototype.init = function() {
	this.cacheElements();
	this.bindEventListeners();
};

Module.prototype.cacheElements = function() {
	this.$geoButton = this.$element.find(this.options.domSelectors.geoButton);
	this.$hiddenLongitude = this.$element.find(this.options.domSelectors.hiddenLongitude);
	this.$hiddenLatitude = this.$element.find(this.options.domSelectors.hiddenLatitude);
	this.$geoLocationTarget = this.$element.siblings(this.options.domSelectors.geoTarget);
	this.$geoLocationSuccessMessageElement = this.$element.prev('p[role="status"]');
};

// -------------------------------------------------------------
// EVENTS
// -------------------------------------------------------------
Module.prototype.bindEventListeners = function() {
	this.$element.on(
		'click' + '.' + this.uuid,
		this.options.domSelectors.geoButton,
		this.getGeolocation.bind(this)
	);

	$document.on(events.setCoordinatesInHiddenFields + '.' + this.uuid, function(evt, coordinates) {
		this.setCoordinatesInHiddenFields(coordinates);
	}.bind(this));

	$document.on(events.geolocationInitState + '.' + this.uuid, function() {
		this.showErrorMessage(false);
	}.bind(this));
};

// -------------------------------------------------------------
// GEOLOCATION
// -------------------------------------------------------------
Module.prototype.getGeolocation = function() {
	var coordinates = {};

	this.$element.addClass(this.options.stateClasses.isAnimating);
	this.showErrorMessage(false);

	Geolocation.prototype.getLocation(this.options.geoLocationRequestOptions)
		.then(function(location) {
			coordinates = {
				lat: location.coords.latitude,
				long: location.coords.longitude
			};

			this.$element.removeClass(this.options.stateClasses.isAnimating);
			this.$geoLocationTarget.val(this.options.i18n.geolocationSuccessVisualLabel);

			if (this.$geoLocationSuccessMessageElement.length > 0) {
				this.$geoLocationSuccessMessageElement.text(this.options.i18n.geolocationSuccessText);
			}

			this.setCoordinatesInHiddenFields(coordinates);
			this.$element.trigger(events.geolocationHasSucceeded);

		}.bind(this))
		.catch(function() {
			this.$geoLocationTarget.val('');
			this.showErrorMessage(true);
			this.$element.removeClass(this.options.stateClasses.isAnimating);
			this.$element.trigger(events.geolocationHasFailed);

		}.bind(this));
};

Module.prototype.showErrorMessage = function(showMessage) {
	var geolocationParsleyInstance = this.$geoLocationTarget.parsley();

	if (showMessage) {
		this.$element.addClass(this.options.stateClasses.hasError);
		geolocationParsleyInstance.addError('error', { message: this.options.i18n.geolocationErrorText, updateClass: true });
	} else {
		geolocationParsleyInstance.removeError('error');
		this.$element.removeClass(this.options.stateClasses.hasError);
	}
};

Module.prototype.setCoordinatesInHiddenFields = function(coordinates) {
	this.$hiddenLongitude.val(coordinates.long);
	this.$hiddenLatitude.val(coordinates.lat);
};

// ---------------------------------------------------------------------------
// MODULE GENERIC
// ---------------------------------------------------------------------------
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 352 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 353 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * Geolocation helper
 *
 * Require geolocation helper, create new instance and call getLocation().
 *
 * Sample code:
 *
 * var Geolocation = require('geolocation.js'),
 * gl = new Geolocation();
 *
 * gl.getLocation().then(successCallbackFunction, errorCallbackFunction);
 *
 * function successCallbackFunction(location) {
 * }
 * function errorCallbackFunction(error) {
 * }
 */



var $ = __webpack_require__(0);

var geoOptions = {
		timeout: 10 * 1000,
		enableHighAccuracy: false,
		maximumAge: 30000 // 30 seconds
	},
	callbackPromise;

function Geolocation() {

}

/**
 * Retrieves user location based on browser location services.
 * @returns {Promise}
 */
Geolocation.prototype.getLocation = function() {

	callbackPromise = $.Deferred();
	if (this.isAvailable()) {
		navigator.geolocation.getCurrentPosition(this._geoSuccess, this._geoError, geoOptions);
	} else {
		setTimeout(function() {
			callbackPromise.reject('Geolocation not available');
		}, 0);
	}

	return callbackPromise;
};

/**
 * Checks if geolocation is available
 * @returns {boolean}
 */
Geolocation.prototype.isAvailable = function() {
	var ret = false;

	if (navigator.geolocation) {
		ret = true;
	}

	return ret;
};

/**
 * Internal method triggered on geolocation successful response. Location object is passed down to promise callback
 * @param location
 * @private
 */
Geolocation.prototype._geoSuccess = function(location) {
	callbackPromise.resolve(location);
};

/**
 * Internal method triggered when geolocation request fails. Promise is rejected with error
 * @param error
 * @private
 */
Geolocation.prototype._geoError = function(error) {
	callbackPromise.reject(error);
};

module.exports = Geolocation;


/***/ }),
/* 354 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(355);


/***/ }),
/* 355 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 356 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(357);


/***/ }),
/* 357 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M95 - Gif Module
 */



// Ensure that the SASS file is built.
__webpack_require__(358);

var $ = __webpack_require__(0),
	getHash = __webpack_require__(65),
	scrollToHash = __webpack_require__(66),
	SuperClass = __webpack_require__(1),
	scrollAnimation = __webpack_require__(52);

var name = 'gif',
	events = {},
	defaults = {
		domSelectors: {
			startStopButton: '[data-' + name + '="startstop_button"]',
			title: '[data-' + name + '="title"]',
			videoWrapper: '[data-' + name + '="video_wrapper"]',
			video: '[data-' + name + '="video"]',
			text: '[data-' + name + '="text"]'
		},
		stateClasses: {
			isActive: 'is_active',
			isPaused: 'is_paused',
			hasLoaded: 'has_loaded'
		},
		loop: false
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	this.hashValue = getHash();
	this.hasPreloaded = false;
	this.isManuallyPaused = false;
	this.hasPlayedThrough = false;

	this._cacheElements();
	this._lookUpLoopAttributeValue();

	this._initPreload();
	this._initIntersectionObserver();
	this._initClickHandler();

	this.initializeAnimations();

	if (this.hashValue) {
		this._scrollToGif();
	}

};

Module.prototype._cacheElements = function() {
	this.$startStopButton = this.$element.find(this.options.domSelectors.startStopButton);
	this.video = this.$element.find(this.options.domSelectors.video)[0];
};

Module.prototype._lookUpLoopAttributeValue = function() {
	this.options.loop = this.video.getAttribute('loop');
};

Module.prototype._initPreload = function() {

	if (this.video.readyState >= 2) {
		this._showGif();
		return;
	}

	this.video.addEventListener('loadeddata', function() {
		if (this.video.readyState >= 2) {
			this._showGif();
		}
	}.bind(this));

};

Module.prototype._showGif = function() {
	this.$element.addClass(this.options.stateClasses.hasLoaded);
	this.video.removeAttribute('preload');
};

Module.prototype._handlePreloadIntersectionObserve = function(changes) {

	changes.forEach(function(change) {
		if (change.isIntersecting) {
			this._preloadVideo();
			this.preloadIntersectionObserver.unobserve(this.video);
		}
	}.bind(this));

};

Module.prototype._initIntersectionObserver = function() {
	var intersectionObserver = new window.IntersectionObserver(this._handleIntersectionObserve.bind(this), {'root': null, 'rootMargin': '-64px 0px 0px 0px'});

	intersectionObserver.observe(this.video);
};

Module.prototype._handleIntersectionObserve = function(changes) {

	changes.forEach(function(change) {
		if (change.isIntersecting && !this.isManuallyPaused) {
			if (!this.inViewport) {
				this.inViewport = true;
				this._playVideo();
			}
		} else if (!this.isManuallyPaused) {
			if (this.inViewport) {
				this.inViewport = false;
				this._pauseVideo();
			}
		}
	}.bind(this));

};

Module.prototype._initClickHandler = function() {

	this.$element.on('click.' + this.uuid, this.options.domSelectors.startStopButton, function() {

		if (this.isManuallyPaused || this.hasPlayedThrough) {
			this.hasPlayedThrough = false;
			this.isManuallyPaused = false;
			this.$startStopButton.removeClass(this.options.stateClasses.isPaused);
			this.video.play();
		} else {
			this.isManuallyPaused = true;
			this.$startStopButton.addClass(this.options.stateClasses.isPaused);
			this.video.pause();
		}

	}.bind(this));

};

Module.prototype._playVideo = function() {

	this.video.play();

	if (!this.options.loop) {
		this.video.addEventListener('ended', function() {
			this.hasPlayedThrough = true;
			this.$startStopButton.addClass(this.options.stateClasses.isPaused);
		}.bind(this));
	}
};

Module.prototype._pauseVideo = function() {

	// Avoid interrupting the play promise to prevent this from happening
	// https://developers.google.com/web/updates/2017/06/play-request-was-interrupted
	if (this.video.readyState >= 2) {
		this.video.pause();
	}

};

Module.prototype.initializeAnimations = function() {

	var hasAnimations = this.$element.hasClass('var_scroll_animation');

	if (!hasAnimations) {
		return;
	}

	scrollAnimation({
		$trigger: this.$element,
		selectors: [
			this.options.domSelectors.title,
			this.options.domSelectors.videoWrapper,
			this.options.domSelectors.text
		],
		$parent: this.$element
	});

};

Module.prototype._scrollToGif = function() {

	var $hashTarget = this.hashValue === this.$element.attr('data-gif-id');

	if ($hashTarget) {
		scrollToHash('data-gif-id', this.hashValue, 0, true, false);
	}

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 358 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 359 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(360);


/***/ }),
/* 360 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M100 - Environment Calculator
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2019.
 */



__webpack_require__(361);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	scrollAnimation = __webpack_require__(52),
	templates = {
		co2Badge: __webpack_require__(362),
		resultIconContainer: __webpack_require__(363),
		resultIcon: __webpack_require__(364),
		resultText: __webpack_require__(365),
		co2: 'CO<sub>2</sub>'
	},
	modelKeys = __webpack_require__(366),
	$document = $(document),
	$window = $(window),
	isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints;

var name = 'envcalculator',
	events = {},
	defaults = {
		domSelectors: {
			title: '[data-' + name + '="title"]',
			lead: '[data-' + name + '="lead"]',
			inputForm: '[data-' + name + '="input_form"]',
			from: '[data-' + name + '="from"]',
			to: '[data-' + name + '="to"]',
			subtitle: '[data-' + name + '="subtitle"]',
			accessories: '[data-' + name + '="accessories"]',
			buy: '[data-' + name + '="buy"]',
			resultsContainer: '[data-' + name + '="results"]',
			resultsText: '[data-' + name + '="result_text"]',
			badgeText: '[data-' + name + '="badge_text"]',
			badgeHiddenText: '[data-' + name + '="badge_hidden_text"]',
			icon: '[data-' + name + '="icon"]',
			icons: '[data-' + name + '="icons"]',
			badge: '[data-' + name + '="co2_badge"]',
			errorIcon: '[data-' + name + '="icon_error"]'
		},
		stateClasses: {
			visible: 'is_visible',
			hidden: 'is_hidden',
			slideIn: 'slide_in',
			slideOut: 'slide_out',
			error: 'var_error'
		},
		constants: {
			ecoApiUrl: 'https://wcms-jsproxy.app.sbb.ch/b2c/v2/trips/ecoBalance/',
			tripsApiUrl: 'https://wcms-jsproxy.app.sbb.ch/b2c/v2/trips/',
			webshopLink: 'https://www.sbb.ch/$3/pages/fahrplan/fahrplan.xhtml?suche=true&vias=%5B%22%22%5D&von=$1&nach=$2&viaField1=&an=false',
			fetchAbortIdentifierEco: 'fetchAbortIdentifierEco',
			fetchAbortIdentifierTrip: 'fetchAbortIdentifierTrip',
			responseKeyDidok: 'didok',
			responseKeyLabel: 'label',
			responseKeyTransportTypeTrain: 'PUBLIC_TRANSPORT',
			responseKeyTransportTypeFlight: 'FLIGHT',
			responseKeyCo2: 'co2',
			responseKeyReconstructionContext: 'reconstructionContext',
			resultType: {
				emission: 'emission',
				difference: 'difference'
			},
			errorIcon: 'SBB_141_exclamation_mark'
		}
	};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ---------------------------------------------------------------------------
// INIT
// ---------------------------------------------------------------------------
Module.prototype.init = function() {
	this.localStorageLang = localStorage.getItem('sbblang');
	this.$resultsContainer = this.$element.find(this.options.domSelectors.resultsContainer);
	this.configAriaLiveContainer(false);
	this.allIcons = this.getAllIconNames();
	this.addMarkup();
	this.cacheElements();
	this.setupInitialState();
	this.addEventListeners();
	this.initAnimations();

	this.selection = {
		origin: this.options.initialStations.origin,
		destination: this.options.initialStations.destination
	};

	this.setInitialValues();

	this.xhrRequestTrip = undefined;
	this.xhrRequestEco = undefined;
	this.initialSubtitle = this.$subtitle.text();
};

Module.prototype.addMarkup = function() {
	this.$resultsContainer.append(templates.co2Badge);
	this.$resultsContainer.append(templates.resultIconContainer);
	this.$resultsContainer.append(templates.resultText);
};

Module.prototype.setInitialValues = function() {
	var origin = this.selection.origin,
		destination = this.selection.destination;

	this.$from.val(origin.label);
	this.$to.val(destination.label);

	this.handleUserSelection();
};

Module.prototype.setupInitialState = function() {
	this.stopAnimationOnNextRun = false;
	this.tempCo2BadgeText = '';
	this.tempCo2BadgeHiddenText = '';
	this.tempResultText = '';
	this.tempResultType = '';
	this.initialRequestDone = false;

	this.renderCo2BadgePlaceholder();
	this.renderResultTextPlaceholder();
	this.animateResultTextHeight();

};

// ---------------------------------------------------------------------------
// PAGE LEVEL ANIMATIONS
// ---------------------------------------------------------------------------
Module.prototype.initAnimations = function() {

	var hasAnimations = this.$element.hasClass('var_scroll_animation');

	if (!hasAnimations) {
		return;
	}

	scrollAnimation({
		$trigger: this.options.domSelectors.title,
		selectors: [
			this.options.domSelectors.title,
			this.options.domSelectors.lead,
			this.options.domSelectors.inputForm,
			this.options.domSelectors.subtitle,
			this.options.domSelectors.accessories,
			this.options.domSelectors.resultsContainer
		],
		$parent: this.$element
	});

};

// ---------------------------------------------------------------------------
// CACHE
// ---------------------------------------------------------------------------
Module.prototype.cacheElements = function() {
	this.$from = this.$element.find(this.options.domSelectors.from);
	this.$to = this.$element.find(this.options.domSelectors.to);
	this.$subtitle = this.$element.find(this.options.domSelectors.subtitle);
	this.$buy = this.$element.find(this.options.domSelectors.buy);
	this.$resultText = this.$element.find(this.options.domSelectors.resultsText);
	this.$badgeText = this.$element.find(this.options.domSelectors.badgeText);
	this.$badgeHiddenText = this.$element.find(this.options.domSelectors.badgeHiddenText);
	this.$iconContainer = this.$element.find(this.options.domSelectors.icons);
	this.$badge = this.$element.find(this.options.domSelectors.badge);
	this.$errorIcon = this.$element.find(this.options.domSelectors.errorIcon);
	this.$resultsContainer = this.$element.find(this.options.domSelectors.resultsContainer);
};

// ---------------------------------------------------------------------------
// COPY
// ---------------------------------------------------------------------------
Module.prototype.getHiddenBuyText = function(from, to) {
	var templateText = this.options.i18n.hiddenBuyText;

	templateText = templateText.replace('$1', from);
	templateText = templateText.replace('$2', to);

	return templateText;
};

Module.prototype.getBuyLink = function(from, to) {
	var fromEncoded = encodeURIComponent(from),
		toEncoded = encodeURIComponent(to),
		url,
		buyComponent = 'de/kaufen';

	if (this.localStorageLang.length > 0) {
		if (this.localStorageLang === 'fr') {
			buyComponent = 'fr/acheter';
		} else if (this.localStorageLang === 'it') {
			buyComponent = 'it/acquistare';
		} else if (this.localStorageLang === 'en') {
			buyComponent = 'en/buying';
		}
	}

	url = this.options.constants.webshopLink.replace('$1', fromEncoded);
	url = url.replace('$2', toEncoded);
	url = url.replace('$3', buyComponent);

	return url;
};

// Get a string like: "219kg CO2 gespart"
Module.prototype.getBadgeLabel = function(resultType, savings) {
	var badgeStringTemplate,
		replacedBadgeString,
		co2String = templates.co2;

	if (resultType === this.options.constants.resultType.difference) {
		badgeStringTemplate = this.options.i18n.badgeTitleDefault;
	} else {
		badgeStringTemplate = this.options.i18n.badgeTitleFallback;
	}

	replacedBadgeString = badgeStringTemplate.replace('$1', savings);
	replacedBadgeString = replacedBadgeString.replace('$2', co2String);

	return replacedBadgeString;
};

// Get a string like: "Ihre Ersparnis für eine Reise von Bern nach Berlin:"
Module.prototype.getBadgeHiddenText = function(resultType) {
	var hiddenBadgeStringTemplate,
		replacedHiddenBadgeString;

	if (resultType === this.options.constants.resultType.difference) {
		hiddenBadgeStringTemplate = this.options.i18n.hiddenResultTextFlight;
	} else {
		hiddenBadgeStringTemplate = this.options.i18n.hiddenResultTextNoFlight;
	}

	replacedHiddenBadgeString = hiddenBadgeStringTemplate.replace('$1', this.selection.origin.label);
	replacedHiddenBadgeString = replacedHiddenBadgeString.replace('$2', this.selection.destination.label);

	return replacedHiddenBadgeString;
};

// Get strings like:
// For difference flight/train: "Genug, um 450'000 Luftballons zu füllen"
// For emission for train only: "120 mal weniger als bei einem Flug von A nach B"
Module.prototype.getCalculatedText = function(resultType, modelItem, co2value) {
	var factor,
		text = modelItem[modelKeys.text],
		resultValue,
		resultText;

	if (resultType === this.options.constants.resultType.emission) {
		factor = modelItem[modelKeys.flightPendant];
		resultValue = this.contentfullRoundedNumber(factor / co2value);
	} else {
		factor = modelItem[modelKeys.kgPerItem];
		resultValue = this.contentfullRoundedNumber(co2value / factor);
	}

	resultText = text.replace('$1', resultValue);

	return resultText;
};

// ---------------------------------------------------------------------------
// NUMBERS
// ---------------------------------------------------------------------------

// - on numbers smaller than 10 leave 2 decimals.
// - on numbers greater than 10 return integer
// - on numbers greater than 1000, add tousands-separator
Module.prototype.contentfullRoundedNumber = function(number) {
	var roundedNumber,
		digits = 0,
		grouping = true;

	if (number < 10) {
		digits = 1;
		grouping = false;
	}

	if (number < 10000) {
		grouping = false;
	}

	roundedNumber = number.toLocaleString('de-CH', {
		minimumFractionDigits: digits,
		maximumFractionDigits: digits,
		useGrouping: grouping
	});

	if (this.localStorageLang.length > 0) {
		if (this.localStorageLang === 'en') {
			return roundedNumber.replace(/’/g, ',');
		}
	}

	return roundedNumber.replace(/’/g, ' ');
};

// ---------------------------------------------------------------------------
// EVENT LISTENERS
// ---------------------------------------------------------------------------
Module.prototype.addEventListeners = function() {
	this.$element.on(
		'selectedData.' + this.uuid,
		this.options.domSelectors.from,
		this.handleSelectionFrom.bind(this)
	);

	this.$element.on(
		'selectedData.' + this.uuid,
		this.options.domSelectors.to,
		this.handleSelectionTo.bind(this)
	);

	this.$element.on(
		'focus.' + this.uuid,
		this.options.domSelectors.to,
		this.handleFocusElement.bind(this, this.$to)
	);

	this.$element.on(
		'focus.' + this.uuid,
		this.options.domSelectors.from,
		this.handleFocusElement.bind(this, this.$from)
	);

	this.$element.on(
		'transitionend.' + this.uuid,
		this.options.domSelectors.icons,
		this.handleIconContainerTransitionEnd.bind(this)
	);
};

// ---------------------------------------------------------------------------
// EVENT HANDLERS
// ---------------------------------------------------------------------------
Module.prototype.handleSelectionFrom = function(_, data) {
	if (data[this.options.constants.responseKeyDidok] && data[this.options.constants.responseKeyLabel]) {
		this.selection.origin = data;
	}

	this.$from.blur();

	this.handleUserSelection();
};

Module.prototype.handleSelectionTo = function(_, data) {
	if (data[this.options.constants.responseKeyDidok] && data[this.options.constants.responseKeyLabel]) {
		this.selection.destination = data;
	}

	this.$to.blur();

	this.handleUserSelection();
};

Module.prototype.hasOriginAndDestination = function() {
	var origin = Object.keys(this.selection.origin).length > 0,
		destination = Object.keys(this.selection.destination).length > 0;

	return origin && destination;
};

Module.prototype.handleUserSelection = function() {

	if (!this.hasOriginAndDestination()) {
		return;
	}

	this.setButtonAttributes(this.selection.origin, this.selection.destination);
	this.fetchData(this.selection.origin, this.selection.destination);
	this.scrollResultIntoView();
};

Module.prototype.handleFocusElement = function($elem) {
	if (isTouchDevice) {
		return;
	}

	$elem.select();
};

// ---------------------------------------------------------------------------
// BUY BUTTON
// ---------------------------------------------------------------------------
Module.prototype.setButtonAttributes = function(origin, destination) {
	var hiddenText = this.getHiddenBuyText(origin.label, destination.label),
		buyLink = this.getBuyLink(origin.label, destination.label);

	this.$buy.attr('aria-label', hiddenText);
	this.$buy.attr('href', buyLink);
};

// ---------------------------------------------------------------------------
// NETWORK
// ---------------------------------------------------------------------------
Module.prototype.fetchData = function(origin, destination) {
	this.$iconContainer.removeClass(this.options.stateClasses.hidden);
	this.$badge.removeClass(this.options.stateClasses.hidden);
	this.$errorIcon.removeClass(this.options.stateClasses.visible);

	this.initLoadingAnimation();

	this.fetchTripData(origin, destination);
};

Module.prototype.fetchTripData = function(origin, destination) {
	var requestData,
		requestOptions;

	if (this.xhrRequestTrip !== undefined) {
		this.xhrRequestTrip.abort(this.options.constants.fetchAbortIdentifierTrip);
	}

	requestData = {
		originValue: origin[this.options.constants.responseKeyDidok],
		destinationValue: destination[this.options.constants.responseKeyDidok]
	};

	requestOptions = {
		url: this.options.constants.tripsApiUrl,
		cache: false,
		data: requestData,
		method: 'GET',
		headers: {
			'accept-language': 'de'
		},
		success: this.fetchTripDataSuccessCallback.bind(this),
		error: this.fetchTripDataErrorCallback.bind(this)
	};

	this.xhrRequestTrip = $.ajax(requestOptions);
};

Module.prototype.fetchTripDataSuccessCallback = function(data) {
	this.xhrRequestTrip = undefined;

	if (!data) {
		this.showNetworkError();
		return;
	}

	if (data.length < 1) {
		this.showNetworkError();
		return;
	}

	if (Object.keys(data[0]).indexOf(this.options.constants.responseKeyReconstructionContext) === -1) {
		this.showNetworkError();
		return;
	}

	this.fetchEcoData(data[0][this.options.constants.responseKeyReconstructionContext]);
};

Module.prototype.fetchTripDataErrorCallback = function(err) {
	this.xhrRequestTrip = undefined;

	if (err.statusText === this.options.constants.fetchAbortIdentifierTrip) {
		return;
	}

	this.showNetworkError();
	this.initialRequestDone = true;
};

Module.prototype.fetchEcoData = function(reconstructionContext) {
	var requestOptions;

	if (this.xhrRequestEco !== undefined) {
		this.xhrRequestEco.abort(this.options.constants.fetchAbortIdentifierEco);
	}

	requestOptions = {
		url: this.options.constants.ecoApiUrl + reconstructionContext,
		cache: false,
		method: 'GET',
		headers: {
			'accept-language': 'de'
		},
		success: this.fetchEcoDataSuccessCallback.bind(this),
		error: this.fetchEcoDataErrorCallback.bind(this)
	};

	this.xhrRequestEco = $.ajax(requestOptions);
};

Module.prototype.fetchEcoDataSuccessCallback = function(data) {
	this.xhrRequestEco = undefined;
	this.proccessFetchedData(data);
	this.initialRequestDone = true;
};

Module.prototype.fetchEcoDataErrorCallback = function(err) {
	this.xhrRequestEco = undefined;

	if (err.statusText === this.options.constants.fetchAbortIdentifierEco) {
		return;
	}

	this.showNetworkError();
	this.initialRequestDone = true;
};

Module.prototype.showNetworkError = function() {
	this.finalIcon = this.options.constants.errorIcon;
	this.stopAnimationOnNextRun = true;
	this.hasError = true;

	this.$badge.addClass(this.options.stateClasses.hidden);

	this.tempResultText = this.options.i18n.networkError;
};

// ---------------------------------------------------------------------------
// DATA PROCESSING
// ---------------------------------------------------------------------------
Module.prototype.proccessFetchedData = function(data) {
	var flightData,
		trainData,
		co2difference,
		flightEmission,
		trainEmission;

	if (!data) {
		this.showNetworkError();
		return;
	}

	flightData = data[this.options.constants.responseKeyTransportTypeFlight];
	trainData = data[this.options.constants.responseKeyTransportTypeTrain];

	if (!trainData) {
		this.showNetworkError();
		return;
	}

	trainEmission = trainData[this.options.constants.responseKeyCo2];

	if (!flightData) {
		if (trainEmission <= 0) {
			this.showNetworkError();
		} else {
			this.processData(this.options.constants.resultType.emission, trainEmission);
		}

		return;
	}

	flightEmission = flightData[this.options.constants.responseKeyCo2];
	co2difference = flightEmission - trainEmission;

	if (co2difference <= 0) {
		this.showNetworkError();
		return;
	}

	this.processData(this.options.constants.resultType.difference, co2difference);
};

Module.prototype.processData = function(resultType, savingsOrEmission) {
	var modelItem,
		calculatedText;

	if (resultType === this.options.constants.resultType.emission) {
		modelItem = this.options.model[modelKeys.default];
		calculatedText = this.getCalculatedText(this.options.constants.resultType.emission, modelItem, savingsOrEmission);
	} else {
		modelItem = this.getRandomModelObject();
		calculatedText = this.getCalculatedText(this.options.constants.resultType.difference, modelItem, savingsOrEmission);
	}

	this.finalIcon = modelItem[modelKeys.icon];
	this.tempResultType = resultType;
	this.stopAnimationOnNextRun = true;

	this.renderCo2Badge(resultType, savingsOrEmission);
	this.tempResultText = calculatedText;
};

// ---------------------------------------------------------------------------
// MODEL
// ---------------------------------------------------------------------------
Module.prototype.getRandomModelObject = function() {
	var model = this.options.model[modelKeys.data],
		randomMax = model.length,
		randomNumber = Math.floor(Math.random() * randomMax);

	return model[randomNumber];
};

Module.prototype.getAllIconNames = function() {
	var model = this.options.model[modelKeys.data],
		icons = [];

	model.forEach(function(modelItem) {
		icons.push(modelItem[modelKeys.icon]);
	});

	// add default icon
	icons.push(this.options.model[modelKeys.default][modelKeys.icon]);

	return icons;
};

Module.prototype.getRandomIconName = function() {
	var randomMax,
		randomNumber,
		iconsCopy = JSON.parse(JSON.stringify(this.allIcons));

	if (this.lastRandomIconName) {
		iconsCopy = iconsCopy.filter(function(icon) {
			return icon !== this.lastRandomIconName;
		}.bind(this));
	}

	randomMax = iconsCopy.length;
	randomNumber = Math.floor(Math.random() * randomMax);

	this.lastRandomIconName = iconsCopy[randomNumber];

	return this.lastRandomIconName;
};

// ---------------------------------------------------------------------------
// RENDER
// ---------------------------------------------------------------------------

Module.prototype.configAriaLiveContainer = function(visible) {
	if (visible) {
		this.$resultsContainer.attr({
			'role': 'status',
			'aria-live': 'polite',
			'aria-atomic': 'false'
		});
	} else {
		this.$resultsContainer.removeAttr('role aria-live aria-atomic');
	}
};

Module.prototype.renderSubtitle = function(resultType) {
	if (resultType === this.options.constants.resultType.difference) {
		this.$subtitle.text(this.initialSubtitle);
	} else {
		this.$subtitle.text(this.options.i18n.alternativeResultTitle);
	}
};

Module.prototype.renderCo2Badge = function(resultType, savings) {
	var roundedSavings = this.contentfullRoundedNumber(savings),
		label = this.getBadgeLabel(resultType, roundedSavings),
		hiddenText = this.getBadgeHiddenText(resultType);

	this.tempCo2BadgeText = label;
	this.tempCo2BadgeHiddenText = hiddenText;
};

Module.prototype.renderCo2BadgePlaceholder = function() {
	var placeholder = this.options.i18n.calculating.replace('$1', templates.co2);

	this.$badgeText.html(placeholder);
	this.$badgeHiddenText.text('');
};

Module.prototype.appendRandomIcon = function() {
	var randomIconName = this.getRandomIconName();

	this.$iconContainer.append(templates.resultIcon({icon: randomIconName}));
};

Module.prototype.appendFinalIcon = function() {
	var templateData = {};

	if (!this.finalIcon) {
		return;
	}

	templateData.icon = this.finalIcon;

	if (this.hasError) {
		templateData.additionalClasses = this.options.stateClasses.error;
	}

	this.$iconContainer.append(templates.resultIcon(templateData));

	this.finalIcon = undefined;
};

Module.prototype.renderResultTextPlaceholder = function() {
	this.$resultText.text('placeholder');
};

Module.prototype.scrollResultIntoView = function() {
	if (!this.initialRequestDone) {
		return;
	}

	var isVisible,
		containerTop = this.$resultsContainer.offset().top,
		containerHeight = this.$resultsContainer.outerHeight(),
		scrollPosition = $document.scrollTop(),
		windowHeight = $window.height(),
		margin = 20;

	isVisible = containerTop + containerHeight + margin < scrollPosition + windowHeight;

	if (isVisible) {
		return;
	}

	$('html, body').animate({
		scrollTop: containerTop - windowHeight + containerHeight + margin
	});

};

// ---------------------------------------------------------------------------
// LOADING ANIMATION
// ---------------------------------------------------------------------------

Module.prototype.handleIconContainerTransitionEnd = function(evt) {

	var srcElementClasses,
		isSlideInElement;

	if (evt.originalEvent.propertyName !== 'transform') {
		return;
	}

	srcElementClasses = evt.originalEvent.srcElement.className.baseVal.split(' ');
	isSlideInElement = srcElementClasses.indexOf(this.options.stateClasses.slideOut) === -1;

	if (!isSlideInElement) {
		return;
	}

	if (this.isFinalAnimation) {
		this.configAriaLiveContainer(true);

		this.$badgeText.html(this.tempCo2BadgeText);
		this.$badgeHiddenText.text(this.tempCo2BadgeHiddenText);
		requestAnimationFrame(function() {
			this.$resultText.html(this.tempResultText);
			this.animateResultTextHeight();
		}.bind(this));

		this.renderSubtitle(this.tempResultType);
		this.showHideBadge(true);
		this.showHideResultText(true);

		return;
	}

	if (this.stopAnimationOnNextRun) {
		this.showHideBadge(false);
		this.loadingAnimationStep(true);
	} else {
		this.loadingAnimationStep(false);
	}

	this.stopAnimationOnNextRun = false;
	this.hasError = false;
};

Module.prototype.initLoadingAnimation = function() {
	this.isFinalAnimation = false;
	this.showHideBadge(true);
	this.showHideResultText(false);
	this.renderCo2BadgePlaceholder();
	this.loadingAnimationStep(false);
};

Module.prototype.showHideBadge = function(show) {
	if (show) {
		this.$badgeText.addClass(this.options.stateClasses.visible);
	} else {
		this.$badgeText.removeClass(this.options.stateClasses.visible);
	}
};

Module.prototype.showHideResultText = function(show) {
	if (show) {
		this.$resultText.addClass(this.options.stateClasses.visible);
	} else {
		this.$resultText.removeClass(this.options.stateClasses.visible);
	}
};

Module.prototype.loadingAnimationStep = function(isFinalAnimation) {
	var $nextIcon;

	if (this.$iconToRemove) {
		this.$iconToRemove.remove();
	}

	this.$iconToRemove = $(this.$iconContainer.find(this.options.domSelectors.icon));

	if (isFinalAnimation) {
		this.isFinalAnimation = true;
		this.lastRandomIconName = undefined;
		this.appendFinalIcon();
	} else {
		this.appendRandomIcon();
	}

	$nextIcon = $(this.$iconContainer.find('.' + this.options.stateClasses.slideIn));

	requestAnimationFrame(function() {
		setTimeout(function() {
			$nextIcon.removeClass(this.options.stateClasses.slideIn);
			this.$iconToRemove.addClass(this.options.stateClasses.slideOut);
		}.bind(this), 0);
	}.bind(this));
};

Module.prototype.animateResultTextHeight = function() {
	var oldHeight,
		newHeight,
		transitionDuration;

	oldHeight = this.$resultText.height();
	transitionDuration = this.$resultText.css('transition-duration');

	this.$resultText.css('transition-duration', '0');
	this.$resultText.height('auto');

	newHeight = this.$resultText.height();
	this.$resultText.height(oldHeight);
	this.$resultText.css('transition-duration', transitionDuration);

	requestAnimationFrame(function() {
		requestAnimationFrame(function() {
			this.$resultText.height(newHeight);
		}.bind(this));
	}.bind(this));

};

// ---------------------------------------------------------------------------
// MODULE GENERIC
// ---------------------------------------------------------------------------
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 361 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 362 */
/***/ (function(module, exports, __webpack_require__) {

var Handlebars = __webpack_require__(8);
function __default(obj) { return obj && (obj.__esModule ? obj["default"] : obj); }
module.exports = (Handlebars["default"] || Handlebars).template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
    return "<p\n	class=\"mod_env_calculator_co2_badge\"\n	data-envcalculator=\"co2_badge\"\n>\n	<span\n		class=\"visuallyhidden\"\n		data-envcalculator=\"badge_hidden_text\"\n	></span>\n	<span\n		data-envcalculator=\"badge_text\"\n		class=\"mod_env_calculator_co2_badge_label\"\n	></span><span class=\"visuallyhidden\">.</span>\n</p>\n";
},"useData":true});

/***/ }),
/* 363 */
/***/ (function(module, exports, __webpack_require__) {

var Handlebars = __webpack_require__(8);
function __default(obj) { return obj && (obj.__esModule ? obj["default"] : obj); }
module.exports = (Handlebars["default"] || Handlebars).template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
    return "<div\n	class=\"mod_env_calculator_result_icons\"\n	data-envcalculator=\"icons\"\n></div>\n";
},"useData":true});

/***/ }),
/* 364 */
/***/ (function(module, exports, __webpack_require__) {

var Handlebars = __webpack_require__(8);
function __default(obj) { return obj && (obj.__esModule ? obj["default"] : obj); }
module.exports = (Handlebars["default"] || Handlebars).template({"1":function(depth0,helpers,partials,data) {
    var helper;

  return " "
    + this.escapeExpression(((helper = (helper = helpers.additionalClasses || (depth0 != null ? depth0.additionalClasses : depth0)) != null ? helper : helpers.helperMissing),(typeof helper === "function" ? helper.call(depth0,{"name":"additionalClasses","hash":{},"data":data}) : helper)));
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
    var stack1, helper;

  return "<svg\n	data-envcalculator=\"icon\"\n	class=\"mod_svgsprite_icon mod_env_calculator_result_icon slide_in"
    + ((stack1 = helpers['if'].call(depth0,(depth0 != null ? depth0.additionalClasses : depth0),{"name":"if","hash":{},"fn":this.program(1, data, 0),"inverse":this.noop,"data":data})) != null ? stack1 : "")
    + "\"\n	focusable=\"false\"\n>\n	<use\n		xmlns:xlink=\"http://www.w3.org/1999/xlink\"\n		xlink:href=\"#"
    + this.escapeExpression(((helper = (helper = helpers.icon || (depth0 != null ? depth0.icon : depth0)) != null ? helper : helpers.helperMissing),(typeof helper === "function" ? helper.call(depth0,{"name":"icon","hash":{},"data":data}) : helper)))
    + "\"\n	></use>\n</svg>\n";
},"useData":true});

/***/ }),
/* 365 */
/***/ (function(module, exports, __webpack_require__) {

var Handlebars = __webpack_require__(8);
function __default(obj) { return obj && (obj.__esModule ? obj["default"] : obj); }
module.exports = (Handlebars["default"] || Handlebars).template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
    return "<p\n	data-envcalculator=\"result_text\"\n	class=\"mod_env_calculator_result_text\"\n\n></p>\n";
},"useData":true});

/***/ }),
/* 366 */
/***/ (function(module, exports) {

module.exports = {
	default: 'default',
	data: 'data',
	keys: 'keys',
	icon: 'icon',
	kgPerItem: 'kgPerItem',
	flightPendant: 'flightPendant',
	text: 'text'
};


/***/ }),
/* 367 */,
/* 368 */,
/* 369 */,
/* 370 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(371);


/***/ }),
/* 371 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * CPM05 - Product Comparison List
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2018.
 */



__webpack_require__(372);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	scrollAnimation = __webpack_require__(52);

var name = 'productcomparison',
	events = {},
	defaults = {
		domSelectors: {
		},
		stateClasses: {
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	this.initializeAnimations();
};

Module.prototype.initializeAnimations = function() {

	var hasAnimations = this.$element.hasClass('var_scroll_animation'),
		$leadText = this.$element.find('[data-' + name + '="lead"]');

	if (!hasAnimations) {
		return;
	}

	scrollAnimation({
		selectors: '[data-' + name + '="title"]',
		$parent: this.$element
	});

	if ($leadText.length > 0) {
		scrollAnimation({
			selectors: '[data-' + name + '="lead"]',
			$parent: this.$element
		});
	}

	scrollAnimation({
		selectors: '[data-' + name + '="list_item"]',
		$parent: this.$element
	});

};

SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 372 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 373 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(374);


/***/ }),
/* 374 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M03 - Homepage Main Teaser
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



// Ensure that the SASS file is built.
__webpack_require__(375);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	HomePageMainTeaser,
	ConsentHelper = __webpack_require__(26);

var name = 'homepagemainteaser',
	events = {
		timetableInputFormFocus: 'timetable-input-form-focus',
		timetableInputFormBlur: 'timetable-input-form-blur'
	},
	defaults = {
		domSelectors: {
			sliderButtons: '[data-' + name + '="slider_button"]',
			startStopButton: '[data-' + name + '="startstop_button"]',
			teaserList: '[data-' + name + '="teaser_list"]',
			teaserItems: '[data-' + name + '="teaser_item"]',
			teaserLink: '[data-' + name + '="teaser_link"]',
			descriptionStart: '[data-' + name + '="description_start"]',
			descriptionStop: '[data-' + name + '="description_stop"]',
			teaserImage: '[data-' + name + '="image"]',
			teaserText: '[data-' + name + '="teaser_text"]',
			whitePanel: '[data-' + name + '="white_panel"]'
		},
		stateClasses: {
			isActive: 'is_active',
			isPaused: 'is_paused',
			isStatic: 'is_static',
			isFadedOut: 'is_faded_out',
			isFadedIn: 'is_faded_in',
			noTransition: 'has_no_transition',
			isHidden: 'is_hidden'
		},
		trackingIsAllowed: false,
		slideInterval: 7000, // ms to wait before the next teaser slides in
		waitBeforeTextFadeIn: 300, // timeout before the text of next
		// teaser element fades in
		imageLowerScale: 0.8, // scale factor for the image to scale before it
		// slides in from the right
		imageUpperScale: 1.0, // scale factor for the image to scale after it
		// slided in from the right
		imageRotation: 10, // amount of rotation (the teaser slightly rotates
		// the image along the Y axis before and after fading in)
		varLh: 'var_lh',
		swipeThresholdPercentage: 25
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	var items = this.$element.find(this.options.domSelectors.teaserItems).length,
		oneTrustIsIntegrated = ConsentHelper.consentIsIntegrated();

	if (items > 1) {
		__webpack_require__.e/* require.ensure */(11).then((function(require) {
				HomePageMainTeaser = __webpack_require__(316);

				Module.prototype = $.extend(true, Module.prototype, HomePageMainTeaser.prototype);

				HomePageMainTeaser.prototype.init.apply(this);

			}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);
	} else {
		this._initSingleItemCase();
	}

	if (oneTrustIsIntegrated) {
		// Tracking is assigned to Performance Cookies in
		// One Trust Cookie Consent. Let's check if the user
		// has allowed this category. If so, we fire an
		// impression for AT-Internet.
		ConsentHelper.activeGroups().then(function(groups) {
			if (groups.allowedPerformance) {
				this.trackingIsAllowed = true;
				this._handleTracking();
			}
		}.bind(this)).catch(function() {});

		// Consent might also be given at a later state, in
		// which case we have to check once more.
		ConsentHelper.consentChangeCallback(function(groups) {
			if (groups.allowedPerformance) {
				this.trackingIsAllowed = true;
				this._handleTracking();
			}
		}.bind(this));
	}

};

Module.prototype._initSingleItemCase = function() {
	var $whitePanel = this.$element.find(this.options.domSelectors.whitePanel),
		$teaserText = this.$element.find(this.options.domSelectors.teaserText);

	$teaserText.removeClass(this.options.stateClasses.isFadedOut);

	if ($whitePanel.length > 0) {
		$whitePanel.css({
			'transition-delay': '0ms'
		}).addClass(this.options.stateClasses.isFadedIn);
	}

	this._handleTracking();

};

/**
 * Make sure that the single teaser is tracked
 */
Module.prototype._handleTracking = function() {
	var $teaserLink = this.$element.find(this.options.domSelectors.teaserLink);

	document.addEventListener('trackingIsReady', function() {
		window.tracking.observeTeaserElement($teaserLink[0]);
		document.removeEventListener('trackingIsReady', function() {});
	});
};

/**
 * Unbind events, remove data, custom teardown
 * @method
 * @public
 */
Module.prototype.destroy = function() {

	// clear interval
	clearTimeout(this.timeoutId);

	// Unbind events, remove data
	SuperClass.prototype.destroy.apply(this);

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 375 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 376 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(377);


/***/ }),
/* 377 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 378 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(379);


/***/ }),
/* 379 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M56 - Survey
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



// Ensure that the SASS file is built.
__webpack_require__(380);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	gsap = __webpack_require__(11).gsap;

var name = 'survey',
	events = {},
	defaults = {
		domSelectors: {
			container: '[data-' + name + '="container"]',
			loading: '[data-' + name + '="loading"]',
			application: '[data-' + name + '="application"]',
			answers: '[data-' + name + '="answers"]',
			save: '[data-' + name + '="save"]',
			form: '.mod_form form',
			results: '[data-' + name + '="results"]',
			pie: '[data-' + name + '="pie"]',
			slice: '[data-' + name + '="slice"]',
			legend: '[data-' + name + '="legend"]'
		},
		stateClasses: {
			varAnswers: 'var_answers',
			varResults: 'var_results',
			isLoading: 'is_loading'
		},
		cookieExpirationTimeInDays: 30,
		sliceTransitionDuration: 1,
		sliceTransitionDelay: 0.5
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	this.surveyId = this.$element.attr('id') || name;
	this.$application = this.$element.find(this.options.domSelectors.application);
	this.$answers = this.$element.find(this.options.domSelectors.answers);
	this.$results = this.$element.find(this.options.domSelectors.results);
	this.$loading = this.$element.find(this.options.domSelectors.loading);

	this._initLoading();

	if (!this._hasCookie()) {

		// show answers
		this.$element.addClass(this.options.stateClasses.varAnswers);
		this.$answers.attr('aria-hidden', false);
		this.$element.on('submit.' + this.uuid, this.options.domSelectors.form, function(event) {
			event.preventDefault();
			this._saveSelection();
		}.bind(this));
	} else {
		this._showResults();
	}

};

/**
 * Helper method to set the cookie.
 * @method
 * @private
 */
Module.prototype._saveSelection = function() {

	var $form = this.$element.find(this.options.domSelectors.form),
		data = $form.serialize() || {};

	this._toggleLoading(false);
	this.$element.removeClass(this.options.stateClasses.varAnswers);
	this.$answers.attr('aria-hidden', true);

	$.ajax({
		url: $form.attr('action'),
		method: $form.attr('method') || 'GET',
		data: data,
		success: function(response) {

			var $results = $(response).find(this.options.domSelectors.results).html();

			this.$results.html($results);

			// save cookie
			this._setCookie(this.$element.find(this.options.domSelectors.answers + ' input:checked').val());
			this._showResults();

		}.bind(this)
	});

};

Module.prototype._showResults = function() {
	var cookieValue = this._getCookieValue() || ' ';

	this._toggleLoading(true);
	this.$element.find(this.options.domSelectors.slice).filter('[data-survey-value="' + cookieValue + '"]').addClass('var_user');
	this.$element.find(this.options.domSelectors.legend).filter('[data-survey-value="' + cookieValue + '"]').addClass('var_user');
	this.$element.find(this.options.domSelectors.answers).attr('aria-hidden', true);
	this.$element.find(this.options.domSelectors.results).attr('aria-hidden', false);
	this.$element.removeClass(this.options.stateClasses.varAnswers).addClass(this.options.stateClasses.varResults);
	this._showPie();
};

Module.prototype._showPie = function() {

	var $slices = this.$element.find(this.options.domSelectors.slice),
		diameter = this.$element.find(this.options.domSelectors.pie).width() - 24,
		umfang = diameter * Math.PI,
		rotation = 0,
		delay = 0;

	$slices.each(function(index, element) {
		var $element = $(element),
			percentage = $element.attr('data-' + name + '-percentage') || (100 / $slices.length);

		$element.attr('stroke-dashoffset', umfang);
		$element.attr('stroke-dasharray', umfang);

		gsap.fromTo($element, {
			rotation: rotation,
			transformOrigin: '50% 50%',
			'stroke-dashoffset': umfang
		}, {
			ease: 'power1.out',
			duration: this.options.sliceTransitionDuration,
			delay: delay,
			'stroke-dashoffset': umfang / 100 * (100 - percentage)
		});

		delay += this.options.sliceTransitionDuration;
		rotation += 360 / 100 * percentage;

	}.bind(this));

};

/**
 * Helper method to set the cookie.
 * @method
 * @private
 */
Module.prototype._setCookie = function(value) {

	var expirationDate = new Date(),
		expires;

	// define expiration date
	expirationDate.setTime(expirationDate.getTime() + (this.options.cookieExpirationTimeInDays* 24 * 60 * 60 * 1000));

	// define expiration date as string
	expires = 'expires=' + expirationDate.toUTCString();

	// set cookie
	document.cookie = this.surveyId + '= ' + value + '; ' + expires;
};

/**
 * Helper method to get the cookie
 * @method
 * @private
 * @return {String} returns true if a cookie for this ghettobox was found. returns false if it was not found.
 */
Module.prototype._hasCookie = function() {

	var name,
		cookieName,
		cookieArray,
		iterator,
		cookie;

	name = this.surveyId + '=';

	cookieArray = document.cookie.split(';');

	for (iterator = 0; iterator < cookieArray.length; iterator++) {
		cookie = cookieArray[iterator];

		while (cookie.charAt(0) === ' ') {
			cookie = cookie.substring(1);
		}

		if (cookie.indexOf(name) === 0) {
			cookieName = cookie.substring(name.length, cookie.length);
			if (cookieName) {
				return true;
			}
		}
	}

	return false;
};

/**
 * Helper method to get the cookie
 * @method
 * @private
 * @return {String} returns true if a cookie for this ghettobox was found. returns false if it was not found.
 */
Module.prototype._getCookieValue = function() {

	var name,
		cookieName,
		cookieArray,
		iterator,
		cookie;

	name = this.surveyId + '=';

	cookieArray = document.cookie.split(';');

	for (iterator = 0; iterator < cookieArray.length; iterator++) {
		cookie = cookieArray[iterator];

		while (cookie.charAt(0) === ' ') {
			cookie = cookie.substring(1);
		}

		if (cookie.indexOf(name) === 0) {
			cookieName = cookie.substring(name.length, cookie.length);
			if (cookieName) {
				return cookieName;
			}
		}
	}

	return false;
};

Module.prototype._initLoading = function() {

	var source = $('[data-template="preloader"]').html();

	this._toggleLoading(true);
	this.$loading.append($(source));
};

Module.prototype._toggleLoading = function(state) {
	this.$application.attr('aria-busy', !state);
	this.$loading.toggleClass('is_hidden', state);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	initEvents: ['ready', 'ajaxload'],
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 380 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 381 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(382);


/***/ }),
/* 382 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M57 - Station Search
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



// Ensure that the SASS file is built.
__webpack_require__(383);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	keymaster = __webpack_require__(14),
	createUrlParams = __webpack_require__(79),
	parseUrl = __webpack_require__(69),
	ajaxRequest = __webpack_require__(16);

var name = 'stationsearch',
	$document = $(document),
	events = {

		// consuming events from autocomplete
		autocompleteItemSelected: 'inputChangeBySelectingAutocompleteItem.autocomplete',

		// triggered events. m00_-_journey_map listens to these.
		removeMarkers: 'removeMarkers.journeymap',
		addMarkers: 'addMarkers.journeymap'
	},
	defaults = {
		domSelectors: {
			form: '[data-search-form]',
			search: '[data-init="autocomplete"]',
			button: '[type="submit"]',
			overlay: '[data-' + name + '="overlay"]',
			tabs: '[data-' + name + '="tablist"]',
			tab: '[data-' + name + '="tabitem"]',
			panel: '[data-' + name + '="tabpanel"]',
			resultList: '[data-' + name + '="result_list"]',
			resultDescription: '[data-' + name + '="result_description"]'
		},
		stateClasses: {
			visible: 'is_visible',
			disabled: 'is_disabled',
			active: 'is_active',
			visuallyhidden: 'visuallyhidden',
			isInitialized: 'is_initialized'
		},
		urlApiAttribute: 'data-options-url',
		urlSearchAttributeName: 'search',
		expandableResultListItemClass: 'mod_station_search_results_list_item',
		expandableResultListItemLinkClass: 'mod_station_search_results_list_item_link'
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	this.results = [];

	// cache elements
	this.cacheElements();

	// add event listeners
	this.addEventListeners();

	if (this.options.showLastActiveTab) {
		var queryParams = parseUrl(window.location.search);

		if (queryParams.mapIsActive) {
			this.showTabWithIndex(1);
		}
	}
};

// Cache all necessary elements
Module.prototype.cacheElements = function() {

	this.$form = this.$element.find(this.options.domSelectors.form);
	this.$search = this.$element.find(this.options.domSelectors.search);
	this.$overlay = this.$element.find(this.options.domSelectors.overlay);
	this.$button = this.$element.find(this.options.domSelectors.button);
	this.$tablist = this.$element.find(this.options.domSelectors.tabs);
	this.$tabitems = this.$element.find(this.options.domSelectors.tab);
	this.$tabpanels = this.$element.find(this.options.domSelectors.panel);
	this.$resultList = this.$element.find(this.options.domSelectors.resultList);
	this.$resultDescription = this.$element.find(this.options.domSelectors.resultDescription);

	if (this.$search.length === 0) {
		this.$search = $('[data-init="servicepoints-input"]').find(this.options.domSelectors.search);
	}

	if (this.$button.length === 0) {
		this.$button = $('[data-init="servicepoints-input"]').find(this.options.domSelectors.button);
	}

	this.getAutocompleteOptions();

};

// Get autocomplete options
// We need to know minChars options from autocomplete so we can use it as
// minChars for enabling/disabling the search button
Module.prototype.getAutocompleteOptions = function() {

	var autocompleteOptions,
		autocompleteInstances;

	// if we should not be able to get the minChars option properly, make sure
	// the sarch button is always active
	this.minChars = 0;

	// case 1: we might find minChars option in the markup
	if (this.$search.attr('data-autocomplete-options')) {
		autocompleteOptions = JSON.parse(this.$search.attr('data-autocomplete-options'));
		if (autocompleteOptions) {
			if (autocompleteOptions.minChars) {
				this.minChars = autocompleteOptions.minChars;

				// initially enable/disable search button
				this.enableDisableSearchButton();

				return;
			}
		}
	}

	// case 2: we need to search for the autocomplete-instance on the
	// window.estatico object and read the option from there
	// setTimeout: we move the following to the call stack to make sure that we
	// ask for the instances AFTER all window.estatico.modules have been initialized
	setTimeout(function() {
		autocompleteInstances = window.estatico.modules.autocomplete.instances;
		$.each(autocompleteInstances, function(key, instance) {

			// let's check if we have found the autocomplete from stationsearch
			if (instance.$element.attr('data-stationsearch')) {
				if (instance.options.minChars) {
					this.minChars = instance.options.minChars;

					// initially enable/disable search button
					this.enableDisableSearchButton();

					return false;
				}
			}
		}.bind(this));
	}.bind(this), 0);

};

// ------------------------------------------------------------------------
// Event listeners
// ------------------------------------------------------------------------
Module.prototype.addEventListeners = function() {

	// listen to form submit / click on search button
	this.$element.on('submit.' + this.uuid, this.$form, this.handleSearch.bind(this));

	// listen to input changes
	$document.on('input.' + this.uuid + ' ' + events.autocompleteItemSelected, this.$search, this.enableDisableSearchButton.bind(this));

	// listen to clicks on tabs
	this.$tabitems.on('click', this.handleTabClick.bind(this));

	// listen to keyboard events
	this.attachKeyboardHandlers();

};

// ------------------------------------------------------------------------
// Event handlers
// ------------------------------------------------------------------------

// Handle search / user clicking on search button
Module.prototype.handleSearch = function(evt) {

	var apiUrl = this.$element.attr(this.options.urlApiAttribute),
		queryString = encodeURIComponent(this.$search.val()),
		urlReplaced = apiUrl.replace('{{query}}', queryString);

	if (this.$search.val().length < 3) {
		return;
	}

	// make sure we don't issue form action
	evt.preventDefault();

	// blur input field and search button
	this.$search.blur();
	this.$button.blur();

	// get content from api
	ajaxRequest(urlReplaced).then(function(results) {

		// process results
		this.processSearchResults(results);

	}.bind(this));
};

// Enable / disable search button
Module.prototype.enableDisableSearchButton = function() {
	if (this.$search.val().length < this.minChars) {
		this.$button.addClass(this.options.stateClasses.disabled);
		this.$button.attr('tabindex', '-1');
	} else {
		this.$button.removeClass(this.options.stateClasses.disabled);
		this.$button.removeAttr('tabindex');
	}

	// make sure that search button gets initialized class
	this.$button.addClass(this.options.stateClasses.isInitialized);
};

// Handle Tab clicks
Module.prototype.handleTabClick = function(evt) {
	var index = $(evt.currentTarget).index();

	if (this.options.showLastActiveTab) {
		var queryParams = parseUrl(window.location.search),
			url;

		if (index === 1) {
			queryParams.mapIsActive = true;
			url = createUrlParams(queryParams);
		} else {
			url = location.pathname;
		}

		history.pushState(null, null, url);
	}

	this.showTabWithIndex(index);
};

/**
 * Bind keyboard events to show previous/next teaser
 */
Module.prototype.attachKeyboardHandlers = function() {
	var scope = this.name + '-' + this.uuid;

	// set keymaster's scope on focusin
	this.$tablist.on('focusin.' + this.uuid, function() {
		keymaster.setScope(scope);
	}).on('focusout.' + this.uuid, function() {
		keymaster.setScope('all');
	});

	// handle jump to next teaser
	keymaster('right, down', scope, function() {
		this.showNextTab();
	}.bind(this));

	// handle jump to previous teaser
	keymaster('left, up', scope, function() {
		this.showPreviousTab();
	}.bind(this));
};

// ------------------------------------------------------------------------
// Tab handling
// ------------------------------------------------------------------------
Module.prototype.showTabWithIndex = function(index) {
	var $panel = this.$tabpanels.eq(index),
		$tab = this.$tabitems.eq(index);

	// only continue if the panel is not already active
	if ($panel.hasClass(this.options.stateClasses.active)) {
		return;
	}

	// add/remove active class and set aria attributes on tabs
	this.$tabitems
		.removeClass(this.options.stateClasses.active)
		.attr({'aria-selected': false, 'tabindex': '-1'});

	$tab
		.addClass(this.options.stateClasses.active)
		.attr({'aria-selected': true, 'tabindex': '0'});

	// add/remove active class and set aria attributes on panel
	this.$tabpanels
		.removeClass(this.options.stateClasses.active)
		.attr('aria-hidden', true);

	$panel
		.addClass(this.options.stateClasses.active)
		.attr('aria-hidden', false);
};

Module.prototype.showNextTab = function() {
	var totalTabs = this.$tabitems.length,
		nextIndex = this.getCurrentlyActiveTabIndex() + 1;

	if (nextIndex > totalTabs - 1) {
		nextIndex = 0;
	}

	this.showTabWithIndex(nextIndex);
};

Module.prototype.showPreviousTab = function() {
	var totalTabs = this.$tabitems.length,
		previousIndex = this.getCurrentlyActiveTabIndex() - 1;

	if (previousIndex < 0) {
		previousIndex = totalTabs - 1;
	}

	this.showTabWithIndex(previousIndex);
};

Module.prototype.getCurrentlyActiveTabIndex = function() {
	var activeItem = 0;

	$.each(this.$tabitems, function(index, item) {
		if ($(item).hasClass(this.options.stateClasses.active)) {
			activeItem = index;
			return false;
		}
	}.bind(this));

	return activeItem;
};

// ------------------------------------------------------------------------
// Process search results
// ------------------------------------------------------------------------
Module.prototype.processSearchResults = function(results) {

	// sanitize the results value
	if (results && results.length === 0) {
		results = [];
	}

	// cache the results
	if (results) {
		this.results = results;
	}

	// handle markers
	this.showOrHideMarkers();

	// populate search result list
	this.populateResultList();

	// handle overlay visability
	this.showOrHideOverlay();

};

// ------------------------------------------------------------------------
// Overlay handling
// ------------------------------------------------------------------------
Module.prototype.showOrHideOverlay = function() {
	if (this.results.length === 0) {
		this.showOverlay();
	} else {
		this.hideOverlay();
	}
};

Module.prototype.showOverlay = function() {
	this.$overlay.addClass(this.options.stateClasses.visible);
};

Module.prototype.hideOverlay = function() {
	this.$overlay.removeClass(this.options.stateClasses.visible);
};

// ------------------------------------------------------------------------
// Marker handling
// ------------------------------------------------------------------------
Module.prototype.showOrHideMarkers = function() {
	if (this.results.length === 0) {
		$document.trigger(events.removeMarkers);
		return;
	}

	// trigger event to remove all markers, handled in m00_-_journey_map
	$document.trigger(events.removeMarkers);

	// trigger event to add new markers, handled in m00_-_journey_map
	$document.trigger(events.addMarkers, [this.results]);
};

// ------------------------------------------------------------------------
// Result list
// ------------------------------------------------------------------------

// Populate results list
Module.prototype.populateResultList = function() {

	var listItems;

	if (this.results.length === 0) {

		// description text
		this.$resultDescription[0].innerHTML = this.options.i18n.noResultsDescription;

		// empty the list
		this.$resultList.empty();

		this.$resultDescription.removeClass(this.options.stateClasses.visuallyhidden);

		return;

	}

	this.$resultDescription.addClass(this.options.stateClasses.visuallyhidden);

	// description text
	$.each(this.$resultDescription, function(index, item) {
		item.innerHTML = this.options.i18n.resultsDescription.replace('$0', this.results.length);
	}.bind(this));

	// create string containing list items and links to append to list
	listItems = this.results.map(function(item) {
		return '<li class="' + this.options.expandableResultListItemClass + '"><a class="' + this.options.expandableResultListItemLinkClass + '" href="' + item.url + '">' + item.label + '</a></li>';
	}.bind(this)).join('');

	// set the contents of the list
	this.$resultList[0].innerHTML = listItems;

};

// ------------------------------------------------------------------------
// Module generics
// ------------------------------------------------------------------------

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 383 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 384 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(385);


/***/ }),
/* 385 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 386 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(387);


/***/ }),
/* 387 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 388 */,
/* 389 */,
/* 390 */,
/* 391 */,
/* 392 */,
/* 393 */,
/* 394 */,
/* 395 */,
/* 396 */,
/* 397 */,
/* 398 */,
/* 399 */,
/* 400 */,
/* 401 */,
/* 402 */,
/* 403 */,
/* 404 */,
/* 405 */,
/* 406 */,
/* 407 */,
/* 408 */,
/* 409 */,
/* 410 */,
/* 411 */,
/* 412 */,
/* 413 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(414);


/***/ }),
/* 414 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M00 - Reveal
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(415);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'reveal',
	events = {},
	defaults = {
		domSelectors: {
			content: '[data-' + name + '="content"]',
			trigger: '[data-' + name + '="trigger"]'
		},
		stateClasses: {
			isRevealed: 'is_revealed'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	this.$content = this.$element.find(this.options.domSelectors.content);
	this.$trigger = this.$element.find(this.options.domSelectors.trigger);

	this.$element
		.on('click.' + this.uuid, this.options.domSelectors.trigger, function(event) {
			event.preventDefault();

			if (!this.$element.hasClass(this.options.stateClasses.isRevealed)) {
				this.$element.addClass(this.options.stateClasses.isRevealed);
				this.$content.attr('aria-hidden', 'false');
				this.$trigger.attr('aria-expanded', 'true');
				return;
			}

			this.$element.removeClass(this.options.stateClasses.isRevealed);
			this.$content.attr('aria-hidden', 'true');
			this.$trigger.attr('aria-expanded', 'false');

		}.bind(this));
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 415 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 416 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(417);


/***/ }),
/* 417 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 418 */,
/* 419 */,
/* 420 */,
/* 421 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(422);


/***/ }),
/* 422 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M37 - iFrames
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */


// Ensure that the SASS file is built.
__webpack_require__(423);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'iframes',
	events = {},
	defaults = {
		domSelectors: {
			iFrame: '[data-' + name + '="iframe"]'
		},
		stateClasses: {}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	// require the iframe-resizer helper
	__webpack_require__.e/* require.ensure */(13).then((function(require) {
			this.iFrameResize = __webpack_require__(410);
			setTimeout(function() {
				this._init();
			}.bind(this), 1);
		}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);
};

Module.prototype._init = function() {

	// resize and scroll options
	var shouldResize = (this.$element.attr('data-options-resize') === 'true'),
		shouldScroll = (this.$element.attr('data-options-scroll') === 'true'),
		$iFrame = this.$element.find(this.options.domSelectors.iFrame);

	// configure iFrame-reiszer plugin
	$iFrame.iFrameResize({
		log: false, // disable logging
		scrolling: shouldScroll, // enable/disable scrolling
		sizeHeight: shouldResize, // true, if the iframe should automatically calculate and set the height to fit the content
		checkOrigin: false,
		messageCallback: function(options) {

			// Messages are sent from within the iframe as follows:
			//
			//     window.parentIFrame.sendMessage('disable-lightbox-iframe');
			//
			// The following two messages can be sent:
			//
			//     * enable-lightbox-iframe
			//     * disable-lightbox-iframe
			//
			// They put the iframe into full screen and back.

			var message = options.message;

			if (!this.$body) {
				this.$body = $('body');
			}

			if (message === 'enable-lightbox-iframe') {
				$iFrame.addClass('is_lightbox_iframe');
				this.$body.addClass('is_lightbox_iframe_open');
				return;
			}

			if (message === 'disable-lightbox-iframe') {
				$iFrame.removeClass('is_lightbox_iframe');
				this.$body.removeClass('is_lightbox_iframe_open');
				return;
			}

		}.bind(this)
	});

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 423 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 424 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(425);


/***/ }),
/* 425 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 426 */,
/* 427 */,
/* 428 */,
/* 429 */,
/* 430 */,
/* 431 */,
/* 432 */,
/* 433 */,
/* 434 */,
/* 435 */,
/* 436 */,
/* 437 */,
/* 438 */,
/* 439 */,
/* 440 */,
/* 441 */,
/* 442 */,
/* 443 */,
/* 444 */,
/* 445 */,
/* 446 */,
/* 447 */,
/* 448 */,
/* 449 */,
/* 450 */,
/* 451 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(452);


/***/ }),
/* 452 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M26.1 - Search Alternative
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(453);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	searchButtonTemplate = __webpack_require__(454);

var name = 'searchalternative',
	events = {},
	defaults = {
		domSelectors: {
			searchField: '[data-init="autocomplete"]',
			clearButton: '[data-autocomplete="clear_button"]',
			searchButton: '[data-searchalt="button"]'
		},
		stateClasses: {
			searchButtonHidden: 'is_searchbutton_hidden'
		},
		i18n: {
			searchAltButtonLabel: 'Aktueller Inhalt im Eingabefeld löschen'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	// cache elements
	this.$searchField = this.$element.find(this.options.domSelectors.searchField);

	// add search button template markup
	this.$searchButton = $(searchButtonTemplate(this.options.i18n)).insertAfter(this.$searchField);

	// attach change handler
	this.attachChangeHandlers();

	// initially handle button visability
	this.updateValue();
};

/**
 * Attach change and click handlers
 */
Module.prototype.attachChangeHandlers = function() {

	// attach value change listener to input field
	this.$element.on('input' + '.' + this.uuid, this.options.domSelectors.searchField, this.updateValue.bind(this));

	// attach click handler on clear button
	this.$element.on('click' + '.' + this.uuid, this.options.domSelectors.clearButton, this.updateValue.bind(this));

	// attach click handler on search button
	this.$element.on('click' + '.' + this.uuid, this.options.domSelectors.searchButton, this.handleSearchButtonClick.bind(this));

	// listen to reset event
	this.$element.on('resetInput' + '.' + this.uuid, this.resetToInitState.bind(this));

	// make it possible to trigger search button click from outside
	this.$element.on('searchButtonClick' + '.' + this.uuid, this.handleSearchButtonClick.bind(this));
};

/**
 * Handle input value change
 */
Module.prototype.updateValue = function() {

	// cache value of search field
	this.value = this.$searchField.val().trim();

	// update search button visabilty
	this.updateSearchButton();
};

/**
 * Show/hide search button
 */
Module.prototype.updateSearchButton = function() {

	// show/hide button depending on if we have a value.
	if (!this.value) {
		this.hideSearchButton();
	} else {
		this.showSearchButton();
	}
};

/**
 * Handle search button click
 */
Module.prototype.handleSearchButtonClick = function() {
	this.$element.addClass(this.options.stateClasses.searchButtonHidden);
	this.$searchButton.fadeOut();
	this.$searchField.blur();
};

/**
 * Show search button
 */
Module.prototype.showSearchButton = function() {

	// we only need to proceed if the button is not already visible
	if (this.$searchButton.is(':visible')) {
		return;
	}

	// remove hidden class
	this.$element.removeClass(this.options.stateClasses.searchButtonHidden);

	// show button
	this.$searchButton.fadeIn();

	// remove aria attributes
	this.$searchButton.removeAttr('role aria-hidden');
};

/**
 * Hide search button
 */
Module.prototype.hideSearchButton = function() {

	// hide button
	this.$searchButton.fadeOut();

	// add aria attributes
	this.$searchButton.attr({
		'aria-hidden': true,
		'role': 'presentation'
	});
};

/**
 * Reset to initial state
 */
Module.prototype.resetToInitState = function() {
	this.$searchField.val('');
	this.updateValue();
	this.$element.find(this.options.domSelectors.clearButton).fadeOut();
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 453 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 454 */
/***/ (function(module, exports, __webpack_require__) {

var Handlebars = __webpack_require__(8);
function __default(obj) { return obj && (obj.__esModule ? obj["default"] : obj); }
module.exports = (Handlebars["default"] || Handlebars).template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
    var helper;

  return "<button type=\"button\" class=\"mod_search_alt_search_button\" data-searchalt=\"button\">\n	<span class=\"visuallyhidden\">"
    + this.escapeExpression(((helper = (helper = helpers.searchAltButtonLabel || (depth0 != null ? depth0.searchAltButtonLabel : depth0)) != null ? helper : helpers.helperMissing),(typeof helper === "function" ? helper.call(depth0,{"name":"searchAltButtonLabel","hash":{},"data":data}) : helper)))
    + "</span>\n	<svg class=\"mod_svgsprite_icon var_search\" focusable=\"false\">\n		<use xlink:href=\"#SBB_12_search\" />\n	</svg>\n</button>\n";
},"useData":true});

/***/ }),
/* 455 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(456);


/***/ }),
/* 456 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33 - Multistep Form
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2021.
 */



__webpack_require__(457);

var $ = __webpack_require__(0),
	$document = $(document),
	SuperClass = __webpack_require__(1),
	ajaxRequest = __webpack_require__(16),
	lscache = __webpack_require__(56);

__webpack_require__(3);

var name = 'multistepform',
	events = {
		multistepFormButtonClicked: 'multistepForm.buttonClicked',
		multistepFormValidationSuccess: 'multistepForm.validationSuccess',
		multistepFormValidationError: 'multistepForm.validationError',
		multistepFormRestore: 'multistepForm.restore',
		multistepFormSubmitSuccess: 'multistepForm.submitSuccess',
		multistepFormUpdateSummary: 'multistepForm.updateSummary',
		multistepFormDisableNextSteps: 'multistepForm.disableNextSteps',
		processIndicatorTriggerFormValidation: 'processIndicator.triggerFormValidation',
		processIndicatorShowStepTarget: 'processIndicator.showStepTarget',
		processIndicatorCurrentStepPosition: 'processIndicator.currentStepPosition',
		checkboxRadioPanelClicked: 'checkboxRadioPanel.clicked',
		updateAllowedDates: 'datepicker.updateAllowedDates',
		removeAllowedTimes: 'datepicker.removeAllowedTimes'
	},
	defaults = {
		domSelectors: {
			form: '[data-' + name + '="form"]',
			notification: '[data-' + name + '="notification"]',
			notificationMessage: '[data-notification="message"]',
			step: '[data-' + name + '="step"]',
			dynamicPanel: '[data-' + name + '="dynamic_panel"]',
			dynamicSelect: '[data-' + name + '="dynamic_select"]',
			dynamicDatepicker: '[data-' + name + '="dynamic_datepicker"]',
			datepicker: '[data-init="datepicker"]',
			datepickerAllowedTimes: '[data-datepicker="allowed_time_item_input"]',
			stepIndependentFields: '[data-' + name + '="step_independent_fields"]',
			hiddenReservationId: '[data-' + name + '="hidden_reservation_id"]',
			hiddenPrevReservationId: '[data-' + name + '="hidden_prev_reservation_id"]',
			prevButton: '[data-' + name + '="prev"]',
			nextButton: '[data-' + name + '="next"]',
			submitButton: '[data-' + name + '="submit"]',
			successPanel: '[data-' + name + '="successPanel"]',
			successPanelHeading: '[data-' + name + '="successPanelHeading"]',
			processIndicatorButton: '[data-processindicator="button"]',
			captcha: '[data-init="captcha"]',
			summaryReasonOfVisit: '[data-multistepformsummary="reasonOfVisit"]',
			summaryLocation: '[data-multistepformsummary="location"]',
			summaryDate: '[data-multistepformsummary="date"]'
		},
		stateClasses: {
			isVisible: 'is_visible',
			isHidden: 'is_hidden'
		},
		hasMultiSteps: true,
		onlineConsultationDates: false,
		prefillEndpoint: false,
		i18n: {
			accessibilityTextPrevButton: 'Zurück zu Schritt {{step}}',
			accessibilityTextNextButton: 'Weiter zu Schritt {{step}}'
		},
		stepIdAttributeKey: 'data-multistepform-step-id',
		dynamicDataAttributeKey: 'data-multistepform-dynamic-data',
		stepTargetIdAttributeKey: 'data-processindicator-step-target-id',
		datepickerOptionsAttributeKey: 'data-datepicker-options',
		localStorageOptions: {
			cacheExpirationTime: 30, // in minutes
			bucketName: name,
			dataKey: '-data',
			uploadedFilesKey: 'ls-uploaded-files'
		},
		stepIndependentFieldsKey: 'step_independent_fields'
	};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ---------------------------------------------------------------------------
// INIT
// ---------------------------------------------------------------------------
Module.prototype.init = function() {
	this.stepId = '1.1';
	this.nextStepWasClicked = false;
	this.reasonOfVisitHasChanged = true;
	this.locationHasChanged = true;
	this.dateHasChanged = true;
	this.prefilledReasonOfVisit = false;
	this.prefilledLocation = false;
	this.prefilledDate = false;

	this._cacheElements();
	this._defineParsleyValidationGroups();
	this._setupEventHandlers();
	this._setupParsleyEventHandlers();

	this.formId = this.$form.attr('id');

	this.currentStep = this.$steps[0];

	this.isPrefilledForm = false;
	this.prefillDate = null;

	this._setInitialState();

	if (this.options.prefill) {
		this.isPrefilledForm = true;
		this._prefillLocation();

		this.prefillDate = {
			id: this.options.prefill.appointmentCapacity,
			start: this.options.prefill.start,
			end: this.options.prefill.end
		};
	}
};

Module.prototype._cacheElements = function() {
	this.$form = this.$element.find(this.options.domSelectors.form);
	this.$notification = this.$element.find(this.options.domSelectors.notification);
	this.$notificationMessage = this.$notification.find(this.options.domSelectors.notificationMessage);
	this.$steps = this.$element.find(this.options.domSelectors.step);
	this.$dynamicSelect = this.$element.find(this.options.domSelectors.dynamicSelect);
	this.$dynamicDatepicker = this.$element.find(this.options.domSelectors.dynamicDatepicker);
	this.$stepIndependentFields = this.$element.find(this.options.domSelectors.stepIndependentFields);
	this.$hiddenReservationId = this.$element.find(this.options.domSelectors.hiddenReservationId);
	this.$hiddenPrevReservationId = this.$element.find(this.options.domSelectors.hiddenPrevReservationId);
	this.$prevButton = this.$element.find(this.options.domSelectors.prevButton);
	this.$nextButton = this.$element.find(this.options.domSelectors.nextButton);
	this.$submitButton = this.$element.find(this.options.domSelectors.submitButton);
	this.$captcha = this.$element.find(this.options.domSelectors.captcha);
	this.$summaryReasonOfVisit = this.$element.find(this.options.domSelectors.summaryReasonOfVisit);
	this.$summaryLocation = this.$element.find(this.options.domSelectors.summaryLocation);
	this.$summaryDate = this.$element.find(this.options.domSelectors.summaryDate);
	this.$successPanel = this.$element.find(this.options.domSelectors.successPanel);
	this.$successPanelHeading = this.$element.find(this.options.domSelectors.successPanelHeading);
};

// AKA reason
Module.prototype._prefillLocation = function() {
	if (!this.prefilledReasonOfVisit) {
		var reasonInputSelector = 'input[value="' + this.options.prefill.reasonOfVisit + '"]',
			$reasonOfVisitPanel = this.$steps.filter('[data-multistepform-dynamic-data="locations"]'),
			$reasonOfVisitInput = $reasonOfVisitPanel.find(reasonInputSelector);

		if ($reasonOfVisitInput.length) {
			$reasonOfVisitInput.attr('checked', 'checked');
			this._addSummaryReasonOfVisit($reasonOfVisitInput.get(0));
		}

		this.prefilledReasonOfVisit = true;
	}
};

// AKA location
Module.prototype._prefillDates = function() {
	if (!this.prefilledLocation) {
		var locationOptionSelector = 'option[value="' + this.options.prefill.posId + '"]',
			$locationOptionsElement = this.$dynamicSelect.find(locationOptionSelector);

		if ($locationOptionsElement.length) {
			$locationOptionsElement.attr('selected', 'selected');
			this._addSummaryLocation(this.$dynamicSelect);
		}

		this.prefilledLocation = true;
	}
};

// AKA Date
Module.prototype._prefillReservation = function() {
	if (!this.prefilledDate) {
		var reservationInputSelector = 'input[value="' + this.options.prefill.appointmentCapacity + '"]',
			$reservationPanel = this.$steps.filter('[data-multistepform-dynamic-data="reservation"]'),
			$reservationInput = $reservationPanel.find(reservationInputSelector),
			$reservationInputLabel = $reservationPanel.find('#' + $reservationInput.prop('id') + '_label');

		if ($reservationInput.length && $reservationInputLabel.length) {
			$reservationInput.prop('checked', true);
			this.$summaryDate.text(this.selectedDate + ', ' + $reservationInputLabel.get(0).innerText);
		}

		this.prefilledDate = true;
	}
};

Module.prototype._setupEventHandlers = function() {
	// trigger an event if prev button was clicked
	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.prevButton,
		function() {
			var isPrevButton = true;

			this.$element.trigger(events.multistepFormButtonClicked, isPrevButton);
		}
	.bind(this));

	// trigger an event if next button was clicked
	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.nextButton,
		function() {
			var isPrevButton = false;

			this.$element.trigger(events.multistepFormButtonClicked, isPrevButton);
		}
	.bind(this));

	$document.on(
		events.processIndicatorTriggerFormValidation + '.' + this.uuid,
		function(_, nextStepWasClicked) {

			// if a previous step was clicked don't validate the form
			if (nextStepWasClicked) {

				var currentStepIndex = this.$form.find('.' + this.options.stateClasses.isVisible).attr('id');

				this.nextStepWasClicked = true;

				this.$form.parsley().whenValidate({
					group: currentStepIndex
				}).done(function() {}.bind(this));

			} else {
				this.nextStepWasClicked = false;

				this.$form.parsley().trigger('form:success');

				this.$form.parsley().reset();
				$document.triggerHandler('formchange', [{id: this.formId}]);

				this._setupParsleyEventHandlers();

			}
		}
	.bind(this));

	$document.on(
		events.processIndicatorShowStepTarget + '.' + this.uuid,
		function(_, stepTargetId) {
			this.stepId = stepTargetId;

			if (this.options.onlineConsultationDates) {
				this._handleOnlineConsultationDates();
			}

			this._toggleVisibilityOfSteps();
		}
	.bind(this));

	$document.on(
		events.processIndicatorCurrentStepPosition + '.' + this.uuid,
		function(_, isFirstStep, isLastStep, prevStepNumber, nextStepNumber) {
			this._updateNavigationButtons(isFirstStep, isLastStep, prevStepNumber, nextStepNumber);
		}
	.bind(this));

	this.$element.on(
		'blur.' + this.uuid,
		'input, textarea, select',
		function() {
			this._saveValuesToLocalStorage();
		}
	.bind(this));

	this.$element.on(
		events.checkboxRadioPanelClicked + '.' + this.uuid,
		function(_, selectedPanel) {
			if (this.options.onlineConsultationDates) {
				this._addSummaryReasonOfVisit(selectedPanel);
			}
		}
	.bind(this));

	this.$element.on(
		'change.' + this.uuid,
		this.options.domSelectors.dynamicSelect,
		function(event) {
			if (this.options.onlineConsultationDates) {
				this._addSummaryLocation($(event.currentTarget));
			}
		}
	.bind(this));

	this.$element.on(
		'change.' + this.uuid,
		this.options.domSelectors.datepickerAllowedTimes,
		function(event) {
			if (this.options.onlineConsultationDates) {
				var selectedTime = event.currentTarget.labels[0].innerText;

				// add selected date and time to the summary
				this.$summaryDate.text(this.selectedDate + ', ' + selectedTime);

				this.dateHasChanged = true;

				// keep the previous reservation id to send it to the backend and cancel the previous reservation
				this.$hiddenPrevReservationId.val(this.$hiddenReservationId.val());
			}
		}
	.bind(this));

	this.$element.on(
		events.multistepFormUpdateSummary + '.' + this.uuid,
		function(_, date) {
			if (this.options.onlineConsultationDates) {
				this.selectedDate = date;
			}
		}
	.bind(this));

	this.$element.on(
		'submit.' + this.uuid,
		function(event) {
			event.preventDefault();
			this._saveValuesToLocalStorage();
			this._handleSubmit();
		}
	.bind(this));
};

Module.prototype._setupParsleyEventHandlers = function() {
	this.$form.parsley().on(
		'form:success', function() {
			this.$element.trigger(events.multistepFormValidationSuccess);
		}
		.bind(this));

	this.$form.parsley().on(
		'form:error', function() {
			this.$element.trigger(events.multistepFormValidationError);
		}.bind(this));
};

Module.prototype._defineParsleyValidationGroups = function() {

	this.$steps.each(function(index, step) {
		var $step = $(step);

		$step.find('input, textarea, select')
			.attr('data-parsley-group', $step.attr('id'));
	});

};

Module.prototype._setInitialState = function() {
	// append localStorageOptions dataKey suffix based on form id
	this.options.localStorageOptions.dataKey += '-' + this.$form.attr('id');

	// check if we have values in the local storage
	this.localStorageEntry = this._readFromLocalStorage();

	if (this.localStorageEntry) {
		this.setInitialValuesBasedOnLocalStorageContent();

		return;
	}

	// if no values in the local storage update the navigation buttons for the first step
	if (this.options.hasMultiSteps) {
		this._updateNavigationButtons(true, false, false, 2);
	} else {
		this._showStepIndependentFields(true);
	}
};

Module.prototype.setInitialValuesBasedOnLocalStorageContent = function() {
	var $element,
		typeOfElement,
		radioCheckboxValue,
		$checkedDynamicPanel;

	// restore input values
	this.localStorageEntry.inputValues.forEach(function(item) {
		$element = this.$form.find('[name="' + item.name + '"]');

		if ($element.length === 0) {
			return;
		}

		typeOfElement = $element[0].type;

		if (typeOfElement === 'radio' || typeOfElement === 'checkbox') {
			$element.each(function(_, radioCheckbox) {
				radioCheckboxValue = $(radioCheckbox).val() === item.value ? true : false;
				$(radioCheckbox).prop('checked', radioCheckboxValue);
			});
		} else {
			$element.val(item.value);
		}

		// for the online consultation dates fetch the locations first
		if (this.options.onlineConsultationDates) {
			if ($element[0] === this.$dynamicSelect[0]) {
				this._fetchOnlineConsultationData('locations', item.value);
			}
		}
	}.bind(this));

	if (this.options.onlineConsultationDates) {
		// if the current step is the third or the fourth, then set the current step
		// to the second because provisional reservations are only valid for five
		// minutes (at qsome api)
		if (this.localStorageEntry.currentStep === '3.1' || this.localStorageEntry.currentStep === '4.1') {
			this.localStorageEntry.currentStep = '2.1';
		}

		// get checked radio in the first step
		$checkedDynamicPanel = $(this.$element.find(this.options.domSelectors.dynamicPanel + ':checked'));

		this._addSummaryReasonOfVisit($checkedDynamicPanel[0]);
	}

	// restore process indicator and the current step
	this.$element.trigger(events.multistepFormRestore, [this.localStorageEntry.currentPath, this.localStorageEntry.currentStep]);
};

// -------------------------------------------------------------
// TOGGLE STEPS
// -------------------------------------------------------------
Module.prototype._toggleVisibilityOfSteps = function() {
	this.$steps.each(function(_, element) {
		var $step = $(element);

		// get the value of the attribute and show the step with this id and hide all other steps
		if ($step.attr(this.options.stepIdAttributeKey) === this.stepId) {
			$step.addClass(this.options.stateClasses.isVisible);
			this.currentStep = $step[0];
		} else {
			$step.removeClass(this.options.stateClasses.isVisible);
		}
	}.bind(this));

	this._saveValuesToLocalStorage();
};

// -------------------------------------------------------------
// UPDATE NAVIGATION BUTTONS
// -------------------------------------------------------------
Module.prototype._updateNavigationButtons = function(isFirstStep, isLastStep, prevStepNumber, nextStepNumber) {

	this.$nextButton.addClass(this.options.stateClasses.isVisible);
	this.$prevButton.addClass(this.options.stateClasses.isVisible);
	this.$submitButton.removeClass(this.options.stateClasses.isVisible);
	this._showStepIndependentFields(false);
	this._updateAria(nextStepNumber, this.$nextButton, this.options.i18n.accessibilityTextNextButton);
	this._updateAria(prevStepNumber, this.$prevButton, this.options.i18n.accessibilityTextPrevButton);

	if (isFirstStep) {
		this.$prevButton.removeClass(this.options.stateClasses.isVisible);
	}

	if (isLastStep) {
		this.$nextButton.removeClass(this.options.stateClasses.isVisible);
		this.$submitButton.addClass(this.options.stateClasses.isVisible);
		this._showStepIndependentFields(true);
	}
};

Module.prototype._updateAria = function(stepNumber, $button, accessibilityText) {
	var ariaLabel = accessibilityText.replace('{{step}}', stepNumber);

	$button.attr('aria-label', ariaLabel);
};

// -------------------------------------------------------------
// STEP INDEPENDENT FIELDS
// -------------------------------------------------------------
Module.prototype._showStepIndependentFields = function(show) {
	// we display the captcha and other hidden fields (which are important for the BE) independently
	// from the last step because we don't want the fields to be rendered more than once
	if (show && this.$captcha.length === 1) {
		this.$stepIndependentFields.addClass(this.options.stateClasses.isVisible);
	} else {
		this.$stepIndependentFields.removeClass(this.options.stateClasses.isVisible);
	}
};

// -------------------------------------------------------------
// LOCAL STORAGE
// -------------------------------------------------------------
Module.prototype._initLocalStorage = function() {
	lscache.setBucket(this.options.localStorageOptions.bucketName);
};

Module.prototype._readFromLocalStorage = function() {
	this._initLocalStorage();

	var savedValues = lscache.get(this.options.localStorageOptions.dataKey);

	if (savedValues) {
		return savedValues;
	}

	return false;
};

Module.prototype._saveValuesToLocalStorage = function() {
	this._initLocalStorage();

	var data = {};

	data.currentStep = this.stepId;
	data.currentPath = this._getCurrentFormPath();
	data.inputValues = this.$form.serializeArray();

	lscache.set(this.options.localStorageOptions.dataKey, data, this.options.localStorageOptions.cacheExpirationTime);
};

Module.prototype._getCurrentFormPath = function() {
	var $processIndicatorButtons = $(this.options.domSelectors.processIndicatorButton),
		stepTargetId,
		formPath = [];

	// get ids of the steps based on the path that the user has taken and push them in an array
	$processIndicatorButtons.each(function(_, button) {
		stepTargetId = $(button).attr(this.options.stepTargetIdAttributeKey);
		formPath.push(stepTargetId);
	}.bind(this));

	return formPath;
};

Module.prototype._cleanLocalStorage = function() {
	this._initLocalStorage();
	lscache.remove(this.options.localStorageOptions.dataKey);
};

// -------------------------------------------------------------
// ONLINE CONSULTATION DATES
// -------------------------------------------------------------
Module.prototype._handleOnlineConsultationDates = function() {
	var dataKey = $(this.currentStep).attr(this.options.dynamicDataAttributeKey);

	if (!this.nextStepWasClicked) {
		return;
	}

	// fetch data if the attribute 'data-multistepform-dynamic-data' is set and a the next button was clicked
	if ((dataKey === 'locations' && this.reasonOfVisitHasChanged) || (dataKey === 'dates' && this.locationHasChanged) || (dataKey === 'reservation' && this.dateHasChanged)
	) {
		this._fetchOnlineConsultationData(dataKey);
	}
};

Module.prototype._fetchOnlineConsultationData = function(dataKey, restoreLocationValue) {
	var $currentStep = $('[' + this.options.dynamicDataAttributeKey + '="' + dataKey + '"]'),
		stepData = $currentStep.serialize();

	if (dataKey === 'locations') {
		this.stepReasonOfVisit = $currentStep.serialize();
	} else {
		stepData += '&' + this.stepReasonOfVisit;
	}

	ajaxRequest(this.options.dynamicDataRequestUrl, true, stepData + '&data=' + dataKey)
		.then(function(response) {
			if (response[dataKey]) {
				if (dataKey === 'locations') {
					this._handleLocations(dataKey, response, restoreLocationValue);
				} else if (dataKey === 'dates') {
					this._handleDates(dataKey, response);
				} else if (dataKey === 'reservation') {
					this._handleReservation(response);
				}

				if (this.options.prefill) {
					// Locations is initial prefilled and reservation will be set on datepicker event callback
					// See _handleDates()
					if (dataKey === 'locations') {
						this._prefillDates();
					}
				}
			} else {
				this._showErrorMessage(this.options.i18n.generalErrorMessage);
			}
		}.bind(this))
		.catch(function() {
			this._showErrorMessage(this.options.i18n.generalErrorMessage);
		}.bind(this));
};

// make locations available for selection in the select
Module.prototype._handleLocations = function(dataKey, response, restoreLocationValue) {
	var selectOptions = this.$dynamicSelect.find('option'),
		$selectOption;

	// remove added select options first
	selectOptions.each(function(index, option) {
		$selectOption = $(option);

		if ($selectOption.val() !== '') {
			$selectOption.remove();
		}
	});

	// add options to the select
	$.each(response[dataKey], function(index, location) {
		this.$dynamicSelect
			.append($('<option>', {
				value: location.id,
				text: location.city
			}));
	}.bind(this));

	if (restoreLocationValue) {
		this.$dynamicSelect.val(restoreLocationValue);
		this._addSummaryLocation(this.$dynamicSelect);
		this._saveValuesToLocalStorage();
	}

	this.reasonOfVisitHasChanged = false;
};

// make dates available for selection in the datepicker
Module.prototype._handleDates = function(dataKey, response) {
	var $datepicker = this.$dynamicDatepicker.closest(this.options.domSelectors.datepicker),
		datepickerOptions = $datepicker.attr(this.options.datepickerOptionsAttributeKey),
		datepickerOptionsObj = JSON.parse(datepickerOptions),
		allowedDatesArray = [];

	// push dates in an array
	$.each(response[dataKey], function(index, dateItem) {
		allowedDatesArray.push(dateItem.start);
	}.bind(this));

	// add the dates and times to the specific option
	datepickerOptionsObj.allowedDates = allowedDatesArray.join();
	datepickerOptionsObj.allowedTimes = response[dataKey];

	if (this.isPrefilledForm) {
		$datepicker.on('datepicker.allowedTimesDisplayed', this._prefillReservation.bind(this));
	}

	// add the updated options to the datepicker options
	$datepicker.attr(this.options.datepickerOptionsAttributeKey, JSON.stringify(datepickerOptionsObj));

	this.$element.trigger(events.updateAllowedDates, datepickerOptionsObj.allowedDates);
	this.locationHasChanged = false;
};

// retrieve the ID of the temporary reservation to submit it on the last step
Module.prototype._handleReservation = function(response) {
	this.$hiddenReservationId.val(response.reservation);
	this.dateHasChanged = false;
};

Module.prototype._addSummaryReasonOfVisit = function(selectedPanel) {
	var selectedReasonOfVisit = selectedPanel.labels[0].innerText,
		splittedReasonOfVisit = selectedReasonOfVisit.split('\n');

	this.$summaryReasonOfVisit.text(splittedReasonOfVisit[0] + ' (' + splittedReasonOfVisit[1] + ')');

	this.reasonOfVisitHasChanged = true;
	this.$element.trigger(events.multistepFormDisableNextSteps);
};

Module.prototype._addSummaryLocation = function($dynamicSelect) {
	var selectedLocation = $dynamicSelect.find(':selected').text();

	this.$summaryLocation.text(selectedLocation);

	this.locationHasChanged = true;
	this.$element.trigger(events.removeAllowedTimes);
	this.$element.trigger(events.multistepFormDisableNextSteps);
};

// -------------------------------------------------------------
// SUBMIT
// -------------------------------------------------------------
Module.prototype._handleSubmit = function() {
	// We need to send the form values with FormData() because the WAF blocks
	// our large base64 strings (files). We send all the values of the form
	// (expect files) as JSON (stringified) to keep the assignment of the fields
	// to the steps. We convert the base64 strings to blobs and send the files
	// separately and have the reference in the JSON (stringified).
	var requestParam = 'frm_form_data'; // the prefix 'frm_' was requested by WAF team to identify our form data and let them pass

	this.formData = new FormData();

	this._getFormValues()
		.then(function(values) {
			this.formData.append(requestParam, JSON.stringify(values));

			$.ajax({
				url: this.options.requestUrl,
				method: 'POST',
				data: this.formData,
				cache: false,
				contentType: false,
				processData: false
			}).then(function(data) {
				if (data.success) {
					this._cleanLocalStorage();
					if (this.isPrefilledForm) {
						this._toggleSuccessPanel();
					} else {
						window.location.href = data.redirectPath;
					}

					this.$element.trigger(events.multistepFormSubmitSuccess);
				} else {
					this._showErrorMessage(data.message);
				}
			}.bind(this))
			.catch(function() {
				this._showErrorMessage(this.options.i18n.generalErrorMessage);
			}.bind(this));
		}.bind(this));
};

Module.prototype._toggleSuccessPanel = function() {
	this.$form.toggle(this.options.stateClasses.isHidden);
	this.$successPanel.toggle(this.options.stateClasses.isHidden);
	this.$successPanelHeading.focus();
};

Module.prototype._getFormValues = function() {
	var currentPath = this._getCurrentFormPath(),
		values = {},
		$step,
		singleStepId,
		promise;

	promise = new Promise(function(resolve) {
		// if the form has no multi steps and no process indicator serialize the whole form
		if (!this.options.hasMultiSteps) {
			singleStepId = this.$steps.attr(this.options.stepIdAttributeKey);
			values[singleStepId] = this.$steps.serializeArray();
		} else {
			// get ids of the steps based on the path that the user has taken and get the values of these steps
			currentPath.forEach(function(stepTargetId) {
				$step = $('[' + this.options.stepIdAttributeKey + '="'+ stepTargetId + '"]');
				values[stepTargetId] = $step.serializeArray();
			}.bind(this));
		}

		// get the values of the captcha and other hidden fields which are important for the BE (:formpath, _charset_, etc.)
		values[this.options.stepIndependentFieldsKey] = this.$stepIndependentFields.serializeArray();

		// get uploaded files
		this._getFileInputs(values)
			.then(function() {
				resolve(values);
			}.bind(this));

	}.bind(this));

	return promise;
};

Module.prototype._getFileInputs = function(valuesArray) {
	var promise,
		$fileInputs = this.$form.find('input[type="file"]'),
		fileInputCounter = 0;

	promise = new Promise(function(resolve) {
		if ($fileInputs.length === 0) {
			resolve();
			return;
		}

		$fileInputs.each(function(_, fileInput) {
			this._getFilesFromLocalStorage(fileInput, valuesArray)
				.then(function() {
					fileInputCounter += 1;

					// not resolving until all files from all inputs are read
					if ($fileInputs.length === fileInputCounter) {
						resolve();
					}
				}.bind(this));
		}.bind(this));
	}.bind(this));

	return promise;
};

Module.prototype._getFilesFromLocalStorage = function(fileInput, valuesArray) {
	var promise,
		$fileInput = $(fileInput),
		fileInputId = $(fileInput).attr('id'),
		uploadedFiles = JSON.parse(window.localStorage.getItem(this.options.localStorageOptions.uploadedFilesKey + '-' + fileInputId)),
		$parentStep = $fileInput.closest(this.options.domSelectors.step),
		stepId = $parentStep.attr(this.options.stepIdAttributeKey),
		fileReferences = [],
		fileReferenceId,
		fileCounter = 0;

	promise = new Promise(function(resolve) {
		if (!uploadedFiles) {
			resolve();
		}

		uploadedFiles.files.forEach(function(file) {
			this._convertBase64ToBlob(file)
				.then(function(blob) {
					fileReferenceId = 'frm_' + file.nameOfInput + '_' + fileCounter; // prefix with _frm to avoid WAF blocking

					// add file to the FormData object
					this.formData.append(fileReferenceId, blob, file.name);

					// push the file reference id into an array
					fileReferences.push({
						name: file.nameOfInput,
						fileReference: fileReferenceId
					});

					fileCounter += 1;
				}.bind(this));
		}.bind(this));

		// find the position where the file references should be pushed into the values array
		valuesArray = valuesArray[stepId];
		valuesArray.push(fileReferences);

		resolve();

	}.bind(this));

	return promise;
};

Module.prototype._convertBase64ToBlob = function(file) {
	var promise,
		base64File = file.value,
		contentType = file.type,
		byteCharacters = window.atob(base64File),
		sliceSize = 512,
		slice,
		byteArrays = [],
		byteNumbers,
		byteArray,
		blob;

	promise = new Promise(function(resolve) {
		// convert the base64 string to a blob
		for (var offset = 0, length = byteCharacters.length; offset < length; offset += sliceSize) {
			slice = byteCharacters.slice(offset, offset + sliceSize);
			byteNumbers = new Array(slice.length);

			for (var i = 0; i < slice.length; i++) {
				byteNumbers[i] = slice.charCodeAt(i);
			}

			byteArray = new Uint8Array(byteNumbers);
			byteArrays.push(byteArray);
		}

		blob = new Blob(byteArrays, {type: contentType});

		resolve(blob);

	}.bind(this));

	return promise;
};

Module.prototype._showErrorMessage = function(errorText) {
	this.$notification.removeClass(this.options.stateClasses.isHidden);
	this.$notificationMessage.text(errorText);
	this.$submitButton.removeAttr('disabled');
};

// ---------------------------------------------------------------------------
// MODULE GENERIC
// ---------------------------------------------------------------------------
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 457 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 458 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(459);


/***/ }),
/* 459 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M150 - Process Indicator
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2021.
 */



__webpack_require__(460);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	$document = $(document),
	createUrlParams = __webpack_require__(79),
	mediaqueries = __webpack_require__(6);

var name = 'processindicator',
	events = {
		processIndicatorTriggerFormValidation: 'processIndicator.triggerFormValidation',
		processIndicatorCurrentStepPosition: 'processIndicator.currentStepPosition',
		processIndicatorShowStepTarget: 'processIndicator.showStepTarget',
		multistepFormButtonClicked: 'multistepForm.buttonClicked',
		multistepFormValidationSuccess: 'multistepForm.validationSuccess',
		multistepFormValidationError: 'multistepForm.validationError',
		multistepFormRestore: 'multistepForm.restore',
		multistepFormDisableNextSteps: 'multistepForm.disableNextSteps',
		checkboxRadioPanelClicked: 'checkboxRadioPanel.clicked',
		multistepFormSubmitSuccess: 'multistepForm.submitSuccess'
	},
	defaults = {
		domSelectors: {
			button: '[data-' + name + '="button"]',
			desc: '[data-' + name + '="desc"]'
		},
		stateClasses: {
			completed: 'var_completed',
			current: 'var_current',
			invalid: 'var_invalid',
			disabled: 'var_disabled'
		},
		i18n: {
			accessibilityTextStep: 'Formularschritt',
			accessibilityTextCurrent: 'derzeit aktiv',
			accessibilityTextCompleted: 'abgeschlossen',
			accessibilityTextDisabled: 'noch deaktiviert',
			accessibilityTextInvalid: 'noch nicht abgeschlossen'
		},
		stepNumberAttributeKey: 'data-processindicator-step-number',
		stepTargetIdAttributeKey: 'data-processindicator-step-target-id',
		triageStepTargetIdAttributeKey: 'data-triage-steptarget-id',
		urlKeys: {
			stepId: 'stepId'
		}
	};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ---------------------------------------------------------------------------
// INIT
// ---------------------------------------------------------------------------
Module.prototype.init = function() {
	this._setInitState();
	this._cacheElements();
	this._setupEventHandlers();

	this._prepareHistory();
};

Module.prototype._setInitState = function() {
	this.$currentStep = this.$element.find('.' + this.options.stateClasses.current);
	this.$currentStepButton = '';
	this.$clickedStep = '';
	this.$clickedStepButton = '';
	this.clickedStepTargetId = '1.1';
	this.currentStepIsRestored = false;
};

Module.prototype._cacheElements = function() {
	this.$button = this.$element.find(this.options.domSelectors.button);
};

Module.prototype._setupEventHandlers = function() {
	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.button,
		function(event) {
			var $eventTarget = $(event.currentTarget);

			this.$clickedStep = $eventTarget.parent();
			this.$clickedStepButton = this.$clickedStep.find(this.options.domSelectors.button);
			this.clickedStepTargetId = $eventTarget.attr(this.options.stepTargetIdAttributeKey);
			this.$currentStepButton = this.$currentStep.find(this.options.domSelectors.button);
			this.currentStepTargetId = this.$currentStepButton.attr(this.options.stepTargetIdAttributeKey);

			this._triggerFormValidation();
		}
	.bind(this));

	$document.on(
		events.multistepFormValidationSuccess + '.' + this.uuid,
		this._handleStepButtonClick.bind(this)
	.bind(this));

	$document.on(
		events.multistepFormSubmitSuccess + '.' + this.uuid,
		function() {
			this.$currentStep
				.removeClass(this.options.stateClasses.current)
				.removeClass(this.options.stateClasses.invalid)
				.addClass(this.options.stateClasses.completed);
			this._updateAria(this.options.i18n.accessibilityTextCompleted, this.$currentStep);
			this._deactivateProcessIndicators();
		}
	.bind(this));

	$document.on(
		events.multistepFormValidationError + '.' + this.uuid,
		this._markCurrentStepAsInvalid.bind(this)
	.bind(this));

	$document.on(
		events.multistepFormRestore + '.' + this.uuid,
		function(_, path, currentStepId) {
			this._restoreStepFromLocalStorage(path, currentStepId);
		}
	.bind(this));

	$document.on(
		events.multistepFormButtonClicked + '.' + this.uuid,
		function(_, prevButtonClicked) {
			this._navigateForwardOrBackward(prevButtonClicked);
		}
	.bind(this));

	$document.on(
		events.multistepFormDisableNextSteps + '.' + this.uuid,
		this._markNextStepsAsDisabled.bind(this)
	.bind(this));

	$document.on(
		events.checkboxRadioPanelClicked + '.' + this.uuid,
		function(_, panel) {
			var stepTargetId = $(panel).attr(this.options.triageStepTargetIdAttributeKey);

			this._updateStepTargetIdOnNextStep(stepTargetId);
		}
	.bind(this));
};

Module.prototype._deactivateProcessIndicators = function() {
	this.$button.each(function(_, button) {
		var $buttonTrigger = $(button),
		buttonParent = $buttonTrigger.parent();

		$buttonTrigger.attr('disabled', true);
		buttonParent.addClass(this.options.stateClasses.disabled);
	}.bind(this));
};

Module.prototype._triggerFormValidation = function() {
	var nextStepWasClicked = true;

	// check if the clicked step is a previous one or the current step was restored
	if (this.currentStepTargetId > this.clickedStepTargetId || this.currentStepIsRestored) {
		nextStepWasClicked = false;
	}

	this.$element.trigger(events.processIndicatorTriggerFormValidation, nextStepWasClicked);
};

Module.prototype._handleStepButtonClick = function() {
	this.$element.trigger(events.processIndicatorShowStepTarget, this.clickedStepTargetId);
	this._handleFocusElement();
	this._markPrevStepAsCompleted(this.$currentStep);
	this._checkStateOfClickedStep();
	this._scrollToTopOfProcessIndicator();
	this._checkPositionOfClickedStep();

	if (!this.restoreStepFromHistory) {
		this._handlePushState();
	}

	this.$currentStep = this.$clickedStep;
};

Module.prototype._markPrevStepAsCompleted = function($currentStep) {
	$currentStep
		.removeClass(this.options.stateClasses.current)
		.removeClass(this.options.stateClasses.invalid)
		.removeClass(this.options.stateClasses.disabled)
		.addClass(this.options.stateClasses.completed);

	this._updateAria(this.options.i18n.accessibilityTextCompleted, $currentStep);
};

Module.prototype._markNextStepsAsDisabled = function() {
	var $nextSteps = this.$currentStep.nextAll();

	$nextSteps.each(function(_, step) {
		var $step = $(step);

		$step
			.removeClass(this.options.stateClasses.current)
			.removeClass(this.options.stateClasses.invalid)
			.removeClass(this.options.stateClasses.completed)
			.addClass(this.options.stateClasses.disabled);

		this._updateAria(this.options.i18n.accessibilityTextDisabled, $step);
	}.bind(this));
};

Module.prototype._markCurrentStepAsInvalid = function() {
	this.$currentStep
		.removeClass(this.options.stateClasses.current)
		.addClass(this.options.stateClasses.invalid);

	this.$clickedStepButton.attr('disabled', true);

	this._updateAria(this.options.i18n.accessibilityTextInvalid, this.$currentStep);
};

Module.prototype._checkStateOfClickedStep = function() {
	if (this.$clickedStep.hasClass(this.options.stateClasses.completed)) {
		this._updateCompletedStep();
	}

	if (this.$clickedStep.hasClass(this.options.stateClasses.disabled)) {
		this._updateDisabledStep();
	}
};

Module.prototype._updateCompletedStep = function() {
	this.$clickedStep
		.removeClass(this.options.stateClasses.completed)
		.addClass(this.options.stateClasses.current);

	this._updateAria(this.options.i18n.accessibilityTextCurrent, this.$clickedStep);
};

Module.prototype._updateDisabledStep = function() {
	this.$clickedStep
		.removeClass(this.options.stateClasses.disabled)
		.addClass(this.options.stateClasses.current);

	this._updateAria(this.options.i18n.accessibilityTextCurrent, this.$clickedStep);
};

Module.prototype._checkPositionOfClickedStep = function() {
	var isFirstStep = false,
		isLastStep = false,
		prevStepNumber,
		nextStepNumber;

	if (this.$clickedStep.prev().length === 0) {
		isFirstStep = true;
	}

	if (this.$clickedStep.next().length === 0) {
		isLastStep = true;
	}

	prevStepNumber = this._getStepNumber(this.$clickedStep.prev());
	nextStepNumber = this._getStepNumber(this.$clickedStep.next());

	this.$element.trigger(events.processIndicatorCurrentStepPosition, [isFirstStep, isLastStep, prevStepNumber, nextStepNumber]);
};

Module.prototype._getStepNumber = function($step) {
	return $step.find(this.options.domSelectors.button).attr(this.options.stepNumberAttributeKey);
};

Module.prototype._navigateForwardOrBackward = function(prevButtonClicked) {
	var $stepToBeClicked,
		$stepButtonToBeClicked;

	if (prevButtonClicked) {
		// if prev button was clicked, click on the prev step button
		this.$currentStep.prev().find(this.options.domSelectors.button).trigger('click');
	} else {
		// if next button was clicked, click on the next step button
		$stepToBeClicked = this.$currentStep.next();
		$stepButtonToBeClicked = $stepToBeClicked.find(this.options.domSelectors.button);

		this._clickOnDisabledStep($stepButtonToBeClicked);
	}
};

Module.prototype._updateStepTargetIdOnNextStep = function(stepTargetId) {
	this.$currentStep
		.next()
		.find(this.options.domSelectors.button)
		.attr(this.options.stepTargetIdAttributeKey, stepTargetId);
};

Module.prototype._scrollToTopOfProcessIndicator = function() {
	var breakpoint = mediaqueries.getCurrentBreakpointName(),
	navHeight = (breakpoint === 'mobilePortrait' || breakpoint === 'mobileLandscape') ? 48 : 64,
	scrollToPaddingTop = 24;

	$('html, body').animate({
		scrollTop: this.$element.offset().top - navHeight - scrollToPaddingTop
	}, 200);
};

Module.prototype._handleFocusElement = function() {
	this.$clickedStepButton.focus();
};

Module.prototype._clickOnDisabledStep = function($step) {
	$step.removeAttr('disabled').trigger('click');
};

Module.prototype._updateAria = function(accessibilityText, $step) {
	var $stepButton = $step.find(this.options.domSelectors.button),
		stepNumber = $stepButton.attr(this.options.stepNumberAttributeKey),
		stepDesc = $step.find(this.options.domSelectors.desc).text();

	$step.attr('aria-current', this.$clickedStep === $step ? 'step' : 'false');
	$stepButton.attr('aria-label', this.options.i18n.accessibilityTextStep + ' ' + stepNumber + ' ' + stepDesc + ' ' + accessibilityText);
};

// -------------------------------------------------------------
// HISTORY
// -------------------------------------------------------------
Module.prototype._prepareHistory = function() {

	// set scrollRestoration to manual
	if ('scrollRestoration' in history) {
		history.scrollRestoration = 'manual';
	}

	// object to save current step for history state
	this.step = {};
	this.step[this.options.urlKeys.stepId] = 1.1;

	// set initial history state
	history.replaceState(this.step, null, null);

	// Add event listener for history's popstate to handle browser history navigation
	window.addEventListener('popstate', this._handlePopState.bind(this), false);
};

Module.prototype._handlePopState = function(event) {
	if (!event.state) {
		return;
	}

	// update step object to keep in sync with state object
	this.step = event.state;

	// restore step
	this._restoreStepFromHistory();
};

Module.prototype._restoreStepFromHistory = function() {
	this.clickedStepTargetId = this.step[this.options.urlKeys.stepId];
	this.restoreStepFromHistory = true;

	// find the step to be restored as the current one
	var $stepToRestore = this.$element.find('[' + this.options.stepTargetIdAttributeKey + '="' + this.clickedStepTargetId + '"]');

	this._clickOnDisabledStep($stepToRestore);
};

Module.prototype._handlePushState = function() {
	this.step = {};
	this.step[this.options.urlKeys.stepId] = this.clickedStepTargetId;
	history.pushState(this.step, null, createUrlParams(this.step));
};

Module.prototype._restoreStepFromLocalStorage = function(path, currentStepId) {
	this.$button.each(function(index, element) {
		var $stepButton = $(element),
			$step = $(element).parent();

		// restore path
		$stepButton.attr(this.options.stepTargetIdAttributeKey, path[index]);

		// as long as the current step is not restored, mark prev steps as completed
		if (!this.currentStepIsRestored) {
			$stepButton.removeAttr('disabled');
			this._markPrevStepAsCompleted($step);
		}

		// restore current step
		if ($stepButton.attr(this.options.stepTargetIdAttributeKey) === currentStepId) {
			this.currentStepIsRestored = true;

			this._clickOnDisabledStep($stepButton);
		}
	}.bind(this));

	this.currentStepIsRestored = false;
};

// ---------------------------------------------------------------------------
// MODULE GENERIC
// ---------------------------------------------------------------------------
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 460 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 461 */,
/* 462 */,
/* 463 */,
/* 464 */,
/* 465 */,
/* 466 */,
/* 467 */,
/* 468 */,
/* 469 */,
/* 470 */,
/* 471 */,
/* 472 */,
/* 473 */,
/* 474 */,
/* 475 */,
/* 476 */,
/* 477 */,
/* 478 */,
/* 479 */,
/* 480 */,
/* 481 */,
/* 482 */,
/* 483 */,
/* 484 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(485);


/***/ }),
/* 485 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 486 */,
/* 487 */,
/* 488 */,
/* 489 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(490);


/***/ }),
/* 490 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M46.1 - Filter Press
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



// Ensure that the SASS file is built.
__webpack_require__(491);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	parseUrl = __webpack_require__(69),
	ajaxRequest = __webpack_require__(16);

var name = 'filterpress',
	log = estatico.helpers.log(name),
	events = {
		keydown: 'keydown.estatico.' + name
	},
	defaults = {
		domSelectors: {
			subjectDropdown: '[data-init="select"]',
			tag: '[data-init="autocomplete"]',
			datepicker: '[data-init="datepicker"]',
			datepickerInput: '[data-datepicker="input"]',
			resetButton: '[data-pressreleasefilter="reset_button"]',
			tagSearchButton: '[data-searchalt="button"]',
			paginationPage: '[data-pagination="page"]',
			paginationPreviousPage: '[data-pagination="previous"]',
			paginationNextPage: '[data-pagination="next"]',
			resultsContainer: '[data-pressrelease="list"]'
		},
		stateClasses: {
			isVisible: 'is_visible'
		},
		urlOptionsAttribute: 'data-options-url',
		tagReturnKey: 13
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	var $datepickers = this.$element.find(this.options.domSelectors.datepicker);

	// object that holds all the filter values
	this.filterValues = {
		page: 1
	};

	// flag to indicate if start page should be reseted
	this.resetStartPage = false;

	// flag used to indicate that history is popping state
	this.isPoppingState = false;

	// cache datepickers and date input fields
	this.$datepicker1 = $($datepickers[0]);
	this.$datepicker2 = $($datepickers[1]);
	this.$dateInput1 = $(this.$element.find(this.options.domSelectors.datepickerInput)[0]);
	this.$dateInput2 = $(this.$element.find(this.options.domSelectors.datepickerInput)[1]);

	// cache subject and tag input fields
	this.$subject = this.$element.find(this.options.domSelectors.subjectDropdown);
	this.$tag = this.$element.find(this.options.domSelectors.tag);

	// cache the reset button
	this.$resetButton = this.$element.find(this.options.domSelectors.resetButton);

	// prepare object to hold datepicker ids along with their dates
	this.datepickerDates = {};
	this.initialDatepickerDates = {};
	this.datepickerDates[this.$datepicker1.attr('id')] = false;
	this.datepickerDates[this.$datepicker2.attr('id')] = false;

	// attach listeners
	this.attachInputListeners();

	// Add event listener for history's popstate to handle browser back button
	window.addEventListener('popstate', this.handlePopState.bind(this), false);
};

/**
 * Listen to input changes and button clicks
 */
Module.prototype.attachInputListeners = function() {

	// listen to valid date selection
	this.$element.on('changedAndValidated' + '.' + this.uuid, this.options.domSelectors.datePicker, this.handleValidDateChange.bind(this));

	// listen to invalid date selection
	this.$element.on('changedAndInvalidated' + '.' + this.uuid, this.options.domSelectors.datePicker, this.handleInvalidDateChange.bind(this));

	// subject dropdown change
	// Do not use delegated event, so that we can dispatch change event from outside.
	$(this.$element.find(this.options.domSelectors.subjectDropdown)).on('change' + '.' + this.uuid, this.immediateRequestAndAppendHtml.bind(this));

	// tag search button
	this.$element.on('click' + '.' + this.uuid, this.options.domSelectors.tagSearchButton, this.immediateRequestAndAppendHtml.bind(this));

	// listen to tag input clear button click
	this.$element.on('cleared' + '.' + this.uuid, this.options.domSelectors.tag, this.handleClearButtonClick.bind(this));

	// listen to reset button click
	this.$element.on('click' + '.' + this.uuid, this.options.domSelectors.resetButton, this.resetForm.bind(this));

	// add click handlers for page links
	this.$element.on('click' + '.' + this.uuid, this.options.domSelectors.paginationPage, this.handlePageClick.bind(this));
	this.$element.on('click' + '.' + this.uuid, this.options.domSelectors.paginationPreviousPage, this.handlePageClick.bind(this));
	this.$element.on('click' + '.' + this.uuid, this.options.domSelectors.paginationNextPage, this.handlePageClick.bind(this));

	// bind keydown event to tag input field
	this.$tag.on(events.keydown, function(event) {
		if (this.options.tagReturnKey === event.keyCode) {
			this.immediateRequestAndAppendHtml();
			this.$tag.blur();
		}
	}.bind(this));
};

/**
 * Initially parse url, set filter values according to url params and send
 * request based on those values
 */
Module.prototype.handleInitialUrl = function() {

	// we only need to restore the ui and issue a request, if we have url params
	if (location.search === '') {
		return;
	}

	this.isPoppingState = true;

	// restore dates
	this.restoreDatesFromUrl();

	// restore subject
	this.restoreSubjectFromUrl();

	// restore tag
	this.restoreTagInputFromUrl();

	// restore page
	this.restorePageFromUrl();

	this.isPoppingState = false;
	this.requestAndAppendHtml();
};

/**
 * Initialize everything needed to use history api
 */
Module.prototype.initHistoryApi = function() {

	// set scrollRestoration to manual
	if ('scrollRestoration' in history) {
		history.scrollRestoration = 'manual';
	}

	// set initial history state
	history.replaceState(this.filterValues, null, null);
};

/**
 * Handle history pop state
 */
Module.prototype.handlePopState = function(event) {

	this.isPoppingState = true;

	// keep filterValues in sync with history's state
	this.filterValues = event.state;

	if (!this.isStateObjectValid()) {
		return;
	}

	// restore filter components
	this.restoreDatesFromHistoryState();
	this.restoreSubjectFromHistoryState();
	this.restoreTagInputFromHistoryState();

	this.isPoppingState = false;

	// request and append html for new state
	this.requestAndAppendHtml();
};

/**
 * Handle history push state
 */
Module.prototype.pushState = function() {
	var date1,
		date2,
		subject,
		tag,
		page;

	// if history is popping state, then immediately return
	if (this.isPoppingState) {
		return;
	}

	if (!this.isStateObjectValid()) {
		return;
	}

	// make sure that we're not pushing the state that the current history
	// already has
	date1 = history.state.date1 === this.filterValues.date1;
	date2 = history.state.date2 === this.filterValues.date2;
	subject = history.state.subject === this.filterValues.subject;
	tag = history.state.tag === this.filterValues.tag;
	page = history.state.page === this.filterValues.page;

	if (date1 && date2 && subject && tag && page) {
		return;
	}

	history.pushState(this.filterValues, null, this.constructUrlForHistory());
};

/**
 * Restore methods for dates
 */
Module.prototype.restoreDatesFromHistoryState = function() {

	// get dates from history's state
	var date1 = this.filterValues.date1,
		date2 = this.filterValues.date2;

	this.restoreDates([date1, date2]);
};

Module.prototype.restoreDatesFromUrl = function() {

	// get dates from URL params
	var params = parseUrl(location.search),
		date1 = params.from,
		date2 = params.to;

	// save new value to history's proxy state object
	this.filterValues.date1 = date1;
	this.filterValues.date2 = date2;

	this.restoreDates([date1, date2]);
};

Module.prototype.restoreDates = function(dates) {

	if (dates.length !== 2) {
		return;
	}

	// restore date1
	this.$dateInput1.val(dates[0]);
	this.$dateInput1.blur(); // in datepicker, there is a listener on blur
	// event which will try to parse a string to a valid date, which is exactly
	// what we need here

	// restore date2
	this.$dateInput2.val(dates[1]);
	this.$dateInput2.blur();
};

/**
 * Restore methods for subject
 */
Module.prototype.restoreSubjectFromHistoryState = function() {

	// get subject from history
	var subject = this.filterValues.subject;

	this.restoreSubject(subject);
};

Module.prototype.restoreSubjectFromUrl = function() {

	// get subject from url
	var params = parseUrl(location.search),
		subject = params.subject;

	this.restoreSubject(subject);
};

Module.prototype.restoreSubject = function(subject) {

	if (!subject || subject.trim() === '') {
		return;
	}

	this.$subject.val(subject);
};

/**
 * Restore methods for tag
 */
Module.prototype.restoreTagInputFromHistoryState = function() {

	// get tag from history
	var tag = this.filterValues.tag;

	this.restoreTagInput(tag);
};

Module.prototype.restoreTagInputFromUrl = function() {

	// get tag from url
	var params = parseUrl(location.search),
		tag = params.tag;

	this.restoreTagInput(tag);
};

Module.prototype.restoreTagInput = function(tag) {

	if (!tag || tag.trim() === '') {
		tag = '';
	}

	this.$tag.val(decodeURIComponent(tag));
};

/**
 * Restore page from url
 */
Module.prototype.restorePageFromUrl = function() {

	// get page from url
	var params = parseUrl(location.search),
		page = params.page;

	this.filterValues.page = page;
};

/**
 * Handle datepicker value changed to valid
 */
Module.prototype.handleValidDateChange = function(datepicker, data) {

	if (!this.isPoppingState && this.firstRequestIsDone) {
		this.resetStartPage = true;
	}

	// save the changed date
	this.datepickerDates[datepicker.target.id] = data.date;

	// request html
	this.requestAndAppendHtml();

	// issue pushState to add url to history
	this.pushState();
};

/**
 * Handle datepicker value changed to invalid
 */
Module.prototype.handleInvalidDateChange = function(datepicker) {

	// save that the date was invalid
	this.datepickerDates[datepicker.target.id] = false;
};

/**
 * Handle click on clear button inside tag input field
 */
Module.prototype.handleClearButtonClick = function() {
	if (!this.isPoppingState) {
		this.resetStartPage = true;
	}

	this.requestAndAppendHtml();

	// issue pushState to add url to history
	this.pushState();
};

/**
 * Handle click on page
 */
Module.prototype.handlePageClick = function(event) {

	var pageLinkUrl;

	// prevent default action on link
	if (event) {
		event.preventDefault();
	}

	this.resetStartPage = false;

	// get the link inside the page element
	pageLinkUrl = $(event.currentTarget).attr('data-url');

	// save the page from url params
	this.filterValues.page = parseUrl(pageLinkUrl).page;

	// get html and append it to panel
	this.requestAndAppendHtml(false, true);

	// issue pushState to add url to history
	this.pushState();
};

/**
 * Immediate proxy to requestAndAppendHtml
 */
Module.prototype.immediateRequestAndAppendHtml = function() {
	if (!this.isPoppingState && this.firstRequestIsDone) {
		this.resetStartPage = true;
	}

	this.requestAndAppendHtml();

	// issue pushState to add url to history
	this.pushState();
};

/**
 * Get html via ajax-request and append result to panel
 */
Module.prototype.requestAndAppendHtml = function(newUrl, isSwitchingPages) {

	var url,
		$resultsContainer = this.$element.find(this.options.domSelectors.resultsContainer);

	// if history is currently popping, immediately return
	if (this.isPoppingState) {
		return;
	}

	// if the form is not valid, return
	if (!this.isFormValid()) {
		return;
	}

	// object to save filter values for history state and filter logic
	this.cacheFilterValues();

	// show or hide the reset button depending on form values
	this.showOrHideResetButton();

	// we can skip the first request after initialization, since BE is initally rendering
	// the first results.
	if (!this.firstRequestIsDone) {
		this.firstRequestIsDone = true;

		// here, the form is valid for the first time. That means, that we can grab
		// the initial dates propagated from the 'changedAndValidated' event here.
		// We need those for being able to reset to those values later.
		this.initialDatepickerDates = $.extend({}, this.datepickerDates);

		// save the initial dates to be able to reset to those values later
		this.initDate1 = this.$dateInput1.val();
		this.initDate2 = this.$dateInput2.val();

		this.initHistoryApi();

		// Initially parse url and show results based on url
		this.handleInitialUrl();

		return;
	}

	url = newUrl || this.constructUrlForRequest();

	ajaxRequest(url).then(function(html) {

		// append the content to the according panel
		$resultsContainer.replaceWith(html);

		// if we're handling a page switch, make sure to scroll the window to the top
		// of the tabs.
		if (isSwitchingPages) {
			this.scrollToTopOfFilter();
		}

	}.bind(this),

	// handle errors
	function(error) {
		log('Bei der Abfrage der Resultate ist leider ein Fehler aufgetreten.');
		log(error);
	});
};

/**
 * Constcut url for history state
 */
Module.prototype.constructUrlForHistory = function() {
	return this.constructUrl(location.origin + location.pathname);
};

/**
 * Constcut url for AJAX-Requst
 */
Module.prototype.constructUrlForRequest = function() {
	return this.constructUrl(this.$element.attr(this.options.urlOptionsAttribute));
};

/**
 * Construct url including paramaters based on passed url
 */
Module.prototype.constructUrl = function(url) {

	var newUrl = url + '?';

	if (!this.isStateObjectValid) {
		return;
	}

	if (this.resetStartPage) {
		this.filterValues.page = 1;
		this.resetStartPage = false;
	}

	newUrl += 'from=' + this.filterValues.date1 + '&';
	newUrl += 'to=' + this.filterValues.date2 + '&';
	newUrl += 'subject=' + this.filterValues.subject + '&';

	if (this.filterValues.tag) {
		newUrl += 'tag=' + this.filterValues.tag + '&';
	}

	newUrl += 'page=' + this.filterValues.page;

	return newUrl;
};

/**
 * Helper to cache the current filter values in an object
 */
Module.prototype.cacheFilterValues = function() {

	// make sure the form is valid
	if (!this.isFormValid) {
		return;
	}

	// get all values and save it to object
	this.filterValues.date1 = this.datepickerDates[this.$datepicker1.attr('id')];
	this.filterValues.date2 = this.datepickerDates[this.$datepicker2.attr('id')];
	this.filterValues.subject = this.$subject.val();
	this.filterValues.tag = encodeURIComponent(this.$tag.val()).trim();
};

/**
 * Check if the form is valid
 */
Module.prototype.isFormValid = function() {

	// get both strored dates
	var date1 = this.datepickerDates[this.$datepicker1.attr('id')],
		date2 = this.datepickerDates[this.$datepicker2.attr('id')];

	// check for falsy date values
	if (!date1 || !date2) {
		return false;
	}

	return true;
};

/**
 * Check if the state object from history is valid
 */
Module.prototype.isStateObjectValid = function() {

	var date1 = this.filterValues.date1 !== undefined,
		date2 = this.filterValues.date2 !== undefined,
		subject = this.filterValues.subject !== undefined,
		tag = this.filterValues.subject !== undefined,
		page = this.filterValues.page !== undefined;

	return (date1 && date2 && subject && tag && page);
};

/**
 * Reset the form to default state
 */
Module.prototype.resetForm = function() {

	this.resetStartPage = true;

	// reset dates
	this.$dateInput1.val(this.initDate1);
	this.$dateInput2.val(this.initDate2);

	// make sure that the new dates are handled in datepicker instance
	this.$dateInput1.trigger('input');
	this.$dateInput2.trigger('input');

	// reset the stored dates
	this.datepickerDates[this.$datepicker1.attr('id')] = this.initialDatepickerDates[this.$datepicker1.attr('id')];
	this.datepickerDates[this.$datepicker2.attr('id')] = this.initialDatepickerDates[this.$datepicker2.attr('id')];

	// reset subject dropdown
	this.$subject.val('all');

	// reset tag input
	this.$tag.trigger('resetInput');

	// issue a new request
	this.requestAndAppendHtml();

	// issue pushState to add url to history
	this.pushState();
};

/**
 * Show or hide reset button depending on form values
 */
Module.prototype.showOrHideResetButton = function() {
	if (this.filterDidChange()) {
		this.showResetButton();
	} else {
		this.hideResetButton();
	}
};

/**
 * Helper to determine if the filter did change from it's initial state
 */
Module.prototype.filterDidChange = function() {
	var date1 = this.datepickerDates[this.$datepicker1.attr('id')],
		date2 = this.datepickerDates[this.$datepicker2.attr('id')],
		initialDate1 = this.initialDatepickerDates[this.$datepicker1.attr('id')],
		initialDate2 = this.initialDatepickerDates[this.$datepicker2.attr('id')],
		date1Changed = false,
		date2Changed = false,
		subjectChanged = this.$subject.val() !== 'all',
		tagChanged = this.$tag.val().trim() !== '';

	if (initialDate1 && initialDate1 !== date1) {
		date1Changed = true;
	}

	if (initialDate2 && initialDate2 !== date2) {
		date2Changed = true;
	}

	return (date1Changed || date2Changed || subjectChanged || tagChanged);
};

/**
 * Show the reset button
 */
Module.prototype.showResetButton = function() {
	this.$resetButton.addClass(this.options.stateClasses.isVisible);
};

/**
 * Hide the reset button
 */
Module.prototype.hideResetButton = function() {
	this.$resetButton.removeClass(this.options.stateClasses.isVisible);
};

/**
 * Helper to scroll the window to the filter
 */
Module.prototype.scrollToTopOfFilter = function() {

	var offset = this.$element.offset().top,
		scrollTopWindow = $(window).scrollTop();

	// scroll window to make filter visible only if it is not already visible.
	if (scrollTopWindow > offset) {
		$('html, body').animate({
			scrollTop: offset
		}, 'slow');
	}
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 491 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 492 */,
/* 493 */,
/* 494 */,
/* 495 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(496);


/***/ }),
/* 496 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 497 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(498);


/***/ }),
/* 498 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 499 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(500);


/***/ }),
/* 500 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 501 */,
/* 502 */,
/* 503 */,
/* 504 */,
/* 505 */,
/* 506 */,
/* 507 */,
/* 508 */,
/* 509 */,
/* 510 */,
/* 511 */,
/* 512 */,
/* 513 */,
/* 514 */,
/* 515 */,
/* 516 */,
/* 517 */,
/* 518 */,
/* 519 */,
/* 520 */,
/* 521 */,
/* 522 */,
/* 523 */,
/* 524 */,
/* 525 */,
/* 526 */,
/* 527 */,
/* 528 */,
/* 529 */,
/* 530 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(531);


/***/ }),
/* 531 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M97 - Countdown Module
 */



// Ensure that the SASS file is built.
__webpack_require__(532);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	cssTransitionEnd = __webpack_require__(4);

var name = 'countdown',
	events = {},
	defaults = {
		domSelectors: {
			timer: '[data-' + name + '="timer"]',
			days: '[data-' + name + '="days"]',
			hours: '[data-' + name + '="hours"]',
			minutes: '[data-' + name + '="minutes"]',
			seconds: '[data-' + name + '="seconds"]',
			endText: '[data-' + name + '="end"]'
		},
		endDate: 'data-countdown-date',
		stateClasses: {
			isVisible: 'is_visible',
			isHidden: 'is_hidden'
		},
		showEndText: true
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	this._cacheElements();
	this._initIntersectionObserver();

	// this.initEndDate();

};

Module.prototype._cacheElements = function() {
	this.endDate = this.$element.attr(this.options.endDate);
	this.$timer = this.$element.find(this.options.domSelectors.timer);
	this.$days = this.$element.find(this.options.domSelectors.days);
	this.$hours = this.$element.find(this.options.domSelectors.hours);
	this.$minutes = this.$element.find(this.options.domSelectors.minutes);
	this.$seconds = this.$element.find(this.options.domSelectors.seconds);
	this.$endText = this.$element.find(this.options.domSelectors.endText);
	this.countdownInterval = null;
};

Module.prototype._initIntersectionObserver = function() {
	var intersectionObserver = new window.IntersectionObserver(this._handleIntersectionObserve.bind(this), {'root': null, 'rootMargin': '200px 0px 200px 0px'});

	intersectionObserver.observe(this.$element[0]);
};

Module.prototype._handleIntersectionObserve = function(changes) {

	changes.forEach(function(change) {
		if (change.isIntersecting) {
			if (!this.inViewport) {
				this.inViewport = true;
				this.initEndDate();
			}
		} else {
			if (this.inViewport) {
				this.inViewport = false;
				clearInterval(this.countdownInterval);
			}
		}
	}.bind(this));

};

Module.prototype.initEndDate = function() {
	var countdownDate = new Date(this.endDate).getTime();

	this.initCountdown(countdownDate);
};

Module.prototype.initCountdown = function(countdownDate) {

	var currentDate = new Date().getTime(),
		differenceDate = countdownDate - currentDate;

	this.prevDays = this.getDays(differenceDate);
	this.prevHours = this.getHours(differenceDate);
	this.prevMinutes = this.getMinutes(differenceDate);
	this.prevSeconds = this.getSeconds(differenceDate);

	if (differenceDate > 0) {

		// Set Initial State
		requestAnimationFrame(function() {
			this.$days.html(this.prevDays);
			this.$hours.html(this.prevHours);
			this.$minutes.html(this.prevMinutes);
			this.$seconds.html(this.prevSeconds);
		}.bind(this));

		// Start Interval
		this.countdownInterval = setInterval(function() {

			currentDate = new Date().getTime();
			differenceDate = countdownDate - currentDate;

			this.days = this.getDays(differenceDate);
			this.hours = this.getHours(differenceDate);
			this.minutes = this.getMinutes(differenceDate);
			this.seconds = this.getSeconds(differenceDate);

			if (this.days !== this.prevDays) {
				this.setDays();
			}

			if (this.hours !== this.prevHours) {
				this.setHours();
			}

			if (this.minutes !== this.prevMinutes) {
				this.setMinutes();
			}

			if (this.seconds !== this.prevSeconds) {
				this.setSeconds();
			}

		}.bind(this), 1000);

	} else {

		clearInterval(this.countdownInterval);

		if (this.options.showEndText) {
			this.$endText.addClass(this.options.stateClasses.isVisible);
		}

		this.$timer.css({
			opacity: 0
		})
		.on(cssTransitionEnd + '.' + this.uuid, function() {
			requestAnimationFrame(function() {
				this.$timer.addClass(this.options.stateClasses.isHidden);
			}.bind(this));
		}.bind(this));

	}

};

Module.prototype.getDays = function(differenceDate) {
	return Math.floor(differenceDate / (1000 * 60 * 60 * 24));
};

Module.prototype.getHours = function(differenceDate) {
	return ('0' + Math.floor((differenceDate % (1000 * 60 * 60 * 24)) / (1000 * 60 * 60))).slice(-2);
};

Module.prototype.getMinutes = function(differenceDate) {
	return ('0' + Math.floor((differenceDate % (1000 * 60 * 60)) / (1000 * 60))).slice(-2);
};

Module.prototype.getSeconds = function(differenceDate) {
	return ('0' + Math.floor((differenceDate % (1000 * 60)) / 1000)).slice(-2);
};

Module.prototype.setDays = function() {
	requestAnimationFrame(function() {
		this.$days.html(this.days);
		this.prevDays = this.days;
	}.bind(this));
};

Module.prototype.setHours = function() {
	requestAnimationFrame(function() {
		this.$hours.html(this.hours);
		this.prevHours = this.hours;
	}.bind(this));
};

Module.prototype.setMinutes = function() {
	requestAnimationFrame(function() {
		this.$minutes.html(this.minutes);
		this.prevMinutes = this.minutes;
	}.bind(this));
};

Module.prototype.setSeconds = function() {
	requestAnimationFrame(function() {
		this.$seconds.html(this.seconds);
		this.prevSeconds = this.seconds;
	}.bind(this));
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 532 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 533 */,
/* 534 */,
/* 535 */,
/* 536 */,
/* 537 */,
/* 538 */,
/* 539 */,
/* 540 */,
/* 541 */,
/* 542 */,
/* 543 */,
/* 544 */,
/* 545 */,
/* 546 */,
/* 547 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(548);


/***/ }),
/* 548 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 549 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(550);


/***/ }),
/* 550 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 551 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(552);



/***/ }),
/* 552 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 553 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(554);


/***/ }),
/* 554 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M43 - Search Results Filter
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */


// Ensure that the SASS file is built.
__webpack_require__(555);

var $ = __webpack_require__(0),
	$document = $(document),
	SuperClass = __webpack_require__(1),
	parseUrl = __webpack_require__(69),
	ajaxRequest = __webpack_require__(16);

var name = 'searchresultsfilter',
	log = estatico.helpers.log(name),
	events = {
		switchTabs: 'switch.estatico.tabs',
		tabsInitialized: 'initialized.estatico.tabs',
		switchFilterTag: 'subfilter.switchFilterTag'
	},
	defaults = {
		domSelectors: {
			tabs: '[data-tabs="tab_list_item"]',
			tabLinks: '[data-tabs="tab_list_item_link"]',
			panelInnerContainer: '[data-tabs="tabs_panel_item_inner"]',
			paginationPage: '[data-pagination="page"]',
			paginationPreviousPage: '[data-pagination="previous"]',
			paginationNextPage: '[data-pagination="next"]'
		},
		stateClasses: {},
		linkUrlAttributeName: 'data-url',
		tabHasTags: 'data-tabs-has-tags',
		linkHrefAttributeName: 'href',
		panelIdPrefix: 'panel-'
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	// cache the tabs
	this.$tabs = this.$element.find(this.options.domSelectors.tabs);

	// attach click handlers
	this.attachTabClickHandlers();

	// add popstate listener for history api
	this.attachHistoryStateHandlers();

	// add click handlers for page links
	var clickTargets = [
		this.options.domSelectors.paginationPage,
		this.options.domSelectors.paginationPreviousPage,
		this.options.domSelectors.paginationNextPage
	].join(',');

	this.$element.on('click.' + this.uuid, clickTargets, this.handlePageClick.bind(this));

	this.setActiveTab();

	$document.on(
		events.switchFilterTag + '.' + this.uuid,
		function(_, requestUrl) {
			this.requestAndAppendHtml(requestUrl, false);
		}
	.bind(this));
};

/**
 * Get Url Parameters
 */
Module.prototype.getUrlParameters = function() {
	return parseUrl(window.location.search);
};

/**
 * Get Combined Url and Target Url Parameters
 * @param targetUrl
 * @returns {string}
 */
Module.prototype.getCombinedUrlAndTargetParameters = function(targetUrl) {

	if (!targetUrl) {
		return;
	}

	var urlParams = this.getUrlParameters(),
		urlWithOutParams = targetUrl.split('?'),
		targetUrlParams = parseUrl(targetUrl),
		modifiedURL = urlWithOutParams[0] + '?';

	Object.keys(targetUrlParams).forEach(function(key) {
		urlParams[key] = targetUrlParams[key];
	});

	modifiedURL += $.param(urlParams);

	return modifiedURL;

};

/**
 * Replace parameter in query
 */
Module.prototype.replaceQueryParam = function(parameter, newValue, search) {
	var regex = new RegExp('([?;&])' + parameter + '[^&;]*[;&]?'),
		query = search.replace(regex, '$1').replace(/&$/, '');

	return (query.length > 2 ? query + '&' : '?') + (newValue ? parameter + '=' + newValue : '');
};

/**
 * Listen to changes in browser history
 */
Module.prototype.attachHistoryStateHandlers = function() {

	var $activeTab = this.$tabs.filter('.is_active'),
		$activeTabLink = $activeTab.find(this.options.domSelectors.tabLinks),
		targetUrl = $activeTabLink.attr(this.options.linkUrlAttributeName),
		modifiedTargetUrl = this.getCombinedUrlAndTargetParameters(targetUrl),
		state;

	// prevent scrolling on popstate event
	if ('scrollRestoration' in history) {
		history.scrollRestoration = 'manual';
	}

	// set state if active tab present
	if ($activeTab.length) {

		state = {
			target: $activeTabLink.attr(modifiedTargetUrl),
			url: $activeTabLink.attr(this.options.linkHrefAttributeName),
			tab: $activeTab.index()
		};

		// set current state
		history.replaceState(state, null, null);
	}

	// Add event listener for popstate to handle browser back button
	window.addEventListener('popstate', function(event) {

		this.requestAndAppendHtml(event.state.target, false);
		$document.trigger(events.switchTabs, event.state.tab);

	}.bind(this), false);
};

/**
 * Attach Click handlers to links inside tabs
 */
Module.prototype.attachTabClickHandlers = function() {

	this.$tabs.on('click', this.handleTabClick.bind(this));
};

/**
 * Initially set current tab based on url parameter 'category'
 */
Module.prototype.setActiveTab = function() {
	var urlParams = this.getUrlParameters(),
		panelId,
		tabToSelect,
		tabToSelectIndex,
		fakeEvent;

	if (!urlParams.category) {
		return;
	}

	panelId = this.options.panelIdPrefix + urlParams.category;
	tabToSelect = this.$element.find('[data-tab-id="' + panelId + '"]');
	tabToSelectIndex = this.$tabs.index(tabToSelect);

	fakeEvent = {
		currentTarget: tabToSelect[0],
		preventDefault: function() { }
	};

	$document.one(events.tabsInitialized, function() {
		this.handleTabClick(fakeEvent, true);
		$document.trigger(events.switchTabs, tabToSelectIndex);
	}.bind(this));

};

/**
 * Handle click on tab
 */
Module.prototype.handleTabClick = function(event, initialLoad) {

	var $tab = $(event.currentTarget),
		$tabLink,
		tabLinkUrl,
		tabLinkHref,
		tagFromUrl = this.getUrlParameters().tag,
		pageFromUrl = this.getUrlParameters().page,
		hasTags = $tab.attr(this.options.tabHasTags);

	// prevent default action on link
	event.preventDefault();

	// get the link inside the tab
	$tabLink = $tab.find(this.options.domSelectors.tabLinks);
	tabLinkUrl = this.getCombinedUrlAndTargetParameters($tabLink.attr(this.options.linkUrlAttributeName));
	tabLinkHref = this.getCombinedUrlAndTargetParameters($tabLink.attr(this.options.linkHrefAttributeName));

	// apply tag and page from URL on initial page load if any tag is set in url
	if (tagFromUrl !== undefined && pageFromUrl !== undefined) {
		if (hasTags && initialLoad) {
			tabLinkUrl = this.replaceQueryParam('tag', tagFromUrl, tabLinkUrl);
			tabLinkHref = this.replaceQueryParam('tag', tagFromUrl, tabLinkHref);
			tabLinkUrl = this.replaceQueryParam('page', pageFromUrl, tabLinkUrl);
			tabLinkHref = this.replaceQueryParam('page', pageFromUrl, tabLinkHref);
		} else {
			tabLinkUrl = this.replaceQueryParam('tag', '0', tabLinkUrl);
			tabLinkHref = this.replaceQueryParam('tag', '0', tabLinkHref);
			tabLinkUrl = this.replaceQueryParam('page', '1', tabLinkUrl);
			tabLinkHref = this.replaceQueryParam('page', '1', tabLinkHref);
		}
	}

	// get html and append it to panel
	this.requestAndAppendHtml(tabLinkUrl, false);
	this.pushState(tabLinkUrl, tabLinkHref, $tab.index());
};

/*
 * Push new state with data.
 */
Module.prototype.pushState = function(target, url, index) {
	var state = {
		tab: index,
		target: target,
		url: url
	};

	history.pushState(state, null, url);
};

/**
 * Handle click on page
 */
Module.prototype.handlePageClick = function(event) {

	var $tab = this.$tabs.filter('is_active'),
		$page = $(event.currentTarget),
		pageLinkUrl,
		pageLinkHref;

	// prevent default action on link
	event.preventDefault();

	// get the link inside the page element
	pageLinkUrl = $page.attr(this.options.linkUrlAttributeName);
	pageLinkHref = $page.attr(this.options.linkHrefAttributeName);

	pageLinkUrl = this.getCombinedUrlAndTargetParameters(pageLinkUrl);
	pageLinkHref = this.getCombinedUrlAndTargetParameters(pageLinkHref);

	this.requestAndAppendHtml(pageLinkUrl, true);
	this.pushState(pageLinkUrl, pageLinkHref, $tab.index());
};

/**
 * Get html via ajax-request and append result to panel
 */
Module.prototype.requestAndAppendHtml = function(target, isSwitchingPages) {

	var urlParameters,
		panelId,
		$panelInnerContainer;

	// url-parameters as object
	urlParameters = parseUrl(target);

	ajaxRequest(target).then(function(html) {

		// construct panelId according to search category
		panelId = this.options.panelIdPrefix + urlParameters.category;

		// get inner panel container
		$panelInnerContainer = $('#' + panelId).find(this.options.domSelectors.panelInnerContainer).first();

		// append the content to the according panel
		$panelInnerContainer.html(html);

		// Triggers ajax load to reinitialize modules inside the tab panel
		window.requestAnimationFrame(function() {
			$document.triggerHandler('ajaxload');
		}.bind(this));

		// if we're handling a page switch, make sure to scroll the window to the top
		// of the tabs
		if (isSwitchingPages) {
			this.scrollToTopOfTabs();
		}

	}.bind(this),

	// handle errors
	function(error) {
		log('Bei der Abfrage der Resultate ist leider ein Fehler aufgetreten.');
		log(error);
	});
};

/**
 * Helper to scroll the window to the top of the tabs
 */
Module.prototype.scrollToTopOfTabs = function() {

	var firstTabOffset = this.$tabs.first().offset().top,
		scrollTopWindow = $(window).scrollTop();

	// scroll window to make tabs visible only, if they are not already visible.
	if (scrollTopWindow > firstTabOffset) {
		$('html, body').animate({
			scrollTop: firstTabOffset
		}, 'slow');
	}
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 555 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 556 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(557);


/***/ }),
/* 557 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 558 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(559);


/***/ }),
/* 559 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 560 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(561);


/***/ }),
/* 561 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 562 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(563);


/***/ }),
/* 563 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 564 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(565);


/***/ }),
/* 565 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 566 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(567);


/***/ }),
/* 567 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


__webpack_require__(568);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'pwainstallnotification',
	events = {},
	defaults = {
		domSelectors: {
		},
		stateClasses: {
		},
		i18n: {
			logo: {
				src: 'https://cdn.app.sbb.ch/favicons/v1/apple-touch-icon-76x76.png'
			},
			title: 'Schneller zum Fahrplan und zu Infos der SBB?',
			description: 'Fügen Sie SBB.ch zu Ihrem Startbildschirm hinzu und nutzen Sie die Seite wie eine App.',
			ctaText: 'Jetzt hinzufügen'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	window.addEventListener(
		'beforeinstallprompt',
		this.showPWAInstallNotification.bind(this)
	);

	if (window.preventedBeforeInstallPromptEvent) {
		// In the head.js, we already registered to the beforeinstallprompt in
		// order to prevent it everywhere.
		// If the init() method was called after the BeforeInstallPromptEvent was
		// triggered, we need to manually call showPWAInstallNotification()
		this.showPWAInstallNotification(
			window.preventedBeforeInstallPromptEvent
		);
	}

};

Module.prototype.showPWAInstallNotification = function showPWAInstallNotification(beforeInstallPromptEvent) {

	this.beforeInstallPromptEvent = beforeInstallPromptEvent;
	var hasCrosslink = !!$('.mod_crosslinkteaser')[0];

	__webpack_require__.e/* require.ensure */(14).then((function(require) {
			var html = __webpack_require__(470)(this.options),
				installPWA = function installPWA() {
					this.beforeInstallPromptEvent.prompt();
					this.$notification.slideUp();
				}.bind(this);

			this.$notification = $(html).hide();

			var observer = new window.IntersectionObserver(function(entries) {

				if (!entries[0].isIntersecting) {
					return;
				}

				observer.disconnect();
			});

			$('main').append(this.$notification);

			this.$notification
			.find('[data-pwa-install-notification="button"]')
			.click(installPWA);

			if (hasCrosslink) {
				this.$notification.find('.mod_pwa_install_notification')
				.addClass('var_crosslink_spacing');
			}

			this.$notification.slideDown(function() {
				observer.observe(this.$notification[0]);
			}.bind(this));

		}.bind(this)).bind(null, __webpack_require__)).catch(__webpack_require__.oe);

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 568 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 569 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(570);


/***/ }),
/* 570 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 571 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(572);


/***/ }),
/* 572 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 573 */,
/* 574 */,
/* 575 */,
/* 576 */,
/* 577 */,
/* 578 */,
/* 579 */,
/* 580 */,
/* 581 */,
/* 582 */,
/* 583 */,
/* 584 */,
/* 585 */,
/* 586 */,
/* 587 */,
/* 588 */,
/* 589 */,
/* 590 */,
/* 591 */,
/* 592 */,
/* 593 */,
/* 594 */,
/* 595 */,
/* 596 */,
/* 597 */,
/* 598 */,
/* 599 */,
/* 600 */,
/* 601 */,
/* 602 */,
/* 603 */,
/* 604 */,
/* 605 */,
/* 606 */,
/* 607 */,
/* 608 */,
/* 609 */,
/* 610 */,
/* 611 */,
/* 612 */,
/* 613 */,
/* 614 */,
/* 615 */,
/* 616 */,
/* 617 */,
/* 618 */,
/* 619 */,
/* 620 */,
/* 621 */,
/* 622 */,
/* 623 */,
/* 624 */,
/* 625 */,
/* 626 */,
/* 627 */,
/* 628 */,
/* 629 */,
/* 630 */,
/* 631 */,
/* 632 */,
/* 633 */,
/* 634 */,
/* 635 */,
/* 636 */,
/* 637 */,
/* 638 */,
/* 639 */,
/* 640 */,
/* 641 */,
/* 642 */,
/* 643 */,
/* 644 */,
/* 645 */,
/* 646 */,
/* 647 */,
/* 648 */,
/* 649 */,
/* 650 */,
/* 651 */,
/* 652 */,
/* 653 */,
/* 654 */,
/* 655 */,
/* 656 */,
/* 657 */,
/* 658 */,
/* 659 */,
/* 660 */,
/* 661 */,
/* 662 */,
/* 663 */,
/* 664 */,
/* 665 */,
/* 666 */,
/* 667 */,
/* 668 */,
/* 669 */,
/* 670 */,
/* 671 */,
/* 672 */,
/* 673 */,
/* 674 */,
/* 675 */,
/* 676 */,
/* 677 */,
/* 678 */,
/* 679 */,
/* 680 */,
/* 681 */,
/* 682 */,
/* 683 */,
/* 684 */,
/* 685 */,
/* 686 */,
/* 687 */,
/* 688 */,
/* 689 */,
/* 690 */,
/* 691 */,
/* 692 */,
/* 693 */,
/* 694 */,
/* 695 */,
/* 696 */,
/* 697 */,
/* 698 */,
/* 699 */,
/* 700 */,
/* 701 */,
/* 702 */,
/* 703 */,
/* 704 */,
/* 705 */,
/* 706 */,
/* 707 */,
/* 708 */,
/* 709 */,
/* 710 */,
/* 711 */,
/* 712 */,
/* 713 */,
/* 714 */,
/* 715 */,
/* 716 */,
/* 717 */,
/* 718 */,
/* 719 */,
/* 720 */,
/* 721 */,
/* 722 */,
/* 723 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * Main JS script for WCMS pages. Inserted before closing body tag.
*/


__webpack_require__(93);

// Modules
__webpack_require__(176);
__webpack_require__(153);
__webpack_require__(197);
__webpack_require__(350);
__webpack_require__(155);
__webpack_require__(160);
__webpack_require__(200);
__webpack_require__(413);
__webpack_require__(241);
__webpack_require__(342);
__webpack_require__(373);
__webpack_require__(110);
__webpack_require__(168);
__webpack_require__(95);
__webpack_require__(215);
__webpack_require__(547);
__webpack_require__(97);
__webpack_require__(81);
__webpack_require__(233);
__webpack_require__(205);
__webpack_require__(220);
__webpack_require__(83);
__webpack_require__(207);
__webpack_require__(59);
__webpack_require__(53);
__webpack_require__(127);
__webpack_require__(115);
__webpack_require__(194);
__webpack_require__(178);
__webpack_require__(129);
__webpack_require__(236);
__webpack_require__(290);
__webpack_require__(292);
__webpack_require__(209);
__webpack_require__(181);
__webpack_require__(238);
__webpack_require__(90);
__webpack_require__(100);
__webpack_require__(416);
__webpack_require__(85);
__webpack_require__(484);
__webpack_require__(164);
__webpack_require__(451);
__webpack_require__(57);
__webpack_require__(131);
__webpack_require__(133);
__webpack_require__(171);
__webpack_require__(549);
__webpack_require__(345);
__webpack_require__(103);
__webpack_require__(67);
__webpack_require__(455);
__webpack_require__(551);
__webpack_require__(118);
__webpack_require__(63);
__webpack_require__(212);
__webpack_require__(136);
__webpack_require__(87);
__webpack_require__(107);
__webpack_require__(191);
__webpack_require__(253);
__webpack_require__(257);
__webpack_require__(166);
__webpack_require__(112);
__webpack_require__(183);
__webpack_require__(186);
__webpack_require__(189);
__webpack_require__(724);
__webpack_require__(421);
__webpack_require__(70);
__webpack_require__(246);
__webpack_require__(249);
__webpack_require__(727);
__webpack_require__(113);
__webpack_require__(553);
__webpack_require__(556);
__webpack_require__(730);
__webpack_require__(732);
__webpack_require__(489);
__webpack_require__(734);
__webpack_require__(736);
__webpack_require__(495);
__webpack_require__(497);
__webpack_require__(738);
__webpack_require__(424);
__webpack_require__(499);
__webpack_require__(378);
__webpack_require__(381);
__webpack_require__(305);
__webpack_require__(354);
__webpack_require__(308);
__webpack_require__(251);
__webpack_require__(558);
__webpack_require__(560);
__webpack_require__(384);
__webpack_require__(218);
__webpack_require__(562);
__webpack_require__(386);
__webpack_require__(741);
__webpack_require__(744);
__webpack_require__(746);
__webpack_require__(564);
__webpack_require__(748);
__webpack_require__(751);
__webpack_require__(753);
__webpack_require__(755);
__webpack_require__(376);
__webpack_require__(356);
__webpack_require__(566);
__webpack_require__(530);
__webpack_require__(359);
__webpack_require__(757);
__webpack_require__(569);
__webpack_require__(571);
__webpack_require__(760);
__webpack_require__(762);
__webpack_require__(141);
__webpack_require__(764);
__webpack_require__(766);
__webpack_require__(458);
__webpack_require__(768);
__webpack_require__(771);
__webpack_require__(773);
__webpack_require__(776);
__webpack_require__(779);
__webpack_require__(781);

// Campaigns
__webpack_require__(370);

// Webshop
__webpack_require__(139);

// Infoportal
__webpack_require__(310);

// Fun
__webpack_require__(784);


/***/ }),
/* 724 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(725);


/***/ }),
/* 725 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M33.25 - Multifield Block
 */



__webpack_require__(726);

var $ = __webpack_require__(0),
	templateSettings = __webpack_require__(27),
	SuperClass = __webpack_require__(1),
	$document = $(document),
	template = __webpack_require__(51),
	mediaqueries = __webpack_require__(6);

var name = 'multifieldblock',
	events = {
		blockPrefilled: 'multifieldblock.blockPrefilled'
	},
	defaults = {
		domSelectors: {
			form: 'form',
			blockTemplate: '[data-' + name + '=block_template]',
			emptyBlockTemplate: '[data-' + name + '=empty_block_template]',
			block: '[data-' + name + '="block"]',
			emptyBlock: '[data-' + name + '="empty_block"]',
			addButton: '[data-' + name + '="add_button"]',
			removeButton: '[data-' + name + '="remove_button"]'
		},
		stateClasses: {
			isVisible: 'is_visible'
		},
		allowRemovingLastBlock: false,
		maxBlocks: false,
		copyFieldContent: false,
		prefillBlocks: false
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ---------------------------------------------------------------------------
// INIT
// ---------------------------------------------------------------------------
Module.prototype.init = function() {
	templateSettings.interpolate = /{{([\s\S]+?)}}/g; // change interpolate settings to accept handlebars syntax

	this.blockIndexCounter = 1;

	this._cacheElements();
	this._addEventListeners();

	if (this.options.prefillBlocks.length > 0) {
		this._prefillBlock();
	} else {
		this._addBlock(false);
	}

	if (this.options.allowRemovingLastBlock) {
		this._addEmptyBlock();
	}
};

// ---------------------------------------------------------------------------
// CACHE ELEMENTS
// ---------------------------------------------------------------------------
Module.prototype._cacheElements = function() {
	this.formId = this.$element.closest(this.options.domSelectors.form).attr('id');
	this.$blockTemplate = this.$element.find(this.options.domSelectors.blockTemplate);
	this.$emptyBlockTemplate = this.$element.find(this.options.domSelectors.emptyBlockTemplate);
};

// ---------------------------------------------------------------------------
// EVENT LISTENERS
// ---------------------------------------------------------------------------
Module.prototype._addEventListeners = function() {
	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.addButton,
		this._addBlock.bind(this)
	);

	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.removeButton,
		this._removeBlock.bind(this)
	);
};

// ---------------------------------------------------------------------------
// BlOCK
// ---------------------------------------------------------------------------
Module.prototype._addBlock = function(event, prefillValues, prefillLastBlock) {
	var $clickedAddButton = $(event.currentTarget),
		$emptyBlock = $clickedAddButton.closest(this.options.domSelectors.emptyBlock),
		$existingBlocks = this._getExistingBlocks(),
		compiledTemplate = template(this.$blockTemplate.html()),
		$newBlock,
		$newAddedAddButton;

	// if the "add" button on the empty block is clicked, hide the empty block
	if ($emptyBlock.length === 1 && this.options.allowRemovingLastBlock) {
		this._hideEmptyBlock();
	}

	$newBlock = $(compiledTemplate({
		'id': this.blockIndexCounter
	}));

	// hide clicked "add" button
	$clickedAddButton.removeClass(this.options.stateClasses.isVisible);

	this._handleRemoveButtonDisabling($existingBlocks, $newBlock, $clickedAddButton);

	$newAddedAddButton = $newBlock.find(this.options.domSelectors.addButton);

	if (this.options.maxBlocks) {
		this._checkForMaxBlocks($existingBlocks, $newAddedAddButton);
	}

	$newBlock.insertBefore(this.$blockTemplate);

	if (this.options.copyFieldContent) {
		this._copyFieldContent($existingBlocks, $newBlock);
	}

	if (this.options.prefillBlocks && prefillValues) {
		// remove new added add button as long the last block is not added
		if (!prefillLastBlock) {
			$newAddedAddButton.removeClass(this.options.stateClasses.isVisible);
		}

		this._fillValues($newBlock, prefillValues);
	}

	$newBlock.addClass(this.options.stateClasses.isVisible);

	// block added on click
	if (event) {
		$newBlock.find('input, select, textarea').first().focus();

		if (mediaqueries.query({ to: 'tabletPortrait' })) {
			this._scrollToTopOfBlock($newBlock);
		}
	}

	window.requestAnimationFrame(function() {
		$document.triggerHandler('ajaxload');
		$document.triggerHandler('formchange', [{id: this.formId}]);
		$document.trigger(events.blockPrefilled);
	}.bind(this));

	this.blockIndexCounter++;
};

Module.prototype._handleRemoveButtonDisabling = function($existingBlocks, $newBlock, $clickedAddButton) {
	var	$newAddedRemoveButton = $newBlock.find(this.options.domSelectors.removeButton),
		$removeButton = $clickedAddButton.siblings(this.options.domSelectors.removeButton);

	if (!this.options.allowRemovingLastBlock && !this.options.prefillBlocks) {
		if ($existingBlocks.length === 0) {
			$newAddedRemoveButton.prop('disabled', true);
		}

		$removeButton.prop('disabled', false);
	}
};

Module.prototype._removeBlock = function(event) {
	var $blockToRemove = $(event.currentTarget).closest(this.options.domSelectors.block),
		$existingBlocks;

	$blockToRemove.remove();
	$existingBlocks = this._getExistingBlocks();

	// if the previous block is removable, show the empty block
	if ($existingBlocks.length === 0 && this.options.allowRemovingLastBlock) {
		this._showEmptyBlock();
		return;
	}

	this._showAddButton($existingBlocks);
};

// ---------------------------------------------------------------------------
// EMPTY BlOCK
// ---------------------------------------------------------------------------
Module.prototype._addEmptyBlock = function() {
	var $emptyBlock = $(this.$emptyBlockTemplate.html());

	$emptyBlock.insertBefore(this.$emptyBlockTemplate);
};

Module.prototype._showEmptyBlock = function() {
	var $emptyBlock = this.$element.find(this.options.domSelectors.emptyBlock),
		$addButton = $emptyBlock.find(this.options.domSelectors.addButton);

	$emptyBlock.addClass(this.options.stateClasses.isVisible);
	$addButton.focus();
};

Module.prototype._hideEmptyBlock = function() {
	var $emptyBlock = this.$element.find(this.options.domSelectors.emptyBlock);

	$emptyBlock.removeClass(this.options.stateClasses.isVisible);
};

// ---------------------------------------------------------------------------
// ADD BUTTON
// ---------------------------------------------------------------------------
Module.prototype._showAddButton = function($existingBlocks) {
	var existingBlocksLength = $existingBlocks.length,
		$prevBlock = this._getPrevBlock($existingBlocks),
		$addButtonToShow = $prevBlock.find(this.options.domSelectors.addButton),
		$removeButtonToDisable = $prevBlock.find(this.options.domSelectors.removeButton);

	// if it's not allowed to remove the previous block, add the disable attribute to the button
	if (existingBlocksLength === 1 && !this.options.allowRemovingLastBlock) {
		$removeButtonToDisable.prop('disabled', true);
	}

	// show "add" button at the previous block
	$addButtonToShow.addClass(this.options.stateClasses.isVisible);
	$addButtonToShow.focus();
};

// ---------------------------------------------------------------------------
// MAXIMUM BlOCKS OPTION
// ---------------------------------------------------------------------------
Module.prototype._checkForMaxBlocks = function($existingBlocks, $newAddedAddButton) {
	// not allow to add another block if the maximum number of blocks is reached
	if ($existingBlocks.length === this.options.maxBlocks) {
		$newAddedAddButton.removeClass(this.options.stateClasses.isVisible);
	}
};

// ---------------------------------------------------------------------------
// COPY FIELD CONTENT OPTION
// ---------------------------------------------------------------------------
Module.prototype._copyFieldContent = function($existingBlocks, $newBlock) {
	var existingBlocksLength = $existingBlocks.length,
		$prevBlock,
		inputValues;

	if (existingBlocksLength >= 1) {
		$prevBlock = this._getPrevBlock($existingBlocks);

		inputValues = $prevBlock.serializeArray();

		this._fillValues($newBlock, inputValues);
	}
};

// ---------------------------------------------------------------------------
// PREFILL
// ---------------------------------------------------------------------------
Module.prototype._prefillBlock = function() {
	var prefillLastBlock = false;

	this.options.prefillBlocks.forEach(function(prefillValues, index, array) {
		if (index === array.length - 1) {
			prefillLastBlock = true;
		}

		this._addBlock(false, prefillValues, prefillLastBlock);
	}.bind(this));
};

Module.prototype._fillValues = function($newBlock, inputValues) {

	// fill the input values into the inputs of the new added block
	inputValues.forEach(function(item) {
		var splittedNameAttr = item.name.substring(0, item.name.lastIndexOf('_')),
			$input = $newBlock.find('[name^="' + splittedNameAttr + '"]'),
			typeOfElement,
			radioCheckboxValue;

		if ($input.length === 0) {
			return;
		}

		typeOfElement = $input[0].type;

		if (typeOfElement === 'radio' || typeOfElement === 'checkbox') {
			$input.each(function(_, radioCheckbox) {
				radioCheckboxValue = $(radioCheckbox).val() === item.value ? true : false;
				$(radioCheckbox).prop('checked', radioCheckboxValue);
			});
		} else {
			$input.val(item.value);
		}
	}.bind(this));
};

// ---------------------------------------------------------------------------
// HELPER METHODS
// ---------------------------------------------------------------------------
Module.prototype._getExistingBlocks = function() {
	return this.$element.find(this.options.domSelectors.block);
};

Module.prototype._getPrevBlock = function($existingBlocks) {
	return $($existingBlocks[$existingBlocks.length - 1]);
};

Module.prototype._scrollToTopOfBlock = function($newBlock) {
	var breakpoint = mediaqueries.getCurrentBreakpointName(),
		navHeight = breakpoint === 'mobilePortrait' || 'mobileLandscape' ? 48 : 64;

	// scroll top of new added block on mobile
	$('html, body').animate({
		scrollTop: $newBlock.offset().top - navHeight
	}, 200);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 726 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 727 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(728);


/***/ }),
/* 728 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M40 - Pressrelease Excerpt
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



// Ensure that the SASS file is built.
__webpack_require__(729);


/***/ }),
/* 729 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 730 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(731);


/***/ }),
/* 731 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 732 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(733);


/***/ }),
/* 733 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 734 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(735);


/***/ }),
/* 735 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 736 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(737);


/***/ }),
/* 737 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 738 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(739);


/***/ }),
/* 739 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M53 - Product Tree
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(740);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'producttree',
	events = {},
	defaults = {
		domSelectors: {
			item: '[data-' + name + '="item"]',
			list: '[data-' + name + '="list"]',
			products: '[data-' + name + '="products"]',
			buttonContainer: '[data-' + name + '="button_container"]',
			button: '[data-' + name + '="button"]'
		},
		stateClasses: {
			isHidden: 'is_hidden',
			varDistributed: 'var_distributed'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {

	this.$lists = this.$element.find(this.options.domSelectors.list);
	this.$products = this.$element.find(this.options.domSelectors.products);

	this._setInitState();
	this.$element.on('click.' + this.uuid, this.options.domSelectors.button, this._showMoreTickets.bind(this));
};

Module.prototype._setInitState = function() {
	this.$lists.each(this._setProductListInitState.bind(this));
};

Module.prototype._setProductListInitState = function(index, list) {
	var $list = $(list),
		$item = $list.find(this.options.domSelectors.item),
		$buttonContainer = $list.find(this.options.domSelectors.buttonContainer),
		maxNumberOfProducts = $list.attr('data-' + name + '-maxproducts') || 10000000,
		numberOfItems = $item.length;

	// slicing list to max number
	if (numberOfItems > maxNumberOfProducts) {
		$buttonContainer.removeClass(this.options.stateClasses.isHidden);
		$item.slice(maxNumberOfProducts).addClass(this.options.stateClasses.isHidden).attr('aria-hidden', true);
		return;
	} else {
		$buttonContainer.remove();
	}

};

Module.prototype._showMoreTickets = function(event) {
	var $button = $(event.currentTarget),
		$list = $button.closest(this.options.domSelectors.products),
		$buttonContainer = $list.find(this.options.domSelectors.buttonContainer),
		$items = $list.find(this.options.domSelectors.item);

	$items.removeClass(this.options.stateClasses.isHidden).attr('aria-hidden', false);
	$buttonContainer.addClass(this.options.stateClasses.isHidden).attr('aria-hidden', true);

	$($items[this.options.maxNumberOfProducts]).find('a').focus();

};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 740 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 741 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(742);


/***/ }),
/* 742 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M89 - Job filter
 */



// Ensure that the SASS file is built.
__webpack_require__(743);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	ajaxRequest = __webpack_require__(16),
	template = __webpack_require__(51),
	$document = $(document),
	$window = $(window),
	mediaqueries = __webpack_require__(6),
	parseUrl = __webpack_require__(69),
	createUrlParams = __webpack_require__(79);

var name = 'jobfilter',
	events = {
		outgoing: {
			triggerSelectionOnItem: 'multiselect.triggerSelectionOnItem',
			clearSelection: 'multiselect.clearSelection',
			setTotalNumerOfItems: 'pagination.setTotalNumerOfItems',
			setStartItem: 'pagination.setStartItem',
			setItemsPerPage: 'pagination.setItemsPerPage'
		},
		incoming: {
			paginationDidInitialize: 'pagination.didInitialize',
			didSwitchToStartItem: 'pagination.didSwitchToStartItem'
		}
	},
	defaults = {
		domSelectors: {
			form: '[data-' + name + '="filter_form"]',
			filter: {
				topic: '[data-' + name + '="topic"]',
				region: '[data-' + name + '="region"]',
				pensum: '[data-' + name + '="pensum"]',
				experience: '[data-' + name + '="experience"]',
				worksmart: '[data-' + name + '="worksmart"]',
				reentry: '[data-' + name + '="reentry"]',
				jobsharing: '[data-' + name + '="jobsharing"]',
				searchField: '[data-' + name + '="search_field"]',
				reset: '[data-' + name + '="reset"]',
				searchButton: '[data-' + name + '="search_button"]'
			},
			subscribe: {
				checkbox: '[data-radiocheckbox="checkbox_item"]',
				group: '[data-' + name + '="subscribe_group"]',
				confirm: '[data-' + name + '="subscribe_confirm"]',
				title: '[data-' + name + '="subscribe_title"]',
				mail: '[data-' + name + '="subscribe_mail"]',
				button: '[data-' + name + '="subscribe_button"]'
			},
			infoContainer: {
				container: '[data-' + name + '="info_container"]',
				resultTitle: '[data-' + name + '="result_title"]',
				resultCount: '[data-' + name + '="result_count"]',
				resultCountAppendix: '[data-' + name + '="result_count_appendix"]',
				tagButton: '[data-' + name + '="tag_button"]',
				tagList: '[data-' + name + '="tag_list"]'
			},
			results: {
				container: '[data-' + name + '="results_container"]',
				resultItem: '[data-' + name + '="result_item"]',
				resultLink: '[data-' + name + '="result_link"]',
				notification: '[data-' + name + '="notification"]',
				errorNotification: '[data-' + name + '="error_notification"]'
			},
			templates: {
				tags: '[data-' + name + '="tag_template"]',
				results: '[data-' + name + '="result_template"]'
			},
			pagination: '[data-init="pagination"]'
		},
		stateClasses: {
			isHidden: 'is_hidden'
		},
		dataAttributes: {
			api: 'data-jobfilter-api',
			selection: 'data-multiselect-values',
			singleCheckboxLabel: 'data-jobfilter-label',
			subscribe: 'data-jobfilter-subscribe',
			tagButtonName: 'data-jobfilter-tagname'
		},
		apiResponseKeys: {
			attributes: 'attributes',
			topic: 20,
			pensum: 50,
			pensumPercent: 160,
			experience: 60,
			experiences: 140,
			jobLocation: 100,
			reentry: 150,
			region: 110,
			worksmart: 130,
			jobsharing: 135,
			jobWording: {
				title: 'title',
				text: 'text'
			},
			links: 'links',
			directLink: 'directlink',
			startDate: 'start_date'
		},
		dateFormat: {
			day: '2-digit',
			month: '2-digit',
			year: 'numeric'
		},
		label: {
			worksmart: 'Work Smart'
		},
		resultsPerPage: 10
	},
	data = {};

// Create an instance of the module
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// -------------------------------------------------------------
// INIT
// -------------------------------------------------------------

Module.prototype.init = function() {
	var hasUrlParams = Object.keys(this.getUrlParamters()).length !== 0;

	this.cacheElements();
	this.cacheTemplates();
	this.bindEventListeners();
	this.historyInit();
	this.fetchJobs()
		.then(function(json) {
			this.jobs = this.transformJobResults(json);
			this.paginationInit(function() {
				if (hasUrlParams) {
					this.restoreFilterValuesForUrlParams();
				}

				if (!this.options.noFilter) {
					this.checkIfKeywordIsTypedIn(true, !hasUrlParams, false);
				} else {
					this.performSearch(true, !hasUrlParams, false);
				}

				this.toggleReset();
			}.bind(this));
		}.bind(this))
		.catch(function() {
			this.handleFetchError();
		}.bind(this));
};

// -------------------------------------------------------------
// CACHING
// -------------------------------------------------------------

// elements
Module.prototype.cacheElements = function() {

	this.elements = {
		filter: {},
		label: {},
		subscribe: {},
		info: {},
		results: {}
	};

	// form element
	this.elements.$form = this.$element.find(this.options.domSelectors.form);

	// subscribe elements
	this.elements.subscribe.$checkbox = this.$element.find(this.options.domSelectors.subscribe.checkbox);
	this.elements.subscribe.$group = this.$element.find(this.options.domSelectors.subscribe.group);
	this.elements.subscribe.$confirm = this.$element.find(this.options.domSelectors.subscribe.confirm);
	this.elements.subscribe.$title = this.$element.find(this.options.domSelectors.subscribe.title);
	this.elements.subscribe.$mail = this.$element.find(this.options.domSelectors.subscribe.mail);
	this.elements.subscribe.$button = this.$element.find(this.options.domSelectors.subscribe.button);

	// info container elements
	this.elements.info.$container = this.$element.find(this.options.domSelectors.infoContainer.container);
	this.elements.info.$resultTitle = this.$element.find(this.options.domSelectors.infoContainer.resultTitle);
	this.elements.info.$resultCount = this.$element.find(this.options.domSelectors.infoContainer.resultCount);
	this.elements.info.$resultCountAppendix = this.$element.find(this.options.domSelectors.infoContainer.resultCountAppendix);
	this.elements.info.$tagList = this.$element.find(this.options.domSelectors.infoContainer.tagList);

	// results elements
	this.elements.results.$container = this.$element.find(this.options.domSelectors.results.container);
	this.elements.results.$notification = this.$element.find(this.options.domSelectors.results.notification);
	this.elements.results.$errorNotification = this.$element.find(this.options.domSelectors.results.errorNotification);

	// pagination
	this.elements.$pagination = this.$element.find(this.options.domSelectors.pagination);

	if (this.options.noFilter) {
		return;
	}

	// filter elements
	this.elements.filter.$topic = this.$element.find(this.options.domSelectors.filter.topic);
	this.elements.filter.$region = this.$element.find(this.options.domSelectors.filter.region);
	this.elements.filter.$pensum = this.$element.find(this.options.domSelectors.filter.pensum);
	this.elements.filter.$experience = this.$element.find(this.options.domSelectors.filter.experience);
	this.elements.filter.$worksmart = this.$element.find(this.options.domSelectors.filter.worksmart);
	this.elements.filter.$reentry = this.$element.find(this.options.domSelectors.filter.reentry);
	this.elements.filter.$jobsharing = this.$element.find(this.options.domSelectors.filter.jobsharing);
	this.elements.filter.$searchField = this.$element.find(this.options.domSelectors.filter.searchField);
	this.elements.filter.$reset = this.$element.find(this.options.domSelectors.filter.reset);
	this.elements.filter.$searchButton = this.$element.find(this.options.domSelectors.filter.searchButton);
};

// templates
Module.prototype.cacheTemplates = function() {
	this.resultTemplate = template(
		this.$element.find(this.options.domSelectors.templates.results).html()
	);

	if (this.options.noFilter) {
		return;
	}

	this.tagTemplate = template(
		this.$element.find(this.options.domSelectors.templates.tags).html()
	);

};

// -------------------------------------------------------------
// EVENTS
// -------------------------------------------------------------

Module.prototype.bindEventListeners = function() {

	// subscribe button
	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.subscribe.button,
		this.handleSubscribe.bind(this)
	);

	// pagination page switch
	this.elements.$pagination.on(
		events.incoming.didSwitchToStartItem,
		this.handlePaginationChange.bind(this)
	);

	// listen to history pop state
	window.addEventListener('popstate', this.historyPop.bind(this));

	if (this.options.noFilter) {
		return;
	}

	// search button
	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.filter.searchButton,
		function(evt) {
			if (evt && !this.options.noFilter) {
				evt.preventDefault();
				this.elements.filter.$searchButton.blur();
			}

			this.checkIfKeywordIsTypedIn(false, true, true);

			if (!this.firstResultIsVisible()) {
				this.scrollToTopOfResultList();
			}
		}.bind(this)
	);

	// reset button
	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.filter.reset,
		this.handleReset.bind(this)
	);

	// tag buttons
	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.infoContainer.tagButton,
		this.handleTagButtonClick.bind(this)
	);

	// listen to input changes on filter
	this.$element.on(
		'change.' + this.uuid,
		[
			this.options.domSelectors.filter.topic,
			this.options.domSelectors.filter.region,
			this.options.domSelectors.filter.pensum,
			this.options.domSelectors.filter.experience,
			this.options.domSelectors.filter.worksmart,
			this.options.domSelectors.filter.reentry,
			this.options.domSelectors.filter.jobsharing,
			this.options.domSelectors.filter.searchField
		],
		this.toggleReset.bind(this)
	);

};

// -------------------------------------------------------------
// SEARCH
// -------------------------------------------------------------

// if a keyword is typed in then send the term to the BE to use the elastic search of the prospective api
Module.prototype.checkIfKeywordIsTypedIn = function(skipHistoryUpdate, resetToFirstPage, focusFirstResult) {
	var keyword = {};

	keyword.value = this.elements.filter.$searchField.val().toLowerCase();

	if (keyword.value) {
		this.fetchJobs(keyword)
			.then(function(json) {
				this.jobsFilteredByKeyword = this.transformJobResults(json);
				this.performSearch(skipHistoryUpdate, resetToFirstPage, focusFirstResult);
			}.bind(this))
			.catch(function() {
				this.handleFetchError();
			}.bind(this));
	} else {
		this.performSearch(skipHistoryUpdate, resetToFirstPage, focusFirstResult);
	}
};

Module.prototype.performSearch = function(skipHistoryUpdate, resetToFirstPage, focusFirstResult) {
	var currentFilterValues;

	// save current page
	if (!this.startItem || resetToFirstPage) {
		this.startItem = 1;
	}

	if (this.options.noFilter) {
		this.filteredResults = this.filterResults();
	} else {
		currentFilterValues = this.getFilterValues();
		this.filteredResults = this.filterResults(currentFilterValues);
	}

	// update pagination
	this.paginationUpdate(this.filteredResults.length, resetToFirstPage);

	// update filter tags
	if (!this.options.noFilter) {
		this.updateFilterTags(currentFilterValues);
	}

	// update results count
	this.updateResultsCount(this.filteredResults.length);

	// update result list
	this.updateResultList(this.startItem - 1);

	// focus first item
	if (focusFirstResult) {
		this.focusFirstResult();
	}

	// add "no results" notification
	if (this.filteredResults.length === 0) {
		this.handleNoResults();
	}

	// update history
	if (!skipHistoryUpdate) {
		this.historyPush();
	}

	this.resetSubscriptionSection();

};

Module.prototype.firstResultIsVisible = function() {
	var $firstResult = $(this.elements.results.$container.find(this.options.domSelectors.results.resultItem)[0]),
		firstResultOffset,
		windowHeight;

	if ($firstResult.length < 1) {
		return true;
	}

	firstResultOffset = $firstResult.offset().top + $firstResult[0].clientHeight;
	windowHeight = $window.height();

	return firstResultOffset < windowHeight;

};

// -------------------------------------------------------------
// FILTER VALUES
// -------------------------------------------------------------

// Get values from all input fields
Module.prototype.getFilterValues = function() {
	var filterValues = {};

	if (!this.options.noFilter) {
		filterValues.topic = JSON.parse(this.elements.filter.$topic.attr(this.options.dataAttributes.selection));
		filterValues.region = JSON.parse(this.elements.filter.$region.attr(this.options.dataAttributes.selection));
		filterValues.pensum = JSON.parse(this.elements.filter.$pensum.attr(this.options.dataAttributes.selection));
		filterValues.experience = JSON.parse(this.elements.filter.$experience.attr(this.options.dataAttributes.selection));
		filterValues.worksmart = this.elements.filter.$worksmart.prop('checked');
		filterValues.reentry = this.elements.filter.$reentry.prop('checked');
		filterValues.jobsharing = this.elements.filter.$jobsharing.prop('checked');
		filterValues.title = this.elements.filter.$searchField.val().toLowerCase();
	}

	return filterValues;
};

Module.prototype.revertCommaInValue = function(value) {
	// before we replaced the comma with a dollar sign to prevent splitting within the value
	// now we have to revert it back to the comma
	return value.replace('$', ', ');
};

// Set filter values based on passed object
Module.prototype.setFilterValues = function(filterValues) {

	if (this.options.noFilter) {
		return;
	}

	// set topic
	if (filterValues.topic) {
		this.elements.filter.$topic.trigger(events.outgoing.clearSelection);
		filterValues.topic.forEach(function(topic) {
			this.elements.filter.$topic.trigger(events.outgoing.triggerSelectionOnItem, this.revertCommaInValue(topic));
		}.bind(this));
	}

	// set region
	if (filterValues.region) {
		this.elements.filter.$region.trigger(events.outgoing.clearSelection);
		filterValues.region.forEach(function(region) {
			this.elements.filter.$region.trigger(events.outgoing.triggerSelectionOnItem, this.revertCommaInValue(region));
		}.bind(this));
	}

	// set pensum
	if (filterValues.pensum) {
		this.elements.filter.$pensum.trigger(events.outgoing.clearSelection);
		filterValues.pensum.forEach(function(pensum) {
			this.elements.filter.$pensum.trigger(events.outgoing.triggerSelectionOnItem, this.revertCommaInValue(pensum));
		}.bind(this));
	}

	// set experience
	if (filterValues.experience) {
		this.elements.filter.$experience.trigger(events.outgoing.clearSelection);
		filterValues.experience.forEach(function(experience) {
			this.elements.filter.$experience.trigger(events.outgoing.triggerSelectionOnItem, this.revertCommaInValue(experience));
		}.bind(this));
	}

	// set worksmart
	this.elements.filter.$worksmart.prop('checked', filterValues.worksmart);

	// set reentry
	this.elements.filter.$reentry.prop('checked', filterValues.reentry);

	// set jobsharing
	this.elements.filter.$jobsharing.prop('checked', filterValues.jobsharing);

	// set search fields
	this.elements.filter.$searchField.val(filterValues.title);
};

// restore filter values based on url params
Module.prototype.restoreFilterValuesForUrlParams = function() {
	var urlParams = this.getUrlParamters(),
		filterValues = {},
		value;

	Object.keys(urlParams).forEach(function(key) {
		value = decodeURIComponent(urlParams[key]);

		if (key === 'startItem' || key === 'title') {
			filterValues[key] = value;
		} else {
			// replace comma with dollar sign to prevent splitting within the value
			filterValues[key] = value.replace(', ', '$').split(',');
		}
	});

	if (filterValues.startItem) {
		this.startItem = parseInt(filterValues.startItem);
	}

	this.setFilterValues(filterValues);
};

// -------------------------------------------------------------
// FILTERING
// -------------------------------------------------------------

// Filter all jobs based on filter criteria
Module.prototype.filterResults = function(filterValues) {
	var jobs,
		matchTopic,
		matchRegion,
		matchExperience,
		matchExperiences,
		matchWorksmart,
		matchReentry,
		matchJobsharing,
		matchPensum,
		apiKeys = this.options.apiResponseKeys,
		jobTopics,
		jobItems = this.jobs;

	if (filterValues) {

		if (filterValues.title) {
			jobItems = this.jobsFilteredByKeyword;
		}

		// iterate job results
		jobs = jobItems.filter(function(item) {

			jobTopics = item[this.options.apiResponseKeys.attributes][this.options.apiResponseKeys.topic];
			matchTopic = this.findMatchInJobItem(filterValues.topic, item, apiKeys.topic);
			matchRegion = this.findMatchInJobItem(filterValues.region, item, apiKeys.region);
			matchExperience = this.findMatchInJobItem(filterValues.experience, item, apiKeys.experience);
			matchExperiences = this.findMatchInJobItem(filterValues.experience, item, apiKeys.experiences);
			matchPensum = this.findMatchInJobItem(filterValues.pensum, item, apiKeys.pensum);
			matchWorksmart = this.findMatchInJobItem(filterValues.worksmart, item, apiKeys.worksmart);
			matchReentry = this.findMatchInJobItem(filterValues.reentry, item, apiKeys.reentry);
			matchJobsharing = this.findMatchInJobItem(filterValues.jobsharing, item, apiKeys.jobsharing);

			return jobTopics && matchTopic && matchRegion && (matchExperience || matchExperiences) && matchPensum && matchWorksmart && matchReentry && matchJobsharing;

		}.bind(this));

	} else {
		jobs = this.jobs;
	}

	return jobs;
};

// Decide if single job item fullfills filter criteria
Module.prototype.findMatchInJobItem = function(filterValue, jobItem, apiKey, isForTitle) {
	var matchFound = false,
		attributesKey = this.options.apiResponseKeys.attributes,
		jobItemValues,
		jobItemValue;

	// Do not filter if the value is undefined
	if (filterValue === undefined) {
		return true;
	}

	// Need the second condition to check if the boolean (worksmart, reentry & jobsharing) is false
	if (filterValue.length === 0 || !filterValue) {
		return true;
	}

	if (filterValue.length || filterValue) {
		if (isForTitle) {
			jobItemValue = jobItem[apiKey.title] + ' ' + jobItem[apiKey.text];

			if (jobItemValue.toLowerCase().indexOf(filterValue) > -1) {
				matchFound = true;
			}
		} else {
			jobItemValues = jobItem[attributesKey][apiKey];

			if (jobItemValues) {
				jobItemValues.forEach(function(jobItemValue) {

					// we need an exception handling for attributes 130, 135, 140 and 150 here
					if (apiKey === this.options.apiResponseKeys.experiences) {
						matchFound = this.findMatchForCommaSeparatedStringInArray(jobItemValue, filterValue);
					} else if (apiKey === this.options.apiResponseKeys.worksmart || apiKey === this.options.apiResponseKeys.reentry || apiKey === this.options.apiResponseKeys.jobsharing) {
						if (filterValue.toString() === jobItemValue) {
							matchFound = true;
						}
					} else {
						if (filterValue.indexOf(jobItemValue) >= 0) {
							matchFound = true;
						}
					}
				}.bind(this));
			}
		}
	}

	return matchFound;
};

Module.prototype.findMatchForCommaSeparatedStringInArray = function(commaSeparatedString, filterValue) {
	var matchFound = false,
		values = commaSeparatedString.split(',');

	values.forEach(function(value) {
		if (filterValue.indexOf(value) !== -1) {
			matchFound = true;
		}
	});

	return matchFound;
};

// -------------------------------------------------------------
// RESET BUTTON
// -------------------------------------------------------------
Module.prototype.handleReset = function() {

	if (!this.options.noFilter) {

		// clear input
		this.elements.filter.$searchField.val('');

		// clear tag buttons
		this.elements.info.$tagList.empty();

		// clear multiselects
		this.elements.filter.$topic.trigger(events.outgoing.clearSelection);
		this.elements.filter.$region.trigger(events.outgoing.clearSelection);
		this.elements.filter.$pensum.trigger(events.outgoing.clearSelection);
		this.elements.filter.$experience.trigger(events.outgoing.clearSelection);
		this.elements.filter.$worksmart.prop('checked', false);
		this.elements.filter.$reentry.prop('checked', false);
		this.elements.filter.$jobsharing.prop('checked', false);
	}

	// reset results and pagination
	this.performSearch(false, true, true);

	// hide reset button
	this.toggleReset();

};

Module.prototype.toggleReset = function() {
	var currentFilterValues,
		inputs;

	if (this.options.noFilter) {
		this.resetSubscriptionSection();
		return;
	}

	currentFilterValues = this.getFilterValues();

	// change the true booleon of the attributes 130, 135 and 150 to a string and set the false boolean to an empty string
	// this is necessary to reduce the array afterwards
	Object.keys(currentFilterValues).forEach(function(key) {
		if (key === 'worksmart' || key === 'reentry' || key === 'jobsharing') {
			if (currentFilterValues[key]) {
				currentFilterValues[key] = currentFilterValues[key].toString();
			} else {
				currentFilterValues[key] = '';
			}
		}
	});

	inputs = Object.keys(currentFilterValues).reduce(function(accumulator, key) {
		return accumulator + currentFilterValues[key].length;
	}, 0);

	if (inputs < 1) {
		this.hideReset();
	} else {
		this.showReset();
	}

};

Module.prototype.showReset = function() {
	this.elements.filter.$reset
		.removeClass(this.options.stateClasses.isHidden)
		.removeAttr('aria-hidden role');
};

Module.prototype.hideReset = function() {
	this.elements.filter.$reset
		.addClass(this.options.stateClasses.isHidden)
		.attr({
			'aria-hidden': true,
			'role': 'presentation'
		});
};

// -------------------------------------------------------------
// NETWORK
// -------------------------------------------------------------

Module.prototype.fetchJobs = function(searchKeyword) {
	var apiUrl = this.$element.attr(this.options.dataAttributes.api);

	return ajaxRequest(apiUrl, false, searchKeyword);
};

Module.prototype.transformJobResults = function(jobs) {
	var pensumPercentage,
		location,
		worksmart;

	jobs.forEach(function(job) {
		pensumPercentage = job[this.options.apiResponseKeys.attributes][this.options.apiResponseKeys.pensumPercent];
		location = job[this.options.apiResponseKeys.attributes][this.options.apiResponseKeys.jobLocation];
		worksmart = job[this.options.apiResponseKeys.attributes][this.options.apiResponseKeys.worksmart];

		// Add pensum percentage to title
		if (pensumPercentage && pensumPercentage.length > 0) {
			job[this.options.apiResponseKeys.jobWording.title] += pensumPercentage[0].length > 0 ? ', ' + pensumPercentage[0] : '';
		}

		// Add location to title
		if (location && location.length > 0) {
			job[this.options.apiResponseKeys.jobWording.title] += location[0].length > 0 ? ', ' + location[0] : '';
		}

		// Add worksmart to title
		if (worksmart && worksmart.length > 0) {
			job[this.options.apiResponseKeys.jobWording.title] += worksmart[0] === 'true' ? ' / ' + this.options.label.worksmart : '';
		}

	}.bind(this));

	return jobs;
};

// -------------------------------------------------------------
// RESULT LIST
// -------------------------------------------------------------

Module.prototype.updateResultList = function(startItem) {

	var resultsToDisplay = JSON.parse(JSON.stringify(this.filteredResults)),
		resultObject = {},
		date,
		formattedDate,
		resultHtml,
		resultTopics;

	resultsToDisplay = resultsToDisplay.splice(startItem, this.options.resultsPerPage);

	this.elements.results.$container.empty();

	resultsToDisplay.forEach(function(result) {

		resultTopics = result[this.options.apiResponseKeys.attributes][this.options.apiResponseKeys.topic];

		resultObject.link = result[this.options.apiResponseKeys.links][this.options.apiResponseKeys.directLink];
		resultObject.title = result[this.options.apiResponseKeys.jobWording.title];
		resultObject.topics = resultTopics.join(', ');

		// jscs:disable
		date = new Date(result[this.options.apiResponseKeys.startDate]);
		// jscs:enable

		formattedDate = date.toLocaleString('de-DE', this.options.dateFormat);

		resultObject.date = formattedDate;

		resultHtml = this.resultTemplate({ result: resultObject});
		this.elements.results.$container.append(resultHtml);
	}.bind(this));
};

Module.prototype.focusFirstResult = function() {
	var $firstResult = $(this.elements.results.$container.find(this.options.domSelectors.results.resultLink)[0]);

	if ($firstResult.length < 1) {
		return;
	}

	$firstResult.focus();
};

Module.prototype.scrollToTopOfResultList = function() {
	var breakpoint = mediaqueries.getCurrentBreakpointName(),
		navHeight = breakpoint === 'mobilePortrait' || 'mobileLandscape' ? 48 : 64;

	$('html, body').animate({
		scrollTop: this.elements.info.$container.offset().top - navHeight
	}, 200);
};

// -------------------------------------------------------------
// RESULT COUNT
// -------------------------------------------------------------

Module.prototype.updateResultsCount = function(count) {
	var tags;

	if (!this.options.noFilter) {
		tags = this.elements.info.$tagList.find(this.options.domSelectors.infoContainer.tagButton);

		if (tags.length < 1) {
			this.elements.info.$resultCountAppendix.addClass(this.options.stateClasses.isHidden);
		} else {
			this.elements.info.$resultCountAppendix.removeClass(this.options.stateClasses.isHidden);
		}
	}

	this.elements.info.$resultCount.html(count);
	this.elements.info.$resultTitle.removeClass(this.options.stateClasses.isHidden);
	this.elements.results.$notification.addClass(this.options.stateClasses.isHidden);
	this.elements.results.$errorNotification.addClass(this.options.stateClasses.isHidden);
};

// -------------------------------------------------------------
// FILTER TAGS
// -------------------------------------------------------------

Module.prototype.updateFilterTags = function(filterValues) {
	var filterValue,
		tagHtml,
		data;

	this.elements.info.$tagList.empty();

	Object.keys(filterValues).forEach(function(key) {
		filterValue = filterValues[key];

		if (!filterValue) {
			return;
		}

		if (key === 'title') {
			return;
		}

		// we need an exception handling for attributes 130, 135 and 150 here
		if (key === 'worksmart' || key === 'reentry' || key === 'jobsharing') {
			if (key === 'worksmart') {
				filterValue = this.elements.filter.$worksmart.attr(this.options.dataAttributes.singleCheckboxLabel);
			} else if (key === 'reentry') {
				filterValue = this.elements.filter.$reentry.attr(this.options.dataAttributes.singleCheckboxLabel);
			} else {
				filterValue = this.elements.filter.$jobsharing.attr(this.options.dataAttributes.singleCheckboxLabel);
			}

			data = {
				tagName: filterValue, tagKey: key
			};

			tagHtml = this.tagTemplate({ data: data });
			this.elements.info.$tagList.append(tagHtml);

			return;
		}

		filterValue.forEach(function(item) {
			data = {
				tagName: item, tagKey: key
			};

			tagHtml = this.tagTemplate({ data: data });
			this.elements.info.$tagList.append(tagHtml);
		}.bind(this));
	}.bind(this));
};

Module.prototype.handleTagButtonClick = function(evt) {
	var tagName = $(evt.currentTarget).attr(this.options.dataAttributes.tagButtonName),
		tagObject = JSON.parse(tagName),
		$clickedElement;

	// remove selection from corresponding element
	$clickedElement = this.$element.find('[data-' + name + '="' + tagObject.tagKey + '"]');

	if (tagObject.tagKey === 'worksmart' || tagObject.tagKey === 'reentry' || tagObject.tagKey === 'jobsharing') {
		$clickedElement.prop('checked', false);
	} else {
		$clickedElement.trigger(events.outgoing.triggerSelectionOnItem, tagObject.tagName);
	}

	// update tags
	this.performSearch(false, true, false);

	// hide reset button
	this.toggleReset();
};

// -------------------------------------------------------------
// NO RESULTS
// -------------------------------------------------------------

Module.prototype.handleNoResults = function() {
	this.elements.info.$resultTitle.addClass(this.options.stateClasses.isHidden);
	this.elements.results.$notification.removeClass(this.options.stateClasses.isHidden);
};

// -------------------------------------------------------------
// FETCH ERROR
// -------------------------------------------------------------

Module.prototype.handleFetchError = function() {
	this.elements.info.$resultTitle.addClass(this.options.stateClasses.isHidden);
	this.elements.results.$errorNotification.removeClass(this.options.stateClasses.isHidden);
};

// -------------------------------------------------------------
// SUBSCRIBE
// -------------------------------------------------------------

Module.prototype.handleSubscribe = function(evt) {
	var validForm = this.elements.$form.parsley().validate(),
		currentFilterValues,
		subject,
		email,
		apiURL,
		topic,
		region,
		pensum,
		experience,
		worksmart,
		reentry,
		jobsharing,
		searchField,
		urlParams;

	this.didSubscribe = true;

	evt.preventDefault();

	// only continue if no validation errors
	if (!validForm) {
		return;
	}

	currentFilterValues = this.getFilterValues();

	email = 'email=' + encodeURIComponent(this.elements.subscribe.$mail.val());
	apiURL = this.$element.attr(this.options.dataAttributes.subscribe);

	if (this.options.noFilter) {
		urlParams = email;

		if (this.options.newsletterName) {
			urlParams += '&name=' + encodeURIComponent(this.options.newsletterName);
		}

		// send subscription request to api
		ajaxRequest(apiURL, false, urlParams);
	} else {
		subject = 'name=' + encodeURIComponent(this.elements.subscribe.$title.val());
		topic = 'f=' + encodeURIComponent(this.options.apiResponseKeys.topic + ':' + currentFilterValues.topic);
		region = 'f=' + encodeURIComponent(this.options.apiResponseKeys.region + ':' + currentFilterValues.region);
		pensum = 'f=' + encodeURIComponent(this.options.apiResponseKeys.pensum + ':' + currentFilterValues.pensum);
		experience = 'f=' + encodeURIComponent(this.options.apiResponseKeys.experience + ':' + currentFilterValues.experience);
		worksmart = 'f=' + encodeURIComponent(this.options.apiResponseKeys.worksmart + ':' + currentFilterValues.worksmart);
		reentry = 'f=' + encodeURIComponent(this.options.apiResponseKeys.reentry + ':' + currentFilterValues.reentry);
		jobsharing = 'f=' + encodeURIComponent(this.options.apiResponseKeys.jobsharing + ':' + currentFilterValues.jobsharing);
		searchField = 'q=' + encodeURIComponent(currentFilterValues.title);

		// send subscription request to api
		ajaxRequest(apiURL, false,
			email + '&' + subject + '&' + topic + '&' + region + '&' + pensum + '&' + experience + '&' + worksmart + '&' + reentry + '&' + jobsharing + '&' + searchField
		);
	}

	// hide subscription form
	this.elements.subscribe.$group.addClass(this.options.stateClasses.isHidden);

	// show confirmation text
	this.elements.subscribe.$confirm.removeClass(this.options.stateClasses.isHidden);

	this.focusFirstResult();
};

// -------------------------------------------------------------
// PAGINATION
// -------------------------------------------------------------
Module.prototype.resetSubscriptionSection = function() {

	if (!this.didSubscribe) {
		return;
	}

	this.didSubscribe = false;

	// empty input fields
	this.elements.subscribe.$title.val('');
	this.elements.subscribe.$mail.val('');

	// uncheck checkbox
	this.elements.subscribe.$checkbox.prop('checked', false).trigger('change');

	// show subscription form
	this.elements.subscribe.$group.removeClass(this.options.stateClasses.isHidden);

	// hide confirmation text
	this.elements.subscribe.$confirm.addClass(this.options.stateClasses.isHidden);
};

// -------------------------------------------------------------
// PAGINATION
// -------------------------------------------------------------

Module.prototype.paginationInit = function(callback) {
	// if pagination is already initialized it has an according data
	// attribute. Otherwise, we need to listen to the init event
	if (this.elements.$pagination.attr('data-pagination-init')) {
		this.paginationSetup(callback);
	} else {
		$document.on(events.incoming.paginationDidInitialize, this.paginationSetup.bind(this, callback));
	}
};

Module.prototype.handlePaginationChange = function(_, item) {
	this.updateResultList(item - 1);
	this.scrollToTopOfResultList();
	this.startItem = item;
	this.historyPush();
	this.focusFirstResult();
};

Module.prototype.paginationSetup = function(callback) {
	this.elements.$pagination.trigger(events.outgoing.setItemsPerPage, this.options.resultsPerPage);
	this.paginationUpdate(this.jobs.length, true);
	callback();
};

Module.prototype.paginationUpdate = function(totalItems, resetToFirstPage) {
	var startItem = resetToFirstPage ? 1 : this.startItem;

	this.elements.$pagination.trigger(events.outgoing.setTotalNumerOfItems, totalItems);
	this.elements.$pagination.trigger(events.outgoing.setStartItem, startItem);
};

// -------------------------------------------------------------
// HISTORY
// -------------------------------------------------------------

Module.prototype.historyInit = function() {

	// prevent scrolling on popstate event
	if ('scrollRestoration' in history) {
		history.scrollRestoration = 'manual';
	}

	history.replaceState({}, null, null);
};

Module.prototype.historyPop = function(evt) {
	this.setFilterValues(evt.state);
	this.startItem = evt.state.startItem;
	this.performSearch(true, false, false);
	this.toggleReset();

	if (evt.state.startItem) {
		this.elements.$pagination.trigger(events.outgoing.setStartItem, evt.state.startItem);
	}
};

Module.prototype.historyPush = function() {
	var stateObject = this.getFilterValues(),
		urlParams;

	stateObject.startItem = this.startItem;

	urlParams = createUrlParams(stateObject);
	history.pushState(stateObject, null, urlParams);
};

// -------------------------------------------------------------
// URL PARAMS
// -------------------------------------------------------------

Module.prototype.getUrlParamters = function() {
	var locationSearch = window.location.search;

	return parseUrl(locationSearch);
};

// -------------------------------------------------------------
// MODULE
// -------------------------------------------------------------

SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 743 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 744 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(745);


/***/ }),
/* 745 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 746 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(747);


/***/ }),
/* 747 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 748 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(749);


/***/ }),
/* 749 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M91 - Circle Progress
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2020.
 */



__webpack_require__(750);

var $ = __webpack_require__(0),
	cssTransitionEnd = __webpack_require__(4),
	mediaqueries = __webpack_require__(6),
	SuperClass = __webpack_require__(1);

var name = 'circleprogress',
	smilHtmlClass = 'no-smil',
	events = {},
	defaults = {
		domSelectors: {
			circle: '[data-' + name + '="circle"]',
			svg: '[data-' + name + '="svg"]',
			text: '[data-' + name + '="text"]',
			hiddenText: '[data-' + name + '="hiddenText"]'
		},
		stateClasses: {
			isInitialized: 'is_initialized'
		}
	},
	paintUpdateInterval = 70,
	defaultTransitionDuration = 400,
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ---------------------------------------------------------------------------
// INIT
// ---------------------------------------------------------------------------
Module.prototype.init = function() {
	var progressPercentage = this.getProgressPercentage();

	// instance variables
	this.lastBreakpoint = mediaqueries.getCurrentBreakpointName();
	this.finishedAnimation = false;
	this.circumference = 0;
	this.animationIterations = 0;

	// instance constants
	this.config = {
		progressPercentage: progressPercentage,
		progressPercentageHumanReadable: parseInt(progressPercentage * 100, 10),
		totalAnimationDuration: this.getTotalAnimationDuration(),
		color: this.getStepColor()
	};

	this.cacheElements();
	this.setCircleProperties(function() {
		this.addEventListeners();
		this.setHiddenText();
		this.$svg.addClass(this.options.stateClasses.isInitialized);

		// if we have 0 progress, we don't need to animate anything
		if (this.config.progressPercentageHumanReadable === 0) {
			return;
		}

		// detect ie, or more general browsers that don't support smil
		if ($('html').hasClass(smilHtmlClass)) {
			this.handleIE();
			return;
		}

		this.animateCircle();
		this.animateText();

	}.bind(this));
};

Module.prototype.cacheElements = function() {
	this.$circle = this.$element.find(this.options.domSelectors.circle);
	this.$svg = this.$element.find(this.options.domSelectors.svg);
	this.circle = this.$circle[0];
	this.$text = this.$element.find(this.options.domSelectors.text);
};

Module.prototype.handleIE = function() {

	// IE has a wired bug, where the origin of the stroke-dasharray is
	// painted on the wrong place on the circle initially. By changing
	// stroke-width, which forces a repaint of the svg, the dasharray
	// will be painted correctly.
	this.$circle.css('stroke-width', '0px');
	setTimeout(function() {
		this.$circle.css('stroke-width', '');
	}.bind(this), 0);

	// since we drop the circle animation for IE, we don't need to animate
	// the text
	this.setTextToFinalValue();
	this.animateCircle();
};

// ---------------------------------------------------------------------------
// EVENTS
// ---------------------------------------------------------------------------
Module.prototype.addEventListeners = function() {
	$(document).on(mediaqueries.event.key, this.handleBreakpointChange.bind(this));
};

Module.prototype.handleBreakpointChange = function() {
	var currentBreakpoint = mediaqueries.getCurrentBreakpointName();

	if (currentBreakpoint === this.lastBreakpoint) {
		return;
	}

	// special case: user is resizing window while animation is still
	// running. So we skip animation
	this.finishedAnimation = true;

	this.setCircleProperties();
	this.lastBreakpoint = currentBreakpoint;
};

// ---------------------------------------------------------------------------
// Circle calculations
// ---------------------------------------------------------------------------
Module.prototype.getCircleDimensions = function() {
	return this.circle.getBoundingClientRect().width;
};

Module.prototype.setCircleProperties = function(setPropertiesCallback) {
	var dashOffset,
		dashOffsetFinished,
		transitionDuration = this.config.totalAnimationDuration / 1000,
		circleWidth = this.getCircleDimensions(),
		circleDiameter = circleWidth / 2;

	this.circumference = Math.PI * 2 * circleDiameter;
	dashOffsetFinished = (1 - this.config.progressPercentage) * this.circumference;
	dashOffset = this.finishedAnimation ? dashOffsetFinished : this.circumference;

	this.$circle.css('transition-duration', '0s');

	window.requestAnimationFrame(function() {
		this.$svg.attr('viewBox', '0 0 ' + circleWidth + ' ' + circleWidth);
		this.circle.setAttribute('stroke-dasharray', this.circumference);
		this.circle.setAttribute('stroke-dashoffset', dashOffset);

		if (!this.finishedAnimation) {
			window.requestAnimationFrame(function() {
				this.$circle.css('transition-duration', transitionDuration + 's');

				window.requestAnimationFrame(function() {
					if (setPropertiesCallback) {
						setPropertiesCallback();
					}
				});
			}.bind(this));
		} else {
			if (setPropertiesCallback) {
				setPropertiesCallback();
			}
		}

	}.bind(this));
};

// ---------------------------------------------------------------------------
// Circle and Text animation
// ---------------------------------------------------------------------------
Module.prototype.animateCircle = function() {
	var dashOffset = this.circumference - this.circumference * this.config.progressPercentage;

	this.animateCircleColor();

	this.$circle.one(cssTransitionEnd + '.' + this.uuid, function() {
		this.finishedAnimation = true;
	}.bind(this));

	setTimeout(function() {
		this.circle.setAttribute('stroke-dashoffset', dashOffset);
	}.bind(this), 1);
};

Module.prototype.animateCircleColor = function() {
	if (!this.config.color) {
		return;
	}

	this.$circle.css('stroke', this.config.color);
};

Module.prototype.animateText = function() {
	// we don't want to have too much dom-updates, so we limit updates to
	// 1 update per 70ms
	var updates = this.config.totalAnimationDuration / paintUpdateInterval,
		updateIncrement = this.config.progressPercentageHumanReadable / updates;

	this.updateText(paintUpdateInterval, updates, updateIncrement);
};

Module.prototype.updateText = function(updateInterval, updates, updateIncrement) {
	if (this.finishedAnimation) {
		this.setTextToFinalValue();
		return;
	}

	setTimeout(function() {
		if (this.animationIterations < updates) {
			this.$text.text(parseInt((this.animationIterations + 1) * updateIncrement) + '%');
			this.animationIterations++;
			this.updateText(updateInterval, updates, updateIncrement);

			// on second last iteration, we would like to set an accent
			// on the progress text
			if (this.animationIterations === parseInt(updates)) {
				this.makeFinalTextAnimation();
			}
		} else {
			// should not happen, but we need to make sure that the %-value
			// shown after the animation is finished is correct
			this.setTextToFinalValue();
		}
	}.bind(this), updateInterval);
};

Module.prototype.setTextToFinalValue = function() {
	this.$text.text(this.config.progressPercentageHumanReadable + '%');
};

Module.prototype.makeFinalTextAnimation = function() {
	var textTransition = this.$text.css('transition'),
		initialTextColor = this.$text.css('color');

	if (!this.config.color) {
		return;
	}

	this.$text.css({
		'transition': 'none'
	});

	window.requestAnimationFrame(function() {
		this.$text.css({
			'color': this.config.color
		});

		window.requestAnimationFrame(function() {
			this.$text.css({
				'transition': textTransition,
				'color': initialTextColor
			});
		}.bind(this));
	}.bind(this));
};

// ---------------------------------------------------------------------------
// A11Y
// ---------------------------------------------------------------------------
Module.prototype.setHiddenText = function() {
	var $hiddenText = this.$element.find(this.options.domSelectors.hiddenText),
		templateText = this.options.i18n.progressHiddenText,
		percentage = this.config.progressPercentageHumanReadable;

	templateText = templateText.replace('$1', percentage);
	$hiddenText.text(templateText);
};

// ---------------------------------------------------------------------------
// GENERAL HELPERS
// ---------------------------------------------------------------------------
Module.prototype.getProgressPercentage = function() {
	var steps = this.options.progressConfig.steps - 1,
		currentStep = this.options.progressConfig.currentStep;

	return currentStep / steps;
};

Module.prototype.getTotalAnimationDuration = function() {
	var transitionDurationForOneStep = this.options.progressConfig.transitionDurationForOneStep || defaultTransitionDuration,
		currentStep = this.options.progressConfig.currentStep;

	return transitionDurationForOneStep * currentStep;
};

Module.prototype.getStepColor = function() {
	var colors = this.options.progressConfig.colors,
		step = this.options.progressConfig.currentStep - 1;

	if (!colors) {
		return false;
	}

	if (step >= colors.length) {
		return false;
	}

	return colors[step];
};

// ---------------------------------------------------------------------------
// MODULE GENERIC
// ---------------------------------------------------------------------------
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 750 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 751 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(752);


/***/ }),
/* 752 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 753 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(754);


/***/ }),
/* 754 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 755 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(756);


/***/ }),
/* 756 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 757 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(758);


/***/ }),
/* 758 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M91 - Circle Progress
 *
 * Author Unic AG
 * Copyright Schweizerische Bundesbahnen SBB, 2020.
 */



__webpack_require__(759);

var $ = __webpack_require__(0),
	ajaxRequest = __webpack_require__(16),
	SuperClass = __webpack_require__(1);

var name = 'userprofileprogress',
	events = {},
	defaults = {
		domSelectors: {
			close: '[data-' + name + '="close"]'
		},
		stateClasses: {
			fadeOut: 'fade_out'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ---------------------------------------------------------------------------
// INIT
// ---------------------------------------------------------------------------
Module.prototype.init = function() {
	this.$element.one('click.' + this.uuid, this.options.domSelectors.close, this.handleClose.bind(this));
};

// ---------------------------------------------------------------------------
// EVENTS
// ---------------------------------------------------------------------------
Module.prototype.handleClose = function() {

	// inform api
	ajaxRequest(this.options.apiUrl, false, {finished: true});

	// hide element
	this.$element.addClass(this.options.stateClasses.fadeOut);
	this.$element.slideUp(300, function() {
		// set focus to next element
		this.focusNextElement();
	}.bind(this));

};

Module.prototype.focusNextElement = function() {
	var elementToFocus,
		$elementToFocus;

	if (!$('html').hasClass('focus-source-key')) {
		return;
	}

	elementToFocus = this.options.focusElementAfterClose;

	if (!elementToFocus) {
		return;
	}

	$elementToFocus = $(elementToFocus);

	if ($elementToFocus.length < 1) {
		return;
	}

	$elementToFocus[0].focus();
};

// ---------------------------------------------------------------------------
// MODULE GENERIC
// ---------------------------------------------------------------------------
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 759 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 760 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(761);


/***/ }),
/* 761 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 762 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(763);


/***/ }),
/* 763 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 764 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(765);


/***/ }),
/* 765 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 766 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(767);


/***/ }),
/* 767 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 768 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(769);


/***/ }),
/* 769 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M155 - Military Special Fields
 */



__webpack_require__(770);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	$document = $(document);

var name = 'militaryspecialfields',
	events = {
		blockPrefilled: 'multifieldblock.blockPrefilled'
	},
	defaults = {
		domSelectors: {
			travelReasonSelect: '[data-' + name + '=travel_reason]',
			firstClassInputs: '[data-' + name + '=firstclass]',
			secondClassInputs: '[data-' + name + '=secondclass]',
			firstClassTotalInput: '[data-' + name + '=firstclass_total]',
			secondClassTotalInput: '[data-' + name + '=secondclass_total]'
		},
		travelReasonAttributeKey: 'data-militaryspecialfields-reason',
		stateClasses: {}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ---------------------------------------------------------------------------
// INIT
// ---------------------------------------------------------------------------
Module.prototype.init = function() {
	this._cacheElements();
	this._addEventListeners();
};

// ---------------------------------------------------------------------------
// CACHE ELEMENTS
// ---------------------------------------------------------------------------
Module.prototype._cacheElements = function() {
	this.$travelReasonSelect = this.$element.find(this.options.domSelectors.travelReasonSelect);
	this.$firstClassInputs = this.$element.find(this.options.domSelectors.firstClassInputs);
	this.$secondClassInputs = this.$element.find(this.options.domSelectors.secondClassInputs);
	this.$firstClassTotalInput = this.$element.find(this.options.domSelectors.firstClassTotalInput);
	this.$secondClassTotalInput = this.$element.find(this.options.domSelectors.secondClassTotalInput);
};

// ---------------------------------------------------------------------------
// EVENT LISTENERS
// ---------------------------------------------------------------------------
Module.prototype._addEventListeners = function() {
	this.$element.on(
		'change.' + this.uuid,
		this.options.domSelectors.travelReasonSelect,
		this.handleSelectChange.bind(this)
	);

	this.$element.on(
		'input.' + this.uuid,
		this.options.domSelectors.firstClassInputs,
		function(event) {
			this._handleInputChange(event, true);
		}.bind(this)
	);

	this.$element.on(
		'input.' + this.uuid,
		this.options.domSelectors.secondClassInputs,
		function(event) {
			this._handleInputChange(event, false);
		}.bind(this)
	);

	$document.on(
		events.blockPrefilled + '.' + this.uuid,
		function() {
			this.handleSelectChange();
			this._calculateTotal(true);
			this._calculateTotal(false);
		}
	.bind(this));
};

Module.prototype.handleSelectChange = function() {
	var selectedOption = this.$travelReasonSelect.find(':selected'),
		travelReason = selectedOption.attr(this.options.travelReasonAttributeKey);

	this.changeTexts(this.options[travelReason]);
};

Module.prototype.changeTexts = function(travelReasonObj) {
	var $searchedElement,
		searchedElementId,
		$labelToChange;

	travelReasonObj.forEach(function(item) {
		$searchedElement = this.$element.find('[data-' + name + '=' + item.nameAttr + ']');
		searchedElementId = $searchedElement.attr('id');
		$labelToChange = this.$element.find('label[for=' + searchedElementId + ']');

		if ($labelToChange.length >= 1) {
			// change only the label text and not replace the hidden text
			$labelToChange[0].childNodes[0].nodeValue = item.text;
		} else {
			// change the text of the region description
			$searchedElement.text(item.text);
		}
	}.bind(this));
};

Module.prototype._handleInputChange = function(event, isFirstClass) {
	var $changedInput = $(event.currentTarget),
		changedInputValue = $changedInput.val();

	// calculate the total if the value is a digit between 0 and 1 or empty
	if (this._checkIfValueIsADigit(changedInputValue) || changedInputValue === '') {
		this._calculateTotal(isFirstClass);
	}
};

Module.prototype._calculateTotal = function(isFirstClass) {
	var $inputsToCalculate = isFirstClass ? this.$firstClassInputs : this.$secondClassInputs,
		$input,
		inputValue,
		totalFirstClass = 0,
		totalSecondClass = 0;

	$inputsToCalculate.each(function(index, input) {
		$input = $(input);
		inputValue = parseInt($input.val());

		if (this._checkIfValueIsADigit(inputValue)) {
			if (isFirstClass) {
				totalFirstClass += inputValue;
				this.$firstClassTotalInput.val(totalFirstClass);
			} else {
				totalSecondClass += inputValue;
				this.$secondClassTotalInput.val(totalSecondClass);
			}
		}
	}.bind(this));
};

Module.prototype._checkIfValueIsADigit = function(value) {
	return /^[0-9]+$/.test(value);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	initEvents: ['ready', 'ajaxload'],
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 770 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 771 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(772);


/***/ }),
/* 772 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 773 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(774);


/***/ }),
/* 774 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * M157 - Search Tag Filter
 */



__webpack_require__(775);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1),
	ajaxRequest = __webpack_require__(16),
	createUrlParams = __webpack_require__(79),
	parseUrl = __webpack_require__(69);

var name = 'searchtagfilter',
	events = {},
	defaults = {
		domSelectors: {
			tags: '[data-' + name + '="tag"]',
			wrapper: '[data-' + name + '="wrapper"]',
			showMore: '[data-' + name + '="show_more"]'
		},
		filterApiUrlAttributeKey: 'data-searchtagfilter-url',
		urlKeys: {
			searchTerm: 'search',
			category: 'category',
			page: 'page'
		},
		stateClasses: {}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

// ---------------------------------------------------------------------------
// INIT
// ---------------------------------------------------------------------------
Module.prototype.init = function() {
	this._cacheElements();
	this._addEventListeners();

	this.filterValues = {};
	this.pageNr = 1;

	this._prepareHistory();
	this._restoreFilterFromUrlParams();
};

// ---------------------------------------------------------------------------
// CACHE ELEMENTS
// ---------------------------------------------------------------------------
Module.prototype._cacheElements = function() {
	this.$tags = this.$element.find(this.options.domSelectors.tags);
	this.$wrapper = this.$element.find(this.options.domSelectors.wrapper);
};

// ---------------------------------------------------------------------------
// EVENT LISTENERS
// ---------------------------------------------------------------------------
Module.prototype._addEventListeners = function() {
	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.tags,
		this._handleTagClicked.bind(this)
	);

	this.$element.on(
		'click.' + this.uuid,
		this.options.domSelectors.showMore,
		this._handleShowMoreClicked.bind(this)
	);
};

// ---------------------------------------------------------------------------
// RESULTS
// ---------------------------------------------------------------------------
Module.prototype._requestAndAppendResults = function(pushHistory, tagClick, handleFocus) {
	var $selectedTag = this._getSelectedTag(),
		requestUrl = $selectedTag.attr(this.options.filterApiUrlAttributeKey);

	ajaxRequest(requestUrl, false, this.filterValues).then(function(results) {
		if (pushHistory) {
			this._handlePushState();
		}

		// if we click on the tab then replace the results
		if (tagClick) {
			this.$wrapper.html(results);
		} else {
			this.$wrapper.append(results);
		}

		if (handleFocus) {
			requestAnimationFrame(function() {
				this.$wrapper.find('li:first-child a').focus();
			}.bind(this));
		}

	}.bind(this)).catch(function() {
		return;
	}.bind(this));
};

Module.prototype._getSelectedTag = function() {
	return this.$element.find(this.options.domSelectors.tags + ':checked');
};

Module.prototype._handleShowMoreClicked = function(event) {
	this.filterValues[this.options.urlKeys.page] = this.pageNr += 1;

	this._requestAndAppendResults(true, false, true);

	// remove clicked button
	event.currentTarget.remove();
};

Module.prototype._handleTagClicked = function() {
	this.filterValues[this.options.urlKeys.page] = this.pageNr = 1;
	this.filterValues[this.options.urlKeys.category] = this._getSelectedTag().val();

	this._requestAndAppendResults(true, true, false);
};

Module.prototype._restoreFilterFromUrlParams = function() {
	var params = parseUrl(decodeURIComponent(location.search)),
		paramSearch = params[this.options.urlKeys.searchTerm],
		paramCategory = params[this.options.urlKeys.category],
		paramPage = params[this.options.urlKeys.page];

	if (paramSearch) {
		this.filterValues[this.options.urlKeys.searchTerm] = paramSearch;
	}

	if (paramCategory) {
		this.filterValues[this.options.urlKeys.category] = paramCategory;
	}

	if (paramCategory) {
		this.filterValues[this.options.urlKeys.page] = paramPage;
	}

	this._restoreTagFromHistory();
	this._restorePageFromHistory();
};

Module.prototype._restoreTagFromHistory = function() {
	var category = this.filterValues[this.options.urlKeys.category],
		$tag,
		checkValue;

	if (!category) {
		return;
	}

	this.$tags.each(function(index, tag) {
		$tag = $(tag);
		checkValue = $tag.val() === category ? true : false;

		$tag.prop('checked', checkValue);
	});
};

Module.prototype._restorePageFromHistory = function() {
	this.pageNr = parseInt(this.filterValues[this.options.urlKeys.page], 10);
};

// -------------------------------------------------------------
// HISTORY
// -------------------------------------------------------------

Module.prototype._prepareHistory = function() {
	// set scrollRestoration to manual
	if ('scrollRestoration' in history) {
		history.scrollRestoration = 'manual';
	}

	// object to save values for history state
	this.filterValues[this.options.urlKeys.searchTerm] = this.options.searchTerm;
	this.filterValues[this.options.urlKeys.category] = $(this.$tags[0]).val(); // category of the first tag
	this.filterValues[this.options.urlKeys.page] = this.pageNr;

	// set initial history state
	history.replaceState(this.filterValues, null, null);

	// Add event listener for history's popstate to handle browser history navigation
	window.addEventListener('popstate', this._handlePopState.bind(this), false);
};

Module.prototype._handlePushState = function() {

	// update filter values for history
	this.filterValues[this.options.urlKeys.category] = this._getSelectedTag().val();
	this.filterValues[this.options.urlKeys.page] = this.pageNr;

	history.pushState(this.filterValues, null, createUrlParams(this.filterValues));
};

Module.prototype._handlePopState = function(event) {
	if (!event.state) {
		return;
	}

	// update filterValues to keep in sync with state object
	this.filterValues = event.state;

	this._restoreTagFromHistory();
	this._restorePageFromHistory();
	this._requestAndAppendResults(false, true, true);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 775 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 776 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(777);


/***/ }),
/* 777 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * online_consultation_cancelation_form
 *
 * Hendrik Wernze (One inside)
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



__webpack_require__(778);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'onlineconsultationcancelationform',
	events = {},
	defaults = {
		domSelectors: {
			notification: '[data-' + name + '="notification"]',
			notificationMessage: '[data-notification="message"]',
			form: '[data-' + name + '="form"]',
			submitButton: '[data-' + name + '="submit"]',
			successContent: '[data-' + name + '="successContent"]',
			successContentTitle: '[data-' + name + '="successPanelHeading"]'
		},
		stateClasses: {
			isHidden: 'is_hidden'
		}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});

}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 */
Module.prototype.init = function() {
	this._cacheElements();
	this._setupEventHandlers();
};

/**
 * Initial dom caching
 * @method
 */
Module.prototype._cacheElements = function() {
	this.$notification = this.$element.find(this.options.domSelectors.notification);
	this.$notificationMessage = this.$notification.find(this.options.domSelectors.notificationMessage);
	this.$form = this.$element.find(this.options.domSelectors.form);
	this.$successContent = this.$element.find(this.options.domSelectors.successContent);
	this.$successContentTitle = this.$element.find(this.options.domSelectors.successContentTitle);
};

/**
 * Bind events
 * @method
 */
Module.prototype._setupEventHandlers = function() {
	this.$element.on(
		'submit.' + this.uuid,
		this.options.domSelectors.form,
		this._handleSubmit.bind(this)
	);
};

/**
 * Handle the submit on the form. Toogle form and success content.
 * @method
 */
Module.prototype._handleSubmit = function(evt) {
	evt.preventDefault();

	$.ajax({
		cache: false,
		url: this.options.apiUrl,
		data: this.$form.serialize(),
		method: 'POST',
		contentType: false,
		processData: false
	}).then(function(data) {
		if (data.success) {
			this.$form.addClass(this.options.stateClasses.isHidden);
			this.$successContent.removeClass(this.options.stateClasses.isHidden);
			this.$successContentTitle.focus();
		} else {
			this._showErrorMessage(data.message);
		}
	}.bind(this))
		.catch(function() {
			this._showErrorMessage(this.options.i18n.generalErrorMessage);
		}.bind(this));
};

/**
 * Show the notification module with the error text.
 * @method
 * @param {string} errorText - Text that will be displayed in the notification module
 */
Module.prototype._showErrorMessage = function(errorText) {
	this.$notification.removeClass(this.options.stateClasses.isHidden);
	this.$notificationMessage.text(errorText);
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 778 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 779 */
/***/ (function(module, exports, __webpack_require__) {

__webpack_require__(780);


/***/ }),
/* 780 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 781 */
/***/ (function(module, exports, __webpack_require__) {

module.exports = __webpack_require__(782);


/***/ }),
/* 782 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";
/**
 * jobPublisher_transsicura
 *
 * Hendrik Wernze (One inside)
 * Copyright Schweizerische Bundesbahnen SBB, 2017.
 */



// Ensure that the SASS file is built.
__webpack_require__(783);

var $ = __webpack_require__(0),
	SuperClass = __webpack_require__(1);

var name = 'jobPublisherTranssicura',
	events = {},
	defaults = {
		domSelectors: {
			container: '[data-' + name + '="container"]'
		},
		stateClasses: {}
	},
	data = {};

/**
 * Create an instance of the module
 * @constructor
 * @param {object} element - The DOM element to bind the module
 * @param {object} options - Options overwriting the defaults
 */
function Module(element, options) {
	this._helper = SuperClass;

	this._helper({
		name: name,
		element: element,
		defaults: defaults,
		options: options,
		events: events,
		data: data
	});
}

Module.prototype = $.extend(true, {}, SuperClass.prototype, Module.prototype);

/**
 * Initialize module, bind events.
 * @method
 * @public
 */
Module.prototype.init = function() {
	this.selectorString = '#ostendisJobs button.action';
	this.observerConfig = {
		childList: true,
		subtree: true,
		attributes: false
	};

	this.cacheElements();
	this.bindEventListeners();
	this.initObserver();

};

/**
 * Caching all relevent elements
 */
Module.prototype.cacheElements = function() {
	this.$container = this.$element.find(this.options.domSelectors.container);
	if (this.$container.length) {
		this.containerElement = this.$container.get(0);
	}
};

/**
 * Bind events
 */
Module.prototype.bindEventListeners = function() {
	document.addEventListener('ostendisLoaderReady', this.initOstendisJobs.bind(this));
};

/**
 * Initialize Mutation Observer
 */
Module.prototype.initObserver = function() {
	this.observer = new MutationObserver(this.onObserverMutation.bind(this));
	this.observer.observe(this.containerElement, this.observerConfig);
};

/**
 * Mutation Observer callback
 *
 * @param mutations
 * @param observer
 */
Module.prototype.onObserverMutation = function(mutations, observer) {
	for (var i = 0; i < mutations.length; i++) {
		for (var j = 0; j < mutations[i].addedNodes.length; j++) {
			var childNode = mutations[i].addedNodes[j];

			if (childNode && childNode.matches && childNode.matches(this.selectorString)) {
				// add custom style
				mutations[i].addedNodes[j].classList.add('text__primarybutton_arrowless');
			}
		}
	}

	observer.disconnect();
};

/**
 * Callback for 'ostendisLoaderReady' to launch OstendisJobs
 */
Module.prototype.initOstendisJobs = function() {
	var OSTENDISJOBS = window.OSTENDISJOBS;

	OSTENDISJOBS.embed(

		// publication place hash
		'd67ff9549aeb415e90c1ea7fdb4b9a10',

		// langauge
		'DE',

		// id of the container to render the job list
		'#ostendisJobs'

		// options. if set, these options will overwrite
		// the options defined in the ostendis admin
		// panel
		// {
		// 	filter: false,
		// 	search: false,
		// 	count:  false,
		// 	sort:   false
		// }
	);
};
/**
 * Unbind events, remove data, custom teardown
 * @method
 * @public
 */
Module.prototype.destroy = function() {
	// Unbind events, remove data
	SuperClass.prototype.destroy.apply(this);

	// Custom teardown (removing added DOM elements etc.)
	// If there is no need for a custom teardown, this method can be removed
};

// Make the plugin available through jQuery (and the global project namespace)
SuperClass.register(Module, name, {
	events: events
});

module.exports = {
	Module: Module,
	events: events
};


/***/ }),
/* 783 */
/***/ (function(module, exports) {

// removed by extract-text-webpack-plugin

/***/ }),
/* 784 */
/***/ (function(module, exports, __webpack_require__) {

"use strict";


var $ = __webpack_require__(0);

window.drunk = function() {

	var $doc = $(document),
		elements = [
			'button:not(.mod_footer_column_link)',
			'a:not(.mod_homepage_main_teaser_item_link):not(.mod_teaserelement_url):not(.mod_lh_teaser_link):not(.mod_quicklinksnav_list_item_link)',
			'input:not(.mod_lh_tag_checkbox)',
			'.mod_deeplinkteaser_list_item',
			'.mod_lh_teaser_homepage_teasers_list_item',
			'.mod_togglebutton',
			'.mod_lh_tag',
			'.mod_footer_column',
			'.mod_clock_wrapper',
			'.mod_quicklinksnav_list'
		],
		$links = $(elements.join(',')),
		isTouchDevice = Modernizr.touchevents || navigator.maxTouchPoints;

	if (isTouchDevice) {
		return;
	}

	function bindDrunkElement(evt) {
		var element = evt.currentTarget,
			$element = $(element),
			elementRect = element.getBoundingClientRect(),
			elementX = elementRect.x,
			elementY = elementRect.y,
			elementWidth = elementRect.width,
			elementHeight = elementRect.height,
			random = Math.floor(Math.random() * 6),
			maxX = elementX + elementWidth,
			maxY = elementY + elementHeight,
			mouseX,
			mouseY,
			deltaX,
			deltaY,
			transfrom,
			transition = 'none';

		$doc.on('mousemove', function(evt2) {
			mouseX = evt2.clientX;
			mouseY = evt2.clientY;
			deltaX = mouseX - elementX + 10;
			deltaY = mouseY - elementY + 10;

			if (mouseX < maxX && mouseY < maxY && mouseX > elementX && mouseY > elementY) {
				if (random === 0) {
					transfrom = 'translate(0px, ' + deltaY + 'px)';
				} else if (random === 1) {
					transfrom = 'translate(' + deltaX + 'px, 0px)';
				} else if (random === 2) {
					transfrom = 'translate(' + -deltaX + 'px, ' + deltaY + 'px)';
				} else if (random === 3) {
					transfrom = 'translate(' + deltaX + 'px, ' + -deltaY + 'px)';
				} else if (random === 4) {
					transfrom = 'translate(' + deltaX + 'px, ' + deltaY + 'px)';
				} else if (random === 5) {
					transfrom = 'rotate3d(1, 0, 0, 90deg)';
					transition = 'transform 100ms ease-in-out';
				}

				$element.css({
					'transform': transfrom,
					'transition': transition
				});
			} else {

				$element.css({
					'transform': 'none',
					'transition': 'none'
				});

				$doc.off('mousemove');

				$links.on('mousemove', function(evt) {
					$links.off('mousemove');
					bindDrunkElement(evt);
				});
			}
		});
	}

	$links.on('mousemove', function(evt) {
		$links.off('mousemove');
		bindDrunkElement(evt);
	});
};


/***/ })
],[723]);